/*
This software module was originally developed by Apple Computer, Inc.
in the course of development of MPEG-4. 
This software module is an implementation of a part of one or 
more MPEG-4 tools as specified by MPEG-4. 
ISO/IEC gives users of MPEG-4 free license to this
software module or modifications thereof for use in hardware 
or software products claiming conformance to MPEG-4.
Those intending to use this software module in hardware or software
products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company,
the subsequent editors and their companies, and ISO/IEC have no
liability for use of this software module or modifications thereof
in an implementation.
Copyright is not released for non MPEG-4 conforming
products. Apple Computer, Inc. retains full right to use the code for its own
purpose, assign or donate the code to a third party and to
inhibit third parties from using the code for non 
MPEG-4 conforming products.
This copyright notice must be included in all copies or
derivative works. Copyright (c) 1999.
*/
#include <stdio.h>
#include <stdlib.h>
#include "MP4Movies.h"
#include "adif.h"
#include "all.h"
#include "port.h"

#define MAX_DSI_LEN			1024	    /* bytes */

#define SHORT_ADTS_HEADER_SIZE		7           /* bytes */
#define LONG_ADTS_HEADER_SIZE		9           /* bytes */
#define INPUT_ADTS_HEADER_SIZE		(adts_header.protection_abs?SHORT_ADTS_HEADER_SIZE:LONG_ADTS_HEADER_SIZE)
#define OUTPUT_ADTS_HEADER_SIZE		(output_adts_CRC?LONG_ADTS_HEADER_SIZE:SHORT_ADTS_HEADER_SIZE)

#define MAX_RDB_PER_FRAME		4
#define START_OF_OUTPUT_FRAME		(num_rdb_this_frame <= 1)
#define END_OF_OUTPUT_FRAME		(num_rdb_this_frame >= output_rdb_per_frame)

#define ROUND_UP(x)			(int)(x+.999999999999)

enum{OUT_MP4FF, OUT_ADIF, OUT_RAW, OUT_ADTS};

struct {
  char *str;
  int fmt;
} fmtstr[] = {
  {"mp4",OUT_MP4FF},
  {"adif",OUT_ADIF},
  {"adts",OUT_ADTS},
  {"raw",OUT_RAW},
  {NULL,-1}
};

MC_Info mc_info_cmdline;

char *inFile, *outFile;
int output_format = -1;
int prog;
int decoderBufferSize = 0;  /* bit */

MP4Movie output_moov;
MP4Media output_media;
MP4Track output_trak;
MP4Handle sampleDataH;
MP4Handle sampleEntryH;
MP4Handle sampleDurationH;
MP4Handle sampleSizeH;
FILE* output_fd;


/* output bitBuffer simulation */
static int bitrates_unknown = 0;
static long input_maxBitrate = 0, output_maxBitrate =-1;
static long input_avgBitrate = 0, output_avgBitrate =-1, input_avgBitrate_cmdline = -1;
static int input_start_bufferFullness = 0, output_start_bitBuffer_len =-1;
static int output_simulated_bitBuffer_len;
static double output_max_bitBuffer_len, output_min_bitBuffer_len;
static long output_total_byte_count = 0;

/* variables for multiple raw data blocks per frame */
static int output_rdb_per_frame = 1;
static int num_rdb_this_frame = 0;
static long frame_byte_count = 0;
static int bufferFullness;

/* other control variables */
static int output_adts_CRC = 0;
static int add_fill = 0;
int del_fill = 0;

void usage(char* s)
{
  PRINT(SE, "ADIF2MP4\n");
  PRINT(SE, "usage:  %s  [options] input_file output_file\n", s);
  PRINT(SE, "	options are:\n");
  PRINT(SE, "	-i     program_number (ADIF header present, 0 is first program)\n");
  PRINT(SE, "	-s     sampling_frequency, Hz (only necessary for RAW input)\n");
  PRINT(SE, "	-r     max. bitrate for output, bit/s (optional)\n");
  PRINT(SE, "	-b     average bitrate for output, bit/s (optional)\n");
  PRINT(SE, "	-bi    average bitrate for input, bit/s (optional)\n");
  PRINT(SE, "	-f     initial buffer fullness (adif_buffer_fullness) (optional)\n");
  PRINT(SE, "	-F [ad] allow changing bitstream by [a]dding/[d]eleting FILL-elements\n");
  PRINT(SE, "	-p     profile  (Main or LC or SSR or LTP, Main is default)\n");
  PRINT(SE, "	-o     output bitstream format: mp4ff (default), adif, adts, raw\n");
  PRINT(SE, "	-n     number of raw data blocks per output frame (default:1)\n");
  PRINT(SE, "	-c     enable CRC in ADTS headers (default:disable)\n");
  PRINT(SE, "	-D [a-Z]    (enable debugging printouts)\n");
  exit(-1);
}

void parse_args(int argc, char **argv)
     /*
      * parse commandline and set the specified variables
      */
{
  char *cmd=argv[0],*s;
  int i, j;
  int *p;

  ARGBEGIN {
  default:
    PRINT(SE, "unknown option\n");
    usage(cmd);

  case 'D':
    for (s = ARGF(); *s; s++) {
      debug[(int)(*s)] = 1;
      PRINT(SE, "debug option %c enabled\n", *s);
    }
    break;

  case 'i':
    adif_header_present = 1;
    prog = atoi(ARGF());
    if(prog > 15) {
      PRINT(SE, "program number out of range: %d\n", prog);
      exit(-1);
    }
    break;

  case 'p':
    s = ARGF();
    p = &mc_info_cmdline.profile;
    if ((!strncmp(s, "Main", 4)) || (!strncmp(s, "main", 4))) {
      *p = Main_Profile; 
      break;
    }
    if ((!strncmp(s, "LC", 2)) || (!strncmp(s, "lc", 2))) {
      *p = LC_Profile; 
      break;
    }
    if ((!strncmp(s, "SSR", 3)) || (!strncmp(s, "ssr", 3))) {
      *p = SSR_Profile;
      break;
    }
    if ((!strncmp(s, "LTP", 3)) || (!strncmp(s, "ltp", 3))) {
      *p = LTP_Profile;
      break;
    }
    PRINT(SE, "Unsupported profile %s\n", s);
    exit(-1);
    break;

  case 's':
    j = atoi(ARGF());
    for (i=0; i<(1<<LEN_SAMP_IDX); i++) {
      if (j == samp_rate_info[i].samp_rate)
        break;
    }
    if (i == (1<<LEN_SAMP_IDX)) {
      PRINT(SE, "Unsupported sampling frequency %d\n", j);
      exit(-1);
    }
    mc_info_cmdline.sampling_rate_idx = i;
    break;


  case 'F':
    for (s = ARGF(); *s; s++) {
      switch (*s) {
      case 'a': add_fill = 1;break;
      case 'd': del_fill = 1;break;
      default:
        PRINT(SE, "Unsupported FILL-element option %c: choose a for add or d for delete or both\n",*s);
        exit(-1);
      }
    }
    break;


  case 'b':
    s = ARGF();
    if (!strcmp(s,"i")) input_avgBitrate_cmdline = atoi(ARGF());
    else output_avgBitrate = atoi(s);
    break;
  case 'r':
    output_maxBitrate = atoi(ARGF());
    break;
  case 'f':
    output_start_bitBuffer_len = atoi(ARGF());
    break;
  case 'n':
    output_rdb_per_frame = atoi(ARGF());
    if (output_rdb_per_frame>MAX_RDB_PER_FRAME) {
      PRINT(SE, "maxiumum of %i raw_data_blocks per frame\n",MAX_RDB_PER_FRAME);
      exit(-1);
    }
    break;
  case 'c':
    output_adts_CRC = 1;
    break;

  case 'o':
    s = ARGF();
    output_format = j = -1;
    while ( fmtstr[++j].str ) {
      if (!strcmp(fmtstr[j].str,s)) {
        output_format = fmtstr[j].fmt;
        break;
      }
    }
    if (output_format==-1) usage(cmd);
    break;
    
  }
  ARGEND

  if(argc != 2) {
    usage(cmd);
  }

  inFile = argv[0];
  outFile = argv[1];

  if (output_format==-1) {
    for (i=strlen(outFile);(i>=(signed int)(strlen(outFile)-5))&&(i>=0);i--) {
      if (outFile[i]=='.') {
        j=-1;
        while ( fmtstr[++j].str ) {
          if (!strcmp(fmtstr[j].str,(outFile+i+1))) {
            output_format = fmtstr[j].fmt;
            break;
          }
        }
      }
    }
  }
  if (output_format==-1) {
    PRINT(SE, "output will be mp4ff\n");
    output_format = OUT_MP4FF;
  }

  if (output_format!=OUT_ADTS) output_rdb_per_frame = 1;
  if ((output_rdb_per_frame!=1)&&(output_adts_CRC)) {
    PRINT(SE, "CRC not supported for multiple raw data blocks per frame... CRC disabled!\n");
    output_adts_CRC = 0;
  }
}


MP4Err write_adif_header(MC_Info mip, int bitrate, int fullness)
     /*
      * write ADIF header to file
      */
{
  char buf[1024];
  ProgConfig my_prog_config;
  ADIF_Header my_adif_header;
  long length;
  int i;

  my_adif_header.adif_id[0] = 'A';
  my_adif_header.adif_id[1] = 'D';
  my_adif_header.adif_id[2] = 'I';
  my_adif_header.adif_id[3] = 'F';
  my_prog_config.comments[0] = 0;

  my_adif_header.copy_id_present = 0;
  my_adif_header.original_copy = 0;
  my_adif_header.home = 0;
  my_adif_header.bitstream_type = 0;
  my_adif_header.bitrate = bitrate;
  my_adif_header.buffer_fullness = fullness;

  my_prog_config.profile = mip.profile;
  my_prog_config.sampling_rate_idx = mip.sampling_rate_idx;

  for (i = my_prog_config.front.num_ele = 0; i<FChans; i++) {
    if (mip.ch_info[i].present && ((!mip.ch_info[i].cpe)^mip.ch_info[i].ch_is_left)) {
      my_prog_config.front.ele_is_cpe[my_prog_config.front.num_ele] = mip.ch_info[i].cpe;
      my_prog_config.front.ele_tag[my_prog_config.front.num_ele] = mip.ch_info[i].tag;
      my_prog_config.front.num_ele++;
    }
  }
  for (my_prog_config.side.num_ele = 0; i<FChans+SChans; i++) {
    if (mip.ch_info[i].present && ((!mip.ch_info[i].cpe)^mip.ch_info[i].ch_is_left)) {
      my_prog_config.side.ele_is_cpe[my_prog_config.side.num_ele] = mip.ch_info[i].cpe;
      my_prog_config.side.ele_tag[my_prog_config.side.num_ele] = mip.ch_info[i].tag;
      my_prog_config.side.num_ele++;
    }
  }
  for (my_prog_config.back.num_ele = 0; i<FChans+SChans+BChans; i++) {
    if (mip.ch_info[i].present && ((!mip.ch_info[i].cpe)^mip.ch_info[i].ch_is_left)) {
      my_prog_config.back.ele_is_cpe[my_prog_config.back.num_ele] = mip.ch_info[i].cpe;
      my_prog_config.back.ele_tag[my_prog_config.back.num_ele] = mip.ch_info[i].tag;
      my_prog_config.back.num_ele++;
    }
  }
  for (my_prog_config.lfe.num_ele = 0; i<FChans+SChans+BChans+LChans; i++) {
    if (mip.ch_info[i].present && ((!mip.ch_info[i].cpe)^mip.ch_info[i].ch_is_left)) {
      my_prog_config.lfe.ele_is_cpe[my_prog_config.lfe.num_ele] = mip.ch_info[i].cpe;
      my_prog_config.lfe.ele_tag[my_prog_config.lfe.num_ele] = mip.ch_info[i].tag;
      my_prog_config.lfe.num_ele++;
    }
  }
  my_prog_config.coupling.num_ele = 0;
  for (i = 0; i<mip.ncch; i++) {
    my_prog_config.coupling.num_ele++;
    my_prog_config.coupling.ele_is_cpe[i] = mip.cc_ind[i];
    my_prog_config.coupling.ele_tag[i] = mip.cc_tag[i];
  }
  my_prog_config.data.num_ele = 0;

  length = put_adif_header( my_adif_header, my_prog_config, buf );

  if (fwrite( buf, 1, length, output_fd) != (size_t)length) return MP4IOErr;
  return MP4NoErr;
}


int write_adts_header(MC_Info mip, int fullness, int frameLen, int num_RDB)
     /*
      * write one ADTS-Header to file
      */
{
  ADTS_Header my_adts_header;
  char buf[16];
  int length;
  int i;

  my_adts_header.syncword = 0xfff;
  my_adts_header.id = 1;
  my_adts_header.layer = 0;
  my_adts_header.protection_abs = (output_adts_CRC?0:1);
  my_adts_header.profile = mip.profile;
  my_adts_header.sampling_freq_idx = mip.sampling_rate_idx;
  my_adts_header.private_bit = 0;
  my_adts_header.original_copy = 0;
  my_adts_header.home = 0;
  my_adts_header.copyright_id_bit = 0;
  my_adts_header.copyright_id_start = 0;
  my_adts_header.frame_length = frameLen;
  my_adts_header.adts_buffer_fullness = fullness;
  my_adts_header.num_of_rdb = num_RDB -1;

  if (mip.nch<=6) my_adts_header.channel_config = mip.nch;else
  if (mip.nch==8) my_adts_header.channel_config = 7;else {
    PRINT(SE, "number of channels\n");
    return 0;
  }

  my_adts_header.crc_check = 0;
  length = put_adts_header( my_adts_header, buf );

  if (output_adts_CRC) {
    init_output_CRC(1);
    for (i=0;i<(CRC_LEN_HDR>>3);i++)
      feed_CRC_calculator(8,(u8)buf[i]);
    my_adts_header.crc_check = calculate_output_CRC();

    length = put_adts_header( my_adts_header, buf );
  }

  if (fwrite(buf, 1, length, output_fd) != (size_t)length) return 0;
  return length;
}


MP4Err writeout_bitstream()
     /*
      * write current frame to disk
      */
{
  MP4Err err;

  if (output_format==OUT_MP4FF) {
    err = MP4AddMediaSamples( output_media, sampleDataH, 1, sampleDurationH, sampleSizeH,
                              sampleEntryH, NULL, NULL ); if (err) return err;
    if ( sampleEntryH ) {
      err = MP4DisposeHandle( sampleEntryH ); if (err) return err;
      sampleEntryH = NULL;
    }
  } else if ((output_format==OUT_RAW)||(output_format==OUT_ADIF)) {
    if ((output_format==OUT_ADIF)&&(!bno)) { /* start of ADIF file */
      err = write_adif_header( mc_info, output_avgBitrate, output_start_bitBuffer_len );
      if (err) return err;
    }
    if (fwrite((char*)*sampleDataH, 1, *(u32*) *sampleSizeH, output_fd) != *(size_t*) *sampleSizeH) return MP4IOErr;
  }
  else if (output_format==OUT_ADTS) {
    static int first_header = 1;

    if (!write_adts_header( mc_info, (bufferFullness/channel_num)>>5,
                            frame_byte_count+OUTPUT_ADTS_HEADER_SIZE, num_rdb_this_frame )) return MP4IOErr;
    if (fwrite((char*)*sampleDataH, 1, *(u32*) *sampleSizeH, output_fd) != *(size_t*) *sampleSizeH) return MP4IOErr;
    if (!first_header) {
      output_total_byte_count += OUTPUT_ADTS_HEADER_SIZE;
    }
    first_header = 0;
  }

  return MP4NoErr;
}


MP4Err addMySamples( MP4Handle decoderSpecificInfoH )
     /*
      * parse input while writing output
      */
{
  MP4Err err = MP4NoErr;
  u32 objectTypeIndication;
  u32 streamType;
  u32 i, block_length;
  u8 *data_block;
  int input_bitBuffer_sim_ready = 0;
  double internal_simulated_bitBuffer_len = 0, output_avgFrameLen = 0;

  err = getDecoderBufferSize( (u32*)&decoderBufferSize ); if (err) goto bail;
  if (output_format == OUT_MP4FF) {
    err = MP4SetMediaLanguage( output_media, "und" ); if (err) goto bail;
    objectTypeIndication = 0x40; /* mpeg-4 audio */
    streamType = 0x05; /* audio stream */
    err = MP4NewHandle( 0, &sampleEntryH ); if (err) goto bail;
    err = MP4NewSampleDescription( output_trak, 
                                   sampleEntryH,
                                   1,
                                   objectTypeIndication,
                                   streamType,
                                   (u32)(decoderBufferSize>>3), /* bytes, not bits */
                                   output_maxBitrate,
                                   output_avgBitrate,
                                   decoderSpecificInfoH ); if (err) goto bail;

    err = MP4NewHandle( sizeof(u32), &sampleDurationH ); if (err) goto bail;
    *((u32*)*sampleDurationH) = aacFrameLength;
  }
  data_block=(u8*)*sampleDataH;
  err = MP4NewHandle( sizeof(u32), &sampleSizeH ); if (err) goto bail;

  for(bno = 0;; bno++) {

    /* --- get input */
    err = getRawDataBlock( data_block, &block_length );

    if (END_OF_OUTPUT_FRAME) {
      num_rdb_this_frame = 0;
      frame_byte_count = 0;
    }
    num_rdb_this_frame++;

    if ( err ) {
      if ( err == MP4EOF ) err = MP4NoErr;
      break;
    }

    /* --- input bitBuffer simulation */
    if (!input_bitBuffer_sim_ready) {
      if ((bno) && (adts_header_present)) input_start_bufferFullness=(adts_header.adts_buffer_fullness<<5)*channel_num;
      input_bitBuffer_sim_ready = init_bitBuffer_sim(BITRATE2FRAMELEN(input_avgBitrate),
                                                     input_start_bufferFullness);
    } else do_bitBuffer_sim();
    check_bitBuffer_sim();

    /* --- output bitBuffer simulation */
    if (!bno) {
      internal_simulated_bitBuffer_len = output_min_bitBuffer_len = output_max_bitBuffer_len =
        output_simulated_bitBuffer_len = output_start_bitBuffer_len;
      output_avgFrameLen = BITRATE2FRAMELEN(output_avgBitrate);
      output_total_byte_count = 0;
    } else {
      internal_simulated_bitBuffer_len+= output_avgFrameLen;
      internal_simulated_bitBuffer_len-= block_length<<3;

      if ((output_format==OUT_ADTS) && START_OF_OUTPUT_FRAME)
        internal_simulated_bitBuffer_len-= OUTPUT_ADTS_HEADER_SIZE<<3;

      /*if ((output_format==OUT_ADTS)&&(adts_header_present))
        internal_simulated_bitBuffer_len = simulated_bitBuffer_len;*/
      /* keep input bufferFullness => output bitrate implicitly changes */

      if ((add_fill)&&(internal_simulated_bitBuffer_len>decoderBufferSize - BITRATE2FRAMELEN(output_avgBitrate))) {
        int tmp = ROUND_UP(internal_simulated_bitBuffer_len-decoderBufferSize+BITRATE2FRAMELEN(output_avgBitrate));
        tmp = (tmp>>3)+(tmp&0x7?1:0);
        if (debug['F'])
          PRINT(SE,"Adding FILL elements - moving end by %i byte, old block-length %i\n",tmp,block_length);
        tmp = move_end(tmp);
        block_length += tmp;
        internal_simulated_bitBuffer_len -= tmp<<3;
        if (debug['F']) PRINT(SE,"-- OK, moved by %i bytes, new block length %i\n", tmp, block_length);
      }

      output_simulated_bitBuffer_len = (int)(internal_simulated_bitBuffer_len);

      output_total_byte_count+= block_length;
    }
    data_block += block_length;
    frame_byte_count += block_length;
    err = MP4SetHandleSize( sampleDataH, frame_byte_count ); if (err) goto bail;
    *(u32*)*sampleSizeH = frame_byte_count;


    if (internal_simulated_bitBuffer_len < output_min_bitBuffer_len)
      output_min_bitBuffer_len = internal_simulated_bitBuffer_len;
    if (internal_simulated_bitBuffer_len > output_max_bitBuffer_len)
      output_max_bitBuffer_len = internal_simulated_bitBuffer_len;

    if (START_OF_OUTPUT_FRAME)
      bufferFullness = output_simulated_bitBuffer_len;
    /* bufferFullnes after "encoding" first raw-data-block in frame */


    /* --- write output */
    if (END_OF_OUTPUT_FRAME) {
      err = writeout_bitstream(); if (err) goto bail;
      data_block=(u8*)*sampleDataH;
      reset_output_CRC();
    }
  }
bail:
  return err;
}


MP4Err closeFile()
     /*
      * close the output file
      */
{
  MP4Err err = MP4NoErr;
  u64 mediaDuration;
  long output_calculatedBitrate;
  long output_proposedBitrate;
  long tmp = 0;
  double tmp1;
  extern int simulated_bitBuffer_len;

  if (!START_OF_OUTPUT_FRAME)
    writeout_bitstream();

  if (!bitrates_unknown)
    tmp1 = ((double)(output_total_byte_count<<3)-output_start_bitBuffer_len+output_simulated_bitBuffer_len)/(bno-1);
  else
    tmp1 = (double)(output_total_byte_count<<3)/(bno-1);
  output_calculatedBitrate = (long int) FRAMELEN2BITRATE(tmp1);

  if (!bitrates_unknown)
    tmp1 = ((double)(output_total_byte_count<<3)-input_start_bufferFullness+simulated_bitBuffer_len)/(bno-1);
  else
    tmp1 = (double)(output_total_byte_count<<3)/(bno-1);
  output_proposedBitrate = (long int) FRAMELEN2BITRATE(tmp1);

  /* close output file ---------------------------------------------- */

  if (output_format==OUT_MP4FF) {
    err = MP4GetMediaDuration( output_media, &mediaDuration ); if (err) goto bail;
    err = MP4EndMediaEdits( output_media ); if (err) goto bail;
    err = MP4InsertMediaIntoTrack( output_trak, 0, 0, mediaDuration, 1 ); if (err) goto bail;
    err = MP4SetTrackEnabled( output_trak, 1 ); if (err) goto bail;
    err = MP4WriteMovieToFile( output_moov, outFile ); if (err) goto bail;
    err = MP4DisposeMovie( output_moov ); if (err) goto bail;
  } else if ((output_format==OUT_RAW)||(output_format==OUT_ADIF)||(output_format==OUT_ADTS)) {
    fclose(output_fd);
    err = MP4NoErr;
  }

  /* report errors -------------------------------------------------- */

  PRINT(SE,
        "INPUT:  used bitrate was: %10ld bit/s (avgFrameLen %8.2f bit)  change with '-bi'\n"
        "        accurate        : %10i bit/s (avgFrameLen %8.2f bit)\n"
        "OUTPUT: used bitrate was: %10ld bit/s (avgFrameLen %8.2f bit)  change with '-b'\n"
        "        accurate        : %10ld bit/s (avgFrameLen %8.2f bit)\n"
        "        proposing       : %10ld bit/s (avgFrameLen %8.2f bit)\n",
        input_avgBitrate, BITRATE2FRAMELEN(input_avgBitrate),
        calculatedBitrate, BITRATE2FRAMELEN(calculatedBitrate),
        output_avgBitrate, BITRATE2FRAMELEN(output_avgBitrate),
        output_calculatedBitrate, BITRATE2FRAMELEN(output_calculatedBitrate),
        output_proposedBitrate, BITRATE2FRAMELEN(output_proposedBitrate));

  PRINT( SE, "problems:\n");
  if ((output_format!=OUT_MP4FF)&&(output_format!=OUT_RAW)) {
    if (output_max_bitBuffer_len > decoderBufferSize - BITRATE2FRAMELEN(output_avgBitrate)) {
      tmp = 1;
      PRINT(SE, "bitBuffer-simulation: violated upper bitBuffer-border... (%8.2f>%i)\n",
            output_max_bitBuffer_len, decoderBufferSize - (int)(BITRATE2FRAMELEN(output_avgBitrate)));
    }
    if (output_min_bitBuffer_len < 0) {
      tmp = 1;
      PRINT(SE, "bitBuffer-simulation: violated lower bitBuffer-border... (%8.2f<0)\n",
            output_min_bitBuffer_len);
    }
  }
  if (tmp)
    PRINT(SE,
          "       You should either specify a different average bitrate with '-b'\n"
          "       or give an initial bufferFullness ('-f') of\n"
          "            at least %6i, and\n"
          "            at most  %6i (currently %i)\n",
          ROUND_UP(output_start_bitBuffer_len-output_min_bitBuffer_len),
          (int)((decoderBufferSize-BITRATE2FRAMELEN(output_avgBitrate))-output_max_bitBuffer_len+output_start_bitBuffer_len),
          output_start_bitBuffer_len);

  if ((output_max_bitBuffer_len-output_min_bitBuffer_len) > decoderBufferSize-BITRATE2FRAMELEN(output_avgBitrate)) {
    tmp += 2;
    PRINT(SE,
          "ERROR in bitBuffer-simulation: buffer_min_size: %i, buffer_max_size %i\n"
          "                            => buffer-limits exceeded!\n"
          "       It may help to give a different bitrate with '-b' (currently %ld bit/s)\n",
          (int)output_min_bitBuffer_len, (int)output_max_bitBuffer_len, output_avgBitrate);
  }

  if (!tmp) PRINT(SE, "no problems found\n");

bail:
  return err;
}


MP4Err createMyMovie( int prog, char *inFile, char *outFile )
     /*
      * get necessary information about input and initiate the output
      */
{
  MP4Err err = MP4NoErr;
  MP4Handle decoderSpecificInfoH;
  u8 OD_profileAndLevel;
  u8 scene_profileAndLevel;
  u8 audio_profileAndLevel;
  u8 visual_profileAndLevel;
  u8 graphics_profileAndLevel;
  u32 initialObjectDescriptorID;
  u32 timeScale;

  /* read input file ------------------------------------------------ */

  err = MP4NewHandle( MAX_FRAME_LEN*MAX_RDB_PER_FRAME, &sampleDataH ); if (err) goto bail;
  err = MP4NewHandle( MAX_DSI_LEN, &decoderSpecificInfoH ); if (err) goto bail;
  err = openMP4InputFile( inFile );
  if (err == MP4NoErr) {
    mpeg4_header_present = 1;
    adif_header_present = 0;
    adts_header_present = 0;
    err = getMP4info( decoderSpecificInfoH, &timeScale, &audio_profileAndLevel,
                      (u32*)&input_maxBitrate, (u32*)&input_avgBitrate, (u32*)&input_start_bufferFullness );
    if (err) goto bail;
    if (debug['n']) PRINT(SE, "MPEG4 header present\n");
  } else {
    mpeg4_header_present = 0;

    err = openInputFile( inFile ); if (err) goto bail;
    if ( adif_id_present() == 1 ) {
      adif_header_present = 1;
      adts_header_present = 0;
      err = getADIFinfo( prog, decoderSpecificInfoH, &timeScale, &audio_profileAndLevel,
                         (u32*)&input_maxBitrate, (u32*)&input_avgBitrate, (u32*)&input_start_bufferFullness );
      if (err) goto bail;
      if (debug['n']) PRINT(SE, "ADIF header present\n");
    } else {
      adif_header_present = 0;
      adts_header_present = 1;
      err = getRAWinfo( decoderSpecificInfoH, &timeScale, &audio_profileAndLevel,
                        (u32*)&input_maxBitrate, (u32*)&input_avgBitrate, (u32*)&input_start_bufferFullness );

      if (err) goto bail;
      if ((debug['n'])&&(adts_header_present)) PRINT(SE, "ADTS syncword found\n");
    }
  }

  if (input_avgBitrate_cmdline!=-1) input_avgBitrate = input_avgBitrate_cmdline;
  if (output_start_bitBuffer_len==-1) output_start_bitBuffer_len = input_start_bufferFullness;
  if (output_avgBitrate==-1) {
    if (input_avgBitrate) {
      double tmp = BITRATE2FRAMELEN(input_avgBitrate)
        -(adts_header_present?INPUT_ADTS_HEADER_SIZE<<3:0)
        +(output_format==OUT_ADTS?OUTPUT_ADTS_HEADER_SIZE<<3:0);
      output_avgBitrate = (long int)(FRAMELEN2BITRATE(tmp)+.5);
    } else {
      PRINT(SE, "bitrates unknown... expect bitBuffer problems!\n");
      bitrates_unknown = 1;
      output_avgBitrate = 0;
    }
  }
  if (input_maxBitrate==0) input_maxBitrate = input_avgBitrate;
  if (output_maxBitrate==-1) output_maxBitrate = input_maxBitrate;

  if (adts_header_present&&(input_avgBitrate_cmdline==-1))
    input_avgBitrate = 0; /* to calculate new (possibly more accurate) bitrate each frame */

  /* open output file ----------------------------------------------- */

  if (output_format==OUT_MP4FF) {

    initialObjectDescriptorID = 1;
    OD_profileAndLevel = 0xff; /* none required */
    scene_profileAndLevel = 0xff; /* none required */
    visual_profileAndLevel = 0xff; /* none required */
    graphics_profileAndLevel = 0xff; /* none required */
    err = MP4NewMovie( &output_moov,
                       initialObjectDescriptorID,
                       OD_profileAndLevel,
                       scene_profileAndLevel,
                       audio_profileAndLevel,
                       visual_profileAndLevel,
                       graphics_profileAndLevel ); if (err) goto bail;
    err = MP4NewMovieTrack( output_moov, MP4NewTrackIsAudio, &output_trak ); if (err) goto bail;
    err = MP4NewTrackMedia( output_trak, &output_media, MP4AudioHandlerType, timeScale, NULL ); if (err) goto bail;
    err = MP4BeginMediaEdits( output_media ); if (err) goto bail;

    err = addMySamples( decoderSpecificInfoH ); if (err) goto bail;

    err = closeFile();

  } else if ((output_format==OUT_RAW)||(output_format==OUT_ADIF)||(output_format==OUT_ADTS)) {
    output_fd = fopen(outFile, "w");
    if(output_fd == NULL) {
      PRINT(SE, "cant create %s\n", outFile);
      exit(-1);
    }

    err = addMySamples( NULL ); if (err) goto bail;

    err = closeFile();
  }
bail:
  MP4DisposeHandle(decoderSpecificInfoH);
  return err;
}


int main( int argc, char **argv )
{
  MP4Err err;
  /* set defaults */
  prog = -1;
  default_config = 1;

  parse_args(argc,argv);
  err = createMyMovie( prog, inFile, outFile );

  if (err) PRINT( SE, "createMyMovie returns %d\n", err );

  return err;
}
