/************************************************************/
/*  13818-7 RMS Verification Software                       */
/*                                                          */
/*  Authors:   CL, et. al.   Chuck Lueck <lueck@ti.com>     */
/*                                                          */
/*  History:                                                */
/*  30-Sep-97  Initial revision.                            */
/************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define N 16     /* sample width in bytes (multiple of 8) */
#define BITS_IN_BYTE 8
#define BYTES_PER_WORD N/BITS_IN_BYTE  /* number of bytes per word (integer) */

#if 0
static char id[] = "$Revision: 1.1 $";
#endif
static char programName[] = "rmsdiff";

void
usage() {
	fprintf(
		stderr,
		"\nusage: %s [-b] [-i n1 [-i n2]] [-v]"
		" [-e] [-s] [-c] [-r] file_1 file_2\n"
"Find the maximum absolute difference between signed %d-bit words in file_1 and\n"
"file_2. Options are:\n\n"
"        -b          byte swap before comparing\n"
"        -i n        ignore n words at the start of both file_1 and file_2\n"
"        -i n1 -i n2 ignore n1 words at the start of file 1, and n2 words\n"
"                    at the start of file2\n"
"        -v          lists non-zero differences\n"
"        -e          lists every difference\n"
"        -s          print signed difference\n"
"        -r          print RMS error level\n"
"        -c          count samples different and samples compared\n",
		programName,N
	);
	exit(1);
}

int
main(int argc, char *argv[]) {
	FILE *	inputFile1;
	FILE *	inputFile2;

	long	maxDifference;
	long    minDifference;
	long	sample1;
	long	sample2;
	unsigned char    byte_sample1;
	unsigned char    byte_sample2;
	double  err_energy=0.0;
	double  rms_error;

	long	ignore1 = -1;
	long	ignore2 = -1;

	int i;
	int done=0;
	int	verbose   = 0;
	int every     = 0;
	int sign      = 0;
	int count     = 0;
	int num_comp  = 0;
	int num_diff  = 0;
	int byte_swap = 0;
	int rms       = 0;

	while (argc > 1 && argv[1][0] == '-') {
		long	arg;

		switch (argv[1][1]) {
		case 'i':
			/*
			 * number of words to ignore at the start
			 * of each file
			 */
			if (argv[1][2] != '\0')
				arg = atol(&argv[1][2]);
			else if (argc > 2) {
				arg = atol(argv[2]);
				argv++;
				argc--;
			}
			else
				usage();

			if (ignore1 < 0)
				ignore1 = arg; /* first */
			else
				ignore2 = arg; /* second */

			break;

		case 'e':
			every = 1;

		case 'v':
			verbose = 1;
			break;

		case 's':
			sign = 1;
			break;

		case 'c':
			count = 1;
			break;

		case 'b':
			byte_swap = 1;
			break;

		case 'r':
			rms = 1;
			break;


		default:
			usage();
		}

		argv++;
		argc--;
	}

	if (argc < 3)
		usage();

	/*
	 * set up the number of words to ignore from the defaults
	 */
	if (ignore1 < 0)
		ignore1 = 0;
	if (ignore2 < 0)
		ignore2 = ignore1;

	/*
	 * open the input files
	 */
	inputFile1 = fopen(argv[1], "rb");
	if (inputFile1 == NULL) {
		fprintf(
			stderr,
			"%s: cannot open input file %s\n",
			programName,
			argv[1]
		);
		exit(1);
	}

	/*
	 * skip over words at the beginning of inputFile1
	 */
	while (
		ignore1-- > 0 &&
		fread(&byte_sample1, BYTES_PER_WORD*sizeof(byte_sample1), 1, inputFile1) == 1
	) ;

	inputFile2 = fopen(argv[2], "rb");
	if (inputFile2 == NULL) {
		fprintf(
			stderr,
			"%s: cannot open input file %s\n",
			programName,
			argv[2]
		);
		exit(1);
	}

	/*
	 * skip over words at the beginning of inputFile2
	 */
	while (
		ignore2-- > 0 &&
		fread(&byte_sample2, BYTES_PER_WORD*sizeof(byte_sample2), 1, inputFile2) == 1
	) ;
	
	/*
	 * read a sample for each channel and make sure all are available
	 *	before writing.
	 */
	maxDifference = 0;
	minDifference = 0;
	while (!done) {
	  sample1=0;
	  sample2=0;
	  for (i=0;i<BYTES_PER_WORD;i++) {
	    done = done ||
	      !(fread(&byte_sample1, sizeof(byte_sample1), 1, inputFile1) == 1 &&
		fread(&byte_sample2, sizeof(byte_sample2), 1, inputFile2) == 1);
	    sample1 = (sample1<<BITS_IN_BYTE) + byte_sample1;
	    sample2 = (sample2<<BITS_IN_BYTE) + byte_sample2;
	  }
	  if (!done) {     
		   
		long difference;
		
		if (byte_swap) {
		  unsigned long u_s1 = (unsigned long) sample1;
		  unsigned long u_s2 = (unsigned long) sample2;
		  
		  /* swap bytes */
		  sample1 = (long) ( ((u_s1 & 0xff000000) >> 24) |
				     ((u_s1 & 0x00ff0000) >> 8)  |
				     ((u_s1 & 0x0000ff00) << 8)  |
				     ((u_s1 & 0x000000ff) << 24) );
		  sample2 = (long) ( ((u_s2 & 0xff000000) >> 24) |
				     ((u_s2 & 0x00ff0000) >> 8)  |
				     ((u_s2 & 0x0000ff00) << 8)  |
				     ((u_s2 & 0x000000ff) << 24) );

		  /* shift right, sign extend */
		  sample1 = (sample1>>(32-N));
		  sample2 = (sample2>>(32-N));

		} else {

		  /* Sign extend data words */
		  sample1 = ((sample1<<(32-N))>>(32-N));
		  sample2 = ((sample2<<(32-N))>>(32-N));
		}

		difference = (sign ? (sample1 - sample2)
			          : abs(sample1 - sample2));
		
		err_energy += (double)difference*(double)difference;

		num_comp++;

		if (difference != 0) {
			num_diff++;
		}

		if (difference > maxDifference)
				maxDifference = difference;

		if (difference < minDifference)
				minDifference = difference;

		if (verbose) {
			if (every || (difference != 0))
			printf(
				"Sample %6d : %6ld %6ld %6ld\n",
				num_comp,
				sample1,
				sample2,
				difference
			);
		}
	  }
	}

	/*
	 * clean up
	 */
	fclose(inputFile1);
	fclose(inputFile2);

	/* Don't print */
	/*
	if (sign) {
		printf("Max : %6d, Min : %6d\n", maxDifference, minDifference);		
		maxDifference = (maxDifference>abs(minDifference)) ? 
		  maxDifference : abs(minDifference);
	} else {
		printf("Max:%d\n", maxDifference);
		} */
	
	if (count) {
		printf("Samples compared : %6d, Samples different : %6d\n",
			   num_comp, num_diff);
	}

	rms_error = sqrt(err_energy/num_comp);
	if (rms) {
		printf("RMS error: %f\n",rms_error);
	}

	/* Check for maximum difference and rms error requirements */
       	if ( (maxDifference>pow(2.0,N-15))  || 
	     (rms_error>(pow(2.0,N-16)/sqrt(12.0))) ) {
	  printf("RMS Verification Test Failed.\n");
	  printf("   Max difference: %ld   Allowed max difference:%ld\n",
		 maxDifference,(long)pow(2.0,N-15)); 
	  printf("   RMS level: %f   Allowed RMS level:%f\n",rms_error,pow(2.0,N-16)/sqrt(12.0));
	} else {
	  /*printf("RMS Verification Test Passed.\n");*/
	  /* Don't print anything if it passes. */
	}

	return(0);
}



