/**********************************************************************
 MPEG-4 Audio VM

 This software module was originally developed by
 
 Bernhard Grill     (Fraunhofer Gesellschaft IIS)
  
 and edited by
 Markus Werner       (SEED / Software Development Karlsruhe) 
 
 in the course of development of the MPEG-2 NBC/MPEG-4 Audio standard
 ISO/IEC 13818-7, 14496-1,2 and 3. This software module is an
 implementation of a part of one or more MPEG-2 NBC/MPEG-4 Audio tools
 as specified by the MPEG-2 NBC/MPEG-4 Audio standard. ISO/IEC gives
 users of the MPEG-2 NBC/MPEG-4 Audio standards free license to this
 software module or modifications thereof for use in hardware or
 software products claiming conformance to the MPEG-2 NBC/ MPEG-4 Audio
 standards. Those intending to use this software module in hardware or
 software products are advised that this use may infringe existing
 patents. The original developer of this software module and his/her
 company, the subsequent editors and their companies, and ISO/IEC have
 no liability for use of this software module or modifications thereof
 in an implementation. Copyright is not released for non MPEG-2
 NBC/MPEG-4 Audio conforming products. The original developer retains
 full right to use the code for his/her own purpose, assign or donate
 the code to a third party and to inhibit third party from using the
 code for non MPEG-2 NBC/MPEG-4 Audio conforming products. This
 copyright notice must be included in all copies or derivative works.
 
 Copyright (c) 2000.
 
 $Id: dec_tf.c,v 1.23.2.14 2001/06/11 14:44:34 cvs03 Exp $
 TF Decoder module
**********************************************************************/

/*******************************************************************************************
 *
 * Master module for T/F based codecs
 *
 ******************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <memory.h>
#include <string.h>

#include "allHandles.h"

#include "bitstream.h"
#include "common_m4a.h"
#include "dec_tf.h"
#include "mod_buf.h"
#include "tf_main.h"

/* Long term predictor */
#include "nok_lt_prediction.h"

/* ---  AAC --- */
#include "aac.h"
#include "all.h"

#include "reorderspec.h"
#include "resilience.h"
#include "concealment.h"
#include "buffers.h"
#include "statistics_aac.h"

#include "mp4_tf.h"
#include "sam_dec.h"

extern int samplFreqIndex [] ;

/*****************************************************************************************
 ***
 *** Function: DecTfInfo
 ***
 *** Purpose:  Provide Info about T/F-based decoder core
 ***
 ****************************************************************************************/

#define PROGVER "general audio decoder core FDAM1 16-Feb-2001"

char *DecTfInfo (FILE *helpStream)
{
  if (helpStream != NULL) 
  {
    fprintf(helpStream,
      PROGVER "\n"
      "decoder parameter string format:\n"
      "possible options:\n"
      "     -aac_raw\n"
      "     -main\n"
      "     -lsf         length_of_scalefactor_data within bitstream\n"
      "     -rvlc        error resilient scalefactor coding\n"
      "     -scfCon      enables scalefactor concealment ( default is muting )\n"
      "     -scfBit      additional bit allowing enhanced concealment within bitstream\n" 
      "     -lcw         length_of_longest_codeword within bitstream\n"
      "     -hcr <mode>  HCR within spectral data, length_of_longest_codeword and\n"
      "                                       length_of_spectral_data within bitstream\n"
      "                  modi: 0 - raw\n"
      "                        1 - codeword pre-sorting (standard)\n"
      "                        2 - consecutive HCR\n"
      "                        3 - codeword pre-sorting and consecutive HCR (standard)\n"
      "     -vcb11       virtual codebooks are used to encode codebook 11 within section data\n");

    ConcealmentHelp(helpStream);
    
    fprintf(helpStream,
      PROGVER "\n"
      "     -aac_sca\n"
      "     -aac_sys_bsac\n"
      "     -bsacDecBr <n> (scalable, 16 <= n <= 64)\n"
      "     -aac_ld : AAC Low Delay mode\n"
      "     -aac_ld_512 : block size 512 (default: 480)\n"
      "     -nttDecLyr <n> (scalable, n=0: base layer)\n"
      "\n");
  }
  
  return PROGVER;
}

/*****************************************************************************************
 ***
 *** Function: DecTfInit
 ***
 *** Purpose:  Initialize the T/F-part and the macro blocks of the T/F part of the VM
 ***
 ****************************************************************************************/

void DecTfInit (char*               decPara,      /* in: decoder parameter string */
                char*               aacDebugStr,
                HANDLE_DECODER_GENERAL hFault,
                FRAME_DATA*         frameData,
                TF_DATA*            tfData,
                LPC_DATA*           lpcData,
                int                 layer,
                char*               infoFileName
                )
{
  AUDIO_SPECIFIC_CONFIG *streamConfig, *layerConfig ;

  int streamChannels = 0 ;  /* number of audio channels in topmost layer */
  int layerChannels = 0 ;      /* number of audio channels in this layer */
  int profile = LC_Profile ;
  int i_ch;


  /*******************************************************************************************/
  /* General Initialization, Evaluate T/F-specific Commandline Switches                      */

  DebugPrintf(2, "  GA scaleable decoding init \n");

  tfData->aacDecoder = NULL ;
 
  tfData->output_select = frameData->scalOutSelect ;
  
  tfData->prev_windowShape [0] = WS_FHG ;
  tfData->prev_windowShape [1] = WS_FHG ;

  tfData->tnsBevCore=0;
  if( strstr( decPara, "-tnsBevCor" ) ) 
  {
    DebugPrintf(1, "\n DON'T calc tns on core spectrum");

    tfData->tnsBevCore=1;
  }

  tfData->mdctCoreOnly=0;
  if( strstr( decPara, "-core_mdct"))
  {
    DebugPrintf (0, "\nIgnoring aac layers, output is core mdct only\n");
    tfData->mdctCoreOnly=1;
  }

  frameData->layer[layer].bitRate = frameData->od->ESDescriptor[layer]->DecConfigDescr.avgBitrate.value;

  streamConfig = &frameData->od->ESDescriptor[frameData->od->streamCount.value-1]->DecConfigDescr.audioSpecificConfig ;
  layerConfig = &frameData->od->ESDescriptor[layer]->DecConfigDescr.audioSpecificConfig;

  tfData->block_size_samples = streamConfig->specConf.TFSpecificConfig.frameLength.value ? 960 : 1024 ;
  
  switch(streamConfig->audioDecoderType.value){
    case AAC_MAIN:
      tfData->pred_type = MONOPRED;
    default:
      tfData->pred_type = NOK_LTP;
    break;
  }

  
  streamChannels = streamConfig->channelConfiguration.value ;
  layerChannels = layerConfig->channelConfiguration.value ;

  if ((streamChannels != 1) && (streamChannels !=2))
    CommonExit(1,"unsupported channel config") ;

  if ((layerChannels != 1) && (layerChannels !=2))
    CommonExit(1,"unsupported channel config") ;

  if (layerConfig->samplingFreqencyIndex.value < 12)
  {
    frameData->layer[layer].sampleRate = samplFreqIndex [layerConfig->samplingFreqencyIndex.value] ;
  }
  else
  {
    CommonExit(1,"unsupported sampleRateIndex") ;
  }


  /*******************************************************************************************/
  /* Initialize Error Resilience Tools                                                       */

  hFault->hVm = CreateBuffer(VIRTUELL);
  hFault->hHcrSpecData = CreateBuffer(AAC_MAX_INPUT_BUF_BITS);
  hFault->hHcrInfo = CreateHcrInfo();
  hFault->hResilience = CreateErrorResilience (decPara, 
                                                   0,
                                                   layerConfig->specConf.TFSpecificConfig.aacSectionDataResilienceFlag.value,
                                                   layerConfig->specConf.TFSpecificConfig.aacScalefactorDataResilienceFlag.value,
                                                   layerConfig->specConf.TFSpecificConfig.aacSpectralDataResilienceFlag.value);

  if( GetEPFlag ( hFault->hResilience ) ) 
  {
    hFault->hEscInstanceData = CreateEscInstanceData ( infoFileName, 0 );
    InitAacSpecificInstanceData ( hFault->hEscInstanceData, hFault->hResilience, layerChannels, strstr(decPara, "tnsNotUsed") ? 1 : 0 );
  }


  ConcealmentInit ( &hFault->hConcealment, decPara,0);
  StatisticsAacInit ( hFault->hEscInstanceData, hFault->hResilience );


  /*******************************************************************************************/
  /* Initialize TwinVQ Base Layer                                                            */

  if ((layerConfig->audioDecoderType.value == TWIN_VQ) ||
      (layerConfig->audioDecoderType.value == ER_TWIN_VQ))
  {
    tvqInitDecoder (decPara, layer, frameData, tfData) ;

    tfData->pred_type = NOK_LTP ;
  }


  /*******************************************************************************************/
  /* Initialize Scalable AAC Layer(s)                                                        */

  switch (streamConfig->audioDecoderType.value)
  {
   
    case AAC_MAIN :
    case AAC_LTP :

      profile = Main_Profile ;

    case AAC_SCAL :
    case AAC_LC :
    case ER_AAC_LC:
    case ER_AAC_LTP:
    case ER_AAC_SCAL:
    case ER_AAC_LD:

      tfData->aacDecoder = AACDecodeInit (frameData->layer[layer].sampleRate,
                                          aacDebugStr,
                                          tfData->block_size_samples,
                                          &tfData->sfbInfo,
                                          tfData->pred_type,
                                          profile) ;
      break ;
  }


  /*******************************************************************************************/
  /* Initialize BSAC Tool                                                                    */

  if (streamConfig->audioDecoderType.value == ER_BSAC)
  {
    char *p_ctmp ;

    tfData->sam_DecBr = 0;

    if( (p_ctmp=strstr( decPara, "-bsacDecBr " )) ) {
      if(sscanf(p_ctmp+11, "%d", &tfData->sam_DecBr) == 0){
        CommonExit(1, "Decode: parameter of -bsacDecBr switch not found");
      }
    }

    DebugPrintf (1, ">> BSAC Decoding BitRate: %d kbits/s/ch\n", tfData->sam_DecBr);

    sam_decode_init(frameData->layer[layer].sampleRate, tfData->block_size_samples);       /* HP 971105 */
  }


  /*******************************************************************************************/
  /* Allocate Frame Scratch Memory, Initialize LTP                                           */

  for (i_ch = 0 ; i_ch < streamChannels ; i_ch++)
  {
    int x ;

    tfData->time_sample_vector[i_ch] = (double*)calloc(tfData->block_size_samples, sizeof(double));
    tfData->overlap_buffer[i_ch] = (double*)calloc(tfData->block_size_samples, sizeof(double));
    tfData->sampleBuf[i_ch] = (float*)malloc(tfData->block_size_samples*sizeof(float));

    for (x = 0 ; x < MAX_TF_LAYER ; x++)
    {
      tfData->spectral_line_vector[x][i_ch] = (double*)calloc(tfData->block_size_samples, sizeof(double));
    }

    if (tfData->pred_type == NOK_LTP)
    {
      tfData->nok_lt_status[i_ch] = (NOK_LT_PRED_STATUS *)malloc(sizeof(*tfData->nok_lt_status[0])) ;
      nok_init_lt_pred(tfData->nok_lt_status[i_ch]) ;
      tfData->nok_lt_status[i_ch]->overlap_buffer = (double*)calloc(tfData->block_size_samples, sizeof(double));

    }
  }


  /*******************************************************************************************/
  /* Allocate CELP Core Sample Delay Modulo Buffer                                           */

  if (frameData->od->ESDescriptor[0]->DecConfigDescr.audioSpecificConfig.audioDecoderType.value==CELP)
  {
    int firstAacLayer = frameData->od->ESDescriptor[0]->DecConfigDescr.audioSpecificConfig.specConf.celpSpecificConfig.numEnhLayers.value + 1;

    int coreDelay= frameData->od->ESDescriptor[firstAacLayer]->DecConfigDescr\
      .audioSpecificConfig.specConf.TFSpecificConfig.coreCoderDelay.value;
    int samplIdxCore= frameData->od->ESDescriptor[0]\
      ->DecConfigDescr.audioSpecificConfig.samplingFreqencyIndex.value;
    int samplIdxTF= frameData->od->ESDescriptor[firstAacLayer]\
      ->DecConfigDescr.audioSpecificConfig.samplingFreqencyIndex.value;
    
    tfData->samplFreqFacCore= samplFreqIndex[samplIdxTF]/samplFreqIndex[samplIdxCore];
    tfData->tfSamplRate=samplFreqIndex[samplIdxTF];
    
    tfData->mdct_overlap_buffer = (double*)malloc(1024*sizeof(double));
    
    tfData->coreModuloBuffer = CreateFloatModuloBuffer(coreDelay 
      + (lpcData->frameNumSample
      * tfData->samplFreqFacCore * 12));

    ZeroFloatModuloBuffer (tfData->coreModuloBuffer, coreDelay) ;
  }
}

/*****************************************************************************************
 ***
 *** Function:    DecTfFree
 ***
 *** Description: free memory allocated by T/F decoders
 ***
 ***
 *****************************************************************************************/

void DecTfFree (TF_DATA  *tfData, 
                HANDLE_DECODER_GENERAL hFault)
{
  if ( hFault->hEscInstanceData && hFault->hResilience && GetReorderSpecFlag (hFault->hResilience ) ) {
    StatisticsAacPrint ( );
  }

  if (tfData)
  {
    unsigned int channel, layer ;

 
    for (channel = 0 ; channel < MAX_TIME_CHANNELS ; channel++)
    {
      if (tfData->time_sample_vector [channel]) free (tfData->time_sample_vector [channel]) ;
      if (tfData->overlap_buffer [channel]) free (tfData->overlap_buffer [channel]) ;
      if (tfData->nok_lt_status[channel]) free (tfData->nok_lt_status[channel]) ;
      if (tfData->sampleBuf[channel]) free (tfData->sampleBuf[channel]) ;

      for (layer=0;layer<MAX_TF_LAYER;layer++)
      {
        if (tfData->spectral_line_vector[layer][channel])
          free (tfData->spectral_line_vector[layer][channel]) ;
      }
    }

    if (tfData->coreModuloBuffer) DeleteFloatModuloBuffer (tfData->coreModuloBuffer) ;
    if (tfData->mdct_overlap_buffer) free (tfData->mdct_overlap_buffer) ;
  }

  if (hFault)
  {
    if (hFault->hVm) 
      DeleteBuffer (hFault->hVm);
    if (hFault->hHcrSpecData) 
      DeleteBuffer (hFault->hHcrSpecData) ;
    if (hFault->hHcrInfo) 
      DeleteHcrInfo (hFault->hHcrInfo) ;
    if (hFault->hResilience)
      DeleteErrorResilience(hFault->hResilience); 

  }

  if (tfData->aacDecoder) 
    AACDecodeFree (tfData->aacDecoder) ;
  if(tfData->tvqDecoder)
    tvqFreeDecoder(tfData->tvqDecoder);


}

/*****************************************************************************************
 ***
 *** Function:    DecTfFrame
 ***
 *** Description: processes a block of time signal input samples into a bitstream
 ***              based on T/F encoding 
 ***
 ***
 *****************************************************************************************/

void DecTfFrame (int          numChannels, 
                 FRAME_DATA*  fd, /* config data ,one bitstream buffer for each layer, obj descr. etc. */
                 TF_DATA*     tfData,
                 LPC_DATA*    lpcData,
                 HANDLE_DECODER_GENERAL hFault)
{
  int i_ch;
  
  
  
  
  ResetReadBitCnt ( hFault->hVm );
  
  switch (fd->od->ESDescriptor [fd->od->streamCount.value-1]->DecConfigDescr.audioSpecificConfig.audioDecoderType.value)
  {
    case ER_BSAC :

      bsacAUDecode (numChannels,
                    fd, /* static configuration data, bitstream buffer, obj descr. etc. */
                    tfData, /* actual frame input/output data */
                    tfData->sam_DecBr);

      break ;

    case AAC_SCAL :
    case TWIN_VQ : 
    case ER_TWIN_VQ : 
  
      tfScaleableAUDecode (numChannels,
                           fd, /* static configuration data, bitstream buffer, obj descr. etc. */
                           tfData, /* actual frame input/output data */
                           lpcData,
                           hFault , /* additional data for error protection */
                           AAC_SCALABLE);

    break ;
    case ER_AAC_SCAL:
      CommonExit(1,"Objectype not supported in this release");
    break;

    case AAC_LC :
    case AAC_MAIN :
    case AAC_LTP :
      aacAUDecode (numChannels, fd, tfData, hFault, MULTICHANNEL) ;

    break;
    
    case ER_AAC_LC:
    case ER_AAC_LTP:
    case ER_AAC_LD:
      aacAUDecode (numChannels, fd, tfData, hFault,MULTICHANNEL_ER) ;
    break ;
  }

  for (i_ch=0; i_ch<numChannels; i_ch++)
  {
    int i;
    freq2buffer (tfData->spectral_line_vector[0][i_ch],
                 tfData->time_sample_vector[i_ch],
                 tfData->overlap_buffer[i_ch],
                 tfData->windowSequence[i_ch],
                 tfData->block_size_samples, /* changed to variables (NI) */
                 tfData->block_size_samples/2,/* not used */
                 tfData->block_size_samples/8,
                 tfData->windowShape[i_ch],
                 tfData->prev_windowShape[i_ch],   
                 OVERLAPPED_MODE,
                 8);

    tfData->prev_windowShape[i_ch] = tfData->windowShape[i_ch];
    
    if(fd->od->ESDescriptor[fd->od->streamCount.value-1]->DecConfigDescr.audioSpecificConfig.audioDecoderType.value == AAC_LTP)
      nok_lt_update(tfData->nok_lt_status[i_ch], tfData->time_sample_vector[i_ch],
		    tfData->overlap_buffer[i_ch], tfData->block_size_samples);


#ifdef DEBUGPLOT
    if (i_ch==0){
      plotSend("L", "timeOut",  MTV_DOUBLE,tfData->block_size_samples,  tfData->time_sample_vector[i_ch] , NULL);
    } else {
      plotSend("R", "timeOutR",  MTV_DOUBLE,tfData->block_size_samples,  tfData->time_sample_vector[i_ch] , NULL);
    }
#endif
     /* convert time data from double to float */
         
    for( i=0; i<(int)tfData->block_size_samples; i++ ) 
      tfData->sampleBuf[i_ch][i] = (float)tfData->time_sample_vector[i_ch][i];
   
  }


}
