///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Copyright (c) 1997 ISO/IEC All Rights Reserved                        //
//                                                                       //
// This software module was developed by                                 //
//                                                                       //
// CSELT (Centro Studi e Laboratori Telecomunicazioni S.p.A)             //
//                                                                       //
// in the course of development of the MPEG-4 standard. This software    //
// module is an implementation of a part of one or more MPEG-4 tools as  //
// specified by the MPEG-4 standard.                                     //
//                                                                       //
// The copyright of this software belongs to ISO/IEC. ISO/IEC gives use  //
// of the MPEG-4 standard free license to use this  software module or   //
// modifications thereof for hardware or software products claiming      //
// conformance to the MPEG-4 standard.                                   //
//                                                                       //
// Those intending to use this software module in hardware or software   //
// products are advised that use may infringe existing  patents. The     //
// original developers of this software module and their companies, the  //
// subsequent editors and their companies, and ISO/IEC have no liability //
// and ISO/IEC have no liability for use of this software module or      //
// modification thereof in an implementation.                            //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//
// File:  TouchSensor2DProxy.cpp
//
// Author: Matthew Leditschke, Gianluca Di Cagno  gianluca.dicagno@cselt.it
//
// Description:
//   The implementation of the proxy object used by the presenter
//   to handle the TouchSensor2D node.
//
// Created: October, 1997.
//
// Modified:
//
///////////////////////////////////////////////////////////////////////////


#include "defs.h"
#include "Nodes-CSELT.h"
#include "TouchSensor2DProxy.h"
#include "trace.h"


// Storage for this static data member
BOOL TouchSensor2DProxy::m_movementGrabbed = FALSE;


// =====================================================================

//
// Presenter::InitTouchSensor2D
//
// Initialise the proxy data structure and attach it to the render data
// pointer of the given media object node.
//

void Presenter::InitTouchSensor2D(TouchSensor2D* pObj)
{
	pObj->SetPresenterData(new TouchSensor2DProxy(pObj));
}




// =====================================================================

//
// TouchSensor2DProxy::OnUserEvent
//
// This function is called when a user event has occurred above a node
// which is in the same grouping node hierarchy as this sensor node.
//
// If this node is registered with the presenter to received events,
// it will be called a second time for this scene draw.
//
// If this touch sensor is enabled, then generate the required events
// and return TRUE, otherwise return FALSE.
//

BOOL TouchSensor2DProxy::OnUserEvent(EventInfo const* pEvent)
{
	TRACE_PRINT("In TouchSensor2DProxy::OnUserEvent():\n");

	// Only generate events if this node is enabled, and no other
	// touch sensor node has grabbed all motion events
	if (((BOOL)(Bool)GetTouchSensor2D()->enabled) &&
		((!IsMovementGrabbed()) ||
		((IsMovementGrabbed()) && (m_mouseDown))))
	{
		// If this is called during pass 2, and the previous pass was
		// also pass 2, then the mouse has moved off the geometry
		// object.
		if ((pEvent->pass == 2) && (m_prevPass == 2))
		{
			// Generate an isOver=FALSE event if this is a mouse move
			if (pEvent->m_event == EVENT_MOUSE_MOVE)
			{
				if ((BOOL)(Bool)GetTouchSensor2D()->isOver)
					GetTouchSensor2D()->isOver = FALSE;
			}

			// Generate an isActive=FALSE event if the button is up
			if (pEvent->m_event == EVENT_MOUSE_UP)
			{
				GetTouchSensor2D()->isActive = FALSE;
				SetGrabbedOff();
				m_mouseDown = FALSE;

				// Also generate a touchTime event (in seconds)
				GetTouchSensor2D()->touchTime = (Time)((Presenter*)GetPresenter())->GetTime() / 1000.0;
			}

			// Unregister with the presenter if the mouse is now up
			if (!m_mouseDown)
			{
				((Presenter*)GetPresenter())->UnregisterEventHandler(this);
				m_registered = FALSE;
			}
			return TRUE;
		}

		// Generate an IsOver event if this is the first time that the
		// mouse has moved here
		if ((pEvent->m_event == EVENT_MOUSE_MOVE) &&
			((BOOL)(Bool)GetTouchSensor2D()->isOver == FALSE))
			GetTouchSensor2D()->isOver = TRUE;

		// If the button is first pressed, then generate an
		// isActive = TRUE event, and grab all movement events
		if ((pEvent->m_event == EVENT_MOUSE_DOWN) && (!m_mouseDown))
		{
			GetTouchSensor2D()->isActive = TRUE;
			SetGrabbedOn();
			m_mouseDown = TRUE;
		}

		// If the button is first released, then generate an
		// isActive = FALSE event, and ungrab all movement events.
		if ((pEvent->m_event == EVENT_MOUSE_UP) && (m_mouseDown))
		{
			GetTouchSensor2D()->isActive = FALSE;
			SetGrabbedOff();
			m_mouseDown = FALSE;

			// Also generate a touchTime event (in seconds)
			GetTouchSensor2D()->touchTime = (Time)((Presenter*)GetPresenter())->GetTime() / 1000.0;
		}

		// Generate a hitPoint_changed event
		Vec2f point(pEvent->m_x, pEvent->m_y);
		GetTouchSensor2D()->hitPoint_changed = point;

		// If not already done, register this node with the presenter
		((Presenter*)GetPresenter())->RegisterEventHandler(this);
		m_registered = TRUE;

		m_prevPass = pEvent->pass;
		return TRUE;
	}

	// Otherwise, if not enabled, return FALSE
	else
		return FALSE;
} /* TouchSensor2DProxy::OnUserEvent */


// =====================================================================

//
// TouchSensor2DProxy::IsFinished
//
// A TouchSensor2D node is considered as finished when it is not enabled.
//

BOOL TouchSensor2DProxy::IsFinished() const
{
	return !((BOOL)(Bool)GetTouchSensor2D()->enabled);
} /* TouchSensor2DProxy::IsFinished */


// ========================================================================
