/*************************************************************************

Module Name:	AnimationQuantizer.cpp

Abstract:	source file for animation streams decoding, quantization part

Author:		Mikael Bourges-Sevenier (France Telecom CNET/DIH/ATI)
Bug fixes by Tony Walsh (Teltec)

Remarks:        This code uses a single adaptative arithmetic decoder for each
                type of animated fields. We have 9 types: position 3D, 
		position 2d, color, angles, normals, scale, rotation angle,
		rotation normals and size.

*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "defs.h"
#include "BifsDec.h"
#include "AnimationQuantizer.h"
#include "AnimStreamDec.h"
#include <math.h>

// log function only for Zvi's Presenter
#include "Executiv.h"
#include "Presenter.h"

int AnimationQuantizer::nbBounds[]={-1,3,2,-1,1,-1,1,1,1,1,1,1,1,1,-1,-1};

AnimationQuantizer::AnimationQuantizer(int category) :
  INbBits(-1),PNbBits(-1),ncomp(-1),
  isUnitVector(FALSE),offset(0),
  type(category),Quantizer(NULL,NULL),models(NULL)
{
  for(int i=0;i<3;i++) 
    IMin[i]=IMax[i]=PMin[i]=PMax[i]=-1;

  if(category<AQ_LAST) {
    nbounds=nbBounds[type];
    ncomp=3; // Optimize necessary!!!
    if (category != AQ_INT) 
      isUnitVector = (category==AQ_ROTATION ||  category== AQ_NORMAL);
  }
  else 
    fprintf(stderr,"AnimationQuantizer : wrong animType = %d\n",category);
}

AnimationQuantizer::~AnimationQuantizer() {
  if(models) 
    delete[] models;
}

/*-----------------------------------------------------------------
   Initializes arithmetic encoding models. 
-----------------------------------------------------------------*/
void AnimationQuantizer::initModels() 
{
  if(!models) 
    models=new ArithModel[ncomp];
}

/*-----------------------------------------------------------------
Resets the statistic of the models
-----------------------------------------------------------------*/
void AnimationQuantizer::resetModels() {
  initModels();
  for (int k=0;k<ncomp;k++)
    models[k].init(1<<PNbBits);
}

/*----------------------------------------------------------------------
Reads initial quantization parameters (during parsing of animation mask)
-----------------------------------------------------------------------*/
void AnimationQuantizer::readInitialParameters(AnimationStreamDecoder *pAS,int category) 
{
  int i;

  type=category;
  nbounds=nbBounds[category];

  if(category>=AQ_LAST || category==AQ_NONE) return;

  useDefault=FALSE;
  initModels();

  switch(type) {
  case AQ_COLOR:
  case AQ_BOUNDFLOAT:
    useDefault=pAS->readBit();
  case AQ_POSITION3D:
  case AQ_POSITION2D:
  case AQ_SIZE3D:
  case AQ_SIZE2D:
  case AQ_FLOAT: {
    if(!useDefault) {
      for(i=0;i<nbounds;i++) 
	IMin[i]=pAS->readGenericFloat(pAS->readBit());
      for(i=0;i<nbounds;i++) 
	IMax[i]=pAS->readGenericFloat(pAS->readBit());
    }
  }
  break;
  case AQ_ANGLE:
    IMin[0]=0; IMax[0]=2*(float) M_PI;
    break;
  case AQ_NORMAL:    
  case AQ_ROTATION:
    for(i=0;i<nbounds;i++) {
      IMin[i]=0;
      IMax[i]=1;
    }
    break;
  case AQ_INT:
    IMin[0]=pAS->readBits(32);
    break;
  }
  INbBits=pAS->readBits(5);
  setINbBits(INbBits);
  for(i=0;i<nbounds;i++) {
    int vq=pAS->readBits(INbBits+1);
    PMin[i]=vq-(1<<INbBits);
  }
  PNbBits=pAS->readBits(4);
}

/*-----------------------------------------------------------------
Re-reads quantization parameters during an Intra Frame
-----------------------------------------------------------------*/
void AnimationQuantizer::readSame(AnimationStreamDecoder *pAS) {
  if(type>=AQ_LAST || type==AQ_NONE) return;

  if(pAS->readBit()) { // hasMinMax
    int i;
    useDefault=FALSE;

    switch(type) {
    case AQ_COLOR:
    case AQ_BOUNDFLOAT:
      useDefault=pAS->readBit();
    case AQ_POSITION3D:
    case AQ_POSITION2D:
    case AQ_SIZE3D:
    case AQ_SIZE2D:
    case AQ_FLOAT: {
      if(!useDefault) {
	for(i=0;i<nbounds;i++) 
	  IMin[i]=pAS->readGenericFloat(pAS->readBit());
	for(i=0;i<nbounds;i++) 
	  IMax[i]=pAS->readGenericFloat(pAS->readBit());
      }
    }
    break;
    case AQ_ANGLE:
      IMin[0]=0; IMax[0]=2*(float) M_PI;
      break;
    case AQ_NORMAL:
    case AQ_ROTATION:
      for(i=0;i<nbounds;i++) {
	IMin[i]=0;
	IMax[i]=1;
      }
      break;
    case AQ_INT:
      IMin[0]=pAS->readBits(32);
      break;
    }

    if(pAS->readBit()) {// hasINbBits
      INbBits=pAS->readBits(5);
      setINbBits(INbBits);
    }

    if(pAS->readBit()) // hasPMinMax
      for(i=0;i<nbounds;i++) {
	int vq=pAS->readBits(INbBits+1);
	PMin[i]=vq-(1<<INbBits);
      }

    if(pAS->readBit()) // hasPNbBits
      PNbBits=pAS->readBits(4);
  }
}

/*-----------------------------------------------------------------
Sets intra nbBits and prepares scaling factor for unquantization.
-----------------------------------------------------------------*/
void AnimationQuantizer::setINbBits(int iNbBits) {
  INbBits=iNbBits;
  
  int  nbBits = max(0,(isUnitVector) ? INbBits-1 : INbBits);
  for (int k=0;k<ncomp;k++) {
    float span = IMax[k]-IMin[k];
    float smallFloat = 1e-8f;
    if(type==AQ_INT)
      unQuantScale[k]=1;
    else if (span<=smallFloat) 
      unQuantScale[k]=0;
    else if(type==AQ_ROTATION || type==AQ_NORMAL) 
      unQuantScale[k] = span/((1<<(nbBits-1)) - 1);
    else 
      unQuantScale[k] = (nbBits==0) ? span*.5f : span/((1<<nbBits) - 1);
  }    
  offset = (isUnitVector) ? (1<<(INbBits-1)) : 0 ;
}  

/*-----------------------------------------------------------------
Dequantifies a value.
  vq     : quantified value
  compIdx: axis of the component (x,y,z)
-----------------------------------------------------------------*/
float AnimationQuantizer::unQuantize(int vq,int compIdx) {
  int aacIdx = (ncomp==1)?0:compIdx;
  return IMin[aacIdx] + unQuantScale[aacIdx] * (float)vq;
 // return IMin[0] + unQuantScale[0] * (float)vq;
}

/*-----------------------------------------------------------------
Decodes Intra component
-----------------------------------------------------------------*/
int AnimationQuantizer::decodeIComp(AnimationStreamDecoder *pAS) 
{
  return pAS->readBits(INbBits)- offset;
}

/*-----------------------------------------------------------------
Decodes Predictive component
-----------------------------------------------------------------*/
int AnimationQuantizer::decodePComp(AnimationStreamDecoder *pAS,ArithModel *models,int k) 
{
  return pAS->getAAD().decodeAndUpdate(models[k])+ PMin[k];
}

/*-----------------------------------------------------------------
Parses an Intra value
-----------------------------------------------------------------*/
void AnimationQuantizer::ParseAnimValue(AnimationStreamDecoder *pAS,BOOL isIntra)
{
  // decoded value is always stored on vDelta
  // in Intra, vr=vDelta
  // in Predictive, vr=vq1+vDelta except for Normal & Rotation


  if(isIntra) {
    resetModels();


    // read Header
    switch(type) {
    case AQ_NORMAL:
      vq1.direction=pAS->readBits(1);
    case AQ_ROTATION:
      vq1.orientation=pAS->readBits(2);
      break;
    }
    
    // read components
    for (int i=0;i<ncomp;i++)
	{
		vq1.vt[i] =  vq1.vq[i] = decodeIComp(pAS);
		if ((type == AQ_NORMAL) || (type == AQ_ROTATION))
			vq1.vt[i] = vq1.vt[i] - (1<<(INbBits-1));
	}
  }
  else {
    // read inverse (for Normals)
    if (type==AQ_NORMAL) 
      vDelta.inverse=pAS->getAAD().decode(bitModel);
    
    // read Components
    for (int i=0;i<ncomp;i++) {
      vDelta.vt[i] = decodePComp(pAS,models,(nbounds==1) ? 0 : i);
	  if ((type == AQ_NORMAL) || (type == AQ_ROTATION))
		vq1.vt[i] = vq1.vt[i] - (1<<(INbBits-1));
    }
    switch(type) {
    case AQ_POSITION3D:
    case AQ_POSITION2D:
    case AQ_COLOR:
    case AQ_ANGLE:
    case AQ_FLOAT:

    case AQ_BOUNDFLOAT:
    case AQ_SIZE3D:
    case AQ_SIZE2D:
    case AQ_INT:
      vq1.add(vDelta,ncomp); // now vq1 contains the whole information
      break;
    case AQ_NORMAL: // addDelta stores result in vq1
      addDelta(vq1,vDelta,INbBits,2);
      break;
    case AQ_ROTATION:
      vq1.direction=vDelta.direction=1;
      addDelta(vq1,vDelta,INbBits,3);
      break;
    }
  }

	if (nbounds == 1)
	  	for(int i=0;i<ncomp;i++)
			vq1.vq[i] = IMin[0] + unQuantScale[0] * (float)vq1.vt[i];
	else
		for(int i=0;i<ncomp;i++)
			vq1.vq[i]=unQuantize(vq1.vt[i],i);
	



}
/*-----------------------------------------------------------------
Parses a Vec3f value
-----------------------------------------------------------------*/
Vec3f AnimationQuantizer::ParseVec3f(AnimationStreamDecoder *pAS,BOOL isIntra)
{
  if(type==AQ_NORMAL) 
    return ParseNormal(pAS,isIntra);
  
  ncomp=3;

  ParseAnimValue(pAS,isIntra);

  Vec3f vr(vq1.vq[0],vq1.vq[1],vq1.vq[2]);

  return vr;
}

/*-----------------------------------------------------------------
Parses a Vec2f value
-----------------------------------------------------------------*/
Vec2f AnimationQuantizer::ParseVec2f(AnimationStreamDecoder *pAS,BOOL isIntra)
{
  ncomp=2;

  ParseAnimValue(pAS,isIntra);
    
  Vec2f vr(vq1.vq[0],vq1.vq[1]);

  return vr;
}

/*-----------------------------------------------------------------
Parses a Color3f value
-----------------------------------------------------------------*/
Color3f AnimationQuantizer::ParseColor(AnimationStreamDecoder *pAS,BOOL isIntra)
{
  ncomp=3;

  ParseAnimValue(pAS,isIntra);
    
  Color3f vr(vq1.vq[0],vq1.vq[1],vq1.vq[2]);
  
  return vr;
}

/*-----------------------------------------------------------------
Parses an Int value
-----------------------------------------------------------------*/
int AnimationQuantizer::ParseInt(AnimationStreamDecoder *pAS,BOOL isIntra)
{
  ncomp=1;

  ParseAnimValue(pAS,isIntra);

  int vr=(int)vq1.vq[0];

  return vr;
}

/*-----------------------------------------------------------------
Parses a float value
-----------------------------------------------------------------*/
float AnimationQuantizer::ParseFloat(AnimationStreamDecoder *pAS,BOOL isIntra)
{
  ncomp=1;

  ParseAnimValue(pAS,isIntra);

  float vr=vq1.vq[0];

  return vr;
}

/*-----------------------------------------------------------------
Parses a Rotation value
-----------------------------------------------------------------*/
Rotation AnimationQuantizer::ParseRotation(AnimationStreamDecoder *pAS,BOOL isIntra)
{
  ncomp=3;

  ParseAnimValue(pAS,isIntra);
  vq1.direction=1;

  // convert to native component
  float natComp[4];
  quantToNat (vq1,natComp,ncomp);

  // convert to rotation
  float alpha2=acos(natComp[0]);
  if(fabs(alpha2)<1e-5) {
    return Rotation(0,1,0,0);
  }


  Rotation vr(natComp[1]/sin(alpha2),natComp[2]/sin(alpha2),natComp[3]/sin(alpha2),2*alpha2);
    
  return vr;
}

/*-----------------------------------------------------------------
Parses a normal
-----------------------------------------------------------------*/
Vec3f AnimationQuantizer::ParseNormal(AnimationStreamDecoder *pAS,BOOL isIntra) {
  ncomp=2;
  ParseAnimValue(pAS,isIntra);

  // convert to native component
  float natComp[3];
  quantToNat (vq1,natComp,ncomp);


  Vec3f vr(natComp[0],natComp[1],natComp[2]);

  return vr;
}

BOOL AnimationQuantizer::operator == (const AnimationQuantizer &qp) const {
  if(type==qp.type) return TRUE;
  return FALSE;
}


/*-----------------------------------------------------------------
Translates from quantized components to native components 
for Normal & Rotation only
-----------------------------------------------------------------*/
// static
void AnimationQuantizer :: quantToNat (QVAL &qv, float *natComp, int ncomp)
{
  int j,len=ncomp+1;
  
  int direction   = qv.direction; // only for Normal else 1
  int orientation = qv.orientation;
			  

  float delta=1;
  float vt[4];
  for (j=0;j<ncomp;j++) {
    float v = qv.vq[j];
    vt[j]= (float) tan (v*M_PI/4);
    delta+= vt[j]*vt[j];
  }
  delta = (float)(direction/sqrt(delta));
  
  natComp[(orientation)%len] = delta;
  for (j=0;j<ncomp;j++) {
    natComp[(orientation+j+1)%len] = vt[j]*delta;
  }


}

/*-----------------------------------------------------------------
Computes the sum of a quantized compressed unit vector and a delta
for Normal & Rotation only
-----------------------------------------------------------------*/
// static
void AnimationQuantizer :: addDelta (QVAL &qVal1,QVAL &qDelta,int NbBits,int NBCOMP) {
  //  NbBits = number of bits the quantization is made on
  //  NBCOMP = number of reduced components : 2 for Normals, 3 for Rotations
  //  qVal1  = NBCOMP quantized coordinates of the 1st compressed object
  //  qVal2  = NBCOMP quantized coordinates of the 2nd compressed object
  //  qDelta = NBCOMP quantized coordinates of the delta
  
  QVAL qVal2;
  int scale = (1<<max(NbBits-1,0))-1;   // highest possible quantized value

  // computes the difference in orientation and direction
  int len=NBCOMP+1; // length of decoded value
  int deltaInverse = (qDelta.inverse==FALSE) ? 1 : -1;
  
  int inv=1;
  int diffDirection   = qVal1.direction * deltaInverse;
  int dOri = 0;
  
  qVal2.orientation=qVal1.orientation;
  qVal2.direction=diffDirection;

  // Computes result's compressed components
  for (int j=0;j<NBCOMP;j++) {
    qVal2.vq[j] = qVal1.vt[j]+qDelta.vt[j];
    if (fabs(qVal2.vq[j])>scale) {  // going out of the cube's face
      dOri = j+1;        // relative orientation we are escaping towards
      inv = (qVal2.vq[j]>=0) ? 1 : -1;   // direction are we escaping towards
      break;
    }
  }
  
  // If need to change face --> reindexation
  if (dOri!= 0) {
    int j,k;
    for (j=0;j<NBCOMP-dOri;j++) {
      k = (j+dOri)%NBCOMP;
      qVal2.vq[j] = inv * (qVal1.vq[k]+qDelta.vq[k]);
    }
    k = dOri-1;
    qVal2.vq[NBCOMP-dOri] = inv * 2 * scale - (qVal1.vq[k]+qDelta.vq[k]);
    for (j=NBCOMP-dOri+1;j<NBCOMP;j++) {
      k = (j+dOri-1)%NBCOMP;
      qVal2.vq[j] = inv * (qVal1.vq[k]+qDelta.vq[k]);
    }

    qVal2.orientation=(qVal1.orientation+dOri) % len;
    qVal2.direction=diffDirection*inv;
  }
  
  qVal1.set(qVal2);
}

