/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Encoder
FILE:       Token.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of Token, the class that handles text parsing.
-------------------------------------------------------------------------*/
#include "defs.h"
#include "Token.h"
#include "MediaObj.h"
#include "Proto.h"

/*-------------------------------------------------------------
Display an error message at exit, if necessary.
-------------------------------------------------------------*/
Token :: ~Token ()
{
	if (m_bHadErrors)
	{	printf (
			"WARNING: Had parsing errors!\07\07\07\n"
			"Hit any key to continue\n"
		);
		getchar ();
	}
}

/*-----------------------------------------------------------------
Initiate the Token object
-----------------------------------------------------------------*/
void Token :: Init (FILE *fInput, FILE *fList)
{
	m_bNewSourceLine  = FALSE;
	m_fInput = fInput;
	if (fList)
		m_fList = fList;
	m_nInputOffset = m_nInputLength = 0;
	m_cNext = GetChar ();
	m_code = TID_UNKNOWN;
	Scan ();
}

/*-----------------------------------------------------------------
Translate the token to an int
-----------------------------------------------------------------*/
int Token :: GetInt ()
{
	int nResult = 0;

	if (m_code == TID_FALSE)
	{	Scan ();
		return 0;
	}
	else if (m_code == TID_TRUE)
	{	Scan ();
		return 1;
	}
	else if (m_code != TID_INTEGER)
		Error ("Integer number expected");
	else
	{	int i = 0, nRadix = 10, c;
		int nSign = 1;
		switch (m_string [0]) {
		case '-':
			nSign = -1;
		case '+':
			++i;
		}
		if (m_string [i] == '0' && tolower (m_string [i + 1]) == 'x')
		{	nRadix = 16;
			i += 2;
		}
		for (;(c = tolower (m_string [i])) != 0; i++)
		{	if (c >= '0' && c <= '9')
				c -= '0';
			else if (nRadix == 16 && c >= 'a' && c <= 'f')
				c -= 'a' - 10;
			else
			{	Error ("Integer number expected");
				break;
			}
			nResult *= nRadix;
			nResult += c;
		}
		nResult *= nSign;
	}

	if (m_code == TID_INTEGER || m_code == TID_NUMBER)
		Scan ();
	return nResult;
}

/*-----------------------------------------------------------------
Translate TRUE/FALSE value
-----------------------------------------------------------------*/
bool Token :: GetBool ()
{
	bool bResult;

	switch (m_code) {
	case TID_TRUE:
		bResult = true;
		break;
	case TID_FALSE:
		bResult = false;
		break;
	case TID_NUMBER:
		Error ("Boolean expected");
		Scan ();
		return FALSE;
	default:
		Error ("Boolean expected");
		return FALSE;
	}
	Scan ();
	return bResult;
}

/*-----------------------------------------------------------------
Translate a real value
-----------------------------------------------------------------*/
double Token :: GetReal ()
{
	double dResult;

	if (m_code != TID_NUMBER && m_code != TID_INTEGER)
	{	Error ("Number expected");
		return 0;
	}

	dResult = atof (m_string);
	Scan ();
	return dResult;
}

/*-----------------------------------------------------------------
If quotes are not used, read all characters till a white space 
-----------------------------------------------------------------*/
const ZString &Token :: GetString ()
{
	if (m_code != TID_STRING)
		while (
			m_cNext != ' ' && m_cNext != '\t' && m_cNext != '\n' &&
			m_cNext != EOF
		)
		{	m_string += m_cNext;
			m_cNext = GetChar ();
		}
	return m_string;
}
	
/*-----------------------------------------------------------------
Scan a symbol and find its code
-----------------------------------------------------------------*/
void Token :: Scan ()
{
	if (m_code == TID_EOF)
		throw ExceptionEndOfFile ();

	do
	{	/*-----------------------------------------------------------------
		Skip all white spaces (and comments)
		-----------------------------------------------------------------*/
		while (
			m_cNext == ' ' || m_cNext == '\t' || m_cNext == '\n' ||
			m_cNext == '\r' || m_cNext == '#' ||
			m_cNext == '/' && m_szLine [m_nInputOffset] == '/'
		)
		{	// Handle a comment, # or // to the end of the line
			if (m_cNext == '#' || m_cNext == '/')
				while (m_cNext != '\n' && m_cNext != EOF)
					m_cNext = GetChar ();
			m_cNext = GetChar ();
		}
		if (m_cNext == EOF)
			break;

		m_string = "";
		m_code = TID_UNKNOWN;

		/*-----------------------------------------------------------------
		Parse a name
		-----------------------------------------------------------------*/
		if (isalpha (m_cNext) || m_cNext == '_' || m_cNext == '$')
		{	while (
				isalnum (m_cNext) || m_cNext == '_' || m_cNext == '$' ||
				m_cNext == '-'
			)
			{	m_string += m_cNext;
				m_cNext = GetChar ();
			}
			m_code = TID_NAME;
		}

		/*-----------------------------------------------------------------
		Parse a number
		-----------------------------------------------------------------*/
		else if (isdigit (m_cNext) || m_cNext == '-' || m_cNext == '+')
		{	m_code = TID_INTEGER;
			while (
				isalnum (m_cNext) || m_cNext == '-' || m_cNext == '+' ||
				m_cNext == '.'
			)
			{	m_string += m_cNext;
				if (m_cNext == '.' || tolower (m_cNext) == 'e')
					m_code = TID_NUMBER;
				m_cNext = GetChar ();
			}
		}

		/*-----------------------------------------------------------------
		Parse a string
		-----------------------------------------------------------------*/
		else if (m_cNext == '"')
		{	while ((m_cNext = GetChar ()) != '"')
			{	if (m_cNext == EOF)
					throw ExceptionEndOfFile ();
				/*-------------------------------------------------------------
				Do UTF-8 translation
				-------------------------------------------------------------*/
				if (m_cNext & 0x80)
				{	m_string += 0xc0 | (m_cNext >> 6);
					m_cNext &= 0x3f;
				}
				m_string += m_cNext;
			}
			m_cNext = GetChar ();
			m_code = TID_STRING;
		}

		/*-------------------------------------------------------------
		Read a complete file into a string
		-------------------------------------------------------------*/
		else if (m_cNext == '<')
		{	while ((m_cNext = GetChar ()) != '>')
			{	if (m_cNext == EOF)
					throw ExceptionEndOfFile ();
				m_string += m_cNext;
			}
			m_cNext = GetChar ();
			m_code = TID_STRING;
			FILE *fExtFile = fopen (m_string, "rb");
			if (fExtFile == NULL)
				Error ("File not found");
			else
			{	int c;
				m_string = "";
				while ((c = fgetc (fExtFile)) != EOF)
					m_string += c;
				fclose (fExtFile);
			}
			return;
		}

		/*-----------------------------------------------------------------
		Parse one character
		-----------------------------------------------------------------*/
		else
		{	m_string += m_cNext;
			m_cNext = GetChar ();
		}

		/*-----------------------------------------------------------------
		Check if a reserved symbol
		-----------------------------------------------------------------*/
		for (int i = TID_LEFT_BRACE; i < TID_NUMBER_OF_RESERVED; i++)
			if (!stricmp (m_string, m_reserved [i]))
			{	m_code = (TOKEN_CODE) i;
				return;
			}
		
		if (m_code != TID_UNKNOWN)
			return;
		Error ("Invalid symbol");
	}
	while (m_cNext != EOF);
	m_code = TID_EOF;
}

/*-----------------------------------------------------------------
Find a name in a list of defined name
-----------------------------------------------------------------*/
int Token :: GetDefinedName () const
{
	for (int i = m_defined.GetLength (); --i >= 0;)
		if (m_string == m_defined [i].m_name)
			return i;
	return -1;
}

/*-----------------------------------------------------------------
Find a name in a list of defined route
-----------------------------------------------------------------*/
int Token :: GetDefinedRoute () const
{
	for (int i = m_routes.GetLength (); --i >= 0;)
		if (m_string == m_routes [i])
			return i;
	return -1;
}

/*-----------------------------------------------------------------
Find a name in a list of defined name
-----------------------------------------------------------------*/
int Token :: GetDefinedProto (MediaObject **ppObject) const
{
	for (int i = m_protos.GetLength (); --i >= 0;)
		if (m_string == m_protos [i].m_name)
		{	if (ppObject)
				*ppObject = m_protos [i].m_pObject;
			return i;
		}
	return -1;
}

/*-----------------------------------------------------------------
Display an error message in the list file
-----------------------------------------------------------------*/
void Token :: Error (LPCSTR szDesc, LPCSTR szItem)
{
	fprintf (
		m_fList,
		"ERROR: %s - %s\n",
		szItem ? szItem : (LPCSTR) m_string,
		szDesc
	);
	m_bHadErrors = TRUE;
}

/*-----------------------------------------------------------------
Add a name to the list of predefined name

MBS: modified to take into account defined nodeID in the script 
     in the forme text_nnn where nnn is the specified nodeID (nodeID < 1024)
Zvi (29/4/98) - Fixed the mess.
-----------------------------------------------------------------*/
int Token :: AddPredefinedName ()
{
	int nResult, nID;

	/*-------------------------------------------------------------
	Verify the a predefined slot is empty, or find an empty slot
	-------------------------------------------------------------*/
	if (!_strnicmp ("id_", m_string, 3) &&
		(nID = atoi ((LPCSTR) m_string + 3)) > 0
	)
	{	if (nID < m_defined.GetLength () && m_defined [nID].m_nCode != INT_MAX)
			return -1;
		nResult = nID;
	}
	else for (nResult = 0; nResult < m_defined.GetLength (); nResult++)
		if (m_defined [nResult].m_nCode == INT_MAX)
			break;

	m_defined.SetAt (nResult, DefinedNode ());
	m_defined [nResult].m_name = m_string;
	m_defined [nResult].m_nCode = 0;
	return nResult;
}

/*-------------------------------------------------------------
Add a name to the list of defined Protos
-------------------------------------------------------------*/
int Token :: AddPredefinedProto (Proto *pObject)
{
	int nResult = m_protos.GetLength ();
	DefinedProto &entry = ++m_protos;
	entry.m_name = m_string;
	entry.m_pObject = pObject;
	return nResult;
}

/*-----------------------------------------------------------------
Add a name to the list of predefined routes
-----------------------------------------------------------------*/
int Token :: AddPredefinedRoute ()
{
	int nResult = m_routes.GetLength ();
	m_routes += m_string;
	return nResult;
}

/*-------------------------------------------------------------
Destructor for the DefinedProto structure
-------------------------------------------------------------*/
Token :: DefinedProto :: ~DefinedProto ()
{
	delete m_pObject;
}

/*-----------------------------------------------------------------
Read one character from the input file. Each input line is copied to the
list file
-----------------------------------------------------------------*/
int Token :: GetChar ()
{
	if (m_bNewSourceLine)
	{	fputs (m_szLine, m_fList);
		m_bNewSourceLine = FALSE;
	}

	if (m_nInputOffset >= m_nInputLength)
	{	if (feof (m_fInput))
			return EOF;
		if (!fgets (m_szLine, sizeof (m_szLine), m_fInput))
			return EOF;
		m_nInputLength = strlen (m_szLine);
		m_nInputOffset = 0;
		m_bNewSourceLine = TRUE;
	}
	return m_szLine [m_nInputOffset++];
}

/*-------------------------------------------------------------
Save old values of the token, and prepare it to parse a secondary file
-------------------------------------------------------------*/
void Token :: Push ()
{
	m_stack.m_code = m_code;
	m_stack.m_string = m_string;
	m_stack.m_cNext = m_cNext;
	m_stack.m_nInputOffset = m_nInputOffset;
	m_stack.m_nInputLength = m_nInputLength;
	m_stack.m_fInput = m_fInput;
	memcpy (m_stack.m_szLine, m_szLine, 300);
	m_stack.m_bNewSourceLine = m_bNewSourceLine;
}

/*-------------------------------------------------------------
Restore old token values after parsing a secondary file
-------------------------------------------------------------*/
void Token :: Pop ()
{
	m_code = m_stack.m_code;
	m_string = m_stack.m_string;
	m_cNext = m_stack.m_cNext;
	m_nInputOffset = m_stack.m_nInputOffset;
	m_nInputLength = m_stack.m_nInputLength;
	m_fInput = m_stack.m_fInput;
	memcpy (m_szLine, m_stack.m_szLine, 300);
	m_bNewSourceLine = m_stack.m_bNewSourceLine;
}

/*-----------------------------------------------------------------
Define the list of reserved tokens
-----------------------------------------------------------------*/
LPCSTR Token :: m_reserved [] = 
{
	"",
	"{",
	"}",
	"[",
	"]",
	",",
	".",
	"",
	"USE",
	"DEF",
	"FALSE",
	"TRUE",
	"ROUTE",
	"TO",
	"AT",
	"INSERT",
	"REPLACE",
	"DELETE",
	"APPEND",
	"LAST",
	"SCENE",
	"BY",
	"UPDATE",
	"REMOVE",
	"OD",
	"ESD",
	"IN",
	"IPMPD",
	"FROM",
	"FILE",
	"PROTO",
	"EXTERNPROTO",
	"FIELD", "EXPOSEDFIELD", "EVENTIN", "EVENTOUT",
	"SFBOOL", "SFFLOAT", "SFTIME", "SFINT32", "SFSTRING",
	"SFVEC3F", "SFVEC2F", "SFCOLOR", "SFROTATION", "SFIMAGE", "SFNODE",
	"MFBOOL", "MFFLOAT", "MFTIME", "MFINT32", "MFSTRING",
	"MFVEC3F", "MFVEC2F", "MFCOLOR", "MFROTATION", "MFIMAGE", "MFNODE",
	"IS"
};		
