/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1997.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Decoder
FILE:       BifsDec.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of BIFSDecoder.
-------------------------------------------------------------------------*/
#include "defs.h"
#include <assert.h>
#include <Executiv.h>
#include "BifsDec.h"
#include "NodeFld.h"
#include "Root.h"
#include "Nodes.h"
#include "Route.h"
#include "Except.h"
#include "MemBifs.h"
#include "Proto.h"

#ifdef BIFSCAR
#include "BifsCar.h"
#else
#include "InLine.h"
#include "ODP.h"
#endif

#define pCoder this

/*-----------------------------------------------------------------
BIFSDecoder constructor
-----------------------------------------------------------------*/
BIFSDecoder :: BIFSDecoder (
	Executive *pExecutive,
	InlineImp *pSource,
	FILE *fList,
	LPCSTR szListTitle,
	int flags,
	int objectTypeIndication
) :
	SysCoder (fList, szListTitle, flags, objectTypeIndication),
	m_pExecutive (pExecutive),
	m_pSource (pSource),
	m_nNodeIDbits (10),
	m_nRouteIDbits (10),
	m_nProtoIDbits (10),
	m_bPixelMetric (FALSE),
	m_nPixelWidth (0),
	m_nPixelHeight (0),
	m_bUse3DMeshCoding (false),
	m_bUsePredictiveMFField (false),
	m_pScene (pSource ? pSource -> GetScene () : NULL)
{
	// MBS: inits current QP node to null
	m_pCurrentQP=NULL;
	// end MBS
}

/*-------------------------------------------------------------
Report an error to the executive
-------------------------------------------------------------*/
void BIFSDecoder :: Error (int nError)
{
	m_pExecutive -> Error ((NON_FATAL_ERROR) nError);
}

/*-------------------------------------------------------------
This virtual function is called if the parser exausted the unit at the
middle of decoding
-------------------------------------------------------------*/
void BIFSDecoder :: EndOfUnit ()
{
	throw BifsException (BIFX_END_OF_STREAM);
}

/*-----------------------------------------------------------------
Parses ObjectDescriptors and scene updates from the stream, and update
the scene accordingly.
-----------------------------------------------------------------*/
void BIFSDecoder :: ParseFrame (LPCBYTE pData, int nLength, DWORD dwTS)
{
	Reset (pData, nLength);
	dwTS;
	TRACEF ("AT %d", dwTS);
	TRACENEWLINE;
	try
	{
		ParseUpdateFrame ();
		m_pExecutive -> RequestChannels ();
	}
	catch (BifsException)
	{
		Error (SCENE_PARSING_ERROR);
	}
}

/*-----------------------------------------------------------------
Parse RoutesDescription
-----------------------------------------------------------------*/
void BIFSDecoder :: ParseRoutesDescription (Root *pRoot)
{
	if (ParseInt (1, "ListDescription"))		// if ListDescription 
		do
			ParseRoute (pRoot);
		while (ParseInt (1, "moreROUTEs"));		// while moreROUTEs
	else
	{	int nBits = ParseInt (5, "nbits");
		int nSize = ParseInt (nBits, "length");
		for (int i = nSize; --i >= 0;)
			ParseRoute (pRoot);
	}
}

/*-----------------------------------------------------------------
Parse one Route and insert it into the scene whose root is pRoot
-----------------------------------------------------------------*/
void BIFSDecoder :: ParseRoute (Root *pRoot, int nID)
{
	/*-----------------------------------------------------------------
	if nID >= we already know the ID
	-----------------------------------------------------------------*/
	ZString routeName;
	if (nID < 0)
	{	if (ParseInt (1, "isUpdatable"))	// if isUpdatable
		{	nID = ParseInt (GetRouteIDbits (), "RouteID");
			if (pRoot -> GetInline () ->IsCodeNames ())
			{	ParseString (routeName);
				TRACEF ("DEF %s ", (LPCSTR) routeName);
			}
			else
				TRACEF ("DEF %d ", nID);
		}
	}

	int nNodeID = ParseInt (GetNodeIDbits (), "NodeID");
	MediaObject *pNode = pRoot -> FindNode (nNodeID);
	if (pNode == NULL)
		throw BifsException (BIFX_NODEID_NOT_FOUND);
	TRACEF ("ROUTE %s.", pNode ->GetNameID ());
	if (!pNode -> AddRoute (nID, this, routeName))
		throw BifsException (BIFX_PARSING_ERROR);
	TRACENEWLINE;
}

/*-----------------------------------------------------------------
Parse an UpdateFrame
-----------------------------------------------------------------*/
void BIFSDecoder :: ParseUpdateFrame ()
{
	TRACEF ("UpdateScene {", 0);
	TRACENEWLINE;
	TRACEINDENT (1);
	
	do
	{	
		int nCommandCode = ParseInt (2, "nCommandCode");	// changed to 2 in FDIS

		switch (nCommandCode) {
		case 0 :
			ParseInsertionCommand ();
			break;
		case 1:
			ParseDeletionCommand ();
			break;
		case 2:
			ParseReplacementCommand ();
			break;
		case 3:
			ParseSceneReplaceCommand ();
			break;
		}

		TRACENEWLINE;
	}
	while (ParseInt (1, "moreCommands"));

	TRACEINDENT (-1);
	TRACEF ("}", 0);
	TRACENEWLINE;
}

/*-----------------------------------------------------------------
Parse Insertion Command
-----------------------------------------------------------------*/
void BIFSDecoder :: ParseInsertionCommand ()
{
	int nID;
	MediaObject *pNode;

	if (!m_pScene)
		throw BifsException (BIFX_NODEID_NOT_FOUND);

	switch (ParseInt (2, "type")) {
	case 0:
		nID = ParseInt (GetNodeIDbits (), "NodeID");
		pNode = m_pScene -> FindNode (nID);
		if (!pNode)
			throw BifsException (BIFX_NODEID_NOT_FOUND);
		TRACEF ("Insert node into %s.children ", pNode ->GetNameID ());
		pNode -> ParseChild (this);
		break;
	case 2:
		nID = ParseInt (GetNodeIDbits (), "NodeID");
		pNode = m_pScene -> FindNode (nID);
		if (!pNode)
			throw BifsException (BIFX_NODEID_NOT_FOUND);
		TRACEF ("Insert into %s.", pNode ->GetNameID ());
		pNode -> InsertValue (this);
		break ;
	case 3:
		ParseRoute (m_pScene);
		break;
	}
}

/*-----------------------------------------------------------------
Parse DeletionCommand
-----------------------------------------------------------------*/
void BIFSDecoder :: ParseDeletionCommand ()
{
	int nID = 0;
	MediaObject *pNode;

	if (!m_pScene)
		throw BifsException (BIFX_NODEID_NOT_FOUND);

	int nType = ParseInt (2, "type");
	if (nType != 3)
		nID = ParseInt (GetNodeIDbits (), "NodeID");
	switch (nType) {
	case 0:
		pNode = m_pScene -> FindNode (nID);
		if (!pNode)
			throw BifsException (BIFX_NODEID_NOT_FOUND);
		TRACEF ("Delete %s", pNode ->GetNameID ());
		pNode -> Delete ();
		break;
	case 2:
		pNode = m_pScene -> FindNode (nID);
		if (!pNode)
			throw BifsException (BIFX_NODEID_NOT_FOUND);
		TRACEF ("Delete %s.", pNode ->GetNameID ());
		pNode -> DeleteValue (this);
		break ;
	case 3:
		nID = ParseInt (GetRouteIDbits (), "RouteID");
		Route *pRoute = m_pScene ->FindRoute (nID);
		if (pRoute)
		{
			TRACEF ("Delete ROUTE %s", pRoute ->GetName ());
		}
		else
			TRACEF ("Delete ROUTE %d", nID);
		delete pRoute;
		break ;
	}
}

/*-----------------------------------------------------------------
Parse replacement command
-----------------------------------------------------------------*/
void BIFSDecoder :: ParseReplacementCommand ()
{
	int nID = 0;
	MediaObject *pNode;

	if (!m_pScene)
		throw BifsException (BIFX_NODEID_NOT_FOUND);

	int nType = ParseInt (2, "type");
	if (nType != 3)
		nID = ParseInt (GetNodeIDbits (), "NodeID");
	switch (nType) {
	case 0:
		pNode = m_pScene -> FindNode (nID);
		if (!pNode)
			throw BifsException (BIFX_NODEID_NOT_FOUND);
		TRACEF ("REPLACE %s BY ", pNode ->GetNameID ());
		pNode -> ReplaceNode (this);
		break;
	case 1:
		pNode = m_pScene -> FindNode (nID);
		if (!pNode)
			throw BifsException (BIFX_NODEID_NOT_FOUND);
		TRACEF ("REPLACE %s.", pNode ->GetNameID ());
		pNode -> ReplaceValue (this);
		break;
	case 2:
		pNode = m_pScene -> FindNode (nID);
		if (!pNode)
			throw BifsException (BIFX_NODEID_NOT_FOUND);
		TRACEF ("REPLACE %s.", pNode ->GetNameID ());
		pNode -> ReplaceIndexedValue (this);
		break ;
	case 3:
		nID = ParseInt (GetRouteIDbits (), "RouteID");
		Route *pRoute = m_pScene ->FindRoute (nID);
		if (pRoute)
		{
			TRACEF ("REPLACE ROUTE %s BY ", pRoute ->GetName ());
		}
		else
			TRACEF ("REPLACE ROUTE %d BY ", nID);
		delete pRoute;;
		ParseRoute (m_pScene, nID);
		break ;
	}
}

/*-----------------------------------------------------------------
Replace the scene (or generate the first scene
-----------------------------------------------------------------*/
void BIFSDecoder :: ParseSceneReplaceCommand ()
{
	/*-------------------------------------------------------------
	Delete the old scene
	-------------------------------------------------------------*/
	m_pSource -> ReplaceScene (NULL);
	m_pScene = m_pSource -> GetScene ();

	/*-----------------------------------------------------------------
	The BIFS parser architecture is based on the principle that every object,
	node or node field, parses itself.  So to parse the scene, we
	instantiate an SFNode object, and call the object to parse itself.
	The parse function will throw an exception if end of stream is 
	unexpectedly encountered.
	-----------------------------------------------------------------*/
	ParseInt (6, "reserved");
	m_pSource -> SetCodeNamesFlags (ParseInt (1, "CODENAMES"));
	m_pScene -> SetInline (m_pSource);
	m_pScene -> ParseProtoList (this);

	SFTopNode node;
	if (node.ParseNodeField (this, m_pScene, NULL, TRUE))
	{	/*-----------------------------------------------------------------
		If parse was successful, "node" now points to the scene root.
		-----------------------------------------------------------------*/
		m_pScene -> Configure (
			m_bPixelMetric,
			m_nPixelWidth,
			m_nPixelHeight,
			m_nNodeIDbits,
			m_nRouteIDbits,
			m_nProtoIDbits
		);

		/*-----------------------------------------------------------------
		Parse Routes
		-----------------------------------------------------------------*/
		if (ParseInt (1, "hasROUTEs"))	// if hasROUTEs
			ParseRoutesDescription (m_pScene);
		m_pSource -> ReplaceScene ((Root *) node.GetNode ());
	}
	else
		throw BifsException (BIFX_PARSING_ERROR);
}

#ifndef BIFSCAR
/*-------------------------------------------------------------
Parse the data in the specificInfo field of the ES descriptor,
and configure the decoder appropriately.
-------------------------------------------------------------*/
bool BIFSDecoder :: Configure (const BYTE *specificInfo, int nInfoLength)
{
	BIFSConfig *pBifsConfig = NULL;
	bool bResult;
	try
	{	if (nInfoLength > 0)
		{	ODDecoder odDec (
				m_pExecutive,
				specificInfo,
				nInfoLength,
				m_pExecutive -> GetTraceFile (),
				"BIFSConfig",
				m_pExecutive -> GetTraceFlags (),
				GetVersion ()
			);
			if (GetVersion () == 0)
				pBifsConfig = new BIFSConfig;
			else
				pBifsConfig = new BIFSv2Config;
			pBifsConfig -> SetLength (nInfoLength * 8);
			pBifsConfig -> EncDec (&odDec, DECODE, NULL);
			if (pBifsConfig -> nodeIDbits > 0)
				m_nNodeIDbits = pBifsConfig -> nodeIDbits;
			if (pBifsConfig -> routeIDbits > 0)
				m_nNodeIDbits = pBifsConfig -> routeIDbits;
			if (pBifsConfig -> protoIDbits > 0)
				m_nProtoIDbits = pBifsConfig -> protoIDbits;
			m_bPixelMetric = pBifsConfig -> pixelMetric;
			m_nPixelWidth = pBifsConfig -> pixelWidth;
			m_nPixelHeight = pBifsConfig -> pixelHeight;
			m_bUse3DMeshCoding = pBifsConfig -> use3DMeshCoding;
			m_bUsePredictiveMFField = pBifsConfig -> usePredictiveMFField;
		}
		bResult = true;
	}
	catch (BifsException)
	{
		m_pExecutive -> Error (INVALID_BIFS_CONFIG);
		bResult = false;
	}
	delete pBifsConfig;
	return bResult;
}
#endif

/*-------------------------------------------------------------
Parse PROTOCode
-------------------------------------------------------------*/
void BIFSDecoder :: ParseProtoCode (Proto *pProto)
{
	/*-------------------------------------------------------------
	Parse the Proto's nested Protos
	-------------------------------------------------------------*/
	if (ParseInt (1, "isExtern"))
	{	MF<Url> ().Parse (this, NULL);
		return;
	}
	TRACEF ("{", 0);
	TRACENEWLINE;
	TRACEINDENT (1);

	/*-------------------------------------------------------------
	Parse nexted Proto, the Proto's nodes and the Routes
	-------------------------------------------------------------*/
	pProto -> ParseProtoList (this);
	pProto -> ParseNodes (this);
	if (ParseInt (1, "hasROUTEs"))	// if hasROUTEs
		ParseRoutesDescription (pProto);
	TRACEINDENT (-1);
	TRACEF ("}", 0);
	TRACENEWLINE;
}

/*-------------------------------------------------------------
Parse an SFString value
-------------------------------------------------------------*/
void BIFSDecoder :: ParseSFString (ZString &string)
{
	string.Cut (0);
	int nBits = ParseInt (5, "nbits");
	int nSize = ParseInt (nBits, "stringLength");

	while (--nSize >= 0)
		string += ParseInt (8, "value");
}

/*-------------------------------------------------------------
Parse a String value
-------------------------------------------------------------*/
void BIFSDecoder :: ParseString (ZString &string)
{
	string.Cut (0);
	char c;
	while ((c = ParseInt (8, "value")) != 0)
		string += c;
}
