/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Core
FILE:       InLine.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of InlineSource, the object that handle the
			streams of the Inline node (or the primary BIFS and OD
			streams)
-------------------------------------------------------------------------*/

#include "defs.h"
#include "InLine.h"
#include "Im1defs.h"
#include "BifsDec.h"
#include "ODImp.h"
#include "Executiv.h"
#include "AL.h"
#include "Channel.h"
#include "StreamCo.h"
#include "Clock.h"
#include "Stream.h"
#include "Nodes.h"
#include "Root.h"
#include "IPMP.h"
#include "ODDec.h"
#include "StrParser.h"

/*-----------------------------------------------------------------
Object constructor
-----------------------------------------------------------------*/

InlineSource :: InlineSource (
	Executive *pExecutive,
	MFGenericNode *pRootNode,
	ALManager *pAL
) :
	m_pExecutive (pExecutive),
	m_pRootNode (pRootNode),
	m_pAL (pAL)
{}

/*-----------------------------------------------------------------
A node of type StreamConsumer registers here when it is instantiated
-----------------------------------------------------------------*/
void InlineSource :: RegisterStreamConsumer (StreamConsumer *pNode)
{
	int nODid = pNode -> GetObjectDescriptorID ();
	int i = m_streamAssociation.Find (nODid);
	if (i < 0)
		(m_streamAssociation += nODid).m_pNode = pNode;
	else
	{	m_streamAssociation [i].m_pNode = pNode;
		if (m_streamAssociation [i].m_pOD != NULL)
			pNode -> SetStream (m_streamAssociation [i].m_pOD);
	}
}

/*-----------------------------------------------------------------
A node of type StreamConsumer unregisters here when it is destroyed
-----------------------------------------------------------------*/
void InlineSource :: UnregisterStreamConsumer  (StreamConsumer *pNode)
{
	for (int i = m_streamAssociation.GetLength (); --i >= 0;)
		if (m_streamAssociation [i].m_pNode == pNode)
		{	m_streamAssociation [i].m_pNode = NULL;
			if (m_streamAssociation [i].m_pOD == NULL)
				m_streamAssociation.DeleteFrom (i);
			return;
		}
}

/*-----------------------------------------------------------------
The BIFS decoder parsed a ReplaceScene command
-----------------------------------------------------------------*/
void InlineSource :: ReplaceScene (Root *pScene)
{
	delete m_pRootNode;
	m_pRootNode = pScene;
	if (pScene != NULL_OBJECT)
	        pScene -> ResetTime ();
	m_pExecutive -> ReplaceScene ();
}

/*-------------------------------------------------------------
Return the address of the dummy field containing the scene
-------------------------------------------------------------*/
NodeField *InlineSource :: GetSceneContainer ()
{
	return m_pRootNode;
}


/*-----------------------------------------------------------------
An update OD command was detected. Locate it in the OD list, or
add it if its not there. Once both the OD and the node which consumes
it are parsed, the proper channel(s) and decoder(s) are instantiated.
-----------------------------------------------------------------*/
void InlineSource :: UpdateOD (ObjectDescriptorImp *pOD)
{
	int nODid = pOD -> ObjectDescriptorID;
	pOD -> SetAL (m_pAL);
	pOD -> SetInline (this);
	int i = m_streamAssociation.Find (nODid);
	if (i < 0)
		(m_streamAssociation += nODid).m_pOD = pOD;
	else if (m_streamAssociation [i].m_pOD != NULL)
	{	m_pExecutive -> Error (OD_UPDATE_NOT_SUPPORTED);
		delete pOD;
	}
	else
	{	m_streamAssociation [i].m_pOD = pOD;
		StreamConsumer *pNode = m_streamAssociation [i].m_pNode;
		pNode -> SetStream (pOD);
	}
}

/*-----------------------------------------------------------------
A destructor for a StreamAssociation object
-----------------------------------------------------------------*/
InlineSource :: StreamAssociation :: ~StreamAssociation ()
{
	delete m_pOD;
}

/*-----------------------------------------------------------------
A remove OD command was deteced. Remove the OD from the list
-----------------------------------------------------------------*/
void InlineSource :: RemoveOD (int nODid)
{
	int i = m_streamAssociation.Find (nODid);
	if (i < 0)
		m_pExecutive -> Error (UNKNOWN_ODID);
	else
	{	delete m_streamAssociation [i].m_pOD;
		m_streamAssociation [i].m_pOD = NULL;
		if (m_streamAssociation [i].m_pNode == NULL)
			m_streamAssociation.DeleteFrom (i);
	}
}

/*-----------------------------------------------------------------
An Update ESD command was detected (not supported in this version)
-----------------------------------------------------------------*/
void InlineSource :: UpdateESD (int /*nODid*/, ES_Descriptor * /*esd*/)
{
	m_pExecutive -> Error (UPDATE_ESD_NOT_SUPPORTED);
}

/*-----------------------------------------------------------------
A remove ESD command was detected (not supported in this version)
-----------------------------------------------------------------*/
void InlineSource :: RemoveESD (int /*nODid*/, int /*nESDid*/)
{
	m_pExecutive -> Error (REMOVE_ESD_NOT_SUPPORTED);
}

/*-----------------------------------------------------------------
The destructor waits till parsers are not active and deletes them
-----------------------------------------------------------------*/
InlineSource :: ~InlineSource ()
{
	for (int i = m_clocks.GetLength (); --i >= 0;)
		delete m_clocks [i];
}

/*-------------------------------------------------------------
Return the clock reference of created by the stream whose id is
nClockID. Create one if none exists
-------------------------------------------------------------*/
ClockReference *InlineSource :: GetClock (int nClockID)
{
	for (int i = m_clocks.GetLength (); --i >= 0;)
		if (m_clocks [i] -> GetID () == nClockID)
			return m_clocks [i];
	return m_clocks += new ClockReference (nClockID);
}

/*-----------------------------------------------------------------
An update IPMPD command was detected. Locate it in the IPMPD list, or
add it if its not there.
-----------------------------------------------------------------*/
void InlineSource :: UpdateIPMPD (IPMP_Descriptor *pIPMPD)
{
	int nIPMPid = pIPMPD -> IPMP_DescriptorID;
	int i = m_IPMPDs.Find (nIPMPid);
	if (i < 0)
		m_IPMPDs += pIPMPD;
	else
	{	delete m_IPMPDs [i];
		m_IPMPDs [i] = pIPMPD;
	}
}

/*-----------------------------------------------------------------
Find an object in the IPMP list
-----------------------------------------------------------------*/
int InlineSource :: IPMPDList :: Find (int nIPMPid) const
{
	for (int i = GetLength (); --i >= 0;)
		if (nIPMPid == (*this) [i] -> IPMP_DescriptorID)
			return i;
	return -1;
}

/*-------------------------------------------------------------
Execute the RemoveIPMPD command. The IPMP_Descriptor is removed from
the list of IPMP descriptors
-------------------------------------------------------------*/
void InlineSource :: RemoveIPMPD (int nIPMPid)
{
	int i = m_IPMPDs.Find (nIPMPid);
	if (i >= 0)
		m_IPMPDs.DeleteFrom (i);
}

/*-------------------------------------------------------------
Return a corresponding IPMP_Descriptor, if available
-------------------------------------------------------------*/
IPMP_Descriptor * InlineSource :: GetIPMP_Descriptor (int nIPMPid)
{
	int i = m_IPMPDs.Find (nIPMPid);
	if (i >= 0)
		return m_IPMPDs [i];
	return NULL;
}
