/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player/ Mux
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  OD Decoder/ Encoder
FILE:       sdl.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementatioan of SDL encoding/parsing
-------------------------------------------------------------------------*/

#include "defs.h"
#include "sdl.h"
#include "ODCoder.h"
#include "Im1Defs.h"

#define pBifs pCoder

/*-------------------------------------------------------------
Decoding / encoding of a descriptor. The length of the descriptor
is parsed/stored, the descriptor content is decoded/encoded, and the
expansion, if exists, is skipped
-------------------------------------------------------------*/
void BaseDescriptor :: EncDec (
	ODCoder *pCoder,
	OPTYPE op,
	BaseDescriptor *pParent
)
{
	int nLengthFieldWidth = 0;

	/*-------------------------------------------------------------
	Encode/decode the length
	-------------------------------------------------------------*/
	switch (op) {
	case ENCODE:
		pCoder -> Align ();
		pCoder -> EncodeInt (GetTag (), 8, GetName ());
		pCoder -> StoreDescLength (m_nLength);
		break;
	case DECODE:
		TRACEF ("%s {", GetName ());
		TRACENEWLINE;
		TRACEINDENT (1);
		if (m_nLength == 0)	// parse length only if not known already
			m_nLength = pCoder -> ReadDescLength (nLengthFieldWidth);
		/*-------------------------------------------------------------
		The total length of the descriptor is its declared length + 
		width of its length fields.
		-------------------------------------------------------------*/
		if (pParent != NULL_DESC)
			pParent -> AddLength (- m_nLength - nLengthFieldWidth);
		break;
	case CALC_LENGTH:
		m_nLength = 0;
		break;
	}

	/*-------------------------------------------------------------
	Encode/decode the content and skip the expansion.
	-------------------------------------------------------------*/
	DescEncDec (pCoder, op);

	switch (op) {
	case DECODE:
		TRACEINDENT (-1);
		TRACEF ("}", 0);
		TRACENEWLINE;
		pCoder -> SkipBytes (m_nLength / 8);
		break;
	case ENCODE:
		pCoder -> Align ();
		break;
	case CALC_LENGTH:
		if (pParent != NULL_DESC)
			pParent -> AddLength (m_nLength + 8 + CalcLengthFieldWidth ());
		break;
	}
}

/*-------------------------------------------------------------
The default encoding/decoding/size calculation of descriptor's content.
The operation is done on all the fields of the descriptor
-------------------------------------------------------------*/
void BaseDescriptor :: DescEncDec (ODCoder *pCoder, OPTYPE op)
{
	int nCount = GetFieldCount ();
	for (int i = 0; i < nCount; i++)
		GetField (i) -> EncDec (pCoder, this, op);
}

/*-----------------------------------------------------------------
Find address of a field in the descriptor, given its index
-----------------------------------------------------------------*/
SDLField *BaseDescriptor :: GetField (int nIndex)
{
	return
		(SDLField *) (((char *) this) + GetFieldTable () [nIndex].m_offset);
}

/*-----------------------------------------------------------------
Look on the field table to find the textual name of a field
-----------------------------------------------------------------*/
LPCSTR BaseDescriptor :: GetFieldName (SDLField *pField) const
{
#ifdef TRACE_ENCODING
	const ODFieldTable *pTable = GetFieldTable ();
	if (pTable == NULL)
		return NULL;
	for (int i = 0; ;i++)
	{	const ODFieldTable *pRow = pTable + i;
		if (pField == (SDLField *) (((char *) this) + pRow -> m_offset))
			return pRow -> m_szName;
	}
#else
	pField;	// avoid compiler warning
	return NULL;
#endif
}

/*-------------------------------------------------------------
Calculate how many bits will be required for a the length field of this
descriptor (the length is stored in one or more bytes while bit 0 zero
of each byte is an escape indicator).
-------------------------------------------------------------*/
int BaseDescriptor :: CalcLengthFieldWidth ()
{
	int i, n = (m_nLength + 7) / 8;
	for (i = 1; (n >>= 7) != 0; ++i);
	return i * 8;
}

/*-------------------------------------------------------------
Instantiate an extension of an abstract class. pTable points to the
table where the matching code to the given tag can be found
-------------------------------------------------------------*/
BaseDescriptor *BaseDescriptor :: CreateExtension (
	int nTag,
	const ODExtensionTable *pTable,
	int nLength
)
{
	for (int i = 0; i < nLength; ++i)
		if (pTable [i].m_tag == nTag)
			return SDLField :: CreateClass (pTable [i].m_code);
	return NULL;
}

/*-------------------------------------------------------------
Instantiate an extension of an abstract class when the extension
class name is given.
-------------------------------------------------------------*/
BaseDescriptor *BaseDescriptor :: CreateExtension (
	LPCSTR szClassName,
	const ODExtensionTable *pTable,
	int nLength
)
{
	for (int i = 0; i < nLength; ++i)
	{	BaseDescriptor *p = SDLField :: CreateClass (pTable [i].m_code);
		if (p != NULL_DESC && !stricmp (p -> GetName (), szClassName))
			return p;
		delete p;
	}
	return NULL_DESC;
}

/*-------------------------------------------------------------
Decoding/encoding/length calucaltion of an integer
-------------------------------------------------------------*/
void SDLInteger :: EncDec (
	ODCoder *pCoder,
	BaseDescriptor *pDesc,
	OPTYPE op
)
{
	switch (op) {
	case DECODE:
		if (m_bAligned)
			pDesc -> AddLength (- pCoder -> Align ());
		m_nValue = pCoder -> ParseInt (m_nBits, pDesc -> GetFieldName (this));
		pDesc -> AddLength (-m_nBits);
		TRACEF ("%s ", pDesc -> GetFieldName (this));
		TRACEF ("%d", m_nValue);
		TRACENEWLINE;
		break;
	case ENCODE:
		if (m_bAligned)
			pCoder -> Align ();
		pCoder -> EncodeInt (m_nValue, m_nBits, pDesc -> GetFieldName (this));
		break;
	case CALC_LENGTH:
		if (m_bAligned)
			pDesc -> AlignLength ();
		pDesc -> AddLength (m_nBits);
		break;
	}
}

/*-------------------------------------------------------------
Decode an integer whose bit size varies
-------------------------------------------------------------*/
void SDLInteger :: EncDecVarSize (
	ODCoder *pODCoder,
	BaseDescriptor *pDesc,
	OPTYPE op,
	int nBits
)
{
	m_nBits = nBits;
	if (nBits > 0)
		EncDec (pODCoder, pDesc, op);
}

/*-------------------------------------------------------------
Decoding/encoding/length calucaltion of a float
-------------------------------------------------------------*/
void SDLFloat :: EncDec (
	ODCoder *pCoder,
	BaseDescriptor *pDesc,
	OPTYPE op
)
{
	switch (op) {
	case DECODE:
		m_fValue = pCoder -> ParseFloat (pDesc -> GetFieldName (this));
		pDesc -> AddLength (-32);
		TRACEF ("%s ", pDesc -> GetFieldName (this));
		TRACEF ("%f", m_fValue);
		TRACENEWLINE;
		break;
	case ENCODE:
		pCoder -> EncodeFloat (m_fValue, pDesc -> GetFieldName (this));
		break;
	case CALC_LENGTH:
		pDesc -> AddLength (32);
		break;
	}
}

/*-------------------------------------------------------------
The constructor of an integer array
-------------------------------------------------------------*/
SDLIntegerArray :: SDLIntegerArray (int nBits)
{
	StoreElementSize (nBits);
}

/*-------------------------------------------------------------
Calculate size of elements
-------------------------------------------------------------*/
void SDLIntegerArray :: StoreElementSize (int nBits)
{
	m_nBits = nBits;
	if (nBits <= 8)
		m_nBytes = 1;
	else if (nBits <= 16)
		m_nBytes = 2;
	else
		m_nBytes = 4;
}

/*-------------------------------------------------------------
Return the ith element, converted to int
-------------------------------------------------------------*/
int SDLIntegerArray :: operator [] (int i) const
{
	const LPBYTE p = (const LPBYTE) (LPCSTR) m_array;
	switch (m_nBytes) {
	case 1:
		return p [i];
	case 2:
		return ((WORD *) p ) [i];
	}
	return ((DWORD *) p) [i];
}

/*-------------------------------------------------------------
Append one element to the array. If index == -1, the element is 
appended at the end of the array.
-------------------------------------------------------------*/
void SDLIntegerArray :: InsertElement (int n, int index)
{
	if (index < 0)
		index = m_array.GetLength () / m_nBytes;
	m_array.InsertAt (index * m_nBytes, 0, m_nBytes);
	unsigned char *p = &m_array [0];
	switch (m_nBytes) {
	case 1:
		p [index] = n;
		return;
	case 2:
		((WORD *) p ) [index] = n;
		return;
	}
	((DWORD *) p) [index] = n;
}

/*-------------------------------------------------------------
Default Encoding/decoding/length calculation of an integer array.
This functions calls EncDecArray. For decoding, it first calculates
the array size according to the remaining descriptor size.
-------------------------------------------------------------*/
void SDLIntegerArray :: EncDec (
	ODCoder *pCoder,
	BaseDescriptor *pDesc,
	OPTYPE op
)
{
	int nArrayLength = GetLength ();
	if (op == DECODE)
		nArrayLength = pDesc -> GetLength () / m_nBits;
	EncDecArray (pCoder, pDesc, op, nArrayLength);
}

/*-------------------------------------------------------------
Encoding/decoding/length calculation of an integer array.
-------------------------------------------------------------*/
void SDLIntegerArray :: EncDecArray (
	ODCoder *pCoder,
	BaseDescriptor *pDesc,
	OPTYPE op,
	int nLength
)
{
	switch (op) {
	case DECODE:
		TRACEF ("%s [", pDesc -> GetFieldName (this));
		break;
	case CALC_LENGTH:
		pDesc -> AddLength (nLength * m_nBits);
		return;
	}

	/*-------------------------------------------------------------
	Encode/decode the elements
	-------------------------------------------------------------*/
	for (int i = 0; i < nLength; i++)
		if (op == DECODE)
		{	if (i > 0)
				TRACEF (", ", 0);
			int n = pCoder -> ParseInt (m_nBits, pDesc -> GetFieldName (this));
			InsertElement (n);
			TRACEF ("%d", n);
		}
		else
			pCoder -> EncodeInt ((*this) [i], m_nBits, pDesc -> GetFieldName (this));

	if (op == DECODE)
	{	TRACEF ("]", 0);			
		TRACENEWLINE;
		pDesc -> AddLength (- (nLength * m_nBits));
	}
}

/*-------------------------------------------------------------
Encoding/decoding/length calculation of one element of an integer array.
-------------------------------------------------------------*/
void SDLIntegerArray :: EncDecArrayElement (
	ODCoder *pCoder,
	BaseDescriptor *pDesc,
	OPTYPE op,
	int nIndex
)
{
	int n;
	switch (op) {
	case DECODE:
		TRACEF ("%s ", pDesc -> GetFieldName (this));
		n = pCoder -> ParseInt (m_nBits, pDesc -> GetFieldName (this));
		InsertElement (n, nIndex);
		TRACEF ("%d", n);
		TRACENEWLINE;
		pDesc -> AddLength (-m_nBits);
		return;
	case ENCODE:
		pCoder -> EncodeInt ((*this) [nIndex], m_nBits, pDesc -> GetFieldName (this));
		return;
	case CALC_LENGTH:
		pDesc -> AddLength (m_nBits);
		return;
	}
}

/*-------------------------------------------------------------
Encoding/decoding/length calculation of a string. A string is an
array of bytes prededed by a length indicator byte.
-------------------------------------------------------------*/
void SDLString :: EncDec (
	ODCoder *pCoder,
	BaseDescriptor *pDesc,
	OPTYPE op
)
{
	EncDecString (pCoder, pDesc, op, false);
}

/*-------------------------------------------------------------
Encoding/decoding/length calculation of a string. A string is an
array of bytes prededed by a length indicator byte. The string may be
a UTF16 string.
-------------------------------------------------------------*/
void SDLString :: EncDecString (
	ODCoder *pCoder,
	BaseDescriptor *pDesc,
	OPTYPE op,
	int bIsUTF16
)
{
	int nLength = m_array.GetLength ();
	switch (op) {
	case DECODE:
		TRACEF ("%s \"", pDesc -> GetFieldName (this));
		nLength = pCoder -> ParseInt (8, "string-size");
		break;
	case ENCODE:
		pCoder -> EncodeInt (nLength, 8, "string-size");
		break;
	case CALC_LENGTH:
		pDesc -> AddLength (nLength * (1 + bIsUTF16) * 8 + 8);
		return;
	}

	/*-------------------------------------------------------------
	Encode/decode the elements
	-------------------------------------------------------------*/
	EncDecStringElements (pCoder, pDesc, op, nLength * (1 + bIsUTF16));
}

/*-------------------------------------------------------------
Encoding/decoding/length calculation of a long string. A string is an
array of bytes prededed by a length indicator byte. When the byte is
255 it is followed by another length bytes, whose value should be
added to the the calculated length, etc.
-------------------------------------------------------------*/
void SDLString :: EncDecLongString (
	ODCoder *pCoder,
	BaseDescriptor *pDesc,
	OPTYPE op,
	int bIsUTF16
)
{
	int nLength = m_array.GetLength ();
	int n;
	switch (op) {
	case DECODE:
		TRACEF ("%s \"", pDesc -> GetFieldName (this));
		do
			nLength += (n = pCoder -> ParseInt (8, "string-size"));
		while (n == 255);
		break;
	case ENCODE:
		for (n = nLength; n >= 255; n -= 255)
			pCoder -> EncodeInt (255, 8, "string-size");
		pCoder -> EncodeInt (n, 8, "string-size");
		break;
	case CALC_LENGTH:
		pDesc -> AddLength (nLength * 8 * (1 + bIsUTF16) + 8 * (1 + nLength / 255));
		return;
	}

	EncDecStringElements (pCoder, pDesc, op, nLength * (1 + bIsUTF16));
}

/*-------------------------------------------------------------
Encode/Decode the elements of a string (called after the size field
has been processed
-------------------------------------------------------------*/
void SDLString :: EncDecStringElements (
	ODCoder *pCoder,
	BaseDescriptor *pDesc,
	OPTYPE op,
	int nLength
)
{
	for (int i = 0; i < nLength; i++)
		if (op == DECODE)
		{	char c = pCoder -> ParseInt (m_nBits, pDesc -> GetFieldName (this));
			m_array += c;
			TRACEF ("%c", c);
		}
		else
			pCoder -> EncodeInt (m_array [i], 8, pDesc -> GetFieldName (this));

	if (op == DECODE)
	{	TRACEF ("\"", 0);			
		TRACENEWLINE;
		pDesc -> AddLength (- (nLength * 8 + 8));
	}
}

/*-------------------------------------------------------------
Encoding/decoding/length calculation of an array of strings.
Since there is no default encoding to complete string array, we just
code it as a placeholder.
-------------------------------------------------------------*/
void SDLStringArray :: EncDec (
	ODCoder *,
	BaseDescriptor *,
	OPTYPE
)
{}

/*-------------------------------------------------------------
Encode/decoder an integer array when the array length is given
-------------------------------------------------------------*/
void SDLStringArray :: EncDecStringArray (
	ODCoder *pCoder,
	BaseDescriptor *pDesc,
	OPTYPE op,
	int length,
	int bIsUTF16
)
{
	for (int i = 0; i < length; i++)
		EncDecStringArrayElement (pCoder, pDesc, op, i, bIsUTF16);
}

/*-------------------------------------------------------------
Encoding/decoding/length calculation of one elemement in an array of strings.
-------------------------------------------------------------*/
void SDLStringArray :: EncDecStringArrayElement (
	ODCoder *pCoder,
	BaseDescriptor *pDesc,
	OPTYPE op,
	int nIndex,
	int bIsUTF16
)
{
	ZString &rString = m_array.ElementAt (nIndex);
	int nLength = rString.GetLength ();
	switch (op) {
	case DECODE:
		TRACEF ("%s \"", pDesc -> GetFieldName (this));
		nLength = pCoder -> ParseInt (8, "string-size");
		break;
	case ENCODE:
		pCoder -> EncodeInt (nLength, 8, "string-size");
		break;
	case CALC_LENGTH:
		pDesc -> AddLength (nLength * (1 + bIsUTF16) * 8 + 8);
		return;
	}

	/*-------------------------------------------------------------
	Encode/decode the elements
	-------------------------------------------------------------*/
	for (int i = 0; i < nLength * (1 + bIsUTF16); i++)
		if (op == DECODE)
		{	char c = pCoder -> ParseInt (8, pDesc -> GetFieldName (this));
			rString += c;
			TRACEF ("%c", c);
		}
		else
			pCoder -> EncodeInt (rString [i], 8, pDesc -> GetFieldName (this));

	if (op == DECODE)
	{	TRACEF ("\"", 0);			
		TRACENEWLINE;
		pDesc -> AddLength (- (nLength * (1 + bIsUTF16) * 8 + 8));
	}
}

/*-------------------------------------------------------------
SDLBaseClass destructor
-------------------------------------------------------------*/
SDLBaseClass :: ~SDLBaseClass ()
{
	if (m_pClass && !m_pClass -> IsProtected ())
		delete m_pClass;
}

/*-------------------------------------------------------------
Encoding/decoding/length calculation of a class field
-------------------------------------------------------------*/
void SDLBaseClass :: EncDec (
	ODCoder *pCoder,
	BaseDescriptor *pDesc,
	OPTYPE op
)
{
	if (op == DECODE)
	{	
		m_pClass = DecodeDescriptor (
			pCoder,
			pDesc,
			pDesc -> GetFieldName (this),
			-1
		);
		if (m_pClass == NULL_DESC && m_nMin > 0)
			pCoder -> Error (MISSING_DESCRIPTOR);
	}
	else if (m_pClass)
		m_pClass -> EncDec (pCoder, op, pDesc);
}

/*-------------------------------------------------------------
Parse one descriptor (may be stand-alone or part of an array). In case
of an array, 'count' shows how many descriptors we have already parsed.
The function returns a pointer to the parsed descriptor, or NULL if
a matching descriptor was not encountered.
-------------------------------------------------------------*/
BaseDescriptor *SDLBaseClass :: DecodeDescriptor (
	ODCoder *pCoder,
	BaseDescriptor *pParent,
	LPCSTR szFieldName,
	int index		// used to trace array members
)
{
	szFieldName;	// avoid a warning in release mode.

	/*-------------------------------------------------------------
	Parse the tag, unless it was already parsed before
	-------------------------------------------------------------*/
	int nTag = pParent -> GetNextTag ();
	if (nTag < 0 && pParent -> GetLength () > 0)
	{	pParent -> AddLength (-pCoder -> Align ());
		nTag = pCoder -> ParseInt (8, "tag");
		pParent -> AddLength (-8);
	}

	/*-------------------------------------------------------------
	Check if the tag is within the tag range of the expected descriptor.
	In any case check if we don't have too few or too many descriptors
	of this type.
	-------------------------------------------------------------*/
	if (!IsMyTag (nTag))
	{	pParent -> SetNextTag (nTag);
		return NULL_DESC;
	}
	pParent -> SetNextTag (-1);

	/*-------------------------------------------------------------
	Create the descriptor and parse it
	-------------------------------------------------------------*/
	BaseDescriptor *pDesc = CreateDescriptor (nTag);
	if (pDesc != NULL_DESC)
	{	if (index <= 0)
			TRACEF ("%s ", szFieldName);
		if (index == 0)
		{	TRACEF ("[", NULL);
			TRACENEWLINE;
			TRACEINDENT (1);
		}
		pDesc -> EncDec (pCoder, DECODE, pParent);
	}
	return pDesc;
}

/*-------------------------------------------------------------
The destructor of SDLBaseClassArray deletes all the array members
-------------------------------------------------------------*/
SDLBaseClassArray :: ~SDLBaseClassArray ()
{
	for (int i = 0; i < m_array.GetLength (); i++)
	{	if (m_array [i] != NULL_DESC && !m_array [i] -> IsProtected ())
			delete m_array [i];
	}
}

/*-------------------------------------------------------------
Encoding/decoding/length calculation of a class array field.
-------------------------------------------------------------*/
void SDLBaseClassArray :: EncDec (
	ODCoder *pCoder,
	BaseDescriptor *pDesc,
	OPTYPE op
)
{
	if (op == DECODE)
	{	BaseDescriptor *p;
		for (
			int nCount = 0;
			(p = DecodeDescriptor (
					pCoder,
					pDesc,
					pDesc -> GetFieldName (this),
					nCount
				 )
			) != NULL_DESC;
			nCount++
		)
		{	
			if (nCount >= m_nMax)
				pCoder -> Error (TOO_MANY_DESCRIPTORS);
			m_array += p;
		}
		if (nCount < m_nMin)
			pCoder -> Error (MISSING_DESCRIPTOR);
		if (nCount > 0)
		{	TRACENEWLINE;
			TRACEINDENT (-1);
			TRACEF ("]", NULL);
			TRACENEWLINE;
		}
	}			
	else
		for (int i = 0; i < m_array.GetLength (); i++)
			m_array [i] -> EncDec (pCoder, op, pDesc);
}

/*-----------------------------------------------------------------
SDLField includeS a static member which is an array of factory
objects which used to instantiate OD classes
-----------------------------------------------------------------*/
ODFactoryBase *SDLField :: m_odFactory [MPEG4_DESC_TABLE_LENGTH];

/*-----------------------------------------------------------------
Stores a pointer to a factory object in the factory array. This
function is called from factory object (ODFactory) constructor.
A factory object for each descriptor is instantiated through the use of the
OD_IMPLEMENT_DESCRIPTOR_START macro.
-----------------------------------------------------------------*/
void SDLField :: StoreFactory (int nCode, ODFactoryBase *pFactory)
{
	assert (nCode >= 0 && nCode < MPEG4_DESC_TABLE_LENGTH);
	m_odFactory [nCode] = pFactory;
}

/*-----------------------------------------------------------------
Use the factory array to instantiate a proper OD class.
-----------------------------------------------------------------*/
BaseDescriptor *SDLField :: CreateClass (int nCode)
{
	if (nCode < 0 || nCode >= MPEG4_DESC_TABLE_LENGTH)
		return NULL;
	if (m_odFactory [nCode] != NULL)
		return m_odFactory [nCode] -> CreateDescriptor ();
	return NULL;
}
