/* ---------------------------------------------------------------------------------------
	PROJECT :	RealTime FlexMux MPEG4 

	FILE :		FmxStruct.h
	OVERVIEW :	Definitions of specific classes for Fifo management 
				& definition of Packet class ... 

	LAST CHANGES :	98/11/13 : KR : creation 

 -----------------------------------------------------------------------------------------*/  

#ifndef _FMXSTRUCT_H_
#define _FMXSTRUCT_H_

#include "ztl.h"

/*-----------------------------------------------------------------------------------------
Definition of a generic Liste class ... 
------------------------------------------------------------------------------------------*/
template <class T> class Liste {

 public:
	 Liste () : First(NULL), Current(NULL), Last(NULL), nbElem(0) {};		
		// constructor with empty initialisation ... 
	 Liste (Liste &pList);
		// copy constructor ... 
	~Liste () { Begin(); while (! IsAtTheEnd()) Remove (); } 		
		// destructor free the reserved memory for the contained elements
 
	void Begin () { Current = First; };
		// let the first element of the list be the current one ... 
	void End () { Current = Last; };		
		// let the last element of the list be the current one ... 
	void GoNext () {if (Current!= NULL) Current = Current->next; }; 		
		// let the next element be the current one ... 
	
	T GetElem ();		// return the current element of the list 

	void Add (const T elt); 	 // add an element at the end of the list ... 
	void Remove ();			 // Remove the current element from the list ... 
	 
	BOOL IsEmpty () { return (First == NULL); };	
		// return true if the list is empty 
	BOOL IsAtTheEnd () { return (Current==NULL); };	
		// return true if the end of the list is reached,
		// i.e.: the last element of the list come before the current elem ... 

	int Size () {return nbElem;}
		// return the nber of elements contained into the list 


 protected:
	
class ListElem { 
		public : 
			T elt;
			ListElem * next;
			ListElem(T elem, ListElem * lstSuite)
					:  elt(elem), next(lstSuite){};
			~ListElem() {};
	 };

	ListElem * First;		// first element of the list 
	ListElem * Last;		// last element of the list 
	ListElem * Current;		// current element in the list 
	int nbElem;				// number of elements contained into the list 
};


// Copy Constructor ...
template <class T> Liste<T>::Liste (Liste<T> &pList)
{ 
	First = NULL; Last = NULL; Current = NULL; nbElem = 0;
	pList.Begin ();
	while (! pList.IsAtTheEnd()) { Add (pList.GetElem()); pList.GoNext (); } 
};

// add a new element at the end of the list ... 
template <class T> void Liste<T>::Add(const T elt)
{ 
	ListElem *pNewElem = new ListElem(elt,NULL);

	if (First == NULL) 
		First = pNewElem; 
	else 
		Last->next = pNewElem; 
  
	nbElem ++; Last = pNewElem;
};

// Return the current element in the list ... 
template <class T> T Liste<T>::GetElem()
{ 
	return ((Current != NULL) ? Current->elt : (T ) NULL);
};

// remove the current element from  the list ... 
template <class T> void Liste<T>::Remove()
{ 
	ListElem * PrevElem = NULL; 
	ListElem * toDelete = NULL;
  
	if (Current != NULL) 
	{
		toDelete = Current;  nbElem --;
		
		if (Current == First) 
		// special case ...
		{ 
			First = Current->next;
			// special case : new list is empty ...
			if (Current == Last) { Last = Current->next;}
			
			Current = Current->next;
		}
		else 
		{	// general case : search the previous element from current ...
		     PrevElem = First;
		     while (PrevElem->next != Current)  PrevElem = PrevElem->next; 
			 PrevElem->next = Current->next; // remove Current from the list ...
	     
			if (Current == Last) 
				// special case : let the last elem be the current one ...
				{ Last = PrevElem; Current = PrevElem; }
			else 
				{Current = Current->next;}
		};
		
		delete (toDelete);
	}
};



/*-----------------------------------------------------------------------------------------
Definition of a generic class to manage a Fifo ... 
------------------------------------------------------------------------------------------*/

template <class T> class Fifo
{
	public : 
		Fifo (int nMaxSize) : m_nMaxSize(nMaxSize) {m_pList = new Liste<T *> ();}
			// constructor : create a new empty list & initialise the Fifo's maximum size
		Fifo (Fifo<T> &pFifo) ;
			// copy constructor 
		~Fifo() {delete (m_pList);} 
			// destructor : delete the list ... 

		int Size () {return (m_pList->Size ());};
			// return the nber of elements contained into the Fifo ...
		T * ReadIn ();
			// return the first element of the List & remove it ... 
		void WriteIn (T *pElem) ;
			// add a new element at the end of the list ...

		BOOL IsFull () {return (m_pList->Size () == m_nMaxSize);}; 
			// return True if the Fifo is Full ...
		BOOL IsEmpty () {return (m_pList->IsEmpty ());}
			// return True if the Fifo is empty ... 

	protected : 
		Liste<T*> *m_pList;	// List of the element contained in the Fifo 
		int m_nMaxSize;		// Maximum number of elements in the Fifo 

	private : 
		Fifo ();
};



// Fifo Constructor : Copy Constructor ... 

template <class T> Fifo<T>::Fifo (Fifo<T> &pFifo)  
{
	if ((pFifo.m_nMaxSize == 0) || (pFifo.m_pList->IsEmpty()))
		m_pList = new Liste<T *> ();
	else 
		m_pList = new Liste<T *> (*(pFifo.m_pList));
}





/*----------------------------------------------------------------------------------------
Fifo Consultation : Fifo Element Consommation 
 Return NULL if there is no element in the buffer 
 Return the first element in the buffer(if the buffer is not empty) and its length 
-----------------------------------------------------------------------------------------*/
template <class T> T * Fifo<T>::ReadIn ()
{
	ASSERT (m_pList != NULL);
	T *pElem = NULL;
	
	// check if the fifo contains at least one element... 	
	if (!(m_pList->IsEmpty())) 
	{
		// Read the first element and delete it from the list ...
		m_pList->Begin();
		pElem = m_pList->GetElem ();
		m_pList->Remove ();	
	}
	
	return pElem;
};

/*----------------------------------------------------------------------------------------
Fifo Consultation : Fifo Element Production 
 Put a new element in the fifo, 
 return True if the insertion is Okay, and return False else...
-----------------------------------------------------------------------------------------*/		
template <class T> void Fifo<T>::WriteIn (T *pElem )
{
	T *pToDelete;

	ASSERT (m_pList != NULL);

	// check if the fifo is full ... 
	if (IsFull())
	{
		// delete the first element of the list (to make place for the new one ...)
		pToDelete = ReadIn ();
		delete (pToDelete);
	}

	// add the new element at the end of the list ...
//	m_pList->Begin ();
	m_pList->Add (pElem);	
};


/*-----------------------------------------------------------------------------------------
Definition of a structure that represent a packet ... 
------------------------------------------------------------------------------------------*/
class Packet {
	public :	
		Packet (const BYTE *pStream, const int nDataLength);
		Packet () : m_ucDataStream(NULL), m_nDataLength(0) {};
		~Packet () { delete [] m_ucDataStream;}

		Packet & operator = (const Packet &p) ;
		BOOL operator == (const Packet &p) const;

		BYTE * m_ucDataStream; 
		int m_nDataLength;
};


/*-----------------------------------------------------------------------------------------
Definition of a default value for the Fifo Size ... 
------------------------------------------------------------------------------------------*/
const int DEFAULT_FIFO_SIZE = 100 ;



#endif 
