/***********************************************************************
    This software module  was  originally  developed by Zvi  Lifshitz
    (Triton R&D Ltd.) and edited by Mauro Rossi (CSELT),
    Guido Franceschini (CSELT) and Michael Stepping (FTK)
    in the course of development of the MPEG-4 DMIF (ISO/IEC 14496-6)
    standard.  This software module is an implementation of a part of
    one or more MPEG-4  DMIF  (ISO/IEC 14496-6)  tools  as  specified
    by  the  MPEG-4 DMIF (ISO/IEC 14496-6)  standard.  ISO/IEC  gives
    users of  the  MPEG-4  DMIF  (ISO/IEC 14496-6)  free  license  to
    this software module or modifications thereof for use in hardware
    or software products claiming  conformance  to  the  MPEG-4  DMIF 
    (ISO/IEC 14496-6). Those intending to use this software module in
    hardware or software  products  are  advised  that  its  use  may
    infringe   existing  patents.  The  original  developer  of  this
    software module and his/her company, the subsequent  editors  and
    their  companies,  and  ISO/IEC have no liability for use of this
    software module or modifications thereof  in  an  implementation.
    Copyright  is not released for non MPEG-4 DMIF  (ISO/IEC 14496-6)
    conforming  products. Triton R&D Ltd.
    retains full  right to use the code for  its  own purpose, assign
    or donate the code to a third party and to inhibit third  parties
    from  using  the  code  for  non  MPEG-4  DMIF  (ISO/IEC 14496-6)
    conforming products. This copyright notice must  be  included  in
    all copies or derivative works. Copyright  1998.
************************************************************************/

// Contact: zvil@csi.com

/**************************************************************************
 * Project Im1Player
 * INTERNATIONAL STANDARD ORGANIZATION
 * ISO/IEC JTC1/SC29/WG11
 * 
 * SUBSYSTEM: DMIF
 * FILE:      DMIFClientFilter.cpp
 * AUTHOR:    Zvi Lifshitz (Triton R&D Ltd.), Mauro Rossi (CSELT)
 * OVERVIEW:  Implementation of FileTrivInstance, the class that implements
 *			  local storage DMIF.
 * 
/***********************************************************************/

#include "FileTrivInstance.h"

/*-------------------------------------------------------------
The exported function that instantiates the proper decoder
-------------------------------------------------------------*/
extern "C" ServiceClient *CreateInstance ()
{
	return new FileTrivInstance;
}

/*-----------------------------------------------------------------
Constructor
-----------------------------------------------------------------*/
FileTrivInstance :: FileTrivInstance (int nID) :
	m_smpWaitForChannel (1, 0)
{
	m_nID = nID;
	m_nChannelWaitedFor = -1;
	m_fInput = NULL;
	m_bRewinding = FALSE;
}

/*-----------------------------------------------------------------
The object destructor makes sure the service is closed. Then it deletes
all auxiliary file objects.
-----------------------------------------------------------------*/
FileTrivInstance :: ~FileTrivInstance ()
{
	m_smpActive.P ();
	if (m_fInput)
		fclose (m_fInput);
	m_smpActive.V ();
	for (int i = m_auxFiles.GetLength (); --i >= 0;)
		delete m_auxFiles [i];
}

/*-----------------------------------------------------------------
Open the file, load the association table, start the file thread,
and notify the application
-----------------------------------------------------------------*/
RESULT FileTrivInstance :: DPI_ServiceAttach_Req(LPCSTR szURL, const BYTE*, int)
{
	ZString initialOD;

	RESPONSE response = Connect (szURL, &initialOD);
	GetSink () -> DPI_ServiceAttach_Cnf (
		(const BYTE *) (LPCSTR) initialOD,
		initialOD.GetLength (),
		response
	);
	if ( response != RESPONSE_OK )
		return ( RESULT_ERROR ) ;
	return ( RESULT_OK ) ;
}

/*-----------------------------------------------------------------
Open the file, load the association table, and start the file thread.
Return a success code.
-----------------------------------------------------------------*/
RESPONSE FileTrivInstance :: Connect (
	LPCSTR szURL,
	ZString *initialOD
)
{
	int c, n;

	/*-------------------------------------------------------------
	If the URL starts with "d-trif:" it is followed by the file name
	-------------------------------------------------------------*/
	if (!_strnicmp (szURL, "d-trif:", 7))
		szURL += 7;
	m_url = szURL;

	/*-------------------------------------------------------------
	If the file was opened before, rewind it. Otherwise open it.
	-------------------------------------------------------------*/
	if (m_fInput)
	{	rewind (m_fInput);
		m_channelMap.Clear ();
	}
	else if ((m_fInput = fopen (szURL, "rb")) == NULL)
		return RESPONSE_URL_NOT_FOUND;

	try
	{
		/*-----------------------------------------------------------------
		Read the initial OD
		-----------------------------------------------------------------*/
		c = GetByte ();
		n = MAKEWORD (GetByte (), c);
		while (--n >= 0)
		{	c = GetByte ();
			if (initialOD != NULL)
				*initialOD += c;
		}
		
		/*-----------------------------------------------------------------
		Read the association map
		-----------------------------------------------------------------*/
		m_nNotReady = n = GetByte ();
		while (--n >= 0)
		{	
			int nESid = 0;
			c = GetByte ();
			nESid = MAKEWORD (GetByte (), c);
			/*-------------------------------------------------------------
			If the initialOD is requested, then this is the main file,
			and channel map entries consists of several entries each with
			its own ESid
			If the initialOD is not requested, then this is an
			auxiliary file, and channel map entries consists of several
			entries each with its own string of the form <filename>:<esid>
			-------------------------------------------------------------*/
			if (initialOD)
			{	
				ChannelAssociation aChannel(nESid);
				m_channelMap += aChannel;
			}
			else
			{	
				ChannelAssociation aChannel;
				
				// temporary build the new string
				char *szTmpURL = new char[ strlen( szURL + 6 ) ] ;
				sprintf(szTmpURL,"%s:%d", szURL, nESid);
				
				aChannel.m_pData = new BYTE[ strlen( szTmpURL ) + 1 ];
				strcpy( (char *)aChannel.m_pData, szTmpURL ) ;
				aChannel.m_nDataLen = strlen( szTmpURL ) + 1 ;

				m_channelMap += aChannel;
			}
		}
	}
	catch (FileException)
	{
		fclose (m_fInput);
		return RESPONSE_REMOTE_ERROR;
	}

	m_bActive = TRUE;
  	if (ZThread<FileTrivInstance> (
			FileThread,
			this,
			THREAD_PRIORITY_BELOW_NORMAL
		).Begin () == -1
	)
	{	fclose (m_fInput);
		return RESPONSE_RESOURCES_EXHAUSTED;
	}
	return RESPONSE_OK;
}


/*-------------------------------------------------------------
A serviceDetach request on closes the main file service and all
auxiliry services
-------------------------------------------------------------*/
RESULT FileTrivInstance :: DPI_ServiceDetach_Req (REASON)
{
	for (int i = m_auxFiles.GetLength (); --i >= 0;)
	{	m_auxFiles [i] -> Disconnect ();
		delete (m_auxFiles [i]);
		m_auxFiles.DeleteFrom (i);
	}
	Disconnect ();
	GetSink () -> DPI_ServiceDetach_Cnf (RESPONSE_OK);

	return ( RESULT_OK ) ;
}


/*-----------------------------------------------------------------
Signal the file thread to terminate
-----------------------------------------------------------------*/
void FileTrivInstance :: Disconnect()
{ 
	m_mtxChannels.P ();
	m_bRewinding = TRUE;	// this way the FileThread will not scan the channelMap at the end
	m_bActive = FALSE;
	if (m_nChannelWaitedFor >= 0)
	{	
		m_nChannelWaitedFor = -1;
		m_smpWaitForChannel.V ();
	}
	for (int i = m_channelMap.GetLength (); --i >= 0;)
	{
		TRACE("ready=%d  ",m_channelMap[i].m_bReady);
		TRACE("needRew=%d  ",m_channelMap[i].m_bNeedRewind);
		TRACE("esid=%d  ",m_channelMap[i].m_nESid);
		TRACE("\n");

		m_channelMap [i].m_bReady = FALSE;
	}
	m_mtxChannels.V ();

	m_smpActive.P ();
	m_smpActive.V ();
}

/*-------------------------------------------------------------
ChannelsAdd requests to open a set of channels. For each channel,
we check whether is belongs to the main file (no URL) or to an auxiliary
file (URL supplied). In the later case, find the appropriate file service
or create a new one
-------------------------------------------------------------*/
RESULT FileTrivInstance :: DPI_ChannelAdd_Req (const DAIChannelAddParam &inParam)
{
	for (int i = 0; i < inParam.GetLength (); i++)
	{	
		const DAIChannelAddElement &element = inParam [i];
		/*-------------------------------------------------------------
		No URL case relates to this object
		-------------------------------------------------------------*/
		if (element.m_nDataLen == 2)
		{	
			OpenChannel (element.m_pData,element.m_nDataLen);
			continue;
		}
		
		/*-------------------------------------------------------------
		Look if container file already intialized
		-------------------------------------------------------------*/
		char *szTmpURL = new char[element.m_nDataLen];
		strcpy(szTmpURL, (char *)element.m_pData);
		char *szEnd = strchr(szTmpURL, ':');
		if (szEnd)	
			memset(szEnd, 0, 1);

		for (int j = m_auxFiles.GetLength (); --j >= 0;)
		{
			if (m_auxFiles [j] -> IsMyFile (szTmpURL))
			{	
				m_auxFiles [j] -> OpenChannel (element.m_pData,element.m_nDataLen);
				break;
			}
		}

		/*-------------------------------------------------------------
		Open a new file
		-------------------------------------------------------------*/
		if (j < 0)
		{	
			const int nID = m_auxFiles.GetLength () + 1;
			
			FileTrivInstance *pNewService = new FileTrivInstance( nID ) ;
				m_auxFiles += pNewService ;
			pNewService -> SetSink (GetSink ());
			
			const RESPONSE_PARAMETER result = pNewService->Connect( szTmpURL );
			if (result != RESPONSE_OK)
			{	
				DPI_TransmuxAdd_CnfParam smt;
				
				StreamMapElement &rMapElement = smt.ElementAt (0);
				rMapElement.m_pData = new BYTE[element.m_nDataLen];
				memcpy(rMapElement.m_pData,element.m_pData,element.m_nDataLen);
				rMapElement.m_nDataLen = element.m_nDataLen;
				GetSink () -> DPI_ChannelAdd_Cnf (NULL, rMapElement, result);
			}
			else
				pNewService -> OpenChannel (element.m_pData,element.m_nDataLen);
		}
	}
	return ( RESULT_OK ) ;
}

/*-----------------------------------------------------------------
Mark that a channel is requested and confirm (or not) with DPI_TransmuxAdd_Cnf.
When channels are requested, the function calls
ServiceClientSink::DPI_TransmuxAdd_Cnf() for each requested channel.

-----------------------------------------------------------------*/
void FileTrivInstance :: OpenChannel (BYTE* pData, int nDataLen)
{

	DPI_TransmuxAdd_CnfParam smt;
	StreamMapElement &rMapElement = smt.ElementAt (0);
	rMapElement.m_pData = new BYTE[nDataLen];
	memcpy(rMapElement.m_pData,pData,nDataLen);
	rMapElement.m_nDataLen = nDataLen;

	const int nTag = TagFromData (pData, nDataLen);

	if (nTag >= 0)
	{	
		rMapElement.m_nChId = GetNewChannelId( ) ;
		m_channelMap[nTag].m_nChid = rMapElement.m_nChId;
		GetSink () -> DPI_TransmuxAdd_Cnf (
			(HANDLE) MAKELONG (nTag, m_nID),
			NULL_FLEXMUX,
			smt );
	}
	else
	{
		rMapElement.m_nChId = 0;

		GetSink () -> DPI_ChannelAdd_Cnf (NULL, rMapElement, RESPONSE_CHANNEL_NOT_FOUND);
	}
}


/*-----------------------------------------------------------------
Play a channel
-----------------------------------------------------------------*/
void FileTrivInstance :: PlayChannel (const int nTag)
{	
	m_mtxChannels.P ();
	/*-----------------------------------------------------------------
	If the stream has already played once, rewind the file and
	play everything again.
	-----------------------------------------------------------------*/
	if (!m_channelMap [nTag].m_bReady)
	{	
		if (m_channelMap [nTag].m_bNeedRewind)
			Rewind ();
		m_channelMap [nTag].m_bReady = TRUE;
		m_channelMap [nTag].m_bNeedRewind = TRUE;

		/*-----------------------------------------------------------------
		If the file thread is waiting for this channel to be ready,
		signal it to go
		-----------------------------------------------------------------*/
		if (m_nChannelWaitedFor == nTag)
		{	
			m_smpWaitForChannel.V ();
			m_nChannelWaitedFor = -1;
		}
	}
	m_mtxChannels.V ();
}

/*-----------------------------------------------------------------
The file thread reads data from the file and pass it to ServiceClientSink
-----------------------------------------------------------------*/
void FileTrivInstance :: FileThread ()
{
	BYTE buffer [256];
	RESPONSE nTerminationCode = RESPONSE_OK;

	m_smpActive.P ();
	while (m_bActive)
	{
		try
		{	
			/*-----------------------------------------------------------------
			If stream transmission has not been requested yet, wait till it
			is requested
			-----------------------------------------------------------------*/
			const int nFmcNo = GetByte ();
			m_mtxChannels.P ();		// critical section
			if (!m_bActive)
			{
				m_mtxChannels.V ();
				break;
			}
			if (!m_channelMap [nFmcNo].m_bReady)
			{	
				m_nChannelWaitedFor = nFmcNo;
				m_mtxChannels.V ();
				m_smpWaitForChannel.P ();
				if (!m_bActive)
					break;
			}
			else
			{	m_channelMap [nFmcNo].m_bInProcess = true;
				m_mtxChannels.V ();
			}

			UINT nLength = GetByte ();
			if (fread (buffer, 1, nLength, m_fInput) < nLength)
				throw FileException ();
			GetSink () -> DPI_DataReceived_Ind (
				(HANDLE) MAKELONG (nFmcNo, m_nID),
				buffer,
				nLength,
				FALSE
			);
			m_mtxChannels.P ();
			m_channelMap [nFmcNo].m_bInProcess = false;
			bool bReady = m_channelMap [nFmcNo].m_bReady;
			m_mtxChannels.V ();
			if (!bReady)
				StopChannel (nFmcNo, RESPONSE_OK);
		}
		catch (FileException except)
		{
			m_bActive = FALSE;
			if (except.m_nCode == FileException :: MEMORY_ERROR)
				nTerminationCode = RESPONSE_RESOURCES_EXHAUSTED;
		}
	}
	m_smpActive.V ();
	m_nChannelWaitedFor = -1;

	/*-------------------------------------------------------------
	If end of a file has reached (and rewinding was not required), notify
	the sink that all the contained channels have closed.
	-------------------------------------------------------------*/
	if (!m_bRewinding)
	{	
		for (int i = m_channelMap.GetLength (); --i >= 0;)
			if (m_channelMap [i].m_bReady ||
				m_channelMap [i].m_bInProcess
			)
			{	m_channelMap [i].m_bInProcess = false;
				StopChannel (i, nTerminationCode);
			}
	}
	else
		m_bRewinding = FALSE;
}

/*-----------------------------------------------------------------
Read one byte and throw exception on end of file
-----------------------------------------------------------------*/
int FileTrivInstance :: GetByte ()
{
	const int c = fgetc (m_fInput);
	if (c == EOF)
		throw FileException ();
	return c;
}

/*-----------------------------------------------------------------
stop a channel from playing
-----------------------------------------------------------------*/
void FileTrivInstance :: StopChannel (
	int nTag,
	RESPONSE_PARAMETER nTerminationCode
)
{
	m_mtxChannels.P ();		// critical section
	m_channelMap [nTag].m_bReady = false;
	bool bInProcess = m_channelMap [nTag].m_bInProcess;
	m_mtxChannels.V ();
	if (!bInProcess)
		GetSink () -> DPI_TransmuxDelete_Cnf (
			(HANDLE) MAKELONG (nTag, m_nID),
			nTerminationCode
		);
}

/*-----------------------------------------------------------------
Rewind the file in order to replay the streams
-----------------------------------------------------------------*/
void FileTrivInstance :: Rewind ()
{
	m_bRewinding = TRUE;
 	Disconnect();						

	rewind (m_fInput);
	/*-----------------------------------------------------------------
	Skip the initial OD
	-----------------------------------------------------------------*/
	int c = GetByte ();
	int n = MAKEWORD (GetByte (), c);
	while (--n >= 0)
		GetByte ();
	
	/*-----------------------------------------------------------------
	Skip the association map
	-----------------------------------------------------------------*/
	m_nNotReady = n = GetByte ();
	while (--n >= 0)
	{	
		GetByte ();
		GetByte ();
	}

	for (int i = m_channelMap.GetLength (); --i >= 0;)
		m_channelMap [i].m_bNeedRewind = FALSE;
	/*-----------------------------------------------------------------
	Restart the file thread
	-----------------------------------------------------------------*/
	m_bActive = TRUE;
	m_bRewinding = false;
  	ZThread<FileTrivInstance> (
			FileThread,
			this,
			THREAD_PRIORITY_BELOW_NORMAL
	).Begin ();
}

/*-------------------------------------------------------------
ChannelsDelete stops the channels in the list.
The function tries to perform this operation on all open files, since
we don't know which file contains this stream.
-------------------------------------------------------------*/
RESULT FileTrivInstance :: DPI_ChannelDelete_Req (const DAIChannelDeleteParam &inParam)
{
	if ( ! CloseChannels( inParam ) )
	{
		int i = m_auxFiles.GetLength () ;
		for ( ; --i >= 0 && !m_auxFiles [i] -> CloseChannels (inParam) ; )
			;
	}
	return ( RESULT_OK ) ;
}

/*-------------------------------------------------------------
Close all the channels in the list and return true if managed to close
all of them (i.e. all are in this file
-------------------------------------------------------------*/
bool FileTrivInstance :: CloseChannels (const DAIChannelDeleteParam &inParam)
{
	bool bAllClosed = true;
	for (int i = inParam.GetLength (); --i >= 0;)
	{	int nTag = TagFromChid ( inParam[i].m_nChId ) ;
		if (nTag >= 0)
		{	
			StopChannel (nTag, RESPONSE_OK);
		}
		else
			bAllClosed = false;
	}
	return bAllClosed;
}			


/*-------------------------------------------------------------
This is my URL if it starts with "file:" or is a valid file name
-------------------------------------------------------------*/
BOOL FileTrivInstance :: IsYourURL (LPCSTR szURL)
{
	return !_strnicmp (szURL, "d-trif:", 7) || ZFileExists (szURL);
}

/*-------------------------------------------------------------
Parse and execute a user command
-------------------------------------------------------------*/
RESULT FileTrivInstance :: DPI_UserCommand_Req(const DAIUserCmdParam& inParam)
{
	RESULT ret = RESULT_OK;
	
	for(int i=0;i<inParam.GetLength();i++)
	{
		assert(inParam[i].m_nLen >= 1);

		int nTag = TagFromChid(inParam[i].m_nChId);
		if (nTag >= 0)
		{
			switch (inParam[i].m_cData[0])
			{
			case DMIF_UC_PLAY:
				PlayChannel(nTag);
				break;
			case DMIF_UC_STOP:
				StopChannel(nTag, RESPONSE_OK);
				break;
			default:
				ret = DMIF_INTERNAL_ERROR;
			}
		}
		else // try with auxiliary files
		{	int j;
			for (j = m_auxFiles.GetLength (); --j >= 0;)
			{
				if ((nTag = m_auxFiles [j] -> TagFromChid(inParam[i].m_nChId)) >= 0)
				{
					switch (inParam[i].m_cData[0])
					{
					case DMIF_UC_PLAY:
						m_auxFiles [j] -> PlayChannel(nTag);
						break;
					case DMIF_UC_STOP:
						m_auxFiles [j] -> StopChannel(nTag, RESPONSE_OK);
						break;
					default:
						ret = DMIF_INTERNAL_ERROR;
					}
					break;
				}
			}
		}

		if (nTag < 0) ret = INVALID_CHANNEL_ID;
		// in anycase, this use of "ret" is not consistent, 
		// since we will return the "ret" relative to the last channel in the list !!
		// I'm not sure what should we do here.
	}
	return ret;
}

/*-------------------------------------------------------------
Returns a new channel id.
-------------------------------------------------------------*/
UINT FileTrivInstance :: GetNewChannelId()
{
	return ++channelIdCounter;
}

int FileTrivInstance :: TagFromChid(UINT chid)
{
	for (int i = m_channelMap.GetLength (); --i >= 0;)
		if (m_channelMap[i].m_nChid == chid)
			return i;
	return -1;
}

int FileTrivInstance :: TagFromData(BYTE *pData, int nDataLen)
{
	for (int i = m_channelMap.GetLength (); --i >= 0;)
		if ((m_channelMap[i].m_nDataLen == nDataLen) && (!memcmp (pData, m_channelMap[i].m_pData,nDataLen)))
			return i;
	return -1;
}

/*-------------------------------------------------------------
Compare the given URL to the file name of this object
-------------------------------------------------------------*/
bool FileTrivInstance :: IsMyFile (LPCSTR szFileName)
{
	/*-------------------------------------------------------------
	If the URL starts with "d-trif:" it is followed by the file name
	-------------------------------------------------------------*/
	if (!_strnicmp (szFileName, "d-trif:", 7))
		szFileName += 7;
	return ( stricmp (szFileName, m_url) == 0 ) ;
}

UINT FileTrivInstance :: channelIdCounter = 0;
