/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  all
FILE:       ztl.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of the template library for data containers.
Currently includes:

ZArray<type>
ZString
-------------------------------------------------------------------------*/
#include "defs.h"
#include "ztl.h"
#include <io.h>

/*-----------------------------------------------------------------
ZArrayBase :: Stress () - make sure we have room for nSize+1 elements
-----------------------------------------------------------------*/
void ZArrayBase :: Stress (int nSize)
{
	int nNewSize = (++nSize) * m_nElementSize;

	/*-----------------------------------------------------------------
	If necessary, extend the buffer by 25% or GROW_BY, whatever is larger,
	but not less than what we need.
	-----------------------------------------------------------------*/
	if (nNewSize > m_nSize)
	{	nNewSize = max (nNewSize, m_nSize + max (GROW_BY, m_nSize / 4));
		if (m_pBuffer == NULL)
			m_pBuffer = malloc (nNewSize);
		else
			m_pBuffer = realloc (m_pBuffer, nNewSize);
		if (m_pBuffer == NULL)
			throw;
		m_nSize = nNewSize;
	}
	if (m_nLength < nSize)
		m_nLength = nSize;
}

/*-----------------------------------------------------------------
Clear all elements
-----------------------------------------------------------------*/
void ZArrayBase :: Clear ()
{
	if (m_pBuffer)
		free (m_pBuffer);
	m_pBuffer = NULL;
	m_nSize = m_nLength = 0;
}

/*-----------------------------------------------------------------
Make room for new element at position i
-----------------------------------------------------------------*/
void ZArrayBase :: MakeRoomAt (int i)
{
	Stress (m_nLength);
	if (m_nLength > i + 1)
		memmove (
			((char *) m_pBuffer) + (i + 1) * m_nElementSize,
			((char *) m_pBuffer) + i * m_nElementSize,
			(m_nLength - i - 1) * m_nElementSize
		);
	memset (((char *) m_pBuffer) + i * m_nElementSize, 0, m_nElementSize);
}

/*-----------------------------------------------------------------
Delete element at position i
-----------------------------------------------------------------*/
void ZArrayBase :: DeleteFrom (int i)
{
	if (m_nLength > i + 1)
		memmove (
			((char *) m_pBuffer) + i * m_nElementSize,
			((char *) m_pBuffer) + (i + 1) * m_nElementSize,
			(m_nLength - i - 1) * m_nElementSize
		);
	--m_nLength;
}

/*-----------------------------------------------------------------
Scans the array and returns a pointer to an element which meets some
criterion, or NULL if none does.
ZCriterion is a base class for objects which have a virtual function,
IsOK(), which receives a pointer to NULL, casts it to pointer to
some object, and checks whether the object meets some criterion
-----------------------------------------------------------------*/
void *ZArrayBase :: Search (ZCriterion *pCriterion)
{
	int i;
	void *p;

	for (i = 0; i < m_nLength; i++)
	{	p = ((char *) m_pBuffer) + i * m_nElementSize;
		if (pCriterion -> IsOK (p))
			return p;
	}
	return NULL;
}

/*-----------------------------------------------------------------
Copy the contents of one array to the other, leaving the source array
empty
-----------------------------------------------------------------*/
void ZArrayBase :: Pump (ZArrayBase &rArray)
{
	m_pBuffer = rArray.m_pBuffer;
	m_nLength = rArray.m_nLength;
	m_nSize = rArray.m_nSize;
	rArray.m_pBuffer = NULL;
	rArray.m_nLength = rArray.m_nSize = 0;
}

/*-------------------------------------------------------------
ZMapBase constructor. Set initial values
-------------------------------------------------------------*/
ZMapBase :: ZMapBase (int nElementSize) :
	m_pIndice (NULL),
	m_pBuffer (NULL),
	m_nElementSize (nElementSize),
	m_nSize (0),
	m_nLength (0),
	m_nMask (0),
	m_nShift (0)
{}

/*-------------------------------------------------------------
ZMapBase destructor deletes the buffers
-------------------------------------------------------------*/
ZMapBase :: ~ZMapBase ()
{
	delete [] m_pBuffer;
	delete [] m_pIndice;
}

/*-------------------------------------------------------------
Find an index in a ZMap table
-------------------------------------------------------------*/
int ZMapBase :: FindIndex (int index) const
{
	int i = (index ^ (index >> m_nShift)) & m_nMask;
	int n = m_nSize;
	for (n = m_nSize; n > 0; n--)
	{	if (m_pIndice [i] == index)
			return i;
		if (i == 0)
			i = m_nSize;
		--i;
	}
	return -1;
}

/*-------------------------------------------------------------
Insert a new index into a ZMap table
-------------------------------------------------------------*/
int ZMapBase :: InsertIndex (int index)
{
	/*-------------------------------------------------------------
	If table is full allocate a new table double in size, copy old table to the new
	one, and delete the old table
	-------------------------------------------------------------*/
	int i, n;
	if (m_nLength == m_nSize)
	{	int nOldSize = m_nSize;
		int *pOldIndice = m_pIndice;
		char *pOldBuffer = m_pBuffer;
		m_nLength = 0;
		m_nSize <<= 1;
		m_nMask <<= 1;
		m_nMask |= 1;
		m_nShift++;
		if (m_nSize == 0)
		{	m_nSize = 64;
			m_nMask = 0x3f;
			m_nShift = 6;
		}
		m_pIndice = new int [m_nSize];
		m_pBuffer = new char [m_nSize * m_nElementSize];
		memset (m_pIndice, -1, m_nSize * sizeof (int));
		for (i = nOldSize; --i >= 0;)
		{	if (pOldIndice [i] >= 0)
			{	n = InsertIndex (pOldIndice [i]);
				memcpy (
					m_pBuffer + n * m_nElementSize,
					pOldBuffer + i * m_nElementSize,
					m_nElementSize
				);
			}
		}
		delete [] pOldIndice;
		delete [] pOldBuffer;
	}

	/*-------------------------------------------------------------
	Now find a slot for the new index
	-------------------------------------------------------------*/
	i = (index ^ (index >> m_nShift)) & m_nMask;
	n = m_nSize;
	while (--n >= 0)
	{	if (m_pIndice [i] == -1)
		{	m_pIndice [i] = index;
			++m_nLength;
			return i;
		}
		if (i == 0)
			i = m_nSize;
		--i;
	}
	return -1;
}

/*-------------------------------------------------------------
Delete an index (if exists) from the table
-------------------------------------------------------------*/
int ZMapBase :: DeleteIndex (int index)
{
	int i = FindIndex (index);
	if (i >= 0)
	{	m_pIndice [i] = -1;
		--m_nLength;
	}
	return i;
}

/*-----------------------------------------------------------------
ZString constructor. Allocate memory and assign value
-----------------------------------------------------------------*/
ZString :: ZString (LPCSTR szStr)
{
	if (szStr == NULL)
	{	m_nLength = m_nSize = 0;
		m_pBuffer = NULL;
	}
	else
	{	m_nLength = strlen (szStr);
		m_nSize = m_nLength + 1;
		if ((m_pBuffer = (unsigned char *) malloc (m_nSize)) == NULL)
			throw;
		memcpy (m_pBuffer, szStr, m_nSize);
	}
}

/*-----------------------------------------------------------------
ZString constructor with string and length
-----------------------------------------------------------------*/
ZString :: ZString (const unsigned char *szStr, int nLength)
{
	m_nLength = nLength;
	m_nSize = m_nLength + 1;
	if ((m_pBuffer = (unsigned char *) malloc (m_nSize)) == NULL)
		throw;
	memcpy (m_pBuffer, szStr, nLength);
	m_pBuffer [nLength] = 0;
}

/*-----------------------------------------------------------------
ZString copy constructor
-----------------------------------------------------------------*/
ZString :: ZString (const ZString &str)
{
	m_nLength = str.m_nLength;
	m_nSize = str.m_nSize;
	m_pBuffer = NULL;
	if (str.m_pBuffer)
	{
        	if ((m_pBuffer = (unsigned char *) malloc (m_nSize)) == NULL)
				throw;
		memcpy (m_pBuffer, str.m_pBuffer, m_nLength + 1);
	}
}

/*-----------------------------------------------------------------
Make sure that the string has enough space for the ith byte.
If bFill is TRUE, extend the string by padding with blanks and appending
a null byte.
-----------------------------------------------------------------*/
void ZString :: Stress (int i, BOOL bFill)
{
	int nNewSize;

	/*-----------------------------------------------------------------
	If necessary, extend the buffer by 25% or GROW_BY, whatever is larger,
	but not less than i + 2 (for the terminating null byte).
	-----------------------------------------------------------------*/
	if (i >= m_nSize)
	{	nNewSize = max (i + 2, m_nSize + max (GROW_BY, m_nSize / 4));
		m_pBuffer = m_pBuffer == NULL ?
			(unsigned char *) malloc (nNewSize) :
			(unsigned char *) realloc (m_pBuffer, nNewSize);
		if (m_pBuffer == NULL)
			throw;
		m_nSize = nNewSize;
		if (i == 0)
			*m_pBuffer = 0;		
	}

	if (!bFill)
		return;

	/*-----------------------------------------------------------------
	If necessary, fill with blanks and append a null byte
	-----------------------------------------------------------------*/
	if (i > m_nLength)
		memset (m_pBuffer + m_nLength, ' ', i - m_nLength);
	if (i >= m_nLength)
		m_pBuffer [m_nLength = i + 1] = 0;
}

/*-----------------------------------------------------------------
Return a reference to position i. Make sure string is large enough
-----------------------------------------------------------------*/
unsigned char & ZString :: operator [] (int i)
{
	Stress (i, TRUE);
	return m_pBuffer [i];
}

/*-----------------------------------------------------------------
ZString assignment
-----------------------------------------------------------------*/
ZString &ZString :: operator = (const ZString &str)
{
	if (str.m_pBuffer == NULL)
	{	if (m_pBuffer)
			free (m_pBuffer);
		m_pBuffer = NULL;
		m_nLength = m_nSize = 0;
	}
	else
	{	Stress (m_nLength = str.m_nLength);
		memcpy (m_pBuffer, str.m_pBuffer, m_nLength + 1);
	}
	return *this;
}

/*-----------------------------------------------------------------
ZString assignment
-----------------------------------------------------------------*/
ZString &ZString :: operator = (LPCSTR szStr)
{
	if (szStr == NULL)
	{	if (m_pBuffer)
			free (m_pBuffer);
		m_pBuffer = NULL;
		m_nLength = m_nSize = 0;
	}
	else
	{	Stress (m_nLength = strlen (szStr));
		strcpy ((LPSTR) m_pBuffer, szStr);
	}
	return *this;
}

/*-----------------------------------------------------------------
Append one character to a string
-----------------------------------------------------------------*/
ZString &ZString :: operator += (char c)
{
	Stress (m_nLength + 1);
	m_pBuffer [m_nLength++] = c;
	m_pBuffer [m_nLength] = 0;
	return *this;
}

/*-----------------------------------------------------------------
Concatenate two strings
-----------------------------------------------------------------*/
ZString &ZString :: operator += (char *szStr)
{
	int n = strlen (szStr);
	Stress (m_nLength + n);
	memcpy (m_pBuffer + m_nLength, szStr, n);
	m_pBuffer [m_nLength += n] = 0;
	return *this;
}

/*-----------------------------------------------------------------
Concatenate two strings
-----------------------------------------------------------------*/
ZString &ZString :: operator += (const ZString &str)
{
	Stress (m_nLength + str.m_nLength);
	memcpy (m_pBuffer + m_nLength, str.m_pBuffer, str.m_nLength);
	m_pBuffer [m_nLength += str.m_nLength] = 0;
	return *this;
}

/*-----------------------------------------------------------------
Concatenate two strings
-----------------------------------------------------------------*/
ZString ZString :: operator + (const ZString &str) const
{
	return ZString (*this) += str;
}

/*-----------------------------------------------------------------
Insert a character into position i
-----------------------------------------------------------------*/
unsigned char ZString :: InsertAt (int i, char c, int n)
{
	m_nLength = max (m_nLength, i) + n;
	Stress (m_nLength - 1, true);
	memmove (m_pBuffer + i + n, m_pBuffer + i, m_nLength - i);
	memset (m_pBuffer + i, c, n);
	return c;
}

/*-------------------------------------------------------------
Insert one string into another
-------------------------------------------------------------*/
const ZString &ZString :: InsertAt (int i, const ZString &str)
{
	int n = str.GetLength ();
	if (n > 0)
	{	m_nLength = max (m_nLength, i) + n;
		Stress (m_nLength - 1, true);
		memmove (m_pBuffer + i + n, m_pBuffer + i, m_nLength - i);
		memcpy (m_pBuffer + i, str, n);
	}
	return str;
}


/*-----------------------------------------------------------------
Delete a character from position i. Returns the deleted character
-----------------------------------------------------------------*/
char ZString :: DeleteFrom (int i)
{
	char cResult = m_pBuffer [i];

	memmove (m_pBuffer + i, m_pBuffer + i + 1, m_nLength - i);
	--m_nLength;
	return cResult;
}

/*-----------------------------------------------------------------
Compare two strings
-----------------------------------------------------------------*/
BOOL ZString :: operator == (const ZString &str) const
{
	return m_nLength == str.m_nLength && !memcmp (*this, str, m_nLength);
}

/*-----------------------------------------------------------------
Find a character c in a ZString, starting the search at offset iOffset.
Returns the distance of the index of the found character from nOffset,
or -1 if did not find
-----------------------------------------------------------------*/
int ZString :: Find (int c, int nOffset) const
{
	int i;

	for (i = nOffset; i < GetLength (); i++)
		if ((*this) [i] == c)
			return i - nOffset;
	return -1;
}

/*-----------------------------------------------------------------
Return a string made of a substring of this string.
If nLength is negative, substring is from nFrom till end of string
-----------------------------------------------------------------*/
ZString ZString :: Substr (int nFrom, int nLength) const
{
	if (nFrom >= m_nLength || nLength == 0)
		return ZString ();
	if (nLength < 0)
		nLength = m_nLength - nFrom;
	return ZString (
		m_pBuffer + nFrom,
		min (m_nLength - nFrom, nLength)
	);
}

/*-----------------------------------------------------------------
Cut nLength chars out of a string, starting at nFrom.
If nLength is negative, cut till end of string
-----------------------------------------------------------------*/
void ZString :: Cut (int nFrom, int nLength)
{
	if (nFrom >= m_nLength || nLength == 0)
		return;
	if (nLength < 0)
		nLength = m_nLength - nFrom;
	memmove (
		m_pBuffer + nFrom,
		m_pBuffer + nFrom + nLength,
		m_nLength - nFrom - nLength + 1
	);
	m_nLength -= nLength;
}

/*-----------------------------------------------------------------
Copy a string to the object. No more than nMax characters
will be copied
-----------------------------------------------------------------*/
ZString & ZString :: Copy (LPCSTR szStr, int nMax)
{
	m_nLength = min ((int) strlen (szStr), nMax);
	Stress (m_nLength);
	memcpy (m_pBuffer, szStr, m_nLength);
	m_pBuffer [m_nLength] = 0;
	return *this;
}

/*-----------------------------------------------------------------
Copy an array of bytes to the object.
-----------------------------------------------------------------*/
ZString & ZString :: Assign (LPCBYTE pData, int nLength)
{
	m_nLength = nLength;
	Stress (m_nLength);
	memcpy (m_pBuffer, pData, m_nLength);
	m_pBuffer [m_nLength] = 0;
	return *this;
}

#ifdef WIN32

/*-------------------------------------------------------------------------
ZSemaphore constructor. Create unnamed non-inherited semaphore.
If lInitialCount is -1, it is set to lMaxCount.
-------------------------------------------------------------------------*/
ZSemaphore :: ZSemaphore (long lMaxCount, long lInitialCount)
{
	if (lInitialCount < 0)
		lInitialCount = lMaxCount;
	m_hSemaph = CreateSemaphore (NULL, lInitialCount, lMaxCount, NULL);
}

/*-------------------------------------------------------------------------
Try to reserve the semphore. Return immediately if semaphore already
reserved. Return TRUE if succeeded.
-------------------------------------------------------------------------*/
BOOL ZSemaphore :: Try ()
{
	return WaitForSingleObject (m_hSemaph, 0) != WAIT_TIMEOUT;
}	 

/*-------------------------------------------------------------------------
Try to reserve the mutex. Return immediately if mutex already
reserved. Return TRUE if succeeded.
-------------------------------------------------------------------------*/
BOOL ZMutex :: Try ()
{
	return WaitForSingleObject (m_hMutex, 0) != WAIT_TIMEOUT;
}	 

#ifdef _DEBUG
void ZMutex :: Trace (LPCSTR szOp)
{
	char szBuffer [80];

	sprintf (
		szBuffer,
		"Mutex handle: %x, operation: %s\n",
		m_hMutex,
		szOp
	);
	OutputDebugString (szBuffer);
}
#endif		// _DEBUG

#elif defined(PTHREADS)

#include <sys/ptimers.h>

/*-------------------------------------------------------------------------
ZSemaphore constructor. Create unnamed non-inherited semaphore.
If lInitialCount is -1, it is set to lMaxCount.
-------------------------------------------------------------------------*/
ZSemaphore :: ZSemaphore (long lMaxCount, long lInitialCount)
{
	if (lInitialCount < 0)
		lInitialCount = lMaxCount;
	m_nCount = lInitialCount;

	::pthread_cond_init (&m_Cond, NULL);
	::pthread_mutex_init (&m_Mutex, NULL);
}

/*-------------------------------------------------------------------------
ZSemaphore destructor.
-------------------------------------------------------------------------*/
ZSemaphore :: ~ZSemaphore ()
{
	::pthread_cond_destroy (&m_Cond);
	::pthread_mutex_destroy (&m_Mutex);
}

/*-------------------------------------------------------------------------
Reserve the semaphore. Return TRUE if succeeded.
-------------------------------------------------------------------------*/
DWORD ZSemaphore :: P (UINT uTimeout)
{
	::pthread_mutex_lock(&m_Mutex);

	if (uTimeout == INFINITE)
		while (m_nCount == 0)
		{	if (::pthread_cond_wait(&m_Cond, &m_Mutex))
			{
				::perror("pthread_cond_wait");
				::pthread_mutex_unlock(&m_Mutex);
				return FALSE;
			}
		}
	else
	{
		struct timeval currenttime;
		struct timespec timeout;

		::gettimeofday(&currenttime);

		timeout.tv_nsec =
			long (currenttime.tv_usec * 1000 + uTimeout * 1000000);
			timeout.tv_sec =
				currenttime.tv_sec + timeout.tv_nsec / 1000000000;
		timeout.tv_nsec %= 1000000;

		while (m_nCount == 0)
			switch (::pthread_cond_timedwait(
				&m_Cond,
				&m_Mutex,
				&timeout)
			){
			default:
				::perror("pthread_cond_timedwait");

			case ETIMEDOUT:
				::pthread_mutex_unlock(&m_Mutex);
				return FALSE;

			case 0:
				;
			}
	}

	m_nCount--;
	::pthread_mutex_unlock(&m_Mutex);
	return TRUE;
}

/*-------------------------------------------------------------------------
Release the semphore. Return TRUE if succeeded.
-------------------------------------------------------------------------*/
BOOL ZSemaphore :: V ()
{
	::pthread_mutex_lock(&m_Mutex);
	if (m_nCount == 0)
		::pthread_cond_signal(&m_Cond);
	m_nCount++;
	::pthread_mutex_unlock(&m_Mutex);
	return FALSE;
}

/*-------------------------------------------------------------------------
Try to reserve the semphore. Return immediately if semaphore already
reserved. Return TRUE if succeeded.
-------------------------------------------------------------------------*/
BOOL ZSemaphore :: Try ()
{
	BOOL ret;
	::pthread_mutex_lock(&m_Mutex);
	if (m_nCount == 0)
		ret = FALSE;
	else
	{
		m_nCount--;
		ret = TRUE;
	}
	::pthread_mutex_unlock(&m_Mutex);
	return ret;
}

/*-------------------------------------------------------------------------
Try to reserve the mutex. Return immediately if mutex already
reserved. Return TRUE if succeeded.
-------------------------------------------------------------------------*/
BOOL ZMutex :: Try ()
{
	switch (::pthread_mutex_trylock(&m_Mutex)) {
	case 0:
		return TRUE;
	default:
		perror("pthread_mutex_trylock");
	case EBUSY:
		return FALSE;
	}
}	 

#else

usptr_t* ZSGIArena::m_pArena = NULL;

/*-------------------------------------------------------------------------
ZSemaphore constructor. Create unnamed non-inherited semaphore.
If lInitialCount is -1, it is set to lMaxCount.
-------------------------------------------------------------------------*/
ZSemaphore :: ZSemaphore (long lMaxCount, long lInitialCount)
{
	if (lInitialCount < 0)
		lInitialCount = lMaxCount;
	
	m_pSema = usnewsema(ZSGIArena::arena(), lInitialCount);
	if (!m_pSema)
	  ::perror("usnewsema");
}

/*-------------------------------------------------------------------------
Try to reserve the mutex. Return immediately if mutex already
reserved. Return TRUE if succeeded.
-------------------------------------------------------------------------*/
BOOL ZMutex :: Try ()
{
	switch (uscsetlock(m_Mutex, 1)) {
	default:
		::perror("uscsetlock");
	case 0:
		return FALSE;
	case 1:
		return TRUE;
	}
}	 

/*-------------------------------------------------------------------------
Reserve the mutex.  Return TRUE if succeeded.
-------------------------------------------------------------------------*/
DWORD ZMutex :: P ()
{
	switch (ussetlock(m_Mutex)) {
	default:
		::perror("ussetlock");
	case 0:
		return FALSE;
	case 1:
		return TRUE;
	}
}	 

#endif

/*-----------------------------------------------------------------
A function that calculates how many bits are required to store a number
which is less than nMax
-----------------------------------------------------------------*/
int BitFieldWidth (int nMax)
{
	int nResult = 0;
	--nMax;
	while (nMax > 0)
	{	++nResult;
		nMax >>= 1;
	}
	return nResult;
}

/*-------------------------------------------------------------
ZRegistry encapsulates registry access. The constructor creates a
handle for the given section. If bCreate is TRUE, the section is created
if does not yet exist.
-------------------------------------------------------------*/
ZRegistry :: ZRegistry (LPCSTR szSection, BOOL bCreate)
{
	DWORD dwDisposition;
	m_hSection = NULL;
	if (bCreate)
		RegCreateKeyEx (
			HKEY_CURRENT_USER,
			szSection,
			0,
			NULL,
			REG_OPTION_NON_VOLATILE,
			KEY_WRITE,
			NULL,
			&m_hSection,
			&dwDisposition
		);
	else
		RegOpenKeyEx(
			HKEY_CURRENT_USER,		// handle of open key
			szSection,				// address of name of subkey to open
			0,						// reserved
			KEY_READ,				// security access mask
			&m_hSection				// address of handle of open key
		);
}

/*-------------------------------------------------------------
ZRegistry destructor closes the key
-------------------------------------------------------------*/
ZRegistry :: ~ZRegistry ()
{
	if (m_hSection)
		RegCloseKey (m_hSection);
}

/*-------------------------------------------------------------
Look into a registry section, find the DLL name that matches the given
key, load the DLL, and calls its "CreateInstance" exported function, which
should instantiate an object and return its address.
-------------------------------------------------------------*/
void * ZRegistry :: CreateInstance (LPCSTR szKey, HINSTANCE &hDLL) const
{
	char szDLL [100];
	DWORD dwcbData = 100;
	typedef void * (*InstantiationFunction) ();
	
	hDLL = NULL;
	if (!m_hSection || RegQueryValueEx (
			m_hSection,
			szKey,
			NULL,
			NULL,
			(LPBYTE) szDLL,
			&dwcbData
		) != ERROR_SUCCESS
	)
		return NULL;
	HINSTANCE hInstance = LoadLibrary (szDLL);
	if (hInstance == NULL)
		return NULL;
	InstantiationFunction function =
		(InstantiationFunction) GetProcAddress (hInstance, "CreateInstance");
	if (function == NULL)
		return NULL;
	void *object = (*function) ();
	if (object == NULL)
	{	FreeLibrary (hInstance);
		hInstance = NULL;
	}
	hDLL = hInstance;
	return object;
}
		
/*-------------------------------------------------------------
Enumerate sub-keys of registry key. Return TRUE is sub-key no. index
exists.
-------------------------------------------------------------*/
BOOL ZRegistry :: Enumerate (int index, LPSTR szBuffer, int nBufferLen)
{
	DWORD dwBufferLen = nBufferLen;

	return
		m_hSection &&
		RegEnumValue (
			m_hSection,
			index,		          // index of value to query
			szBuffer,
			&dwBufferLen,
			NULL,
			NULL,
			NULL,
			0
		) == ERROR_SUCCESS;
}

/*-------------------------------------------------------------
Verifies that a certain file exists
-------------------------------------------------------------*/
BOOL ZFileExists (LPCSTR szFile)
{
	/*-------------------------------------------------------------
	Use _findfirst to find the file then make sure this name uniquely
	identifies a file (not a wildcard name)
	-------------------------------------------------------------*/
    struct _finddata_t fileDetails;
    long hFile;
	BOOL bResult = FALSE;

	if ((hFile = _findfirst (szFile, &fileDetails)) != -1L)
	{	if (_findnext (hFile, &fileDetails) != 0)
			bResult = TRUE;
		_findclose (hFile);
	}
	return bResult;
}

/*-------------------------------------------------------------
Replace the file type (suffix) of szFileName, by szFileType
-------------------------------------------------------------*/
ZString ComposeFileName (LPCSTR szFileName, LPCSTR szFileType)
{
	LPCSTR szSuffix = strrchr (szFileName, '.');
	int nLength =
		szSuffix == NULL ? strlen (szFileName) : szSuffix - szFileName;
	return (ZString ((const LPBYTE) szFileName, nLength) += '.') += szFileType;
}

/*-------------------------------------------------------------
Expand a relative to an absolute URL.
Returns 0 if successful (enough buffer size of the result) or required buffer size if not.
-------------------------------------------------------------*/
int ZExpandURL (LPCSTR szRootURL, LPCSTR szRelURL, LPSTR szAbsURL, int nSize)
{
	int nRootLength = 0;
	if (szRootURL != NULL &&
		strspn (szRelURL, " \t") < strcspn (szRelURL, "/\\") &&		// relURL doesn't start with \ or /
		strcspn (szRelURL, "/\\?") <= strcspn (szRelURL, ":|")		// relURL doesn't start with xxx: or xxx|
	)																// then this is a relative URL
	{	for (nRootLength = strlen (szRootURL); --nRootLength >= 0;)
		{	char c = szRootURL [nRootLength];
			if (c == '/' || c == '\\' || c == ':' || c == '|')
				break;
		}
	}

	/*-------------------------------------------------------------
	Check if there is enough space for the combined URL
	-------------------------------------------------------------*/
	++nRootLength;
	int n = nRootLength + strlen (szRelURL);
	if (n > nSize)
		return n;

	/*-------------------------------------------------------------
	Concatenate the name
	-------------------------------------------------------------*/
	memcpy (szAbsURL, szRootURL, nRootLength);
	strcpy (szAbsURL + nRootLength, szRelURL);
	return 0;
}
