// This software module was originally developed by Wei Wu(Rockwell Science
// Center) and Wenlong Dong(Rockwell Science Center) in the course of 
// development of the MPEG-4 Systems standard
// (ISO/IEC 14496-1).  This software module is an implementation of a part
// of one or more MPEG-4 Systems tools as specified by the MPEG-4 Systems
// standard.  ISO/IEC gives users of the MPEG-4 Systems standard free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4 Systems
// standard.  Those intending to use this software module in hardware or
// software products are advised that its use may infringe existing
// patents.  The original developer of this software module and his/her
// company, the subsequent editors and their companies, and ISO/IEC have no
// liability for use of this software module or modifications thereof in an
// implementation.  Copyright is not released for non MPEG-4 Systems
// conforming products.  The original developer retains full right to use
// the code for his/her own purpose, assign or donate the code to a third
// party and to inhibit third parties from using the code for non MPEG-4
// Systems conforming products.  This copyright notice must be included in
// all copies or derivative works.
//
// For more information or to receive an up to date version of this module,
// you may call Wei Wu at weiwu@jupiter.risc.rockwell.com or Wenlong Dong 
// at wdong@usc.edu.
//
// Copyright 1998.

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:	Script Node
FILE:       Expr.cpp
AUTHOR:		Wei Wu       (Rockwell Science Center)
			Wenlong Dong (Rockwell Science Center)
		    Yuval Fisher (Rockwell Science Center)
OVERVIEW:   Implementation of the CParser class
--------------------------------------------------------------------*/
#include "defs.h"
#include "..\..\BifsEnc\BifsEnc.h"
#include "CParser.h"
#include "VSNodeField.h"
#include "..\..\BifsEnc\Token.h"
#include "VScript.h" // NODEFIELD
#include "..\..\Core\MediaObj.h" // NODEFIELD
#include "..\..\Core\Root.h" // NODEFIELD
#include "..\..\Core\NodeFld.h" // NODEFIELD

static const char *exprDesc[] =
{
	"ET_IDENTIFIER",
	"ET_STRING",
	"ET_NUMBER",
	"ET_CURVED_EXPR",
	"ET_OBJECT_METHOD_CALL",
	"ET_OBJECT_MEMBER_ACCESS",
	"ET_FUNCTION_CALL",
	"ET_ARRAY_DEREFERENCE",
	"ET_NEW",
	"ET_NOT",
	"ET_ONESCOMP",
	"ET_NEGATIVE",
	"ET_INCREMENT",
	"ET_DECREMENT",
	"ET_POST_INCREMENT",
	"ET_POST_DECREMENT",
	"ET_MULTIPLY",
	"ET_DIVIDE",
	"ET_MOD",
	"ET_PLUS",
	"ET_MINUS",
	"ET_LSHIFT",
	"ET_RSHIFT",
	"ET_RSHIFTFILL",
	"ET_LT",
	"ET_LE",
	"ET_GT",
	"ET_GE",
	"ET_EQ",
	"ET_NE",
	"ET_AND",
	"ET_XOR",
	"ET_OR",
	"ET_LAND",
	"ET_LOR",
	"ET_CONDTEST",
	"ET_ASSIGN",
	"ET_PLUSEQ",
	"ET_MINUSEQ",
	"ET_MULTIPLYEQ",
	"ET_DIVIDEEQ",
	"ET_MODEQ",
	"ET_LSHIFTEQ",
	"ET_RSHIFTEQ",
	"ET_RSHIFTFILLEQ",
	"ET_ANDEQ",
	"ET_XOREQ",
	"ET_OREQ"
};

#define NUMBER_OF_RANK	15

static int rank[NUMBER_OF_EXPR_TYPE] =
{
	0, 0, 0,			// variable, number, string
	1, 1, 1, 1, 1,			// curved expr, call, member, array
	2, 2, 2, 2, 2, 2, 2, 2,		// unary operator
	3, 3, 3,			// multiply, divide, mod
	4, 4,				// add, subtract
	5, 5, 5,			// bitwise shift
	6, 6, 6, 6,			// relational
	7, 7,				// equality
	8,				// bitwise and
	9,				// bitwise xor
	10,				// bitwise or
	11,				// logical and
	12,				// logical or
	13,				// conditional
	14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14	// assignment
};

static int leftAssoc[NUMBER_OF_RANK] = {1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0};

#define CheckToken(x)	if (curTC!=x) { \
				fprintf(stderr, "Error: Token %d read, %d expected\n", \
					curTC, x); \
				exit (-1); }

EXPR_TYPE CParser::Expression(BIFSEncoder *pBifs, int startPos, int endPos, BOOL n_flag)
{
	int n = startPos;
	int curPos = 0, finalPos = 0;
	CTOKEN_CODE curTC, prevTC;
	EXPR_TYPE curExprType = 0, exprType = 0;
	int curRank, maxRank=0;

	curTC = prevTC = m_buffer[n];

	do {
		curPos = n;
		switch (curTC) {
		case CTID_IDENTIFIER:
			curExprType = ET_IDENTIFIER;
			break;
		case CTID_NUMBER:
			curExprType = ET_NUMBER;
			break;
		case CTID_STRING:
			curExprType = ET_STRING;
			break;
		case CTID_LEFT_CURVE:
			if (prevTC==CTID_IDENTIFIER) {
				curExprType = ET_FUNCTION_CALL;
			} else {
				curExprType = ET_CURVED_EXPR;
			}
			n = MoveUntil(CTID_RIGHT_CURVE, n, endPos);
			if (n!=-1) {
				curTC = CTID_RIGHT_CURVE;
			} else {
				fprintf(stderr, "Error: can't find closing )\n");
				exit(-1);
			}
			break;
		case CTID_LEFT_BRACKET:
			curExprType = ET_ARRAY_DEREFERENCE;
			n = MoveUntil(CTID_RIGHT_BRACKET, n, endPos);
			if (n!=-1) {
				curTC = CTID_RIGHT_BRACKET;
			} else {
				fprintf(stderr, "Error: can't find closing )\n");
				exit(-1);
			}
			break;
		case CTID_PERIOD:
			curTC = m_buffer[++n];
			CheckToken(CTID_IDENTIFIER);
			if (m_buffer[n+1]==CTID_LEFT_CURVE) {
				curExprType = ET_OBJECT_METHOD_CALL;
				n = MoveUntil(CTID_RIGHT_CURVE, n+1, endPos);
				if (n!=-1){
					curTC = CTID_RIGHT_CURVE;
				} else {
					fprintf(stderr, "Error: can't find closing )\n");
					exit(-1);
				}
			} else {
				curExprType = ET_OBJECT_MEMBER_ACCESS;
			}
			break;
		case CTID_NEW:
			curExprType = ET_NEW;
			curTC = m_buffer[++n];
			CheckToken(CTID_IDENTIFIER);
			curTC = m_buffer[++n];
			CheckToken(CTID_LEFT_CURVE);
			n = MoveUntil(CTID_RIGHT_CURVE, n, endPos);
			if (n!=-1) {
				curTC = CTID_RIGHT_CURVE;
			} else {
				fprintf(stderr, "Error: can't find closing )\n");
				exit(-1);
			}
			break;
		case CTID_MINUS:
			if (
				prevTC==CTID_RIGHT_CURVE || prevTC==CTID_RIGHT_BRACKET ||
				prevTC==CTID_IDENTIFIER || prevTC==CTID_NUMBER ||
				prevTC==CTID_STRING || prevTC==CTID_INCREMENT ||
				prevTC==CTID_DECREMENT
			) {
				curExprType = ET_MINUS;
			} else {
				curExprType = ET_NEGATIVE;
				curTC = CTID_NEGATIVE;
			}
			break;
		case CTID_INCREMENT:
			curExprType = ET_INCREMENT;
			break;
		case CTID_DECREMENT:
			curExprType = ET_DECREMENT;
			break;
		case CTID_NOT:
			curExprType = ET_NOT;
			break;
		case CTID_ONESCOMP:
			curExprType = ET_ONESCOMP;
			break;
		case CTID_CONDTEST:
			curExprType = ET_CONDTEST;
			break;
		case CTID_CONDSEP:
			break;
		default: // hack to reduce typing
			if (curTC < CTID_MULTIPLY || curTC > CTID_OREQ) {
				fprintf(stderr, "Error: illegal token %d read\n", 
					curTC);
				exit(-1);
			}
			curExprType =(EXPR_TYPE)(ET_MULTIPLY + (curTC - CTID_MULTIPLY));
		}
		if (curTC == CTID_CONDSEP) {
			prevTC = curTC;
			curTC = m_buffer[++n];
			continue;
		}
		curRank = rank[curExprType];
		if (curRank > maxRank) {
			maxRank = curRank;
			exprType = curExprType;
			finalPos = curPos;
		} else if (curRank == maxRank && leftAssoc[curRank]) {
			exprType = curExprType;
			finalPos = curPos;
		}
		prevTC = curTC;
		curTC = m_buffer[++n];
	} while (n<endPos);

	if (exprType == ET_INCREMENT) {
		if (finalPos==startPos);
		else if (finalPos==endPos-1) exprType = ET_POST_INCREMENT;
		else {
			fprintf(stderr, "Error: illegal Increment expression\n");
			exit(-1);
		}
	} else if (exprType == ET_DECREMENT) {
		if (finalPos==startPos);
		else if (finalPos==endPos-1) exprType = ET_POST_DECREMENT;
		else {
			fprintf(stderr, "Error: illegal Decrement expression\n");
			exit(-1);
		}
	}

	pBifs -> EncodeInt ((int) exprType, NUMBITS_EXPR_TYPE, exprDesc[exprType]);

	if (exprType >= ET_MULTIPLY && exprType <= ET_LOR) { // hack
		Expression(pBifs, startPos, finalPos);
		Expression(pBifs, finalPos+1, endPos);
	} else if(exprType >= ET_ASSIGN && exprType <=ET_OREQ) { // hack
		EXPR_TYPE ret = Expression(pBifs, startPos, finalPos);
		if (
			ret != ET_IDENTIFIER && ret != ET_OBJECT_MEMBER_ACCESS &&
			ret != ET_ARRAY_DEREFERENCE
		) {
			fprintf(stderr, "Error: LeftVariable expected, %s returned\n",
				exprDesc[ret]);
			exit(-1);
		}
		Expression(pBifs, finalPos+1, endPos);
	} else if(exprType == ET_IDENTIFIER) {
		const char *str = m_strBuffer.getStr();
		if(n_flag) // n_flag is set when the call is from Object Method call or Member access
		{
			nNodeID = -1; // NODEFIELD: should be named nNodeCode
			if(!strcmp(str, "Math"))
				pBifs->EncodeInt(1, 1, "Math identifier"); 
			else {
				pBifs->EncodeInt(0, 1, "Not Math identifier");
				PutIdentifier(str, pBifs);
				for (int i=0; i<pSNode->FieldList.GetLength(); i++) { // NODEFIELD
					if(!strcmp(pSNode->FieldList[i]->GetName(), str)) {
						nNodeID = pSNode->FieldList[i]->GetID();
						break;
					}
				}
				if (nNodeID == -1) {
					fprintf(stderr, "Error: node ID of %s not found\n", str);
					exit(-1);
				}
			}
		}
		else
			PutIdentifier(str, pBifs);

	} else if(exprType == ET_NUMBER) {
		const char *str = m_strBuffer.getStr();
		PutNumber(str, pBifs);
	} else if(exprType == ET_STRING) {
		const char *str = m_strBuffer.getStr();
		PutString(str, pBifs);
	} else if (
		exprType == ET_NEGATIVE || exprType == ET_INCREMENT || 
		exprType == ET_DECREMENT || exprType == ET_NOT || 
		exprType == ET_ONESCOMP
	) {
		Expression(pBifs, finalPos+1, endPos);
	} else if (exprType == ET_CURVED_EXPR) {
		compoundExpression(pBifs, finalPos+1, endPos-1);
	} else if (exprType == ET_POST_INCREMENT || exprType == ET_POST_DECREMENT) {
		Expression(pBifs, startPos, finalPos);
	} else if (exprType == ET_FUNCTION_CALL) {
		FunctionCall(pBifs, startPos, finalPos, endPos);
	} else if (exprType == ET_OBJECT_MEMBER_ACCESS) {
		ObjectMemberAccess(pBifs, startPos, finalPos, endPos);
	} else if (exprType == ET_OBJECT_METHOD_CALL) {
		ObjectMethodCall(pBifs, startPos, finalPos, endPos);
	} else if (exprType == ET_ARRAY_DEREFERENCE) {
		ArrayDereference(pBifs, startPos, finalPos, endPos);
	} else if (exprType == ET_NEW) {
		ObjectConstruct(pBifs, startPos, finalPos, endPos);
	} else if (exprType == ET_CONDTEST) {
		ConditionTest(pBifs, startPos, finalPos, endPos);
	} else {
		fprintf(stderr, "Error: illegal expression type %d\n", exprType);
		exit(-1);
	}
	return exprType;
}

void CParser::FunctionCall(BIFSEncoder *pBifs, int startPos, int, int endPos)
{
	CTOKEN_CODE curTC;

	curTC = m_buffer[startPos++];
	CheckToken(CTID_IDENTIFIER);
	const char *str = m_strBuffer.getStr();
	PutIdentifier(str, pBifs);
	curTC = m_buffer[startPos++];
	CheckToken(CTID_LEFT_CURVE);
	Params(pBifs, startPos, endPos-1);
	curTC = m_buffer[endPos-1];
	CheckToken(CTID_RIGHT_CURVE);
}

void CParser::ObjectMemberAccess(BIFSEncoder *pBifs, int startPos, int opPosition, int endPos)
{
	CTOKEN_CODE curTC;

	Expression(pBifs, startPos, opPosition, TRUE); // The flag is set for "Math" checking
	curTC = m_buffer[opPosition];
	CheckToken(CTID_PERIOD);
	curTC = m_buffer[endPos-1];
	CheckToken(CTID_IDENTIFIER);
	const char *str = m_strBuffer.getStr();
	if (nNodeID == -1) { // NODEFIELD: Math case
		PutString(str, pBifs); // Put the string directly
	} else { 
		MediaObject *pNode = SFGenericNode().CreateNode(nNodeID);
		const BifsFieldTable *pEntry = pNode->GetBifsEntry(str);
		if (pEntry) {
			char buffer[10];
//			pBifs->EncodeInt(pEntry->m_defID, pNode->GetBifsBounds()->m_nBitsPerDef, "fieldReference");
			_itoa(pEntry->m_defID, buffer, 10);
			PutNumber(buffer, pBifs);
		} else {
			fprintf(stderr, "unrecognized field\n");
			exit(-1);
		}
	}
}

void CParser::ObjectMethodCall(BIFSEncoder *pBifs, int startPos, int opPosition, int endPos)
{
	CTOKEN_CODE curTC;

	Expression(pBifs, startPos, opPosition, TRUE); // The flag is set for "Math" checking
	curTC = m_buffer[opPosition];
	CheckToken(CTID_PERIOD);
	curTC = m_buffer[opPosition+1];
	CheckToken(CTID_IDENTIFIER);
	const char *str = m_strBuffer.getStr();
	PutString(str, pBifs); // Put the string directly
	curTC = m_buffer[opPosition+2];
	CheckToken(CTID_LEFT_CURVE);
	Params(pBifs, opPosition+3, endPos-1);
	curTC = m_buffer[endPos-1];
	CheckToken(CTID_RIGHT_CURVE);
}

void CParser::ArrayDereference(BIFSEncoder *pBifs, int startPos, int opPosition, int endPos)
{
	CTOKEN_CODE curTC;

	Expression(pBifs, startPos, opPosition);
	curTC = m_buffer[opPosition];
	CheckToken(CTID_LEFT_BRACKET);
	compoundExpression(pBifs, opPosition+1, endPos-1);
	curTC = m_buffer[endPos-1];
	CheckToken(CTID_RIGHT_BRACKET);
}

void CParser::ObjectConstruct(BIFSEncoder *pBifs, int startPos, int, int endPos)
{
	CTOKEN_CODE curTC;

	curTC = m_buffer[startPos++];
	CheckToken(CTID_NEW);
	curTC = m_buffer[startPos++];
	CheckToken(CTID_IDENTIFIER);
	const char *str = m_strBuffer.getStr();
	int i;
	FIELD_TYPE type = UNKNOWN_TYPE;
//	for(i = SFWORLDNODE; i <= MFVEC3F; i++)
	for(i = SFNODE; i<=MFVEC3F; i++)
	{
		if(!strcmp(str, VSNodeField::field_reserved[i]))
		{
			type = (FIELD_TYPE)i;
			break;
		}
	}
	
	pBifs -> EncodeInt((int)type, 6, "Field ID number for new method");
	if(type == UNKNOWN_TYPE)
		nextToken.Error("Unrecognized Object type");

	curTC = m_buffer[startPos++];
	CheckToken(CTID_LEFT_CURVE);
	Params(pBifs, startPos, endPos-1);
	curTC = m_buffer[endPos-1];
	CheckToken(CTID_RIGHT_CURVE);
}

void CParser::ConditionTest(BIFSEncoder *pBifs, int startPos, int opPosition, int endPos)
{
	CTOKEN_CODE curTC;

	Expression(pBifs, startPos, opPosition);
	curTC = m_buffer[opPosition];
	CheckToken(CTID_CONDTEST);
	startPos = opPosition+1;
	opPosition = MoveUntil(CTID_CONDSEP, opPosition, endPos-1);
	if (opPosition != -1) {
		Expression(pBifs, startPos, opPosition);
		curTC = m_buffer[opPosition];
		CheckToken(CTID_CONDSEP);
		Expression(pBifs, opPosition+1, endPos);
	} else {
		fprintf(stderr, "Error: can't find : for condition test\n");
		exit(-1);
	}
}

void CParser::Params(BIFSEncoder *pBifs, int startPos, int endPos)
{
	CTOKEN_CODE curTC;

	curTC = m_buffer[startPos];
	if (curTC != CTID_RIGHT_CURVE) {
		pBifs -> EncodeInt (1, 1, "Params");
		compoundExpression(pBifs, startPos, endPos);
	} else {
		pBifs -> EncodeInt (0, 1, "No_More_Params");
	}
}
