// This software module was originally developed by Wei Wu(Rockwell Science
// Center) and Wenlong Dong(Rockwell Science Center) in the course of 
// development of the MPEG-4 Systems standard
// (ISO/IEC 14496-1).  This software module is an implementation of a part
// of one or more MPEG-4 Systems tools as specified by the MPEG-4 Systems
// standard.  ISO/IEC gives users of the MPEG-4 Systems standard free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4 Systems
// standard.  Those intending to use this software module in hardware or
// software products are advised that its use may infringe existing
// patents.  The original developer of this software module and his/her
// company, the subsequent editors and their companies, and ISO/IEC have no
// liability for use of this software module or modifications thereof in an
// implementation.  Copyright is not released for non MPEG-4 Systems
// conforming products.  The original developer retains full right to use
// the code for his/her own purpose, assign or donate the code to a third
// party and to inhibit third parties from using the code for non MPEG-4
// Systems conforming products.  This copyright notice must be included in
// all copies or derivative works.
//
// For more information or to receive an up to date version of this module,
// you may call Wei Wu at weiwu@jupiter.risc.rockwell.com or Wenlong Dong 
// at wdong@usc.edu.
//
// Copyright 1998.

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:	Script Node
FILE: 	    DecScript.cpp
AUTHOR:     Wei Wu       (Rockwell Science Center)
		    Wenlong Dong (Rockwell Science Center)
		    Yuval Fisher (Rockwell Science Center)
OVERVIEW:   Implementation of the DecScript class
----------------------------------------------------------------------*/
#include "defs.h"
#include "DecScript.h"
#include "..\..\Core\BifsDec.h"
#include "..\..\Core\NodeFld.h"
#include "vsexpr.h"
#include "vsfunction.h"
#include "vsstatement.h"

char* DecScript::binaryOp[] = 
{
	"*", "/", "%", "+", "-", "<<", ">>", ">>>", "<", "<=",
	">", ">=", "==", "!=", "&", "^", "|", "&&", "||"
};

char* DecScript::assignmentOp[] =
{
	"=", "+=", "-=", "*=", "/=", "%=", "<<=", ">>=", ">>>=",
	"&=", "^=", "|="
};

char* DecScript::unaryOp[] =
{
	"!", "~", "-", "++", "--"
};

DecScript::DecScript()
{
  numIdentifier=0;
  flag=0;
  IdCnt=0;
}

DecScript::~DecScript()
{
}

void 
DecScript::InsertExID(int num)
{
	for(int i=0; i<num; i++) {
		char ret[10];
		sprintf(ret, "var%d", i);
		savedStrBuffer.insert(ret);
		IdCnt ++;
	}
}

LPCSTR
DecScript::GetTextString()
{
	return outBuff;
}

void
DecScript::Decode(BIFSDecoder *pBifs)
{
	VsStatementList *StList;
	while (pBifs -> ParseInt(1)) {
		outBuff += "function ";
		LPCSTR str = GetIdentifier(pBifs);
		VsFunctionDef *func = beginFunction(str);
		outBuff += str;
		outBuff += "(";
		Arguments(pBifs, func); 
		outBuff += ")\n";
		StList = StatementBlock(pBifs);
		func -> endDefinition(StList);
		functions += func;
	}
	outBuff += "\0";

	CheckUnimp();
}

void 
DecScript::Arguments(BIFSDecoder *pBifs, VsFunctionDef *f)
{
	int val = pBifs -> ParseInt(1);
	ASSERT (f != NULL);
	f;	// avoid compiler warning
	while(val) {
		LPCSTR str = GetIdentifier(pBifs);
		f -> addArg(str);

		outBuff += str;
		val = pBifs -> ParseInt(1);
		if (val) {
			outBuff += ", ";
		}
	}
}

VsStatementList *
DecScript::StatementBlock(BIFSDecoder *pBifs)
{
	VsStatementList *stlist = new VsStatementList;
	VsStatement *stmnt;

	if(pBifs -> ParseInt(1)) {
		outBuff += "{\n";
		while (pBifs -> ParseInt(1)) {
			stmnt = Statement(pBifs);
			*stlist += stmnt;
		}
		outBuff += "}\n";
	} 
	else 
	{
		stmnt = Statement(pBifs);
		*stlist += stmnt;
	}

	return stlist;
}

VsStatement *
DecScript::Statement(BIFSDecoder *pBifs)
{
	int val;
	VsStatement *stmnt;

	val = pBifs -> ParseInt(NUMBITS_STATEMENT);
	switch (val) {
	case ST_IF:
		stmnt = ifStatement(pBifs);
		break;
	case ST_FOR:
		stmnt = forStatement(pBifs);
		break;
	case ST_WHILE:
		stmnt = whileStatement(pBifs);
		break;
	case ST_RETURN:
		stmnt = returnStatement(pBifs);
		outBuff += ";\n";
		break;
	case ST_BREAK:
		stmnt = breakStatement();
		outBuff += ";\n";
		break;
	case ST_CONTINUE:
		stmnt = continueStatement();
		outBuff += ";\n";
		break;
	case ST_COMPOUND_EXPR:
		stmnt = CompoundExpression(pBifs);
		outBuff += ";\n";
		break;
	default:
		stmnt = NULL;
//		fprintf(stderr, "Error: Unrecognized statement type %d\n", val);
	}

	return stmnt;
}

VsStatement *
DecScript::ifStatement(BIFSDecoder *pBifs)
{
	int val;
	VsExpr *stmnt_expr;
	VsStatementList *stmnt_else, *stmnt_blk;

	outBuff += "if (";
	stmnt_expr = CompoundExpression(pBifs);
	outBuff += ")\n";
	stmnt_blk = StatementBlock(pBifs);
	
	val = pBifs -> ParseInt(1);
	if (val) {
		outBuff += "else\n";
		stmnt_else = StatementBlock(pBifs);
	}
	else
		stmnt_else = NULL;

	return new VsIf(stmnt_expr, stmnt_blk, stmnt_else);

}

VsStatement *
DecScript::forStatement(BIFSDecoder *pBifs)
{
	VsExpr *expr1, *expr2, *expr3;
	VsStatementList *stmntlist;

	outBuff += "for (";
	expr1 = OptionalExpression(pBifs);
	outBuff += "; ";
	expr2 = OptionalExpression(pBifs);
	outBuff += "; ";
	expr3 = OptionalExpression(pBifs);
	outBuff += ")\n";
	stmntlist = StatementBlock(pBifs);

	return new VsFor(expr1, expr2, expr3, stmntlist);
}

VsStatement *
DecScript::whileStatement(BIFSDecoder *pBifs)
{
	VsExpr *expr;
	VsStatementList *stmntlist;
	outBuff += "while (";
	expr = CompoundExpression(pBifs);
	outBuff += ")\n";
	stmntlist = StatementBlock(pBifs);

	return new VsWhile(expr, stmntlist);
}

VsStatement *
DecScript::returnStatement(BIFSDecoder *pBifs)
{
	VsExpr *expr = NULL;

	outBuff += "return ";

	if (pBifs -> ParseInt(1)) {
		expr = CompoundExpression(pBifs);
	}

	return new VsReturn(expr);
}

VsStatement *
DecScript::breakStatement()
{
	outBuff += "break";

	return new VsBreak();
}

VsStatement *
DecScript::continueStatement()
{
	outBuff += "continue";

	return new VsContinue();
}

VsExpr *
DecScript::CompoundExpression(BIFSDecoder *pBifs)
{
	VsExpr *expr1, *expr2;

	expr1 = Expression(pBifs);
	if(!pBifs -> ParseInt(1))
		return expr1;
	else
	{
		outBuff += ", ";
		expr2 = CompoundExpression(pBifs);
		return new VsComma(expr1, expr2);
	}
}

VsExpr *
DecScript::OptionalExpression(BIFSDecoder *pBifs)
{
	if (pBifs -> ParseInt(1)) {
		return CompoundExpression(pBifs);
	}

	return NULL;
}

VsExpr *
DecScript::Expression(BIFSDecoder *pBifs, BOOL n_flag)
{
	VsExpr *expr1, *expr2, *expr3, *expr = NULL;
	int val = pBifs -> ParseInt(NUMBITS_EXPR_TYPE);
	LPCSTR str;

	switch(val)
	{
	case ET_IDENTIFIER:
		if(n_flag)
		{
			if(pBifs->ParseInt(1))
				outBuff += "Math"; // In this case, expr = NULL
			else
			{
				str = GetIdentifier(pBifs);
				expr = VsFunctionDef::findIdentifier(str);
				outBuff += str;
			}
		}
		else
		{
			str = GetIdentifier(pBifs);	
			expr = VsFunctionDef::findIdentifier(str);
			outBuff += str;
		}
		break;
	case ET_STRING:
		outBuff += "'";
		str = GetString(pBifs);
		expr = new VsConstant(str);
		outBuff += str;
        outBuff += "'";// ???
		break;
	case ET_NUMBER:
		str = GetNumber(pBifs);
		expr = new VsConstant(atof(str)); 
		outBuff += str;
		break;
	case ET_CURVED_EXPR:
		outBuff += "(";
		expr = CompoundExpression(pBifs);
		outBuff += ")";
		break;
	case ET_OBJECT_METHOD_CALL:
		expr = ObjectMethodCall(pBifs);
		break;
	case ET_OBJECT_MEMBER_ACCESS:
		expr = ObjectMemberAccess(pBifs);
		break;
	case ET_FUNCTION_CALL:
		expr = FunctionCall(pBifs);
		break;
	case ET_ARRAY_DEREFERENCE:
		expr = ArrayDeref(pBifs);
		break;
	case ET_NEW:
		expr = NewObject(pBifs);
		break;
	case ET_NOT:
		str = unaryOp[0];
		outBuff += str;
		expr1 = Expression(pBifs);
		expr = new VsULogicOp(expr1, ET_NOT);
		break;
	case ET_ONESCOMP:
		str = unaryOp[1];
		outBuff += str;
		expr1 = Expression(pBifs);
		expr = new VsUArithOp(expr1, ET_ONESCOMP);
		break;
	case ET_NEGATIVE:
		str = unaryOp[2];
		outBuff += str;
		expr1 = Expression(pBifs);
		expr = new VsUArithOp(expr1, ET_MINUS);
		break;
	case ET_INCREMENT:
		str = unaryOp[3];
		outBuff += str;
		expr1 = Expression(pBifs);
		expr = new VsIncDec(expr1, TRUE, TRUE);
		break;
	case ET_DECREMENT:
		str = unaryOp[4];
		outBuff += str;
		expr1 = Expression(pBifs);
		expr = new VsIncDec(expr1, FALSE, TRUE);
		break;
	case ET_POST_INCREMENT:
		expr1 = Expression(pBifs);
		outBuff += "++";
		expr = new VsIncDec(expr1, TRUE, FALSE);
		break;
	case ET_POST_DECREMENT:
		expr1 = Expression(pBifs);
		outBuff += "--";
		expr = new VsIncDec(expr1, FALSE, FALSE);
		break;
	case ET_MULTIPLY:
	case ET_DIVIDE:
	case ET_MOD:
	case ET_PLUS:
	case ET_MINUS:
	case ET_LSHIFT:
	case ET_RSHIFT:
	case ET_RSHIFTFILL:
	case ET_AND:
	case ET_XOR:
	case ET_OR:
		expr1 = Expression(pBifs);
		outBuff += binaryOp[val - ET_MULTIPLY];
		expr2 = Expression(pBifs);
		expr = new VsArith(expr1, expr2, val);
		break;
	case ET_LT:
	case ET_LE:
	case ET_GT:
	case ET_GE:
	case ET_EQ:
	case ET_NE:
	case ET_LAND:
	case ET_LOR:
		expr1 = Expression(pBifs);
		outBuff += binaryOp[val - ET_MULTIPLY];
		expr2 = Expression(pBifs);
		expr = new VsLogicOp(expr1, expr2, val);
		break;	
	case ET_CONDTEST:
		expr1 = Expression(pBifs);
		outBuff += " ? ";
		expr2 = Expression(pBifs);
		outBuff += " : ";
		expr3 = Expression(pBifs);
		expr = new VsCondExpr(expr1, expr2, expr3);
		break;
	case ET_ASSIGN:
		expr1 = Expression(pBifs);
		outBuff += assignmentOp[0];
		expr2 = Expression(pBifs);
		expr = new VsAssign((VsVar *)expr1, expr2);
		break;
	case ET_PLUSEQ:
		expr1 = Expression(pBifs);
		outBuff += assignmentOp[1];
		expr2 = Expression(pBifs);
		expr3 = new VsArith(expr1, expr2, ET_PLUS);
		expr = new VsAssign((VsVar *)expr1, expr3);
		break;
	case ET_MINUSEQ:
		expr1 = Expression(pBifs);
		outBuff += assignmentOp[2];
		expr2 = Expression(pBifs);
		expr3 = new VsArith(expr1, expr2, ET_MINUS);
		expr = new VsAssign((VsVar *)expr1, expr3);
		break;
	case ET_MULTIPLYEQ:
		expr1 = Expression(pBifs);
		outBuff += assignmentOp[3];
		expr2 = Expression(pBifs);
		expr3 = new VsArith(expr1, expr2, ET_MULTIPLY);
		expr = new VsAssign((VsVar *)expr1, expr3);
		break;
	case ET_DIVIDEEQ:
		expr1 = Expression(pBifs);
		outBuff += assignmentOp[4];
		expr2 = Expression(pBifs);
		expr3 = new VsArith(expr1, expr2, ET_DIVIDE);
		expr = new VsAssign((VsVar *)expr1, expr3);
		break;
	case ET_MODEQ:
		expr1 = Expression(pBifs);
		outBuff += assignmentOp[5];
		expr2 = Expression(pBifs);
		expr3 = new VsArith(expr1, expr2, ET_MOD);
		expr = new VsAssign((VsVar *)expr1, expr3);
		break;
	case ET_LSHIFTEQ:
		expr1 = Expression(pBifs);
		outBuff += assignmentOp[6];
		expr2 = Expression(pBifs);
		expr3 = new VsArith(expr1, expr2, ET_LSHIFT);
		expr = new VsAssign((VsVar *)expr1, expr3);
		break;
	case ET_RSHIFTEQ:
		expr1 = Expression(pBifs);
		outBuff += assignmentOp[7];
		expr2 = Expression(pBifs);
		expr3 = new VsArith(expr1, expr2, ET_RSHIFT);
		expr = new VsAssign((VsVar *)expr1, expr3);
		break;
	case ET_RSHIFTFILLEQ:
		expr1 = Expression(pBifs);
		outBuff += assignmentOp[8];
		expr2 = Expression(pBifs);
		expr3 = new VsArith(expr1, expr2,  ET_RSHIFTFILL);
		expr = new VsAssign((VsVar *)expr1, expr3);
		break;
	case ET_ANDEQ:
		expr1 = Expression(pBifs);
		outBuff += assignmentOp[9];
		expr2 = Expression(pBifs);
		expr3 = new VsArith(expr1, expr2, ET_AND);
		expr = new VsAssign((VsVar *)expr1, expr3);
		break;
	case ET_XOREQ:
		expr1 = Expression(pBifs);
		outBuff += assignmentOp[10];
		expr2 = Expression(pBifs);
		expr3 = new VsArith(expr1, expr2, ET_XOR);
		expr = new VsAssign((VsVar *)expr1, expr3);
		break;
	case ET_OREQ:
		expr1 = Expression(pBifs);
		outBuff += assignmentOp[11];
		expr2 = Expression(pBifs);
		expr3 = new VsArith(expr1, expr2, ET_OR);
		expr = new VsAssign((VsVar *)expr1, expr3);
		break;
	default:
		Error("Error: illegal expression type");
	}

	return expr;
}

VsExpr *
DecScript::NewObject(BIFSDecoder *pBifs)
{
	VsExpr *expr = NULL;
	VsExprList *exprlist;
	FIELD_TYPE fType;

	outBuff += "new ";
	fType = (FIELD_TYPE)pBifs -> ParseInt(6);
	outBuff += "Field"; // do not care the output
	outBuff += "(";
	exprlist = Params(pBifs);
	outBuff += ")";

	Error("The new method has not been implemented yet");
	expr = new VsNew(fType, *exprlist);

	return expr;
}

VsExpr *
DecScript::ArrayDeref(BIFSDecoder *pBifs)
{
	VsExpr *expr1, *expr2;
	expr1 = Expression(pBifs);
	outBuff += "[";
	expr2 = CompoundExpression(pBifs);
	outBuff += "]";

	return new VsArrayVar(expr1, expr2);
}

VsExpr *
DecScript::FunctionCall(BIFSDecoder *pBifs)
{
	VsExpr *expr1, *expr = NULL;
	VsExprList *expr2;
	LPCSTR str;
	str = GetIdentifier(pBifs);
	expr1 = new VsConstant(str);
	outBuff += str;
	outBuff += " (";
	expr2 = Params(pBifs);
	outBuff += ")";

	VsFunctionDef *funcDef = findFunction(str);
	if(funcDef != NULL) 
	{
		expr = new VsFunctionCall(funcDef, *expr2);
	}
	else if (VsFunctionCall::isAnUnimplementedFunction(str))
	{
		char msg[256];

		sprintf(msg, "Built-in function \'%s\' is not implemented yet", str);
		Error(msg); 
	}
	else
	{
		/*
		char msg[256];

		sprintf(msg, "Function \'%s\' must be defined before it can be used", str);
		Error(msg);
		*/
		expr = new VsFunctionCall(NULL, *expr2);
		Unimplist += (VsFunctionCall *)expr;
		UnimpNames += str; 
	}

	return expr;
}		

VsExpr *
DecScript::ObjectMemberAccess(BIFSDecoder *pBifs)
{
	VsExpr *expr1, *expr = NULL;
	LPCSTR str;

	expr1 = Expression(pBifs, TRUE); // TRUE for "Math" checking
	outBuff += ".";
	if(expr1==NULL) { // NODEFIELD
		str = GetString(pBifs);
	} else {
		str = GetNumber(pBifs);
	}
	outBuff += str;	
//	if(expr1 == NULL)
		expr = new VsMemberAccess(expr1, str); // NODEFIELD
//	else
//	{
//		char msg[256];

//		sprintf(msg, "Access to object member \'%s\' is not implemented yet", str);
//		Error(msg);
//	}

	return expr;
}

VsExpr *
DecScript::ObjectMethodCall(BIFSDecoder *pBifs)
{
	VsExpr *expr1, *expr = NULL;
	VsExprList *exprlist;
	LPCSTR str;

	expr1 = Expression(pBifs, TRUE); // TRUE for "Math" checking

	outBuff += ".";
	str = GetString(pBifs);
	outBuff += str;
	outBuff += "(";
	exprlist = Params(pBifs);
	outBuff += ")";

	if(expr1 == NULL)
		expr = new VsMethodCall(expr1, str, *exprlist);
	else 
	{
		char msg[256];

		sprintf(msg, "Object method call \'%s\' is not implemented yet", str);
		Error(msg);
	}

	return expr;
}

VsExprList *
DecScript::Params(BIFSDecoder *pBifs)
{
	int val;
	VsExprList *exprlist = new VsExprList;
	VsExpr *expr;

	val = pBifs -> ParseInt(1);
	while (val) {
		expr = Expression(pBifs);
		*exprlist += expr;

		val = pBifs -> ParseInt(1);
		if(val) {
			outBuff += ", ";		
		}
	}

	return exprlist;
}

LPCSTR
DecScript::GetIdentifier(BIFSDecoder *pBifs)
{
	int calNbits(int num);
	char *str, *ret;

	if (pBifs -> ParseInt(1)) {
		int nbits, index;

		if (flag) {
			nbits = calNbits(savedStrBuffer.getPos()-1);
			index = pBifs -> ParseInt(nbits);
			str = savedStrBuffer[index];
			ret = _strdup(str);
		}
		else {
			nbits = calNbits(IdCnt-1);
			index = pBifs -> ParseInt(nbits);
			ret = (char *) malloc(10);
			sprintf(ret, "var%d", index);
		}
		return ret;
	} else {
		if (flag) {
			ret = (char *)GetString(pBifs);
			savedStrBuffer.insert(ret);
		} else {
			ret = (char *) malloc(10);
			sprintf(ret, "var%d", IdCnt);
			IdCnt++;
		}
		return ret;
	}
}

LPCSTR
DecScript::GetNumber(BIFSDecoder *pBifs)
{
	int val;
	char *ret;

	ret = (char *) malloc(40);
	if (pBifs -> ParseInt(1)) { // integer
		int nbits;

		nbits = pBifs -> ParseInt(5);
		val = pBifs -> ParseInt(nbits);
		sprintf(ret, "%d", val);
	} else { // real number
		ZString str;
		str = "";
		val = pBifs -> ParseInt(4);
		while (val != 15) {
			if (val>=0 && val<=9) {
				str += (char) (val + '0');
			} else if (val==10) {
				str += '.';
			} else if (val==11) {
				str += 'E';
			} else {
				fprintf(stderr, "Error: %d is not a valid \
real number character code\n", val);
				exit (-1);
			}
			val = pBifs -> ParseInt(4);
		}
		strcpy(ret, (LPCSTR) str);
	}
	return ret;
}

LPCSTR
DecScript::GetString(BIFSDecoder *pBifs)
{
	SFString string;
	LPCSTR str;
	
	string.Parse(pBifs, NULL);
	str = _strdup((LPCSTR)string);
	return str;
}

void
DecScript::addGlobalVar(LPCSTR name, VsVar *var)
{
    globalNames += name;
    globalVars += var;

    var->ref();
}

VsVar *
DecScript::getGlobalVar(LPCSTR name)
{
    // Look through globals:
    for (int i = 0; i < globalVars.GetLength(); i++) {
		if(!strcmp(globalNames[i], name))
			return globalVars[i];
    }

    return NULL;
}

VsFunctionDef *
DecScript::beginFunction(LPCSTR name)
{
    VsFunctionDef *result = new VsFunctionDef(name, this);

    // ??? Warn if two functions with same name ???
    // Call findFunction to test if the current function name is 
    // already used.    --Dong
    if(findFunction(name) != NULL) {
      fprintf(stderr, "Error: The function: \"%s\" is already defined\n",name);
      exit(1);
    //  return NULL;
    }      

    functions += result;
    return result;
}

VsFunctionDef *
DecScript::findFunction(LPCSTR name)
{
    for (int i = 0; i < functions.GetLength(); i++) {
	VsFunctionDef *f = (VsFunctionDef *)functions[i];
	if (strcmp(f->getName(), name) == 0)
	    return f;
    }
    return NULL;
}

void 
DecScript::CheckUnimp()
{
	for(int i = 0; i < Unimplist.GetLength(); i++)
	{
		VsFunctionDef *f = findFunction(UnimpNames[i]);

		if(f == NULL)
		{
			char msg[256];
			sprintf(msg, "Function: %s is not defined", UnimpNames[i]);
			Error(msg);
		}
		else
			Unimplist[i]->SetFuncDef(f);
	}

	UnimpNames.Clear();
}

void
DecScript::Error(LPCSTR errmsg)
{
//    char *msg = (char *)malloc(strlen(errmsg)+50);

//    sprintf(msg, "Script node: %s\n",errmsg);
//    ::MessageBox((HWND)NULL, msg, "VrmlScript syntax error", MB_OK );

//    free(msg);
	fprintf(stderr, "Script node: %s\n", errmsg);
}








