/*
 * VsStatement.h
 *
 *      Declaration of classes:
 *		VsValue
 *        VsStatement
 *        VsReturn
 *        VsIf
 *        VsFor
 *        VsStatementList
 *
 * Statement support.  Functions are lists of statements that get
 * executed.
 * 
 * Copyright (C) 1996, Silicon Graphics,  Inc.
 */

/*---------------------------------------------------------------
This module is copied from the VRML Script interpreter owned by
Silicon Graphics, Inc. It is modified by Wenlong Dong, Wei Wu and
Yuval Fisher(Rockwell Science Center) in the course of development
of the MPEG-4 Systems standard(ISO/IEC 14496-1) and added to the 
subsystem: Script Node.
---------------------------------------------------------------*/

#ifndef _VS_STATEMENT_
#define _VS_STATEMENT_

#include "ztl.h"

class VsStatementList;
class VsFunctionCall;
class VsExpr;
class VsVar;

//
// VsValues are passed around and represent values of various types.
// Since JavaScript is pretty much untyped, having one value class
// makes things a lot cleaner.  KEEP THIS STRUCTURE AS SMALL AS
// POSSIBLE!  VsValues tend to be passed around by value, so should be kept
// as lightweight as possible.
//

typedef int int32_t;
typedef long uint32_t;
  
struct FloatArray 
{
    float *array;
    int32_t length;
};

struct Int32Array 
{
    int32_t *array;
    int32_t length;
};

struct VoidArray 
{
    void **array;
    int32_t length;
};

struct StringArray 
{
    ZString *array;
    int32_t length;
};

struct VsValue 
{
    VsValue(double num);
    VsValue(void *ptr);
    VsValue(char *str);
    VsValue() { };

    int		getBool() const;  // Returns value as boolean (0 or 1)

    union 
	{
		double number;
		char  *string;
		void  *data;
		FloatArray *floatArray;
		Int32Array *int32Array;
		VoidArray  *voidArray;
		StringArray *stringArray;
    };

    enum Type 
	{
		vUNKNOWN, vVOID,
		vNUMBER, vSTRING,
		vOBJECT, vFUNCTION, 
		vFLOAT_ARRAY,
		vINT32_ARRAY,
		vSTRING_ARRAY,
		vVOID_ARRAY,
		vROTATION,
    };

    unsigned char type; // One of the above ENUM values

    static const VsValue nothing;
};    

//
// Statement base class:
//
class VsStatement 
{
public:
    // These status codes are returned from statements to tell the
    // enclosing controls structure (loop, if-test, function) what to
    // do:
    enum Status 
	{
		NORMAL,   // Normal return, continue executing statements
		BREAK,    // Break out of enclosing loop/if
		CONTINUE, // Continue enclosing loop/if
		RETURN,   // Return from enclosing function
    };

    VsStatement();
    virtual ~VsStatement();

    // Statement subclasses should override this to do whatever they
    // do and return a Status code.  References to function arguments
    // (which are buried down inside expressions) need to know what
    // function they're executing inside of, which is why a
    // VsFunctionCall is passed in to this routine:
    virtual Status	execute(VsFunctionCall *) = 0;
};

//
// Return statement.
//
class VsReturn : public VsStatement 
{
public:
    VsReturn(VsExpr *expr);
    virtual ~VsReturn();

    // Standard execute.  Evalutes expression (if any given), returns
    // a status of RETURN.
    virtual VsStatement::Status	execute(VsFunctionCall *);

private:
    VsExpr *expression;
};

//
// If statement.
//
class VsIf : public VsStatement 
{
public:
    // Pass in expression for condition, if statements, and else
    // statements (NULL if just a "bare" if).
    VsIf(VsExpr *condition, VsStatementList *ifStatements,
	VsStatementList *elseStatements);
    virtual ~VsIf();

    // Evaluates.  Depending on condition expression, either if or
    // else statements are executed.
    virtual VsStatement::Status	execute(VsFunctionCall *);

private:
    VsExpr *condition;
    VsStatementList *ifStatements, *elseStatements;
};

//
// For statement.
//
class VsFor : public VsStatement 
{
public:

    VsFor(VsExpr *initial1, VsExpr *condition1, VsExpr *increment1, 
	VsStatementList *body1);
    virtual ~VsFor();

    // Execute statements repeatedly until terminate expression is
    // true.
    virtual VsStatement::Status	execute(VsFunctionCall *);

private:

    VsExpr *initial, *increment, *condition;
    VsStatementList *body;
};

//
// For..In statement.
//
class VsForIn : public VsStatement 
{
public:
    VsForIn(VsVar *variable, VsExpr *object, VsStatementList *body);
    virtual ~VsForIn();

    virtual VsStatement::Status	execute(VsFunctionCall *);
    
private:
    VsVar  *variable;
    VsExpr *object;
    VsStatementList *body;
};

//
// While statement.
//
class VsWhile : public VsStatement 
{
public:
    VsWhile(VsExpr *condition, VsStatementList *body);
    virtual ~VsWhile();

    // Execute statements repeatedly until terminate expression is
    // true.
    virtual VsStatement::Status	execute(VsFunctionCall *);

private:
    VsExpr *condition;
    VsStatementList *body;
};

//
// Break statement.
//
class VsBreak : public VsStatement 
{
public:
    VsBreak();

    virtual ~VsBreak();

    // Execute statements repeatedly until terminate expression is
    // true.
    virtual VsStatement::Status	execute(VsFunctionCall *);

};

//
// Continue statement.
//
class VsContinue : public VsStatement 
{
public:

    VsContinue();
    virtual ~VsContinue();

    // Execute statements repeatedly until terminate expression is
    // true.
    virtual VsStatement::Status	execute(VsFunctionCall *);
};

//
// With statement.
//
class VsWith : public VsStatement 
{
public:
    VsWith(VsExpr *object, VsStatementList *body );

    virtual ~VsWith();

    // Execute statements repeatedly until terminate expression is
    // true.
    virtual VsStatement::Status	execute(VsFunctionCall *);

private:
	VsExpr* object;
	
	VsStatementList* body;
};

class VsStatementList : public ZArray<VsStatement *>
{
public:

	VsStatement::Status execute(VsFunctionCall *);


};

#endif /* _VS_STATEMENT_ */
