/***********************************************************************
This software module was originally developed by Mikael Bourges-Sevenier
(France Telecom CNET) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. France Telecom CNET. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@vdo.net or zvil@compuserve.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Decoder/ Compositor
FILE:       EncQuant.cpp
AUTHOR:     Mikael Bourges-Sevenier (France Telecom CNET)
OVERVIEW:   Implementation of Quantizer functions for BIFS encoding
-------------------------------------------------------------------------*/

#include "defs.h"
#include "Quantize.h"
#include "BifsEnc.h"
#include <float.h>
#include <math.h>
#include "ztl.h"

#define QPVAL(type, id) ((type) *(SF<type>*) m_pQP->GetField(id))

#ifndef BIFSCAR
BOOL Quantizer::HasEfficientFloat() 
{ 
  if(!m_pQP)  return FALSE;
  return QPVAL(Bool,EFLOAT); 
}

/*-----------------------------------------------------------------
Test is a quantizer is used
-----------------------------------------------------------------*/
BOOL Quantizer::CanQuantize () {
  if(!m_pQP) return FALSE;

  if(QPVAL(Bool,EFLOAT)) return TRUE;

  switch(m_nQuantizationCode) {
  case Q_3D:
    return QPVAL(Bool,POS3DQ); 
  case Q_2D:
    return QPVAL(Bool,POS2DQ); 
  case Q_DRAWORDER:
    return QPVAL(Bool,DOQ); 
  case Q_COLOR:
    return QPVAL(Bool,COLORQ); 
  case Q_TEXTURECOORDINATE:
    return QPVAL(Bool,TCQ); 
  case Q_ANGLE:
    return QPVAL(Bool,ANGLEQ); 
  case Q_SCALE:
    return QPVAL(Bool,SCALEQ); 
  case Q_KEY:
    return QPVAL(Bool,KEYQ); 
  case Q_NORMAL:
  case Q_ROTATION:
    return QPVAL(Bool,NORMALQ); 
  case Q_SIZE3D:
    return QPVAL(Bool,SIZEQ); 
  case Q_SIZE2D:
    return QPVAL(Bool,SIZEQ); 
  case Q_LSQ:
    return TRUE; 
  }
  return FALSE;
}
#endif

/*-----------------------------------------------------------------
Encode a float
-----------------------------------------------------------------*/
void Quantizer::EncodeFloat (BIFSEncoder *pBifs, float nValue) {
  if(m_nQuantizationCode == Q_NONE)  return;
  
  if(HasEfficientFloat()) {
    EncodeEfficientFloat (pBifs, nValue);
    return;
  }

  //fprintf(stderr,"Quantizing a float QP %d\n",m_nQuantizationCode);    
  float mini=0,maxi=0,nBits=0;
    
  switch(m_nQuantizationCode) {
  case Q_DRAWORDER:
    mini=max(m_pMin[0],QPVAL(float,DOMIN));		  
    maxi=min(m_pMax[0],QPVAL(float,DOMAX));
    nBits= QPVAL(int,DONB);
    break;
  case Q_COLOR:
    mini=max(m_pMin[0],QPVAL(float,COLORMIN));		  
    maxi=min(m_pMax[0],QPVAL(float,COLORMAX));
    nBits= QPVAL(int,COLORNB);
    break;
  case Q_ANGLE:
    mini=max(m_pMin[0],QPVAL(float,ANGLEMIN));		  
    maxi=min(m_pMax[0],QPVAL(float,ANGLEMAX));
    nBits= QPVAL(int,ANGLENB);
    break;
  case Q_SCALE:
    mini=max(m_pMin[0],QPVAL(float,SCALEMIN));		  
    maxi=min(m_pMax[0],QPVAL(float,SCALEMAX));
    nBits= QPVAL(int,SCALENB);
    break;
  case Q_KEY:
    mini=max(m_pMin[0],QPVAL(float,KEYMIN));		  
    maxi=min(m_pMax[0],QPVAL(float,KEYMAX));
    nBits= QPVAL(int,KEYNB);
    break;
  case Q_SIZE2D: 
  case Q_SIZE3D: {    
    mini=max(m_pMin[0],QPVAL(float,SIZEMIN));
    maxi=min(m_pMax[0],QPVAL(float,SIZEMAX));
    nBits=QPVAL(int,SIZENB);
   break;
  }
  case Q_LSQ:
    mini=0.f;
    maxi=1<<m_nBits;
    nBits=m_nBits;
    break;
  }
  pBifs -> EncodeInt (Quantize(mini, maxi, nBits,nValue), nBits, "quantized value");
}

/*-----------------------------------------------------------------
Encode a Vec3f
-----------------------------------------------------------------*/
void Quantizer::EncodeVec3f (BIFSEncoder *pBifs, Vec3f &nValue) {
  if(m_nQuantizationCode == Q_NONE) return;

  if(HasEfficientFloat()) {
    EncodeEfficientFloat (pBifs, nValue.x);
    EncodeEfficientFloat (pBifs, nValue.y);
    EncodeEfficientFloat (pBifs, nValue.z);
    return;
  }

  float mini[3],maxi[3],nBits=0;

  switch(m_nQuantizationCode) {
  case Q_3D:
    //fprintf(stderr,"Quantizing a Vec3f\n");
    mini[0]=max(m_pMin[0],QPVAL(Vec3f,POS3DMIN).x);
    mini[1]=max(m_pMin[0],QPVAL(Vec3f,POS3DMIN).y);
    mini[2]=max(m_pMin[0],QPVAL(Vec3f,POS3DMIN).z);
      
    maxi[0]=min(m_pMax[0],QPVAL(Vec3f,POS3DMAX).x);
    maxi[1]=min(m_pMax[0],QPVAL(Vec3f,POS3DMAX).y);
    maxi[2]=min(m_pMax[0],QPVAL(Vec3f,POS3DMAX).z);
      
    nBits= QPVAL(int,POS3DNB);
    break;
  case Q_NORMAL:
    //fprintf(stderr,"Quantizing a Normal\n");
    nBits= QPVAL(int,NORMALNB);
    EncodeNormal(pBifs,nValue,nBits);
    return;
    break;
  case Q_SCALE: {
    //fprintf(stderr,"Quantizing a Scale3D\n");
    for(int i=0;i<3;i++) {
      mini[i]=max(m_pMin[0],QPVAL(float,SCALEMIN));
      maxi[i]=min(m_pMax[0],QPVAL(float,SCALEMAX));
    }
      
    nBits= QPVAL(int,SCALENB);
    break;
  }
  case Q_SIZE3D: {
      for(int i=0;i<3;i++) {
	  mini[i]=max(m_pMin[0],QPVAL(float,SIZEMIN));	  
	  maxi[i]=min(m_pMax[0],QPVAL(float,SIZEMAX));
      }
      nBits=QPVAL(int,SIZENB);
  }
  break;
  }

  pBifs -> EncodeInt (Quantize(mini[0], maxi[0], nBits,nValue.x), nBits, "quantized value X");
  pBifs -> EncodeInt (Quantize(mini[1], maxi[1], nBits,nValue.y), nBits, "quantized value Y");
  pBifs -> EncodeInt (Quantize(mini[2], maxi[2], nBits,nValue.z), nBits, "quantized value Z");
}

/*-----------------------------------------------------------------
Encode a Vec2f
-----------------------------------------------------------------*/
void Quantizer::EncodeVec2f (BIFSEncoder *pBifs, Vec2f &nValue) {
  if(m_nQuantizationCode == Q_NONE) return;

  if(HasEfficientFloat()) {
    EncodeEfficientFloat (pBifs, nValue.x);
    EncodeEfficientFloat (pBifs, nValue.y);
    return;
  }

  float mini[2],maxi[2];
  int nBits=0;

  switch(m_nQuantizationCode) {
  case Q_2D:
    //fprintf(stderr,"Quantizing a Vec2f\n");
    mini[0]=max(m_pMin[0],QPVAL(Vec2f,POS2DMIN).x);
    mini[1]=max(m_pMin[0],QPVAL(Vec2f,POS2DMIN).y);
      
    maxi[0]=min(m_pMax[0],QPVAL(Vec2f,POS2DMAX).x);
    maxi[1]=min(m_pMax[0],QPVAL(Vec2f,POS2DMAX).y);
    nBits= QPVAL(int,POS2DNB);
    break;
  case Q_TEXTURECOORDINATE:
    //fprintf(stderr,"Quantizing a Texture Coordinate\n");
    mini[0]=max(m_pMin[0],QPVAL(float,TCMIN));
    mini[1]=max(m_pMin[0],QPVAL(float,TCMIN));
    
    maxi[0]=min(m_pMax[0],QPVAL(float,TCMAX));
    maxi[1]=min(m_pMax[0],QPVAL(float,TCMAX));
    nBits= QPVAL(int,TCNB);
    break;
  case Q_SCALE: {
      //fprintf(stderr,"Quantizing a Scale2D\n");
      for(int i=0;i<2;i++) {
	  mini[i]=max(m_pMin[0],QPVAL(float,SCALEMIN));
	  maxi[i]=min(m_pMax[0],QPVAL(float,SCALEMAX));
      }
      
      nBits= QPVAL(int,SCALENB);
      break;
  }
  case Q_SIZE2D: {
      for(int i=0;i<2;i++) {
	  mini[i]=max(m_pMin[0],QPVAL(float,SIZEMIN));
	  maxi[i]=min(m_pMax[0],QPVAL(float,SIZEMAX));
      }
      nBits=QPVAL(int,SIZENB);
  }
  break;
  }

  pBifs -> EncodeInt (Quantize(mini[0], maxi[0], nBits,nValue.x), nBits, "quantized value X");
  pBifs -> EncodeInt (Quantize(mini[1], maxi[1], nBits,nValue.y), nBits, "quantized value Y");
}

/*-----------------------------------------------------------------
Encode a Rotation
-----------------------------------------------------------------*/
//MBS 110598: new version for FDIS
void Quantizer::EncodeRotation (BIFSEncoder *pBifs, Rotation &nValue) {
  if(m_nQuantizationCode == Q_NONE) return;

  if(HasEfficientFloat()) {
    EncodeEfficientFloat (pBifs, nValue.x);
    EncodeEfficientFloat (pBifs, nValue.y);
    EncodeEfficientFloat (pBifs, nValue.z);
    EncodeEfficientFloat (pBifs, nValue.theta);
    return;
  }

  float vrmlComp[]={ nValue.x, nValue.y, nValue.z, nValue.theta };
  float quatComp[4];
  int i;

  // quaternion form (vrml components -> quaternion components
  float angle=nValue.theta;
  quatComp[0]=cos(angle/2);
  
  float delta = 0;
  for (i=0;i<3;i++) 
    delta += vrmlComp[i]*vrmlComp[i];

  delta = sin(angle/2)/sqrt(delta);
  for (i=1;i<4;i++) 
    quatComp[i] = vrmlComp[i-1]*delta;
  
  // encode
  EncodeOnUnitSphere(pBifs, quatComp, 3, QPVAL(int,NORMALNB));
}

/* old FCD N2201 version
void Quantizer::EncodeRotation (BIFSEncoder *pBifs, Rotation &nValue) {
  if(m_nQuantizationCode == Q_NONE) return;
  //fprintf(stderr,"Quantizing a Rotation\n");

  int nBitsNormal= QPVAL(int,NORMALNB);
  int nBitsAngle = QPVAL(int,ANGLENB);

  Vec3f norm(nValue.x,nValue.y,nValue.z);

  EncodeNormal(pBifs,norm,nBitsNormal);

  float mini=max(m_pMin[0],QPVAL(float,ANGLEMIN));		  
  float maxi=min(m_pMax[0],QPVAL(float,ANGLEMAX));

  pBifs -> EncodeInt (Quantize(mini, maxi, nBitsAngle,nValue.theta), nBitsAngle, "quantized value theta");
}
*/
/*-----------------------------------------------------------------
Encode a Rotation2D
-----------------------------------------------------------------*/
void Quantizer::EncodeRotation2D (BIFSEncoder * /*pBifs*/, Rotation2D & /*nValue*/) {
  if(m_nQuantizationCode == Q_NONE) return;
  fprintf(stderr,"Arrrghhh: Don't know how to quantize a Rotation2D\n");
}

/*-----------------------------------------------------------------
Encode a Color3f
-----------------------------------------------------------------*/
void Quantizer::EncodeColor3f (BIFSEncoder *pBifs, Color3f &nValue) {
  if(m_nQuantizationCode == Q_NONE) return;
  if(HasEfficientFloat()) {
    EncodeEfficientFloat (pBifs, nValue.r);
    EncodeEfficientFloat (pBifs, nValue.g);
    EncodeEfficientFloat (pBifs, nValue.b);
    return;
  }

  //fprintf(stderr,"Quantizing a Color3f");
  float mini= max(m_pMin[0],QPVAL(float,COLORMIN));
  float maxi= min(m_pMax[0],QPVAL(float,COLORMAX));
  int nBits= QPVAL(int,COLORNB);
  
  pBifs -> EncodeInt (Quantize(mini, maxi, nBits,nValue.r), nBits, "quantized value R");
  pBifs -> EncodeInt (Quantize(mini, maxi, nBits,nValue.g), nBits, "quantized value G");
  pBifs -> EncodeInt (Quantize(mini, maxi, nBits,nValue.b), nBits, "quantized value B");
}

/*-----------------------------------------------------------------
Linear Scalar Quantizer
-----------------------------------------------------------------*/
int Quantizer::Quantize(float mini, float maxi, int nbits, float v) {
  if(v<=mini) return 0;
  if(v>=maxi) return (1<<nbits)-1;
  return (int) floor( ((v-mini)/(maxi-mini)*((1<<nbits)-1)) + 0.5);
}

/*---------------------------------------------------------------------
Encode on a unit sphere. For rotation (quaternions) & normals (sphere)
----------------------------------------------------------------------*/
void Quantizer::EncodeOnUnitSphere(BIFSEncoder *pBifs, float *comp, int nbComp, int nBits) 
{
  int len=nbComp+1;
  int orientation =-1;
  float maxTmp=-FLT_MAX;

  for (int i=0;i<len;i++) {
    if (fabs(comp[i])>maxTmp) {
      maxTmp=fabs(comp[i]);
      orientation=i;
    }
  }

  if(nbComp==2) {
    pBifs -> EncodeInt (orientation, 2, "quantized normal: orientation");
    pBifs -> EncodeInt (((comp[orientation]>0) ? 1 : 0), 1, "quantized normal: direction");
  }
  else
    pBifs -> EncodeInt (orientation, 2, "quantized rotation: orientation");

  for (int j=0;j<nbComp;j++) {
    float v = 4/M_PI * atan( comp[(orientation+j+1) % len] / comp[orientation]);
    if(nbComp==2) 
      pBifs -> EncodeInt ( (1<<(nBits-1))+Quantize(0, 1, nBits-1, v), nBits, "quantized normal: reduced component");
    else
      pBifs -> EncodeInt (  (1<<(nBits-1))+Quantize(0, 1, nBits-1, v), nBits, "quantized rotation: reduced component");
  }
}

/*-----------------------------------------------------------------
Encode a normal
-----------------------------------------------------------------*/
//MBS 110598: new version for FDIS
void Quantizer::EncodeNormal(BIFSEncoder *pBifs, Vec3f &norm, int nbBits) {
  if(m_nQuantizationCode == Q_NONE) return;

  float natComp[]={ norm.x, norm.y, norm.z };

  // normalize vector
  float delta=0;
  int i;

  for(i=0;i<3;i++)
    delta += natComp[i] * natComp[i];
  delta=sqrt(delta);

  if(delta != 0)
    for(i=0;i<3;i++)
      natComp[i] /= delta;
  
  // encode
  EncodeOnUnitSphere(pBifs, natComp, 2, nbBits);
}

/*-----------------------------------------------------------------
Encode an efficient float value
MBS 110698
-----------------------------------------------------------------*/
void Quantizer :: EncodeEfficientFloat (BIFSEncoder *pBifs, float fValue)
{
#ifdef WIN32
  union
  {	
    float f;
    long l;
  } value;
  value.f = fValue;
  
  // 1b sign, 8b exponent + 127 bias, 23b mantissa & leading one removed
  // in BIFS: mantissa is max 15b, exponent 7b and sign bits for both!
  
  if(fValue==0) {
    pBifs->EncodeInt(0,4,"mantissaLength");	  
    return;
  }
  
  unsigned int mantissaSign=((value.l & 0x80000000)>>31) & 0x1;
  unsigned int mantissa=(value.l & 0x007FFFFF)>>8;
  unsigned int mantissaLength=15;
  unsigned int exponentSign=0;
  char exponent=(((value.l & 0x7F800000)>>23)-127);
  unsigned int exponentLength=8;
  
  // coding mantissa
  if(mantissa !=0) {
    int shrink=0;
    while((mantissa & (1<<shrink))==0)
      shrink++;
    mantissa>>=shrink;
    mantissaLength -= shrink;
  }
  else 
    mantissaLength=1;
  
  // coding exponent
  if(exponent != 0) {
    if(exponent < 0) {
      exponentSign=1;
      exponent=-exponent;	    
    }
    
    while((exponent & (1<<(--exponentLength)))==0);
    
    // remove leading 1
    exponent &= ~(1<<exponentLength);
    exponentLength++;
  }
  else 
    exponentLength=0;

  pBifs->EncodeInt(mantissaLength,4,"mantissaLength");
  if(mantissaLength) {
    pBifs->EncodeInt(exponentLength,3,"exponentLength");
    pBifs->EncodeInt(mantissaSign,1,"mantissaSign");
    pBifs->EncodeInt(mantissa,mantissaLength,"mantissa");
    if(exponentLength) {
      pBifs->EncodeInt(exponentSign,1,"exponentSign");
      pBifs->EncodeInt(exponent,exponentLength,"exponent");      
    }
  }
#endif		// WIN32
}

/* old version of CD N2201
   void Quantizer::EncodeNormal(BIFSEncoder *pBifs, Vec3f &norm, int nbBits) {
   int octant=0;
   double maxi=-1.;
   int triant=0;

   octant<<=1;
   if(norm.z<0) octant |=1;    
   if(fabs(norm.x)>maxi) {
   maxi=fabs(norm.x);
   triant=0;
   }
   octant<<=1;
   if(norm.y<0) octant |=1;    
   if(fabs(norm.y)>maxi) {
   maxi=fabs(norm.y);
   triant=1;
   }
   octant<<=1;
   if(norm.x<0) octant |=1;    
   if(fabs(norm.z)>maxi) {
   maxi=fabs(norm.z);
   triant=2;
   }
   pBifs -> EncodeInt (octant, 3, "quantized normal: octant");
   pBifs -> EncodeInt (triant, 2, "quantized normal: triant");
  
   double a=0.,b=0.;
   switch(triant) {
   case 0: // x>y, x>z
   a=atan(fabs(norm.y/norm.x));
   b=atan(fabs(norm.z/norm.x));
   break;
   case 1: // y>x, y>z
   a=atan(fabs(norm.x/norm.y));
   b=atan(fabs(norm.z/norm.y));
   break;
   case 2: // z>x, z>y
   a=atan(fabs(norm.x/norm.z));
   b=atan(fabs(norm.y/norm.z));
   break;
   }
   a*=4/M_PI;
   b*=4/M_PI;

   pBifs -> EncodeInt ((int) (a*((1<<nbBits)-1)), nbBits, "quantized normal: xPos");
   pBifs -> EncodeInt ((int) (b*((1<<nbBits)-1)), nbBits, "quantized normal: yPos");
   }
*/
