/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) and edited by Mauro Rossi (CSELT), Guido Franceschini
(CSELT), George Tselikis (Xbind Inc.) and Michael Stepping (FTK)
in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Core
FILE:       AL.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of ALManager
-------------------------------------------------------------------------*/
#include <iostream.h>

#include "defs.h"
#include "AL.h"
#include "Channel.h"
#include "Executiv.h"
#include "StreamCo.h"
#include "ODP.h"
#include "Stream.h"
#include "Inline.h"
#include "DAI.h"

/*-----------------------------------------------------------------
Create a DataChannel object, configure it, and add it to a list
of channels waiting to be opened.
-----------------------------------------------------------------*/
DataChannel *ALManager :: AttachStream (
	ES_Descriptor *pESD,
	MediaStreamImp *pStream,
	SLConfigDescriptor *pConfigParams,
	InlineImp *pInline,
	int nExpectedAULength
)
{
	int nESid = pESD -> ES_ID;
	DataChannel *pChannel = new DataChannel (this, nESid, pESD);
	m_mtxChannels.P ();
	m_channels += pChannel;
	/*-------------------------------------------------------------
	Attach the stream to the proper clock reference (create one if
	necessary)
	-------------------------------------------------------------*/
	if (pStream)
	{	int nClockID = pConfigParams -> OCR_ES_Id;
		if (nClockID == 0)
			nClockID = nESid;
		ClockReference *pClock = pInline -> GetClock (nClockID);
		if (nClockID == nESid)		// reset the clock if I am the OCR
			pClock -> Reset ();
		pStream -> SetClockReference (pClock);
	}
	pChannel -> Config (pStream, pConfigParams, nExpectedAULength);
	m_mtxChannels.V ();
	return pChannel;
}

/*-----------------------------------------------------------------
Request the opening of all the channels which are marked READY.
If service is not connected yet, only mark the channels as "required"
If service has been detached, reattach it.
-----------------------------------------------------------------*/
void ALManager :: RequestChannels ()
{
	bool bReadyChannels = false;
	m_mtxChannels.P ();
	for (int i = 0; i < m_channels.GetLength (); i++)
		bReadyChannels |= m_channels [i] -> Request ();
	m_mtxChannels.V ();
	if (bReadyChannels)
	{	if (m_state == SS_IDLE)
			Connect ();
		if (m_state == SS_CONNECTED)
			AddChannels ();
	}
}

/*-------------------------------------------------------------
Request the service to add all the channels that are marked "required"
-------------------------------------------------------------*/
void ALManager :: AddChannels ()
{
	DAIChannelAddParam inParam;
	
	m_mtxChannels.P();
	for (int i = 0; i < m_channels.GetLength(); i++)
	{
		if(m_channels[i] -> IsRequired())
		{
			DAIChannelAddElement &inEl = ++inParam;
			inEl.m_pData = m_channels [i] -> GetuuData ();
			inEl.m_nDataLen = m_channels [i] -> GetuuDataLength ();
			inEl.m_pQoSStream = m_channels [i] -> GetQoS() ;
			inEl.m_eDirection	= DOWNSTREAM;
			inEl.m_nTmpId	= (UINT) m_channels [i];
		}
	}
	m_mtxChannels.V();

	if (inParam.GetLength() > 0)
		m_pExecutive -> ChannelAdd (m_nServiceID, inParam);
}

/*-----------------------------------------------------------------
When service is connected, request all pending channels and notify
the application.
-----------------------------------------------------------------*/
void ALManager :: OnServiceAttach (
	UINT nServiceID, 
	const BYTE* pUuData,
	int nUuDataLen,
	RESPONSE eResponse
)
{
	if (eResponse == RESPONSE_OK)
	{	
		m_nServiceID = nServiceID;
		m_state = SS_CONNECTED;
		AddChannels ();		
		/*-------------------------------------------------------------
		If this is a secondary session with remoteOD, notify the node
		that the session is established.
		-------------------------------------------------------------*/
		if (m_pNode)
			m_pNode -> OnConnect (pUuData, nUuDataLen);
	}
	else
	{	/*-------------------------------------------------------------
		Notifies the application and mark the service as idle if failed.
		-------------------------------------------------------------*/
		m_pExecutive -> Error (INVALID_URL);
		ChannelsClosed ();
		m_state = SS_UNAVAILABLE;
		m_pExecutive ->DeleteService (this);
	}
}

/*-------------------------------------------------------------
Close all channels when DMIF reports service detach
-------------------------------------------------------------*/
void ALManager :: OnServiceDetach (REASON reason)
{
	ChannelsClosed ();
	m_state = SS_IDLE;
	m_pExecutive ->DeleteService (this);
	if (reason != REASON_OK)
	{	m_pExecutive -> Error (UNEXPECTED_SERVICE_DETACH);
		m_state = SS_UNAVAILABLE;
	}
}

/*-----------------------------------------------------------------
When channels are added notify the DataChannel objects.
-----------------------------------------------------------------*/
void ALManager :: OnChannelAdd (const DAIChannelAddParam &cnfParam)
{
	for (int j=0; j < cnfParam.GetLength(); j++)
	{	DataChannel *pChannel = NULL;
		m_mtxChannels.P ();
		for (int i = m_channels.GetLength (); --i >= 0;)
		{
			if ((UINT) m_channels [i] == cnfParam [j].m_nTmpId)
			{	pChannel = m_channels [i];
				pChannel -> Lock ();
				break;
			}
		}
		m_mtxChannels.V ();
		if (pChannel != NULL)
		{	pChannel -> OnChannelOpen (cnfParam [j].m_nChid, cnfParam [j].m_eResponse);
			pChannel -> Unlock ();
		}
	}
}

/*-------------------------------------------------------------
When service terminated notify all the channels to close
-------------------------------------------------------------*/
void ALManager :: ChannelsClosed ()
{
	m_mtxChannels.P ();
	for (int i = m_channels.GetLength (); --i >= 0;)
		m_channels [i] -> OnChannelDelete (REASON_OK);
	m_mtxChannels.V ();
}

/*-----------------------------------------------------------------
Connect to the given URL
-----------------------------------------------------------------*/
void ALManager :: Connect ()
{
	m_state = SS_CONNECTING;
	m_pExecutive -> ServiceAttach (this, m_url);
}

/*-----------------------------------------------------------------
Request a service detach
-----------------------------------------------------------------*/
void ALManager :: Disconnect ()
{
	bool bWasConnected = false;
	m_pExecutive ->DeleteService (this);
	Lock ();
	if (m_state == SS_CONNECTING || m_state == SS_CONNECTED)
	{	m_state = SS_IDLE;
		bWasConnected = true;
	}
	Unlock ();
	if (bWasConnected)
		m_pExecutive -> ServiceDetach (this, REASON_OK);
}

/*-----------------------------------------------------------------
Make sure that the service is closed before the object is deleted.
-----------------------------------------------------------------*/
ALManager :: ~ALManager ()
{
	Disconnect ();
}

/*-------------------------------------------------------------
This function is called by DMIF in order to notify the application
about the result of the channel release request.
-------------------------------------------------------------*/
void ALManager :: OnChannelDeleteCnf (const DAIChannelDeleteParam& cnfParam)
{
	for (int j=0; j < cnfParam.GetLength(); j++)
		if (cnfParam [j].m_eResponse != RESPONSE_OK)
			m_pExecutive ->Error (CHANNEL_DELETE_FAILED);
}

/*-------------------------------------------------------------
Remove a channel from the list
-------------------------------------------------------------*/
void ALManager :: DeleteChannel (DataChannel *pChannel)
{
	m_mtxChannels.P ();
	m_channels -= pChannel;
	m_mtxChannels.V ();
}

static UINT nCommands = 0;
/*-------------------------------------------------------------
This function is called in order to send a user command,like PLAY,to the server.
This user command should be always referred to an already established channel.
For i.e in case of PLAY command this channel is going to receive the data.
-------------------------------------------------------------*/
void ALManager :: SendCommand (char *szCommand, UINT nChannelID)
{
	DAIUserCmdParam inParam;
	inParam.m_nCmdId = nCommands++;
	inParam.m_pChid = &nChannelID;
	inParam.m_nChLen = 1;
	inParam.m_nDataLen = strlen (szCommand) + 1;
	inParam.m_pData = szCommand;
	m_pExecutive -> SendCommand (GetServiceID (), inParam);
}
 
/*-------------------------------------------------------------
Request the DAI to delete a channel
-------------------------------------------------------------*/
void ALManager :: RequestChannelDelete (DataChannel *pChannel)
{
	m_pExecutive ->DeleteChannel (pChannel);
	DAIChannelDeleteParam param;
	DAIChannelDeleteElement &element = ++param;
	element.m_nChid = pChannel ->GetChannelID ();
	element.m_eResponse = RESPONSE_OK;
	element.m_eReason = REASON_OK;
	m_pExecutive ->ChannelDelete (m_nServiceID, param);
}

/*-------------------------------------------------------------
Display a message error in case user command rejected
-------------------------------------------------------------*/
void ALManager :: OnUserCommandCnf (const DAIUserCmdParam &outParam)
{
	if (outParam.m_eResponse != RESPONSE_OK)
	{	m_mtxChannels.P ();
		for (int j = outParam.m_nChLen; --j >= 0;)
		{	for (int i = m_channels.GetLength (); --i >= 0;)
				if (m_channels [i] ->GetChannelID () == outParam.m_pChid [j])
				{	m_channels [i] -> OnUserCommandReject (outParam.m_eResponse);
					break;
				}
		}
		m_mtxChannels.V ();
	}
}
