/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Buffer management
FILE:       Clock.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of ClockReference.
-------------------------------------------------------------------------*/
#include "defs.h"
#include "Clock.h"
#include "ztl.h"

#ifndef WIN32

#include <sys/time.h>
#include <unistd.h>
DWORD timeGetTime ()
{
	struct timeval time;
	::gettimeofday(&time);
	return (DWORD) (time.tv_sec * 1000 + time.tv_usec / 1000);
}

void Sleep (DWORD dw)
{
	usleep ((long long) dw * 1000);
}

#endif

/*-----------------------------------------------------------------
Static members intialization
-----------------------------------------------------------------*/
int ClockReference :: m_nGranularity = 5;

/*-----------------------------------------------------------------
Reset the clock to zero
-----------------------------------------------------------------*/
void ClockReference :: Reset ()
{
	m_dwRealStartTime = m_dwStartTime = timeGetTime ();
	m_bReset = true;
	m_dwInitOCR = 0;
	for (int i = 0; i < NUM_TAPS; i++)
	{
		m_dDriftEstimate [i] = 1;
	}
}

/*-----------------------------------------------------------------
Return current value of the object
-----------------------------------------------------------------*/
DWORD ClockReference :: GetTime ()
{
	DWORD dwTime = m_dFactor * (timeGetTime () - m_dwStartTime) + m_dwInitOCR;
	return dwTime - min (int (dwTime), m_nAdjust);
}

/*-----------------------------------------------------------------
Wait till the object reaches the given time
-----------------------------------------------------------------*/
void ClockReference :: WaitOn (ZSemaphore &rsmpControl, DWORD dwTime)
{
	DWORD dw = GetTime ();
	if (dwTime < dw + m_nGranularity)
		return;
	DWORD dwWaitTime = (dwTime - dw - m_nGranularity) / m_dFactor;
	rsmpControl.P (dwWaitTime);
}

/*-----------------------------------------------------------------
Compare two time stamps
-----------------------------------------------------------------*/
int ClockReference :: CompareTimes (DWORD dwTime, DWORD dwCompareTo)
{
	int nResult;

	/*-------------------------------------------------------------
	Timestamp of zero means no OCR was available at the time the AU
	was dispatched, therefore it should be immediately available.
	-------------------------------------------------------------*/
	if (dwTime == 0)
		return 0;

	if (dwCompareTo == 0)
		dwCompareTo = GetTime ();
	nResult = (long) dwTime - (long) dwCompareTo;
	if (abs (nResult) <= m_nGranularity)
		nResult = 0;
	return nResult;
}

/*-----------------------------------------------------------------
Set the clock to a given time.
This function is called to initialize the clock, or to update it when 
a new OCR is received. In the first case, we initialize the clock by
storing the current time (minus the initialization value). In the later
case, we calculate the speed of the clock relatively to our system clock
and store a value that will enable retrieving the controlled speed of
the clock.
-----------------------------------------------------------------*/
void ClockReference :: SetTime (DWORD dwOCR)
{
	m_dwLastOCR = dwOCR;
	if (m_bReset)
	{	m_dwRealStartTime = m_dwStartTime = timeGetTime ();
		m_dwInitOCR = dwOCR;
		m_bReset = false;
	}
	else
	{	
		/*-------------------------------------------------------------
		OCR adjustment made by Guido Franceschini (CSELT)

		We estimate the drift between the timebase conveyed by the OCR
		and the STB. We assume that the drift cannot exceed MAX_DRIFT in
		either direction and constrain estimations inside those limits.
		We compute the adjustment factor and use it in subsequent
		clock references only if there is a time slip of more than
		THREESOLD ms between the STB and the time conveyed by the OCR
		We also recalculate the start time in order to avoid discontinuities
		in the STB. However, when estimating the drift, the real start
		time is considered.
		-------------------------------------------------------------*/
		DWORD dwOCRTime = dwOCR - m_dwInitOCR;
		DWORD dwCurrentTime = timeGetTime ();
		DWORD dwElapsedTime = dwCurrentTime - m_dwRealStartTime;
		DWORD dwClockTime = m_dFactor * dwElapsedTime;
		double dSumDrift = 0;
		for (int i=0; i<NUM_TAPS-1; i++)
		{
			m_dDriftEstimate [i] = m_dDriftEstimate [i+1];
			dSumDrift += m_dDriftEstimate [i];
		}
		m_dDriftEstimate [NUM_TAPS-1] = (((double)dwOCRTime/(double)dwElapsedTime) + dSumDrift)/NUM_TAPS;
		if (m_dDriftEstimate [NUM_TAPS-1] > (1 + MAX_DRIFT))
			m_dDriftEstimate [NUM_TAPS-1] = (1 + MAX_DRIFT);
		if (m_dDriftEstimate [NUM_TAPS-1] < (1 - MAX_DRIFT))
			m_dDriftEstimate [NUM_TAPS-1] = (1 - MAX_DRIFT);
		if (abs ((int) (dwClockTime - dwOCRTime)) >= THREESOLD) {
			double dOldFactor = m_dFactor;
			DWORD dwOldStartTime = m_dwStartTime;
			m_dFactor = m_dDriftEstimate [NUM_TAPS-1];
			m_dwStartTime = dwCurrentTime - ((dOldFactor / m_dFactor) * (dwCurrentTime - dwOldStartTime));
		}
	}
}

/*-----------------------------------------------------------------
Adjustment value is the time elapsed since the clock was set to 
a certain value till a unit whose timestamp equals that value is rendered.
Normally this equals the time it takes to decode the unit.
-----------------------------------------------------------------*/
void ClockReference :: Adjust (DWORD dwTime)
{
	m_nAdjust =
		m_dFactor * (timeGetTime () - m_dwStartTime) -
		(dwTime - m_dwInitOCR);
}

/*-------------------------------------------------------------
Given residual value of a timestamp (dwTS), its size in bits (nLength)
and resolution (units per second), compare the value to the last OCR used
to set the clock and return the offset which should be added to the timestamp
in order to reconstruct the real timestamp value
-------------------------------------------------------------*/
DWORD ClockReference :: AdjustTimeOffset (
	DWORD dwTS,
	int nLength,
	float fResolutionFactor
)
{
	DWORD dwBase = 1 << nLength;
	DWORD dwOCR = (double) m_dwLastOCR / fResolutionFactor;
	DWORD dwOffset = dwOCR & ~(dwBase - 1);
	DWORD dwResult = dwOffset + dwTS;
	DWORD dwOption2 = dwResult - dwBase;
	DWORD dwOption3 = dwResult + dwBase;

	if (abs ((int) (dwOption2 - dwOCR)) < abs ((int) (dwResult - dwOCR)))
		dwResult = dwOption2;
	if (abs ((int) (dwOption3 - dwOCR)) < abs ((int) (dwResult - dwOCR)))
		dwResult = dwOption3;
	return dwResult & ~(dwBase - 1);
}
