/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1997.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Decoder/ Compositor
FILE:       NodeFld.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of NodeField and derivations.
-------------------------------------------------------------------------*/
#include "defs.h"
#include "Nodes.h"
#include "NodeFld.h"
#include "MediaObj.h"
#include "BifsDec.h"
#include "Route.h"
#include "Root.h"
#include "Quantize.h"
#include "NodeCode.h"
#include "Proto.h"
#include "Inline.h"

/*-----------------------------------------------------------------
NodeField destructor. Delete all routes
-----------------------------------------------------------------*/
NodeField :: ~NodeField ()
{
	while (m_pFirstRoute)
		delete m_pFirstRoute;  // this unlinks the route automatically
}

/*-----------------------------------------------------------------
Add a Route to the nodeField
-----------------------------------------------------------------*/
void NodeField :: LinkRoute (Route *pRoute)
{
	pRoute -> Link (m_pFirstRoute);
}

/*-----------------------------------------------------------------
Unlink a Route from the list of routes
-----------------------------------------------------------------*/
void NodeField :: UnlinkRoute (Route *pRoute)
{
	pRoute -> Unlink (m_pFirstRoute);
}

/*-----------------------------------------------------------------
Activate attached routes on nodeField update
The paramter bFirst is used to eliminate loops.
-----------------------------------------------------------------*/
void NodeField :: OnFieldUpdate (bool bIsFirst)
{
	for (Route *pRoute = m_pFirstRoute;
		pRoute != NULL;
		pRoute = pRoute -> GetNext ()
	)
		pRoute -> Activate (bIsFirst);
#ifdef BIFSCAR
	m_bNeedsEncoding = true;
#endif
}

/*-----------------------------------------------------------------
Parse an SFInt
-----------------------------------------------------------------*/
BOOL SF<int> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject*, 
	Quantizer *pQuant
)
{
	if(!pQuant || !pQuant -> IsActive ()) 
		m_value = pCoder -> ParseInt (32, "intValue");
	else 
		m_value = pQuant -> ParseInt (pCoder);

	TRACEF ("%d", m_value);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFBool
-----------------------------------------------------------------*/
BOOL SF<Bool> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	m_value = pCoder -> ParseInt (1, "boolValue") != 0;
	TRACEF ("%s", m_value ? "TRUE" : "FALSE");
	return TRUE;
}

/*-----------------------------------------------------------------
Parse and SFFloat
MBS: modified to take into account quantization parameters if any
-----------------------------------------------------------------*/
BOOL SF<float> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * pQuant
)
{
	if(!pQuant || !pQuant -> IsActive ()) 
		m_value = pCoder -> ParseFloat ("floatValue");
	else 
		m_value = pQuant->ParseFloat(pCoder);

	TRACEF ("%.2f", m_value);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFTime
-----------------------------------------------------------------*/
BOOL SF<Time> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	m_value = pCoder -> ParseDouble("doubleValue");
	TRACEF ("%.2lf", m_value);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFVec3f
MBS: modified to take into account quantization parameters if any
-----------------------------------------------------------------*/
BOOL SF<Vec3f> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * pQuant
)
{
	if(!pQuant || !pQuant->IsActive())
	{	m_value.x = pCoder -> ParseFloat("x");
		m_value.y = pCoder -> ParseFloat("y");
		m_value.z = pCoder -> ParseFloat("z");
	}
	else
		m_value = pQuant -> ParseVec3f(pCoder);

	TRACEF3 ("%.2f %.2f %.2f", m_value.x, m_value.y, m_value.z);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFVec2f
MBS: modified to take into account quantization parameters if any
-----------------------------------------------------------------*/
BOOL SF<Vec2f> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * pQuant
)
{
	if(!pQuant || !pQuant->IsActive())
	{	m_value.x = pCoder -> ParseFloat("x");
		m_value.y = pCoder -> ParseFloat("y");
	}
	else
		m_value = pQuant -> ParseVec2f(pCoder);

	TRACEF2 ("%.2f %.2f", m_value.x, m_value.y);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFRotation
MBS: modified to take into account quantization parameters if any
-----------------------------------------------------------------*/
BOOL SF<Rotation> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * pQuant
)
{
	if(!pQuant || !pQuant->IsActive())
	{	m_value.x = pCoder -> ParseFloat("x");
		m_value.y = pCoder -> ParseFloat("y");
		m_value.z = pCoder -> ParseFloat("z");
		m_value.theta = pCoder -> ParseFloat("theta");
	}
	else
		m_value = pQuant -> ParseRotation(pCoder);

	TRACEF4 ("%.2f %.2f %.2f %.2f", m_value.x, m_value.y, m_value.z, m_value.theta);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFRotation2D
MBS: modified to take into account quantization parameters if any
-----------------------------------------------------------------*/
BOOL SF<Rotation2D> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * pQuant
)
{
  if(!pQuant || !pQuant->IsActive()) {
    m_value.x = pCoder -> ParseFloat("x");
    m_value.y = pCoder -> ParseFloat("y");
    m_value.theta = pCoder -> ParseFloat("theta");
  }
  else
    m_value = pQuant -> ParseRotation2D(pCoder);

  TRACEF3 ("%2.f %.2f %.2f", m_value.x, m_value.y, m_value.theta);
  return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFColor
MBS: modified to take into account quantization parameters if any
-----------------------------------------------------------------*/
BOOL SF<Color3f> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * pQuant
)
{
	if(!pQuant || !pQuant->IsActive())
	{	m_value.r = pCoder -> ParseFloat("r");
		m_value.g = pCoder -> ParseFloat("g");
		m_value.b = pCoder -> ParseFloat("b");
	}
	else
		m_value = pQuant -> ParseColor3f(pCoder);

	TRACEF3 ("%.2f %.2f %.2f", m_value.r, m_value.g, m_value.b);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFObjectID
-----------------------------------------------------------------*/
BOOL SF<Url> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	
	if (pCoder -> ParseInt (1, "isOD"))
	{	m_value.ODid = pCoder -> ParseInt (10, "ODid");
		TRACEF ("%d", m_value.ODid);
	}
	else
	{	m_value.ODid = -1;
		m_value.urlValue.Parse (pCoder, NULL);
	}
	return TRUE;
}

/*-----------------------------------------------------------------
Parse SFString
MBS 110698: updated for UTF-8 String
Zvi 021799: redo MBS change. The string length is given in bytes, not
in characters. The handling of UTF-8 coding shall be done by the compositor -
it is not the matter of the parser.
-----------------------------------------------------------------*/
BOOL SF<ZString> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	pCoder ->ParseSFString (m_value);
	TRACEF ("%s", (LPCSTR) m_value);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse SFImage
MBS 110498: this is a new type
-----------------------------------------------------------------*/
BOOL SF<Image> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	m_value.width = pCoder -> ParseInt (12, "width");
	m_value.height = pCoder -> ParseInt (12, "height");
	m_value.numComponents = pCoder -> ParseInt (2, "numComponents");

	//Alvar 981223: Added pixel array allocation and refcount stuff
	int nPixels = m_value.numComponents * m_value.width * m_value.height;
	BYTE *p = m_value.pixels = new BYTE [nPixels];
	m_value.m_pRefCnt = new int;
	*(m_value.m_pRefCnt) = 1;  
	for (int i = 0; i < nPixels; i++ )
		*p++ = pCoder -> ParseInt (8, "pixel");
	// end of Alvar changes

	TRACEF ("Image %d ", m_value.width);
	TRACEF ("Image %d ", m_value.height);
	TRACEF ("Image %d ", m_value.numComponents);
	return TRUE;
}

/*-----------------------------------------------------------------
A dummy function
-----------------------------------------------------------------*/
BOOL SF<MediaObject *> :: Parse (
	BIFSDecoder *,
	MediaObject *, 
	Quantizer *
)
{
	return FALSE;
}

/*-----------------------------------------------------------------
SFGenericNode Destructor.  Deletes the MediaObject, unless it is defined
by other node and only USEd here.
-----------------------------------------------------------------*/
SFGenericNode :: ~SFGenericNode ()
{
	if (m_pNode)
		m_pNode -> RemoveContainer (this);
}

/*-----------------------------------------------------------------
Assign a value into a generic-node. If the old value is not a copy 
of another nodeField, delete the old node
-----------------------------------------------------------------*/
MediaObject *SFGenericNode :: SetObject (MediaObject *node, MediaObject *pParent, bool bIsFirst)
{
	if (m_pNode)
		m_pNode -> RemoveContainer (this);
	m_pNode = node;
	if (node)
		node -> SetContainer (this, false, pParent);
	OnFieldUpdate (bIsFirst);
	#ifdef BIFSCAR
	if (node == NULL_OBJECT)
		SetNeedsEncoding (false);
	#endif
	return node;
}

/*-----------------------------------------------------------------
Parse a node field which is a sub-node
-----------------------------------------------------------------*/
BOOL SFGenericNode :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject *pNode, 
	Quantizer * /*pQuant*/
)
{
	m_pParent = pNode;
	return ParseNodeField (pCoder, pNode, this, FALSE);
}

/*-------------------------------------------------------------
This function does the actual work of node parsing when the node
is either a scalar field in a parent node (in which case the function
is called from SFGenericNode :: Parse), an element in a vector field,
or when it is parsed in ReplaceNode command. The later two cases ParseNode
is called directly, not through Parse().
-------------------------------------------------------------*/
BOOL SFGenericNode :: ParseNodeField (
	BIFSDecoder *pCoder, 
	MediaObject *pNode,			// the node this field belongs to
	NodeField *pContainer,		// a pointer to the container field
								// (this is not always 'this' because
								// a temporary NodeField object can be
								// used to parse the node, and then the
								// node is assigned or inserted in another
								// NodeField object.
	BOOL bISMF					// TRUE when the field is an MF field
)
{
	if (pCoder -> ParseInt (1, "isReused"))	// if isReused
	{	int nID = pCoder -> ParseInt (pCoder -> GetNodeIDbits (), "NodeID");
		if (pNode &&
			(m_pNode = pNode -> GetRoot () -> FindNode (nID)) != NULL
		)
			m_pNode -> SetContainer (pContainer, bISMF, pNode);
		else
			return FALSE;
		TRACEF ("USE %s", m_pNode ->GetNameID ());
	}
	else
	{	
		/*-----------------------------------------------------------------
		Get object type, instantiate the object, and let it parse itself
		-----------------------------------------------------------------*/
		int nID = -1;
		int nNodeType =
			pCoder -> ParseInt (
				BitFieldWidth (GetNodeTableSize () - GetV2ExtSize () + 1),
				"nodeType"
			);
		int nProtoID = -1;

		/*-------------------------------------------------------------
		If nodeType is zero this is an escape to V2
		-------------------------------------------------------------*/
		if (nNodeType == 0)
		{	int nExt = pCoder -> ParseInt (BitFieldWidth (GetV2ExtSize () + 2), "nodeType");
			if (nExt == 0)
				return false;		// another escape which we don't support in V2
			else if (nExt == 1)
			{	nNodeType = -1;		// negative value signals a PROTO
				nProtoID = pCoder -> ParseInt (pCoder -> GetProtoIDbits (), "PROTOnodeType");
			}
			else
				nNodeType = GetNodeTableSize () - GetV2ExtSize () + nExt - 1;
		}

		/*-------------------------------------------------------------
		m_pNode may already point to another node if the function is called
		during a ReplaceField command.
		-------------------------------------------------------------*/
		if (m_pNode)
			m_pNode -> RemoveContainer (pContainer, pNode);

		/*-----------------------------------------------------------------
		The root of the new object is the same as of the container 
		object. If the container object is itslef a root, the root is
		the container object.
		-----------------------------------------------------------------*/
		Root *pRoot = pNode -> GetRoot () ? pNode -> GetRoot () : (Root *) pNode;

		/*-------------------------------------------------------------
		Create a new node. If this is a proto, instantiate the proto
		-------------------------------------------------------------*/
		if (nNodeType >= 0)
			m_pNode = CreateNode (nNodeType);
		else
		{	/*-------------------------------------------------------------
			Locate the Proto, and check whether its first node is of appropriate type.
			-------------------------------------------------------------*/
			Proto *pProto = pRoot -> FindProto (nProtoID);
			if (pProto == NULL)
				return NULL;
			m_pNode = pProto -> Instantiate (pRoot);
			MediaObject *pFirst = ((Proto *)m_pNode) ->GetFirstNode ();
			if (!pFirst || !IsTypeOK (pFirst))
			{	m_pNode -> Delete ();
				m_pNode = NULL;
			}
		}
		if (!m_pNode)
			return FALSE;

		/*-------------------------------------------------------------
		Parse the DEF id, if any
		-------------------------------------------------------------*/
		if (pCoder -> ParseInt (1, "isUpdatable")) // if isUpdatable
		{	
			nID = pCoder -> ParseInt (pCoder -> GetNodeIDbits (), "NodeID");
			if (pRoot -> GetInline () ->IsCodeNames ())
			{	m_pNode ->ParseNameID (pCoder);
				TRACEF ("DEF %s ", m_pNode ->GetNameID ());
			}
			else
				TRACEF ("DEF %d ", nID);
		}

		//ML - moved SetExecutive to after SetRoot
		//m_pNode -> SetExecutive (pCoder -> GetExecutive ());
		m_pNode -> SetID (nID);
		m_pNode -> SetContainer (pContainer, bISMF, pNode);
		m_pNode -> SetRoot (pRoot);
		m_pNode -> SetExecutive (pCoder -> GetExecutive ());

		/*-------------------------------------------------------------
		Parse the node
		-------------------------------------------------------------*/

		// MBS: sets current QP in decoder if we have a QP node
		if(m_pNode->GetNodeCode() == QuantizationParameter :: GetCode()) 
		  pCoder->SetCurrentQP(m_pNode);
		// end MBS

		// BenA (7/30/99): moved this here from above if (m_pNode->...)
		m_pNode -> Parse (pCoder);
		// end BenA
	}
	return TRUE;
}

/*-------------------------------------------------------------
Clone a node field when a Proto is instantiated
-------------------------------------------------------------*/
NodeField *SFGenericNode :: Clone (NodeField *pField, MediaObject *pParent)
{
	if (pField == NULL)
		pField = new SFGenericNode;
	if (m_pNode)
	{	((SFGenericNode*) pField) ->m_pNode =  m_pNode -> Clone (pParent -> GetRoot ());
		((SFGenericNode*) pField) -> m_pNode -> SetContainer (this, false, pParent);
	}
	return pField;
}

/*-----------------------------------------------------------------
Parse an MF nodeField
-----------------------------------------------------------------*/
BOOL MField :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject *pNode, 
	Quantizer *pQuant
)
{
	TRACEF ("[", 0);
	bool bFirst = true;

	Clear ();
	/*-----------------------------------------------------------------
	MF nodeField may be coded as ListMFDescription or VectorMFDescription
	-----------------------------------------------------------------*/
	// MBS: variables for quantization
	BOOL localQP=FALSE;
	MediaObject *qpOld = NULL;
	// end MBS

	// BenA (7/30/99): save qp for this level
	MediaObject *qpForThisLevel = pCoder->GetCurrentQP();
	// end BenA

	if (!pCoder -> ParseInt (1, "reserved"))		// reserved, added in FDIS
	{	if (pCoder -> ParseInt (1, "isListDescription"))	// isListDescription
			while (pCoder -> ParseInt (1, "end") == 0)	// changed for  FDIS
			{
				if (bFirst)
				{	bFirst = false;
					TRACENEWLINE;
					TRACEINDENT (1);
				}
				// MBS: saves old QP
				if(!localQP) qpOld=pCoder->GetCurrentQP();
				// end MBS

				if (!AppendElement (pCoder, pNode, pQuant))
					return FALSE;
		
				// MBS: restores old QP

				if(pCoder->hasLocalQP()) {
					if(localQP) {
						localQP=FALSE;
						pCoder->SetCurrentQP(qpOld);
					}
					else {
						localQP=TRUE;
					}
				}
				// end MBS

				TRACENEWLINE;
			}
		else
			for (int i = pCoder -> ParseInt (pCoder -> ParseInt (5, "nBits"), "size"); --i >= 0;)
			{
				if (bFirst)
				{	bFirst = false;
					TRACENEWLINE;
					TRACEINDENT (1);
				}

				// MBS: saves old QP
				if(!localQP) qpOld=pCoder->GetCurrentQP();
				// end MBS

				if (!AppendElement (pCoder, pNode, pQuant))
					return FALSE;

				// MBS: restores old QP

				if(pCoder->hasLocalQP()) {
				  if(localQP) {
					localQP=FALSE;
					pCoder->SetCurrentQP(qpOld);
				  }
				  else {
					localQP=TRUE;
				  }
				}
				// end MBS

				TRACENEWLINE;
			}

		// BenA (7/30/99): restore quantization for this level
		pCoder->SetCurrentQP(qpForThisLevel);
		// end BenA
	}

#ifdef TRACE_ENCODING
	if (!bFirst)
		TRACEINDENT (-1);
	TRACEF ("]", 0);
	TRACENEWLINE;
#endif
		
	return TRUE;
}

/*-----------------------------------------------------------------
Deletes all nodes
-----------------------------------------------------------------*/
void MFGenericNode :: RemoveAll ()
{
	for (int i = GetLength (); --i >= 0;)
		if (m_array [i])
			m_array [i] -> RemoveContainer (this, m_pParent);
}

/*-----------------------------------------------------------------
Manipulation of MFGenericNode arrays
-----------------------------------------------------------------*/
void MFGenericNode :: InsertAt (int index, MediaObject *value, MediaObject *pParent, bool bRoute)
{
	m_array.InsertAt (index, value);
	value ->SetContainer (this, true, pParent);
	if (bRoute)
		OnFieldUpdate ();
}

void MFGenericNode :: Append (MediaObject *value, MediaObject *pParent, bool bRoute)
{
	if (value -> GetNodeCode () != CODE_Script)	// SCRIPT
	{	m_array += value;
		value ->SetContainer (this, true, pParent);
	}
	if (bRoute)
		OnFieldUpdate ();
}

void MFGenericNode :: SetAt (int index, MediaObject *value, MediaObject *pParent, bool bRoute)
{
	if (index < GetLength () && m_array [index] != NULL)
		m_array [index] -> RemoveContainer (this, m_pParent);
	m_array.SetAt (index, value);
	if (value != NULL)
		value ->SetContainer (this, true, pParent);
	if (bRoute)
		OnFieldUpdate ();
}

void MFGenericNode :: DeleteFrom (int index, BOOL bRoute)
{
	if (index < GetLength () && m_array [index] != NULL)
		m_array [index] -> RemoveContainer (this, m_pParent);
	m_array.DeleteFrom (index);
	if (bRoute)
		OnFieldUpdate ();
	#ifdef BIFSCAR
	if (m_array.GetLength () == 0)
		SetNeedsEncoding (false);
	#endif
}

/*-----------------------------------------------------------------
Assigment operator activated in case of Routes. We assume that
the nodeField argument is always of the same type as this, therefor
allow for deliberate downcasting. This assumption is dangerous.
Commercial application should implement safe casting or proper
conversion mechanism.
-----------------------------------------------------------------*/
void MFGenericNode :: Copy (NodeField &nodeField, MediaObject *pParent, bool bIsFirst)
{
	MFGenericNode &fld = (MFGenericNode &) nodeField;
	for (int i = GetLength (); --i >= 0;)
		DeleteFrom (i, false);		
	for (i = 0; i < fld.GetLength (); i++)
	{	m_array += fld.m_array [i];
		fld.m_array [i] -> SetContainer (this, true, pParent);
	}
	OnFieldUpdate (bIsFirst);
}

/*-------------------------------------------------------------
The node in this field has been replaced by ReplaceNode command
-------------------------------------------------------------*/
void SFGenericNode :: ReplaceNode (MediaObject *pNew)
{
	m_pNode = pNew;
	pNew -> SetContainer (this, FALSE, m_pParent);
}

/*-------------------------------------------------------------
Node pOld in this MFxxxNode field has been replaced by a 
ReplaceNode command.
-------------------------------------------------------------*/
void MFGenericNode :: ReplaceNode (MediaObject *pOld, MediaObject *pNew)
{
	for (int i = m_array.GetLength (); --i >= 0; i)
		if (m_array [i] == pOld)
		{	m_array [i] = pNew;
			pNew -> SetContainer (this, TRUE, m_pParent);
			return;
		}
}

/*-------------------------------------------------------------
Clone a vector node field when a Proto is instantiated
-------------------------------------------------------------*/
NodeField *MFGenericNode :: Clone (NodeField *pField, MediaObject *pParent)
{
	if (pField == NULL)
		pField = new MFGenericNode;
	for (int i = 0; i < m_array.GetLength (); i++)
	{	if (m_array [i])
		{	MediaObject *pNode = m_array [i] -> Clone (pParent -> GetRoot ());
			((MFGenericNode *) pField) -> m_array += pNode;
			pNode -> SetContainer (this, true, pParent);
		}
	}
	return pField;
}