/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1997.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Buffer management
FILE:       Stream.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of MediaStreamImp.
			(This version simulation only. Just makes sure Allocate,
			Reallocate, and Dispatch are called properly.
-------------------------------------------------------------------------*/
#include "defs.h"
#include <assert.h>
#include "Stream.h"
#include "Im1defs.h"

int MediaStreamImp :: m_nStreamCount;
FILE * MediaStreamImp :: m_fTraceFile;
ZMutex MediaStreamImp :: m_mtxLockTraceFile;

/*-----------------------------------------------------------------
MediaStreamImp constructor. The simulation allocate a buffer large enough 
for all possible AUs
-----------------------------------------------------------------*/
MediaStreamImp :: MediaStreamImp (ClockReference *pClock, int nBufferSize) :
	m_smpWaitForMemory (1, 0),
	m_smpWaitForData (1, 0)
{
	/*-----------------------------------------------------------------
	Allocate the stream buffer. If nBufferSize < 0, this is dynamic
	buffer, i.e., memory when Allocate or Reallocate are called.
	-----------------------------------------------------------------*/
	m_nTakenStart = m_nBufferSize = nBufferSize;
	m_nNextFetch = -1;
	m_nUsedBy = m_nFetched = 0;
	m_pBuffer = NULL;
	m_bDynamic = nBufferSize < 0;
	if (!m_bDynamic)
		m_pBuffer = (LPBYTE) malloc (nBufferSize + 3);
	m_nSN = m_nStreamCount++;		// For tracing
	m_pClock = pClock;
	Reset ();
}

/*-----------------------------------------------------------------
Reset the stream
-----------------------------------------------------------------*/
void MediaStreamImp :: Reset ()
{
	/*-------------------------------------------------------------
	Reset member variables
	-------------------------------------------------------------*/
	m_nTakenStart = m_nBufferSize;
	m_nNextFetch = -1;
	m_nFreeStart = m_nFetched = 0;
	m_nAllocatedLength = -1;
	m_bWaitingForMemory = m_bWaitingForData = false;
	m_bReallocInProgress = m_bEndOfStream = false;
	m_bFirstRetrieved = m_bWaiting = false;
	m_bTrace = FALSE;
	m_bOpenToReceive = m_bOpenToSend = false;
}

/*-----------------------------------------------------------------
Open for receiving
-----------------------------------------------------------------*/
void MediaStreamImp :: Start ()
{
	m_smpReceiving.P ();
	m_smpInUse.P ();
	m_smpTimer.P ();
	if (m_nUsedBy++ == 0)	// stream is in used if after opened by both sides
	{	Reset ();			// the first who opens resets
		m_smpInUse.V ();
	}
	m_bOpenToReceive = true;
}		

/*-----------------------------------------------------------------
Open for sending
-----------------------------------------------------------------*/
void MediaStreamImp :: Open ()
{
	m_smpSending.P ();
	m_smpInUse.P ();
	if (m_nUsedBy++ == 0)	// stream is in used if after opened by both sides
	{	Reset ();			// the first who opens resets
		m_smpInUse.V ();
	}
	m_bOpenToSend = true;
}		

/*-----------------------------------------------------------------
MediaStreamImp destructor
-----------------------------------------------------------------*/
MediaStreamImp :: ~MediaStreamImp ()
{
	/*-----------------------------------------------------------------
	Wait till both sides close the stream
	-----------------------------------------------------------------*/
	m_smpReceiving.P ();
	m_smpSending.P ();
	m_mtxLockStream.P ();
	free (m_pBuffer);
	m_mtxLockStream.V ();
	m_smpSending.V ();
	m_smpReceiving.V ();

	/*-----------------------------------------------------------------
	Following lines for tracing
	-----------------------------------------------------------------*/
	if (--m_nStreamCount == 0 && m_fTraceFile != NULL)
	{	fclose (m_fTraceFile);
		m_fTraceFile = NULL;
	}
}

/*-----------------------------------------------------------------
Allocate() allocates a block of nSize byte and returns its address
-----------------------------------------------------------------*/
LPVOID MediaStreamImp :: Allocate (int nSize, BOOL bWait, BOOL bShove)
{
	if (m_bEndOfStream)
		return END_OF_STREAM;
	assert (m_bOpenToSend);
	nSize += sizeof (int) + sizeof (DWORD);	// for block length and time stamp

	/*-----------------------------------------------------------------
	Return NULL if last allocated block has not been dispatched
	-------------------------------------------------------------*/
	if (m_nAllocatedLength >= 0 && !m_bReallocInProgress)
		return NULL;

	/*-------------------------------------------------------------
	If this is an empty dynamic buffer reallocate buffer if necessary
	-------------------------------------------------------------*/	
	if (nSize > m_nBufferSize)
		if (m_bDynamic &&
			m_nFreeStart == 0 && m_nTakenStart == m_nBufferSize
		)
		{	if (!m_pBuffer)
				m_pBuffer = (LPBYTE) malloc (nSize);
			else
				m_pBuffer = (LPBYTE) realloc (m_pBuffer, nSize);
			m_nTakenStart = m_nBufferSize = nSize;
		}
		else
			return NULL;

	/*-----------------------------------------------------------------
	Check available (contiguous) space.
	-----------------------------------------------------------------*/
	m_mtxLockStream.P ();
	if (m_bEndOfStream)
	{	m_mtxLockStream.V ();
		return END_OF_STREAM;
	}
	while (
		nSize > m_nTakenStart - m_nFreeStart &&
		(	m_nTakenStart > m_nFreeStart ||
			m_nTakenStart == m_nFreeStart && !m_bReallocInProgress ||
			nSize > m_nBufferSize - m_nFreeStart &&
			nSize > m_nTakenStart
		)
	)
	{
		/*-----------------------------------------------------------------
		Not enough space available. One reason might be that the buffer
		is empty, and this is a reallocation of a block which is
		currently at the middle of the buffer
		-----------------------------------------------------------------*/
		if (m_nTakenStart == m_nFreeStart && m_bReallocInProgress)
		{	m_nFreeStart = 0;
			m_nTakenStart = m_nBufferSize;
			m_nNextFetch = -1;
		}
		/*-----------------------------------------------------------------
		If bShove it TRUE, shove last block
		(or wait for a currently fetched block to be discarded).
		-----------------------------------------------------------------*/
		else if (bShove)
		{	if (m_nFetched > 0)
			{	m_mtxLockStream.V ();
				m_smpLockOutBlock.P ();
				m_smpLockOutBlock.V ();
				m_mtxLockStream.P ();
			}
			else
				Discard ();
		}
		/*-----------------------------------------------------------------
		If the wait flag is true, wait till a space is available
		-----------------------------------------------------------------*/
		else if (bWait)
		{	m_bWaitingForMemory = true;
			m_mtxLockStream.V ();
			m_smpWaitForMemory.P ();
			m_mtxLockStream.P ();
			if (m_bEndOfStream)
			{	m_mtxLockStream.V ();
				return END_OF_STREAM;
			}
		}
		else
		{	m_mtxLockStream.V ();
			return NULL;
		}
	}

	/*-----------------------------------------------------------------
	Allocate a block of the desired length. Starting either at the next
	available byte, or cycle around the buffer
	-----------------------------------------------------------------*/
	if (m_nBufferSize - m_nFreeStart < nSize)
	{	STOREINT (-1, m_pBuffer + m_nFreeStart);	// Mark as not used
		if (m_nTakenStart == m_nFreeStart)
			m_nTakenStart = m_nBufferSize;
		if (m_nNextFetch == m_nFreeStart)
			m_nNextFetch = -1;
		m_nFreeStart = 0;
	}
	m_nAllocatedLength = nSize - (sizeof (int) + sizeof (DWORD));
	m_bWait = bWait;			// save for realloc
	m_bShove = bShove;
	m_mtxLockStream.V ();
	return m_pBuffer + m_nFreeStart + sizeof (int) + sizeof (DWORD);
}

/*-----------------------------------------------------------------
Reallocate() expands the reserved block, if necessary
-----------------------------------------------------------------*/
LPVOID MediaStreamImp :: Reallocate (int nSize)
{
	if (m_bEndOfStream)
		return END_OF_STREAM;
	assert (nSize >= 0);
	assert (m_bOpenToSend);

	/*-----------------------------------------------------------------
	Reallocation is done by zeroing m_nAllocatedLength and calling
	Allocate with the new size.
	-----------------------------------------------------------------*/

	m_mtxLockStream.P ();

	int nPreviousStart = m_nFreeStart;
	int nPreviousLength = m_nAllocatedLength;
	LPVOID pResult;
	
	m_bReallocInProgress = TRUE;
	m_mtxLockStream.V ();
	pResult = Allocate (nSize, m_bWait, m_bShove);
	m_bReallocInProgress = FALSE;

	/*-----------------------------------------------------------------
	If the stream was closed by the receiving side, tell the receiver
	not to wait till block is discarded, and return END_OF_STREAM
	-----------------------------------------------------------------*/
	if (pResult == END_OF_STREAM)
	{	m_nAllocatedLength = -1;
		if (m_bWaitingForData)
		{	m_bWaitingForData = false;
			m_smpWaitForData.V ();
		}
		return END_OF_STREAM;
	}

	/*-----------------------------------------------------------------
	If realloc successful, but block was shifted in memory, copy the
	contents of the old block.
	-----------------------------------------------------------------*/
	if (pResult != NULL && m_nFreeStart != nPreviousStart)
		memcpy (
			pResult,
			m_pBuffer + nPreviousStart + sizeof (int) + sizeof (DWORD),
			nPreviousLength
		);
	return pResult;
}

/*-----------------------------------------------------------------
Dispatch() dispatches the allocated block into the stream
-----------------------------------------------------------------*/
BOOL MediaStreamImp :: Dispatch (int nSize, DWORD dwTime, BOOL /*bEOU*/)
{
	if (m_bEndOfStream)
		return false;
	assert (nSize >= 0);
	assert (m_bOpenToSend);

	if (m_nAllocatedLength < 0)
		return FALSE;

	/*-----------------------------------------------------------------
	store the size and time information. Signal to waiting threads.
	-----------------------------------------------------------------*/
//	if (!m_bClockSet)
//		SetClock (dwTime);
	m_mtxLockStream.P ();

	nSize += sizeof (int) + sizeof (DWORD);
	STOREINT (nSize, m_pBuffer + m_nFreeStart);
	STOREDWORD (dwTime, m_pBuffer + m_nFreeStart + sizeof (int));

	/*-------------------------------------------------------------
	If this is the only unit in a dynamic buffer, trim the buffer to the
	required size
	-------------------------------------------------------------*/
	if (m_bDynamic &&
		m_nFreeStart == 0 && m_nTakenStart == m_nBufferSize &&
		m_nBufferSize > nSize
	)
	{	m_pBuffer = (LPBYTE) realloc (m_pBuffer, nSize);
		m_nTakenStart = m_nBufferSize = nSize;
	}

	if (m_nNextFetch < 0)
		m_nNextFetch = m_nFreeStart;
	m_nFreeStart += nSize;
	if (m_nFreeStart == m_nBufferSize)
		m_nFreeStart = 0;
	if (m_nTakenStart == m_nBufferSize)
		m_nTakenStart = 0;
	m_nAllocatedLength = -1;
	if (m_bWaitingForData)
	{
    m_bWaitingForData = FALSE;
    TRACE("(2) [MSI::Dis] buf %08x msi %08x \n", 
		m_pBuffer, this);
		m_smpWaitForData.V ();
	}
	m_mtxLockStream.V ();
	return TRUE;
}

/*-----------------------------------------------------------------
FetchNext fetches the next available block, ignoring its time stamp (which
is only returned in rdwTime.  If bWait is true, the function waits till
a block becomes available.
If successful, the function locks the fetched block.
If end of stream reached, the function returns END_OF_STREAM.
-----------------------------------------------------------------*/
LPVOID MediaStreamImp :: FetchNext (
	int &rnSize,
	DWORD &rdwTime,
	BOOL &bEOU,
	BOOL bWait
)
{
	if (m_bEndOfStream)
		return END_OF_STREAM;
	assert (m_bOpenToReceive);
	bEOU = FALSE;		// Only in this version. Needs correction.
	/*-----------------------------------------------------------------
	Check if a loaded block is available
	-----------------------------------------------------------------*/
	m_mtxLockStream.P ();
	while (
		m_nNextFetch < 0 ||
		m_nNextFetch == m_nFreeStart && m_nAllocatedLength >= 0
	)
	{
		/*-----------------------------------------------------------------
		If end of stream, return END_OF_STREAM
		-----------------------------------------------------------------*/
		if (m_bEndOfStream)
		{	m_mtxLockStream.V ();
			return END_OF_STREAM;
		}
		/*-----------------------------------------------------------------
		If necessary, wait for a block, otherwise return NULL
		-----------------------------------------------------------------*/
		if (bWait)
		{	m_bWaitingForData = TRUE;
			m_mtxLockStream.V ();
			m_smpWaitForData.P ();
			if (m_bEndOfStream)
				return END_OF_STREAM;
			m_mtxLockStream.P ();
		}
		else
		{	m_mtxLockStream.V ();
			return NULL;
		}
	}
	
	/*-----------------------------------------------------------------
	Lock a block and return its address, size and time stamp
	-----------------------------------------------------------------*/
	if (m_nFetched++ == 0)
		m_smpLockOutBlock.P ();
	rnSize =
		GetBlockLength (m_nNextFetch) - sizeof (int) - sizeof (DWORD);
	rdwTime = GetTimeStamp (m_nNextFetch);
	LPVOID pBuffer =
		m_pBuffer + sizeof (int) + sizeof (DWORD) + m_nNextFetch;
	m_nNextFetch = GetNext (m_nNextFetch);
	m_mtxLockStream.V ();
	return pBuffer;
}

/*-----------------------------------------------------------------
Fetch() fetches a block whose time stamp matured.
It will return TRUE rbRecurrent if it's the same block returned on
previous call.
If bSkip is TRUE it will skip all blocks till the newest one whose time
matches.
If bWait is TRUE it will wait till one block matures.
-----------------------------------------------------------------*/
LPVOID MediaStreamImp :: Fetch (
	int &rnSize,
	DWORD &rdwTime,
	BOOL &bEOU,
	int &rnStatus,
	BOOL bSkip,
	BOOL bWait,
	long lDelta,
	BOOL bDontClose
)
{
	if (m_bEndOfStream)
		return END_OF_STREAM;
	assert (m_bOpenToReceive);
	/*-----------------------------------------------------------------
	Returns NULL if a block is already fetched
	-----------------------------------------------------------------*/
	if (m_nFetched > 0)
		return NULL;

	bEOU = FALSE;		// Only in this version. Needs correction.
	/*-----------------------------------------------------------------
	If first block has been already retrieved and bWait is true, discard
	that block.
	-----------------------------------------------------------------*/
	m_mtxLockStream.P ();
	if (m_bFirstRetrieved && bWait)
		Discard ();

	/*-----------------------------------------------------------------
	Check if a loaded block is available
	-----------------------------------------------------------------*/
	while (
		m_nTakenStart >= m_nBufferSize ||
		m_nTakenStart == m_nFreeStart && m_nAllocatedLength >= 0
	)
	{
		/*-----------------------------------------------------------------
		If end of stream, return END_OF_STREAM
		-----------------------------------------------------------------*/
		if (m_bEndOfStream)
		{	m_mtxLockStream.V ();
			return END_OF_STREAM;
		}
		/*-----------------------------------------------------------------
		If necessary, wait for a block, otherwise return NULL
		-----------------------------------------------------------------*/
		if (bWait)
		{	m_bWaitingForData = TRUE;
			m_mtxLockStream.V ();
			m_smpWaitForData.P ();
			m_mtxLockStream.P ();
		}
		else
		{	m_mtxLockStream.V ();
			return NULL;
		}
	}

	/*-----------------------------------------------------------------
	Find the next block which matured. If bSkip is TRUE, find the newest
	matured block.
	-----------------------------------------------------------------*/
	int nNext = m_nTakenStart;
	int nFollowingAU = nNext;
	DWORD dwTime;
	DWORD dwCurrentTime = (long) m_pClock -> GetTime () + lDelta;

	if (m_bFirstRetrieved)
		nFollowingAU = GetNext (nNext);
	while (
		nFollowingAU >= 0 &&
		m_pClock -> CompareTimes (
			dwTime = GetTimeStamp (nFollowingAU),
			dwCurrentTime
		) <= 0
	)
	{	nNext = nFollowingAU;
		if (!bSkip)
			break;
		nFollowingAU = GetNext (nNext);
	}

	/*-----------------------------------------------------------------
	Now nNext points to the most suitable block.
	If it is a matured block, lock it and return its address.
	Otherwise wait or return NULL
	-----------------------------------------------------------------*/
	dwTime = GetTimeStamp (nNext);
	if (m_pClock -> CompareTimes (dwTime, dwCurrentTime) > 0)
	{	
		m_mtxLockStream.V ();
		if (bWait)
		{	
			m_pClock -> WaitOn (m_smpTimer, dwTime);
			m_mtxLockStream.P ();
		}
		else
			return NULL;
	}

	/*-----------------------------------------------------------------
	Now we can lock the block and return all the required information
	-----------------------------------------------------------------*/
	++m_nFetched;
	m_smpLockOutBlock.P ();
	rnSize = GetBlockLength (nNext) - sizeof (int) - sizeof (DWORD);
	rdwTime = GetTimeStamp (nNext);
	rnStatus = nNext == m_nTakenStart && m_bFirstRetrieved;
	m_bFirstRetrieved = TRUE;
	m_nNextFetch = GetNext (m_nTakenStart);

	/*-----------------------------------------------------------------
	Discard old blocks
	-----------------------------------------------------------------*/
	if (nNext != m_nTakenStart)
	{	m_nTakenStart = nNext;
		if (m_bWaitingForMemory)
		{	m_bWaitingForMemory = FALSE;
			m_smpWaitForMemory.V ();
		}
	}			
	m_mtxLockStream.V ();
	if (rnStatus > 0 && m_bEndOfStream && nFollowingAU < 0)
	{	rnStatus = -1;
		if (!bDontClose)
			return END_OF_STREAM;
	}

	return m_pBuffer + sizeof (int) + sizeof (DWORD) + m_nTakenStart;
}

/*-----------------------------------------------------------------
Releases a previously fetched output block. If bDiscard is TRUE, or its
time stamp is zero, discard the block from the stream
-----------------------------------------------------------------*/
void MediaStreamImp :: Release (BOOL bDiscard)
{
	/*-----------------------------------------------------------------
	Do nothing if no block is fetched
	-----------------------------------------------------------------*/
	if (m_nFetched == 0)
		return;

	m_mtxLockStream.P ();
	if (--m_nFetched == 0)
		m_smpLockOutBlock.V ();
	if (bDiscard)
		Discard ();
	m_mtxLockStream.V ();
}

/*-----------------------------------------------------------------
Close the stream by the sender
-----------------------------------------------------------------*/
void MediaStreamImp :: Close ()
{
	if (!m_bOpenToSend)
		return;
	m_mtxLockStream.P ();
	m_bEndOfStream = true;
	if (m_bWaitingForData)
	{	m_bWaitingForData = false;
		m_smpWaitForData.V ();
	}
	if (m_bWaitingForMemory)
	{	m_bWaitingForMemory = false;
		m_smpWaitForMemory.V ();
	}
	if (--m_nUsedBy == 0)
		m_smpInUse.V ();	// stream is in use till released by both sides
	m_mtxLockStream.V ();
	assert (m_nSN >= 0);
	m_bOpenToSend = false;
	m_smpSending.V ();
}
	
/*-----------------------------------------------------------------
Discard next block
-----------------------------------------------------------------*/
void MediaStreamImp :: Discard ()
{
	m_nTakenStart += GetBlockLength (m_nTakenStart);
	if (m_nTakenStart >= m_nBufferSize ||
		m_nTakenStart != m_nFreeStart &&
		GetBlockLength (m_nTakenStart) == -1
	)
		m_nTakenStart = 0;

	/*-----------------------------------------------------------------
	If the buffer is now empty and no block is currently
	allocated, we can reset the TakenStart and FreeStart values
	-----------------------------------------------------------------*/
	if (m_nTakenStart == m_nFreeStart && m_nAllocatedLength < 0)
	{	m_nFreeStart = 0;
		m_nTakenStart = m_nBufferSize;
	}
	m_bFirstRetrieved = FALSE;
	if (m_bWaitingForMemory)
	{	m_bWaitingForMemory = FALSE;
		m_smpWaitForMemory.V ();
	}
}

/*-----------------------------------------------------------------
Find the offset of the block following the one whose offset is given
Returns -1 if there are no more available blocks.
-----------------------------------------------------------------*/
int MediaStreamImp :: GetNext (int nOffset)
{
	int i = GetBlockLength (nOffset);

	if (i == -1)
		nOffset = 0;
	else
		nOffset += i;
	if (nOffset >= m_nBufferSize)
		nOffset = 0;
	if (nOffset == m_nFreeStart)
		return -1;
	if (GetBlockLength (nOffset) == -1)
		nOffset = 0;
	if (nOffset == m_nFreeStart)
		return -1;
	return nOffset;
}

/*-----------------------------------------------------------------
Trace a stream action into a file
-----------------------------------------------------------------*/
void MediaStreamImp :: Trace (int nOffset, LPCSTR szAction)
{
	if (!m_bTrace)
		return;

	m_mtxLockTraceFile.P ();
	if (m_fTraceFile == NULL)
		m_fTraceFile = fopen ("Trace.txt", "w");
	fprintf (
		m_fTraceFile,
		"Time: %d, stream #: %d, action: %s, size: %d, time-stamp: %d\n",
		m_pClock -> GetTime (),
		m_nSN,
		szAction,
		GetBlockLength (nOffset),
		GetTimeStamp (nOffset)
	);
	m_mtxLockTraceFile.V ();
}

/*-----------------------------------------------------------------
Close the stream by the receiving peer
-----------------------------------------------------------------*/
void MediaStreamImp :: Terminate ()
{
	if (!m_bOpenToReceive)
		return;
	Release ();
	m_mtxLockStream.P ();
	m_bEndOfStream = true;
	m_smpTimer.V ();

	/*-----------------------------------------------------------------
	If the sender waits for memory tell it to go
	-----------------------------------------------------------------*/
	if (m_bWaitingForMemory)
	{	m_bWaitingForMemory = false;
		m_smpWaitForMemory.V ();
	}
	/*-------------------------------------------------------------
	If another thread is the receiver, tell it to go
	-------------------------------------------------------------*/
	if (m_bWaitingForData)
	{	m_bWaitingForData = false;
		m_smpWaitForData.V ();
	}

	m_nAllocatedLength = -1;
	if (--m_nUsedBy == 0)
		m_smpInUse.V ();	// stream is in use till released by both sides
	m_mtxLockStream.V ();

	assert (m_nSN >= 0);
	m_bOpenToReceive = false;
	m_smpReceiving.V ();
	m_smpTimer.V ();
}