/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1997.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS/OD Decoder / Encoder
FILE:       BifsDec.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of SysCoder - the base class for encoding/decoding
			of Systems (OD, BIFS etc.) streams.
-------------------------------------------------------------------------*/
#include "defs.h"
#include <assert.h>
#include "SysCoder.h"

/*-----------------------------------------------------------------
SysCoder constructor for decoding
-----------------------------------------------------------------*/
SysCoder :: SysCoder (FILE *fList, LPCSTR szListTitle, int flags, int version) :
	m_fList (fList),
	m_szType (szListTitle),
	m_flags (flags),
	m_nVersion (version)
{
	m_bDecode = true;
	m_nLastBit = 0;
	m_bNewLine = TRUE;
}

/*-------------------------------------------------------------
SysCoder constructor for encoding
-------------------------------------------------------------*/
SysCoder :: SysCoder (
	FILE *fOutput,
	FILE *fList,
	LPCSTR szListTitle,
	int flags,
	int version
) :
	m_fOutput (fOutput),
	m_fList (fList),
	m_szType (szListTitle),
	m_flags (flags),
	m_nLastByte (0),
	m_nLastBit (0),
	m_bDecode (false),
	m_nVersion (version)
{}

/*-------------------------------------------------------------
Reset the coder for decoding. The coded unit address and length is
stored in member variables
-------------------------------------------------------------*/
void SysCoder :: Reset (LPCBYTE pData, int nLength)
{
	m_pData = pData;
	m_nSize = nLength;
	m_nLastBit = 8;
}

/*-------------------------------------------------------------
For Decode:
Make sure next bit would be fetched from byte boundary. Return the 
number of skipped bits
For Encode:
Pad the current byte and make sure subsequent encoding starts on
a byte boundary.
-------------------------------------------------------------*/
int SysCoder :: Align ()
{
	if (m_bDecode)
	{	int nResult = 8 - m_nLastBit;
		if (nResult > 0)
			ParseInt (nResult, "Align");
		return nResult;
	}
	else if (m_nLastBit > 0)
		EncodeInt (0, 8 - m_nLastBit, "Align");
	return 0;
}

/*-----------------------------------------------------------------
Fetch one byte from the stream
-----------------------------------------------------------------*/
int SysCoder :: GetByte ()
{
	if (m_nSize == 0)
		EndOfUnit ();
	return (int) *m_pData++;
}

/*-----------------------------------------------------------------
Retrieves nBits bits from the BifStream and returns their integer value
Methodology based on GetByte() and GetBits()
-----------------------------------------------------------------*/
int SysCoder :: ParseInt (int nBits, LPCSTR szDesc)
{
	int nReturn = 0, nSaveBits = nBits;
	while (nBits-- > 0)
	{
		nReturn <<= 1;
		nReturn |= GetBit ();
	}
	if((m_flags & TRACE_VERBOSE) && m_fList && nSaveBits > 0)
		fprintf (
			m_fList,
			"%s: bit (%d) %s = %d (0x%x)\n",
			m_szType,
			nSaveBits,
			szDesc,
			nReturn,
			nReturn
		);

	return nReturn;

}

/*-----------------------------------------------------------------
Retrieves one bit from the stream
-----------------------------------------------------------------*/
int SysCoder :: GetBit ()
{
	/*-----------------------------------------------------------------
	Get next byte, if necessary
	-----------------------------------------------------------------*/
	if (m_nLastBit >= 8)
	{	
	 	m_nLastByte = GetByte ();
		m_nLastBit = 0;
		m_nSize--;
	}

	/*-----------------------------------------------------------------
	Retrieve the first bit of the byte, shift the byte so next time
	second bit will be the first
	-----------------------------------------------------------------*/
	m_nLastByte <<= 1;
	++m_nLastBit;
	return (m_nLastByte & 0x100) >> 8;
}

/*-------------------------------------------------------------
Skip the given number of input bytes (excluding the current)
-------------------------------------------------------------*/
void SysCoder :: SkipBytes (int nLength)
{
	Align ();
	if (nLength == 0)
		return;
	m_nSize -= nLength;
	if (m_nSize < 0)
		EndOfUnit ();
	m_pData += nLength;

	if((m_flags & TRACE_VERBOSE) && m_fList)
		fprintf (
			m_fList,
			"%s: bit (%d) skip unrecognized data\n",
			m_szType,
			8 * nLength
		);
}

/*-----------------------------------------------------------------
The following is the float parse function 
-----------------------------------------------------------------*/
float SysCoder :: ParseFloat (LPCSTR szDesc)
{	
	float fResult;
#ifdef WIN32
	char buf [4] = "\0\0\0";
	int i;
	for (i = 0; i < 32; i++)
		buf[3-i/8] |= GetBit () << (7 - i%8);

	fResult = * (float *) buf;

#else
	fResult = 0.0;
#endif

	if((m_flags & TRACE_VERBOSE) && m_fList)
		fprintf (m_fList, "%s: float %s = %f\n", m_szType, szDesc, fResult);
	return fResult;
}


/*-----------------------------------------------------------------
The following is the double parse function
-----------------------------------------------------------------*/
double SysCoder :: ParseDouble (LPCSTR szDesc)
{
	double dResult;

#ifdef WIN32	
	char buf [8] = "\0\0\0\0\0\0\0";
	int i;
	for (i = 0; i < 64; i++)
		buf[7-i/8] |= GetBit () << (7 - i%8);
	
	dResult = * (double *) buf;

#else
	dResult = 0.0;
#endif

	if((m_flags & TRACE_VERBOSE) && m_fList)
		fprintf (m_fList, "%s: double %s = %lf\n", m_szType, szDesc, dResult);

	return dResult;
}

int SysCoder :: m_nIndent = 0;

/*-----------------------------------------------------------------
Trace functions for debug mode
-----------------------------------------------------------------*/
#ifdef TRACE_ENCODING
void SysCoder :: Tracef (LPCSTR szFormat, ...)
{
	if (!(m_flags & TRACE_PARSE) || m_fList == NULL)
		return;
	if (m_bNewLine)
	{	for (int i = m_nIndent; --i >= 0;)
			putc ('\t', m_fList);
		m_bNewLine = FALSE;
	}
	va_list argptr;
	va_start (argptr, szFormat);
	vfprintf (m_fList, szFormat, argptr);
	va_end (argptr);
}
	
void SysCoder :: TraceIndent (int nDirection)
{
	if (m_flags & TRACE_PARSE)
		m_nIndent += nDirection;
}

void SysCoder :: TraceNewLine ()
{
	if (!(m_flags & TRACE_PARSE) || m_fList == NULL)
		return;
	if (!m_bNewLine)
	{	m_bNewLine = TRUE;
		putc ('\n', m_fList);
	}
}
#endif

/*-----------------------------------------------------------------
Encode nValue into nBits bits, and report this action in the list file
-----------------------------------------------------------------*/
void SysCoder :: EncodeInt (int nValue, int nBits, LPCSTR szDesc)
{
	if((m_flags & TRACE_VERBOSE) && m_fList)
	{	if (nBits < sizeof (int) * 8)
			nValue &= (1 << nBits) - 1;
		fprintf (
			m_fList,
			"%s: bit (%d) %s = %d (0x%x)\n",
			m_szType,
			nBits,
			szDesc,
			nValue,
			nValue
		);
	}

	nValue <<= sizeof (int) * 8 - nBits;
	while (--nBits >= 0)
	{	WriteBit (nValue < 0);
		nValue <<= 1;
	}
}

/*-----------------------------------------------------------------
Encode a float value
-----------------------------------------------------------------*/
void SysCoder :: EncodeFloat (float fValue, LPCSTR szDesc)
{
	// writes GenericFloat
#ifdef WIN32
	union
	{	float f;
		char sz [4];
	} value;
	value.f = fValue;

	for (int i = 0; i < 32; i++)
		WriteBit ((value.sz [3 - i / 8] & 1 << (7 - i % 8)) != 0);
#endif		// WIN32

	if((m_flags & TRACE_VERBOSE) && m_fList)
		fprintf (m_fList, "%s: float %s = %f\n", m_szType, szDesc, fValue);
}

/*-----------------------------------------------------------------
Encode a double value
-----------------------------------------------------------------*/
void SysCoder :: EncodeDouble (double dValue, LPCSTR szDesc)
{
#ifdef WIN32
	union
	{	double d;
		char sz [8];
	} value;
	value.d = dValue;

	for (int i = 0; i < 64; i++)
		WriteBit ((value.sz [7 - i / 8] & 1 << (7 - i % 8)) != 0);
#endif		// WIN32

	if((m_flags & TRACE_VERBOSE) && m_fList)
		fprintf (m_fList, "%s: double %s = %lf\n", m_szType, szDesc, dValue);
}

/*-----------------------------------------------------------------
Add one bit to the output stream
-----------------------------------------------------------------*/
void SysCoder :: WriteBit (BOOL bit)
{
	m_nLastByte <<= 1;
	m_nLastByte |= bit;
	if (++m_nLastBit == 8)
	{	m_nLastBit = 0;
		WriteByte ();
		m_nLastByte = 0;
	}
}

/*-----------------------------------------------------------------
Write the last byte to the output stream
-----------------------------------------------------------------*/
void SysCoder :: WriteByte ()
{
	m_buffer += m_nLastByte;
}

/*-------------------------------------------------------------
Reset the buffer and counters
-------------------------------------------------------------*/
void SysCoder :: Reset ()
{
	m_buffer.Cut (0);
	m_nLastByte = 0;
	m_nLastBit = 0;
}

/*-----------------------------------------------------------------
Flush the buffer when finished
-----------------------------------------------------------------*/
void SysCoder :: Flush (DWORD dwTS)
{
	Align ();
	if (m_buffer.GetLength () > 0)
	{	fputc (HIBYTE (HIWORD (dwTS)), m_fOutput);
		fputc (LOBYTE (HIWORD (dwTS)), m_fOutput);
		fputc (HIBYTE (LOWORD (dwTS)), m_fOutput);
		fputc (LOBYTE (LOWORD (dwTS)), m_fOutput);
		Write ();
	}
}

/*-----------------------------------------------------------------
Write the encoded data, preceded by the length, to the output file.
If bShort is true the length is encoded as 16-bit big-endian number,
otherwise it is a 32-bit number.
-----------------------------------------------------------------*/
void SysCoder :: Write (bool bShort)
{
	int nLength = m_buffer.GetLength ();
	if (!bShort)
	{	fputc (HIBYTE (HIWORD (nLength)), m_fOutput);
		fputc (LOBYTE (HIWORD (nLength)), m_fOutput);
	}
	fputc (HIBYTE (LOWORD (nLength)), m_fOutput);
	fputc (LOBYTE (LOWORD (nLength)), m_fOutput);
	fwrite (m_buffer, m_buffer.GetLength (), 1, m_fOutput);
	m_buffer.Cut (0);
}

/*-------------------------------------------------------------
Replace the target encoded file
-------------------------------------------------------------*/
void SysCoder :: ReplaceOutputFile (FILE *fNewOutputFile)
{
	if (m_fOutput)
		fclose (m_fOutput);
	m_fOutput = fNewOutputFile;
}

/*-------------------------------------------------------------
Concatenate the buffer of another coder with this one's
-------------------------------------------------------------*/
void SysCoder :: Concatenate (SysCoder *pSource)
{
	int nTotalBits = pSource ->GetOffset ();
	pSource ->Align ();
	pSource ->Reset (pSource->GetData (), pSource ->GetDataLength ());
	while (--nTotalBits >= 0)
		WriteBit (pSource ->GetBit ());
}