/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Decoder
FILE:       Decoder.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of Decoder - base class for all decoders.
-------------------------------------------------------------------------*/
#include "defs.h"
#include "DecoderImp.h"
#include "MediaStream.h"

/*-----------------------------------------------------------------
Decoder constructor
-----------------------------------------------------------------*/
DecoderImp :: DecoderImp () :
	m_decoderThread (DecoderThread, this)
{
	m_pOutStream = m_pInStream = NULL;
	m_bActive = false;
}

/*-----------------------------------------------------------------
Decoder destructor
-----------------------------------------------------------------*/
DecoderImp :: ~DecoderImp ()
{	
	/*-----------------------------------------------------------------
	Make sure the decoder thread terminates before the object is deleted.
	-----------------------------------------------------------------*/
	ASSERT (m_decoderThread.GetThread () == NULL);
}

/*-----------------------------------------------------------------
The default implementation of Stop() aborts any "wait for time" access
on the input stream and signals the decoder thread to stop. 
-----------------------------------------------------------------*/
void DecoderImp :: Stop ()
{
	m_bActive = false;
	if (m_pInStream)
	{	m_pInStream -> Abort ();
		m_pInStream -> Terminate ();
	}
	if (m_pOutStream)
		m_pOutStream -> Close ();
	WaitForThread ();
}

/*-----------------------------------------------------------------
Start() resets the input stream and open both input and output stream
thread.
-----------------------------------------------------------------*/
void DecoderImp :: Start ()
{
	ASSERT (m_decoderThread.GetThread () == NULL);
	m_bActive = true;
	if (m_pInStream)
		m_pInStream -> Start ();
	if (m_pOutStream)
		m_pOutStream -> Open ();
	if (StartDecoderThread () == -1)
		throw;
}

/*-----------------------------------------------------------------
Run() retrieves units from the input stream, decode them, and dispatch
units to the output stream, till the input stream's end is exceeded.
-----------------------------------------------------------------*/
void DecoderImp :: Run ()
{
	LPBYTE pInData, pOutData;
	int nInSize, nOutSize;
	DWORD dwTime;
	BOOL bEOU;

	while (
		m_bActive &&
		(pInData =
			(LPBYTE) m_pInStream -> FetchNext (nInSize, dwTime, bEOU, TRUE)
		) !=
		END_OF_STREAM
	)
	{
		TRACE("(3) Dec::Run FetchNext done %08x \n", m_pInStream);
		if (m_bActive)
		{	nOutSize = GetOptimalOutputSize (),
			pOutData = (LPBYTE) m_pOutStream -> Allocate (nOutSize, TRUE, FALSE);
			if (pOutData == NULL)
				throw;
			if (pOutData == END_OF_STREAM)
				return;
			while (!Decode (pInData, nInSize, pOutData, nOutSize))
			{	pOutData = (LPBYTE) m_pOutStream -> Reallocate (nOutSize);
				if (pOutData == NULL)
					throw;
				if (pOutData == END_OF_STREAM)
					return;
			}
			m_pOutStream -> Dispatch (nOutSize, dwTime);
		}
		m_pInStream -> Release (TRUE);
	}
}

/*-----------------------------------------------------------------
The Decoder thread function calls the virtual Run() function and 
release the flag semaphore when finishes.
-----------------------------------------------------------------*/
void DecoderImp :: DecoderThread ()
{
	if (Init ())
		Run ();
	Terminate ();
}

/*-----------------------------------------------------------------
Default implementation of SetFormat, picks up the first supplied
format and returns some default values.
-----------------------------------------------------------------*/
BOOL DecoderImp :: SetFormat (
	Capabilities &formatOptions,
	DecoderParams &decoderParams
)
{
	if (formatOptions.GetFormatCapabilityLength () == 0)
		return FALSE;
	decoderParams.m_nStreamNumber = 0;
	decoderParams.m_format = formatOptions.GetFormatCapability (0);
	decoderParams.m_nOptimalAUSize = 256;
	decoderParams.m_nDBSize = 1024;
	decoderParams.m_nCBSize = 1024;
	decoderParams.m_bRequiresFullAUs = TRUE;
	return TRUE;
}
