/*
 * This file contains a function which will decode a JPEG frame in memory.
 *
 * Matthew Leditschke
 * CSELT
 * July, 1997.
 */


#include <stdio.h>
#undef FAR
#include <malloc.h>
#include <setjmp.h>

#include "memdec.h"

/* Initialise the data source manager */
void init_source (j_decompress_ptr cinfo)
{
	mem_source_mgr* srcMgr = (mem_source_mgr*) cinfo->src;

	srcMgr->base.next_input_byte = srcMgr->buffer;
	srcMgr->base.bytes_in_buffer = srcMgr->size;
}

/* Fill the input buffer which the block of memory containing the JPEG */
boolean fill_input_buffer (j_decompress_ptr cinfo)
{
	mem_source_mgr* srcMgr = (mem_source_mgr*) cinfo->src;

	srcMgr->base.next_input_byte = srcMgr->buffer;
	srcMgr->base.bytes_in_buffer = srcMgr->size;
	return TRUE;
}

/* Skip over the given number of bytes */
void skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
	if (num_bytes > 0)
	{
		cinfo->src->next_input_byte += num_bytes;
		cinfo->src->bytes_in_buffer -= num_bytes;
	}
}

/* Just an empty function */
void term_source (j_decompress_ptr cinfo) {}


/*
 * The error manager for the JPEG library.
 */

/* Expanded error manager object for use in this function */
struct mem_error_mgr {
  struct jpeg_error_mgr base;	/* Inherited fields */

  jmp_buf setjmp_buffer;        /* For return to caller */
};

/*
 * This function is called when a fatal error has occurred, and the
 * library must exit.  Return control to JpegDecodeInMemory().
 */
void mem_error_exit (j_common_ptr cinfo)
{
	struct mem_error_mgr* errMgr = (struct mem_error_mgr*) cinfo->err;

	/* Return control to the setjmp point */
	longjmp(errMgr->setjmp_buffer, 1);
}


/*
 * JpegDecodeInMemory
 *
 * Given a compressed JPEG image in the source buffer, decode it,
 * storing the result in the given output buffer.  Also return the
 * width and height of the image.
 *
 * If the output buffer isn't large enough, return FALSE, and set
 * destLength to be the required length.
 *
 * If the image is successfully decoded, return TRUE, and set
 * destLength to the actual length of the output image.
 *
 * If an error occurs decoding the image, the width, height and
 * destLength fields will be set to 0, and TRUE will be returned.
 */

int JpegDecodeInMemory (unsigned char* pSource, int sourceLength,
			unsigned char* pDest, int* destLength,
			unsigned short* width, unsigned short* height)
{
	struct jpeg_decompress_struct cinfo;
	struct mem_error_mgr jerr;
	mem_source_mgr srcMgr;
	int requiredOutputSize;
	JSAMPARRAY rows;
	unsigned int i;


	/* We set up the normal JPEG error routines, then override error_exit. */
	cinfo.err = jpeg_std_error(&jerr.base);
	jerr.base.error_exit = mem_error_exit;

	/* Establish the setjmp return context for mem_error_exit to use */
	if (setjmp(jerr.setjmp_buffer))
	{
		/*
		 * If we get here, the JPEG code has signaled an error.
		 * We need to clean up the JPEG object, close the input file,
		 * and return.
		 */
		jpeg_destroy_decompress(&cinfo);

		/* Return an empty image */
		*width = 0;
		*height = 0;
		*destLength = 0;
		return TRUE;
	}

	/* Initialise the decompression object */
	jpeg_create_decompress(&cinfo);

	/* Specify the data source manager */
	srcMgr.base.init_source		  = init_source;
	srcMgr.base.fill_input_buffer = fill_input_buffer;
	srcMgr.base.skip_input_data	  = skip_input_data;
	srcMgr.base.resync_to_restart = jpeg_resync_to_restart;
	srcMgr.base.term_source		  = term_source;
	srcMgr.buffer = pSource;
	srcMgr.size   = sourceLength;
	cinfo.src = (struct jpeg_source_mgr *)&srcMgr;

	/* Read the header of the image */
	jpeg_read_header(&cinfo, TRUE);
	*width  = cinfo.image_width;
	*height = cinfo.image_height;

	/* and determine how big the output buffer should be */
	/* (allowing space for the image dimensions to be stored) */
	requiredOutputSize = cinfo.image_width * cinfo.image_height * 3;

	/* If the output buffer is not big enough, return FALSE */
	if (*destLength < requiredOutputSize)
	{
		jpeg_destroy_decompress(&cinfo);
		*destLength = requiredOutputSize;
		return FALSE;
	}

	/* Get ready to start decoding */
	jpeg_start_decompress(&cinfo);

	/* Make up an array of pointers to the start of every row in the */
	/* output image */
	rows = (JSAMPARRAY) malloc(sizeof(JSAMPROW) * cinfo.output_height);
	for (i = 0; i < cinfo.output_height; i++)
		rows[i] = pDest + i * cinfo.output_width * 3;

	/* Decode the image */
	while (cinfo.output_scanline < cinfo.output_height)
	{
		jpeg_read_scanlines(&cinfo,
				    &(rows[cinfo.output_scanline]),
				    cinfo.output_height);
	}
	free(rows);

	/* Finish decompression */
	jpeg_finish_decompress(&cinfo);

	/* Release the decompression object */
	jpeg_destroy_decompress(&cinfo);

	/* Inidicate the amount of the output buffer used */
	*destLength = requiredOutputSize;

	return TRUE;

} /* JpegDecodeInMemory */
