/**
Author: Stefano Battista, STMicroelectronics
Date:   1999/04/02
**/

#include "Mpeg4Class.H"

/**
 * static variables for Yuv to Rgb Conversion
 */

struct  Mpeg4Class::yuvconvtable Mpeg4Class::convtab[256];
int     Mpeg4Class::cliptab[768];

/**
 * the entry point to the decoder DLL
 */

void * CreateInstance()
{
  return new Mpeg4Class;
}

/**
 * Constructor and desctructor
 */

Mpeg4Class::Mpeg4Class()
{
  mbInitialised = FALSE;
  mpoDecoder = new MPEG4Decoder;

  // Initialise Yuv to Rgb conversion tables
  {
    for (int i = 0; i < 256; ++i) {
      convtab[i].u1 = int(float(i - 128) * 1.37f);
      convtab[i].u2 = int(float(i - 128) * 0.698f);
      convtab[i].v1 = int(float(i - 128) * 0.336f);
      convtab[i].v2 = int(float(i - 128) * 1.73f);
    }
    
    for (i = 0; i < 256; ++i)
      cliptab[i] = 0;
    for (; i < 512; ++i)
      cliptab[i] = i - 256;
    for (; i < 768; ++i)
      cliptab[i] = 255;
  }
}

Mpeg4Class::~Mpeg4Class()
{
  if (mpoDecoder != NULL)
    delete mpoDecoder;
}

/**
 * Utility methods for Input & Output Formats and Buffers
 */

MediaFormat Mpeg4Class::SetFormat(
    Capabilities * /* poCaps */)
{
#ifdef _MPEG4_RGBA_
  return MediaFormat::RGBA;
#else
  return MediaFormat::RGBA;
#endif
}

MediaFormat Mpeg4Class::GetFormat()
{
#ifdef _MPEG4_RGBA_
  return MediaFormat::RGBA;
#else
  return MediaFormat::RGBA;
#endif
}

int Mpeg4Class::GetOptimalInputSize()
{
  return M4V_OPTIMALINPUTSIZE;
}

int Mpeg4Class::GetOptimalOutputSize()
{
  return miOptimalOutputSize;
}

int Mpeg4Class::GetRecommendedInputBufferSize()
{
  return M4V_RECOMMENDEDINPUTSIZE;
}

int Mpeg4Class::GetRecommendedOutputBufferSize()
{
  return miRecommendedOutputSize;
}

bool Mpeg4Class::SetOutputBufferFormat(int iFormat)
{
  int iComponents;
  iComponents = (GetFormat() == MediaFormat::RGBA) ? 4 : 3;
  switch (iFormat)
  {
  case 2:
    miOutputFormat = iFormat;
    miOutputSize = 352 * 288 * iComponents + 4;
    break;
  case 1:
    miOutputFormat = iFormat;
    break;
  case 0:
    miOutputFormat = iFormat;
    miOutputSize = 176 * 144 * iComponents + 4;
    break;
  default:
    miOutputFormat = 0;
  }

  return true;
}

/**
 * The real thing:
 * Init;
 * Run the Main Loop;
 * Decode a single "time instance" of the Video Object.
 */

bool Mpeg4Class::Init()
{
  TRACE("[Mpeg4::Run] \n");

  mpoDecoder->initcddecoder();

  miWidth = mpoDecoder->GetWidth();
  miHeight = mpoDecoder->GetHeight();

  return true;
}

void Mpeg4Class::Run()
{
  TRACE("[Mpeg4::Run] \n");

  Init();

  SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
  DecoderImp::Run();
}

BOOL Mpeg4Class::Decode(
    unsigned char * pucInput,
    int iInputLength,
    unsigned char * pucOutput,
    int & riOutputLength)
{

  TRACE("[Mpeg4::Decode] \n");

  mpoDecoder->Decode(
    pucInput,
    iInputLength,
    pucOutput,
    riOutputLength);

  return TRUE;
}

/**
 * Yuv 422 ==> Rgb 24 bits (without Alpha)
 */

void Mpeg4Class::Yuv422ToRgb(
  unsigned char* rgbdata,
  short int* yptr,
  short int* uptr,
  short int* vptr,
  int lumcols, int lumrows)
{
  // row loop
  
  for (int j = lumrows >> 1; j > 0; --j) {
    short int* tuptr = uptr;
    short int* tvptr = vptr;
    int i;
    
    // loop chrominance columns 
    
    for (i = lumcols >> 1; i > 0; --i) {
      int u = *tuptr++;
      int v = *tvptr++;
      
      // horizontal downsampling
      
      for (int k = 2; k > 0; --k) { 
        int y = *yptr++;
        
        *rgbdata++ = (cliptab + 256)[y                 + convtab[v].v2]; // r
        *rgbdata++ = (cliptab + 256)[y - convtab[u].u2 - convtab[v].v1]; // g
        *rgbdata++ = (cliptab + 256)[y + convtab[u].u1                ]; // b
      }
    }
    
    for (i = lumcols >> 1; i > 0; --i) {
      // calculate intermediate values that can be recycled
      
      int u = *uptr++;
      int v = *vptr++;
      
      // horizontal downsampling   
      
      for (int k = 2; k > 0; --k) { 
        int y = *yptr++;
        
        *rgbdata++ = (cliptab + 256)[y                 + convtab[v].v2]; // r
        *rgbdata++ = (cliptab + 256)[y - convtab[u].u2 - convtab[v].v1]; // g
        *rgbdata++ = (cliptab + 256)[y + convtab[u].u1                ]; // b
      }
    }
  }
}

/**
 * Yuv 422 ==> Rgb 32 bits (with Alpha)
 */

void Mpeg4Class::Yuv422ToRgba(
  unsigned char* rgbdata,
  short int* yptr,
  short int* uptr,
  short int* vptr,
  short int* aptr,
  int lumcols, int lumrows)
{
  // row loop
  
  for (int j = lumrows >> 1; j > 0; --j) {
    short int* tuptr = uptr;
    short int* tvptr = vptr;
    int i;
    
    // loop chrominance columns 
    
    for (i = lumcols >> 1; i > 0; --i) {
      int u = *tuptr++;
      int v = *tvptr++;
      
      // horizontal downsampling
      
      for (int k = 2; k > 0; --k) { 
        int y = *yptr++;
        int a = *aptr++;
        
        if (a)
        {
          *rgbdata++ = (cliptab + 256)[y                 + convtab[v].v2]; // r
          *rgbdata++ = (cliptab + 256)[y - convtab[u].u2 - convtab[v].v1]; // g
          *rgbdata++ = (cliptab + 256)[y + convtab[u].u1                ]; // b
          *rgbdata++ = a;
        }
        else
        {
          *rgbdata++ = 0;
          *rgbdata++ = 0;
          *rgbdata++ = 0;
          *rgbdata++ = 0;
        }
      }
    }
    
    for (i = lumcols >> 1; i > 0; --i) {
      // calculate intermediate values that can be recycled
      
      int u = *uptr++;
      int v = *vptr++;
      
      // horizontal downsampling   
      
      for (int k = 2; k > 0; --k) { 
        int y = *yptr++;
        int a = *aptr++;
        
        if (a)
        {
          *rgbdata++ = (cliptab + 256)[y                 + convtab[v].v2]; // r
          *rgbdata++ = (cliptab + 256)[y - convtab[u].u2 - convtab[v].v1]; // g
          *rgbdata++ = (cliptab + 256)[y + convtab[u].u1                ]; // b
          *rgbdata++ = a;
        }
        else
        {
          *rgbdata++ = 0;
          *rgbdata++ = 0;
          *rgbdata++ = 0;
          *rgbdata++ = 0;
        }
      }
    }
  }
}

