/*****************************************************************************
 *                                                                          
 * This software module was originally developed by 
 *
 * J. Ignacio Ronda (UPM-GTI / ACTS-MoMuSys)
 *                                                                      
 * and edited by                                                        
 *
 * Fernando Jaureguizar (UPM-GTI / ACTS-MoMuSys)
 * Martina Eckert       (UPM-GTI / ACTS-MoMuSys)
 *
 * in the course of development of the MPEG-4 Video (ISO/IEC 14496-2) standard.
 * This software module is an implementation of a part of one or more MPEG-4 
 * Video (ISO/IEC 14496-2) tools as specified by the MPEG-4 Video (ISO/IEC 
 * 14496-2) standard. 
 *
 * ISO/IEC gives users of the MPEG-4 Video (ISO/IEC 14496-2) standard free 
 * license to this software module or modifications thereof for use in hardware
 * or software products claiming conformance to the MPEG-4 Video (ISO/IEC 
 * 14496-2) standard. 
 *
 * Those intending to use this software module in hardware or software products
 * are advised that its use may infringe existing patents. The original 
 * developer of this software module and his/her company, the subsequent 
 * editors and their companies, and ISO/IEC have no liability for use of this 
 * software module or modifications thereof in an implementation. Copyright is 
 * not released for non MPEG-4 Video (ISO/IEC 14496-2) Standard conforming 
 * products. 
 *
 * ACTS-MoMuSys partners retain full right to use the code for his/her own 
 * purpose, assign or donate the code to a third party and to inhibit third 
 * parties from using the code for non MPEG-4 Video (ISO/IEC 14496-2) Standard
 * conforming products. This copyright notice must be included in all copies or
 * derivative works. 
 *
 * Copyright (c) 1997
 *
 *****************************************************************************/

/***********************************************************HeaderBegin*******
 *                                                                         
 * File:        rc.h
 *
 * Author:      J. Ignacio Ronda, UPM-GTI
 *
 * Created:     18-06-97
 *                                                                         
 * Description: Rate control data types and generic interface
 *              Header for all rate control files
 *
 * Notes:       
 *
 * Modified:
 *      13.11.97 Martina Eckert: Headers, comments, cleaning
 *
 ************************************************************HeaderEnd*******/

#ifndef RC_INC
#define RC_INC

#include "momusys.h"
#include "vm_config.h"
#include "vm_stats.h"


/* ------------------------------------------------------------------------
   RC_HIST: Data structure to store the history of a VOL
   ------------------------------------------------------------------------ */

#define RC_WINDOW_LENGTH 200

typedef struct
   {
   Int  size;           /* Storage capacity */
      
   Int  n;              /* Number of actually stored values */
   Int  prev;           /* 1 if first set of data already stored */
   Int  ptr;            /* Position where next data are going to be stored */
   Int  proc_stat;      /* Flag = 1 if VO is in process */

   Int    *bits_text;   /* Bits for texture */
   Int    *pixels;      /* Number of coded pixels */
   Double *mad_text;    /* Mean average prediction error for texture */
   Double *dist;        /* Distortion for texture */
   Int    *bits_vop;    /* Total bits */ 
   Int    *qp;          /* Employed QP */

   Int     total_skipped;   /* Count of skipped frames */
   Double  total_dist;      /* Total texture distortion */
   Int     total_bits_text; /* Total bits texture */
   Int     total_bits;      /* Total bits */
   Int     total_frames;    /* Total frames */

   Int    last_pred_bits_text;   /* Predicted bits for texture */

   } RC_HIST;

#define rch_get_size(rch)       ((rch)->size)
#define rch_get_n(rch)          ((rch)->n)
#define rch_get_ptr(rch)        ((rch)->ptr)
#define rch_proc_stat(rch)      ((rch)->proc_stat)

#define rch_get_bits_text(rch)  ((rch)->bits_text)
#define rch_get_pixels(rch)     ((rch)->pixels)
#define rch_get_mad_text(rch)   ((rch)->mad_text)
#define rch_get_dist(rch)       ((rch)->dist)
#define rch_get_bits_total(rch) ((rch)->bits_vop) 
#define rch_get_qp(rch)         ((rch)->qp)

#define rch_first(rch)         ((rch)->n==0)

#define rch_get_pred(rch)  ((rch)->last_pred_bits_text)

#include "rc_hist.p"

/* ------------------------------------------------------------------------
   RCQ2_DATA: Q2 algorithm data and parameters
   ------------------------------------------------------------------------ */

#define START_RATE_CONTROL     5    /* Unused */
#define START_MODEL_UPDATING   1
#define START_OUTLIER_REMOVING 3    /* According to the VM should be 1! */
#define MAX_SLIDING_WINDOW     20
#define PAST_PERCENT           0.05
#define SAFETY_MARGIN          0.10
#define SKIP_MARGIN            0.8  /* 80% of VBV (Modif. UPM-GTI) */
#define MAX_QUANT              31
#define MIN_QUANT              1
#define DISCOUNT_NTB           0    /* If 1, not-texture bits of the previous
                                       frame are discounted to compute QP */

/* VOL-level parameters
*/
typedef struct
   {
   Int     bit_rate; /* in bits/second */
   Int     numFrame; /* # of frames remaining of each type in this GOP */
   /*Int     S;*/    /* number of bits generated by this picture type */
   Int     T;        /* target # of bits for the next frame of each type */
   /*Double  X;*/    /* global complexity measure for each picture type */
   /*Double  Q;*/    /* average q-parameter for each picture type */
   Int     R;        /* remaining number of bits assigned to the GOP. */
   Double  Bpp;      /* bit_rate / picture_rate */
   Int     VBV_size;
   Int     VBV_fullness;    /* buffer occupancy */
   Double  X1;       /* second order complexity measure */
   Double  X2;       /* second order complexity measure */
   Int     skip;
   Int     totalskipped;
   Int     codedFr;
   Int     dataNumber;
   Int     headerBitCount;
   Int     QPfirst;  
   } RCQ2_DATA;

#define rcQ2_get_QPfirst(rcc)         ((rcc)->QPfirst)

#include "rc_q2data.p"

/* ------------------------------------------------------------------------
   UPM algorithm data and parameters
   ------------------------------------------------------------------------ */

typedef struct
   {
   Double  X1;       /* second order complexity measure - rate */
   Double  X2;       /* second order complexity measure - rate */
   Double  X3;       /* second order complexity measure - rate */
   Double  Y1;       /* second order complexity measure - distortion */
   Double  Y2;       /* second order complexity measure - distortion */
   Double  Y3;       /* second order complexity measure - distortion */
   } RC_UPM_DATA;

/* For translation of aleatoric id's into a consecutive list: */

typedef struct
   { 
   Int vo_id;         /* vo_id of list element */
   Int vol_id;        /* vol_id of list element */
   } LIST;

#define rcd_get_X1(rcd)      ((rcd)->X1)
#define rcd_get_X2(rcd)      ((rcd)->X2)
#define rcd_get_X3(rcd)      ((rcd)->X3)
#define rcd_get_Y1(rcd)      ((rcd)->Y1)
#define rcd_get_Y2(rcd)      ((rcd)->Y2)
#define rcd_get_Y3(rcd)      ((rcd)->Y3)

#include "rc_upmdata.p"


/* ------------------------------------------------------------------------
   Q2 rate control functions
   ------------------------------------------------------------------------ */

#include "rc_q2.p"

/* ------------------------------------------------------------------------
   UPM rate control functions
   ------------------------------------------------------------------------ */

#include "rc_upm_model.p"
#include "rc_upm1.p"
#include "rc_upm2.p"

/* ------------------------------------------------------------------------
   Generic rate control interface
   ------------------------------------------------------------------------ */

/* Algorithms */
#define RC_ALGORITHM_Q2   1
#define RC_ALGORITHM_UPM1 2 /* Minimization of sum alpha_i D_i */
#define RC_ALGORITHM_UPM2 3 /* Supporting priority list of VOPs */
#define RC_ALGORITHM_UPM3 4 /* Fixed D_i/D_0 ratios */
                            /* with D_i = distortio of VOP i, 
                               alpha_i = weights permitted by the user */

/* Statistical models */
#define RC_MODEL_Q2      0  /* Second order estimation function */
#define RC_MODEL_UPM     1  /* Variable order estimation function */
#define RC_MODEL_FIX     2  /* Fix tables */
#define RC_MODEL_NN      3  /* Neuronal net prediction */

#include "rc_if.p"

/* ------------------------------------------------------------------------
   Buffer control 
   ------------------------------------------------------------------------ */

/* Data type BUFFER to simulate the channel adaptation buffer */

typedef struct 
   {
   Int size;
   Int occup;
   Int frame_bit_rate;
   } BUFFER;

#define buffer_get_occup(buffer) ((buffer)->occup)

#include "rc_buffer.p"

/* ------------------------------------------------------------------------
   Target bit control 
   ------------------------------------------------------------------------ */

/* Parameter struct for decisions of the total number of bits 
   for the next frame */

typedef struct 
   {
   int target_level;
   int buffer_size;
   int bits_frame;
   int N;
   } RC_TB_PARS;

#include "rc_total_bits_upm.p"
#include "rc_total_bits_q2.p"

/* ------------------------------------------------------------------------
   Auxiliary functions
   ------------------------------------------------------------------------ */

#include "rc_util.p"


#endif
/*********************************************************** End of file ***/
