/*****************************************************************************
 *
 * This software module was originally developed by
 *
 *   Dietmar Zaig (Siemens AG / ACTS-MoMuSyS)
 *
 * and edited by
 * 
 *   Fernando Jaureguizar (UPM / ACTS-MoMuSyS)
 *
 * in the course of development of the MPEG-4 Video (ISO/IEC 14496-2) standard.
 * This software module is an implementation of a part of one or more MPEG-4
 * Video (ISO/IEC 14496-2) tools as specified by the MPEG-4 Video (ISO/IEC
 * 14496-2) standard.
 *
 * ISO/IEC gives users of the MPEG-4 Video (ISO/IEC 14496-2) standard free
 * license to this software module or modifications thereof for use in hardware
 * or software products claiming conformance to the MPEG-4 Video (ISO/IEC
 * 14496-2) standard.
 *
 * Those intending to use this software module in hardware or software products
 * are advised that its use may infringe existing patents. The original
 * developer of this software module and his/her company, the subsequent
 * editors and their companies, and ISO/IEC have no liability for use of this
 * software module or modifications thereof in an implementation. Copyright is
 * not released for non MPEG-4 Video (ISO/IEC 14496-2) Standard conforming
 * products.
 *
 * ACTS-MoMuSys partners retain full right to use the code for his/her own
 * purpose, assign or donate the code to a third party and to inhibit third
 * parties from using the code for non MPEG-4 Video (ISO/IEC 14496-2) Standard
 * conforming products. This copyright notice must be included in all copies or
 * derivative works.
 *
 * Copyright (c) 1996
 *
 *****************************************************************************/

/*****************************************************************************/
/*         This software module has been speed optimized                     */
/*                                                                           */
/*         by  Bruno Loret - France Telecom / CNET                           */
/*                                                                           */
/*         in the framework of the ACTS Project MoMuSys                      */
/*****************************************************************************/

/***********************************************************HeaderBegin*******
 *
 * File:        mot_decode.c
 *
 * Author:      Siemens AG - Dietmar Zaig
 * Created:     03.08.96
 *
 * Description: Includes functions needed for the Motion Vector encoding
 *              process. Two of them are also used in the motion estima-
 *              tion, while finding the predicted motion vector
 *
 * Notes:       Based on tmndecode, (C) 1995 Telenor R&D
 *
 * Modified:    Version 1.0: used no shape information
 *              Version 1.1: used inform. about transparent blocks on MB level
 *              Version 1.2: used inform. about transparent blocks on block
 *                 level. Included new rules defined in 3.3.2.7
 *              Version 2.1: motion decoding according to VM 2.1 (without
 *                 B-Vops)
 *              10.09.96 F. Jaureguizar: Reformatted to make the code more
 *                 readable.
 *              18.09.96 F. Jaureguizar: Now new B_TRANSP and MB_TRANSP
 *                 defines are in mot_util.h. Their calls are modified.
 *              01.05.97 Luis Ducla-Soares: Changes to allow 1D prediction
 *                                          used in error resilient mode and
 *                                          to set the predictions to zero after
 *                                          resync markers.
 *              14.05.97 F. Jaureguizar: new commentaries about new f_code
 *                       range according to VM7.0
 *              21.05.97 A. Pacheco (UPM): Modifications in several functions
 *                       to have just one function to compute the MVs prediction
 *                       in both encoder/decoder (find_pmvs). Deletetion of the
 *                       old find_pmv, moved to mot_util.c, and modified.
 *              16.06.97 Angel Pacheco: unified the TRANSPARENT modes.
 *
 ***********************************************************HeaderEnd*********/

/************************    INCLUDE FILES    ********************************/

#include "momusys.h"
#include "mom_bitstream_d.h"
#include "mot_decode.h"
#include "mot_util.h"

/*ok*/static Void
CallingConvention
DeScaleMVD (
    Int  f_code,       /* <-- MV range in 1/2 units: 1=32,2=64,...,7=2048     */
    Int  residual,     /* <-- part of the MV Diff. FLC coded                  */
    Int  vlc_code_mag, /* <-- part of the MV Diff. VLC coded                  */
    Int  pred_vector,  /* <-- MV Prediction component in 1/2 units            */
    Int  *vector       /* --> Obtained MV component in 1/2 units              */
    );

/***********************************************************CommentBegin******
 *
 * -- DecodeVopMotion --
 *
 * Author :
 *
 *
 * Created :
 *      
 *
 * Purpose :
 *      Decodes and stores all vectors coded in the separated motion data
 *      bitstream for one VOP
 *
 * Arguments in :
 *      Vop        *vop,            pointer to VOP structure
 *      Bitstream  *vop_bitstream,  bitstream to be parsed
 *      Trace      *trace,          
 *
 * Arguments in/out :
 *      Image      *B_decisions,    field with number of vectors per MB
 *                                  and indicating which blocks are transparent
 *
 * Arguments /out :
 *      Image      *mot_x,          x-motion vector field
 *      Image      *mot_y,          y-motion vector field
 *      Int        *error_flag      set if an error occured
 *
 * Return values : 
 *      none    
 *
 * Side effects :
 *
 *
 * Description : 
 *      Assumptions:
 *        - The output arrays have already been allocated before
 *          (mot_x, mot_y, B_decisions)
 *        - B_decisions-> is of type 'short int' (otherwise the code must be
 *          written dependable on 'image_type'
 *
 *      Notes :
 *        For discriminating transparent, skipped and intra MB's, 
 *        the array 'B_decisions' is assumed to hold:
 *          - INTRA (= 0) with all zero vectors
 *          - SKIPPED (= 3) with all zero vectors
 *          - MBM_TRANSPARENT (= 2) or transparent, with all zero vectors
 *          - INTER (= 1 or 4) with decoded vectors
 *        In this procedure, only the distinction MBM_TRANSPARENT/non-MBM_TRANSPARENT
 *        is used for prediction. For intra, transparent, skipped,  zero
 *        number of vectors is expected, and zero vectors are written.
 *        All blocks already having been labeled as MBM_TRANSPARENT remain unchanged,
 *        all others having no vectors get the label INTRA (0). This is
 *        supposed to be corrected during texture decoding, when the
 *        information intra/skipped is available.
 *
 *        Structure of the 'mot_[xy]'-fields: 
 *          - size = 4*number of MB's in extended VOP
 *        
 *          - The vector of the block in `MB at pos. (xpos, ypos),
 *            block = 1,2,3,4
 *            xpos = 0,..,w-1, ypos = 0, ...,h-1 is:
 *
 *            *(mot->data->f + (2*ypos+((block-1)>>1))*mot->x + 2*xpos + 
 *             ((block-1)&0x1)) = mv 
 *
 *        B_decisions has the same structure
 *
 *        No B-Vops are supported now
 *
 *        No vector limit checking is done yet
 *
 *        Error signalling should be adapted to the method used in all other
 *        modules. Just now, the flag 'error_flag' is set.
 *
 *        Memory access could be made faster if using ptr[i][j] instead of
 *        *(ptr+i*width + j)
 *
 * See also :
 *
 *
 * Modified :
 *      20-Jan-97 added a trace line ; Jan De Lameillieure (HHI)
 *      01.05.97 Luis Ducla-Soares: added arguments to DecodeMBVec()
 *                                  function call.
 *      08.10.97 Minhua Zhou: added vector padding in the case of INTER4V
 *
 ***********************************************************CommentEnd********/

#if 0
Void
CallingConvention
DecodeVopMotion(
    Vop        *vop,
    Bitstream  *vop_bitstream,
    Trace      *trace,
    Image      *mot_x,
    Image      *mot_y,
    Image      *B_decisions, 
    Int        *error_flag
    )
{
    Int k, h, w, xB, xM, ypos, xpos;
    Int vecnum;            /* number of vectors for current MB */
    Float mvx, mvy;        /* vector components */
    Short *dcsn_data = (Short *) GetImageData(B_decisions);
    Float *motxdata = (Float *) GetImageData(mot_x);
    Float *motydata = (Float *) GetImageData(mot_y);
    
    trace->trace = trace->MB_motion;
    
    h = GetImageSizeY(GetVopY(vop))/16;
    w = GetImageSizeX(GetVopY(vop))/16;
    xB = GetImageSizeX(B_decisions);
    xM = GetImageSizeX(mot_x);
    
    /* for all MBs in the Vop */
    for (ypos=0; ypos<h; ypos++)
        for (xpos=0; xpos<w; xpos++)
        {
            if (trace->trace)           /* added by JDL 20-Jan-97 */
                fprintf(trace->fp_trace, 
                        "\n>>>   Decoding motion for Macroblock %d ........- separate Motion/Shape/Text\n", 
                        (int)(ypos*w+xpos));
            
            /* for all non-transparent MBs */
            
            if (!MB_TRANSP(dcsn_data, xB, xpos, ypos, MBM_TRANSPARENT))
            {
                vecnum = GetMVnum(vop_bitstream, trace);
                
                /* remark: we could also have stored all vectors within
                   DecodeMBVec(), but haven't done it for flexibility reasons.
                   assumption: mot_x,y have type FLOAT 
                   */
                
                if (vecnum == 4){
                    for (k = 0; k < 4; k++)
                    {
                        if (!B_TRANSP(dcsn_data, xB, xpos, ypos, k&0x1, k>>1, MBM_TRANSPARENT))
                        {
                            /* we do not want to destroy information about transparency
                               of blocks */
                            BV(dcsn_data, xB, xpos, ypos, k&0x1, k>>1) = FOURVEC;
                            
/* MVP/Noel */
                            DecodeMBVec (mot_x, mot_y, NULL, B_decisions, vop_bitstream, 
                                         trace, xpos, ypos, 
                                         k + 1, GetVopFCodeFor(vop), 1, 0, error_flag,
                                         &mvx, &mvy, (Int **)NULL); 
                            
                            if (trace->trace)
                            {
                                fprintf(trace->fp_trace, "  >> mvx: %g\n", mvx);
                                fprintf(trace->fp_trace, "  >> mvy: %g\n", mvy);
                            }
                            
                            BV(motxdata, xM, xpos, ypos, k&0x1, k>>1) = mvx; 
                            BV(motydata, xM, xpos, ypos, k&0x1, k>>1) = mvy; 
                            
                        }
                        else
                        {
                            BV(motxdata, xM, xpos, ypos, k&0x1, k>>1) = 0.0; 
                            BV(motydata, xM, xpos, ypos, k&0x1, k>>1) = 0.0; 
                        }
                        
                    }
                    
                    /* VPT / Minhua */
                    if (B_TRANSP(dcsn_data, xB, xpos, ypos, 0, 0, MBM_TRANSPARENT)) {
                        if (!B_TRANSP(dcsn_data, xB, xpos, ypos, 1, 0, MBM_TRANSPARENT)) {
                            BV(motxdata, xM, xpos, ypos, 0, 0) = BV(motxdata, xM, xpos, ypos, 1, 0); 
                            BV(motydata, xM, xpos, ypos, 0, 0) = BV(motxdata, xM, xpos, ypos, 1, 0); 
                        } else {
                            BV(motxdata, xM, xpos, ypos, 0, 0) = BV(motxdata, xM, xpos, ypos, 0, 1); 
                            BV(motydata, xM, xpos, ypos, 0, 0) = BV(motxdata, xM, xpos, ypos, 0, 1); 
                        }
                    }
                    if (B_TRANSP(dcsn_data, xB, xpos, ypos, 1, 0, MBM_TRANSPARENT)) {
                        if (!B_TRANSP(dcsn_data, xB, xpos, ypos, 0, 0, MBM_TRANSPARENT)) {
                            BV(motxdata, xM, xpos, ypos, 1, 0) = BV(motxdata, xM, xpos, ypos, 0, 0); 
                            BV(motydata, xM, xpos, ypos, 1, 0) = BV(motydata, xM, xpos, ypos, 0, 0); 
                        } else {
                            BV(motxdata, xM, xpos, ypos, 1, 0) = BV(motxdata, xM, xpos, ypos, 1, 1); 
                            BV(motydata, xM, xpos, ypos, 1, 0) = BV(motydata, xM, xpos, ypos, 1, 1); 
                        }
                    }
                    if (B_TRANSP(dcsn_data, xB, xpos, ypos, 0, 1, MBM_TRANSPARENT)) {
                        if (!B_TRANSP(dcsn_data, xB, xpos, ypos, 1, 1, MBM_TRANSPARENT)) {
                            BV(motxdata, xM, xpos, ypos, 0, 1) = BV(motxdata, xM, xpos, ypos, 1, 1); 
                            BV(motydata, xM, xpos, ypos, 0, 1) = BV(motydata, xM, xpos, ypos, 1, 1); 
                        } else {
                            BV(motxdata, xM, xpos, ypos, 0, 1) = BV(motxdata, xM, xpos, ypos, 0, 0); 
                            BV(motydata, xM, xpos, ypos, 0, 1) = BV(motydata, xM, xpos, ypos, 0, 0); 
                        }
                    }
                    if (B_TRANSP(dcsn_data, xB, xpos, ypos, 1, 1, MBM_TRANSPARENT)) {
                        if (!B_TRANSP(dcsn_data, xB, xpos, ypos, 0, 1, MBM_TRANSPARENT)) {
                            BV(motxdata, xM, xpos, ypos, 1, 1) = BV(motxdata, xM, xpos, ypos, 0, 1); 
                            BV(motydata, xM, xpos, ypos, 1, 1) = BV(motydata, xM, xpos, ypos, 0, 1); 
                        } else {
                            BV(motxdata, xM, xpos, ypos, 1, 1) = BV(motxdata, xM, xpos, ypos, 1, 0); 
                            BV(motydata, xM, xpos, ypos, 1, 1) = BV(motydata, xM, xpos, ypos, 1, 0); 
                        }
                    }
                    
                } /* end if (vecnum == 4) */
                
                else /* 0 or 1 vectors */
                {
                    if (vecnum == ONEVEC)
                    {
/* MVP/Noel */
                        DecodeMBVec (mot_x, mot_y, NULL, B_decisions, vop_bitstream, trace, xpos, ypos,  
                                     0, GetVopFCodeFor(vop), 1, 0, error_flag, &mvx, &mvy,
                                     (Int **)NULL);  
                    }
                    else
                    {
                        mvx = mvy = 0.0;
                        vecnum = INTRA;
                        /* this is supposed to be corrected during texture decoding, 
                           when the information intra/skipped is available
                           */
                    }
                    if (trace->trace)
                    {
                        fprintf(trace->fp_trace, "  >> mvx: %g\n", mvx);
                        fprintf(trace->fp_trace, "  >> mvy: %g\n", mvy);
                    }
                    
                    for (k = 0; k < 4; k++)  /* every block gets the same values */
                    {
                        /* we do not want to destroy information about transparency of
                           blocks */
                        if (!B_TRANSP(dcsn_data, xB, xpos, ypos, k&0x1, k>>1, MBM_TRANSPARENT))
                            BV(dcsn_data, xB, xpos, ypos, k&0x1, k>>1) = vecnum;
                        
                        BV(motxdata, xM, xpos, ypos, k&0x1, k>>1) = mvx; 
                        BV(motydata, xM, xpos, ypos, k&0x1, k>>1) = mvy; 
                    }
                } /* end 0 or 1 vectors */
                
            } /* for all non-transparent MBs */
            
            else /* write zero-vectors */
            {
                for (k = 0; k < 4; k++)
                {
                    BV(motxdata, xM, xpos, ypos, k&0x1, k>>1) = 0.0; 
                    BV(motydata, xM, xpos, ypos, k&0x1, k>>1) = 0.0; 
                }
                if (trace->trace)
                {
                    fprintf(trace->fp_trace, "  >> mvx: %g\n", 0.0);
                    fprintf(trace->fp_trace, "  >> mvy: %g\n", 0.0);
                }
            }
        }
    
#ifdef _DEBUG_ 
    if(trace->trace)
    {
        /* motion vectors */
        fprintf(trace->fp_trace, "\nMotion Vectors:\n");
        for (ypos=0; ypos<h; ypos++)
        {
            for (xpos=0; xpos<w; xpos++)
            {
                if (BV(dcsn_data, xB, xpos, ypos, 0, 0) != INTRA &&
                    (!MB_TRANSP(dcsn_data, xB, xpos, ypos, MBM_TRANSPARENT)))
                {
                    if (!B_TRANSP(dcsn_data, xB, xpos, ypos, 0, 0, MBM_TRANSPARENT))
                        fprintf(trace->fp_trace, " (%+05.1f %+05.1f)", 
                                BV(motxdata, xM, xpos, ypos, 0, 0),
                                BV(motydata, xM, xpos, ypos, 0, 0));
                    else if (!B_TRANSP(dcsn_data, xB, xpos, ypos, 1, 0, MBM_TRANSPARENT))
                        fprintf(trace->fp_trace, " (%+05.1f %+05.1f)", 
                                BV(motxdata, xM, xpos, ypos, 1, 0),
                                BV(motydata, xM, xpos, ypos, 1, 0));
                    else if (!B_TRANSP(dcsn_data, xB, xpos, ypos, 0, 1, MBM_TRANSPARENT))
                        fprintf(trace->fp_trace, " (%+05.1f %+05.1f)", 
                                BV(motxdata, xM, xpos, ypos, 0, 1),
                                BV(motydata, xM, xpos, ypos, 0, 1));
                    else 
                        fprintf(trace->fp_trace, " (%+05.1f %+05.1f)", 
                                BV(motxdata, xM, xpos, ypos, 1, 1),
                                BV(motydata, xM, xpos, ypos, 1, 1));
                }
                else
                    fprintf(trace->fp_trace, " (   .     . )");
            }
            fprintf(trace->fp_trace, "\n");
        }
        
        fprintf(trace->fp_trace, "\nBlock 0 Motion Vectors:\n");
        for (ypos=0; ypos<h; ypos++)
        {
            for (xpos=0; xpos<w; xpos++)
            {
                if (BV(dcsn_data, xB, xpos, ypos, 0, 0) != INTRA &&
                    (!B_TRANSP(dcsn_data, xB, xpos, ypos, 0, 0, MBM_TRANSPARENT)))
                    fprintf(trace->fp_trace, " (%+05.1f %+05.1f)", 
                            BV(motxdata, xM, xpos, ypos, 0, 0),
                            BV(motydata, xM, xpos, ypos, 0, 0));
                else fprintf(trace->fp_trace, " (   .     . )");
            }
            fprintf(trace->fp_trace, "\n");
        }
        
        fprintf(trace->fp_trace, "\nBlock 1 Motion Vectors:\n");
        for (ypos=0; ypos<h; ypos++)
        {
            for (xpos=0; xpos<w; xpos++)
            {
                if (BV(dcsn_data, xB, xpos, ypos, 1, 0) != INTRA &&
                    (!B_TRANSP(dcsn_data, xB, xpos, ypos, 1, 0, MBM_TRANSPARENT)))
                    fprintf(trace->fp_trace, " (%+05.1f %+05.1f)", 
                            BV(motxdata, xM, xpos, ypos, 1, 0),
                            BV(motydata, xM, xpos, ypos, 1, 0));
                else
                    fprintf(trace->fp_trace, " (   .     . )");
            }
            fprintf(trace->fp_trace, "\n");
        }
        
        fprintf(trace->fp_trace, "\nBlock 2 Motion Vectors:\n");
        for (ypos=0; ypos<h; ypos++)
        {
            for (xpos=0; xpos<w; xpos++)
            {
                if (BV(dcsn_data, xB, xpos, ypos, 0, 1) != INTRA &&
                    (!B_TRANSP(dcsn_data, xB, xpos, ypos, 0, 1, MBM_TRANSPARENT)))
                    fprintf(trace->fp_trace, " (%+05.1f %+05.1f)", 
                            BV(motxdata, xM, xpos, ypos, 0, 1),
                            BV(motydata, xM, xpos, ypos, 0, 1));
                else
                    fprintf(trace->fp_trace, " (   .     . )");
            }
            fprintf(trace->fp_trace, "\n");
        }
        
        fprintf(trace->fp_trace, "\nBlock 3 Motion Vectors:\n");
        for (ypos=0; ypos<h; ypos++)
        {
            for (xpos=0; xpos<w; xpos++)
            {
                if (BV(dcsn_data, xB, xpos, ypos, 1, 1) != INTRA &&
                    (!B_TRANSP(dcsn_data, xB, xpos, ypos, 1, 1, MBM_TRANSPARENT)))
                    fprintf(trace->fp_trace, " (%+05.1f %+05.1f)", 
                            BV(motxdata, xM, xpos, ypos, 1, 1),
                            BV(motydata, xM, xpos, ypos, 1, 1));
                else fprintf(trace->fp_trace, " (   .     . )");
            }
            fprintf(trace->fp_trace, "\n");
        }
        
        fprintf(trace->fp_trace, "\nMB Modes:\n");
        /* MB modes */
        for (ypos=0; ypos<h; ypos++)
        {
            for (xpos=0; xpos<w; xpos++)
            {
                fprintf(trace->fp_trace, "%d ", BV(dcsn_data, xB, xpos, ypos, 0, 0));
            }
            fprintf(trace->fp_trace, "\n");
        }
        
        fprintf(trace->fp_trace, "\nBlock Modes:\n");
        /* block modes */
        for (ypos=0; ypos<2*h; ypos++)
        {
            for (xpos=0; xpos<2*w; xpos++)
            {
                fprintf(trace->fp_trace, "%d ", BV(dcsn_data, xB, xpos/2, 
                                                   ypos/2, xpos&1, ypos&1));
            }
            fprintf(trace->fp_trace, "\n");
        }
    }
    
#endif
}
#endif

/***********************************************************CommentBegin******
 *
 * -- DecodeSpriteVopMotion --
 *
 * Author :Cecile Dufour
 *
 *
 * Created :22-03.97
 *      
 *
 * Purpose :
 *      Decodes and stores all vectors coded in the separated motion data
 *      bitstream for one VOP
 *
 * Arguments in :
 *      Vop        *vop,            pointer to VOP structure
 *      Bitstream  *vop_bitstream,  bitstream to be parsed
 *      Trace      *trace,          
 *
 * Arguments in/out :
 *      Image      *B_decisions,    field with number of vectors per MB
 *                                  and indicating which blocks are transparent
 *
 * Arguments /out :
 *      Image      *mot_x,          x-motion vector field
 *      Image      *mot_y,          y-motion vector field
 *      Int        *error_flag      set if an error occured
 *
 * Return values : 
 *      none    
 *
 * Side effects :
 *
 *
 * Description : 
 *
 * See also : DecodeVopMotion below
 *
 *
 * Modified :
 *      01.05.97 Luis Ducla-Soares: added arguments to DecodeMBVec()
 *                                  function call.
 *
 ***********************************************************CommentEnd********/

#if(__SPRITES__)
Void
CallingConvention
DecodeSpriteVopMotion(
    Vop        *vop,
    Bitstream  *vop_bitstream,
    Trace      *trace,
    Image      *mot_x,
    Image      *mot_y,
    Image      *B_decisions, 
    Int        *error_flag
    )
{
    Int k, h, w, xB, xM, ypos, xpos;
    Int vecnum;            /* number of vectors for current MB */
    Float mvx, mvy;        /* vector components */
    Short *dcsn_data = (Short *) GetImageData(B_decisions);
    Float *motxdata = (Float *) GetImageData(mot_x);
    Float *motydata = (Float *) GetImageData(mot_y);
    
    trace->trace = trace->MB_motion;
    
    h = GetImageSizeY(GetVopY(vop))/16;
    w = GetImageSizeX(GetVopY(vop))/16;
    xB = GetImageSizeX(B_decisions);
    xM = GetImageSizeX(mot_x);
    
    /* for all MBs in the Vop */
    for (ypos=0; ypos<h; ypos++)
        for (xpos=0; xpos<w; xpos++)
        {
            if (trace->trace)           /* added by JDL 20-Jan-97 */
                fprintf(trace->fp_trace, 
                        "\n>>>   Decoding motion for Macroblock %d ........- separate Motion/Shape/Text\n", 
                        (int)(ypos*w+xpos));
            
            /* for all non-transparent MBs */
            
            if (!MB_TRANSP(dcsn_data, xB, xpos, ypos, MBM_TRANSPARENT))
            {
                vecnum = GetSpriteMVnum(vop_bitstream, trace);
                
                if (vecnum == 4){
                    for (k = 0; k < 4; k++)
                    {
                        if (!B_TRANSP(dcsn_data, xB, xpos, ypos, k&0x1, k>>1, MBM_TRANSPARENT))
                        {
                            /* we do not want to destroy information about transparency
                               of blocks */
                            BV(dcsn_data, xB, xpos, ypos, k&0x1, k>>1) = FOURVEC;
/* MVP/Noel */
                            DecodeMBVec (mot_x, mot_y, NULL, B_decisions, vop_bitstream, trace, xpos, ypos,
                                         k + 1, GetVopFCodeFor(vop), 1, 0, error_flag, &mvx,
                                         &mvy, (Int **)NULL);
                            
                            if (trace->trace)
                            {
                                fprintf(trace->fp_trace, "  >> mvx: %g\n", mvx);
                                fprintf(trace->fp_trace, "  >> mvy: %g\n", mvy);
                            }
                            
                            BV(motxdata, xM, xpos, ypos, k&0x1, k>>1) = mvx; 
                            BV(motydata, xM, xpos, ypos, k&0x1, k>>1) = mvy; 
                            
                        }
                        else
                        {
                            BV(motxdata, xM, xpos, ypos, k&0x1, k>>1) = 0.0; 
                            BV(motydata, xM, xpos, ypos, k&0x1, k>>1) = 0.0; 
                        }
                        
                    }
                } /* end if (vecnum == 4) */
                
                else /* 0 or 1 vectors */
                {
                    if (vecnum == ONEVEC)
                    {
/* MVP/Noel */
                        DecodeMBVec (mot_x, mot_y, NULL, B_decisions, vop_bitstream, trace, xpos, ypos,
                                     0, GetVopFCodeFor(vop), 1, 0, error_flag, &mvx, &mvy,
                                     (Int **)NULL);
                    }
                    else
                    {
                        mvx = mvy = 0.0;
                        vecnum = INTRA;
                        /* this is supposed to be corrected during texture decoding, 
                           when the information intra/skipped is available
                           */
                    }
                    if (trace->trace)
                    {
                        fprintf(trace->fp_trace, "  >> mvx: %g\n", mvx);
                        fprintf(trace->fp_trace, "  >> mvy: %g\n", mvy);
                    }
                    
                    for (k = 0; k < 4; k++)  /* every block gets the same values */
                    {
                        /* we do not want to destroy information about transparency of
                           blocks */
                        if (!B_TRANSP(dcsn_data, xB, xpos, ypos, k&0x1, k>>1, MBM_TRANSPARENT))
                            BV(dcsn_data, xB, xpos, ypos, k&0x1, k>>1) = vecnum;
                        
                        BV(motxdata, xM, xpos, ypos, k&0x1, k>>1) = mvx; 
                        BV(motydata, xM, xpos, ypos, k&0x1, k>>1) = mvy; 
                    }
                } /* end 0 or 1 vectors */
                
            } /* for all non-transparent MBs */
            
            else /* write zero-vectors */
            {
                for (k = 0; k < 4; k++)
                {
                    BV(motxdata, xM, xpos, ypos, k&0x1, k>>1) = 0.0; 
                    BV(motydata, xM, xpos, ypos, k&0x1, k>>1) = 0.0; 
                }
                if (trace->trace)
                {
                    fprintf(trace->fp_trace, "  >> mvx: %g\n", 0.0);
                    fprintf(trace->fp_trace, "  >> mvy: %g\n", 0.0);
                }
            }
        }
    
    
}
#endif

/***********************************************************CommentBegin******
 * 
 * -- DecodeMBVec -- 
 *
 * Author :
 *      
 *
 * Created :
 *      
 *
 * Purpose :
 *      Calculates a component of a block (or MB) vector by decoding 
 *      the magnitude & residual of the diff. vector, making the prediction, 
 *      and combining the decoded diff. and the predicted values
 *
 * Arguments in :
 *      Image       *mot,            x/y-motion vector field
 *      Image       *B_decisions,    field with number of vectors per MB
 *                                   (needed for prediction)
 *      Bitstream   *vop_bitstream,  bitstream to be parsed
 *      Trace       *trace,
 *      Int         xpos,            xpos of MB in multiples of 16 (0,..,w-1)
 *      Int         ypos,            ypos of MB in multiples of 16 (0,..,h-1)
 *      Int         block,           0 if one vector per MB, 1..4 else
 *      Int         f_code,          f_code for curretn Vop
 *
 * Arguments out :
 *      Int         *error_flag      set if an error occured
 *
 * Return values  : 
 *      Reconstructed vector component [ in half-pixels units ]    
 *      0 if an error occured
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *      01.05.97 Luis Ducla-Soares: Changes to set the prediction to zero
 *                                  after a resync marker. This is needed in
 *                                  the error resilient mode.
 *      21.05.97 Angel Pacheco: Interface changes to use one call for decoding
 *                              at the same time mvx and mvy (needed to unify
 *                              the predictors in both encoder and decoder).
 *                24.09.97 Noel Brady : added MB_decisions to paramter list and passed
 *                                                                                                                      it into find_pmvs()
 *
 *
 ***********************************************************CommentEnd********/

Int
CallingConvention
DecodeMBVec(
    Image       *mot_x,
    Image       *mot_y,
    Image *MB_decisions,                                /* MVP/Noel */  
    Image       *B_decisions,
    Bitstream   *vop_bitstream, 
    Trace       *trace,
    Int         xpos,
    Int         ypos,
    Int         block,
    Int         f_code,
    Int         error_res_disable,
    Int         after_marker,
    Int         *error_flag,
    Float       *mv_x,       /* pixel units */
    Float       *mv_y,       /* pixel units */
    Int         **slice_nb
    )

{
    Int   pmvx=0, pmvy=0, mvx=0, mvy=0, 
        residualx=0, residualy=0, 
        vlc_code_magx=0, vlc_code_magy=0;
    
    /* decode component x */
    vlc_code_magx = VlcDecMV(vop_bitstream , error_flag , trace);
    if (*error_flag)
        return 0;
    
    if ( (f_code > 1) && (vlc_code_magx!=0) )
        residualx= BitstreamReadBits(vop_bitstream, f_code-1, "MV residual",
                                     trace, CODE);
    else
        residualx=0;
    
    /* decode component y */ 
    vlc_code_magy = VlcDecMV(vop_bitstream , error_flag , trace);
    if (*error_flag)
        return 0;    
    
    if ( (f_code > 1) && (vlc_code_magy!=0) )
        residualy= BitstreamReadBits(vop_bitstream, f_code-1, "MV residual",
                                     trace, CODE);
    else   
        residualy=0; 
    
    /* make prediction */
    if (error_res_disable)
    {
/* MVP/Noel */
        find_pmvs(mot_x, mot_y, MB_decisions, B_decisions, xpos, ypos,
                  block, MBM_TRANSPARENT, error_flag, &pmvx, &pmvy);
    }
    else
    {
        find_pmvsErrRes(mot_x, mot_y, MB_decisions, B_decisions, xpos, ypos,    
                        block, MBM_TRANSPARENT, error_flag, &pmvx, &pmvy, slice_nb);
    }
    
    if (*error_flag)
        return 0;
    
    /* descale */
    DeScaleMVD(f_code, residualx, vlc_code_magx, pmvx, &mvx);
    DeScaleMVD(f_code, residualy, vlc_code_magy, pmvy, &mvy);
    
    *mv_x=((Float)mvx)/2;
    *mv_y=((Float)mvy)/2;
    
    return 1;
}


/***********************************************************CommentBegin******
 *
 * -- DeScaleMVD -- Makes the inverse scaling in the MVD component acording
 *                  to the MV range
 *
 * Author :        
 *    UPM - Fernando Jaureguizar
 *
 * Created :        
 *    8/May/96
 *
 * Purpose :        
 *    Makes the inverse scaling of a Motion Vector Difference (MVD)
 *    component (x or y) according to the MV range. The maximum range that
 *    can be represented is determined by the f_code encoded in the VOP
 *    header.
 *    The MV component is obtained adding the computed MVD component and
 *    the input MV Prediction component. This MV component is corrected
 *    to to fall in the correct range in half pixel units.
 * 
 * Arguments in :     
 *    Int f_code,        MV range in 1/2 units: 1=32,2=64,...,7=2048
 *    Int residual,      FLC coded part from the scaled MV Difference
 *    Int vlc_code_mag,  VLC coded part from the scaled MV Difference
 *    Int pred_vector,   MV Prediction component in 1/2 units
 *
 * Arguments in/out :    
 *
 *
 * Arguments out :    
 *      Int *vector,       Obtained MV component in 1/2 units
 *
 * Return values :    
 *    none
 *
 * Side effects :    
 *    
 *
 * Description :    
 *   1) The MV Diff. component is obtained from the code_mag and from
 *    the residual, according to the f_code.
 *   2) The MV Diff. component is added to the MV Prediction componet to obtain
 *    the MV component.
 *   3) The MV component is inverse scaled to fall in the correct range.
 *
 * See also :
 *    
 *
 * Modified :        
 *    11/May/96: UPM - Fernando Jaureguizar: New input (PMV) and output (MV)
 *               Reconstructed MV component is computed inside this function.
 *    
 ***********************************************************CommentEnd********/

/*ok*/static Void
CallingConvention
DeScaleMVD (
    Int  f_code,       /* <-- MV range in 1/2 units: 1=32,2=64,...,7=2048     */
    Int  residual,     /* <-- part of the MV Diff. FLC coded                  */
    Int  vlc_code_mag, /* <-- part of the MV Diff. VLC coded                  */
    Int  pred_vector,  /* <-- MV Prediction component in 1/2 units            */
    Int  *vector       /* --> Obtained MV component in 1/2 units              */
    )
{
    Int   range;
    Int   scale_factor;
    Int   r_size;
    Int   low;
    Int   high;
    Int   diff_vector;
    
    r_size = f_code-1;
    scale_factor = 1<<r_size;
    range = 32*scale_factor;
    low   = -range;
    high  =  range-1;
    
    if (scale_factor==1 || vlc_code_mag==0)
        diff_vector = vlc_code_mag;
    else
    {
        diff_vector = ((ABS(vlc_code_mag)-1)<<r_size) + residual + 1;
        if (vlc_code_mag < 0)
            diff_vector = -diff_vector;
    }
    
    *vector = pred_vector + diff_vector;
    
    if (*vector < low)
        *vector += 2*range;
    else if (*vector > high)
        *vector -= 2*range;
}

