/******************************************************************************
 *
 * This software module was originally developed by
 *
 *   J.Ignacio Ronda (UPM-GTI / ACTS-MoMuSys)
 *   Angel Pacheco (UPM-GTI / ACTS-MoMuSys)
 *   Fernando Jaureguizar (UPM-GTI / ACTS-MoMuSys)
 *
 * and edited by
 *
 *   Angel Pacheco (UPM-GTI / ACTS-MoMuSys)
 *   Fernando Jaureguizar (UPM-GTI / ACTS-MoMuSys)
 *   Robert Danielsen (Telenor / ACTS-MoMuSys)
 *   Ferran Marques (UPC / ACTS-MoMuSys)
 *   Minhua Zhou (HHI / ACTS-MoMuSys)
 *   C.S. Boon (Matsushita Corporation)
 *   J. Takahashi
 *
 * in the course of development of the MPEG-4 Video (ISO/IEC 14496-2) standard.
 * This software module is an implementation of a part of one or more MPEG-4
 * Video (ISO/IEC 14496-2) tools as specified by the MPEG-4 Video (ISO/IEC
 * 14496-2) standard.
 *
 * ISO/IEC gives users of the MPEG-4 Video (ISO/IEC 14496-2) standard free
 * license to this software module or modifications thereof for use in hardware
 * or software products claiming conformance to the MPEG-4 Video (ISO/IEC
 * 14496-2) standard.
 *
 * Those intending to use this software module in hardware or software products
 * are advised that its use may infringe existing patents. The original
 * developer of this software module and his/her company, the subsequent
 * editors and their companies, and ISO/IEC have no liability for use of this
 * software module or modifications thereof in an implementation. Copyright is
 * not released for non MPEG-4 Video (ISO/IEC 14496-2) standard conforming
 * products.
 *
 * ACTS-MoMuSys partners retain full right to use the code for his/her own
 * purpose, assign or donate the code to a third party and to inhibit third
 * parties from using the code for non MPEG-4 Video (ISO/IEC 14496-2) standard
 * conforming products. This copyright notice must be included in all copies or
 * derivative works.
 *
 * Copyright (c) 1997
 *
 *****************************************************************************/

/*****************************************************************************/
/*         This software module has been speed optimized                     */
/*                                                                           */
/*         by  Bruno Loret - France Telecom / CNET                           */
/*                                                                           */
/*         in the framework of the ACTS Project MoMuSys                      */
/*****************************************************************************/

/***********************************************************HeaderBegin*******
 *
 * File:        mot_padding.c
 *
 * Authors:     UPM - J.Ignacio Ronda / Angel Pacheco
 * Created:     02.02.96
 *
 * Description: Repetitive padding for vops (Y/YUV images) and separate images
 *
 * Flags:       -D_DEBUG_        :  debugging messages are printed
 *              -D_NO_PADDING_   :  no padding is performed
 *              -D_NO_TEXTURE_PADDING_ : TexturePadding() is not carried out
 *
 * Modified:    
 *      version 1.1: Angel Pacheco: changes in the interface to allow
 *               one Vop as input, and to develop the padding over YUV values 
 *               or only over Y, depending on the lumi field
 *                 lumi==1 only pad Y field
 *                 lumi!=1 pad YUV fields
 *      version 1.2: Angel Pacheco: fixes a bug that produces 
 *               incorrect accesses in *_FinalVerticalPadding
 *      version 1.2.1: Angel Pacheco (UPM-GTI): bypassed a bug that produces 
 *               incorrect padding if the shape is one pixel width and is 
 *               touching the right side.
 *               THIS WILL BE FIXED IN THE NEXT VERSION (coming soon)
 *      21.04.96 Robert Danielsen: Reformatted. New headers.
 *      version 2.0: Angel Pacheco: fixed the bug bypassed of 
 *               version 1.2.1 and modified the alpha plane subsampling for 
 *               the chrominaces, acording to the VM2.x
 *      30.08.96 Angel Pacheco & Fernando Jaureguizar:
 *               Padding modification to perform TexturePadding as required 
 *               for VM3.0.
 *      22.10.96 Angel Pacheco: New MB-based padding scheme as 
 *               VM 4.0 describes.
 *      28.11.96 Ferran Marques: introduces changes of (22.10.96)
 *      05.02.97 Angel Pacheco: modified to include the new low 
 *               pass extrapolation padding technique of the VM5.1
 *      05.02.97 Angel Pacheco: modified to fill the pixels 
 *               acording the original (=!reconstructed) alpha values (VM5.1)
 *      17.04.97 Minhua Zhou: added vm_compos.h, modified  
 *               MacroblockBasedPadding.
 *      09.04.97 MInhua Zhou: rewrote MacroblockBasedImagePadding
 *      09.04.97 Minhua Zhou: modified BlockRepetitivePadding
 *      26.05.97 Minhua Zhou: removed DoAlphaAndStatusPlanes(),
 *               Removed DoAlphaUVPlane(), added VopPadding().
 *      16.06.97 Angel Pacheco: renaming the B_* modes by MBM_ modes.
 *               default initialization of some variables to avoid warnings.
 *      13.11.97 Angel Pacheco: formating and new header.
 *           
 ***********************************************************HeaderEnd*********/

/************************    INCLUDE FILES    ********************************/

#include "mot_padding.h"
#include "mot_util.h"
#include "mot_est.h"
#include "vm_compos.h"
#include "mom_image.h"
#include "mom_vop.h"
#include "ring.h"

/*ok*/static Void
CallingConvention
PrintOutMB( Char *s,
            SInt mb_mask[16][16],
            SInt mb_data[16][16],
            FILE *fout
            );
/*ok*/static void
CallingConvention
LoadArea_16(SInt *block, SInt *im, Int x, Int y, Int lx);
/*ok*/static void
CallingConvention
LoadArea_8(SInt *block, SInt *im, Int x, Int y, Int lx);
/*ok*/static Void
CallingConvention
SetArea_16(SInt *block, SInt *im, Int x, Int y, Int lx);
/*ok*/static Void
CallingConvention
SetArea_8(SInt *block, SInt *im, Int x, Int y, Int lx);


/***********************************************************CommentBegin******
 *
 * --    VopPadding --
 *
 * Author :
 *      HHI - Minhua Zhou
 *
 * Created :
 *      26.05.97
 *
 * Purpose :
 *      padding a VOP
 *
 * Arguments in :
 *
 *
 * Arguments in/out :
 *      Vop *curr_vop
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
CallingConvention
VopPadding(
    Vop *curr_vop,
	Int *zero_edge
    )
{    
    Int edge = 16;
    Vop *curr_vop_padded;
    
/*A.Pacheco:1/04/98:static variables are not allowed in MTSafe code*/
/*
ok static Int zero_edge;
*/
    
/*** 10/30 TEST */
/*  printf("PAD "); */
/* 10/30 TEST ***/
    curr_vop_padded = AllocVop (GetVopWidth(curr_vop) + 2 * edge, 
                                GetVopHeight(curr_vop) + 2 * edge);
    curr_vop_padded->hor_spat_ref=curr_vop->hor_spat_ref-edge;  
    curr_vop_padded->ver_spat_ref=curr_vop->ver_spat_ref-edge; 
    PutVopWidth(GetVopWidth(curr_vop)+2*edge,curr_vop_padded);
    /* Store width and height */
    PutVopHeight(GetVopHeight(curr_vop)+2*edge,curr_vop_padded);
    /* as integer multiples of 16 */
    
    
#ifdef _OPTIMIZE_
    if(GetVopShape(curr_vop) /*&& GetVopOBMCDisable(curr_vop)==0*/)
        *zero_edge = 1;
    
    if(*zero_edge!=0)
    {
        MakeImageEdge(GetVopY(curr_vop), edge,GetVopY(curr_vop_padded));
        MakeImageEdge(GetVopU(curr_vop), edge/2, GetVopU(curr_vop_padded));
        MakeImageEdge(GetVopV(curr_vop), edge/2, GetVopV(curr_vop_padded));
        if(GetVopShape(curr_vop))
        {
            MakeImageEdge(GetVopA(curr_vop), edge, GetVopA(curr_vop_padded));
            MakeImageEdge(GetVopAuv(curr_vop), edge/2, GetVopAuv(curr_vop_padded));
        }
        else
        {
            SetConstantImageI(GetVopA(curr_vop_padded), 255);
            SetConstantImageI(GetVopAuv(curr_vop_padded), 255);
        }    
    }
    else
    {
        MakeImageEdgeNoZero(GetVopY(curr_vop), edge, GetVopY(curr_vop_padded));
        MakeImageEdgeNoZero(GetVopU(curr_vop), edge/2, GetVopU(curr_vop_padded));
        MakeImageEdgeNoZero(GetVopV(curr_vop), edge/2, GetVopV(curr_vop_padded));
        if(GetVopShape(curr_vop))
        {
            MakeImageEdgeNoZero(GetVopA(curr_vop), edge, GetVopA(curr_vop_padded));     
            MakeImageEdgeNoZero(GetVopAuv(curr_vop), edge/2, GetVopAuv(curr_vop_padded));
        }
        else
        {
            SetConstantImageI(GetVopA(curr_vop_padded), 255);
            SetConstantImageI(GetVopAuv(curr_vop_padded), 255);
        }
    }
    
    /* set zero_edge for next time */
    /*zero_edge = ((GetVopWidth(curr_vop)-width) || (GetVopHeight(curr_vop)-height) || offset_x || offset_y);*/
    if(GetVopShape(curr_vop) /* && GetVopOBMCDisable(curr_vop)==0 */)
        *zero_edge = 1;
#else
    MakeImageEdge(GetVopY(curr_vop), edge, GetVopY(curr_vop_padded));
    MakeImageEdge(GetVopU(curr_vop), edge/2, GetVopU(curr_vop_padded));
    MakeImageEdge(GetVopV(curr_vop), edge/2, GetVopV(curr_vop_padded));
    MakeImageEdge(GetVopA(curr_vop), edge, GetVopA(curr_vop_padded));
    MakeImageEdge(GetVopAuv(curr_vop), edge/2, GetVopAuv(curr_vop_padded));
#endif
    
    
    curr_vop_padded->arbitrary_shape=GetVopShape(curr_vop);
    if(GetVopShape(curr_vop)==0)
        RepetitivePadding (0, curr_vop_padded);         
    else 
        MacroblockBasedPadding (GetVopA(curr_vop_padded), curr_vop_padded);
    
#ifndef _OPTIMIZE_
    PutVopA(GetVopA(curr_vop_padded),curr_vop);
    PutVopAuv(GetVopAuv(curr_vop_padded),curr_vop);  
    PutVopY(GetVopY(curr_vop_padded),curr_vop);
    PutVopU(GetVopU(curr_vop_padded),curr_vop);
    PutVopV(GetVopV(curr_vop_padded),curr_vop);
#else
    curr_vop->a_chan->f = curr_vop->a_chan->data->s;
    FreeImage(curr_vop->a_chan);
    curr_vop->a_chan = curr_vop_padded->a_chan;
    
    curr_vop->a_uv_chan->f = curr_vop->a_uv_chan->data->s;
    FreeImage(curr_vop->a_uv_chan);
    curr_vop->a_uv_chan = curr_vop_padded->a_uv_chan;
    
    curr_vop->y_chan->f = curr_vop->y_chan->data->s;
    FreeImage(curr_vop->y_chan);
    curr_vop->y_chan = curr_vop_padded->y_chan;
    
    curr_vop->u_chan->f = curr_vop->u_chan->data->s;
    FreeImage(curr_vop->u_chan);
    curr_vop->u_chan = curr_vop_padded->u_chan;
    
    curr_vop->v_chan->f = curr_vop->v_chan->data->s;
    FreeImage(curr_vop->v_chan);
    curr_vop->v_chan = curr_vop_padded->v_chan;
    
#endif
    
    
    FreeImage(GetVopQP(curr_vop_padded));
    /* PutVopQP(GetVopQP(curr_vop_padded),curr_vop); */
    
    SfreeVop(curr_vop_padded);
    
    curr_vop->hor_spat_ref-=edge;  
    curr_vop->ver_spat_ref-=edge; 
    PutVopWidth(GetVopWidth(curr_vop)+2*edge,curr_vop);
    PutVopHeight(GetVopHeight(curr_vop)+2*edge,curr_vop);
}

/***********************************************************CommentBegin******
 *
 * --    CloneVop_TMP --
 *
 * Author :
 *      HHI - Minhua Zhou
 *
 * Created :
 *      29.05.97
 *
 * Purpose :
 *      Clone a VOP
 *
 * Arguments in :
 *     Vop *curr_vop
 *
 * Arguments in/out 
 *
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Vop *
CallingConvention
CloneVop_TMP(
    Vop *curr_vop
    )
{
    Int  edge = 16;
    Vop  *curr_vop_cut;
    SInt *data_in,*data_out;
    Int  height = GetVopHeight(curr_vop),i,j;
    Int  width  = GetVopWidth(curr_vop);
    
    curr_vop_cut = AllocVop (GetVopWidth(curr_vop) - 2 * edge, 
                             GetVopHeight(curr_vop) - 2 * edge);
    
    SetConstantImageI(curr_vop_cut->a_chan, 255);
    SetConstantImageI(curr_vop_cut->a_uv_chan, 255);
    SetConstantImageI(curr_vop_cut->y_chan, 0);
    SetConstantImageI(curr_vop_cut->u_chan, 128);
    SetConstantImageI(curr_vop_cut->v_chan, 128);
    
#if 0
    DirtyTrick(curr_vop_cut);
#endif
    
    PutVopHorSpatRef(GetVopHorSpatRef(curr_vop)+edge,curr_vop_cut);
    PutVopVerSpatRef(GetVopVerSpatRef(curr_vop)+edge,curr_vop_cut);
    
    data_in = (SInt *) GetImageData(GetVopA(curr_vop));
    data_out = (SInt *) GetImageData(GetVopA(curr_vop_cut));
    for (i=0;i<height-2*edge;i++)
        for (j=0;j<width-2*edge;j++)
            data_out[i*(width-2*edge)+j] = data_in[(i+edge)*width+j+edge];
    
    data_in = (SInt *) GetImageData(GetVopY(curr_vop));
    data_out = (SInt *) GetImageData(GetVopY(curr_vop_cut));
    for (i=0;i<height-2*edge;i++)
        for (j=0;j<width-2*edge;j++)
            data_out[i*(width-2*edge)+j] = data_in[(i+edge)*width+j+edge];
    
    width/=2;
    height/=2;
    data_in = (SInt *) GetImageData(GetVopU(curr_vop));
    data_out = (SInt *) GetImageData(GetVopU(curr_vop_cut));
    for (i=0;i<height-edge;i++)
        for (j=0;j<width-edge;j++)
            data_out[i*(width-edge)+j] = data_in[(i+edge/2)*width+j+edge/2];
    
    data_in = (SInt *) GetImageData(GetVopV(curr_vop));
    data_out = (SInt *) GetImageData(GetVopV(curr_vop_cut));
    for (i=0;i<height-edge;i++)
        for (j=0;j<width-edge;j++)
            data_out[i*(width-edge)+j] = data_in[(i+edge/2)*width+j+edge/2];
    
    data_in = (SInt *) GetImageData(GetVopAuv(curr_vop));
    data_out = (SInt *) GetImageData(GetVopAuv(curr_vop_cut));
    for (i=0;i<height-edge;i++)
        for (j=0;j<width-edge;j++)
            data_out[i*(width-edge)+j] = data_in[(i+edge/2)*width+j+edge/2];
    
    MoMuSysCopyImage(GetVopQP(curr_vop),GetVopQP(curr_vop_cut));
    
    return curr_vop_cut;
}

/***********************************************************CommentBegin******
 *
 * -- MacroblockBasedPadding --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      21.10.96
 *
 * Purpose :
 *      Macroblock-based padding of the Vop images, according to the
 *      VM 5.1
 *
 * Arguments in :
 *      Image   *prev_rec_alpha,   Previous reconstructed alpha plane data
 *
 * Arguments in/out :
 *      Vop     *cur_vop,          Vop to pad 
 *
 * Arguments out :
 *
 * Return values :
 *      1 on success, <0 on error
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the padding of the YUV images of a given VOP
 *      needed in the motion estimation and compensation (see section 3.3.1).
 *
 *      1) The padded Vop is NOT extended 16 pixels in all directions (it is
 *         supposed that this extension has been previously done).
 *      2) Padding is performed SEPARATELY FOR EACH of the Y, U/V bands, but
 *         grouped in MB/B units.
 *      3) Padding of the MB/B uses the RECONSTRUCTED ALPHA values of the
 *         luminance or chrominance in this block (input parameter).
 *      4) Padding of chroma 8x8 block is performed by decimating the 16x16
 *         alpha block (by descarding every other row and column of pixels,
 *         starting from the second row and column, with an OR operation
 *         over the alpha pixel values).
 *
 * See also :
 *
 *
 * Modified :
 *      28-01-97 Robert Danielsen: Fix for freeing alpha_modes. Fix by
 *                Noel O'Connor.
 *      06.02.97 Angel Pacheco: the padding of the transparent blocks
 *                 is made looking at the most recently padded boundary block.
 *      17.04.97 Minhua Zhou: rewrote this program
 *      27.07.97 C.S. Boon: Changed to VM8.0 padding -- simplified padding
 *
 ***********************************************************CommentEnd********/

Int
CallingConvention
MacroblockBasedPadding(
    Image   *prev_rec_alpha,/* <-- previous reconstructed alpha plane data  */
    Vop     *cur_vop       /* <-> Vop to pad                                */
    )
{
    Image * alpha_sub_ima;
    
    SInt  * alpha_modes;
    SInt
        * alpha,
        * alpha_uv,
        * alpha_sub;
    
    Int
        i,j,
        i_ref,j_ref, padding_type;
    
#if 0
    SInt
        * buf,
        * buf_a;
    Int
        h,v,
#endif
        
#ifdef _DEBUG
        Image  * y_ima, *u_ima, *v_ima;
#endif
    
#ifdef _NO_PADDING_
    return(2);
#endif
    
#ifndef _OPTIMIZE_    
    Int h_size,w_size;
    Int a_w,a_h;
    
    a_w=GetImageSizeX(GetVopY(cur_vop));
    a_h=GetImageSizeY(GetVopY(cur_vop));
    
    w_size=a_w/MB_SIZE;
    h_size=a_h/MB_SIZE;
#else
    const Int a_w=GetImageSizeX(GetVopY(cur_vop));
    const Int a_h=GetImageSizeY(GetVopY(cur_vop));
    
    const Int w_size=a_w/MB_SIZE;
    const Int h_size=a_h/MB_SIZE;
    
    const Int b_size_2 = MB_SIZE_2/4;
#endif
    
    /*
     * MACROBLOCK-BASED REPETITIVE PADDING 
     */
    
#ifdef _DEBUG
    /* checkings */
    if ( (cur_vop==NULL) || (prev_rec_alpha==NULL) || 
         ((y_ima=GetVopY(cur_vop))==NULL) || ((u_ima=GetVopU(cur_vop))==NULL) || 
         ((v_ima=GetVopV(cur_vop))==NULL) || 
         ((a_w<1))  || 
         ((a_h<1))  || 
         ((alpha=(SInt*)GetImageData(prev_rec_alpha))==NULL) ||
         (a_w!=GetImageSizeX(prev_rec_alpha)) ||
         (a_h!=GetImageSizeY(prev_rec_alpha)) || 
         (GetVopShape(cur_vop)==0)
         )
    {
        fprintf(stderr,"MacroblockBasedPadding: incorrect input parameters\n\n");
        return(-1);
    }
#endif
    
    alpha=(SInt*)GetImageData(prev_rec_alpha);
    
    alpha_sub_ima=GetVopAuv(cur_vop);  
    
    /* array with MBM_TRANSPARENT/MBM_OPAQUE/MBM_BOUNDARY modes per MB */
    alpha_modes=(SInt*)malloc(sizeof(SInt)*w_size*h_size);
    subsamp_alpha(alpha,a_w,a_h,0/*MB-based*/,alpha_modes);
    
    alpha_uv= (SInt*)malloc(sizeof(SInt)*w_size*h_size);
    alpha_sub=(SInt*)GetImageData(alpha_sub_ima);
    subsamp_alpha(alpha_sub,a_w/2,a_h/2,1/*B-based*/,alpha_uv);
    
#ifndef _OPTIMIZE_
    buf = (SInt *)GetImageData(GetVopY(cur_vop));
    for (i=0;i<a_w*a_h;i++)
        if (alpha[i]==0)
            buf[i]=128;
    
    buf = (SInt *)GetImageData(GetVopU(cur_vop));
    for (i=0;i<a_w*a_h/4;i++)
        if (alpha_sub[i]==0)
            buf[i]=128;
    
    buf = (SInt *)GetImageData(GetVopV(cur_vop));
    for (i=0;i<a_w*a_h/4;i++)
        if (alpha_sub[i]==0)
            buf[i]=128; 
#else
#if 0 /* Not very clear why it is not necessary */
    for(v=0;v<h_size;v++)
    {
        for(h=0;h<w_size;h++)
        {
            if(alpha_modes[w_size*v+h]==MBM_TRANSPARENT)
            {
                buf = (SInt *)GetImageData(GetVopY(cur_vop))
                    + v*MB_SIZE*a_w + h*MB_SIZE;
                for(i=0; i<MB_SIZE; i++)
                    for(j=0; j<MB_SIZE; j++)
                        buf[i*MB_SIZE + j] = 128;
                
                buf = (SInt *)GetImageData(GetVopU(cur_vop))
                    + v*b_size*a_w/2 + h*b_size;
                for(i=0; i<B_SIZE; i++)
                    for(j=0; j<B_SIZE; j++)
                        buf[i*B_SIZE + j] = 128;
                
                buf = (SInt *)GetImageData(GetVopV(cur_vop))
                    + v*B_SIZE*a_w/2 + h*B_SIZE;
                for(i=0; i<B_SIZE; i++)
                    for(j=0; j<B_SIZE; j++)
                        buf[i*B_SIZE + j] = 128;
            }
            
            if(alpha_modes[w_size*v+h]==MBM_BOUNDARY)
            {
                buf = (SInt *)GetImageData(GetVopY(cur_vop))
                    + v*MB_SIZE*a_w + h*MB_SIZE;
                buf_a = alpha
                    + v*MB_SIZE*a_w + h*MB_SIZE;                    
                for(i=0; i<MB_SIZE; i++)
                    for(j=0; j<MB_SIZE; j++)
                        if(buf_a[i*MB_SIZE + j] == 0)
                            buf[i*MB_SIZE + j] = 0;
                
                buf = (SInt *)GetImageData(GetVopU(cur_vop))
                    + v*B_SIZE*a_w/2 + h*B_SIZE;
                buf_a = alpha_sub
                    + v*B_SIZE*a_w/2 + h*B_SIZE;
                for(i=0; i<B_SIZE; i++)
                    for(j=0; j<B_SIZE; j++)
                        if(buf_a[i*B_SIZE + j] == 0)
                            buf[i*B_SIZE + j] = 0;
                
                buf = (SInt *)GetImageData(GetVopV(cur_vop))
                    + v*B_SIZE*a_w/2 + h*B_SIZE;
                buf_a = alpha_sub
                    + v*B_SIZE*a_w/2 + h*B_SIZE;
                for(i=0; i<B_SIZE; i++)
                    for(j=0; j<B_SIZE; j++)
                        if(buf_a[i*B_SIZE + j] == 0)
                            buf[i*B_SIZE + j] = 0;
            }
        }
    }        
#endif
#endif
    
    padding_type=1; /*==repetitive padding */
    
    /* Padding of boundary blocks */
    for(j=1;j<(h_size-1);j++)
        for(i=1;i<(w_size-1);i++)
        {
            if(alpha_modes[w_size*j+i]==MBM_BOUNDARY) 
                BlockPadding(j,i,padding_type, MB_SIZE, prev_rec_alpha,0,cur_vop);
            
            if(alpha_uv[w_size*j+i]==MBM_BOUNDARY) 
                BlockPadding(j,i,padding_type, B_SIZE, alpha_sub_ima,1,cur_vop);
        }  
    
    /* Blocks surrounded by transparent blocks should be also
       treated as boundary blocks in the next step of block repetiting */
    for(j=1;j<(h_size-1);j++)
        for(i=1;i<(w_size-1);i++)
        {
            if ((alpha_modes[w_size*j+i]==MBM_OPAQUE)&&
                ((alpha_modes[w_size*j+i-1]==MBM_TRANSPARENT)||
                 (alpha_modes[w_size*j+i+1]==MBM_TRANSPARENT)||
                 (alpha_modes[w_size*(j+1)+i]==MBM_TRANSPARENT)||
                 (alpha_modes[w_size*(j-1)+i]==MBM_TRANSPARENT)))
                alpha_modes[w_size*j+i]=MBM_BOUNDARY;
            
            if ((alpha_uv[w_size*j+i]==MBM_OPAQUE)&&
                ((alpha_uv[w_size*j+i-1]==MBM_TRANSPARENT)||
                 (alpha_uv[w_size*j+i+1]==MBM_TRANSPARENT)||
                 (alpha_uv[w_size*(j+1)+i]==MBM_TRANSPARENT)||
                 (alpha_uv[w_size*(j-1)+i]==MBM_TRANSPARENT)))
                alpha_uv[w_size*j+i]=MBM_BOUNDARY;
        } 
    
    /* priority order for extended padding according to VM8.0. boon */
    for (j=0;j<h_size;j++)
        for (i=0;i<w_size;i++)
        {
            if(alpha_modes[w_size*j+i]==MBM_TRANSPARENT)
            {
                i_ref=j_ref=-1;
                if ((i-1>=0)&&(alpha_modes[w_size*j+i-1]==MBM_BOUNDARY))
                {     /*LEFT*/
                    j_ref =j;i_ref =i-1;
                }
                else if ((j-1>=0)&&(alpha_modes[w_size*(j-1)+i]==MBM_BOUNDARY))
                {   /*TOP*/
                    j_ref =j-1;i_ref =i;
                }
                else if ((i+1<=w_size-1)&&(alpha_modes[w_size*j+i+1]==MBM_BOUNDARY))
                { /*RIGHT*/
                    j_ref =j;i_ref =i+1;
                }
                else if ((j+1<=h_size-1)&&(alpha_modes[w_size*(j+1)+i]==MBM_BOUNDARY))
                { /*BOTTOM*/
                    j_ref =j+1;i_ref =i;
                }
                if (i_ref!=-1) 
                    BlockRepetitivePadding(j,i,j_ref,i_ref,MB_SIZE,
                                           prev_rec_alpha,0,cur_vop);
            }
            
            if (alpha_uv[w_size*j+i]==MBM_TRANSPARENT)
            {
                i_ref=j_ref=-1;
                if ((i-1>=0)&&(alpha_uv[w_size*j+i-1]==MBM_BOUNDARY))
                {  /*LEFT*/
                    j_ref =j;i_ref =i-1;
                }
                else if ((j-1>=0)&&(alpha_uv[w_size*(j-1)+i]==MBM_BOUNDARY))
                {   /*TOP*/
                    j_ref =j-1;i_ref =i;
                }
                else if ((i+1<=w_size-1)&&(alpha_uv[w_size*j+i+1]==MBM_BOUNDARY))
                { /*RIGHT*/
                    j_ref =j;i_ref =i+1;
                }
                else if ((j+1<=h_size-1)&&(alpha_uv[w_size*(j+1)+i]==MBM_BOUNDARY))
                { /*BOTTOM*/
                    j_ref =j+1;i_ref =i;
                }
                if (i_ref!=-1) 
                    BlockRepetitivePadding(j,i,j_ref,i_ref,B_SIZE,
                                           alpha_sub_ima,1,cur_vop);
            }
        }
    
    /* Free alpha_sub_ima  & alpha_uv before return - NO'C, 13/Sep/96 */
    free((Char *)alpha_uv);
    
    free((Char *)alpha_modes);
    return 1;
}

/***********************************************************CommentBegin******
 *
 * -- BlockRepetitivePadding --
 *
 * Author :
 *      UPM-GTI - Angel Pacheco
 *
 * Created :
 *      22.10.96
 *
 * Purpose :
 *      Repetitive padding of an image blocks (section 3.3.1 and figs. 2,3
 *      of VM4.0)
 *
 * Arguments in :
 *      Int   v,               vertical coord. (in block units) of the
 *                             block within the lumi image or chroma images
 *      Int   h,               horizontal coord. (in block units) of the
 *                             block within the lumi image or chroma images
 *      Int   vref,            vertical reference coord. (in block units) of the
 *                             block within the lumi image or chroma images
 *      Int   href,            horizontal reference coord. (in block units) of
 *                             the block within the lumi image or chroma images
 *      Int   block_size,      B_SIZE or MB_SIZE
 *      Image *prev_rec_alpha, previous reconstructed alpha plane (Y or
 *                             subsampled UV)
 *      Int   image_type,      0=Y; 1=U and V
 *
 * Arguments in/out :
 *      Vop   *cur_vop         Vop to pad 
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the repetitive padding of an block (Y or U and V)
 *      of a given VOP. Needed in the estimation/compensation process.
 *
 *      1) Padding is performed for an luminance block (image_type=0)
 *         or for both chrominance blocks separately (image_type=1)
 *      2) Padding of the block must use the RECONSTRUCTED ALPHA values of the
 *         luminance or chrominance in this block.
 *      3) If image_type = 0, the prev_rec_alpha must be the previous
 *         reconstructed alpha plane.
 *         If image_type = 1, the prev_rec_alpha must be the previous
 *         reconstructed alpha plane decimated according to VM.
 *      4) The average of the values of the differents MB/Bs used to pad is
 *         performed in the calling code (here, it is no possible to know how
 *         many reference MB/Bs will infuence the current MB/B).
 *      5) No input checkings are performed!!
 *
 * See also :
 *
 *
 * Modified :
 *      09.05.97 Minhua Zhou: changed operation "+=" to "=" 
 *
 ***********************************************************CommentEnd********/

Void
CallingConvention
BlockRepetitivePadding (
    Int     v,                /* <-- vertical coord. (in block units)        */
    Int     h,                /* <-- horizontal coord. (in block units)      */
    Int     vref,             /* <-- vertical ref. coord. (in block units)   */
    Int     href,             /* <-- horizontal ref. coord. (in block units) */
    Int     block_size,       /* <-- B_SIZE or MB_SIZE                       */
    Image   *prev_rec_alpha,  /* <-- previous reconstructed alpha plane data */
    Int     image_type,       /* <-- 0=Y; 1=U and V                          */
    Vop     *cur_vop          /* <-> Vop to pad                              */
    )

{
#ifndef _OPTIMIZE_
    SInt * y_data,*y_block,*yref_block;
    SInt * u_data,*u_block,*uref_block;
    SInt * v_data,*v_block,*vref_block;
    SInt * a_data;
#else
    SInt * y_data;
    SInt * u_data;
    SInt * v_data;
    
    SInt y_block[16][16],yref_block[16][16];
    SInt u_block[8][8],uref_block[8][8];
    SInt v_block[8][8],vref_block[8][8];
#endif
    Int    direction;
    Int    size_w,pos;
    Int    i,j;
    SInt   *a, *b;
    SInt   val;
    
    /* prev_rec_alpha has, depending on image_type, the right alpha plane
       subsampled for the UV data or for the Y data */
    
    /* block-based repetitive padding */
    
    y_data=(SInt*)GetImageData(GetVopY(cur_vop));
    u_data=(SInt*)GetImageData(GetVopU(cur_vop));
    v_data=(SInt*)GetImageData(GetVopV(cur_vop));
    
    
    if (vref==v)
        direction=((href-h)<0)?LEFT_REFERENCE:RIGHT_REFERENCE;
    else /* href==h */
        direction=((vref-v)<0)?UPPER_REFERENCE:LOWER_REFERENCE;
    
    if(image_type==0)  /* Y */
    {
        size_w=GetImageSizeX(GetVopY(cur_vop));
        
        /* Y repetitive-padding */
#ifndef _OPTIMIZE_
        y_block=(SInt*)LoadArea(y_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
        yref_block=(SInt*)LoadArea(y_data,href*block_size,vref*block_size,
                                   block_size,block_size,size_w);
#else
        LoadArea_16(&y_block[0][0],    y_data, h*16,    v*16,    size_w);
        LoadArea_16(&yref_block[0][0], y_data, href*16, vref*16, size_w);
#endif        
        
#ifdef _OPTIMIZE_
        if ((direction==UPPER_REFERENCE)||
            (direction==LOWER_REFERENCE))
        { 
            /*pos=(direction==UPPER_REFERENCE)?240:0;*/
            if(direction==UPPER_REFERENCE)
                b = &yref_block[15][0];
            else
                b = &yref_block[0][0];
            
            for(j=0;j<16;j++)
            {
                a = &y_block[j][0];
                memcpy(a, b, 16*sizeof(SInt));
            }            
        }
        else
        {
            pos=(direction==LEFT_REFERENCE)?15:0;
            for(j=0;j<16;j++)
            {
#ifndef _DEBUG
                /*printf("Coucou 1 \n");*/
                /*fflush(stdout);*/
                a = &y_block[j][0];
                /*printf("Coucou 2 \n");*/
                /*fflush(stdout);*/
                /*val = 0;*/
                val = yref_block[j][pos];
                /*printf("Coucou 3 \n");*/
                /*fflush(stdout);                */
#else
                a = &y_block[j][0];
                val = yref_block[j][pos];
#endif
                for (i = 0; i < 16; i++)
                {
                    *a++ = val;
                }
            }
        }
        
        SetArea_16(&y_block[0][0],y_data,h*16,v*16,size_w);
#else           
        if ((direction==UPPER_REFERENCE)||
            (direction==LOWER_REFERENCE))
        { 
            pos=(direction==UPPER_REFERENCE)?block_size*(block_size-1):0;
            for(j=0;j<block_size;j++)
                for(i=0;i<block_size;i++)
                    y_block[j*block_size+i]=yref_block[pos+i];
        }
        else
        {
            pos=(direction==LEFT_REFERENCE)?(block_size-1):0;
            for(j=0;j<block_size;j++)
                for(i=0;i<block_size;i++)
                    y_block[j*block_size+i]=yref_block[j*block_size+pos];   
        }
        
        SetArea(y_block,h*block_size,v*block_size,block_size,block_size,
                size_w,y_data);
        
        free((Char*)y_block);
        free((Char*)yref_block);
#endif
        
    }
    else              /* U-V */
    {
        size_w=GetImageSizeX(GetVopU(cur_vop));
        
        /* U-V repetitive-padding */
#ifndef _OPTIMIZE_
        u_block=(SInt*)LoadArea(u_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
        v_block=(SInt*)LoadArea(v_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
        uref_block=(SInt*)LoadArea(u_data,href*block_size,vref*block_size,
                                   block_size,block_size, size_w);
        vref_block=(SInt*)LoadArea(v_data,href*block_size,vref*block_size,
                                   block_size,block_size, size_w);
#else
        LoadArea_8(&u_block[0][0],    u_data, h*8,    v*8,    size_w);
        LoadArea_8(&v_block[0][0],    v_data, h*8,    v*8,    size_w);
        LoadArea_8(&uref_block[0][0], u_data, href*8, vref*8, size_w);
        LoadArea_8(&vref_block[0][0], v_data, href*8, vref*8, size_w);
#endif
        
#ifdef _OPTIMIZE_
        if ((direction==UPPER_REFERENCE)||
            (direction==LOWER_REFERENCE))
        { 
            /*pos=(direction==UPPER_REFERENCE)?56:0;*/
            if(direction==UPPER_REFERENCE)
                b = &uref_block[7][0];
            else
                b = &uref_block[0][0];
            
            for(j=0;j<8;j++)
            {
                a = &u_block[j][0];
                memcpy(a, b, 8*sizeof(SInt));
#if 0
                for (i = 0; i < 8; ++i)
                {
                    *a++ = *b++;
                }
#endif                
            }
            
            if(direction==UPPER_REFERENCE)
                b = &vref_block[7][0];
            else
                b = &vref_block[0][0];
            
            for(j=0;j<8;j++)
            {
                a = &v_block[j][0];
                memcpy(a, b, 8*sizeof(SInt));
#if 0
                for (i = 0; i < 8; ++i)
                {
                    *a++ = *b++;
                }
#endif                
            }
        }
        else
        {
            pos=(direction==LEFT_REFERENCE)?7:0;
            for(j=0;j<8;j++)
            {
                a = &u_block[j][0];
                val = uref_block[j][pos];
                /*memset(a, i, 8*sizeof(SInt));*/
                for (i = 0; i < 8; ++i)
                {
                    *a++ = val;
                }
            }
            
            for(j=0;j<8;j++)
            {
                a = &v_block[j][0];
                val = vref_block[j][pos];
                /*memset(a, i, 8*sizeof(SInt));*/
                for (i = 0; i < 8; ++i)
                {
                    *a++ = val;
                }                
            }
        }
        
        SetArea_8(&u_block[0][0],u_data,h*8,v*8,size_w);
        SetArea_8(&v_block[0][0],v_data,h*8,v*8,size_w);
#else           
        if ((direction==UPPER_REFERENCE)||(direction==LOWER_REFERENCE))
        {   
            pos=(direction==UPPER_REFERENCE)?block_size*(block_size-1):0;
            for(j=0;j<block_size;j++)
                for(i=0;i<block_size;i++)
                {
                    u_block[j*block_size+i]=uref_block[pos+i];   
                    v_block[j*block_size+i]=vref_block[pos+i];   
                }
        }   
        else  
        { 
            pos=(direction==LEFT_REFERENCE)?(block_size-1):0;
            for(j=0;j<block_size;j++)
                for(i=0;i<block_size;i++) 
                {
                    u_block[j*block_size+i]=uref_block[j*block_size+pos];    
                    v_block[j*block_size+i]=vref_block[j*block_size+pos];    
                }
        }   
        
        SetArea(u_block,h*block_size,v*block_size,
                block_size,block_size,size_w,u_data);
        SetArea(v_block,h*block_size,v*block_size,
                block_size,block_size,size_w,v_data);
        
        free((Char*)u_block);
        free((Char*)v_block);
        free((Char*)uref_block);
        free((Char*)vref_block);
#endif
    }
    
} /* BlockRepetitivePadding */

#if 0
Void
BlockRepetitivePadding (
    Int     v,                /* <-- vertical coord. (in block units)        */
    Int     h,                /* <-- horizontal coord. (in block units)      */
    Int     vref,             /* <-- vertical ref. coord. (in block units)   */
    Int     href,             /* <-- horizontal ref. coord. (in block units) */
    Int     block_size,       /* <-- B_SIZE or MB_SIZE                       */
    Image   *prev_rec_alpha,  /* <-- previous reconstructed alpha plane data */
    Int     image_type,       /* <-- 0=Y; 1=U and V                          */
    Vop     *cur_vop          /* <-> Vop to pad                              */
    )

{
    SInt * y_data,*y_block,*yref_block;
    SInt * u_data,*u_block,*uref_block;
    SInt * v_data,*v_block,*vref_block;
    SInt * a_data;
    Int    direction;
    Int    size_w,pos;
    Int    i,j;
    
    /* prev_rec_alpha has, depending on image_type, the right alpha plane
       subsampled for the UV data or for the Y data */
    
    /* block-based repetitive padding */
    
    y_data=(SInt*)GetImageData(GetVopY(cur_vop));
    u_data=(SInt*)GetImageData(GetVopU(cur_vop));
    v_data=(SInt*)GetImageData(GetVopV(cur_vop));
    a_data=(SInt*)GetImageData(prev_rec_alpha);
    
    if (vref==v)
        direction=((href-h)<0)?LEFT_REFERENCE:RIGHT_REFERENCE;
    else /* href==h */
        direction=((vref-v)<0)?UPPER_REFERENCE:LOWER_REFERENCE;
    
    if(image_type==0)  /* Y */
    {
        size_w=GetImageSizeX(GetVopY(cur_vop));
        
        /* Y repetitive-padding */
        y_block=(SInt*)LoadArea(y_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
        yref_block=(SInt*)LoadArea(y_data,href*block_size,vref*block_size,
                                   block_size,block_size,size_w);
        
        if ((direction==UPPER_REFERENCE)||(direction==LOWER_REFERENCE))
        { 
            pos=(direction==UPPER_REFERENCE)?block_size*(block_size-1):0;
            for(j=0;j<block_size;j++)
                for(i=0;i<block_size;i++)
                    y_block[j*block_size+i]=yref_block[pos+i];
        }
        else
        {
            pos=(direction==LEFT_REFERENCE)?(block_size-1):0;
            for(j=0;j<block_size;j++)
                for(i=0;i<block_size;i++)
                    y_block[j*block_size+i]=yref_block[j*block_size+pos];   
        }
        
        SetArea(y_block,h*block_size,v*block_size,block_size,block_size,
                size_w,y_data);
        
        free((Char*)y_block);
        free((Char*)yref_block);
    }
    else              /* U-V */
    {
        size_w=GetImageSizeX(GetVopU(cur_vop));
        
        /* U-V repetitive-padding */
        u_block=(SInt*)LoadArea(u_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
        v_block=(SInt*)LoadArea(v_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
        uref_block=(SInt*)LoadArea(u_data,href*block_size,vref*block_size,
                                   block_size,block_size, size_w);
        vref_block=(SInt*)LoadArea(v_data,href*block_size,vref*block_size,
                                   block_size,block_size, size_w);
        
        if ((direction==UPPER_REFERENCE)||(direction==LOWER_REFERENCE))
        {   
            pos=(direction==UPPER_REFERENCE)?block_size*(block_size-1):0;
            for(j=0;j<block_size;j++)
                for(i=0;i<block_size;i++)
                {
                    u_block[j*block_size+i]=uref_block[pos+i];   
                    v_block[j*block_size+i]=vref_block[pos+i];   
                }
        }   
        else  
        { 
            pos=(direction==LEFT_REFERENCE)?(block_size-1):0;
            for(j=0;j<block_size;j++)
                for(i=0;i<block_size;i++) 
                {
                    u_block[j*block_size+i]=uref_block[j*block_size+pos];    
                    v_block[j*block_size+i]=vref_block[j*block_size+pos];    
                }
        }   
        
        SetArea(u_block,h*block_size,v*block_size,
                block_size,block_size,size_w,u_data);
        SetArea(v_block,h*block_size,v*block_size,
                block_size,block_size,size_w,v_data);
        
        free((Char*)u_block);
        free((Char*)v_block);
        free((Char*)uref_block);
        free((Char*)vref_block);
    }
    
} /* BlockRepetitivePadding */
#endif

/***********************************************************CommentBegin******
 *
 * --   BlockDefaultImagePadding --
 *
 * Author :
 *      UPM-GTI - Angel Pacheco
 *
 * Created :
 *      06.02.97
 *
 * Purpose :
 *      Padding of an 8x8/16x16 image blocks with a default value
 *
 * Arguments in :
 *      Int   v,                  vertical coord. (in block units) of the
 *                                  block within the lumi image or chroma images
 *      Int   h,                  horizontal coord. (in block units) of the
 *                                  block within the lumi image or chroma images
 *      Int   block_size,         B_SIZE or MB_SIZE (block unit)
 *      Int   default_pad_value,  padding value
 *
 * Arguments in/out :
 *      Image cur_ima             image to pad 
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the padding of an block_size x block_size
 *      block from data
 *      No input checkings are performed
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/
#if 0
Void
CallingConvention
BlockDefaultImagePadding (
    Int     v,                /* <-- vertical coord. (in block units)  */
    Int     h,                /* <-- horizontal coord. (in block units)*/
    Int     block_size,       /* <-- B_SIZE or MB_SIZE                 */
    Int     default_pad_value,/* <-- padding value                     */
    Image   *cur_ima          /* <-> Image to pad                      */
    )

{
/*ok*/    static SInt y_pad_block[MB_SIZE*MB_SIZE]= {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
                                               0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
                                               0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
                                               0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
                                               0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
                                               0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
                                               0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
                                               0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
                                               0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
                                               0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
                                               0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
                                               0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
                                               0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
                                               0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
                                               0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, 
                                               0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
/*ok*/    static SInt uv_pad_block[B_SIZE*B_SIZE]=  {128,128,128,128,128,128,128,128,
                                               128,128,128,128,128,128,128,128,
                                               128,128,128,128,128,128,128,128,
                                               128,128,128,128,128,128,128,128,
                                               128,128,128,128,128,128,128,128,
                                               128,128,128,128,128,128,128,128,
                                               128,128,128,128,128,128,128,128,
                                               128,128,128,128,128,128,128,128};
    SInt * data;
    Int    size_w;
    
    data=(SInt*)GetImageData(cur_ima);
    size_w=GetImageSizeX(cur_ima);
    
    if(default_pad_value==0) /* Y image */
        SetArea(y_pad_block,h*block_size,v*block_size,
                block_size,block_size,size_w,data);
    else /* default_pad_value==128 , UV image*/
        SetArea(uv_pad_block,h*block_size,v*block_size,
                block_size,block_size,size_w,data);
    
} /* BlockDefaultImagePadding */
#endif

/***********************************************************CommentBegin******
 *
 * -- BlockPadding --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / Fernando Jaureguizar
 *
 * Created :
 *      30.08.96
 *
 * Purpose :
 *      Repetitive/zero padding of an 8x8/16x16 Y or UV image blocks
 *
 * Arguments in :
 *      Int     v,                  vertical coord. (in block units) of the
 *                                  block within the lumi image or chroma images
 *      Int     h,                  horizontal coord. (in block units) of the
 *                                  block within the lumi image or chroma images
 *      Int     padding_type,       0=zeroes padding; 1=repetitive padding
 *      Int     block_size,         B_SIZE or MB_SIZE
 *      Image   *prev_rec_alpha,    previous reconstructed alpha plane (Y or
 *                                  subsampled UV)
 *      Int     image_type,         0=Y; 1=U and v
 *
 * Arguments in/out :
 *      Vop     *cur_vop            Vop to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the padding of an 8x8 block (Y or U and V) of a
 *      given VOP. Needed in the texture coding.
 *
 *      1) Padding is performed for an 8x8/16x16 luminance block (image_type=0)
 *         or for both chrominance blocks separately (image_type=1)
 *      2) Padding of the block uses the RECONSTRUCTED ALPHA values of the
 *         luminance or chrominance in this block.
 *      3) If image_type = 0, the prev_rec_alpha must be the previous
 *         reconstructed alpha plane.
 *         If image_type = 1, the prev_rec_alpha must be the previous
 *         reconstructed alpha plane decimated according to VM.
 *      4) If all the pixels in an 8x8 block are transparent, their values
 *         are replaced by zero.
 *      5) Performs two paddings depending on the padding_type parameter:
 *           0   ==> padding with zeroes; for residue 8x8 blocks (prediction
 *                   error after the MC)
 *           1   ==> Repetitive padding as in ME/MC; for intra 8x8 blocks
 *      No input checkings are performed
 *
 * See also :
 *
 *
 * Modified :
 *      22.10.96 A. Pacheco: modified to allow blocks of 8 and 16 units.
 *                  ATTENTION: if block_size==16 the padding_type must be 1!!
 *                  in the same way, there are combinations of parameters not 
 *                  checked because they would be not used at all.
 *         27.07.97 C.S. Boon. Modified for VM8.0 padding -- simplified padding
 *
 ***********************************************************CommentEnd********/

Void
CallingConvention
BlockPadding (
    Int     v,                /* <-- vertical coord. (in block units)        */
    Int     h,                /* <-- horizontal coord. (in block units)      */
    Int     padding_type,     /* <-- 0=zeroes padding; 1=repetitive padding  */
    Int     block_size,       /* <-- B_SIZE or MB_SIZE                       */
    Image   *prev_rec_alpha,  /* <-- previous reconstructed alpha plane data */
    Int     image_type,       /* <-- 0=Y; 1=U and V                          */
    Vop     *cur_vop          /* <-> Vop to pad                              */
    )
{
#ifdef _OPTIMIZE_
    SInt * a_data;
    SInt * y_data;
    SInt * u_data;
    SInt * v_data;
    
    SInt a_block[16][16];
    SInt y_block[16][16];
    SInt u_block[8][8];
    SInt v_block[8][8];
    SInt auv_block[8][8];
#else
    SInt * a_data,*a_block;
    SInt * y_data,*y_block;
    SInt * u_data,*u_block;
    SInt * v_data,*v_block;
    
    SInt * status_block;
    
    Byte * ca_block;
    Byte * cy_block;
    Byte * cu_block;
    Byte * cv_block;
#endif
    
    Int    size_w;
    
    Int size = block_size*block_size;
    
    /* prev_rec_alpha has, depending on image_type, the right alpha plane
       subsampled for the UV data or for the Y data */
    
    if(padding_type==0) /* Zero pading */
    {
        ZeroFillBlock (v,h,image_type,prev_rec_alpha,cur_vop);
        return;
    }
    
    /* block-based repetitive padding */
    
    y_data=(SInt*)GetImageData(GetVopY(cur_vop));
    u_data=(SInt*)GetImageData(GetVopU(cur_vop));
    v_data=(SInt*)GetImageData(GetVopV(cur_vop));
    a_data=(SInt*)GetImageData(prev_rec_alpha);
    
    /* creates and inits status_block */
#ifndef _OPTIMIZE_
    if ((status_block=(SInt*)malloc(block_size*block_size*sizeof(SInt)))==NULL)
    {
        fprintf(stderr,"BlockPadding: malloc failed\n");
        return;
    }
    
    for (j=0;j<(block_size*block_size);j++)
        status_block[j]=NO_PAD;
#else
#   if 0
    status_block = AllocImage(block_size, block_size, SHORT_TYPE);
    SetConstantImageI(status_block, NO_PAD);
#   endif
#endif
    
    if(image_type==0)  /* Y */
    {
        size_w=GetImageSizeX(GetVopY(cur_vop));
        
        /* Y repetitive-padding */
#ifndef _OPTIMIZE_
        y_block=(SInt*)LoadArea(y_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
        a_block=(SInt*)LoadArea(a_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
#else
        LoadArea_16(&y_block[0][0], y_data, h*16, v*16, size_w);
        LoadArea_16(&a_block[0][0], a_data, h*16, v*16, size_w);        
#endif
        
        /* modified for VM8.0 padding. boon */
#ifndef _OPTIMIZE_
        cy_block=(Byte *)malloc(sizeof(Byte)*block_size*block_size);
        ca_block=(Byte *)malloc(sizeof(Byte)*block_size*block_size);
#endif
        
#ifndef _OPTIMIZE_
        for(i=0;i<16;i++)
        {
            for (j=0;j<16 ;j++)
            {
                cy_block[i][j] = (Byte) y_block[i][j];
                ca_block[i][j] = (Byte) a_block[i][j];
            }
        }
#endif
        
        
#ifdef _DEBUG
        /*PrintOutMB("Y Before padding",a_block,y_block, stdout); */
#endif
        
        simpl_pad(&y_block[0][0],&a_block[0][0],16,16);
        
#ifdef _DEBUG
        /*PrintOutMB("Y Before padding",a_block,y_block, stdout); */
#endif
        
#ifndef _OPTIMIZE_
        for(i=0;i<16;i++)
        {
            for (j=0;j<16 ;j++)
            {
                y_block[i][j] = (SInt) cy_block[i][j];
                a_block[i][j] = (SInt) ca_block[i][j];
            }
        }
#endif
        
#ifndef _OPTIMIZE_
        free(cy_block);
        free(ca_block);
#endif
        
#ifdef _OPTIMIZE_
        SetArea_16(&y_block[0][0],y_data,h*16,v*16,size_w);
#else
        SetArea(&y_block[0][0],h*16,v*16,16,16,size_w,y_data);
#endif
        
#ifndef _OPTIMIZE_
        free((Char*)y_block);
        free((Char*)status_block);
        free((Char*)a_block);
#endif
    }
    else              /* U-V */
    {
#ifndef _OPTIMIZE_
        cu_block=(Byte *)malloc(sizeof(Byte)*block_size*block_size);
        cv_block=(Byte *)malloc(sizeof(Byte)*block_size*block_size);
        ca_block=(Byte *)malloc(sizeof(Byte)*block_size*block_size);
#endif
        
        size_w=GetImageSizeX(GetVopU(cur_vop));
        
#ifndef _OPTIMIZE_
        /* U-V repetitive-padding */
        u_block=(SInt*)LoadArea(u_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
        v_block=(SInt*)LoadArea(v_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
        a_block=(SInt*)LoadArea(a_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
#else
        LoadArea_8(&u_block[0][0],   u_data,   h*8, v*8, size_w);
        LoadArea_8(&v_block[0][0],   v_data,   h*8, v*8, size_w);
        LoadArea_8(&auv_block[0][0], a_data, h*8, v*8, size_w);
#endif
        
        
#ifndef _OPTIMIZE_
        for(i=0;i<8;i++)
        {
            for (j=0;j<8 ;j++)
            {
                cu_block[i][j] = (Byte) u_block[i][j];
                cv_block[i][j] = (Byte) v_block[i][j];
                cauv_block[i][j] = (Byte) auv_block[i][j];
            }
        }
#endif
        
        simpl_pad(&u_block[0][0],&auv_block[0][0],8,8); /*UV simplif. padding*/
        simpl_pad(&v_block[0][0],&auv_block[0][0],8,8); /*UV simplif. padding*/
        
        
#ifndef _OPTIMIZE_
        for(i=0;i<8;i++)
        {
            for (j=0;j<8 ;j++)
            {
                u_block[i][j] = (SInt) cu_block[i][j];
                v_block[i][j] = (SInt) cv_block[i][j];
                auv_block[i][j] = (SInt) cauv_block[i][j];
            }
        }
#endif
        
#ifndef _OPTIMIZE_
        free(cu_block);
        free(cv_block);
        free(ca_block);
#endif
        
#ifdef _OPTIMIZE_
        SetArea_8(&u_block[0][0],u_data,h*8,v*8,size_w);
        SetArea_8(&v_block[0][0],v_data,h*8,v*8,size_w);
#else
        SetArea(&u_block[0][0],h*8,v*8,8,8,size_w,u_data);
        SetArea(&v_block[0][0],h*8,v*8,8,8,size_w,v_data);
#endif
        
#ifndef _OPTIMIZE_
        free((Char*)a_block);
        free((Char*)u_block);
        free((Char*)v_block);
        free((Char*)status_block);
#endif
    }
} /* BlockPadding */

/***********************************************************CommentBegin******
 *
 * -- simpl_pad --
 *
 * Authors :
 *      C. S. Boon, J. Takahashi
 *
 * Created :
 *      31.07.97
 *
 * Purpose :
 *      Simplified padding of an 8x8/16x16 Y or UV image blocks 
 *
 * Arguments in :
 *      Byte *org_shape    reference shape for padding
 *      int  width         size of image
 *      int  height        size of image
 *
 * Arguments in/out :
 *      Byte *dt           in-data to be padded, out-padded data
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
CallingConvention
simpl_pad(
    SInt *dt,
    SInt *org_shape,
    int  width,
    int height
    )
{
    int  size = width*height;
    int  v, h, adr, prev_adr, tmp_adr, start_adr;
    int  cnt=0, dest;
    SInt other[2];
    SInt  fill_dt;
    SInt src[16];
    SInt *hrep_shp;
    Int i;
    
    hrep_shp=(SInt *)malloc(sizeof(SInt)*size);
    
    /* H phase */
    memcpy(hrep_shp, org_shape, size*sizeof(SInt));
    
    adr = 0;
    for (v=0;v<height;v++)
    {
        tmp_adr = 0;
        prev_adr = -1;
        other[1] = 0;
        while (tmp_adr < width)
        {
            cnt = 0;
            while ( (tmp_adr < width) && (*(hrep_shp+adr+tmp_adr) == 0) )
            {
                cnt++;
                tmp_adr++;
            }
            
            if ( (cnt > 0) && (tmp_adr < width) )
            {
                start_adr = adr + prev_adr + 1;
                if (other[1])
                {            /* left & right */
                    fill_dt = (other[0] + *(dt+adr+tmp_adr) + 1 )>>1;
                    
                    /*memset(dt+start_adr, fill_dt, cnt*sizeof(SInt));*/
                    for(i=0; i<cnt; i++)
                        *(dt+start_adr +i) = fill_dt;
                    /*memset(hrep_shp+start_adr, 255, cnt*sizeof(SInt));*/
                    for(i=0; i<cnt; i++)
                        *(hrep_shp+start_adr +i) = 255;
                }
                else
                {        /* right only */
                    fill_dt = (SInt)*(dt+adr+tmp_adr);
                    /*memset((SInt*)(dt+start_adr), (SInt)fill_dt, (SInt)cnt*sizeof(SInt));*/
                    for(i=0; i<cnt; i++)
                        *(dt+start_adr +i) = fill_dt;
                    /*memset(hrep_shp+start_adr, 255, cnt*sizeof(SInt));*/
                    for(i=0; i<cnt; i++)
                        *(hrep_shp+start_adr +i) = 255;
                }
            }
            
            if (tmp_adr < width)
            {
                cnt = 0;
                other[0] = *(dt + adr + tmp_adr);
                other[1] = *(hrep_shp + adr + tmp_adr);
                prev_adr = tmp_adr;
            }
            tmp_adr++;
        }
        
        if ((cnt > 0)&&(other[1]))
        {  /* left only */
            start_adr = adr + prev_adr + 1;
            fill_dt = other[0];
            /*memset(dt+start_adr, fill_dt, cnt);*/
            for(i=0; i<cnt; i++)
                *(dt+start_adr +i) = fill_dt;
            /*memset(hrep_shp+start_adr, 255, cnt);*/
            for(i=0; i<cnt; i++)
                *(hrep_shp+start_adr +i) = 255;
        }
        adr += width;
    } /*for*/
    
    
    /* V phase */
    adr = 0;
    prev_adr = -1;
    while (adr < size)
    {
        if ( *(hrep_shp+adr) != 0 )
        {
            if (prev_adr < 0)
            {
                prev_adr = 0;
                while (prev_adr < adr)
                {
                    memcpy(dt+prev_adr, dt+adr, width*sizeof(SInt));
                    prev_adr += width;
                }
            }
            else
            {
                if (prev_adr+width < adr)
                {
                    for (h=0;h<width;h++)
                    {
                        src[h] = (*(dt+prev_adr+h) + *(dt+adr+h) + 1 )/2;
                    }
                    prev_adr += width;
                    while (prev_adr < adr)
                    {
                        memcpy(dt+prev_adr, src, width*sizeof(SInt));
                        prev_adr += width;
                    }
                }
                prev_adr = adr;
            }
        }
        adr += width;
    }
    
    dest = prev_adr + width;
    while (dest < adr)
    {
        memcpy(dt+dest, dt+prev_adr, width*sizeof(SInt));
        dest += width;
    }
    free(hrep_shp);
}

/***********************************************************CommentBegin******
 *
 * -- TextureBlockPadding --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / Fernando Jaureguizar
 *
 * Created :
 *      30.08.96
 *
 * Purpose :
 *      LPE/zero padding of an 8x8/16x16 Y or UV image blocks 
 *
 * Arguments in :
 *      Int     v,                  vertical coord. (in block units) of the
 *                                  block within the lumi image or chroma images
 *      Int     h,                  horizontal coord. (in block units) of the
 *                                  block within the lumi image or chroma images
 *      Int     padding_type,       0=zeroes padding; 1=repetitive padding
 *      Int     block_size,         B_SIZE or MB_SIZE 
 *      Image   *prev_ori_alpha,    previous original alpha plane (Y or
 *                                  subsampled UV)
 *      Int     image_type,         0=Y; 1=U and v
 *
 * Arguments in/out :
 *      Vop     *cur_vop            Vop to pad 
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the padding of an 8x8 block (Y or U and V) of a
 *      given VOP. Needed in the texture coding.
 *
 *      1) Padding is performed for an 8x8/16x16 luminance block (image_type=0)
 *         or for both chrominance blocks separately (image_type=1)
 *      2) Padding of the block uses the ORIGINAL ALPHA values of the
 *         luminance or chrominance in this block.
 *      3) If image_type = 0, the prev_rec_alpha must be the previous
 *         original alpha plane.
 *         If image_type = 1, the prev_rec_alpha must be the previous
 *         original alpha plane decimated according to VM.
 *      4) If all the pixels in an 8x8 block are transparent, their values
 *         are replaced by zero.
 *      5) Performs two paddings depending on the padding_type parameter:
 *           0   ==> padding with zeroes; for residue 8x8 blocks (prediction
 *                   error after the MC)
 *           1   ==> Low Pass Extrapolation padding
 *      No input checkings are performed
 *
 * See also :
 *
 *
 * Modified :
 *      22.10.96 A. Pacheco: modified to allow blocks of 8 and 16 units.
 *               ATTENTION: if block_size==16 the padding_type must be 1!! in
 *               the same way, there are combinations of parameters not checked
 *               because they would be not used at all.
 *      05.02.97 A. Pacheco: modified to include the new low pass extrapolation
 *               padding technique of the VM5.1
 *      05.02.97 A. Pacheco: modified to fill the pixels acording the original
 *               alpha values (VM5.1)
 *
 ***********************************************************CommentEnd********/
#if 0
Void
CallingConvention
TextureBlockPadding (
    Int     v,                /* <-- vertical coord. (in block units)        */
    Int     h,                /* <-- horizontal coord. (in block units)      */
    Int     padding_type,     /* <-- 0=zeroes padding; 1=repetitive padding  */
    Int     block_size,       /* <-- B_SIZE or MB_SIZE                       */
    Image   *prev_ori_alpha,  /* <-- previous original alpha plane data */
    Int     image_type,       /* <-- 0=Y; 1=U and V                          */
    Vop     *cur_vop          /* <-> Vop to pad                              */
    )

{
    SInt *a_data,*a_block;
    SInt *y_data,*y_block;
    SInt *u_data,*u_block;
    SInt *v_data,*v_block;
    Int  size_w;
    
    /* prev_ori_alpha has, depending on image_type, the right alpha plane
       subsampled for the UV data or for the Y data */
    
    if(padding_type==0) /* Zero pading */
    {
        ZeroFillBlock (v,h,image_type,prev_ori_alpha,cur_vop); 
        return;
    }
    
    /* block-based repetitive padding */
    
    y_data=(SInt*)GetImageData(GetVopY(cur_vop));
    u_data=(SInt*)GetImageData(GetVopU(cur_vop));
    v_data=(SInt*)GetImageData(GetVopV(cur_vop));
    a_data=(SInt*)GetImageData(prev_ori_alpha);
    
    if(image_type==0)  /* Y */
    {
        size_w=GetImageSizeX(GetVopY(cur_vop));
        
        /* Y repetitive-padding */
        y_block=(SInt*)LoadArea(y_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
        a_block=(SInt*)LoadArea(a_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
        
        LPEPaddingYBlock(a_block,y_block);
        
        SetArea(y_block,h*block_size,v*block_size,
                block_size,block_size,size_w,y_data);
        
        free((Char*)y_block);
        free((Char*)a_block);
    }
    else              /* U-V */
    {
        size_w=GetImageSizeX(GetVopU(cur_vop));
        
        /* U-V repetitive-padding */
        u_block=(SInt*)LoadArea(u_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
        v_block=(SInt*)LoadArea(v_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
        a_block=(SInt*)LoadArea(a_data,h*block_size,v*block_size,
                                block_size,block_size,size_w);
        
        LPEPaddingUVBlock(a_block,u_block,v_block);
        
        SetArea(u_block,h*block_size,v*block_size,
                block_size,block_size,size_w,u_data);
        SetArea(v_block,h*block_size,v*block_size,
                block_size,block_size,size_w,v_data);
        
        free((Char*)a_block);
        free((Char*)u_block);
        free((Char*)v_block);
    }
    
} /* TextureBlockPadding */
#endif

/***********************************************************CommentBegin******
 *
 * -- ZeroFillBlock --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / Fernando Jaureguizar
 *
 * Created :
 *      30.08.96
 *
 * Purpose :
 *      Fills an 8x8 block of the Y image or the 8x8 blocks of the U and V
 *      images with zeroes, depending on the prev_ori_alpha
 *
 * Arguments in :
 *      Int     v,                  vertical coord. (in block units) of the
 *                                  block within the image
 *      Int     h,                  horizontal coord. (in block units) of the
 *                                  block within the image
 *      Int     image_type,         0=Y; 1=U and V
 *      Image   *prev_ori_alpha,    original reconstructed alpha plane data
 *
 * Arguments in/out :
 *      Vop     *cur_vop            Vop to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the zeroes filling of an 8x8 block of the Y
 *      image or the 8x8 blocks of the U and V images of a given VOP, depending
 *      on the prev_ori_alpha values.
 *      Needed in the texture coding.
 *
 *      1) Filling is performed for an 8x8 luminance block (image_type=0)
 *         or for both chrominance blocks separately (image_type=1)
 *
 *      No input checkings are performed
 *
 * See also :
 *
 *
 * Modified :
 *      05.02.97 A. Pacheco: modified to fill the pixels acording the original
 *               alpha values (VM5.1)
 *
 ***********************************************************CommentEnd********/
Void
CallingConvention
ZeroFillBlock (
    Int     v,                /* <-- vertical coord. (in block units)        */
    Int     h,                /* <-- horizontal coord. (in block units)      */
    Int     image_type,       /* <-- 0=Y; 1=U and V                          */
    Image   *prev_ori_alpha,  /* <-- previous original alpha plane data */
    Vop     *cur_vop          /* <-> Vop to pad                              */
    )
{
    SInt * a_data,*a_aux;
    SInt * y_data,*y_aux;
    SInt * u_data,*u_aux;
    SInt * v_data,*v_aux;
    Int    size_w,pos,j;
    
    if(image_type==0)
    {
        /* Y zero-padding */
        size_w=GetImageSizeX(GetVopY(cur_vop));
        
        y_data=y_aux=(SInt*)GetImageData(GetVopY(cur_vop));
        a_data=a_aux=(SInt*)GetImageData(prev_ori_alpha);
        
        for(j=0;j<B_SIZE;j++)
        {
            pos=(v*B_SIZE+j)*size_w+h*B_SIZE;
            
            if(a_data[pos]==0) y_data[pos]=0; pos++;
            if(a_data[pos]==0) y_data[pos]=0; pos++;
            if(a_data[pos]==0) y_data[pos]=0; pos++;
            if(a_data[pos]==0) y_data[pos]=0; pos++;
            
            if(a_data[pos]==0) y_data[pos]=0; pos++;
            if(a_data[pos]==0) y_data[pos]=0; pos++;
            if(a_data[pos]==0) y_data[pos]=0; pos++;
            if(a_data[pos]==0) y_data[pos]=0; 
        }
    }
    else
    {
        /* U-V zero-padding */
        size_w=GetImageSizeX(GetVopU(cur_vop));
        
        u_data=u_aux=(SInt*)GetImageData(GetVopU(cur_vop));
        v_data=v_aux=(SInt*)GetImageData(GetVopV(cur_vop));
        a_data=a_aux=(SInt*)GetImageData(prev_ori_alpha);
        
        for(j=0;j<B_SIZE;j++)
        {
            pos=(v*B_SIZE+j)*size_w+h*B_SIZE;
            
            if(a_data[pos]==0) {u_data[pos]=v_data[pos]=0;} pos++;
            if(a_data[pos]==0) {u_data[pos]=v_data[pos]=0;} pos++;
            if(a_data[pos]==0) {u_data[pos]=v_data[pos]=0;} pos++;
            if(a_data[pos]==0) {u_data[pos]=v_data[pos]=0;} pos++;
            
            if(a_data[pos]==0) {u_data[pos]=v_data[pos]=0;} pos++;
            if(a_data[pos]==0) {u_data[pos]=v_data[pos]=0;} pos++;
            if(a_data[pos]==0) {u_data[pos]=v_data[pos]=0;} pos++;
            if(a_data[pos]==0) {u_data[pos]=v_data[pos]=0;} 
        }
    }
    
} /* ZeroFillBlock */

/***********************************************************CommentBegin******
 *
 * -- AllZeroFillBlock --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / Fernando Jaureguizar
 *
 * Created :
 *      30.08.96
 *
 * Purpose :
 *      Fills an 8x8 block of the Y image or the 8x8 blocks of the U and V
 *      images with zeroes.
 *
 * Arguments in :
 *      Int     v,                  vertical coord. (in block units) of the
 *                                  block within the image
 *      Int     h,                  horizontal coord. (in block units) of the
 *                                  block within the image
 *      Int     image_type,         0=Y; 1=U and V
 *
 * Arguments in/out :
 *      Vop     *cur_vop            Vop to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the zeroes filling of an 8x8 block of the Y
 *      image or the 8x8 blocks of the U and V images of a given VOP.
 *      Needed in the texture coding.
 *
 *      1) Filling is performed for an 8x8 luminance block (image_type=0)
 *         or for both chrominance blocks separately (image_type=1)
 *
 *      No input checkings are performed
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/
#if 0
Void
CallingConvention
AllZeroFillBlock (
    Int     v,                /* <-- vertical coord. (in block units)        */
    Int     h,                /* <-- horizontal coord. (in block units)      */
    Int     image_type,       /* <-- 0=Y; 1=U and V                          */
    Vop     *cur_vop          /* <-> Vop to pad                              */
    )
{
    SInt * y_data,*y_aux;
    SInt * u_data,*u_aux;
    SInt * v_data,*v_aux;
    Int    size_w,pos,j;
    
    
    if(image_type==0)
    {
        /* Y zero-padding */
        size_w=GetImageSizeX(GetVopY(cur_vop));
        
        y_data=y_aux=(SInt*)GetImageData(GetVopY(cur_vop));
        
        for(j=0;j<B_SIZE;j++)
        {
            pos=(v*B_SIZE+j)*size_w+h*B_SIZE;
            
            y_data[pos++]=0;
            y_data[pos++]=0;
            y_data[pos++]=0;
            y_data[pos++]=0;
            
            y_data[pos++]=0;
            y_data[pos++]=0;
            y_data[pos++]=0;
            y_data[pos  ]=0;
        }
    }
    else
    {
        /* U-V zero-padding */
        size_w=GetImageSizeX(GetVopU(cur_vop));
        
        u_data=u_aux=(SInt*)GetImageData(GetVopU(cur_vop));
        v_data=v_aux=(SInt*)GetImageData(GetVopV(cur_vop));
        
        for(j=0;j<B_SIZE;j++)
        {
            pos=(v*B_SIZE+j)*size_w+h*B_SIZE;
            
            u_data[pos]=v_data[pos]=0; pos++;
            u_data[pos]=v_data[pos]=0; pos++;
            u_data[pos]=v_data[pos]=0; pos++;
            u_data[pos]=v_data[pos]=0; pos++;
            
            u_data[pos]=v_data[pos]=0; pos++;
            u_data[pos]=v_data[pos]=0; pos++;
            u_data[pos]=v_data[pos]=0; pos++;
            u_data[pos]=v_data[pos]=0; 
        }
    }
    
} /* AllZeroFillBlock */
#endif

/***********************************************************CommentBegin******
 *
 * -- LPEPaddingYBlock; --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / Fernando Jaureguizar
 *
 * Created :
 *      5.02.97
 *
 * Purpose :
 *      Fills an 8x8 block of the Y data following the LPE technique
 *
 * Arguments in :
 *      SInt*   a_data,             Alpha block
 *
 * Arguments in/out :
 *      SInt*   y_data,             Y block to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the LPE filling of an 8x8 block of the Y
 *      data. Needed in the texture coding.
 *
 *      No input checkings are performed
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/
#if 0
Void
CallingConvention
LPEPaddingYBlock (
    SInt*   a_data,  /* Alpha block                                      */
    SInt*   y_data   /* Y block to pad                                   */
    )
{
    Int    i,j,mean,cont;
    SInt * a_aux,*y_aux;
    
    /* compute the block mean */
    mean=cont=0;
    a_aux=a_data;y_aux=y_data;
    for(i=0;i<64;i++,a_data++,y_data++)
        if (*a_data!=0)
        {
            mean+=*y_data; cont++;
        }
    if(cont==0) 
    {
        fprintf(stdout,"\nERROR: cont==0 and the shape cannot be transparent!\n");
        fflush(stdout);
    } 
    else 
    { 
        mean= mean/cont;
    }
    a_data=a_aux;y_data=y_aux;
    
    /* assign the mean value to the transparent pixels */
    for(i=0;i<64;i++,a_data++,y_data++)  
        if (*a_data==0)  
            *y_data=mean;
    a_data=a_aux;y_data=y_aux; 
    
    /* perform the low pass extrapolation operator over the transparent pixels */
    
    /* first column, first line */
    if(a_data[0]==0)
        y_data[0]=(y_data[8]+y_data[1])/2;
    /* rest, first line */
    for(i=1;i<7;i++)
        if(a_data[i]==0) y_data[i]=(y_data[i-1]+y_data[i+1]+y_data[i+8])/3;
    /* last column, first line */
    if(a_data[7]==0) y_data[7]=(y_data[6]+y_data[15])/2;
    
    for(j=1;j<7;j++)
    {
        /* first column */
        if(a_data[j*8]==0)
            y_data[j*8]=(y_data[(j-1)*8]+y_data[j*8+1]+y_data[(j+1)*8])/3;
        
        for(i=1;i<7;i++)
            if(a_data[j*8+i]==0) 
                y_data[j*8+i]=(y_data[(j-1)*8+i]+y_data[j*8+i-1]+y_data[j*8+i+1]+
                               y_data[(j+1)*8+i])/4;
        
        /* last column */
        if(a_data[j*8+7]==0)
            y_data[j*8+7]=(y_data[(j-1)*8+7]+y_data[j*8+6]+y_data[(j+1)*8+7])/3;
    }
    
    /* first column, last line*/
    if(a_data[56]==0) y_data[56]=(y_data[48]+y_data[57])/2;
    /* rest, last line*/
    for(i=1;i<7;i++)
        if(a_data[56+i]==0)
            y_data[56+i]=(y_data[48+i]+y_data[56+i-1]+y_data[56+i+1])/3; 
    /* last column, last line*/
    if(a_data[63]==0)
        y_data[63]=(y_data[55]+y_data[62])/2;
    
} /* LPEPaddingYBlock */
#endif

/***********************************************************CommentBegin******
 *
 * -- LPEPaddingUVBlock; --
 *
 * Author :
 *      UPM-GTI - Angel Pacheco
 *
 * Created :
 *      5.02.97
 *
 * Purpose :
 *      Fills two 8x8 blocks of the UV data following the LPE technique
 *
 * Arguments in :
 *      SInt*   a_data,             Alpha block
 *
 * Arguments in/out :
 *      SInt*   u_data,             U block to pad
 *      SInt*   v_data,             V block to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the LPE filling of two 8x8 blocks of the UV
 *      data. Needed in the texture coding.
 *
 *      No input checkings are performed
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/
#if 0
Void
CallingConvention
LPEPaddingUVBlock (
    SInt*   a_data,  /* Alpha block                                      */
    SInt*   u_data,  /* U block to pad                                   */
    SInt*   v_data   /* V block to pad                                   */
    )
{
    Int  i,j,mean_u, mean_v,cont;
    SInt *a_aux,*u_aux,*v_aux;
    
    /* compute the block mean */
    mean_u=mean_v=cont=0;
    a_aux=a_data; 
    u_aux=u_data; 
    v_aux=v_data;
    
    for(i=0;i<64;i++,a_data++,u_data++,v_data++)
        if (*a_data!=0)
        {
            mean_u+=*u_data;
            mean_v+=*v_data; cont++;
        }
    if(cont==0)
    {
        fprintf(stdout,"\nERROR: cont==0 and the shape cannot be transparent!\n");
        fflush(stdout);
    }
    else
    {
        mean_u=mean_u/cont;
        mean_v=mean_v/cont;
    }
    a_data=a_aux;
    u_data=u_aux; 
    v_data=v_aux;
    
    /* assign the mean value to the transparent pixels */
    for(i=0;i<64;i++,a_data++,u_data++,v_data++)  
        if (*a_data==0)  
        {
            *u_data=mean_u;
            *v_data=mean_v;
        }
    a_data=a_aux;
    u_data=u_aux; 
    v_data=v_aux;
    
    /* perform the low pass extrapolation operator over the transparent pixels */
    
    /* first column, first line */
    if(a_data[0]==0)
    {
        u_data[0]=(u_data[8]+u_data[1])/2;
        v_data[0]=(v_data[8]+v_data[1])/2;
    }
    /* rest, first line */
    for(i=1;i<7;i++)
        if(a_data[i]==0)
        {
            u_data[i]=(u_data[i-1]+u_data[i+1]+u_data[i+8])/3;
            v_data[i]=(v_data[i-1]+v_data[i+1]+v_data[i+8])/3;
        }
    /* last column, first line */
    if(a_data[7]==0)
    {
        u_data[7]=(u_data[6]+u_data[15])/2;
        v_data[7]=(v_data[6]+v_data[15])/2;
    }
    
    for(j=1;j<7;j++)
    {
        /* first column */
        if(a_data[j*8]==0)
        {
            u_data[j*8]=(u_data[(j-1)*8]+u_data[j*8+1]+u_data[(j+1)*8])/3;
            v_data[j*8]=(v_data[(j-1)*8]+v_data[j*8+1]+v_data[(j+1)*8])/3;
        }
        
        for(i=1;i<7;i++)
            if(a_data[j*8+i]==0) 
            {
                u_data[j*8+i]=(u_data[(j-1)*8+i]+u_data[j*8+i-1]+u_data[j*8+i+1]+
                               u_data[(j+1)*8+i])/4;
                v_data[j*8+i]=(v_data[(j-1)*8+i]+v_data[j*8+i-1]+v_data[j*8+i+1]+
                               v_data[(j+1)*8+i])/4;
            }
        
        /* last column */
        if(a_data[j*8+7]==0)
        {
            u_data[j*8+7]=(u_data[(j-1)*8+7]+u_data[j*8+6]+u_data[(j+1)*8+7])/3;
            v_data[j*8+7]=(v_data[(j-1)*8+7]+v_data[j*8+6]+v_data[(j+1)*8+7])/3;
        }
    }
    
    /* first column, last line*/
    if(a_data[56]==0)
    {
        u_data[56]=(u_data[48]+u_data[57])/2;
        v_data[56]=(v_data[48]+v_data[57])/2;
    }
    /* rest, last line*/
    for(i=1;i<7;i++)
        if(a_data[56+i]==0)
        {
            u_data[56+i]=(u_data[48+i]+u_data[56+i-1]+u_data[56+i+1])/3; 
            v_data[56+i]=(v_data[48+i]+v_data[56+i-1]+v_data[56+i+1])/3; 
        }
    /* last column, last line*/
    if(a_data[63]==0)
    {
        u_data[63]=(u_data[55]+u_data[62])/2;
        v_data[63]=(v_data[55]+v_data[62])/2;
    }
    
} /* LPEPaddingUVBlock */
#endif

/***********************************************************CommentBegin******
 *
 * -- PadMB --
 *
 * Author :  Noel Brady
 *      
 *
 * Created :
 *      14.11.97
 *
 * Purpose :
 *      
 *
 * Arguments in :
 *      Vop     *rec_vop          alpha plane data
 *      Image   *MB_decisions     encoding mode for each MB
 *      Int     i
 *      Int     j
 *
 * Arguments in/out
 *      Vop     *cur_vop          Vop to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :  Pads the blocks of a MB. Should be used prior to encoding
 *                                                              the MB.
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/
#if 0
Void
PadMB(
    Vop     *rec_vop,         /* <-- previous original alpha plane data */
    Image   *MB_decisions,    /* <-- encoding mode for each MB               */
    Vop     *cur_vop,         /* <-> Vop to pad                              */
    Int     i,
    Int     j  
    )
{
    Image *alpha = GetVopA(rec_vop);
    Image *alpha_uv = GetVopAuv(rec_vop);
    Image *block = AllocImage(8,8,SHORT_TYPE);
    SInt  *mb_modes = (SInt *)GetImageData(MB_decisions);
    SInt  *p = (SInt *)GetImageData(block);
    Int         w_size = GetImageSizeX(MB_decisions),
        mb_pos,
        padding_type,
        transparent_nb,
        mb_pos_x = i*MB_SIZE,
        mb_pos_y = j*MB_SIZE,
        type,k;
    
    
    mb_pos=w_size*j+i;
    if (mb_modes[mb_pos]==MBM_INTRA)
        padding_type=1;
    else /* MBM_INTER16) || MBM_INTER8 or (MBM_TRANSPARENT||MBM_OUT??*/
        padding_type=0;
    
    /* Y blocks */
    
    transparent_nb = 0;
    GetSubImage(alpha,block,mb_pos_x,mb_pos_y);
    for (k=0;k<64;k++) if (!p[k]) transparent_nb++;
    if (transparent_nb == 64) type = MBM_TRANSPARENT;
    else if (transparent_nb == 0) type = MBM_OPAQUE;
    else type = MBM_BOUNDARY;   
    
    if(type==MBM_TRANSPARENT)
        AllZeroFillBlock(j*2  ,i*2  ,0,cur_vop);
    else if (type==MBM_BOUNDARY)
        TextureBlockPadding(j*2,i*2,padding_type,B_SIZE,alpha,
                            0,cur_vop);
    /* else MBM_OPAQUE, do nothing */
    
    transparent_nb = 0;
    GetSubImage(alpha,block,mb_pos_x+B_SIZE,mb_pos_y);
    for (k=0;k<64;k++) if (!p[k]) transparent_nb++;
    if (transparent_nb == 64) type = MBM_TRANSPARENT;
    else if (transparent_nb == 0) type = MBM_OPAQUE;
    else type = MBM_BOUNDARY;   
    
    if(type==MBM_TRANSPARENT)
        AllZeroFillBlock(j*2  ,i*2+1  ,0,cur_vop);
    else if (type==MBM_BOUNDARY)
        TextureBlockPadding(j*2,i*2+1,padding_type,B_SIZE,alpha,
                            0,cur_vop);
    /* else MBM_OPAQUE, do nothing */
    
    transparent_nb = 0;
    GetSubImage(alpha,block,mb_pos_x,mb_pos_y+B_SIZE);
    for (k=0;k<64;k++) if (!p[k]) transparent_nb++;
    if (transparent_nb == 64) type = MBM_TRANSPARENT;
    else if (transparent_nb == 0) type = MBM_OPAQUE;
    else type = MBM_BOUNDARY;   
    
    if(type==MBM_TRANSPARENT)
        AllZeroFillBlock(j*2+1  ,i*2  ,0,cur_vop);
    else if (type==MBM_BOUNDARY)
        TextureBlockPadding(j*2+1,i*2,padding_type,B_SIZE,alpha,
                            0,cur_vop);
    /* else MBM_OPAQUE, do nothing */
    
    transparent_nb = 0;
    GetSubImage(alpha,block,mb_pos_x+B_SIZE,mb_pos_y+B_SIZE);
    for (k=0;k<64;k++) if (!p[k]) transparent_nb++;
    if (transparent_nb == 64) type = MBM_TRANSPARENT;
    else if (transparent_nb == 0) type = MBM_OPAQUE;
    else type = MBM_BOUNDARY;   
    
    if(type==MBM_TRANSPARENT)
        AllZeroFillBlock(j*2+1  ,i*2+1  ,0,cur_vop);
    else if (type==MBM_BOUNDARY)
        TextureBlockPadding(j*2+1,i*2+1,padding_type,B_SIZE,alpha,
                            0,cur_vop);
    /* else MBM_OPAQUE, do nothing */
    
    /* U-V blocks */
    transparent_nb = 0;
    GetSubImage(alpha_uv,block,mb_pos_x/2,mb_pos_y/2);
    for (k=0;k<64;k++) if (!p[k]) transparent_nb++;
    if (transparent_nb == 64) type = MBM_TRANSPARENT;
    else if (transparent_nb == 0) type = MBM_OPAQUE;
    else type = MBM_BOUNDARY;   
    
    if(type==MBM_TRANSPARENT)
        AllZeroFillBlock(j,i,1,cur_vop);
    else if(type==MBM_BOUNDARY)
        /* has alpha_sub*/
        TextureBlockPadding(j,i ,padding_type,B_SIZE, alpha_uv,1,cur_vop);
    /* else MBM_OPAQUE, do nothing */
    
    FreeImage(block);
    return;
}
#endif

/*ok*/static void
CallingConvention
LoadArea_16(    SInt *block,      /* <-- pointer to block to fill */
                SInt   *im,       /* <-- pointer to image         */
                Int    x,         /* <-- horizontal pos           */
                Int    y,         /* <-- vertical  pos            */
                Int    lx         /* <-- width of the image data  */
                )
{
    Int i, j;
    
    SInt   *in;
    
    in  = im + (y*lx) + x;
    
    for (i = 0; i < 16; i++)
    {
        for (j = 0; j < 4; j++)
        {
            *block++ = *in++;
            *block++ = *in++;
            *block++ = *in++;
            *block++ = *in++;
        }
        in += lx - 16;
    }
}

/*ok*/static void
CallingConvention
LoadArea_8(     SInt *block,      /* <-- pointer to block to fill */
                SInt   *im,       /* <-- pointer to image         */
                Int    x,         /* <-- horizontal pos           */
                Int    y,         /* <-- vertical  pos            */
                Int    lx         /* <-- width of the image data  */
                )
{
    Int i, j;
    
    SInt   *in;
    
    in  = im + (y*lx) + x;
    
    for (i = 0; i < 8; i++)
    {
        for (j = 0; j < 2; j++)
        {
            *block++ = *in++;
            *block++ = *in++;
            *block++ = *in++;
            *block++ = *in++;
        }
        in += lx - 8;
    }
}


/*ok*/static Void
CallingConvention
SetArea_16(
    SInt   *block,    /* <-- pointer to array         */  
    SInt   *im,       /* --> pointer to image         */  
    Int    x,         /* <-- horizontal pos in image  */  
    Int    y,         /* <-- vertical  pos in image   */  
    Int    lx        /* <-- width of the image data  */
    )
{
    SInt   *out;
    Int    i, j;
    
    out = im + (y*lx) + x;
    
    for (i = 0; i < 16; i++)
    {
        for (j = 0; j < 4; j++)
        {
            *out++ = *block++;
            *out++ = *block++;
            *out++ = *block++;
            *out++ = *block++;
        }
        out += lx - 16;
    }
}

/*ok*/static Void
CallingConvention
SetArea_8(
    SInt   *block,    /* <-- pointer to array         */  
    SInt   *im,       /* --> pointer to image         */  
    Int    x,         /* <-- horizontal pos in image  */  
    Int    y,         /* <-- vertical  pos in image   */  
    Int    lx        /* <-- width of the image data  */
    )
{
    SInt   *out;
    Int    i, j;
    
    out = im + (y*lx) + x;
    
    for (i = 0; i < 8; i++)
    {
        for (j = 0; j < 2; j++)
        {
            *out++ = *block++;
            *out++ = *block++;
            *out++ = *block++;
            *out++ = *block++;
        }
        out += lx - 8;
    }
}

/*ok*/static Void
CallingConvention
PrintOutMB( Char *s,
            SInt mb_mask[16][16],
            SInt mb_data[16][16],
            FILE *fout
            )  
{
    Int i, j;
    
    i=0;
    
    
    fprintf(fout,"%s:\n",s);
    
    
    
    for (i=0;i<16;i++)
    {
        for(j=0;j<16;j++)
        {
            /*if (mb_mask[i][j])*/
                fprintf(fout," %3d",mb_data[i][j]);
            /*else 
                fprintf(fout,"   -");*/
            if (j==7)
                fprintf(fout,"   ");
        }
        fprintf(fout,"\n");
        if(i==7)
            fprintf(fout, "\n");
    }
}

#if 0
/***********************************************************CommentBegin******
 *
 * -- Y_HorizontalPadding -- Horizontal padding of the Y data
 *
 * Author :
 *      UPM-GTI - Angel Pacheco
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Horizontal padding of the Y data
 *
 * Arguments in :
 *      SInt   *alpha_plane,    alpha plane data
 *      Int    w,               alpha width
 *      Int    h,               alpha height
 *
 * Arguments in/out :
 *      SInt   *status_plane,   status plane data
 *      SInt   *y_plane         Y plane data
 *
 * Arguments out :
 *
 *
 * Return values :
 *      none
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
CallingConvention
Y_HorizontalPadding (
    SInt   *alpha_plane,  /* <-- alpha plane data  */
    Int    w,             /* <-- alpha width       */
    Int    h,             /* <-- alpha height      */
    SInt   *status_plane, /* --> status plane data */
    SInt   *y_plane,      /* --> Y plane data      */
    int    arbitrary_shape
    )
{
    Int    i, j;
    Int    k=0;
    Int    status;
    Int    previous;
    SInt   value;
    
    if(arbitrary_shape)
    {
        i=0;
        while(i<h)
        {
            j=previous=0;
            
            status=(alpha_plane[INDEX(j,i,w,h)]!=0) ? INSIDE : FIRST_PAD;
            if(status==INSIDE)
                status_plane[INDEX(j,i,w,h)]=INSIDE;
            
            j++;
            while(j<w)
            {
                if(alpha_plane[INDEX(j,i,w,h)]!=0)
                {
                    if(status==FIRST_PAD)
                    {
                        value=y_plane[INDEX(j,i,w,h)];
                        for(k=0;k<j;k++)
                        {
                            y_plane[INDEX(k,i,w,h)]=value;
                            status_plane[INDEX(k,i,w,h)]=HOR_PAD;
                        }
                        status=INSIDE;
                        status_plane[INDEX(j,i,w,h)]=INSIDE;
                    }
                    else if (status==MIDDLE_PAD)
                    {
                        value=((Int)y_plane[INDEX(previous,i,w,h)]+
                               (Int)y_plane[INDEX(j,i,w,h)]) >> 1;
                        for(k=previous+1;k<j;k++)
                        {
                            y_plane[INDEX(k,i,w,h)] = value;
                            status_plane[INDEX(k,i,w,h)]=HOR_PAD;
                        }
                        status_plane[INDEX(j,i,w,h)]=INSIDE;
                        status=INSIDE;
                    }
                    else /* INSIDE */
                    {
                        status_plane[INDEX(j,i,w,h)]=INSIDE;
                    }
                }
                else /* alpha_plane[INDEX(j,i,w,h)]==0 */
                {
                    if(status==INSIDE)
                    {
                        status=MIDDLE_PAD;
                        previous=j-1;
                    }/* MIDDLE_PAD or FIRST_PAD do nothig */
                }
                j++;
            } /* while j */
            
            /* last part of the line */
            if(status==FIRST_PAD)
            {
                /* all line long */
                for(k=0;k<w;k++)
                {
                    y_plane[INDEX(k,i,w,h)]=0;
                    /* NO_PAD is the default for status_plane */
                }
            }
            else if(status==MIDDLE_PAD)
            {
                value=y_plane[INDEX(previous,i,w,h)];
                for(k=previous+1;k<w;k++)
                {
                    y_plane[INDEX(k,i,w,h)]=value;
                    status_plane[INDEX(k,i,w,h)]=HOR_PAD;
                }
            } /* if INSIDE, do nothing */
            i++;
        } /* while i */
    }
}

/***********************************************************CommentBegin******
 *
 * -- Y_VerticalOverlapedPadding -- Vertical and overlapped padding
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Vertical and overlapped (horizontal and vertical zones) padding of the
 *      Y data
 *
 * Arguments in :
 *      SInt   *alpha_plane,    original alpha plane
 *      Int    w,               original width  (Y or Alpha)
 *      Int    h,               original height (Y or Alpha)
 *
 * Arguments in/out :
 *      SInt   *status_plane,   status plane
 *      SInt   *y_plane         Y plane
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      It would be possible to compute here the padding of the null (zero)
 *      values in the vertical direction
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
CallingConvention
Y_VerticalOverlapedPadding (
    SInt   *alpha_plane,  /* <-- original alpha plane         */
    Int    w,             /* <-- original width  (Y or Alpha) */
    Int    h,             /* <-- original height (Y or Alpha) */
    SInt   *status_plane, /* --> status plane                 */
    SInt   *y_plane,      /* --> y plane                      */
    int    arbitrary_shape
    )
{
    Int    i, j;
    Int    k=0;
    Int    status;
    Int    previous;
    SInt   value;
    
    if(arbitrary_shape)
    {
        j=0;
        while(j<w)
        {
            i=previous=0;
            
            status=(alpha_plane[INDEX(j,i,w,h)]!=0) ? INSIDE:FIRST_PAD;
            /* status_plane[j,i] is updated if INSIDE */
            
            i++;
            while(i<h)
            {
                if(alpha_plane[INDEX(j,i,w,h)]!=0)
                {
                    if(status==FIRST_PAD)
                    {
                        value=y_plane[INDEX(j,i,w,h)];
                        for(k=0;k<i;k++)
                        {
                            if(status_plane[INDEX(j,k,w,h)]==HOR_PAD)
                            {
                                y_plane[INDEX(j,k,w,h)]=((Int)y_plane[INDEX(j,k,w,h)]+
                                                         (Int)value)>>1;
                            }
                            else
                            {
                                y_plane[INDEX(j,k,w,h)]=value;
                                status_plane[INDEX(j,k,w,h)]=VER_PAD;
                            }
                        }
                        status=INSIDE;
                    }
                    else if (status==MIDDLE_PAD)
                    {
                        value=((Int)y_plane[INDEX(j,previous,w,h)]+
                               (Int)y_plane[INDEX(j,i,w,h)]) >> 1;
                        for(k=previous+1;k<i;k++)
                        {
                            if(status_plane[INDEX(j,k,w,h)]==HOR_PAD)
                            {
                                y_plane[INDEX(j,k,w,h)]=((Int)y_plane[INDEX(j,k,w,h)]+
                                                         (Int)value)>>1;
                            }
                            else if (status_plane[INDEX(j,k,w,h)]==NO_PAD)
                            {
                                y_plane[INDEX(j,k,w,h)]=value;
                                status_plane[INDEX(j,k,w,h)]=VER_PAD;
                            } /*else is INSIDE, do nothing */
                        }
                        status=INSIDE;
                    } /* INSIDE do nothing */
                }
                else /* alpha_plane[INDEX(j,i,w,h)]==0 */
                {
                    if(status==INSIDE)
                    {
                        status=MIDDLE_PAD;
                        previous=i-1;
                    }/* MIDDLE_PAD or FIRST_PAD do nothig */
                }
                i++;
            } /* while i */
            
            /* last part of the column */
            if(status==MIDDLE_PAD)
            {
                value=y_plane[INDEX(j,previous,w,h)];
                for(k=previous+1;k<h;k++)
                {
                    if(status_plane[INDEX(j,k,w,h)]==HOR_PAD)
                    {
                        y_plane[INDEX(j,k,w,h)]=((Int)y_plane[INDEX(j,k,w,h)]+
                                                 (Int)value)>>1;
                    }
                    else if (status_plane[INDEX(j,k,w,h)]==NO_PAD)
                    {
                        y_plane[INDEX(j,k,w,h)]=value;
                        status_plane[INDEX(j,k,w,h)]=VER_PAD;
                    }/* else INSIDE no change in the y_value */
                }/*for*/
            } /* else if(status==FIRST_PAD) */
            /* here we could perform the vertical padding in this column, i.e.,
               the same process made inside the first while of the
               Y_FinalVerticalPadding() fuction. The process would be the filling
               of the empty regions corresponding to the zones padded in the
               last part of the padding process (only the vertical value, which
               will be used to padd the region with the mean between the nearest
               horizontal&vertical values */
            
            j++;
        } /* while j */
    }
}

/***********************************************************CommentBegin******
 *
 * -- Y_FinalHorizontalPadding -- Final horizontal zero padding
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 02.02.96
 *
 * Purpose :
 *      Final zero padding in the horizontal direction of the Y data
 *
 * Arguments in :
 *      Int    w,               original width  (Y or Alpha)
 *      Int    h,               original height (Y or Alpha)
 *
 * Arguments in/out :
 *      SInt   *status_plane,   status plane
 *      SInt   *y_plane         Y plane
 *
 * Arguments out :
 *
 *
 * Return values :
 *      none
 *
 * Side effects :
 *
 *
 * Description :
 *      It would be modified if VerticalOverlapedPadding() performs the
 *      vertical padding in the null (zero) values.
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
CallingConvention
Y_FinalHorizontalPadding (
    Int    w,             /* <-- original width  (Y or Alpha) */
    Int    h,             /* <-- original height (Y or Alpha) */
    SInt   *status_plane, /* --> status plane                 */
    SInt   *y_plane,      /* --> y plane                      */
    int    arbitrary_shape
    )
{
    Int    i, j;
    Int    k=0;
    Int    status;
    Int    previous;
    SInt   value;
    Int    middle;
    
    if(arbitrary_shape)
    {
        i=0;
        while(i<h)
        {
            j=previous=0;
            
            /* perhaps, the first part of the line */
            if((status_plane[INDEX(j,i,w,h)]==HOR_PAD)||
               (status_plane[INDEX(j,i,w,h)]==INSIDE))
            {
                i++;
            }
            else
            {
                if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
                {
                    /* first part of the line */
                    
                    j++;
                    while(j<(w-1))
                    {
                        if(status_plane[INDEX(j,i,w,h)]==VER_PAD)
                        {
                            value=y_plane[INDEX(j,i,w,h)];
                            for(k=0;k<j;k++)
                                y_plane[INDEX(k,i,w,h)]=value;
                            j++;
                            /*previous=j-1; really needed?*/
                            break;
                        }
                        else
                            j++;
                    }
                }
                
                /* the inner parts */
                
                status=VER_PAD;
                while(j<(w-1))
                {
                    if (status==VER_PAD)
                    {
                        j++;
                        if ((status=status_plane[INDEX(j,i,w,h)])==NO_PAD)
                        {
                            previous=j-1;
                        }
                    }
                    else /* NO_PAD */
                    {
                        j++;
                        if ((status=status_plane[INDEX(j,i,w,h)])==VER_PAD)
                        {
                            /*from the begin to the middle */
                            middle=(previous+j) >> 1;
                            value=y_plane[INDEX(previous,i,w,h)];
                            for(k=previous+1;k<middle;k++)
                                y_plane[INDEX(k,i,w,h)]=value;
                            /*from the middle to the end */
                            value=y_plane[INDEX(j,i,w,h)];
                            for(k=middle;k<j;k++)
                                y_plane[INDEX(k,i,w,h)]=value;
                        }
                    }
                }   /* while j */
                
                /* the last part of the line */
                if(status==VER_PAD)
                {
                    if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
                    {
                        y_plane[INDEX(j,i,w,h)]=y_plane[INDEX(j-1,i,w,h)];
                    }
                    else if(status_plane[INDEX(j,i,w,h)]==VER_PAD)
                    {
                        /* the case of an empty line with a pixel object at the
                           last column */
                        value=y_plane[INDEX(j,i,w,h)];
                        for(k=0;k<j;k++)
                            y_plane[INDEX(k,i,w,h)]=value;
                    }
                }
                else
                {
                    if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
                    {
                        value=y_plane[INDEX(previous,i,w,h)];
                        for(k=previous+1;k<w;k++)
                            y_plane[INDEX(k,i,w,h)]=value;
                    }
                    else
                    {
                        /*from the begin to the middle */
                        middle=(previous+j) >> 1;
                        value=y_plane[INDEX(previous,i,w,h)];
                        for(k=previous+1;k<middle;k++)
                            y_plane[INDEX(k,i,w,h)]=value;
                        /*from the middle to the end */
                        value=y_plane[INDEX(j,i,w,h)];
                        for(k=middle;k<w;k++)
                            y_plane[INDEX(k,i,w,h)]=value;
                    }
                } /* else VER_PAD */
                
                i++;
            }
        } /* while i */
    }
}

/***********************************************************CommentBegin******
 *
 * -- Y_FinalVerticalPadding -- Final zero padding in the vertical direction
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Final zero padding in the vertical direction of the Y data
 *
 * Arguments in :
 *      Int    w,               original width  (Y or Alpha)
 *      Int    h,               original height (Y or Alpha)
 *
 * Arguments in/out :
 *      SInt   *status_plane,   status plane
 *      SInt   *y_plane         Y plane
 *
 * Arguments out :
 *
 *
 * Return values :
 *      none
 *
 * Side effects :
 *
 *
 * Description :
 *      It would disapear if VerticalOverlapedPadding() performs the vertical
 *      zero padding (see the comment in the last part of the that function
 *      code)
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
CallingConvention
Y_FinalVerticalPadding (
    Int    w,             /* <-- original width  (Y or Alpha) */
    Int    h,             /* <-- original height (Y or Alpha) */
    SInt   *status_plane, /* --> status plane                 */
    SInt   *y_plane,      /* --> y plane                      */
    int    arbitrary_shape
    
    )
{
    Int    i, j;
    Int    k=0;
    Int    status;
    Int    previous;
    SInt   value;
    Int    middle;
    
    if(arbitrary_shape)
    {
        j=0;
        while(j<w)
        {
            i=previous=0;
            
            /* perhaps, the first part of the column */
            if((status_plane[INDEX(j,i,w,h)]==VER_PAD)||
               (status_plane[INDEX(j,i,w,h)]==INSIDE))
            {
                j++;
            }
            else
            {
                if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
                {
                    /* first part of the column */
                    
                    i++;
                    while(i<(h-1))
                    {
                        if(status_plane[INDEX(j,i,w,h)]==HOR_PAD)
                        {
                            value=y_plane[INDEX(j,i,w,h)];
                            for(k=0;k<i;k++)
                                y_plane[INDEX(j,k,w,h)]=(y_plane[INDEX(j,k,w,h)]+value)>>1;
                            i++;
                            /* previous=i-1; really needed?*/
                            break;
                        }
                        else
                            i++;
                    }
                }
                
                /* the inner parts */
                status=HOR_PAD;
                while(i<(h-1))
                {
                    i++;
                    if (status==HOR_PAD)
                    {
                        if ((status=status_plane[INDEX(j,i,w,h)])==NO_PAD)
                        {
                            previous=i-1;
                        }
                    }
                    else if(status==NO_PAD) /* NO_PAD */
                    {
                        if ((status=status_plane[INDEX(j,i,w,h)])==HOR_PAD)
                        {
                            /*from the begin to the middle */
                            middle=(previous+i) >> 1;
                            value=y_plane[INDEX(j,previous,w,h)];
                            for(k=previous+1;k<middle;k++)
                                y_plane[INDEX(j,k,w,h)]=(value+y_plane[INDEX(j,k,w,h)])>>1;
                            /*from the middle to the end */
                            value=y_plane[INDEX(j,i,w,h)];
                            for(k=middle;k<i;k++)
                                y_plane[INDEX(j,k,w,h)]=(value+y_plane[INDEX(j,k,w,h)])>>1;
                        }
                    }/* else VER_PAD */
                } /* while i */
                
                /* the last part of the column */
                if(status==HOR_PAD)
                {
                    if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
                        y_plane[INDEX(j,i,w,h)]=y_plane[INDEX(j,i-1,w,h)];
                }
                else
                {
                    if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
                    {
                        value=y_plane[INDEX(j,previous,w,h)];
                        for(k=previous+1;k<h;k++)
                            y_plane[INDEX(j,k,w,h)]=(value+y_plane[INDEX(j,k,w,h)])>>1;
                    }
                } /* else HOR_PAD */
                
                j++;
            } /* else skip the column */
            
        } /* while j */
    }
}

/***********************************************************CommentBegin******
 *
 * -- UV_HorizontalPadding -- Horizontal padding of the Y data
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Horizontal padding of the Y data
 *
 * Arguments in :
 *      SInt   *alpha_plane,    subsampled alpha plane
 *      Int    w,               subsampled width  (U or V)
 *      Int    h,               subsampled height (U or V)
 *
 * Arguments in/out :
 *      SInt   *status_plane,   subsampled status plane
 *      SInt   *u_plane,        U plane
 *      SInt   *v_plane         V plane
 *
 * Arguments out :
 *
 *
 * Return values :
 *      none
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
CallingConvention
UV_HorizontalPadding (
    SInt   *alpha_plane,  /* <-- subsampled alpha plane     */
    Int    w,             /* <-- subsampled width  (U or V) */
    Int    h,             /* <-- subsampled height (U or V) */
    SInt   *status_plane, /* --> subsampled status plane    */
    SInt   *u_plane,      /* --> u plane                    */
    SInt   *v_plane       /* --> v plane                    */
    )
{
    Int    i, j;
    Int    k=0;
    Int    status;
    Int    previous;
    SInt   value1,value2;
    
    i=0;
    while(i<h)
    {
        j=previous=0;
        
        status=(alpha_plane[INDEX(j,i,w,h)]!=0) ? INSIDE:FIRST_PAD;
        if(status==INSIDE)
            status_plane[INDEX(j,i,w,h)]=INSIDE;
        
        j++;
        while(j<w)
        {
            if(alpha_plane[INDEX(j,i,w,h)]!=0)
            {
                if(status==FIRST_PAD)
                {
                    value1=u_plane[INDEX(j,i,w,h)];
                    value2=v_plane[INDEX(j,i,w,h)];
                    for(k=0;k<j;k++)
                    {
                        u_plane[INDEX(k,i,w,h)]=value1;
                        v_plane[INDEX(k,i,w,h)]=value2;
                        status_plane[INDEX(k,i,w,h)]=HOR_PAD;
                    }
                    status=INSIDE;
                    status_plane[INDEX(j,i,w,h)]=INSIDE;
                }
                else if (status==MIDDLE_PAD)
                {
                    value1=((Int)u_plane[INDEX(previous,i,w,h)]+
                            (Int)u_plane[INDEX(j,i,w,h)]) >> 1;
                    value2=((Int)v_plane[INDEX(previous,i,w,h)]+
                            (Int)v_plane[INDEX(j,i,w,h)]) >> 1;
                    for(k=previous+1;k<j;k++)
                    {
                        u_plane[INDEX(k,i,w,h)] = value1;
                        v_plane[INDEX(k,i,w,h)] = value2;
                        status_plane[INDEX(k,i,w,h)]=HOR_PAD;
                    }
                    status_plane[INDEX(j,i,w,h)]=INSIDE;
                    status=INSIDE;
                }
                else /* INSIDE */
                {
                    status_plane[INDEX(j,i,w,h)]=INSIDE;
                }
            }
            else /* alpha_plane[INDEX(j,i,w,h)]==0 */
            {
                if(status==INSIDE)
                {
                    status=MIDDLE_PAD;
                    previous=j-1;
                }/* MIDDLE_PAD or FIRST_PAD do nothig */
            }
            j++;
        } /* while j */
        
        /* last part of the line */
        if(status==FIRST_PAD)
        {
            /* all line long */
            for(k=0;k<w;k++)
            {
                u_plane[INDEX(k,i,w,h)]=0;
                v_plane[INDEX(k,i,w,h)]=0;
                /* NO_PAD is the default for status_plane */
            }
        }
        else if(status==MIDDLE_PAD)
        {
            value1=u_plane[INDEX(previous,i,w,h)];
            value2=v_plane[INDEX(previous,i,w,h)];
            for(k=previous+1;k<w;k++)
            {
                u_plane[INDEX(k,i,w,h)]=value1;
                v_plane[INDEX(k,i,w,h)]=value2;
                status_plane[INDEX(k,i,w,h)]=HOR_PAD;
            }
        } /* if INSIDE, do nothing */
        i++;
    } /* while i */
}

/***********************************************************CommentBegin******
 *
 * -- UV_VerticalOverlapedPadding -- Vertical and overlapped padding
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Vertical and overlapped (horizontal and vertical zones) padding of the
 *      Y data
 *
 * Arguments in :
 *      SInt   *alpha_plane,    subsampled alpha plane
 *      Int    w,               subsampled width  (U or V)
 *      Int    h,               subsampled height (U or V)
 *
 * Arguments in/out :
 *      SInt   *status_plane,   subsampled status plane
 *      SInt   *u_plane,        U plane
 *      SInt   *v_plane         V plane
 *
 * Arguments out :
 *
 *
 * Return values :
 *      none
 *
 * Side effects :
 *
 *
 * Description :
 *    It would be possible to compute here the zero padding in the vertical
 *    direction
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
CallingConvention
UV_VerticalOverlapedPadding (
    SInt   *alpha_plane,  /* <-- subsampled alpha plane     */
    Int    w,             /* <-- subsampled width  (U or V) */
    Int    h,             /* <-- subsampled height (U or V) */
    SInt   *status_plane, /* --> subsampled status plane    */
    SInt   *u_plane,      /* --> u plane                    */
    SInt   *v_plane       /* --> v plane                    */
    )
{
    Int    i, j;
    Int    k=0;
    Int    status;
    Int    previous;
    SInt   value1,value2;
    
    j=0;
    while(j<w)
    {
        i=previous=0;
        
        status=(alpha_plane[INDEX(j,i,w,h)]!=0) ? INSIDE:FIRST_PAD;
        /* status_plane[j,i] is updated if INSIDE */
        
        i++;
        while(i<h)
        {
            if(alpha_plane[INDEX(j,i,w,h)]!=0)
            {
                if(status==FIRST_PAD)
                {
                    value1=u_plane[INDEX(j,i,w,h)];
                    value2=v_plane[INDEX(j,i,w,h)];
                    
                    for(k=0;k<i;k++)
                    {
                        if(status_plane[INDEX(j,k,w,h)]==HOR_PAD)
                        {
                            u_plane[INDEX(j,k,w,h)]=((Int)u_plane[INDEX(j,k,w,h)]+
                                                     (Int)value1)>>1;
                            v_plane[INDEX(j,k,w,h)]=((Int)v_plane[INDEX(j,k,w,h)]+
                                                     (Int)value2)>>1;
                        }
                        else
                        {
                            u_plane[INDEX(j,k,w,h)]=value1;
                            v_plane[INDEX(j,k,w,h)]=value2;
                            status_plane[INDEX(j,k,w,h)]=VER_PAD;
                        }
                    }
                    status=INSIDE;
                }
                else if (status==MIDDLE_PAD)
                {
                    value1=((Int)u_plane[INDEX(j,previous,w,h)]+
                            (Int)u_plane[INDEX(j,i,w,h)]) >> 1;
                    value2=((Int)v_plane[INDEX(j,previous,w,h)]+
                            (Int)v_plane[INDEX(j,i,w,h)]) >> 1;
                    for(k=previous+1;k<i;k++)
                    {
                        if(status_plane[INDEX(j,k,w,h)]==HOR_PAD)
                        {
                            u_plane[INDEX(j,k,w,h)]=((Int)u_plane[INDEX(j,k,w,h)]+
                                                     (Int)value1)>>1;
                            v_plane[INDEX(j,k,w,h)]=((Int)v_plane[INDEX(j,k,w,h)]+
                                                     (Int)value2)>>1;
                        }
                        else
                        {
                            u_plane[INDEX(j,k,w,h)]=value1;
                            v_plane[INDEX(j,k,w,h)]=value2;
                            status_plane[INDEX(j,k,w,h)]=VER_PAD;
                        }
                    }
                    status=INSIDE;
                } /* INSIDE do nothing */
                
            }
            else /* alpha_plane[INDEX(j,i,w,h)]==0 */
            {
                if(status==INSIDE)
                {
                    status=MIDDLE_PAD;
                    previous=i-1;
                }/* MIDDLE_PAD or FIRST_PAD do nothig */
            }
            i++;
        } /* while i */
        
        /* last part of the column */
        if(status==MIDDLE_PAD)
        {
            value1=u_plane[INDEX(j,previous,w,h)];
            value2=v_plane[INDEX(j,previous,w,h)];
            for(k=previous+1;k<h;k++)
            {
                if(status_plane[INDEX(j,k,w,h)]==HOR_PAD)
                {
                    u_plane[INDEX(j,k,w,h)]=((Int)u_plane[INDEX(j,k,w,h)]
                                             +(Int)value1)>>1;
                    v_plane[INDEX(j,k,w,h)]=((Int)v_plane[INDEX(j,k,w,h)]
                                             +(Int)value2)>>1;
                }
                else if (status_plane[INDEX(j,k,w,h)]==NO_PAD)
                {
                    u_plane[INDEX(j,k,w,h)]=value1;
                    v_plane[INDEX(j,k,w,h)]=value2;
                    status_plane[INDEX(j,k,w,h)]=VER_PAD;
                }/* else INSIDE no change in the y_value */
            }/*for*/
        } /*else if(status==FIRST_PAD) */
        /* here we could perform the vertical padding in this column, i.e.,
           the same process made inside the first while of the
           UV_FinalVerticalPadding() fuction. The process would be the filling
           of the empty regions corresponding to the zones padded in the
           last part of the padding process (only the vertical value, which
           will be used to padd the region with the mean between the nearest
           horizontal&vertical values */
        
        j++;
    } /* while j */
}

/***********************************************************CommentBegin******
 *
 * -- UV_FinalHorizontalPadding -- Final zero padding in horizontal direction
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Final zero padding in the horizontal direction of the Y data
 *
 * Arguments in :
 *      Int    w,               subsampled width  (U or V)
 *      Int    h,               subsampled height (U or V)
 *
 * Arguments in/out :
 *      SInt   *status_plane,   subsampled status plane
 *      SInt   *u_plane,        U plane
 *      SInt   *v_plane         V plane
 *
 * Arguments out :
 *
 *
 * Return values :
 *      none
 *
 * Side effects :
 *
 *
 * Description :
 *      It would be modified if VerticalOverlapedPadding() performs the
 *      vertical zero padding
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
CallingConvention
UV_FinalHorizontalPadding (
    Int    w,             /* <-- subsampled width  (U or V) */
    Int    h,             /* <-- subsampled height (U or V) */
    SInt   *status_plane, /* --> subsampled status plane    */
    SInt   *u_plane,      /* --> u plane                    */
    SInt   *v_plane       /* --> v plane                    */
    )
{
    Int    i, j;
    Int    k=0;
    Int    status;
    Int    previous;
    SInt   value1,value2;
    Int    middle;
    
    i=0;
    while(i<h)
    {
        j=previous=0;
        
        /* perhaps, the first part of the line */
        if((status_plane[INDEX(j,i,w,h)]==HOR_PAD)||
           (status_plane[INDEX(j,i,w,h)]==INSIDE))
        {
            i++;
        }
        else
        {
            if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
            {
                /* first part of the line */
                
                j++;
                while(j<(w-1))
                {
                    if(status_plane[INDEX(j,i,w,h)]==VER_PAD)
                    {
                        value1=u_plane[INDEX(j,i,w,h)];
                        value2=v_plane[INDEX(j,i,w,h)];
                        for(k=0;k<j;k++)
                        {
                            u_plane[INDEX(k,i,w,h)]=value1;
                            v_plane[INDEX(k,i,w,h)]=value2;
                        }
                        j++;
                        /*previous=j-1; really needed?*/
                        break;
                    }
                    else
                        j++;
                }
            }
            
            /* the inner parts */
            
            status=VER_PAD;
            while(j<(w-1))
            {
                if (status==VER_PAD)
                {
                    j++;
                    if ((status=status_plane[INDEX(j,i,w,h)])==NO_PAD)
                    {
                        previous=j-1;
                    }
                }
                else /* NO_PAD */
                {
                    j++;
                    if ((status=status_plane[INDEX(j,i,w,h)])==VER_PAD)
                    {
                        /*from the begin to the middle */
                        middle=(previous+j) >> 1;
                        value1=u_plane[INDEX(previous,i,w,h)];
                        value2=v_plane[INDEX(previous,i,w,h)];
                        for(k=previous+1;k<middle;k++)
                        {
                            u_plane[INDEX(k,i,w,h)]=value1;
                            v_plane[INDEX(k,i,w,h)]=value2;
                        }
                        /*from the middle to the end */
                        value1=u_plane[INDEX(j,i,w,h)];
                        value2=v_plane[INDEX(j,i,w,h)];
                        for(k=middle;k<j;k++)
                        {
                            u_plane[INDEX(k,i,w,h)]=value1;
                            v_plane[INDEX(k,i,w,h)]=value2;
                        }
                    }
                }
            }  /* while j */
            
            /* the last part of the line */
            if(status==VER_PAD)
            {
                if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
                {
                    u_plane[INDEX(j,i,w,h)]=u_plane[INDEX(j-1,i,w,h)];
                    v_plane[INDEX(j,i,w,h)]=v_plane[INDEX(j-1,i,w,h)];
                }
                else if(status_plane[INDEX(j,i,w,h)]==VER_PAD)
                {
                    /* the case of an empty line with a pixel object at the
                       last column */
                    value1=u_plane[INDEX(j,i,w,h)];
                    value2=v_plane[INDEX(j,i,w,h)];
                    for(k=0;k<j;k++)
                    {
                        u_plane[INDEX(k,i,w,h)]=value1;
                        v_plane[INDEX(k,i,w,h)]=value2;
                    }
                }
                
            }
            else
            {
                if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
                {
                    value1=u_plane[INDEX(previous,i,w,h)];
                    value2=v_plane[INDEX(previous,i,w,h)];
                    for(k=previous+1;k<w;k++)
                    {
                        u_plane[INDEX(k,i,w,h)]=value1;
                        v_plane[INDEX(k,i,w,h)]=value2;
                    }
                }
                else
                {
                    /*from the begin to the middle */
                    middle=(previous+j) >> 1;
                    value1=u_plane[INDEX(previous,i,w,h)];
                    value2=v_plane[INDEX(previous,i,w,h)];
                    for(k=previous+1;k<middle;k++)
                    {
                        u_plane[INDEX(k,i,w,h)]=value1;
                        v_plane[INDEX(k,i,w,h)]=value2;
                    }
                    /*from the middle to the end */
                    value1=u_plane[INDEX(j,i,w,h)];
                    value2=v_plane[INDEX(j,i,w,h)];
                    for(k=middle;k<w;k++)
                    {
                        u_plane[INDEX(k,i,w,h)]=value1;
                        v_plane[INDEX(k,i,w,h)]=value2;
                    }
                }
            } /* else VER_PAD */
            
            i++;
        }
    } /* while i */
}

/***********************************************************CommentBegin******
 *
 * -- UV_FinalVerticalPadding -- Final zero padding in the vertical direction
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Final zero padding in the vertical direction of the Y data
 *
 * Arguments in :
 *      Int    w,               subsampled width  (U or V)
 *      Int    h,               subsampled height (U or V)
 *
 * Arguments in/out :
 *      SInt   *status_plane,   subsampled status plane
 *      SInt   *u_plane,        U plane
 *      SInt   *v_plane         V plane
 *
 * Arguments out :
 *
 *
 * Return values :
 *      none
 *
 * Side effects :
 *
 *
 * Description :
 *      It would disapear if VerticalOverlapedPadding() performs the vertical
 *      zero padding
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Void
CallingConvention
UV_FinalVerticalPadding (
    Int    w,             /* <-- subsampled width  (U or V) */
    Int    h,             /* <-- subsampled height (U or V) */
    SInt   *status_plane, /* --> subsampled status plane    */
    SInt   *u_plane,      /* --> u plane                    */
    SInt   *v_plane       /* --> v plane                    */
    )
{
    Int    i, j;
    Int    k=0;
    Int    status;
    Int    previous;
    SInt   value1,value2;
    Int    middle;
    
    j=0;
    while(j<w)
    {
        i=previous=0;
        
        /* perhaps, the first part of the column */
        if((status_plane[INDEX(j,i,w,h)]==VER_PAD)||
           (status_plane[INDEX(j,i,w,h)]==INSIDE))
        {
            j++;
        }
        else
        {
            if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
            {
                /* first part of the column */
                
                i++;
                while(i<(h-1))
                {
                    if(status_plane[INDEX(j,i,w,h)]==HOR_PAD)
                    {
                        value1=u_plane[INDEX(j,i,w,h)];
                        value2=v_plane[INDEX(j,i,w,h)];
                        for(k=0;k<i;k++)
                        {
                            u_plane[INDEX(j,k,w,h)]=(u_plane[INDEX(j,k,w,h)]+value1)>>1;
                            v_plane[INDEX(j,k,w,h)]=(v_plane[INDEX(j,k,w,h)]+value2)>>1;
                        }
                        /* previous=i-1; really needed?*/
                        i++;
                        break;
                    }
                    else
                        i++;
                }
            }
            
            /* the inner parts */
            status=HOR_PAD;
            while(i<(h-1))
            {
                i++;
                if (status==HOR_PAD)
                {
                    if ((status=status_plane[INDEX(j,i,w,h)])==NO_PAD)
                    {
                        previous=i-1;
                    }
                }
                else if(status==NO_PAD) /* NO_PAD */
                {
                    if ((status=status_plane[INDEX(j,i,w,h)])==HOR_PAD)
                    {
                        /*from the begin to the middle */
                        middle=(previous+i) >> 1;
                        value1=u_plane[INDEX(j,previous,w,h)];
                        value2=v_plane[INDEX(j,previous,w,h)];
                        for(k=previous+1;k<middle;k++)
                        {
                            u_plane[INDEX(j,k,w,h)]=(value1+u_plane[INDEX(j,k,w,h)])>>1;
                            v_plane[INDEX(j,k,w,h)]=(value2+v_plane[INDEX(j,k,w,h)])>>1;
                        }
                        /*from the middle to the end */
                        value1=u_plane[INDEX(j,i,w,h)];
                        value2=v_plane[INDEX(j,i,w,h)];
                        for(k=middle;k<i;k++)
                        {
                            u_plane[INDEX(j,k,w,h)]=(value1+u_plane[INDEX(j,k,w,h)])>>1;
                            v_plane[INDEX(j,k,w,h)]=(value2+v_plane[INDEX(j,k,w,h)])>>1;
                        }
                    }
                }
            } /* while i */
            
            /* the last part of the column */
            if(status==HOR_PAD)
            {
                if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
                {
                    u_plane[INDEX(j,i,w,h)]=u_plane[INDEX(j,i-1,w,h)];
                    v_plane[INDEX(j,i,w,h)]=v_plane[INDEX(j,i-1,w,h)];
                }
            }
            else
            {
                if(status_plane[INDEX(j,i,w,h)]==NO_PAD)
                {
                    value1=u_plane[INDEX(j,previous,w,h)];
                    value2=v_plane[INDEX(j,previous,w,h)];
                    for(k=previous+1;k<h;k++)
                    {
                        u_plane[INDEX(j,k,w,h)]=(value1+u_plane[INDEX(j,k,w,h)])>>1;
                        v_plane[INDEX(j,k,w,h)]=(value2+v_plane[INDEX(j,k,w,h)])>>1;
                    }
                    
                }
            } /* else HOR_PAD */
            
            j++;
        } /* else skip the column */
        
    }/* while j */
}
#endif

/********************************************************** *CommentBegin******
 *
 * -- RepetitivePadding -- Repetitive padding of the YUV images
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Repetitive padding of the YUV images of vop_ima depending on lumi field
 *
 * Arguments in :
 *      Int   lumi,      1=only padd Y image, !=1 padd YUV images
 *
 * Arguments in/out :
 *      Vop   *vop_ima   Vop to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *      1 on success, <0 on error
 *
 * Side effects :
 *
 *
 * Description :
 *      The downsampling of the alpha plane is done, if necesary, acording
 *      VM 2.x (valid alpha values are only the odd columns/files ones)
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Int
CallingConvention
RepetitivePadding (
    Int   lumi,    /* <-- 1=only padd Y image, !=1 padd YUV images  */
    Vop   *vop_ima /* --> Vop to pad                                */
    )
{
#ifndef _OPTIMIZE_
    SInt    *alpha_plane;
    SInt    *y_plane;
    SInt    *u_plane;
    SInt    *v_plane;
    SInt    *status_plane;
    SInt    *alp_plane;
    Int     i, j, base;
    Image   *alpha_ima;
    Image   *y_ima;
    Image   *u_ima;
    Image   *v_ima;
    
    Int     y_x, y_y;
    Int     u_x, u_y;
    Int     v_x, v_y;
    
    Int     h;
    Int     w;
#endif
    
    Int     arbitrary_shape;
    
    /* checkings */
    
#ifdef _NO_PADDING_
    return(2);
#endif
    
    arbitrary_shape = GetVopShape(vop_ima);
    
    if(!arbitrary_shape)
        return(2);
    
    
#ifndef _OPTIMIZE_
    if( (vop_ima==NULL) || ((alpha_ima=vop_ima->a_chan)==NULL) ||
        ((y_ima=vop_ima->y_chan)==NULL) || ((u_ima=vop_ima->u_chan)==NULL) ||
        ((v_ima=vop_ima->v_chan)==NULL) ||
        ((alpha_plane=(SInt *)GetImageData(alpha_ima))==NULL) ||
        ((y_plane=(SInt *)GetImageData(y_ima))==NULL) ||
        ((u_plane=(SInt *)GetImageData(u_ima))==NULL)||
        ((v_plane=(SInt *)GetImageData(v_ima))==NULL) )
    {
        fprintf(stderr,"RepetitivePadding: NULL input pointers\n");
        return(0);
    }
    
    if ( (alpha_ima->grid=='h') || (y_ima->grid=='h') ||
         (u_ima->grid=='h')||(v_ima->grid=='h') )
    {
        fprintf(stderr,"RepetitivePadding: hexadecimal grid not supported\n");
        return(-1);
    }
    
    w=y_x=(Int)GetImageSizeX(y_ima); h=y_y=(Int)GetImageSizeY(y_ima);
    
    if ( (y_x!=GetImageSizeX(alpha_ima)) || (y_y!=GetImageSizeY(alpha_ima)) )
    {
        fprintf(stderr,"RepetitivePadding: different image dimensions \n");
        return(0);
    }
    
    
    if(lumi!=1)
    { /* do YUV padding */
        u_x=(Int)GetImageSizeX(u_ima); u_y=(Int)GetImageSizeY(u_ima);
        v_x=(Int)GetImageSizeX(v_ima); v_y=(Int)GetImageSizeY(v_ima);
        
        if ( (u_x!=v_x) || (u_y!=v_y) || ((y_y/2)!=u_y) ||
             (!((y_x/2==u_x)||(y_x==u_x)))  )
        {
            fprintf(stderr,"RepetitivePadding: different image dimensions \n");
            return(0);
        }
    }
    
    if ((status_plane= (SInt*)malloc(h*w*sizeof(SInt)))==NULL)
    {
        fprintf(stderr,"RepetitivePadding: malloc failed\n");
        return(-3);
    }
    
    for(j=0;j<h;j++)
    {
        base=j*w;
        for(i=0;i<w;i++)
            status_plane[base+i]= NO_PAD;
    }
    
    arbitrary_shape = GetVopShape(vop_ima);
    
    /* do Y padding */
    
    Y_HorizontalPadding (alpha_plane, w, h, status_plane, y_plane, arbitrary_shape);
    
    Y_VerticalOverlapedPadding (alpha_plane, w, h, status_plane, y_plane, arbitrary_shape);
    
    Y_FinalHorizontalPadding (w, h, status_plane, y_plane, arbitrary_shape);
    
    Y_FinalVerticalPadding (w, h, status_plane, y_plane, arbitrary_shape);
    
    if (lumi!=1) 
    { 
        free ((Char*)status_plane);
        
        /* do UV padding */
        
        h=h/2;
        w=w/2;
        if ((status_plane= (SInt*)malloc(h*w*sizeof(SInt)))==NULL)
        {
            fprintf(stderr,"RepetitivePadding: malloc failed\n");
            return(-3);
        }
        
        for(j=0;j<h;j++)
        {
            base=j*w;
            for(i=0;i<w;i++)
                status_plane[base+i]= NO_PAD;
        }
        alp_plane = (SInt *)GetImageData(vop_ima->a_uv_chan); 
        
        if(arbitrary_shape)
        {
            UV_HorizontalPadding (alp_plane, w, h, status_plane, u_plane,v_plane);
            
            UV_VerticalOverlapedPadding (alp_plane, w, h, status_plane,u_plane,v_plane);
            
            UV_FinalHorizontalPadding (w, h, status_plane, u_plane,v_plane);
            
            UV_FinalVerticalPadding (w, h, status_plane, u_plane,v_plane);
        }
        
        free ((Char*)status_plane);
        
        return(1);
    }
    else
    {
        free ((Char*)status_plane);
        
        return(-4);
    }
#endif
    
}

/***********************************************************CommentBegin******
 *
 * -- RepetitiveImagePadding --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      02.02.96
 *
 * Purpose :
 *      Repetitive padding of the SInt data of ima
 *
 * Arguments in :
 *      Image   *alpha_ima,   alpha plane to check for shape
 *
 * Arguments in/out :
 *      Image   *ima          image to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *      1 on success, <0 on error
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/
#if 0
Int
RepetitiveImagePadding (
    Image   *alpha_ima, /* <-- alpha plane to check for shape */
    Image   *ima        /* --> image to pad                   */
    )
{
    SInt    *alpha_plane;
    SInt    *y_plane;
    SInt    *status_plane;
    Int     i, j, base;
    Int     y_x,y_y;
    Int     h;
    Int     w;
    
#ifdef _NO_PADDING_
    return 2;
#endif
    
    /* checkings */
    if( (ima==NULL) || (alpha_ima==NULL) ||
        ((alpha_plane=(SInt *)GetImageData(alpha_ima))==NULL) ||
        ((y_plane=(SInt *)GetImageData(ima))==NULL) )
    {
        fprintf(stderr,"RepetitivePadding: NULL input pointers\n");
        return(0);
    }
    
    if ( (alpha_ima->grid=='h') || (ima->grid=='h') )
    {
        fprintf(stderr,"RepetitivePadding: hexadecimal grid not supported\n");
        return(-1);
    }
    
    w=y_x=(Int)GetImageSizeX(ima); h=y_y=(Int)GetImageSizeY(ima);
    
    if ( (y_x!=GetImageSizeX(alpha_ima)) || (y_y!=GetImageSizeY(alpha_ima)) )
    {
        fprintf(stderr,"RepetitivePadding: different image dimensions \n");
        return(0);
    }
    
    if ((status_plane= (SInt*)malloc(h*w*sizeof(SInt)))==NULL)
    {
        fprintf(stderr,"RepetitivePadding: malloc failed\n");
        return(-3);
    }
    
    for(j=0;j<h;j++)
    {
        base=j*w;
        for(i=0;i<w;i++)
            status_plane[base+i]= NO_PAD;
    }
    
    /* do Y padding */
    
    Y_HorizontalPadding (alpha_plane, w, h, status_plane, y_plane);
    
    Y_VerticalOverlapedPadding (alpha_plane, w, h, status_plane, y_plane);
    
    Y_FinalHorizontalPadding (w, h, status_plane, y_plane);
    
    Y_FinalVerticalPadding (w, h, status_plane, y_plane);
    
    free ((Char*)status_plane);
    
    return (1);
}
/***********************************************************CommentBegin******
 *
 * -- CheckParameters --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      30.08.96
 *
 * Purpose :
 *      Checks the Input Parameters the TexturePadding fuction.
 *
 * Arguments in :
 *      Image   *prev_rec_alpha,   previous reconstructed alpha plane data
 *      Image   *alpha_decisions,  the previous reconstructed alpha plane
 *                                 subsampled (in block units) which values
 *                                 are:
 *                                  MBM_TRANSPARENT = all pixels are transparent
 *                                        (outside the shape)
 *                                  MBM_OPAQUE      = all pixels are not transp.
 *                                        (inside the shape)
 *                                  MBM_BOUNDARY    = transparent and not
 *                                        transparent pixels at the same time
 *                                        (in the boundary of the shape)
 *      Image   *MB_decisions,     encoding mode for each MB. If this Image is
 *                                  NULL      => Frame based repetitive padd.
 *                                  otherwise => Block based repetitive and
 *                                               zero padding.
 *
 * Arguments in/out :
 *      Vop     *cur_vop           Vop to pad
 *
 * Arguments out :
 *
 *
 * Return values :
 *      1 on success, <0 on error
 *
 * Side effects :
 *
 *
 * Description :
 *
 *
 * See also :
 *
 *
 * Modified :
 *
 *
 ***********************************************************CommentEnd********/

Int
CallingConvention
CheckParameters (
    Image   *prev_rec_alpha,  /* <-- previous reconstructed alpha plane data */
    Image   *alpha_decisions, /* <-- subsampled (per block) prev_rec_alpha   */
    Image   *MB_decisions,    /* <-- encoding mode for each MB               */
    Vop     *cur_vop          /* <-> Vop to pad                              */
    )
{
    
    Image *y_ima,*u_ima,*v_ima;
    Int   w,h,y_x,y_y,u_x,v_x,u_y,v_y;
    
    if( (cur_vop==NULL) || 
        ((y_ima=cur_vop->y_chan)==NULL) ||((u_ima=cur_vop->u_chan)==NULL) ||
        ((v_ima=cur_vop->v_chan)==NULL) ||
        (GetImageData(prev_rec_alpha)==NULL) ||
        (GetImageData(y_ima)==NULL) || 
        (GetImageData(u_ima)==NULL) ||
        (GetImageData(v_ima)==NULL) ||
        (alpha_decisions==NULL) || (MB_decisions==NULL)  || 
        (GetImageData(alpha_decisions)==NULL) || 
        (GetImageData(MB_decisions)==NULL) )
    {
        fprintf(stderr,"CheckParameters: NULL input pointers\n");
        return(0);
    }
    
    if ( (prev_rec_alpha->grid=='h') || (y_ima->grid=='h') ||
         (u_ima->grid=='h')||(v_ima->grid=='h') )
    {
        fprintf(stderr,"CheckParameters: hexadecimal grid not supported\n");
        return(-1);
    }
    
    w=y_x=(Int)GetImageSizeX(y_ima); h=y_y=(Int)GetImageSizeY(y_ima);
    
    if ( (y_x!=GetImageSizeX(y_ima)) || (y_y!=GetImageSizeY(y_ima)) )
    {
        fprintf(stderr,"CheckParameters: different image dimensions \n");
        return(-1);
    }
    
    u_x=(Int)GetImageSizeX(u_ima); u_y=(Int)GetImageSizeY(u_ima);
    v_x=(Int)GetImageSizeX(v_ima); v_y=(Int)GetImageSizeY(v_ima);
    
    if ( (u_x!=v_x) || (u_y!=v_y) || ((y_y/2)!=u_y) ||
         (!((y_x/2==u_x)||(y_x==u_x)))  )
    {
        fprintf(stderr,"CheckParameters: different image dimensions \n");
        return(-1);
    }
    
    return 1;
}

/***********************************************************CommentBegin******
 *
 * -- MacroblockBasedImagePadding --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      21.10.96
 *
 * Purpose :
 *      Macroblock-based image padding of the Vop images, according to the
 *      VM 5.1
 *
 * Arguments in :
 *      Image   *prev_rec_alpha,   Previous reconstructed alpha plane data
 *
 * Arguments in/out :
 *      Image   *cur_ima,          Image to pad 
 *
 * Arguments out :
 *
 *
 * Return values :
 *      1 on success, <0 on error
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the padding of the a image needed in the motion 
 *      estimation and compensation (see section 3.3.1).
 *
 *      1) The padded Image is NOT extended 16 pixels in all directions
 *         (it is supposed that this extension has been previously done).
 *      2) Padding is performed grouped in MB units.
 *      3) Padding of the MB uses the RECONSTRUCTED ALPHA values of the
 *         luminance in this MB (input parameter).
 *
 * See also :
 *
 *
 * Modified :
 *      06.02.97 Angel Pacheco: the padding of the transparent blocks
 *               is made looking at the most recently padded boundary
 *               block.
 *      09.05.97 Minhua Zhou: rewrote this fuction
 *     
 ***********************************************************CommentEnd********/
Int
MacroblockBasedImagePadding (
    Image   *prev_rec_alpha,  /* <-- Previous reconstructed alpha plane data */
    Int     default_pad_value,/* <-- Value used tp pad the non padded block  */
    Image   *cur_ima          /* <-> Image to pad                            */
    )
{
    SInt  * alpha,
        * alpha_modes,*cur_ima_data;
    Int     a_w,a_h,i,j;
    Int     h_size,w_size,i_ref,j_ref;
    
#ifdef _NO_PADDING_
    return(2);
#endif
    
    
/*
 * MACROBLOCK-BASED REPETITIVE PADDING 
 */
    
    /* checkings */
    if ( (cur_ima==NULL) || (prev_rec_alpha==NULL) || 
         ((a_w=GetImageSizeX(cur_ima))<1)  || 
         ((a_h=GetImageSizeY(cur_ima))<1)  || 
         ((alpha=(SInt*)GetImageData(prev_rec_alpha))==NULL) ||
         (a_w!=GetImageSizeX(prev_rec_alpha)) ||
         (a_h!=GetImageSizeY(prev_rec_alpha)) )
    {
        fprintf(stderr,
                "MacroblockBasedImagePadding: incorrect input parameters\n\n");
        return(-1);
    }
    
    w_size=a_w/MB_SIZE;
    h_size=a_h/MB_SIZE;
    
    /* array with MBM_TRANSPARENT/MBM_OPAQUE/MBM_BOUNDARY modes per MB */
    alpha_modes=(SInt*)malloc(sizeof(SInt)*w_size*h_size);
    subsamp_alpha(alpha,a_w,a_h,0/*MB-based*/,alpha_modes);
    cur_ima_data=(SInt*)GetImageData(cur_ima);
    for (i=0;i<a_w*a_h;i++)
        if (alpha[i]==0) cur_ima_data[i]=128;
    
    /* Padding of boundary blocks */
    for(j=1;j<(h_size-1);j++)
        for(i=1;i<(w_size-1);i++) 
            if(alpha_modes[w_size*j+i]==MBM_BOUNDARY) 
                BlockImagePadding(j,i, MB_SIZE, prev_rec_alpha,cur_ima);
    
    /* Blocks surrounded by transparent blocks should be also
       treated as boundary blocks in the next step of block repetiting */
    for(j=1;j<(h_size-1);j++)
        for(i=1;i<(w_size-1);i++)
        {
            if ((alpha_modes[w_size*j+i]==MBM_OPAQUE)&&
                ((alpha_modes[w_size*j+i-1]==MBM_TRANSPARENT)||
                 (alpha_modes[w_size*j+i+1]==MBM_TRANSPARENT)||
                 (alpha_modes[w_size*(j+1)+i]==MBM_TRANSPARENT)||
                 (alpha_modes[w_size*(j-1)+i]==MBM_TRANSPARENT)))
                alpha_modes[w_size*j+i]=MBM_BOUNDARY;
        }  
    
    for (j=0;j<h_size;j++)
        for (i=0;i<w_size;i++) 
            if(alpha_modes[w_size*j+i]==MBM_TRANSPARENT)
            {
                i_ref=j_ref=-1;
                if ((i-1>=0)&&(alpha_modes[w_size*j+i-1]==MBM_BOUNDARY))
                {
                    j_ref =j;i_ref =i-1;
                }
                else if ((i+1<=w_size-1)&&(alpha_modes[w_size*j+i+1]==MBM_BOUNDARY))
                {
                    j_ref =j;i_ref =i+1;
                }
                else if ((j-1>=0)&&(alpha_modes[w_size*(j-1)+i]==MBM_BOUNDARY))
                {
                    j_ref =j-1;i_ref =i;
                }
                else if ((j+1<=h_size-1)&&(alpha_modes[w_size*(j+1)+i]==MBM_BOUNDARY))
                {
                    j_ref =j+1;i_ref =i;
                }
                if (i_ref!=-1) 
                    BlockRepetitiveImagePadding(j,i,j_ref,i_ref,MB_SIZE,
                                                prev_rec_alpha,cur_ima);
            }
    
    
    free((Char*)alpha_modes);
    return 1;
}

/***********************************************************CommentBegin******
 *
 * -- TexturePadding --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / J. Ignacio Ronda
 *
 * Created :
 *      30.08.96
 *
 * Purpose :
 *      LPE/zero block-based padding of the YUV images
 *
 * Arguments in :
 *      Image *prev_ori_alpha,   previous original alpha plane data
 *      Image *alpha_decisions,  the previous original alpha plane
 *                               subsampled (in block units) which values
 *                               are:
 *                                MBM_TRANSPARENT = all pixels are transparent
 *                                      (outside the shape)
 *                                MBM_OPAQUE      = all pixels are not transp.
 *                                      (inside the shape)
 *                                MBM_BOUNDARY    = transparent and not
 *                                      transparent pixels at the same time
 *                                      (in the boundary of the shape)
 *      Image *MB_decisions,   encoding mode for each MB. If this Image is
 *                                NULL      => Frame based repetitive padd.
 *                                otherwise => Block based LPE and zero padding.
 *
 * Arguments in/out :
 *      Vop   *cur_vop         Vop to pad 
 *
 * Arguments out :
 *
 *
 * Return values :
 *      1 on success, <0 on error
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the padding of the YUV images of a given VOP
 *      needed in the texture coding.
 *
 *      0) Padding is performed in those MB that lie on the boundary of
 *         the shape, and in some neihbourgs.
 *      1) Padding is performed SEPARATELY FOR EACH of the Y, U and V 8x8
 *         blocks.
 *      2) Padding of the block uses the ORIGINAL ALPHA values of the
 *         luminance or chrominance in this block.
 *      3) Padding of chroma 8x8 block is performed by decimating the 16x16
 *         alpha block (by descarding every other row and column of pixels,
 *         starting from the second row and column).
 *      4) If all the pixels in an 8x8 block are transparent, their values
 *         are replaced by zero (not strictly necessary).
 *      5) There are two cases depending on the MB type:
 *           o intra 8x8 blocks 
 *              ==> LPE padding
 *           o residue 8x8 blocks (prediction error after the MC)
 *              ==> padding with zeroes
 *
 * See also :
 *
 *
 * Modified :
 *      05.02.97 A. Pacheco: modified to include the new low pass extrapolation
 *                 padding technique of the VM5.1
 *      05.02.97 A. Pacheco: modified to fill the pixels acording the original
 *                 alpha values (VM5.1)
 *      26.05.97 Minhua Zhou: Removed DoAlphaUVPlane() 
 *
 ***********************************************************CommentEnd********/

Int
CallingConvention
TexturePadding (
    Vop      *rec_vop,  /* <-- previous original alpha plane data */
    Image   *alpha_decisions, /* <-- subsampled (per block) prev_ori_alpha   */
    Image   *MB_decisions,    /* <-- encoding mode for each MB               */
    Vop     *cur_vop          /* <-> Vop to pad                              */
    )
{
    Image * aux_a, *prev_ori_alpha=GetVopA(rec_vop);
    Image * alpha_sub_ima;
    SInt  * mb_modes;
    SInt  * alpha,
        * a_sub,
        * alpha_uv,
        * alpha_sub;
    Int     error,a_w,a_h,i,j,a_sub_pos, mb_pos,padding_type;
    Int     h_size,w_size,uv_w_size;
    
    
#ifdef _NO_TEXTURE_PADDING_
    return(2);
#endif
    
    
    /* choose between frame-based or block based repetitive padding */
    
    if(MB_decisions==NULL)
    {
        
        /*
         * FRAME-BASED REPETITIVE PADDING 
         */
        
        /* update alpha plane to pad with */
        aux_a=GetVopA(cur_vop);
        cur_vop->a_chan=GetVopA(rec_vop); 
        
        /* do the repetitive frame-based padding */
        error=RepetitivePadding(0,cur_vop);
        
        /* update the vop */
        cur_vop->a_chan=aux_a;
        if(error!=1)
            fprintf(stderr,"TexturePadding: error in RepetitivePadding\n");
        
        return (error);
    }
    
    /*
     * BLOCK-BASED REPETITIVE PADDING 
     */
    
    /* checkings */
    if (CheckParameters(GetVopA(rec_vop),alpha_decisions,MB_decisions,cur_vop)!=1)
    {
        fprintf(stderr,"TexturePadding: incorrect input parameters\n\n");
        return(-1);
    }
    
    a_sub=   (SInt*)GetImageData(alpha_decisions);
    mb_modes=(SInt*)GetImageData(MB_decisions);
    alpha=(SInt*)GetImageData(GetVopA(rec_vop));
    a_w=GetImageSizeX(GetVopY(cur_vop)); a_h=GetImageSizeY(GetVopY(cur_vop));
    
    alpha_sub=(SInt*)GetImageData(GetVopAuv(rec_vop));
    
    w_size=GetImageSizeX(GetVopY(cur_vop))/MB_SIZE;
    h_size=GetImageSizeY(GetVopY(cur_vop))/MB_SIZE;
    uv_w_size=a_w/B_SIZE/2;
    alpha_sub_ima= GetVopAuv(rec_vop);
    alpha_uv= (SInt*)malloc(sizeof(SInt)*w_size*h_size);
    alpha_sub=(SInt*)GetImageData(alpha_sub_ima);
    subsamp_alpha(alpha_sub,a_w/2,a_h/2,1/*B-based*/,alpha_uv);
    
    for(j=0;j<h_size;j++)
        for(i=0;i<w_size;i++)
        {
            mb_pos=w_size*j+i;
            if (mb_modes[mb_pos]==MBM_INTRA)
                padding_type=1;
            else /* MBM_INTER16) || MBM_INTER8 or (MBM_TRANSPARENT||MBM_OUT??*/
                padding_type=0;
            
            /* Y blocks */
            a_sub_pos=w_size*2*j*2+i*2;
            if(a_sub[a_sub_pos]==MBM_TRANSPARENT)
                AllZeroFillBlock(j*2  ,i*2  ,0,cur_vop);
            else if (a_sub[a_sub_pos]==MBM_BOUNDARY)
                TextureBlockPadding(j*2,i*2,padding_type,B_SIZE,prev_ori_alpha,
                                    0,cur_vop);
            /* else MBM_OPAQUE, do nothing */
            
            a_sub_pos++;
            if(a_sub[a_sub_pos]==MBM_TRANSPARENT)
                AllZeroFillBlock(j*2  ,i*2+1,0,cur_vop);
            else if (a_sub[a_sub_pos]==MBM_BOUNDARY)
                TextureBlockPadding(j*2,i*2+1,padding_type,B_SIZE,prev_ori_alpha,
                                    0,cur_vop);
            /* else MBM_OPAQUE, do nothing */
            
            a_sub_pos=w_size*2*(2*j+1)+i*2;
            if(a_sub[a_sub_pos]==MBM_TRANSPARENT)
                AllZeroFillBlock(j*2+1,i*2  ,0,cur_vop);
            else if (a_sub[a_sub_pos]==MBM_BOUNDARY)
                TextureBlockPadding(j*2+1,i*2,padding_type,B_SIZE,prev_ori_alpha,
                                    0,cur_vop);
            /* else MBM_OPAQUE, do nothing */
            
            a_sub_pos++;
            if(a_sub[a_sub_pos]==MBM_TRANSPARENT)
                AllZeroFillBlock(j*2+1,i*2+1,0,cur_vop);
            else if (a_sub[a_sub_pos]==MBM_BOUNDARY)
                TextureBlockPadding(j*2+1,i*2+1,padding_type,B_SIZE, prev_ori_alpha,
                                    0,cur_vop);
            /* else MBM_OPAQUE, do nothing */
            
            /* U-V blocks */
            a_sub_pos=uv_w_size*j+i;
            if(alpha_uv[a_sub_pos]==MBM_TRANSPARENT)
                AllZeroFillBlock(j,i,1,cur_vop);
            else if(alpha_uv[a_sub_pos]==MBM_BOUNDARY)
                /* has alpha_sub*/
                TextureBlockPadding(j,i ,padding_type,B_SIZE, alpha_sub_ima,1,cur_vop);
            /* else MBM_OPAQUE, do nothing */
        }
    
    /* Free alpha_sub_ima  & alpha_uv before return - NO'C, 13/Sep/96 */
    free((Char *)alpha_uv);
    
    return 1;
}
/***********************************************************CommentBegin******
 *
 * -- BlockRepetitiveImagePadding --
 *
 * Author :
 *      UPM-GTI - Angel Pacheco
 *
 * Created :
 *      22.10.96
 *
 * Purpose :
 *      Repetitive padding of an image blocks (section 3.3.1 and figs. 2,3
 *      of VM4.0)
 *
 * Arguments in :
 *      Int   v,               vertical coord. (in block units) of the
 *                             block within the lumi image or chroma images
 *      Int   h,               horizontal coord. (in block units) of the
 *                             block within the lumi image or chroma images
 *      Int   vref,            vertical reference coord. (in block units) of the
 *                             block within the lumi image or chroma images
 *      Int   href,            horizontal reference coord. (in block units) of
 *                             the block within the lumi image or chroma images
 *      Int   block_size,      B_SIZE or MB_SIZE
 *      Image *prev_rec_alpha, previous reconstructed alpha plane (Y or
 *                             subsampled UV)
 * Arguments in/out :
 *      Image *cur_ima         Image to pad 
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the repetitive padding of an block (Y or U and V)
 *      of a given VOP. Needed in the estimation/compensation process.
 *
 *      2) Padding of the block must use the RECONSTRUCTED ALPHA values.
 *      4) The average of the values of the differents MB/Bs used to pad is
 *         performed in the calling code (here, it is no possible to know how
 *         many reference MB/Bs will infuence the current MB/B).
 *      5) No input checkings are performed!!
 *
 * See also :
 *
 *
 * Modified :
 *      06.02.97 Angel Pacheco: the padding of the transparent blocks
 *               is made looking at the most recently padded boundary
 *               block.
 *
 ***********************************************************CommentEnd********/
Void
BlockRepetitiveImagePadding (
    Int     v,                /* <-- vertical coord. (in block units)        */
    Int     h,                /* <-- horizontal coord. (in block units)      */
    Int     vref,             /* <-- vertical ref. coord. (in block units)   */
    Int     href,             /* <-- horizontal ref. coord. (in block units) */
    Int     block_size,       /* <-- B_SIZE or MB_SIZE                       */
    Image   *prev_rec_alpha,  /* <-- previous reconstructed alpha plane data */
    Image   *cur_ima          /* <-> Image to pad                            */
    )

{
    SInt * data, *block,*ref_block;
    Int    size_w,pos;
    Int    direction;
    Int    i,j;
    
    /* block-based repetitive padding */
    
    data=(SInt*)GetImageData(cur_ima);
    
    if (vref==v)
        direction=((href-h)<0)?LEFT_REFERENCE:RIGHT_REFERENCE;
    else /* href==h */
        direction=((vref-v)<0)?UPPER_REFERENCE:LOWER_REFERENCE;
    
    size_w=GetImageSizeX(cur_ima);
    
    /* Y repetitive-padding */
    block=(SInt*)LoadArea(data,h*block_size,v*block_size,
                          block_size,block_size,size_w);
    ref_block=(SInt*)LoadArea(data,href*block_size,vref*block_size,
                              block_size,block_size,size_w);
    
    if ((direction==UPPER_REFERENCE)||(direction==LOWER_REFERENCE))
    { 
        pos=(direction==UPPER_REFERENCE)?block_size*(block_size-1):0;
        for(j=0;j<block_size;j++)
            for(i=0;i<block_size;i++)
                block[j*block_size+i]+=ref_block[pos+i];
    }
    else
    {
        pos=(direction==LEFT_REFERENCE)?(block_size-1):0;
        for(j=0;j<block_size;j++)
            for(i=0;i<block_size;i++)
                block[j*block_size+i]+=ref_block[j*block_size+pos];   
    }
    
    SetArea(block,h*block_size,v*block_size,block_size,block_size,size_w,data);
    
    free((Char*)block);
    free((Char*)ref_block);
    
} /* BlockRepetitiveImagePadding */

/***********************************************************CommentBegin******
 *
 * -- BlockImagePadding --
 *
 * Authors :
 *      UPM-GTI - Angel Pacheco / Fernando Jaureguizar
 *
 * Created :
 *      23.10.96
 *
 * Purpose :
 *      Repetitive/zero padding of an 8x8/16x16 image blocks 
 *
 * Arguments in :
 *      Int     v,              vertical coord. (in block units) of the
 *                              block within the lumi image or chroma images
 *      Int     h,              horizontal coord. (in block units) of the
 *                              block within the lumi image or chroma images
 *      Int     block_size,     B_SIZE or MB_SIZE 
 *
 * Arguments in/out :
 *      Image   *cur_ima        Ima to pad 
 *
 * Arguments out :
 *
 *
 * Return values :
 *
 *
 * Side effects :
 *
 *
 * Description :
 *      This function performs the padding of an 8x8 block (Y or U and V) of a
 *      given VOP.
 *
 *      1) Padding is performed for an 8x8/16x16 block
 *      2) Padding of the block uses the RECONSTRUCTED ALPHA values.
 *
 *      No input checkings are performed
 *
 * See also :
 *
 *
 * Modified :
 *      23.10.96 A. Pacheco: modified to allow blocks of 8 and 16 units.
 *               ATTENTION: if block_size==16 the padding_type must be 1!! in
 *               the same way, there are combinations of parameters not checked
 *               because they would be not used at all.
 *
 ***********************************************************CommentEnd********/
Void
CallingConvention
BlockImagePadding (
    Int     v,                /* <-- vertical coord. (in block units)        */
    Int     h,                /* <-- horizontal coord. (in block units)      */
    Int     block_size,       /* <-- B_SIZE or MB_SIZE                       */
    Image   *prev_rec_alpha,  /* <-- previous reconstructed alpha plane data */
    Image   *cur_ima          /* <-> Image to pad                            */
    )

{
    SInt * a_data,*a_block;
    SInt * data,*block;
    SInt * status_block,*s_aux;
    Int    size_w,j;
    
    /* prev_rec_alpha has, depending on image_type, the right alpha plane
       subsampled for the UV data or for the Y data */
    
    /* block-based repetitive padding */
    
    data=(SInt*)GetImageData(cur_ima);
    a_data=(SInt*)GetImageData(prev_rec_alpha);
    
    /* creates and inits status_block */
    if ((status_block=s_aux=(SInt*)malloc(block_size*block_size*sizeof(SInt)))
        ==NULL)
    {
        fprintf(stderr,"BlockImagePadding: malloc failed\n");
        return;
    }
    for (j=0;j<(block_size*block_size);j++)
        s_aux[j]=NO_PAD;
    
    size_w=GetImageSizeX(cur_ima);
    
    /* Y repetitive-padding */
    block=(SInt*)LoadArea(data,h*block_size,v*block_size,
                          block_size,block_size,size_w);
    a_block=(SInt*)LoadArea(a_data,h*block_size,v*block_size,
                            block_size,block_size,size_w);
    
    Y_HorizontalPadding        (a_block,block_size,block_size,status_block,block);
    Y_VerticalOverlapedPadding (a_block,block_size,block_size,status_block,block);
    Y_FinalHorizontalPadding   (block_size, block_size, status_block, block);
    Y_FinalVerticalPadding     (block_size, block_size, status_block, block);
    
    SetArea(block,h*block_size,v*block_size,block_size,block_size,size_w,data);
    
    free((Char*)block);
    free((Char*)status_block);
    free((Char*)a_block);
    
} /* BlockImagePadding */

#endif
