// -*- c++ -*-

// This software module was originally developed by Matthew Leditschke
// (CSELT) and Karl Anders ygard (Telenor R&D), in the course of
// development of the MPEG-4 Systems standard (ISO/IEC 14496-1).  This
// software module is an implementation of a part of one or more MPEG-4
// Systems tools as specified by the MPEG-4 Systems standard.  ISO/IEC
// gives users of the MPEG-4 Systems standard free license to this software
// module or modifications thereof for use in hardware or software products
// claiming conformance to the MPEG-4 Systems standard.  Those intending to
// use this software module in hardware or software products are advised
// that its use may infringe existing patents.  The original developer of
// this software module and his/her company, the subsequent editors and
// their companies, and ISO/IEC have no liability for use of this software
// module or modifications thereof in an implementation.  Copyright is not
// released for non MPEG-4 Systems conforming products.  The original
// developer retains full right to use the code for his/her own purpose,
// assign or donate the code to a third party and to inhibit third parties
// from using the code for non MPEG-4 Systems conforming products.  This
// copyright notice must be included in all copies or derivative works.
//
// Copyright 1997, 1998.

#include "OpenGl/OpenGl.h"

#include "AudioSourceProxy.h"
#include "opengl/base/OpenGLPresenter.h"

#include "Stream.h"
#include "opengl/base/AudioRenderer.h"

#include <assert.h>


#ifndef WIN32

// Constructor

AudioSourceProxy::AudioSourceProxy(AudioSource* poAudioSource)
: MediaObjectProxy(poAudioSource),
m_skippedHeader(FALSE),
m_nOffset(0)
{
  m_format.samplesPerSecond = 0;
  m_format.bitsPerSample = 0;
  m_format.numChannels = 0;
}

// render the node

BOOL
AudioSourceProxy::Render(
  OpenGLVisualRenderer* /* pRenderer */, 
  DWORD /* dwTime */)
{
  return TRUE;
}

#else

// constructor

AudioSourceProxy::AudioSourceProxy(AudioSource* poAudioSource)
: MediaObjectProxy(poAudioSource),
miStreamId(-1),
m_finished(FALSE),
m_bFirst(TRUE)
{
}

#endif

// Render object description

/* virtual */ void
AudioSourceProxy::RenderDescription(OpenGLVisualRenderer* pRenderer, DWORD)
{
  RenderDescriptionBox(pRenderer, "AudioSource");
}

// Get audio format

const DecoderAudioFormat*
AudioSourceProxy::GetFormat()
{
  AudioSource* poAudioSource = (AudioSource*) GetMediaObject();
  MediaStream* pStream = poAudioSource->GetStream();
  
  // Only proceed once a stream has been attached
  
  if (pStream) {
    /* poAudioSource-> */ TestActive();
    
    if (/* poAudioSource-> */ GetActive()) {
      if (m_format.samplesPerSecond == 0) {
        int dataSize;
        DWORD rdwTime;
        BOOL rbEOU;
        LPVOID pvDataPtr = pStream->FetchNext(
          dataSize, rdwTime, rbEOU, TRUE);
        
        TRACE("AudioSourceProxy::GetFormat after FetchNext %08x", this);
        
        if (pvDataPtr != NULL &&
          pvDataPtr != END_OF_STREAM &&
          dataSize >= (int) sizeof(DecoderAudioFormat)) {
          ::memcpy(&m_format, pvDataPtr, sizeof(DecoderAudioFormat));
          
          // Release the media stream buffer (but don't discard)
          pStream->Release(FALSE);
        } else
          COMPLAIN("[AudioSourceProxy::GetFormat]: buffer error trying to get audio format\n");
      }
    }
  }
  
  return &m_format;
}

// ============================================================================

//
// AudioSourceProxy::GetData
//
// This function is used by the audio renderer to get more data to play.
// The audio render will ask for a given number of bytes (which will be a
// multiple of the number of bytes required for a single sample), and the
// audio source node will return a pointer to some data, and the number of
// bytes of data pointed to.
//
// The number of bytes returned can be less than that asked for.  If fewer
// bytes were returned to the audio renderer than expected, then the audio
// renderer will ask for more.  If zero bytes are returned, then the audio
// renderer will assume that no data can be given, and will insert silence.
//

void
AudioSourceProxy::GetData(int nBytesRequired,
                          char*& rpData,
                          int& rnBytesGiven)
{
  AudioSource* poAudioSource = (AudioSource*) GetMediaObject();
  MediaStream* pStream = poAudioSource->GetStream();
  
  // Only proceed once a stream has been attached
  
  if (pStream) {
    /* poAudioSource-> */ TestActive();
    
    if (/* poAudioSource-> */ GetActive()) {
      int dataSize;
      DWORD rdwTime;
      BOOL rbEOU;	
      char* pData = (char*) pStream->FetchNext(
        dataSize, rdwTime, rbEOU, FALSE);
      
      TRACE("AudioSourceProxy::GetData after FetchNext %08x", this);

      if (pData)
        // If end of stream has been reached, close the stream and return
        // zero bytes
        
        if (pData != END_OF_STREAM) {
          if (dataSize) {
            // Skip the header data if this is the first time data has been
            // read from this buffer by this function
            
            if (!m_skippedHeader) {
              m_skippedHeader = TRUE;
              m_nOffset += sizeof(DecoderAudioFormat);
            }
            
            // Put in possible offset in data block
            
            pData += m_nOffset;
            dataSize -= m_nOffset;
            
            rpData = pData;
            
            // If the number of bytes requested is less than the number read
            // from the media stream, only return some, and remember where to
            // get the rest from on the next call.
            
            if (dataSize > nBytesRequired) {
              // we only took some of the data
              
              m_nOffset += nBytesRequired;
              rnBytesGiven = nBytesRequired;
            } else {
              // we took all the data; clear m_nOffset
              
              m_nOffset = 0;
              rnBytesGiven = dataSize;
            }
            
            return;
          }
          
          pStream->Release();
        } else
          pStream->Terminate();
    }
  }
  
  rnBytesGiven = 0;
  rpData = NULL;
}

// Signal that we've used the data

void AudioSourceProxy::DataUsed(char* /*pData*/, int /*nBytesGiven*/)
{
  AudioSource* poAudioSource = (AudioSource*) GetMediaObject();
  
  // Release the media stream buffer.  If only part of this buffer was
  // given to the audio renderer, then don't discard it.
  
  if (m_nOffset)
    poAudioSource->GetStream()->Release(FALSE);
  else
    poAudioSource->GetStream()->Release();
}

#ifdef WIN32

// destructor

AudioSourceProxy::~AudioSourceProxy()
{
  // close down the stream

  if (miStreamId != -1)
    ((OpenGLPresenter *) ((MediaObject*) GetMediaObject())->GetExecutive()->GetPresenter())->GetAudioRenderer()->CloseOutputStream(miStreamId);


}

// render the node

BOOL
AudioSourceProxy::Render(
  OpenGLVisualRenderer* pRenderer, 
  DWORD /* dwTime */)
{
  mpoRenderer = pRenderer;
  AudioSource* poAudioSource = (AudioSource*) GetMediaObject();
  MediaStream* pStream = poAudioSource->GetStream();

// riitta200999
  MediaObject* parent_sound = ((MediaObject*)poAudioSource)->GetParent(0);


  TRACE("[AudioSourceProxy::Render] Stream Ptr %08x \n", pStream);

  // SB 99/03/03 $$$$ Start Stream & Register Channel
  if (m_bFirst)
  {
    Start();
    poAudioSource->GetExecutive()->RequestChannels();
    m_bFirst = FALSE;
    mlTimeInit = GetMediaObject()->GetRoot()->GetTime();
    mlTimeLocal = 0;
    mlCycles = 0;
    TRACE("[AudioSourceProxy::Render] StreamConsumer started \n");
  }
  
  // Only proceed once a stream has been attached
  if (pStream)
  {
    /* poAudioSource-> */ TestActive();
    
    if (/* poAudioSource-> */ GetActive())
    {
      // Get a pointer to the audio renderer
      AudioRenderer* pAR = pRenderer->GetPresenter()->GetAudioRenderer();
      
      TRACE("[AudioSourceProxy::Render] active with StreamId %4d \n", miStreamId);

      // If this is the first time that this function is called, set up
      // a connection with the audio renderer
      if (miStreamId == -1)
      {
        // Create an audio stream to play this file
//        miStreamId = pAR->OpenOutputStream(
//          pStream, 
//          8000, 2, 16, SAMPLEFMT_SIGNED_LITTLE, 240);


		  miStreamId = pAR->OpenOutputStream(
			  pStream, 
			  8000, 
			  2, 
			  16, 
			  SAMPLEFMT_SIGNED_LITTLE, 
			  240, 
			  parent_sound);

        ZThread<AudioSourceProxy>(
            AudioSourceThread, 
            this, 
            THREAD_PRIORITY_TIME_CRITICAL).Begin();
      }

      /******
      // ========================================== MAIN LOOP >>>>
      // Repeat until either there is no input data, 
      // or the output buffer is full.
      int iLoopingFlag = 0;
      while (!iLoopingFlag)
      {
        // SB $$$$ 99/03/09 added "local" timing for Audio Stream
        mulTimeLocal = GetMediaObject()->GetRoot()->GetTime() - mulTimeInit;
        mulCycles ++;
        TRACE("[AudiosourceProxy::Render] local time %8d \n", mulTimeLocal);

        if (mulTimeLocal < 30 * mulCycles)
        {
          TRACE("[AudiosourceProxy::Render] sleeping %8d \n", 
            30 * mulCycles - mulTimeLocal);
          Sleep(30 * mulCycles - mulTimeLocal);
        }

        // Get the next block of data from the decoder (via the media stream)
        int dataSize;
        DWORD rdwTime;
        BOOL rbEOU;
        BOOL bWait = FALSE;
        LPVOID pvDataPtr = pStream->FetchNext(
          dataSize, rdwTime, rbEOU, bWait);
        
        // If there is no data, we have finished for now
        if (pvDataPtr == NULL)
        {
          iLoopingFlag = 1;
        }
        
        // If the end of the stream has been reached, we have also finished
        else
        { // SB $$$$
          if (pvDataPtr == END_OF_STREAM)
          {
            m_finished = TRUE;
            iLoopingFlag = 1;
          }
          else
          { // SB $$$$
            // Otherwise see if this block fits in the output buffer
            if (pAR->QueueFrameForPlaying(miStreamId, pvDataPtr))
            {
              // If it does, we have finished with this block from
              // the stream
              pStream->Release(TRUE);
            }
            else
            {
              // Otherwise put this block of data back to be read later
              // when there is space in the output buffer
              pStream->Release(FALSE);
              iLoopingFlag = 1;
            }
          }
        }
      }
      // ========================================== MAIN LOOP <<<<
      ******/

    }
  }
  
  return TRUE;
}

#endif // WIN32


void AudioSourceProxy::AudioSourceThread()
{
  AudioSource* poAudioSource = (AudioSource*) GetMediaObject();
  MediaStream* poStream = poAudioSource->GetStream();
  AudioRenderer* poRendAudio = mpoRenderer->GetPresenter()->GetAudioRenderer();

  // Repeat until either there is no input data, 
  // or the output buffer is full.
  int iLoopingFlag = 1;
  while (iLoopingFlag)
  {
    // SB $$$$ 99/03/09 added "local" timing for Audio Stream
    mlTimeLocal = GetMediaObject()->GetRoot()->GetTime() - mlTimeInit;
    mlCycles ++;

    TRACE("[ASP::Thread] local %8d cycle %8d \n", 
      mlTimeLocal, 30 * mlCycles);

    if (mlTimeLocal < (30 * mlCycles - 270))
    {
      TRACE("[ASP::Thread] sleeping %8d \n", 
        (30 * mlCycles - 270) - mlTimeLocal);
      Sleep((30 * mlCycles - 270) - mlTimeLocal);
    }

    // Get the next block of data from the decoder (via the media stream)
    int iDataSize;
    DWORD rdwTime;
    BOOL rbEOU;
    BOOL bWait = TRUE;
    LPVOID pvDataPtr = poStream->FetchNext(
      iDataSize, rdwTime, rbEOU, bWait);
    
    // If there is no data, we have finished for now
    if (pvDataPtr == NULL)
    {
      iLoopingFlag = 0;
    }
    
    // If the end of the stream has been reached, we have also finished
    else
    { // SB $$$$
      if (pvDataPtr == END_OF_STREAM)
      {
        m_finished = TRUE;
        iLoopingFlag = 0;
      }
      else
      { // SB $$$$
        // Otherwise see if this block fits in the output buffer
        if (poRendAudio->QueueFrameForPlaying(miStreamId, pvDataPtr))
        {
          // If it does, we have finished with this block from
          // the stream
          poStream->Release(TRUE);
        }
        else
        {
          // Otherwise put this block of data back to be read later
          // when there is space in the output buffer
          poStream->Release(FALSE);
		//	 poStream->Release(TRUE);
          // SB 99/08/05 ?!?! does this make sense ?!?!
          // iLoopingFlag = 0;
        }
      }
    }
  }
}
