// -*- c++ -*-

// This software module was originally developed by Matthew Leditschke
// (CSELT) and Karl Anders ygard (Telenor R&D), in the course of
// development of the MPEG-4 Systems standard (ISO/IEC 14496-1).  This
// software module is an implementation of a part of one or more MPEG-4
// Systems tools as specified by the MPEG-4 Systems standard.  ISO/IEC
// gives users of the MPEG-4 Systems standard free license to this software
// module or modifications thereof for use in hardware or software products
// claiming conformance to the MPEG-4 Systems standard.  Those intending to
// use this software module in hardware or software products are advised
// that its use may infringe existing patents.  The original developer of
// this software module and his/her company, the subsequent editors and
// their companies, and ISO/IEC have no liability for use of this software
// module or modifications thereof in an implementation.  Copyright is not
// released for non MPEG-4 Systems conforming products.  The original
// developer retains full right to use the code for his/her own purpose,
// assign or donate the code to a third party and to inhibit third parties
// from using the code for non MPEG-4 Systems conforming products.  This
// copyright notice must be included in all copies or derivative works.
//
// Copyright 1997, 1998.

#include "defs.h"

#include "AudioSourceProxy.h"
#include "opengl/base/OpenGLPresenter.h"

#include "Stream.h"
#include "opengl/base/AudioRenderer.h"

#include <assert.h>

// Stefano Battista 98/11/12
// #include "group/Root.h"
#include "Nodes.h"

#ifndef WIN32

// Constructor

AudioSourceProxy::AudioSourceProxy(AudioSource* pAudioSource)
: MediaObjectProxy(pAudioSource),
m_skippedHeader(FALSE),
m_nOffset(0)
{
  m_format.samplesPerSecond = 0;
  m_format.bitsPerSample = 0;
  m_format.numChannels = 0;
}

// render the node

BOOL
AudioSourceProxy::Render(OpenGLVisualRenderer* /* pRenderer */, DWORD /* dwTime */)
{
  return TRUE;
}

#else

// constructor

AudioSourceProxy::AudioSourceProxy(AudioSource* pAudioSource)
: MediaObjectProxy(pAudioSource),
streamID(-1),
m_finished(FALSE),
m_bFirst(TRUE)
{
}

#endif

// Render object description

/* virtual */ void
AudioSourceProxy::RenderDescription(OpenGLVisualRenderer* pRenderer, DWORD)
{
  RenderDescriptionBox(pRenderer, "AudioSource");
}

// Get audio format

const DecoderAudioFormat*
AudioSourceProxy::GetFormat()
{
  AudioSource* pAudioSource = (AudioSource*) GetMediaObject();
  MediaStream* pStream = pAudioSource->GetStream();
  
  // Only proceed once a stream has been attached
  
  if (pStream) {
    /* pAudioSource-> */ TestActive();
    
    if (/* pAudioSource-> */ GetActive()) {
      if (m_format.samplesPerSecond == 0) {
        int dataSize;
        DWORD rdwTime;
        BOOL rbEOU;
        LPVOID dataPtr = pStream->FetchNext(
          dataSize, rdwTime, rbEOU, TRUE);
        
        TRACE("AudioSourceProxy::GetFormat after FetchNext %08x", this);
        
        if (dataPtr != NULL &&
          dataPtr != END_OF_STREAM &&
          dataSize >= (int) sizeof(DecoderAudioFormat)) {
          ::memcpy(&m_format, dataPtr, sizeof(DecoderAudioFormat));
          
          // Release the media stream buffer (but don't discard)
          pStream->Release(FALSE);
        } else
          COMPLAIN("[AudioSourceProxy::GetFormat]: buffer error trying to get audio format\n");
      }
    }
  }
  
  return &m_format;
}

// ============================================================================

//
// AudioSourceProxy::GetData
//
// This function is used by the audio renderer to get more data to play.
// The audio render will ask for a given number of bytes (which will be a
// multiple of the number of bytes required for a single sample), and the
// audio source node will return a pointer to some data, and the number of
// bytes of data pointed to.
//
// The number of bytes returned can be less than that asked for.  If fewer
// bytes were returned to the audio renderer than expected, then the audio
// renderer will ask for more.  If zero bytes are returned, then the audio
// renderer will assume that no data can be given, and will insert silence.
//

void
AudioSourceProxy::GetData(int nBytesRequired,
                          char*& rpData,
                          int& rnBytesGiven)
{
  AudioSource* pAudioSource = (AudioSource*) GetMediaObject();
  MediaStream* pStream = pAudioSource->GetStream();
  
  // Only proceed once a stream has been attached
  
  if (pStream) {
    /* pAudioSource-> */ TestActive();
    
    if (/* pAudioSource-> */ GetActive()) {
      int dataSize;
      DWORD rdwTime;
      BOOL rbEOU;	
      char* pData = (char*) pStream->FetchNext(
        dataSize, rdwTime, rbEOU, FALSE);
      
      COMPLAIN("AudioSourceProxy::GetData after FetchNext %08x", this);

      if (pData)
        // If end of stream has been reached, close the stream and return
        // zero bytes
        
        if (pData != END_OF_STREAM) {
          if (dataSize) {
            // Skip the header data if this is the first time data has been
            // read from this buffer by this function
            
            if (!m_skippedHeader) {
              m_skippedHeader = TRUE;
              m_nOffset += sizeof(DecoderAudioFormat);
            }
            
            // Put in possible offset in data block
            
            pData += m_nOffset;
            dataSize -= m_nOffset;
            
            rpData = pData;
            
            // If the number of bytes requested is less than the number read
            // from the media stream, only return some, and remember where to
            // get the rest from on the next call.
            
            if (dataSize > nBytesRequired) {
              // we only took some of the data
              
              m_nOffset += nBytesRequired;
              rnBytesGiven = nBytesRequired;
            } else {
              // we took all the data; clear m_nOffset
              
              m_nOffset = 0;
              rnBytesGiven = dataSize;
            }
            
            return;
          }
          
          pStream->Release();
        } else
          pStream->Terminate();
    }
  }
  
  rnBytesGiven = 0;
  rpData = NULL;
}

// Signal that we've used the data

void AudioSourceProxy::DataUsed(char* /*pData*/, int /*nBytesGiven*/)
{
  AudioSource* pAudioSource = (AudioSource*) GetMediaObject();
  
  // Release the media stream buffer.  If only part of this buffer was
  // given to the audio renderer, then don't discard it.
  
  if (m_nOffset)
    pAudioSource->GetStream()->Release(FALSE);
  else
    pAudioSource->GetStream()->Release();
}

#ifdef WIN32

// destructor

AudioSourceProxy::~AudioSourceProxy()
{
  // close down the stream
  
  if (streamID != -1)
    ((OpenGLPresenter *) ((MediaObject*) GetMediaObject())->GetExecutive()->GetPresenter())->GetAudioRenderer()->CloseOutputStream(streamID);
}

// render the node

BOOL
AudioSourceProxy::Render(OpenGLVisualRenderer* pRenderer, DWORD /* dwTime */)
{
  AudioSource* pAudioSource = (AudioSource*) GetMediaObject();
  MediaStream* pStream = pAudioSource->GetStream();
  
// riitta042599
  MediaObject* parent_sound = ((MediaObject*)pAudioSource)->GetParent(0);


  // SB 99/03/03 $$$$ Start Stream & Register Channel
  if (m_bFirst)
  {
    Start();
    pAudioSource->GetExecutive()->RequestChannels();
    m_bFirst = FALSE;
    TRACE("[AudioSourceProxy::Render] StreamConsumer started \n");
  }
  
  // Only proceed once a stream has been attached
  if (pStream)
  {
    /* pAudioSource-> */ TestActive();
    
    if (/* pAudioSource-> */ GetActive())
    {
      // Get a pointer to the audio renderer
      AudioRenderer* pAR = pRenderer->GetPresenter()->GetAudioRenderer();
      
      //	TRACE_PRINT1("AudioProxy, stream = 0x%08x\n", pStream);
      
      // If this is the first time that this function is called, set up
      // a connection with the audio renderer
      if (streamID == -1)
      {
        // Create an audio stream to play this file
//        streamID = pAR->OpenOutputStream(pStream, 8000, 2, 16,
//          SAMPLEFMT_SIGNED_LITTLE,
//          240);
		  streamID = pAR->OpenOutputStream(pStream, 8000, 2, 16,
          SAMPLEFMT_SIGNED_LITTLE,
          240, parent_sound);
      }
      
      // Repeat until either there is no input data, or the output buffer
      // is full.
      int finishedLooping = 0;
      while (!finishedLooping)
      {
        // Get the next block of data from the decoder (via the media stream)
        int dataSize;
        DWORD rdwTime;
        BOOL rbEOU;
        // SB $$$$ 99/03/02 to Wait or not to Wait ?!?!
        BOOL bWait = FALSE;
        LPVOID dataPtr = pStream->FetchNext(
          dataSize, rdwTime, rbEOU, bWait);
        
        TRACE("AudioSourceProxy::Render after FetchNext %08x \n", this);
        
        // If there is no data, we have finished for now
        if (dataPtr == NULL)
        {
           finishedLooping = 1;
          TRACE("AudioProxy: No data read at TIME %d \n", 
            GetMediaObject()->GetRoot()->GetTime());
        }
        
        // If the end of the stream has been reached, we have also finished
        else
        { // SB $$$$
          if (dataPtr == END_OF_STREAM)
          {
            m_finished = TRUE;
            finishedLooping = 1;
            TRACE("AudioProxy: End of stream at TIME %d \n",
              GetMediaObject()->GetRoot()->GetTime());
          }
          else
          { // SB $$$$
            // Otherwise see if this block fits in the output buffer
            if (pAR->QueueFrameForPlaying(streamID, dataPtr))
            {
              // If it does, we have finished with this block from
              // the stream
              pStream->Release(TRUE);
              TRACE("AudioProxy: Queueing for play TIME %d at TIME %d\n",
                rdwTime,
                GetMediaObject()->GetRoot()->GetTime());
            }
            else
            {
              // Otherwise put this block of data back to be read later
              // when there is space in the output buffer
				//riitta 070399 pStream->Release(FALSE);
				pStream->Release(TRUE);
              finishedLooping = 1;
              TRACE("AudioProxy: Putting back time %d (output buffer full)\n",
				rdwTime);
            }
          } // if (dataPtr == END_OF_STREAM) else {...
        } // if (dataPtr == END_OF_STREAM) else {...
      } // while (!finishedLooping) {...
    } // if (/* pAudioSource-> */ GetActive()) {...
  } // else // if (pStream) 
	 // pStream->CloseStream();
  
  return TRUE;
}

#endif
