// This software module was originally developed by Matthew Leditschke
// (CSELT) and Karl Anders ygard (Telenor R&D), in the course of
// development of the MPEG-4 Systems standard (ISO/IEC 14496-1).  This
// software module is an implementation of a part of one or more MPEG-4
// Systems tools as specified by the MPEG-4 Systems standard.  ISO/IEC
// gives users of the MPEG-4 Systems standard free license to this software
// module or modifications thereof for use in hardware or software products
// claiming conformance to the MPEG-4 Systems standard.  Those intending to
// use this software module in hardware or software products are advised
// that its use may infringe existing patents.  The original developer of
// this software module and his/her company, the subsequent editors and
// their companies, and ISO/IEC have no liability for use of this software
// module or modifications thereof in an implementation.  Copyright is not
// released for non MPEG-4 Systems conforming products.  The original
// developer retains full right to use the code for his/her own purpose,
// assign or donate the code to a third party and to inhibit third parties
// from using the code for non MPEG-4 Systems conforming products.  This
// copyright notice must be included in all copies or derivative works.
//
// Copyright 1997.

#include "defs.h"
#include "ztl.h"
#include <assert.h>
#include "AudioRendererSgi.h"
#include "Caps.h"
#include "Executiv.h"
#include "Presenter.h"
#include "Stream.h"
#include "audio/Sound.h"
#include "opengl/audio/AudioObjectProxy.h"
#include "opengl/audio/SoundProxy.h"
#include "opengl/audio/Sound2DProxy.h"
#include "audio/AudioObject.h"
#include "audio/Sound.h"
#include <stropts.h>
#include <poll.h>
// Stefano Battista 98/11/12
// #include "group/Root.h"
// #include "bindable/Viewpoint.h"
#include "Nodes.h"

// The length of the output queue (in milliseconds)

const int QUEUE_LENGTH = 1000;

// Stops rendering and performs cleanups

/* virtual */ void
AudioRendererSgi::Terminate()
{
  // Flag that the audio renderer is no longer active
  m_bActive = FALSE;
  
  // Wait for this thread to terminate
  m_mtxThread.P();
  m_mtxThread.V();
  
  // Release the audio output related resources
  if (m_audioPort)
    ::ALcloseport(m_audioPort);
  
  if (m_pMixBuffer)
    delete m_pMixBuffer;

  if (m_pSpatializeBuffer)
    delete m_pSpatializeBuffer;

  delete [] sound_loc;
  delete [] sound_dir;
  delete [] ellip_def;

  for (int j = 0; j < nof_sources; j++) {
    delete ellip_array[j];
    delete pan_array[j];
  }
}

// Initialise the renderer

/* virtual */ void
AudioRendererSgi::Init()
{
  // Start the main thread of the audio renderer
  if (ZThread<AudioRendererSgi>(Main, this).Begin() == -1)
    throw;
}

// Init audio

void
AudioRendererSgi::InitAudioOutput(int samplesPerSec,
				  int bitsPerSample,
				  int numChannels)
{
  numChannels = 2;

  ALconfig audioConfig;

  // Remove the default error handler
  ::ALseterrorhandler(0);

  // Specify the configuration of the output audio port
  audioConfig = ::ALnewconfig();

  ::ALsetsampfmt(audioConfig, AL_SAMPFMT_TWOSCOMP);

  if (bitsPerSample <= 8)
    ::ALsetwidth(audioConfig, AL_SAMPLE_8);

  ::ALsetchannels(audioConfig, (long) numChannels);

  // Determine and set the length of the output queue (in samples)
  m_outputQueueSize = numChannels * samplesPerSec * QUEUE_LENGTH / 1000;

  ::ALsetqueuesize(audioConfig, m_outputQueueSize);

  // Create the audio port
  m_audioPort = ::ALopenport("im1", "w", audioConfig);

  // Free the config data; we don't need it any more
  ::ALfreeconfig(audioConfig);

  if (!m_audioPort)
    switch (::oserror()) {
    case AL_BAD_NO_PORTS:
      COMPLAIN("System is out of audio ports!\n");
      break;
	  
    case AL_BAD_DEVICE_ACCESS:
      COMPLAIN("Couldn't access audio device!\n");
      break;
	  
    case AL_BAD_OUT_OF_MEM:
      COMPLAIN("Out of memory!\n");
      break;
    }
  
  {
    // Change the output sample rate
	
    long al_params[] = {AL_OUTPUT_RATE, 0};

    al_params[1] = samplesPerSec;
	
    ::ALsetparams(AL_DEFAULT_DEVICE, al_params, 2);
    ::ALgetparams(AL_DEFAULT_DEVICE, al_params, 2);

    if (al_params[1] != samplesPerSec)
      TRACE("given sample rate (%d) was invalid, current is %d\n", samplesPerSec, al_params[1]);
  }

  // Remember the output port settings
  m_samplesPerSec  = samplesPerSec;
  m_bitsPerSample  = bitsPerSample;
  m_numChannels    = numChannels;
  m_bytesPerSample = numChannels * bitsPerSample / 8;
  
  // Allocate a buffer to mix samples from different streams in
  m_pMixBuffer = new char[m_outputQueueSize * m_bytesPerSample];

  m_pSpatializeBuffer = new short[m_outputQueueSize * m_bytesPerSample];
  
  for (int j = 0; j < 10; j++){
    pan_array[j] = NULL;
    ellip_array[j] = NULL;
  }
}

// Main rendering loop

void
AudioRendererSgi::Main()
{
  BOOL first = true;
  
  // Mark the thread as running.
  m_mtxThread.P();

  // Flag that the audio renderer is now active
  m_bActive = TRUE;
  
  {
    ZArray<Sound*> soundSources3D;
    ZArray<Sound2D*> soundSources2D;
    ZArray<Matrix> soundTransformations3D;
    ZArray<Matrix> soundTransformations2D;
    int numberOfSources;
    
    GetPresenter()->GetExecutive()->LockScene();

    numberOfSources = ((MediaObjectProxy*) GetPresenter()->GetScene()->GetRenderingObject())->FindSoundSources(soundSources3D, soundTransformations3D,
                                                                                                             soundSources2D, soundTransformations2D);

    GetPresenter()->GetExecutive()->UnlockScene();

    if (numberOfSources > 0) {
      const DecoderAudioFormat* pFormat;

      if (soundSources3D.GetLength())
	pFormat = ((AudioObjectProxy*) soundSources3D[0]->GetSound()->GetRenderingObject())->GetFormat();
      else
	pFormat = ((AudioObjectProxy*) soundSources2D[0]->GetSound()->GetRenderingObject())->GetFormat();

      if (pFormat) {
	InitAudioOutput(pFormat->samplesPerSecond,
			pFormat->bitsPerSample,
			pFormat->numChannels);

	TRACE("[Audio renderer]: params %d, %d, %d\n", pFormat->samplesPerSecond,
			                                  pFormat->bitsPerSample,
			                                  pFormat->numChannels);
      } else
	// Initialise the audio device with defaults

	InitAudioOutput(8000, 16, 2);
    } else
      // Initialise the audio device with defaults

      InitAudioOutput(8000, 16, 2);
  }

  // Read data from the active buffers, merging the
  // results together into the mixing buffer.

  ZArray<Sound*> soundSources3D;
  ZArray<Sound2D*> soundSources2D;
  ZArray<Matrix> soundTransformations3D;
  ZArray<Matrix> soundTransformations2D;

  ((MediaObjectProxy*) GetPresenter()->GetScene()->GetRenderingObject())->FindSoundSources(soundSources3D, soundTransformations3D,soundSources2D, soundTransformations2D);

  nof_sources = soundSources3D.GetLength();

  sound_loc = new Vec3f[nof_sources];
  sound_dir = new Vec3f[nof_sources];
  ellip_def = new float[nof_sources * 4];
  GetViewPosition();

  for (int i = 0; i < nof_sources; i++){
    GetSoundPosition(i, soundSources3D[i], soundTransformations3D[i]);
    InitSpatializeAndEllipse(i, soundSources3D[i]);
    
    pan_array[i] = new AmPan(viewp,
			     sound_loc[i],
			     sound_dir[i]);
    
    ellip_array[i] = new EllipseSound(viewp,
				      sound_loc[i],
				      sound_dir[i],
				      &ellip_def[i * 4]);
  }

  // This loop runs while the audio renderer is active, ie. up to the next
  // Terminate() call.

  while (m_bActive) {
    // how many samples to try to write

    long samplesToWrite = m_outputQueueSize / 2;

    // Only read data from the streams if there is room

    if (samplesToWrite > 0) {
      // Read data from the active buffers, merging the
      // results together into the mixing buffer.

      ZArray<Sound*> soundSources3D;
      ZArray<Sound2D*> soundSources2D;
      ZArray<Matrix> soundTransformations3D;
      ZArray<Matrix> soundTransformations2D;
      int numberOfSources = 0;
      BOOL bufferEmpty = TRUE;

      GetPresenter()->GetExecutive()->LockScene();

      if (GetPresenter()->GetScene())
        numberOfSources = ((MediaObjectProxy*) GetPresenter()->GetScene()->GetRenderingObject())->FindSoundSources(soundSources3D, soundTransformations3D,
                                                                                                                 soundSources2D, soundTransformations2D);

      GetPresenter()->GetExecutive()->UnlockScene();

      if (numberOfSources > 0) {
	int i;

	for (i = soundSources3D.GetLength() - 1; i >= 0; --i)  {
	  // Determine number of bytes to read from this source

	  const DecoderAudioFormat* pFormat = ((AudioObjectProxy*) soundSources3D[i]->GetSound()->GetRenderingObject())->GetFormat();

	  if (pFormat) {
	    int bytesPerSample = pFormat->bitsPerSample / 8;
	    int bytesToWrite = samplesToWrite * bytesPerSample;
	    int bytesWritten = 0;
			
	    // Repeatedly read data for this buffer, until either the buffer
	    // is full, or there is no more data

	    int bytesRead;

	    do {
	      // Read data from the audio node

	      char* pData;

	      ((AudioObjectProxy*) soundSources3D[i]->GetSound()->GetRenderingObject())->GetData(bytesToWrite,
											       pData,
											       bytesRead);
	      if (bytesRead > 0) {
		GetSoundPosition(i, soundSources3D[i], soundTransformations3D[i]);
		
		if (pan_array[i])
		  pan_array[i]->ChangeGains(viewp,
					    sound_loc[i],
					    sound_dir[i]);
		if (ellip_array[i])
		  ellip_array[i]->ChangeGain(viewp,
					     sound_loc[i],
					     sound_dir[i],
					     &ellip_def[i * 4]);
		
		short* temp = NULL;
		float y = 0.0;

		temp = (short*) pData;
		for (int j = 0; j < bytesRead; j += 2) {
		  
		  if (ellip_array[i])
		    y = (*ellip_array[i]) (float(*temp));
		  else
		    y = float(*temp);

		  if (pan_array[i]) {
		    x = (*pan_array[i]) (float(y));
		    m_pSpatializeBuffer[j] = short(x[0]);
		    m_pSpatializeBuffer[j + 1] = short(x[1]);
		  } else {
		    // y = (float)(*temp);
		    m_pSpatializeBuffer[j] = short(y);
		    m_pSpatializeBuffer[j + 1] = short(y);
		  }
		  
		  temp++;
		}
		
		if (first) {
		  // Adjust system overhead first time we get data

		  soundSources3D[i]->GetSound()->GetStream()->AdjustClock();
		  first = false;
		}

		// If this is the first stream, copy it into
		// the mixing buffer.

		if (bufferEmpty)
//		  ::memcpy(m_pMixBuffer + bytesWritten, pData, bytesRead);
                  ::memcpy(m_pMixBuffer + bytesWritten, m_pSpatializeBuffer, bytesRead * 2);
		else
		  // otherwise combine it with the existing data

//		  CombineData(bytesWritten, pData, bytesRead);
                  CombineData(bytesWritten, (char*) m_pSpatializeBuffer, bytesRead * 2);
				
		// Release this data

		((AudioObjectProxy*) soundSources3D[i]->GetSound()->GetRenderingObject())->DataUsed(pData,
												  bytesRead);
			  
		// Update how much data is left to get

		bytesToWrite -= bytesRead;
//		bytesWritten += bytesRead;
		bytesWritten += bytesRead * 2;
	      }
	    } while (bytesToWrite > 0 && bytesRead > 0);
			
	    if (bufferEmpty && bytesWritten > 0) {
	      bufferEmpty = FALSE;

	      if (bytesToWrite)
		samplesToWrite = bytesWritten / bytesPerSample;
	    }
	  }
	}

	for (i = soundSources2D.GetLength() - 1; i >= 0; --i)  {
	  // Determine number of bytes to read from this source
	  const DecoderAudioFormat* pFormat = ((AudioObjectProxy*) soundSources2D[i]->GetSound()->GetRenderingObject())->GetFormat();

	  if (pFormat) {
	    int bytesPerSample = pFormat->bitsPerSample / 8;
	    int bytesToWrite = samplesToWrite * bytesPerSample;
	    int bytesWritten = 0;
			
	    // Repeatedly read data for this buffer, until either the buffer
	    // is full, or there is no more data

	    int bytesRead;

	    do {
	      // Read data from the audio node

	      char* pData;

	      ((AudioObjectProxy*) soundSources2D[i]->GetSound()->GetRenderingObject())->GetData(bytesToWrite,
											       pData,
											       bytesRead);

	      if (bytesRead > 0) {
		if (first) {
		  // Adjust system overhead first time we get data

		  soundSources2D[i]->GetSound()->GetStream()->AdjustClock();
		  first = false;
		}

		// If this is the first stream, copy it into
		// the mixing buffer.

		if (bufferEmpty)
		  ::memcpy(m_pMixBuffer + bytesWritten, pData, bytesRead);
		else
		  // otherwise combine it with the existing data

		  CombineData(bytesWritten, pData, bytesRead);
				
		// Release this data

		((AudioObjectProxy*) soundSources2D[i]->GetSound()->GetRenderingObject())->DataUsed(pData,
												  bytesRead);
			  
		// Update how much data is left to get

		bytesToWrite -= bytesRead;
		bytesWritten += bytesRead;
	      }
	    } while (bytesToWrite > 0 && bytesRead > 0);
			
	    if (bufferEmpty && bytesWritten > 0) {
	      bufferEmpty = FALSE;

	      if (bytesToWrite)
		samplesToWrite = bytesWritten / bytesPerSample;
	    }
	  }
	}

	if (!bufferEmpty)
	  // Write combined samples to the audio port

	  ::ALwritesamps(m_audioPort, m_pMixBuffer, samplesToWrite);
	else
	  // Wait a little instead, so we don't chew up CPU just looking
	  // for more samples

	  ::usleep(1000);
      }
    }
  }
  
  // Flag that the thread is terminating
  m_mtxThread.V();
}

/// Mix audio

void
AudioRendererSgi::CombineData(int offset,
			      char* pSrc,
			      int lenSrc)
{
  DWORD numSamps  = lenSrc / m_bytesPerSample;

  if (m_bitsPerSample <= 8) {
    // Assume 8 bit unsigned format
    unsigned char* oldSamps = (unsigned char*) (m_pMixBuffer + offset);
    unsigned char* newSamps = (unsigned char*) pSrc;

    for (DWORD i = 0; i < numSamps; i++) {
      unsigned short temp = oldSamps[i] + newSamps[i];

      if (temp > 255)
	oldSamps[i] = 255;
      else
	oldSamps[i] = (unsigned char) temp;
    }
  } else {
    // Assume 16 bit signed format
    signed short* oldSamps = (signed short*) (m_pMixBuffer + offset);
    signed short* newSamps = (signed short*) pSrc;

    for (DWORD i = 0; i < numSamps; i++) {
      signed long temp = oldSamps[i] + newSamps[i];

      if (temp < -32768)
	oldSamps[i] = -32768;
      else
	if (temp > 32767)
	  oldSamps[i] = 32767;
	else
	  oldSamps[i] = (signed short) temp;
    }
  }
}

void
AudioRendererSgi::GetViewPosition()
{
  //ML - changed the following line
  Viewpoint* pActiveViewpoint = GetPresenter()->GetScene()->GetBoundViewpoint();
  //Viewpoint* pActiveViewpoint = GetPresenter()->GetScene()->GetRoot()->GetBoundViewpoint();
  Vec3f position(pActiveViewpoint->GetPosition());
  Rotation orientation(pActiveViewpoint->GetOrientation());

  viewp[0] = position.x;
  viewp[1] = position.y;
  viewp[2] = position.z;
  viewp[3] = orientation.theta*orientation.y;
}

void
AudioRendererSgi::GetSoundPosition(int i, Sound* pSound3D, const Matrix& transf_matrix)
{
  const Vec3f sound_pos = pSound3D->GetLocation();
  const Vec3f sound_d = pSound3D->GetDirection();
  //  spat = soundSources3D[stream_id]->GetSpatialize();

  sound_loc[i].x = sound_pos.x + transf_matrix[12]; 
  sound_loc[i].y = sound_pos.y + transf_matrix[13];
  sound_loc[i].z = sound_pos.z + transf_matrix[14];

  sound_dir[i] = sound_d;

  ellip_def[i * 4]     = pSound3D->GetMinBack();
  ellip_def[i * 4 + 1] = pSound3D->GetMinFront();
  ellip_def[i * 4 + 2] = pSound3D->GetMaxBack();
  ellip_def[i * 4 + 3] = pSound3D->GetMaxFront();
}

void
AudioRendererSgi::InitSpatializeAndEllipse(int i, Sound* pSound3D)
{
  Bool spat = pSound3D->GetSpatialize();

  if (spat)
    pan_array[i] = new AmPan(viewp,
			     sound_loc[i],
			     sound_dir[i]);
  
  if ((ellip_def[i * 4]     && ellip_def[i * 4 + 1]) ||
      (ellip_def[i * 4 + 2] && ellip_def[i * 4 + 3]))
    ellip_array[i] = new EllipseSound(viewp,
				      sound_loc[i],
				      sound_dir[i],
				      &ellip_def[i * 4]);
}
