// This software module was originally developed by Edward Cooke (Teltec
// Ireland) and Karl Anders ygard (Telenor R&D) in the course of
// development of the MPEG-4 Systems standard (ISO/IEC 14496-1).  This
// software module is an implementation of a part of one or more MPEG-4
// Systems tools as specified by the MPEG-4 Systems standard.  ISO/IEC
// gives users of the MPEG-4 Systems standard free license to this software
// module or modifications thereof for use in hardware or software products
// claiming conformance to the MPEG-4 Systems standard.  Those intending to
// use this software module in hardware or software products are advised
// that its use may infringe existing patents.  The original developer of
// this software module and his/her company, the subsequent editors and
// their companies, and ISO/IEC have no liability for use of this software
// module or modifications thereof in an implementation.  Copyright is not
// released for non MPEG-4 Systems conforming products.  The original
// developer retains full right to use the code for his/her own purpose,
// assign or donate the code to a third party and to inhibit third parties
// from using the code for non MPEG-4 Systems conforming products.  This
// copyright notice must be included in all copies or derivative works.
//
// Copyright 1997, 1998.

#include "OpenGl/OpenGl.h"
#include "OpenGLVisualRenderer.h"
#include "OpenGLPresenter.h"
#include "OpenGLApplication.h"
#include "NodeProxy.h"
#include "Node3DProxy.h"
#include "Matrix.h"
#include "opengl/bindable/FogProxy.h"
#include "opengl/bindable/BackgroundProxy.h"
#include "opengl/sensors/TouchSensorProxy.h"
#include "opengl/group/AnchorProxy.h"
#ifdef PRE_FDIS
#include "opengl/group/Anchor2DProxy.h"
#endif // PRE_FDIS
#include "opengl/sensors/DragSensorProxy.h"
#include "opengl/sensors/TimeSensorProxy.h"

// Stefano Battista 98/11/12
/***
#include "group/Root.h"
#include "bindable/Viewpoint.h"
#include "bindable/Fog.h"
#include "bindable/NavigationInfo.h"
***/

#ifdef MMI
#include "opengl/base/SelectList.h"
// #include "group/Transform.h"
#endif

#if defined(GLUT)
#include <GL/glut.h>
#endif
//#include <GL/glu.h>

#ifndef WIN32
#include <sys/time.h>
#endif

// Constructor

OpenGLVisualRenderer::OpenGLVisualRenderer(OpenGLPresenter* pPresenter)
: VisualRenderer(pPresenter),
m_pPresenter(pPresenter),
m_fFrameRate(25.0f),
m_nLastTime(-1),
m_fFrames(0.0f),
m_bCollisionDetection(FALSE),
m_nDrawSceneGraph(0),
m_bDrawBBoxes(FALSE),
m_bClearScreen(TRUE),
m_bZBuffer(TRUE),
m_nMouseButtons(0),
m_nMouseX(-1),
m_nMouseY(-1),
m_nMouseFirstX(0),
m_nMouseFirstY(0),
#if defined(FASTER_PUSSYCAT)
m_nMinFunc(GL_NEAREST),
m_nMipmapMinFunc(GL_NEAREST_MIPMAP_NEAREST),
m_nMagFunc(GL_NEAREST),
#else
m_nMinFunc(GL_LINEAR),
m_nMipmapMinFunc(GL_LINEAR_MIPMAP_LINEAR),
m_nMagFunc(GL_LINEAR),
#endif
m_dAspectRatio(1.0f),
m_dNearClipping(0.5f),
m_dFarClipping(400.0f),
m_nViewportWidth(100),
m_nViewportHeight(100),
m_pCurrentTexture(NULL),
m_pCurrentMaterial(NULL),
m_pActiveNode(NULL),
#ifdef MMI
m_pHeadSelectedNode(NULL),
m_cMMovePoint(0),
m_pTailSelectedNode(NULL),
m_cPoint(0),
m_bMovement(FALSE),
m_nNumberofMovements(0),
m_bSelection(FALSE),
#endif
m_nStereoDisplayList(0),
m_nActiveFont(-1)
{
  m_bColorMatrixExtension = IsExtensionSupported("GL_SGI_color_matrix");
  ::sscanf((const char*) ::glGetString(GL_VERSION), "%f", &m_fOpenGLVersion);
  
  SetFontFamily("SERIF", FALSE, FALSE);
}

// Find whether opengl extension is supported

BOOL
OpenGLVisualRenderer::IsExtensionSupported(const char* sExtension) const
{
  const char* ext = (const char*) ::glGetString(GL_EXTENSIONS);
  
  if (ext) {
    int len = ::strlen(sExtension);
    
    if (len <= 0)
      return 0;
    
    for (;;) {
      const char* next = ::strchr(ext, ' ');
      
      if (next) {
        if (next - ext == len &&
          ::strncmp(ext, sExtension, len) == 0)
          return TRUE;
      } else
        // end of extension list
        
        return ::strcmp(ext, sExtension) == 0;
      
      ext = next + 1;
    }
  } else
    COMPLAIN("glGetString() returned 0; probably called outside GL context\n");
  
  return FALSE;
}

// Terminate

/* virtual */ void
OpenGLVisualRenderer::Terminate()
{
  //  Stop();
  
  m_pPresenter->GetExecutive()->LockScene();
  m_pPresenter->GetExecutive()->UnlockScene();
}

// Paint

/* virtual */ void
OpenGLVisualRenderer::Paint(RECT* pRect)
{
  m_mtxRendererActive.P();
  
  Executive* pExecutive = m_pPresenter->GetExecutive();
  
  pExecutive->LockScene();
  
  //ML - added this next line, as pScene is also pRoot
  Root* pRoot = m_pPresenter->GetScene();
  MediaObject* pScene = pRoot;

  if (pScene != NULL)
  { // >> (active scene)

    MediaObjectProxy * pSceneRenderer = (MediaObjectProxy*) pScene->GetRenderingObject();
	//ML - removed this next line, as pScene is also pRoot
    //Root * pRoot = pScene->GetRoot();
    Viewpoint * pActiveViewpoint = pRoot->GetBoundViewpoint();
    Fog * pActiveFog = pRoot->GetBoundFog();
    Background * pActiveBackground = pRoot->GetBoundBackground();
    
    // Update time sensors
    TimeSensorProxy::UpdateTimeSensors();

    // SB 99/05/14
#ifdef _DEBUG
    // pSceneRenderer->DumpTrace();
#endif // _DEBUG

    Drag();
    Select(m_nMouseButtons, m_nMouseButtons);
    
    if (m_bZBuffer)
      ::glEnable(GL_DEPTH_TEST);
    else
      ::glDisable(GL_DEPTH_TEST);
    
    if (pActiveViewpoint != NULL)
    { // >> (active viewpoint)
      ViewpointProxy * poViewpointProxy = (ViewpointProxy *) pActiveViewpoint->GetRenderingObject();

      { // >> (compute new viewpoint, checking mouse buttons)
        
        Vec3f position(poViewpointProxy->GetPosition());
        Rotation orientation(poViewpointProxy->GetOrientation());
        
        Vec3f oldPosition = position;
        Rotation oldOrientation = orientation;

        // TRACE("[RendVideo::Paint] MouseButtons %04x \n", m_nMouseButtons);
        if (m_nMouseButtons)
        { // >> (some mouse button)
          Vec3f move;
          float theta = ::cos(orientation.theta * 0.5f);
          Vec3f dir(orientation.x, orientation.y, orientation.z);
          float sw;
          
          // normalise dir
          dir.Normalize();
          
          sw = ::sin(orientation.theta * 0.5f);
          dir.x *= sw;
          dir.y *= sw;
          dir.z *= sw;
          
          if (((m_nMouseButtons & LEFT_BUTTON) &&
            (m_nMouseButtons & RIGHT_BUTTON)) ||
            (m_nMouseButtons & MIDDLE_BUTTON))
          { // >> (2 muose buttons)
            {
              float omega = 5.0f * float(m_nMouseFirstX - m_nMouseX) / float(m_nViewportWidth);
              
              if (omega != 0.0f)
              {
                float s1 =        ::cos(omega * M_PI / 360.0f);
                float sin_angle = ::sin(omega * M_PI / 360.0f);
                float aa = dir.x * dir.x,                     
                  ac = dir.x * dir.z, 
                  ad = dir.x * theta,
                  bb = dir.y * dir.y, 
                  bc = dir.y * dir.z, 
                  bd = dir.y * theta,
                  cc = dir.z * dir.z, 
                  dd = theta * theta;
                
                Vec3f v1(2.0 * (ac + bd) * sin_angle, 2.0 * (bc - ad) * sin_angle, (- aa - bb + cc + dd) * sin_angle);
                Vec3f v2 = dir;
                
                dir.x = s1 * v2.x + theta * v1.x + v1.y * v2.z - v1.z * v2.y;
                dir.y = s1 * v2.y + theta * v1.y + v1.z * v2.x - v1.x * v2.z;
                dir.z = s1 * v2.z + theta * v1.z + v1.x * v2.y - v1.y * v2.x;
                theta = theta * s1 - (v1.x * v2.x + v1.y * v2.y + v1.z * v2.z);
              }
            }
            
            {
              float omega = 5.0f * float(m_nMouseFirstY - m_nMouseY) / float(m_nViewportHeight);
              
              if (omega != 0.0f)
              { 
                float s1 =        ::cos(omega * M_PI / 360.0f);
                float sin_angle = ::sin(omega * M_PI / 360.0f);
                float aa = dir.x * dir.x, 
                  ab = dir.x * dir.y, 
                  ac = dir.x * dir.z,                    
                  bb = dir.y * dir.y,                    
                  bd = dir.y * theta,	     
                  cc = dir.z * dir.z,
                  cd = dir.z * theta,
                  dd = theta * theta;
                
                Vec3f v1((aa - bb - cc + dd) * sin_angle,  2.0 * (ab + cd) * sin_angle, 2.0 * (ac - bd) * sin_angle);
                Vec3f v2 = dir;
                
                dir.x = s1 * v2.x + theta * v1.x + v1.y * v2.z - v1.z * v2.y;
                dir.y = s1 * v2.y + theta * v1.y + v1.z * v2.x - v1.x * v2.z;
                dir.z = s1 * v2.z + theta * v1.z + v1.x * v2.y - v1.y * v2.x;
                theta = theta * s1 - (v1.x * v2.x + v1.y * v2.y + v1.z * v2.z);
              }
            }
          } // << (2 muose buttons)
          else
          { // >> (only 1 muose button)
            if (m_nMouseButtons & LEFT_BUTTON)
            {
              float omega = 5.0f * float(m_nMouseFirstX - m_nMouseX) / float(m_nViewportWidth);
              
              if (omega != 0.0)
              {
                float aa = dir.x * dir.x, 
                  ab = dir.x * dir.y,
                  ad = dir.x * theta,
                  bb = dir.y * dir.y, 
                  bc = dir.y * dir.z,
                  cc = dir.z * dir.z, 
                  cd = dir.z * theta,
                  dd = theta * theta;
                
                float s1 =        ::cos(omega * M_PI / 360.0f);
                float sin_angle = ::sin(omega * M_PI / 360.0f);
                Vec3f v1(2.0 * (ab - cd) * sin_angle, (- aa + bb - cc + dd) * sin_angle,  2.0 * (bc + ad) * sin_angle);
                Vec3f v2 = dir;
                float s2 = theta;
                
                dir.x = s1 * v2.x + s2 * v1.x + v1.y * v2.z - v1.z * v2.y;
                dir.y = s1 * v2.y + s2 * v1.y + v1.z * v2.x - v1.x * v2.z;
                dir.z = s1 * v2.z + s2 * v1.z + v1.x * v2.y - v1.y * v2.x;
                theta = s2 * s1 - (v1.x * v2.x + v1.y * v2.y + v1.z * v2.z);
              }
              
              move.z = - float(m_nMouseFirstY - m_nMouseY) / float(m_nViewportHeight);    
            }
            else
            {
              if (m_nMouseButtons & RIGHT_BUTTON)
              {
                move.x =   float(m_nMouseX - m_nMouseFirstX) / float(m_nViewportWidth);
                move.y = - float(m_nMouseY - m_nMouseFirstY) / float(m_nViewportHeight);
              }
            }
          } // << (only 1 muose button)
          
          {  
            orientation.x = dir.x;
            orientation.y = dir.y;
            orientation.z = dir.z;
            orientation.theta = 2.0 * ::acos(theta);
            NavigationInfo* pNavigationInfo = pRoot->GetBoundNavigationInfo();
            float speed = (pNavigationInfo ? ((NavigationInfo *) pNavigationInfo->GetRenderingObject())->speed /* GetSpeed() */ : 1.0f) / m_fFrameRate;
            
            orientation.Normalize();
            
            float aa = dir.x * dir.x, ab = dir.x * dir.y, ac = dir.x * dir.z, ad = dir.x * theta;
            float                     bb = dir.y * dir.y, bc = dir.y * dir.z, bd = dir.y * theta;
            float                                         cc = dir.z * dir.z, cd = dir.z * theta;
            float                                                             dd = theta * theta;
            
            float T[3][3] = {{aa - bb - cc + dd,     2.0 * (ab - cd),      2.0 * (ac + bd)},
            {  2.0 * (ab + cd), - aa + bb - cc + dd,      2.0 * (bc - ad)},
            {  2.0 * (ac - bd),     2.0 * (bc + ad), - aa - bb + cc + dd}};
            
            float t[3];
            
            for (int j = 0; j < 3; j++)
              t[j] = move.x * T[j][0] + move.y * T[j][1] + move.z * T[j][2];
            
            position.x += t[0] * speed;
            position.y += t[1] * speed;
            position.z += t[2] * speed;
          }
          
          pActiveViewpoint->position = /* SetPosition */ position;
          pActiveViewpoint->orientation = /* SetOrientation */ orientation;
          
          if (m_bCollisionDetection)
          {
            /* Node */ MediaObject * pNode;
            
            // Do collision detection
            pNode = poViewpointProxy->DetectCollision();
            if (pNode) {
              COMPLAIN("Collision\n");
              
              pActiveViewpoint->position = /* SetPosition */ oldPosition;
              pActiveViewpoint->orientation = /* SetOrientation */ oldOrientation;
            }
          }
        }  // << (some mouse button)
      } // << (compute new viewpoint, checking mouse buttons)
      
      Matrix transform;
      
      // load projection matrix from viewpoint
      
      ::glMatrixMode(GL_PROJECTION);
      
      ::glLoadIdentity();
      
      ::gluPerspective(pActiveViewpoint->fieldOfView /* GetFieldOfView() */ * 180.0f / M_PI,
        m_dAspectRatio,
        m_dNearClipping, m_dFarClipping);
      
      ::glMatrixMode(GL_MODELVIEW);
      
      ::glLoadIdentity();
      
      {
        float xmax, ymax;
        
        ymax = 0.5f * ::tan(pActiveViewpoint->fieldOfView /* GetFieldOfView() */ * 0.5f);
        xmax = ymax * m_dAspectRatio;
        
        Vec3f ll(-xmax, -ymax, -0.5f);
        Vec3f ul(-xmax,  ymax, -0.5f);
        Vec3f lr( xmax, -ymax, -0.5f);
        Vec3f ur( xmax,  ymax, -0.5f);
        
        m_leftFrustumPlane =   ul;
        m_rightFrustumPlane =  lr;
        m_topFrustumPlane =    ur;
        m_bottomFrustumPlane = ll;
        
        m_leftFrustumPlane.Cross(ll);
        m_rightFrustumPlane.Cross(ur);
        m_topFrustumPlane.Cross(ul);
        m_bottomFrustumPlane.Cross(lr);
      }
      
      // render background, if we have any
      
      if (pActiveBackground)
        ((BackgroundProxy*) pActiveBackground->GetRenderingObject())->RenderBackground(this, m_pPresenter->GetTime());
      
      // enable fog, if we have any
      
      if (pActiveFog)
        ((FogProxy*) pActiveFog->GetRenderingObject())->Enable(this, m_pPresenter->GetTime());
      
      if (!pActiveBackground)
      {
        // clear screen
        if (m_bClearScreen)
        {
          if (m_bZBuffer)
            ::glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
          else
            ::glClear(GL_COLOR_BUFFER_BIT);
        }
      }
      
      // set viewpoint to world transformation
      
      poViewpointProxy->ViewpointToWorldMatrix(m_pPresenter->GetTime(), transform);
      ::glLoadMatrixf(transform);
      
      // activate lights
      
      m_nNextGLLight = GL_LIGHT0;
      pSceneRenderer->RenderLights(this, m_pPresenter->GetTime());
      
      // did we get any lights?
      
      if (m_nNextGLLight > GL_LIGHT0)
      {
        CullFlag cullFlag = PARTIALLY_INSIDE;
        
        // render scene
        
        pSceneRenderer->CullOrRender(transform, cullFlag,
          this, m_pPresenter->GetTime()); 
        // StereoPaint(m_pPresenter->GetTime());
        
        // disable lights
        
        for (int i = GL_LIGHT0; i < m_nNextGLLight; ++i) {
          ::glDisable((GLenum) i);
        }
      } 
      else
      {
        CullFlag cullFlag = PARTIALLY_INSIDE;
        
        // render scene
        
        pSceneRenderer->CullOrRender(transform, cullFlag,
          this, m_pPresenter->GetTime()); 
        // StereoPaint(m_pPresenter->GetTime());
      }
      
      // disable fog, if we have any
      
      if (pActiveFog)
        ((FogProxy*) pActiveFog->GetRenderingObject())->Disable(this, m_pPresenter->GetTime());
    } // << (active viewpoint)
    else
    { // >> (NO active viewpoint)
      Matrix transform;
      
      // we don't have a viewpoint; use an orthogonal viewpoint instead
      
      ::glMatrixMode(GL_PROJECTION);
      ::glLoadIdentity();
      
      ::glOrtho(-m_nViewportWidth / 2, m_nViewportWidth / 2,
        -m_nViewportHeight / 2, m_nViewportHeight / 2,
        0, 512); 
      
      ::glMatrixMode(GL_MODELVIEW);
      
      ::glLoadIdentity();
      
      m_leftFrustumPlane = Vec3f(-m_nViewportWidth / 2, 0.0f, 0.0f);
      m_rightFrustumPlane = Vec3f(m_nViewportWidth / 2, 0.0f, 0.0f);
      m_topFrustumPlane = Vec3f(0.0f, m_nViewportHeight / 2, 0.0f);
      m_bottomFrustumPlane = Vec3f(0.0f, -m_nViewportHeight / 2, 0.0f);
      
      if (m_bClearScreen)
      {
        if (m_bZBuffer)
          ::glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        else
          ::glClear(GL_COLOR_BUFFER_BIT);
      } 
      // activate lights
      
      m_nNextGLLight = GL_LIGHT0;
      pSceneRenderer->RenderLights(this, m_pPresenter->GetTime());
      
      // did we get any lights?
      if (m_nNextGLLight > GL_LIGHT0)
      {
        CullFlag cullFlag = COMPLETELY_INSIDE;
        
        // render scene
        
        pSceneRenderer->CullOrRender(transform, cullFlag,
          this, m_pPresenter->GetTime());
        
        // disable lights
        
        for (int i = GL_LIGHT0; i < m_nNextGLLight; ++i)
          ::glDisable((GLenum) i);
      } 
      else
      {
        CullFlag cullFlag = COMPLETELY_INSIDE;
        
        // render scene
        
        pSceneRenderer->CullOrRender(transform, cullFlag,
          this, m_pPresenter->GetTime());
      }
    } // << (NO active viewpoint)
    
    if (m_nDrawSceneGraph)
    {
      ::glMatrixMode(GL_PROJECTION);
      ::glLoadIdentity();
      
      ::glOrtho(0.0, 1.0,
        -9.0, 1.0,
        -1.0, 1.0);
      
      ::glMatrixMode(GL_MODELVIEW);
      
      ::glLoadIdentity();
      
      ::glPushAttrib(GL_ENABLE_BIT);
      
      ::glDisable(GL_ALPHA_TEST);
      ::glDisable(GL_DEPTH_TEST);
      ::glEnable(GL_BLEND);
      ::glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      
      pSceneRenderer->RenderDescription(this, m_pPresenter->GetTime());
      
      ::glPopAttrib();
    }

#ifdef MESA_3DFX
    {
      ::glMatrixMode(GL_PROJECTION);
      ::glLoadIdentity();
      
      ::glOrtho(0.0, m_nViewportWidth,
        m_nViewportHeight, 0,
        -1.0, 1.0);
      
      ::glMatrixMode(GL_MODELVIEW);
      
      ::glLoadIdentity();
      
      ::glPushAttrib(GL_ENABLE_BIT);
      
      ::glDisable(GL_ALPHA_TEST);
      ::glDisable(GL_DEPTH_TEST);
      
      ::glBegin(GL_LINES);
      ::glVertex2f(m_nMouseX - 2, m_nMouseY);
      ::glVertex2f(m_nMouseX + 1, m_nMouseY);
      ::glVertex2f(m_nMouseX, m_nMouseY - 1);
      ::glVertex2f(m_nMouseX, m_nMouseY + 2);
      ::glEnd();
      
      ::glPopAttrib();
    }
#endif
  } // << (active scene)
  else
  { // >> (NO active scene)
    ::glClearColor(0.0f, 0.0f, 0.0f, 0.0f); 
    ::glClear(GL_COLOR_BUFFER_BIT);
    
    // COMPLAIN("OpenGLVisualRenderer::Paint(): no scene?\n");
  } // << (NO active scene)
  
  pExecutive->UnlockScene();
  
  {
    unsigned long now;
    
#ifndef WIN32
    struct timeval time;
    
    ::gettimeofday(&time, NULL);
    now = time.tv_sec * 1000 + time.tv_usec / 1000;
#else
    now = timeGetTime();
#endif
    
    if (m_nLastTime == -1)
      m_nLastTime = now;
    else {
      m_fFrames++;
      
      if (m_nLastTime + 500 < (int) now) {
        m_fFrameRate = m_fFrames * 1000.0f / float(now - m_nLastTime);
#ifdef _DEBUG
        // COMPLAIN("current frame rate: %.2f per second\n", m_fFrameRate);
#endif
        m_nLastTime = now;
        m_fFrames = 0.0f;
      }
    }
  }
  
#if defined(GLUT)
  ::glutSwapBuffers();
#elif defined(WIN32)
  ::glFinish();
  ::SwapBuffers(wglGetCurrentDC());
#endif
  
#ifdef _DEBUG
  {
    GLenum error;
    
    error = ::glGetError();
    
    while (error != GL_NO_ERROR) {
      TRACE("GL error: %d\n", error);
      error = ::glGetError();
    }
  }
#endif
  
  VisualRenderer::Paint(pRect);
  
  m_mtxRendererActive.V();
}

// Set mouse state

/* virtual */ void
OpenGLVisualRenderer::MouseState(int nButton, int nX, int nY)
{
  int oldButton = m_nMouseButtons;
  
  if (nButton & ~m_nMouseButtons) {
    m_nMouseFirstX = nX;
    m_nMouseFirstY = nY;
  }
  
  m_nMouseButtons = nButton;
  m_nMouseX = nX;
  m_nMouseY = nY;
  
  // do Select() only on mouse up/down
  
  if ((nButton & RIGHT_BUTTON) !=
    (oldButton & RIGHT_BUTTON)) {
    m_mtxRendererActive.P();
    
    Executive* pExecutive = m_pPresenter->GetExecutive();
    
    pExecutive->LockScene();
    Select(nButton, oldButton);
    pExecutive->UnlockScene();
    
    m_mtxRendererActive.V();
  }
}

// Render string

void
OpenGLVisualRenderer::RenderString(const char* pString)
{
#if defined(GLTT)
  if (m_nActiveFont >= 0) {
    m_generatedFonts[m_nActiveFont]->font->output(pString);
    return;
  }
  
  // By not making GLTT and GLUT font rendering mutually exclusive, we can
  // fallback to GLUT font rendering if GLTT font rendering fails
  
#endif
#if defined(GLUT)
  for (int i = 0; pString[i]; ++i)
    ::glutStrokeCharacter(GLUT_STROKE_ROMAN, pString[i]);
#endif
#if defined(WIN32)
  if (m_nActiveFont >= 0) {
    ::glListBase(m_generatedFonts[m_nActiveFont]->displayListBase);
    
    ::glPushAttrib(GL_POLYGON_BIT);
    ::glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    ::glCallLists(::strlen(pString), GL_UNSIGNED_BYTE, (const GLvoid*) pString);
    ::glPopAttrib();
    
    ::glListBase(0);
  }
  
#ifdef _DEBUG
  {
    GLenum error;
    
    error = ::glGetError();
    
    while (error != GL_NO_ERROR) {
      TRACE("OpenGLVisualRenderer::RenderString(): GL error %d\n", error);
      error = ::glGetError();
    }
  }
#endif
#endif
}

// Get width of string

/* virtual */ float
OpenGLVisualRenderer::GetStringWidth(const char* pString)
{
#if defined(GLTT)
  if (m_nActiveFont >= 0)
    return m_generatedFonts[m_nActiveFont]->font->getWidth(pString);
  
  // By not making GLTT and GLUT font rendering mutually exclusive, we can
  // fallback to GLUT font rendering if GLTT font rendering fails
  
#endif
#if defined(GLUT)
  float width = 0.0f;
  
  for (int i = 0; pString[i]; ++i)
    width += ::glutStrokeWidth(GLUT_STROKE_ROMAN, pString[i]);
  
  return width;
#endif  
#if defined(WIN32)
  float width = 0.0f;
  
  if (m_nActiveFont >= 0)
    for (int i = 0; pString[i]; ++i)
      width += m_generatedFonts[m_nActiveFont]->width[pString[i]];
    
    return width;
#endif
}

// Get height of string

/* virtual */ float
OpenGLVisualRenderer::GetFontHeight()
{
#if defined(GLTT)
  if (m_nActiveFont >= 0)
    return m_generatedFonts[m_nActiveFont]->height;
  
  // By not making GLTT and GLUT font rendering mutually exclusive, we can
  // fallback to GLUT font rendering if GLTT font rendering fails
  
#endif
#if defined(GLUT)
  return 152.38;
#endif
#if defined(WIN32)
  if (m_nActiveFont >= 0)
    return m_generatedFonts[m_nActiveFont]->height;
#endif
  
  return 0;
}

// Get baseline of string

float
OpenGLVisualRenderer::GetFontBaseline()
{
#if defined(GLTT)
  if (m_nActiveFont >= 0)
    return m_generatedFonts[m_nActiveFont]->baseline;
#endif
#if defined(GLUT)
  return 33.33f;
#endif
#if defined(WIN32)
  return 0.0f;
#endif  
}

/* virtual */ BOOL
OpenGLVisualRenderer::SetFontFamily(const char* pFontname, BOOL bBold, BOOL bItalic)
{
#if defined(GLTT) || defined(WIN32)
  int fontNumber = 0;
  
  while (fontNumber < m_generatedFonts.GetLength())
    if (m_generatedFonts[fontNumber]->italic == bItalic &&
      m_generatedFonts[fontNumber]->bold == bBold &&
      ::strcmp(m_generatedFonts[fontNumber]->fontname, pFontname) == 0) {
      m_nActiveFont = fontNumber;
      
      return TRUE;
    } else
      ++fontNumber;
#endif
    
#if defined(GLTT)
    
    FTFace* pFace = new FTFace;
    GLTTFont* pFont;
    ZString fontname(pFontname);
    
    if (::strcmp(pFontname, "SERIF") == 0)
      fontname = "times";
    else
      if (::strcmp(pFontname, "SANS") == 0)
        fontname = "arial";
      else
        if (::strcmp(pFontname, "TYPEWRITER") == 0)
          fontname = "cour";
        
        if (bBold && !bItalic)
          fontname += "bd";
        else {
          if (bBold)
            fontname += "b";
          if (bItalic)
            fontname += "i";
        }
        
        fontname += ".ttf";
        
        if (pFace)
          if (pFace->open(fontname)) {
            pFont = new GLTTFont(pFace);
            
            if (pFont) {
              pFont->setPrecision(25);
              
              if (pFont->create(240)) {
                GeneratedFont* pGeneratedFont = new GeneratedFont;
                
                pFont->load();
                
                if (pGeneratedFont) {
                  pGeneratedFont->fontname = pFontname;
                  pGeneratedFont->face = pFace;
                  pGeneratedFont->font = pFont;
                  pGeneratedFont->bold = bBold;
                  pGeneratedFont->italic = bItalic;
                  pGeneratedFont->height = pFont->getHeight();
                  pGeneratedFont->baseline = -pFont->getDescender();
                  
                  m_generatedFonts += pGeneratedFont;
                  
                  m_nActiveFont = fontNumber;
                  
                  return TRUE;
                } else
                  COMPLAIN("Out of memory creating font; cannot do font rendering\n");
              } else
                COMPLAIN("Failed to create outline font; cannot do font rendering\n");
              
              delete pFont;
            } else
              COMPLAIN("Failed to create font; cannot do font rendering\n");
          }
          
          delete pFace;
          
          return FALSE;
          
#endif
#if defined(GLUT)
          return TRUE;
#endif
#ifdef WIN32
          CWnd* mainWnd = AfxGetMainWnd();
          CDC* pDC = mainWnd->GetDC();
          
          if (pDC == NULL ||
            pFontname == NULL)
            return FALSE;
          
          LOGFONT logfont;
          GLYPHMETRICSFLOAT gmf[256];
          
          // lfHeight can't be used to change the font size
          logfont.lfHeight		= -12; // use glScale to change size
          logfont.lfWidth		= 0;
          logfont.lfEscapement		= 0;
          logfont.lfOrientation		= logfont.lfEscapement;
          logfont.lfWeight		= bBold ? 700 : 400;
          logfont.lfItalic		= bItalic;
          logfont.lfUnderline		= FALSE;
          logfont.lfStrikeOut		= FALSE;
          logfont.lfCharSet		= ANSI_CHARSET;
          logfont.lfOutPrecision	= OUT_TT_ONLY_PRECIS;
          logfont.lfClipPrecision	= CLIP_DEFAULT_PRECIS;
          logfont.lfQuality		= DEFAULT_QUALITY;
          logfont.lfPitchAndFamily      = FF_DONTCARE|DEFAULT_PITCH;
          
          if (::strcmp(pFontname, "SERIF") == 0)
            ::lstrcpy(logfont.lfFaceName, "Times New Roman");
          else
            if (::strcmp(pFontname, "SANS") == 0)
              ::lstrcpy(logfont.lfFaceName, "Arial");
            else
              if (::strcmp(pFontname, "TYPEWRITER") == 0)
                ::lstrcpy(logfont.lfFaceName, "Courier New");
              else
                ::lstrcpy(logfont.lfFaceName, pFontname);
              
              CFont newfont;
              
              // returns 0 if it fails
              
              BOOL success = newfont.CreateFontIndirect(&logfont);
              CFont* oldfont = pDC->SelectObject(&newfont);
              GeneratedFont* pGeneratedFont;
              
              ASSERT(0 != oldfont);
              
              // Create a set of display lists based on the glyphs of the TrueType font 
              // notice that we really waste the first 32 spaces....
              // if there's a problem delete the display lists
              
              if (success == FALSE) {
                TRACE("CreateFontIndirect failed: probably couldn't find desired font '%s', using default font\n", logfont.lfFaceName);
                
                return FALSE;
              }
              
              pGeneratedFont = new GeneratedFont();
              
              if (pGeneratedFont) {
                pGeneratedFont->displayListBase = ::glGenLists(256);
                
                if (pGeneratedFont->displayListBase &&
                  ::wglUseFontOutlines(pDC->m_hDC, 0, 256, pGeneratedFont->displayListBase, 1.0f, 0.0f, WGL_FONT_POLYGONS, gmf)) {
                  float height = 0.0f;
                  
                  pDC->SelectObject(oldfont);
                  
                  pGeneratedFont->fontname = pFontname;
                  pGeneratedFont->bold = bBold;
                  pGeneratedFont->italic = bItalic;
                  
                  for (int textdimension = 0; textdimension < 256; ++textdimension) {
                    pGeneratedFont->width[textdimension] = gmf[textdimension].gmfBlackBoxX;
                    
                    if (height < gmf[textdimension].gmfBlackBoxY)
                      height = gmf[textdimension].gmfBlackBoxY;
                  }
                  
                  pGeneratedFont->height = height;
                  
                  m_generatedFonts += pGeneratedFont;
                  
                  m_nActiveFont = fontNumber;
                  
                  return TRUE;
                } else {
                  TRACE("Couldn't generate display lists for font '%s', using default font\n", logfont.lfFaceName);
                  
                  ::glDeleteLists(pGeneratedFont->displayListBase, 256);
                  delete pGeneratedFont;
                }
              }
              
              return FALSE;
#endif
}

// Drag object

/* virtual */ void
OpenGLVisualRenderer::Drag()
{
  for (int i = 0; i < m_pActiveDragSensors.GetLength(); ++i)
    m_pActiveDragSensors[i]->Drag(this, m_nMouseX, m_nMouseY);
}

// Select object

/* virtual */ int OpenGLVisualRenderer::Select(
  int nNewButtons, int nOldButtons)
{
  int pickObject = 0;
  
  //ML - changed MediaObject* pScene to Root* pRoot
  //MediaObject* pScene = m_pPresenter->GetScene();
  Root* pRoot = m_pPresenter->GetScene();

  if (pRoot)
  {
    ZArray<MediaObjectProxy*> activatedSensors;
    MediaObjectProxy* intersectingObject = NULL;
    GLdouble objX = 0.0, objY = 0.0, objZ = 0.0;
    
    // if mouse coords are "valid" >>>>
    if (0 <= m_nMouseX && m_nMouseX < m_nViewportWidth &&
      0 <= m_nMouseY && m_nMouseY < m_nViewportHeight)
    {
      Viewpoint* pActiveViewpoint = pRoot->GetBoundViewpoint();
      ViewpointProxy * poViewpointProxy;
      if (pActiveViewpoint != NULL)
        poViewpointProxy = (ViewpointProxy *) pActiveViewpoint->GetRenderingObject();
      else
        poViewpointProxy = NULL;
      MediaObjectProxy* pSceneRenderer = (MediaObjectProxy*) pRoot->GetRenderingObject();
      GLint viewport[] = { 0, 0, m_nViewportWidth, m_nViewportHeight };
      
      ZArray<MediaObjectProxy*> dummyArray;
      unsigned int depth = UINT_MAX;
      Matrix transform;
      
      ::glMatrixMode(GL_PROJECTION);
      ::glLoadIdentity();
      
      ::gluPickMatrix(double(m_nMouseX), double(viewport[3] - m_nMouseY), 1.0, 1.0, viewport);
      
      if (pActiveViewpoint)
      {
        CullFlag cullFlag = PARTIALLY_INSIDE;
        
        // load projection matrix from viewpoint
        
        ::gluPerspective(pActiveViewpoint->fieldOfView /* GetFieldOfView() */ * 180.0f / M_PI,
          m_dAspectRatio,
          m_dNearClipping, m_dFarClipping);
        
        ::glMatrixMode(GL_MODELVIEW);
        
        poViewpointProxy->ViewpointToWorldMatrix(m_pPresenter->GetTime(), transform);
        ::glLoadMatrixf(transform);
        
        // check for intersection
        
        ((MediaObjectProxy*) pSceneRenderer)->
        // ((NodeProxy *) pSceneRenderer)->
            CullOrSelect(transform, cullFlag,
						    this, m_pPresenter->GetTime(),
                intersectingObject, activatedSensors, depth, dummyArray);
      }
      else
      {
        CullFlag cullFlag = COMPLETELY_INSIDE;
        
        // we don't have a viewpoint; use an orthogonal viewpoint instead
        
        ::glOrtho(-m_nViewportWidth / 2, m_nViewportWidth / 2,
          -m_nViewportHeight / 2, m_nViewportHeight / 2,
          -256, 256);
        
        ::glMatrixMode(GL_MODELVIEW);
        
        ::glLoadIdentity();
        
        // check for intersection
        
        ((MediaObjectProxy*) pSceneRenderer)->
        // ((NodeProxy *) pSceneRenderer)->
            CullOrSelect(transform, cullFlag,
						    this, m_pPresenter->GetTime(),
                intersectingObject, activatedSensors, depth, dummyArray);
      }
      
      if (intersectingObject)
      {
        GLdouble projection[16];
        GLdouble modelview[16];
        GLint viewport[4];
        
        ::glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
        ::glGetDoublev(GL_PROJECTION_MATRIX, projection);
        ::glGetIntegerv(GL_VIEWPORT, viewport);
        
        if (!::gluUnProject(double(m_nMouseX), double(m_nMouseY), double(depth) / 4294967295.0,
          modelview,
          projection,
          viewport,
          &objX, &objY, &objZ))
          COMPLAIN("gluUnProject() failed?!\n");
      }
    }
    // if mouse coords are "valid" <<<<
    
#ifdef MMI    
    if (intersectingObject) {
      pickObject = 1;
      
      if (m_pHeadSelectedNode == NULL) {
        m_pHeadSelectedNode = new SelectedNode;
        m_pTailSelectedNode = new SelectedNode;
        m_pHeadSelectedNode->InsertAfterNode(m_pTailSelectedNode);
        m_nLastUndo = 0;
      }
      
      if (m_pHeadSelectedNode->next != NULL)
        if (m_pHeadSelectedNode->next->m_nMoveNumber >= 9) {
          for (SelectedNode* i = m_pTailSelectedNode->prev; i->m_nMoveNumber == m_nLastUndo; i = m_pTailSelectedNode->prev) {
            i->FreeSelectedNode();	
            delete i;
          }
          
          m_nLastUndo++;
        }		
        
        SelectedNode* newSelectedNode = new SelectedNode;
        
        newSelectedNode-> selectedObject = (Group*)intersectingObject->GetMediaObject();
        newSelectedNode->m_nMoveNumber = m_nNumberofMovements;
        m_pHeadSelectedNode->InsertAfterNode(newSelectedNode);
    }
#endif
    
    {
      int i;
      
      /* First, if any sensors are already active, and if we have registered
      a mouse click of some sorts, those sensors should be triggered. */
      
      if (!(nNewButtons & RIGHT_BUTTON) &&
        (nOldButtons & RIGHT_BUTTON)) 
      {
        int index;
        
        for (int i = activatedSensors.GetLength() - 1; i >= 0; --i)
        {
          if ((index = m_pActiveTouchSensors.Find((TouchSensorProxy*) activatedSensors[i])) != -1)
          {
            m_pActiveTouchSensors[index]->Trigger(this);
          }
          else
          {
            if ((index = m_pActiveAnchors.Find((AnchorProxy*) activatedSensors[i])) != -1)
              m_pActiveAnchors[index]->Trigger(this);
#ifdef PRE_FDIS
            else
              if ((index = m_pActiveAnchors2D.Find((Anchor2DProxy*) activatedSensors[i])) != -1)
                m_pActiveAnchors2D[index]->Trigger(this);
#endif // PRE_FDIS
          }
        }

        // Deactivate all dragsensors
        while (m_pActiveDragSensors.GetLength())
          m_pActiveDragSensors[0]->Deactivate(this, m_nMouseX, m_nMouseY);
      }
      
      if ((nNewButtons & RIGHT_BUTTON) &&
        !(nOldButtons & RIGHT_BUTTON)) 
      {
        // If we got a mouse down, first deactivate all sensors
        
        while (m_pActiveTouchSensors.GetLength())
          m_pActiveTouchSensors[0]->Deactivate(this, FALSE);
        
        while (m_pActiveAnchors.GetLength())
          m_pActiveAnchors[0]->Deactivate(this, FALSE);
        
#ifdef PRE_FDIS
        while (m_pActiveAnchors2D.GetLength())
          m_pActiveAnchors2D[0]->Deactivate(this, FALSE);
#endif // PRE_FDIS

        // Activate all sensors
        for (i = activatedSensors.GetLength() - 1; i >= 0; --i)
        {
          activatedSensors[i]->Activate(this, objX, objY, objZ, m_nMouseX, m_nMouseY, true);
        }
      } 
      else 
      {
        // Deactivate sensors that we no longer have
        for (i = 0; i < m_pActiveTouchSensors.GetLength();)
        {
          if (activatedSensors.Find(m_pActiveTouchSensors[i]) != -1)
            activatedSensors.DeleteFrom(activatedSensors.Find(m_pActiveTouchSensors[i++]));
          else
            m_pActiveTouchSensors[i]->Deactivate(this, nNewButtons & RIGHT_BUTTON);
        }
        for (i = 0; i < m_pActiveAnchors.GetLength();)
        {
          if (activatedSensors.Find(m_pActiveAnchors[i]) != -1)
            activatedSensors.DeleteFrom(activatedSensors.Find(m_pActiveAnchors[i++]));
          else
            m_pActiveAnchors[i]->Deactivate(this, nNewButtons & RIGHT_BUTTON);
        }
#ifdef PRE_FDIS
        for (i = 0; i < m_pActiveAnchors2D.GetLength();)
        {
          if (activatedSensors.Find(m_pActiveAnchors2D[i]) != -1)
            activatedSensors.DeleteFrom(activatedSensors.Find(m_pActiveAnchors2D[i++]));
          else
            m_pActiveAnchors2D[i]->Deactivate(this, nNewButtons & RIGHT_BUTTON);
        }
#endif // PRE_FDIS
        for (i = 0; i < m_pActiveDragSensors.GetLength(); ++i)
        {
          if (activatedSensors.Find(m_pActiveDragSensors[i]) != -1)
            activatedSensors.DeleteFrom(activatedSensors.Find(m_pActiveDragSensors[i]));
        }

        // Activate new sensors
        for (i = activatedSensors.GetLength() - 1; i >= 0; --i)
        {
          activatedSensors[i]->Activate(this, objX, objY, objZ, m_nMouseX, m_nMouseY, false);
        }
      }
      
      if (m_pActiveTouchSensors.GetLength() ||
        m_pActiveAnchors.GetLength() ||
#ifdef PRE_FDIS
        m_pActiveAnchors2D.GetLength() ||
#endif // PRE_FDIS
        m_pActiveDragSensors.GetLength()) 
        m_pActiveNode = intersectingObject;
      else
        m_pActiveNode = NULL;
    }
  } 
  else
  {
    COMPLAIN("OpenGLVisualRenderer::Select(): no scene?\n");
  }

#ifdef _DEBUG  
  {
    GLenum error;
    
    error = ::glGetError();
    
    while (error != GL_NO_ERROR) {
      TRACE("GL error: %d\n", error);
      error = ::glGetError();
    }
  }
#endif
  
  return pickObject;
}

// Resize the display

/* virtual */ void
OpenGLVisualRenderer::ReSize(int nWidth, int nHeight)
{
  m_nViewportWidth = nWidth;
  m_nViewportHeight = nHeight;
  
  ::glViewport(0, 0, nWidth, nHeight);
  
  m_dAspectRatio = double(nWidth) / double(nHeight);
  
  if (m_nStereoDisplayList) {
    ::glDeleteLists(m_nStereoDisplayList, 1);
    m_nStereoDisplayList = 0;
  }
}

#ifdef _DEBUG

// Check for OpenGL errors

BOOL
OpenGLVisualRenderer::CheckGLError(const char* pString)
{
  GLenum error = ::glGetError();
  
  if (error) {
    ::printf(pString);
    
    do {
      ::printf(" %d", error);
	  error = ::glGetError();
    } while (error);
	//ML - changed this line to remove a warning
    //} while ((error = ::glGetError()));

    ::printf("\n");
    
#ifndef WIN32
    ((OpenGLApplication*) m_pPresenter->GetExecutive()->GetApplication())->Debug();
#endif
    
    return TRUE;
  } else
    return FALSE;
}

#endif

/* Generate stereo perspective matrix.

  This routine does a stereoscopic offset perspective projection for one
  eye's view.  The "center of interest" of the scene, which will be
  rendered at a zero parallax setting, is assumed to be at the coordinate
  system origin (0.0, 0.0, 0.0).
  
    Parameters:
    
      fovy_degrees = field of view angle in vertical dimension, specified
      in degrees
      
        distance = the distance from the virtual cameras to the "center of
        interest" of the scene
        
          left_eye = TRUE or FALSE
          
            znear, zfar = the near and far clipping plane, relative to the
            "center of interest" of the scene
            
              aspect = the aspect ratio of the frustrum that will be set up
              appropriate to the dimensions of the viewport
              
called by redraw()   */

void
OpenGLVisualRenderer::StereoPerspective(float fFovyDegrees, float fDistance, BOOL bLeftEye,
                                        float fZNear, float fZFar, float fAspect)
{
  double left, right, top, bottom;
  double clip_near, clip_far;
  double camera_separation;
  double eyeview;
  double horz_tan, vert_tan, fovy_radians, perspective_offset;
  
  clip_near = fDistance - fZNear;
  clip_far = fDistance - fZFar;
  
  fovy_radians = double(fFovyDegrees) * M_PI / 180.0;
  horz_tan = ::tan(fovy_radians * fAspect / 2.0);
  vert_tan = ::tan(fovy_radians / 2.0);
  
  camera_separation = fDistance * horz_tan / 10.0f;
  eyeview = camera_separation / 2.0;
  if (bLeftEye)
    eyeview = -eyeview;
  
  top = vert_tan * clip_near;
  bottom = -top;
  right = horz_tan * clip_near;
  left = -right;
  
  perspective_offset = eyeview * (clip_near / fDistance);
  right -= perspective_offset;
  left -= perspective_offset;
  
  ::glFrustum(left, right, bottom, top, clip_near, clip_far);
  ::glTranslated(-eyeview, 0.0, -fDistance);
}

// Stereo paint method

void
OpenGLVisualRenderer::StereoPaint(DWORD dwTime)
{
  //ML - changed MediaObject* pScene to Root* pRoot
  //MediaObject * pScene = m_pPresenter->GetScene();
  Root* pRoot = m_pPresenter->GetScene();
  MediaObjectProxy* pSceneRenderer = (MediaObjectProxy*) pRoot->GetRenderingObject();
  Viewpoint* pActiveViewpoint = pRoot->GetBoundViewpoint();
  ViewpointProxy * poViewpointProxy = (ViewpointProxy *) pActiveViewpoint->GetRenderingObject();
  
  GLint displayList;
  
  Matrix dummy;
  CullFlag cullFlag = PARTIALLY_INSIDE;
  
  displayList = ::glGenLists(1);
  
  ::glNewList(displayList, GL_COMPILE);
  pSceneRenderer->CullOrRender(dummy, cullFlag,
    this, dwTime);
  ::glEndList();
  
  {
    Matrix transform;
    
    // load projection matrix from viewpoint
    
    ::glMatrixMode(GL_PROJECTION);
    ::glLoadIdentity();
    StereoPerspective(pActiveViewpoint->fieldOfView /* GetFieldOfView() */ * 180.0f / M_PI,
      0.5f, TRUE,
      -m_dNearClipping, -m_dFarClipping, m_dAspectRatio);
    poViewpointProxy->ViewpointToWorldMatrix(0, transform);
    ::glMultMatrixf(transform);
    
    // render left eye
    
    ::glViewport(0, 0, m_nViewportWidth, m_nViewportHeight / 2);
    ::glMatrixMode(GL_MODELVIEW);
    ::glCallList(displayList);
  }
  
  {  
    Matrix transform;
    
    // load projection matrix from viewpoint
    
    ::glMatrixMode(GL_PROJECTION);
    ::glLoadIdentity();
    
    StereoPerspective(pActiveViewpoint->fieldOfView /* GetFieldOfView() */ * 180.0f / M_PI,
      0.5f, FALSE,
      -m_dNearClipping, -m_dFarClipping, m_dAspectRatio);
    poViewpointProxy->ViewpointToWorldMatrix(0, transform);
    ::glMultMatrixf(transform);
    
    // render right eye
    
    ::glViewport(0, m_nViewportHeight / 2, m_nViewportWidth, (m_nViewportHeight + 1) / 2);
    ::glMatrixMode(GL_MODELVIEW);
    ::glCallList(displayList);
  }  
  
  ::glDeleteLists(displayList, 1);
  
  if (m_nStereoDisplayList)
    ::glCallList(m_nStereoDisplayList);
  else {
    m_nStereoDisplayList = ::glGenLists(1);
    ::glNewList(m_nStereoDisplayList, GL_COMPILE_AND_EXECUTE);
    
    ::glMatrixMode(GL_PROJECTION);
    ::glLoadIdentity();
    ::glOrtho(0.0, m_nViewportWidth, 0.0, m_nViewportHeight, -1.0, 1.0);
    ::glMatrixMode(GL_MODELVIEW);
    ::glLoadIdentity();
    ::glViewport(0, 0, m_nViewportWidth, m_nViewportHeight);
    
    // Disable stuff that's likely to slow down glDrawPixels.
    
    ::glPushAttrib(GL_ENABLE_BIT);
    
    ::glDisable(GL_ALPHA_TEST);
    ::glDisable(GL_DEPTH_TEST);
    ::glDisable(GL_FOG);
    ::glDisable(GL_LIGHTING);
    
#ifdef GL_SGIS_multisample
    ::glDisable(GL_MULTISAMPLE_SGIS);
#endif
    
    {
      int width = m_nViewportWidth;
      int height = m_nViewportHeight;
      
      // textures must have size and width == 2^n
      
      int numshifts;
      
      // align width
      
      numshifts = -1;
      
      for (int i = 0; i < 16; ++i) {
        if (width & 1)
          numshifts = i;
        
        width >>= 1;
      }
      
      width = 1 << numshifts;
      if (width < m_nViewportWidth)
        width <<= 1;
      
      // align height
      
      numshifts = -1;
      
      for (int j = 0; j < 15; ++j) {
        if (height & 1)
          numshifts = j;
        
        height >>= 1;
      }
      
      height = 1 << numshifts;
      if (height < m_nViewportHeight)
        height <<= 1;
      
      ::glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
      ::glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
      ::glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
      
      ::glEnable(GL_TEXTURE_2D);
      
      ::glMatrixMode(GL_TEXTURE);
      ::glLoadIdentity();
      
      if (width != m_nViewportWidth ||
        height != m_nViewportHeight)
        ::glScalef(float(m_nViewportWidth) / float(width),
        float(m_nViewportHeight) / float(height), 1.0f);
      
      ::glMatrixMode(GL_MODELVIEW);
      
#ifndef GL_VERSION_1_1
      ::glCopyTexImage2DEXT(GL_TEXTURE_2D, 0, GL_RGB,
        0, 0,
        width, height,
        0);
#else
      ::glCopyTexImage2D(GL_TEXTURE_2D, 0, GL_RGB,
        0, 0,
        width, height,
        0);
#endif
      
      ::glBegin(GL_LINES);
      
      {
        int swap = 0;
        
        for (int k = 0; k < m_nViewportHeight; ++k) {
          ::glTexCoord2f(0.0f, (swap * 0.5f) + 0.5f * float(k) / m_nViewportHeight);
          ::glVertex2i(0, k);
          ::glTexCoord2f(1.0f, (swap * 0.5f) + 0.5f * float(k) / m_nViewportHeight);
          ::glVertex2i(m_nViewportWidth - 1, k);
          
          swap = !swap;
        }
      }
      
      ::glEnd();
    }
    
    ::glPopAttrib();
    
    ::glEndList();
  }
}

#ifdef MMI

Viewpoint* OpenGLVisualRenderer::GetActiveViewPoint()
{
  //ML - changed MediaObject* pScene to Root* pRoot
  //MediaObject* pScene = m_pPresenter->GetScene();
  Root* pRoot = m_pPresenter->GetScene();
  
  if (!pRoot) return NULL;
  Viewpoint* pActiveViewpoint = pRoot->GetBoundViewpoint();
  return pActiveViewpoint;
  
}

void OpenGLVisualRenderer::SetActiveViewpoint(Viewpoint * pViewpoint)
{
  //ML - changed MediaObject* pScene to Root* pRoot
  //MediaObject* pScene = m_pPresenter->GetScene();
  Root* pRoot = m_pPresenter->GetScene();
  
  if (pRoot) 
    pRoot->PushViewpoint(pViewpoint);
  
}

void OpenGLVisualRenderer::LBDownAction(UINT nFlags, CPoint point)
{
  m_cPoint = point;
  m_nFlags = nFlags;
  m_bSelection = Select((int) point.x, (int) point.y);
}

void OpenGLVisualRenderer::LBUpAction(UINT /* nFlags */, CPoint /* point */)
{
#if 0
  float xvalue,yvalue;
  BOOL success = FALSE;
  m_nFlags = nFlags;
  
  if (m_bSelection && m_pHeadSelectedNode != NULL && m_bMovement) {
    point.x = m_cMMovePoint.x - m_cPoint.x;
    point.y = m_cMMovePoint.y - m_cPoint.y;
    int currentindex = m_pHeadSelectedNode->next->m_nMoveNumber;
    for (SelectedNode* i = m_pHeadSelectedNode->next; i != m_pTailSelectedNode; i = i->next) {
      if (currentindex == i->m_nMoveNumber) {
        m_pParentGroup = i->selectedObject;
        
        const ZArray<Group*>& groups = m_pParentGroup->GetGroupParents();
        
        while (dynamic_cast<Group *>(m_pParentGroup) == NULL)
          for (int pcounter = 0; pcounter < groups.GetLength(); ++pcounter)
            if (dynamic_cast<Group *>(groups[pcounter]) != NULL) {
              m_pParentGroup = groups[pcounter];
              pcounter = groups.GetLength();
            }
            
            Transform*	pUseTransform;
            Group* pParentMObject;
            pUseTransform = new Transform;
            
			//ML - changed MediaObject* pScene to Root* pRoot
            //Root* pRoot = m_pPresenter->GetScene();
            MediaObject** pRoot = m_pPresenter->GetScene();
            Viewpoint* pActiveViewpoint = pRoot->GetBoundViewpoint();
            if (pActiveViewpoint) {
              xvalue = (point.x * m_dAspectRatio) / 100;			
              yvalue = (point.y * -m_dAspectRatio)/ 100;
            }
            else {
              xvalue = point.x;
              yvalue = point.y * -1;
            }
            
            Vec3f translation1(xvalue, yvalue, 0);
            pUseTransform->SetTranslation(translation1);
            pUseTransform->SetExecutive(m_pPresenter->GetExecutive());
            pUseTransform->RegisterObject();
            
            pParentMObject = (Group*)m_pParentGroup->GetFirstChild();
            ((Transform *)pUseTransform)->AddChild((Node3D *)pParentMObject);
            ((Transform *)m_pParentGroup)->AddChild((Node3D *)pUseTransform);
            ((Group *)m_pParentGroup)->RemoveChild((Node3D *)pParentMObject);
            ((Node3D *)pParentMObject)->AddParent((Group *)pUseTransform);
            ((Node3D *)pUseTransform)->AddParent((Group *)m_pParentGroup);
            
            i->selectedObject = pUseTransform;
      }
    }
    ++m_nNumberofMovements;
    TRACE( "\n ++ %d", m_nNumberofMovements );
  } 
  m_bMovement = FALSE;
  
  return;
#endif
}

void OpenGLVisualRenderer::MouseMovement(UINT nFlags, CPoint point)
{
  switch(nFlags) {
		case MK_LBUTTON :
      if (m_pHeadSelectedNode != NULL) 
        m_cMMovePoint = point;
      m_bMovement = TRUE;
      break;
    default :
      m_bMovement = FALSE;
  }
  
}

BOOL OpenGLVisualRenderer::AllowUndo()
{
  if ((m_pHeadSelectedNode == NULL) || (m_pHeadSelectedNode->next == m_pTailSelectedNode))
    return FALSE;
  else
    return TRUE;
}

BOOL OpenGLVisualRenderer::UndoTransformation()
{
  //#ED
  /*
  Group *pChildObject;
  BOOL success = FALSE;
  Group* pObjectChoosen;
  
    int currentindex = m_pHeadSelectedNode->next->m_nMoveNumber;
    SelectedNode* p;
    
      for (SelectedNode* i = m_pHeadSelectedNode->next; i != m_pTailSelectedNode;) {
      if (currentindex == i->m_nMoveNumber) {
      pChildObject = i->selectedObject;
      
        pObjectChoosen = (Group*)pChildObject->GetFirstChild();
        
          const ZArray<Group*>& groups = pChildObject->GetGroupParents();
          
            ((Transform *)groups[0])->AddChild((Node3D *)pObjectChoosen);
            ((Group *)pChildObject)->RemoveChild((Node3D *)pObjectChoosen);
            ((Node3D *)pObjectChoosen)->AddParent((Group *)groups[0]);
            
              groups[0]->RemoveChild((Node3D*) pChildObject);
              ((Node3D *)pObjectChoosen)->RemoveParent(pChildObject);
              
                p = i->next;
                i->FreeSelectedNode();	
                delete i;
                i = p;
                }
                else
                i = i->next;
                }
                --m_nNumberofMovements;
                TRACE( "\n -- %d", m_nNumberofMovements );
  */
  return TRUE;
}

#endif
