// -*- c++ -*-

// This software module was originally developed by Karl Anders ygard
// (Telenor R&D) in the course of development of the MPEG-4 Systems
// standard (ISO/IEC 14496-1).  This software module is an implementation
// of a part of one or more MPEG-4 Systems tools as specified by the MPEG-4
// Systems standard.  ISO/IEC gives users of the MPEG-4 Systems standard
// free license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4 Systems
// standard.  Those intending to use this software module in hardware or
// software products are advised that its use may infringe existing
// patents.  The original developer of this software module and his/her
// company, the subsequent editors and their companies, and ISO/IEC have no
// liability for use of this software module or modifications thereof in an
// implementation.  Copyright is not released for non MPEG-4 Systems
// conforming products.  The original developer retains full right to use
// the code for his/her own purpose, assign or donate the code to a third
// party and to inhibit third parties from using the code for non MPEG-4
// Systems conforming products.  This copyright notice must be included in
// all copies or derivative works.
//
// For more information or to receive an up to date version of this module,
// mail karlo@nta.no or check <http://televr.fou.telenor.no/compositor/>.
//
// Copyright 1997, 1998.

#include "OpenGl/OpenGl.h"
#include "IndexedFaceSetProxy.h"
#include "opengl/base/OpenGLVisualRenderer.h"


// traverse scene graph, checking for intersection

/* virtual */ BOOL
IndexedFaceSetProxy::Select(
    OpenGLVisualRenderer* pRenderer, 
    DWORD dwTime, 
    unsigned int& depth)
{
  GLuint selectBuffer[4];

  ::glSelectBuffer(4, selectBuffer);
  ::glRenderMode(GL_SELECT);

  Render(pRenderer, dwTime);

#ifdef _DEBUG
  if (pRenderer->CheckGLError("IndexedFaceSetProxy:"))
    TRACE("IndexedFaceSetProxy::Select() [%08x]: ", this);
#endif

  if (::glRenderMode(GL_RENDER) > 0)
    if (depth > selectBuffer[1]) {
      depth = selectBuffer[1];

      return TRUE;
    }

  return FALSE;
}

// Render object

/* virtual */ BOOL
IndexedFaceSetProxy::Render(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
  IndexedFaceSet* pIndexedFaceSet = (IndexedFaceSet*) GetMediaObject();

  const MF<int>& coordIndex = pIndexedFaceSet->coordIndex;
  /* const */ Coordinate * pCoord = pIndexedFaceSet->coord;
  const MF<int>& colorIndex = pIndexedFaceSet->colorIndex;
  /* const */ Color* pColor = pIndexedFaceSet->color;
  const MF<int>& normalIndex = pIndexedFaceSet->normalIndex;
  /* const */ Normal* pNormal = pIndexedFaceSet->normal;
  const MF<int>& texCoordIndex = pIndexedFaceSet->texCoordIndex;
  /* const */ TextureCoordinate* pTexCoord = pIndexedFaceSet->texCoord;
  BOOL colorPerVertex = (Bool) (pIndexedFaceSet->colorPerVertex);
  BOOL normalPerVertex = (Bool) (pIndexedFaceSet->normalPerVertex);
  
  const MF<Vec3f>& coordarray = pCoord->point;
  const MF<Color3f>& colorarray = pColor->color;
  const MF<Vec3f>& normalarray = pNormal->vector;
  const MF<Vec2f>& texCoordarray = pTexCoord->point;

  ::glPushAttrib(GL_POLYGON_BIT | GL_CURRENT_BIT);
  
  if ((BOOL) (Bool) (pIndexedFaceSet->solid) /* GetSolid() */)
    ::glEnable(GL_CULL_FACE);
  
  if (! (BOOL) (Bool) (pIndexedFaceSet->ccw) /* GetCCW() */)
    ::glFrontFace(GL_CW);

  if (! pNormal)
    ::glDisable(GL_LIGHTING);

  if (colorIndex.GetLength() == 0 && colorPerVertex &&
    normalIndex.GetLength() == 0 && normalPerVertex &&
    texCoordIndex.GetLength() == 0)
  {
    TRACE("[IFS::Render] no ColorIndex, no NormalIndex, no TexCoordIndex \n");

    // special case fast path 
    
    int coords;
    
    if (coordIndex.GetLength() == 0)
      coords = coordarray.GetLength();
    else
      coords = coordIndex.GetLength();

    if (long(pIndexedFaceSet->GetUpdateTime()) > m_nTimeStamp ||
      (pCoord && (long(pCoord->GetUpdateTime()) > m_nTimeStamp)) ||
      (pColor && (long(pColor->GetUpdateTime()) > m_nTimeStamp)) ||
      (pNormal && (long(pNormal->GetUpdateTime()) > m_nTimeStamp)) ||
      (pTexCoord && (long(pTexCoord->GetUpdateTime()) > m_nTimeStamp)))
    {
      if (! m_nDisplayList)
      {
        m_nDisplayList = ::glGenLists(1);
      }

      m_nTimeStamp = long(dwTime);

      TRACE("BEFORE NewList -- DisplayList %4d \n", m_nDisplayList);
      ::glNewList(m_nDisplayList, GL_COMPILE_AND_EXECUTE);
      TRACE("AFTER  NewList -- DisplayList %4d \n", m_nDisplayList);
      
      if (pRenderer->GetOpenGLVersion() < 1.1) {
#if defined(GL_EXT_vertex_array) && !defined(WIN32)
        ::glVertexPointerEXT(3, GL_FLOAT, sizeof(Vec3f), coords, &coordarray[0].x);
        ::glEnable(GL_VERTEX_ARRAY_EXT);
        if (pColor) {
          ::glColorPointerEXT(3, GL_FLOAT, sizeof(Color3f), coords, &colorarray[0].r);
          ::glEnable(GL_COLOR_ARRAY_EXT);
        }
        if (pNormal) {
          ::glNormalPointerEXT(GL_FLOAT, sizeof(Vec3f), coords, &normalarray[0].x);
          ::glEnable(GL_NORMAL_ARRAY_EXT);
        }
        if (pTexCoord) {
          ::glTexCoordPointerEXT(2, GL_FLOAT, sizeof(Vec2f), coords, &texCoordarray[0].x);
          ::glEnable(GL_TEXTURE_COORD_ARRAY_EXT);
        }
        
        if (coordIndex.GetLength() == 0)
          ::glDrawArraysEXT(GL_POLYGON, 0, coords);
        else {
          ::glBegin(GL_POLYGON);
          
          for (int i = 0; i < coords; ++i) {
            int j = coordIndex[i];
            
            if (j != -1)
              ::glArrayElementEXT(j);
            else {
              ::glEnd();
              ::glBegin(GL_POLYGON);
            }
          }
          
          ::glEnd();
        }
        
        ::glDisable(GL_VERTEX_ARRAY_EXT);
        ::glDisable(GL_COLOR_ARRAY_EXT);
        ::glDisable(GL_NORMAL_ARRAY_EXT);
        ::glDisable(GL_TEXTURE_COORD_ARRAY_EXT);
#endif
      } 
      else 
      {
        ::glVertexPointer(3, GL_FLOAT, sizeof(Vec3f), &coordarray[0].x);
        ::glEnableClientState(GL_VERTEX_ARRAY);
        if (pColor) {
          ::glColorPointer(3, GL_FLOAT, sizeof(Color), &colorarray[0].r);
          ::glEnableClientState(GL_COLOR_ARRAY);
        }
        if (pNormal) {
          ::glNormalPointer(GL_FLOAT, sizeof(Vec3f), &normalarray[0].x);
          ::glEnableClientState(GL_NORMAL_ARRAY);
        }
        if (pTexCoord) {
          ::glTexCoordPointer(2, GL_FLOAT, sizeof(Vec2f), &texCoordarray[0].x);
          ::glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        }
        
        if (coordIndex.GetLength() == 0)
        {
          ::glDrawArrays(GL_POLYGON, 0, coords);
        }
        else 
        {
          TRACE("[IFS::Render] with CoordIndex \n");
          ::glBegin(GL_POLYGON);

          int i, j;

          for (i = 0; i < coords; ++i) {
            j = coordIndex[i];
            
            if (j != -1)
              ::glArrayElement(j);
            else {
              ::glEnd();
              ::glBegin(GL_POLYGON);
            }
          }

          ::glEnd();
        }
        
        ::glDisableClientState(GL_VERTEX_ARRAY);
        ::glDisableClientState(GL_COLOR_ARRAY);
        ::glDisableClientState(GL_NORMAL_ARRAY);
        ::glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      }
      
      ::glEndList();
    } 
    else
    {
      TRACE("BEFORE CallList -- DisplayList %4d \n", m_nDisplayList);
      ::glCallList(m_nDisplayList);
      TRACE("AFTER  CallList -- DisplayList %4d \n", m_nDisplayList);
    }
  }
  else
  {
    TRACE("[IFS::Render] either ColorIndex, or NormalIndex, or TexCoordIndex \n");
    /* The following is a little obscure; the reason why the
    inlinedRenderLoop() method isn't started right away is that passing
    inlinedRenderLoop() static boolean arguments allows the compiler to
    evalute parts of the code at compile time and inline the code much
    better.  This of course leads to much larger code; and the compiler
    probably will not inline all of the code.  Still, the code should be
    more efficient.  Doing arc profiling would probably give the
    compiler a chance to do a better job. */
    if (coordIndex.GetLength() != 0)
    { // 1
      if (colorIndex.GetLength() != 0)
      { // 2
        if (colorPerVertex)
        { // 3
          if (pRenderer->GetCurrentTexture() && pTexCoord)
          {
            inlinedRenderLoop(TRUE,
              TRUE, TRUE,
              normalIndex.GetLength() != 0, normalPerVertex,
              texCoordIndex.GetLength() != 0, pTexCoord,
              dwTime);
          }
          else
          {
            inlinedRenderLoop(TRUE,
              TRUE, TRUE,
              normalIndex.GetLength() != 0, normalPerVertex,
              FALSE, NULL,
              dwTime);
          }
        } // 3
        else
        { // 3
          if (pRenderer->GetCurrentTexture() && pTexCoord)
          {
            inlinedRenderLoop(TRUE,
              TRUE, FALSE,
              normalIndex.GetLength() != 0, normalPerVertex,
              texCoordIndex.GetLength() != 0, pTexCoord,
              dwTime);
          }
          else
          {
            inlinedRenderLoop(TRUE,
              TRUE, FALSE,
              normalIndex.GetLength() != 0, normalPerVertex,
              FALSE, NULL,
              dwTime);
          }
        } // 3
      } // 2
      else
      { // 2
        if (colorPerVertex)
        { // 3
          if (pRenderer->GetCurrentTexture() && pTexCoord)
          {
            inlinedRenderLoop(TRUE,
              FALSE, TRUE,
              normalIndex.GetLength() != 0, normalPerVertex,
              texCoordIndex.GetLength() != 0, pTexCoord,
              dwTime);
          }
          else
          {
            inlinedRenderLoop(TRUE,
              FALSE, TRUE,
              normalIndex.GetLength() != 0, normalPerVertex,
              FALSE, NULL,
              dwTime);
          }
        } // 3
        else
        { // 3
          if (pRenderer->GetCurrentTexture() && pTexCoord)
          {
            inlinedRenderLoop(TRUE,
              FALSE, FALSE,
              normalIndex.GetLength() != 0, normalPerVertex,
              texCoordIndex.GetLength() != 0, pTexCoord,
              dwTime);
          }
          else
          {
            inlinedRenderLoop(TRUE,
              FALSE, FALSE,
              normalIndex.GetLength() != 0, normalPerVertex,
              FALSE, NULL,
              dwTime);
          }
        } // 3
      } // 2
    } // 1
    else
    { // 1
      if (colorIndex.GetLength() != 0)
      { // 2
        if (colorPerVertex)
        { // 3
          if (pRenderer->GetCurrentTexture() && pTexCoord)
          {
            inlinedRenderLoop(FALSE,
              TRUE, TRUE,
              normalIndex.GetLength() != 0, normalPerVertex,
              texCoordIndex.GetLength() != 0, pTexCoord,
              dwTime);
          }
          else
          {
            inlinedRenderLoop(FALSE,
              TRUE, TRUE,
              normalIndex.GetLength() != 0, normalPerVertex,
              FALSE, NULL,
              dwTime);
          }
        } // 3
        else
        { // 3
          if (pRenderer->GetCurrentTexture() && pTexCoord)
          {
            inlinedRenderLoop(FALSE,
              TRUE, FALSE,
              normalIndex.GetLength() != 0, normalPerVertex,
              texCoordIndex.GetLength() != 0, pTexCoord,
              dwTime);
          }
          else
          {
            inlinedRenderLoop(FALSE,
              TRUE, FALSE,
              normalIndex.GetLength() != 0, normalPerVertex,
              FALSE, NULL,
              dwTime);
          }
        } // 3
      } // 2
      else
      { // 2
        if (colorPerVertex)
        { // 3
          if (pRenderer->GetCurrentTexture() && pTexCoord)
          {
            inlinedRenderLoop(FALSE,
              FALSE, TRUE,
              normalIndex.GetLength() != 0, normalPerVertex,
              texCoordIndex.GetLength() != 0, pTexCoord,
              dwTime);
          }
          else
          {
            inlinedRenderLoop(FALSE,
              FALSE, TRUE,
              normalIndex.GetLength() != 0, normalPerVertex,
              FALSE, NULL,
              dwTime);
          }
        } // 3
        else
        { // 3
          if (pRenderer->GetCurrentTexture() && pTexCoord)
          {
            inlinedRenderLoop(FALSE,
              FALSE, FALSE,
              normalIndex.GetLength() != 0, normalPerVertex,
              texCoordIndex.GetLength() != 0, pTexCoord,
              dwTime);
          }
          else
          {
            inlinedRenderLoop(FALSE,
              FALSE, FALSE,
              normalIndex.GetLength() != 0, normalPerVertex,
              FALSE, NULL,
              dwTime);
          }
        } // 3
      } // 2
    } // 1
  }

  ::glPopAttrib();
                              
#ifdef _DEBUG
  if (pRenderer->CheckGLError("IndexedFaceSetProxy:"))
    TRACE("IndexedFaceSetProxy::Render() [%08x] \n", this);
#endif
                              
  return TRUE;
}

// Render object description

/* virtual */ void
IndexedFaceSetProxy::RenderDescription(OpenGLVisualRenderer* pRenderer, DWORD /* dwTime */)
{
  // IndexedFaceSet* pIndexedFaceSet = (IndexedFaceSet*) GetMediaObject();
  int children = ((GetCoord() ? 1 : 0) +
		  (GetColor() ? 1 : 0) +
		  (GetNormal() ? 1 : 0) +
		  (GetTexCoord() ? 1 : 0));

  GLfloat color[] = { 1.0f, 0.0f, 1.0f, 0.75f};

  ::glColor4fv(color);

  RenderDescriptionBox(pRenderer, "IndexedFaceSet");

  ::glPushMatrix();

  ::glScalef(1.0f / children, 1.0f, 1.0f);
  ::glTranslatef(0.0f, -1.0f, 0.0f);

  if (/* pIndexedFaceSet-> */ GetCoord()) {
    RenderDescriptionBox(pRenderer, "Coordinate");
    
    ::glTranslatef(1.0f, 0.0f, 0.0f);
  }

  if (/* pIndexedFaceSet-> */ GetColor()) {
    RenderDescriptionBox(pRenderer, "Color");
    
    ::glTranslatef(1.0f, 0.0f, 0.0f);
  }

  if (/* pIndexedFaceSet-> */ GetNormal()) {
    RenderDescriptionBox(pRenderer, "Normal");
    
    ::glTranslatef(1.0f, 0.0f, 0.0f);
  }

  if (/* pIndexedFaceSet-> */ GetTexCoord())
    RenderDescriptionBox(pRenderer, "TextureCoordinate");

  ::glPopMatrix();
}


// Stefano Battista
// moved from IndexedFaceSet to Proxy

// ===========================================================
// Update bounding box

void IndexedFaceSetProxy::UpdateBbox()
{
  // Node::UpdateBbox();
  for (int i = m_CollisionParents.GetLength() - 1; i >= 0; --i)
    ( (MediaObjectProxy*) ( m_CollisionParents.ElementAt(i) ->
      GetRenderingObject() ) ) ->
      UpdateBbox();
}


// ===========================================================
// Compute normals

void IndexedFaceSetProxy::ComputeNormals()
{
  // See if we have normals that may need updating
  if (! GetNormal()) {
    // If we do, see if we have placeholders

    if (! m_pAutoNormals){
      m_pAutoNormals = new Normal;
      m_pAutoNormals->SetExecutive(GetIndexedFaceSet()->GetExecutive());
      // m_pAutoNormals->RegisterObject();
      MediaObjectProxy * poNormalProxy = new MediaObjectProxy(m_pAutoNormals);
      m_pAutoNormals->SetRenderingObject(poNormalProxy);
    }
    
    GenerateNormals();
  }
}

// ===========================================================
// Generate normals

void IndexedFaceSetProxy::GenerateNormals()
{
  if (GetCoord()) {
    const MF<Vec3f>& coordarray = GetCoord()->point /* GetPoint() */;
    const MF<int>& coordIndex = GetCoordIndex();
    
    if (coordIndex.GetLength() > 0) {
      ZArray<Vec3f> faceNormals;
      
      ZArray<ZArray<int> > vertexDependencies;
      ZArray<int> uniqueVertexMap;
      
      uniqueVertexMap.Expand(coordarray.GetLength());
      
      {
        ZArray<Vec3f> uniqueVertices;
        int numberOfUniqueVertices = 0;
        
        int point = 0;
        int face = 0;
        
        // first, compute normals of all faces and find all vertex dependencies
        
        for (int i = 0; i < coordIndex.GetLength(); ++i) {
          int coord = coordIndex[i];
          
          if (coord != -1) {
            int uniqueVertex = uniqueVertices.Find(coordarray[coord]);
            
            if (uniqueVertex == -1) {
              uniqueVertices += coordarray[coord];
              
              ++vertexDependencies;
              uniqueVertex = numberOfUniqueVertices++;
            }
            
            uniqueVertexMap[coord] = uniqueVertex;
            vertexDependencies[uniqueVertex] += face;
            
            // Compute normal when we have enough vertices
            
            if (i - point == 2) {
              Vec3f normal = ComputeNormal(
                coordarray[coordIndex[point]],
                coordarray[coordIndex[point + 1]],
                coordarray[coordIndex[point + 2]]);
              
              normal.Normalize();
              
              faceNormals += normal;
            }
          } else {
            face++;
            point = i + 1;
          }
        }
      }
      
      {
        MF<Vec3f> normals;
        
        int face = 0;
        int counter = 0;
        
        m_autoNormalIndex.Clear();
        
        for (int j = 0; j < coordIndex.GetLength(); ++j)
          if (coordIndex[j] != -1) {
            const ZArray<int>& faces = vertexDependencies[uniqueVertexMap[coordIndex[j]]];
            Vec3f normal = faceNormals[face];
            Vec3f vertexNormal;
            
            for (int k = 0; k < faces.GetLength(); ++k)
              if (normal == faceNormals[faces[k]] ||
                ::acos(normal * faceNormals[faces[k]]) <= GetIndexedFaceSet()->creaseAngle)
                vertexNormal += faceNormals[faces[k]];
              
              normals += vertexNormal;
              m_autoNormalIndex += counter++;
          } else {
            m_autoNormalIndex += -1;
            ++face;
          }
          
          m_pAutoNormals->vector /* SetVector */ = (normals);
      }
    } else
      for (int i = 0; i < GetCoord()->point /* GetPoint() */ .GetLength(); ++i)
        printf("IndexedFaceSet::GenerateNormals: nonimplemented stuff here\n");
  }
  
  m_pAutoNormals->SetUpdateTime();
}

// ===========================================================
// Compute normal

Vec3f IndexedFaceSetProxy::ComputeNormal(
    const Vec3f& point_a,
		const Vec3f& point_b,
		const Vec3f& point_c)
{
  Vec3f vector_a(point_a);
  Vec3f vector_b(point_a);

  vector_a -= point_b;
  vector_b -= point_c;

  vector_a.Cross(vector_b);
  vector_a.Normalize();

  return vector_a;
}

