#include "OpenGl/OpenGl.h"
#include "OrientationInterpolatorProxy.h"

OrientationInterpolatorProxy::OrientationInterpolatorProxy(MediaObject * poNode) :
InterpolatorNodeProxy(poNode)
{}

inline BOOL OrientationInterpolatorProxy::SetValue(
    OrientationInterpolator * poOI, int iIndex)
{
  poOI->value_changed = poOI->keyValue[iIndex];
  return TRUE;
}

BOOL OrientationInterpolatorProxy::Evaluate()
{
  OrientationInterpolator * poOI = 
    (OrientationInterpolator *) GetMediaObject();

  // Check if we have any keys
  if (poOI->key.GetLength() > 0)
  {
    // If we only have one key
    if (poOI->key.GetLength() == 1)
    {
      return SetValue(poOI, 0);
    }
    else // more keys
    {
      // If set_fraction is smaller than smallest key
      if (poOI->set_fraction <= poOI->key[0])
      {
        return SetValue(poOI, 0);
      }
      else
      {
        // If set_fraction is larger than largest key
        if (poOI->set_fraction >= poOI->key[poOI->key.GetLength() - 1])
        {
          return SetValue(poOI, poOI->key.GetLength() - 1);
        }
        else
        {
          // Search the key space for fraction
          for (int i = poOI->key.GetLength() - 2; i >= 0 ; --i)
          {
            if (poOI->key[i] == poOI->set_fraction)
            {
              // Found exact match
              return SetValue(poOI, i);
            }
            else
            {
              if (poOI->key[i] < poOI->set_fraction)
              {
                // Didn't find exact match; must interpolate
                return InterpolateValue( 
                  i, i + 1, 
                  ((poOI->key[i + 1] - poOI->set_fraction) / 
                  (poOI->key[i + 1] - poOI->key[i])) );
              }
            }
          } // for
        }
      }
    } // one or more keys
  } // some key

  return FALSE;
}

BOOL OrientationInterpolatorProxy::InterpolateValue(
    int nMinIndex, int nMaxIndex, float fFraction)
{
  OrientationInterpolator * poOI = 
    (OrientationInterpolator *) GetMediaObject();

  assert(poOI->keyValue.GetLength() == poOI->key.GetLength());

  // Interpolate orientation
  float cos_t, alpha, beta;
  float epsilon(0.00001f);
  BOOL flip = FALSE;
  float sw;

  Rotation a(poOI->keyValue[nMinIndex]);
  Rotation b(poOI->keyValue[nMaxIndex]);
  Rotation tempRotation;

  // Make quaternions of the OrientationInterpolators.  To optimize this
  // function: generated quaternions when the OrientationInterpolators are
  // set.

  // let's fry some CPU cycles
  
  a.Normalize();
  b.Normalize();
  
  // Generate quaternions
  sw = ::sin(a.theta * 0.5f);

  a.x *= sw;
  a.y *= sw;
  a.z *= sw;
  
  a.theta = ::cos(0.5f * a.theta);

  // Assuming radians
  sw = ::sin(b.theta * 0.5f);
  
  b.x *= sw;
  b.y *= sw;
  b.z *= sw;

  b.theta = ::cos(0.5f * b.theta);
  
  // What is the angle between the two 4D vectors (quaternions)?
  cos_t = a.x * b.x + a.y * b.y + a.z * b.z;

  if (::fabs(cos_t - M_PI) > epsilon) {
    cos_t += a.theta * b.theta;
    // if the angle is beyond PI
    
    if (cos_t <= 0.0f) {
      cos_t = - cos_t;
      flip = true;
    }

    if ((1.0f - cos_t) < epsilon) {
    
      // linear interpolate
      beta = 1.0f - fFraction;
      alpha = fFraction;
    
    } else {
     
      // else find the slerp
      float theta = ::acos(cos_t);
      float sin_t = 1.0f / ::sin(theta);
      
      beta = ::sin((1.0f - fFraction) * theta) * sin_t;
      alpha = ::sin(fFraction * theta) * sin_t;
    }

    if (flip)
      beta = -beta;

    tempRotation.x = alpha * a.x + beta * b.x;
    tempRotation.y = alpha * a.y + beta * b.y;
    tempRotation.z = alpha * a.z + beta * b.z;
    tempRotation.theta = 2.0 * ::acos(alpha * a.theta + beta * b.theta);
    
    poOI->value_changed = tempRotation;
  } 
  else
  {
    TRACE("Warning: Interpolating between diagonally opposite orientations impossible\n");
  }

  // SB 99/05/12
  TRACE("OI::Interpolate -- value_changed %4f %4f %4f %4f \n", 
    tempRotation.x, tempRotation.y, tempRotation.z, tempRotation.theta);

  return TRUE;
}
