/*
 * VsExpr.h
 *
 *      Declaration of classes:
 *        VsExpr
 *        VsConstant
 *	  VsVar
 *        VsAssign
 *        VsArith
 *        VsUArithOp
 *        VsLogicOp
 *        VsULogicOp
 *	  VsIncDec
 *	  VsComma
 *	  VsExprList
 *
 * Expression support
 * 
 * Copyright (C) 1996, Silicon Graphics,  Inc.
 */

/*---------------------------------------------------------------
This module is copied from the VRML Script interpreter owned by
Silicon Graphics, Inc. It is modified by Wenlong Dong, Wei Wu and
Yuval Fisher(Rockwell Science Center) in the course of development
of the MPEG-4 Systems standard(ISO/IEC 14496-1) and added to the 
subsystem: Script Node.
---------------------------------------------------------------*/

#ifndef _VS_EXPR_
#define _VS_EXPR_

#include "..\encoder\ScriptDefs.h"
#include "vsstatement.h"

class VsFunctionCall;
class VsMethodCall;
class VsMemberAccess;

//
// Expression base class:
//
class VsExpr : public VsStatement 
{
public:
    // Constructor, just initializes reference count to zero
    VsExpr();

    // Expressions are also statements, and so must have an execute()
    // method.  The execute() method just calls evaluate() and returns
    // a NORMAL status
    virtual VsStatement::Status	execute(VsFunctionCall *);

    // Evaluates this expression, returning resulting value.
    virtual VsValue	evaluate(VsFunctionCall *) = 0;

    // Expressions must be reference counted because variables may be
    // multiply instanced in the parse tree
    void	ref();
    void	unref();

protected:
    virtual ~VsExpr();
private:
    short refCount;
};

//
// Constant numbers and/or strings:
//
class VsConstant : public VsExpr 
{
public:
    VsConstant(const char *string); // String is copied in
    VsConstant(double number);

    // Evaluation just returns constant value:
    virtual VsValue		evaluate(VsFunctionCall *);

protected:
    virtual ~VsConstant();
private:
    VsValue myValue;
};

//
// Abstract base class:
//
class VsVar : public VsExpr 
{
public:
    // Assignment operator.
    virtual VsValue		assign(const VsValue &, VsFunctionCall *) = 0;
    
    // Assign to a component of an array.
    virtual VsValue		assign(const VsValue &, int32_t index, VsFunctionCall *);

    // Note: derived classes will also need to define an execute()
    // function to return the var's value.
};

//
// Assignment expression:
//
class VsAssign: public VsExpr {
public:

    // Pass variable = expression.  Destructor will delete expression,
    // the enclosing scope must delete the variable (since it may
    // appear more than once).
    VsAssign(VsVar *, VsExpr *);
    
    // Evaluate.  Assigns to variable, result is value of variable.
    virtual VsValue		evaluate(VsFunctionCall *);

protected:
    virtual ~VsAssign();

private:

	VsVar *var;
	VsExpr *expr;
};

//
// Handles arrays on the right hand side of an expression:
//
class VsArrayVar : public VsVar {
public:

    VsArrayVar(VsExpr *array, VsExpr *subscript);

    // Evaluation just returns constant value:
    virtual VsValue		evaluate(VsFunctionCall *);
    
    // Assignment operator.
    virtual VsValue		assign(const VsValue &, VsFunctionCall *);

protected:
    virtual ~VsArrayVar();
    
private:
    VsVar   *array;	    
    VsExpr  *subscript;
};

//
// Binary arithmetic operators.  One class is used for all of them to
// cut down on code bloat.
//
class VsArith: public VsExpr 
{
public:
    VsArith(VsExpr *e1, VsExpr *e2, int op);

    virtual VsValue		evaluate(VsFunctionCall *);

protected:
    virtual ~VsArith();
private:
    VsExpr *expr1, *expr2;
    int operation;
};

//
// Increment/decrement operators.  Separate class to allow
// for correct prefix/postfix behavior
//
class VsIncDec: public VsExpr 
{
public:
    VsIncDec(VsExpr *var, int inc, int pref);

    virtual VsValue		evaluate(VsFunctionCall *);

protected:
    virtual ~VsIncDec();
private:
    VsVar *var;
    int	  increment;
    int	  prefix;
};

//
// Comma operator.  Separate class to allow for expression of
// mixed type
//
class VsComma: public VsExpr 
{
public:
    VsComma(VsExpr *e1, VsExpr *e2);

    virtual VsValue		evaluate(VsFunctionCall *);

protected:
    virtual ~VsComma();
private:
    VsExpr *expr1, *expr2;
};

//
// Conditional expression operator.
//
class VsCondExpr: public VsExpr 
{
public:
    VsCondExpr(VsExpr *test, VsExpr *e1, VsExpr *e2);

    virtual VsValue		evaluate(VsFunctionCall *);

protected:
    virtual ~VsCondExpr();
private:
    VsExpr *test,  *expr1, *expr2;
};

//
// Unary minus operator
//
class VsUArithOp: public VsExpr 
{
public:
    VsUArithOp(VsExpr *e, int op);

    virtual VsValue		evaluate(VsFunctionCall *);

protected:
    virtual ~VsUArithOp();
private:
    VsExpr *expr;
    
    int	operation;
};

//
// Binary logic operators.  One class is used for all of them to
// cut down on code bloat.
//
class VsLogicOp: public VsExpr 
{
public:
    VsLogicOp(VsExpr *e1, VsExpr *e2, int op);

    virtual VsValue		evaluate(VsFunctionCall *);

protected:
    virtual ~VsLogicOp();
private:
    VsExpr *expr1, *expr2;
    int operation;
};

//
// Unary logic operators.  One class is used for all of them to
// cut down on code bloat.
//
class VsULogicOp: public VsExpr 
{
public:
    VsULogicOp(VsExpr *e, int op);

    virtual VsValue		evaluate(VsFunctionCall *);

protected:
    virtual ~VsULogicOp();
private:
    VsExpr *expr;
    int operation;
};

class VsExprList : public ZArray<VsExpr *>
{
};

#endif /* _VS_EXPR_ */
