/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

  SUBSYSTEM:  Decoder
  FILE:       AnimationStreamDec.cpp
  AUTHOR:     Mikael Bourges-Sevenier (France Telecom CNET/DIH/ATI)
  OVERVIEW:   Implementation of AnimationStreamDecoder
-------------------------------------------------------------------------*/
#include "defs.h"

#ifdef ANIMATION

#include <assert.h>
#include "nodes.h"
#include "AnimStreamDec.h"
#include "Stream.h"
#include "ODImp.h"
#include "Executiv.h"

#define ANIM_START_CODE 0x00000018c
#define ANIM_END_CODE 0x00000018d
#define DEC_ERROR -1

AnimationStreamDecoder :: AnimationStreamDecoder ()
  : initialised(FALSE),
    numNodes(0),
    m_pMask(NULL), maskSize(0)
{
  //m_mtxProtectGlobalVars.P();
  //m_mtxProtectGlobalVars.V();
}

/*-----------------------------------------------------------------
Setup decoder (parse BIFS Animation mask)
-----------------------------------------------------------------*/
BOOL AnimationStreamDecoder :: Setup (
				      const BYTE *pSpecificInfo,
				      int nSpecificInfoLength,
				      int /*nESNumber*/
				      )
{
  maskSize=nSpecificInfoLength;
  if(m_pMask) delete[] m_pMask;
  m_pMask=NULL;

  if(pSpecificInfo && nSpecificInfoLength > 0 ) {
    /* init variable
       nodeMask: mask setting the animated nodes between frames
       pNodeMask: array of pointers to node in the scene potentially animated
       */
    m_pMask=new unsigned char[nSpecificInfoLength];
    memcpy(m_pMask,(unsigned char *) pSpecificInfo,nSpecificInfoLength);
  }
  return TRUE;
}

bool AnimationStreamDecoder :: ParseMask ( ) {
  if(!m_pMask || !maskSize) return FALSE;

  //fprintf(stderr,"===== Parsing Animation Mask        =====\n");
  
  // parses specificInfo buffer
  setInputBuffer(m_pMask);
  setBufSize(maskSize);
  
  nodeIDbits=readBits(5);
  routeIDbits=readBits(5);

  if(readBit()) {
    fprintf(stderr,"AnimationStream Setup ERROR: This is a BIFS Update stream!!!");
    return FALSE;
  }
  
  randomAccess=readBit();

	GetExecutive () -> LockScene ();
  do {
    int nodeID=readBits(nodeIDbits);
    //fprintf(stderr,"[Mask] nodeID: %d\n",nodeID);
    MediaObject *node=GetRoot()->FindNode(nodeID);
    pNodeMask += node;
    nodeMask += TRUE;
    numNodes++;
    
    if(node) {
#ifdef TRACE_ENCODING
      ZString nodeName=node->GetName();
      //fprintf(stderr,"[Mask] name %s\n",(const char *) nodeName);
#endif
      // calls the node's parseAnimationMask method
      node->parseAnimationMask(this);
    }
  } while(readBit());
	GetExecutive () -> UnlockScene ();
  
  maskSize=0;
  if(m_pMask) delete[] m_pMask;
  m_pMask=NULL;

  return TRUE;
}

/*-----------------------------------------------------------------
Find a compatible format and initialize the decoder
-----------------------------------------------------------------*/
BOOL AnimationStreamDecoder :: SetFormat (
	Capabilities &,
	DecoderParams &decoderParams
)
{
	decoderParams.m_nStreamNumber = 0;
	decoderParams.m_nOptimalAUSize = 256;
	decoderParams.m_nDBSize = 256;
	decoderParams.m_nCBSize = 0;		// ZL
	return TRUE;
}

void AnimationStreamDecoder :: Run ()
{
  LPBYTE pInData;
  int nInSize;
  DWORD dwTime;
  BOOL bRecurrent, bEOU;

  while ((pInData =(LPBYTE) GetInputStream() -> Fetch (
						  nInSize,
						  dwTime,
						  bEOU,
						  bRecurrent, // is it the same block as previously ?
						  FALSE,
						  TRUE)
	  ) != END_OF_STREAM ) {

    if (IsActive() && !bRecurrent) {
      static int frameID;

      if(!initialised) {
	ParseMask();
	initialised=TRUE;
	frameID=0;
      }

      //fprintf(stderr,"AnimationStreamDec: Unit size: %d, Time stamp: %lu\n",nInSize,dwTime);

      // reads animation frames
      setInputBuffer(pInData);
      setBufSize(nInSize);
      
      // parses frame header
      readAFrame();
      GetInputStream() -> Release (TRUE);
    }
  }
}

int AnimationStreamDecoder :: readAFrame()
{
  if(readFrameHeader()==DEC_ERROR) {
    fprintf(stderr,"Error in frame header\n");
    return DEC_ERROR;
  }
  readFrameData();
  return TRUE;
}

int AnimationStreamDecoder :: readFrameHeader()
{
  if(checkAnimStartCode()) {
    isIntra=readBit();

    for (int i = 0; i < numNodes; i++) {
      nodeMask[i] = readBit ();
    }
    
    if(isIntra) {
      if(readBit()) // is_frame_rate
	readFrameRate();
      if(readBit()) //is_time_code
	readTimeCode();
    }
    if(readBit()) // is_skip_frames
      if(!readSkipFrames()) return DEC_ERROR;
  }
  else 
    return DEC_ERROR;
  return 1; // OK
}

void AnimationStreamDecoder :: readFrameData()
{
	GetExecutive () -> LockScene ();
  //fprintf(stderr,"In readFrameData()\n");
  aad.setStream(this);
  if (!isIntra) aad.reset(); // reset Adaptive Arithmetic Decoder 
                             //MBS: cause problem with FBA, 2D mesh
                             // is it an error in FDIS ?

  for (int i=0; i<numNodes;i++) { // for all defined nodes in mask
    if (nodeMask[i]) {// if node i is animated in this frame
      // calls the node's animate method
      pNodeMask[i]->animate(this,isIntra); 
    }
  }
  if (!(isIntra)) aad.adjustBits();
	GetExecutive () -> UnlockScene ();
}

bool AnimationStreamDecoder :: checkAnimStartCode()
{
  unsigned char *bufptrtmp=bufptr;
  
  int code=readBits(32);
  
  if((code & 0xfffffe00)==0x0) {
    fprintf(stderr,"Start code %d found\n",code);
    incnt=0;
  }
  else 
    bufptr=bufptrtmp;
  
  return TRUE;    
}

float AnimationStreamDecoder :: readFrameRate()
{
  int frame_rate,seconds;
  BOOL frequency_offset;
  
  m_aTime.setFrameRate(frame_rate=readBits(8));
  m_aTime.setFrameRateSeconds(seconds=readBits(4));
  m_aTime.setFrameRateOffset(frequency_offset=readBit());    
  
  return (float)1000/(1000+(frequency_offset ? 1 : 0))*
    (float)(frame_rate+(float)seconds/16);
}


void  AnimationStreamDecoder :: readTimeCode()
{
  m_aTime.setTimeCodeHours(readBits(5));
  m_aTime.setTimeCodeMinutes(readBits(6));
  m_aTime.setTimeCodeMarkerBit(readBit());
  m_aTime.setTimeCodeSeconds(readBits(6));
}

bool AnimationStreamDecoder :: readSkipFrames()
{
  int nfs,value;
  
  nfs = readBits(4);
  
  if(nfs<15) {
    m_aTime.setNumSkipFrames(nfs);
    return TRUE;
  }
  
  value = nfs;
  
  do {
    value = readBits(4);
    nfs+=value;
  }
  while (value == 15);
  
  if(nfs>254) {
    fprintf(stderr,"AnimationStreamDecoder: \n Maximum allowable frames skip is 254\n");
    return FALSE;
  }
  
  m_aTime.setNumSkipFrames(nfs);
  return TRUE;
}

/*-----------------------------------------------------------------
Bit manipulation routines: we use a FIFO model for the bit buffer.
The floating-point parsing reads IEEE floats
-----------------------------------------------------------------*/

/*-----------------------------------------------------------------
Sets the input buffer
-----------------------------------------------------------------*/
void AnimationStreamDecoder::setInputBuffer(unsigned char *AUptr) {
  //fprintf(stderr,"setting input buffer to %ld\n",AUptr);
  startbuf=bufptr=AUptr;
  initBits();
}

/*-----------------------------------------------------------------
Initializes the bit counter and buffer
-----------------------------------------------------------------*/
void AnimationStreamDecoder::initBits() {
  incnt = 0;
  inused = 0;
  inbytecnt = 0;
}

/*-----------------------------------------------------------------
Reads n bits
-----------------------------------------------------------------*/
int AnimationStreamDecoder::readBits(int n) {
  int i;
  int ibuf = 0;

  if(n>32) {
    fprintf(stderr, "Cannot fetch more than 32 bits from readbits()\n");
    exit(-1);
  }

  for(i=0; i<n; i++) { /* MSB first */
    ibuf <<= 1;
    ibuf += readBit();
  }

  return ibuf;
}


/*-----------------------------------------------------------------
Reads a bit
-----------------------------------------------------------------*/
bool AnimationStreamDecoder::readBit() {
  int t;
 
  if(incnt==0) {
    if(bufptr-startbuf>=bufSize) 
      inbfr=0;
    else
      inbfr = *(bufptr++);
    incnt = 8;
    inbytecnt++;
  }
  t = (inbfr&0x80)>>7;
  inbfr <<= 1;
  incnt --;
  inused ++;
  return t != 0;
}


/*-----------------------------------------------------------------
Resets bit buffer
-----------------------------------------------------------------*/
void AnimationStreamDecoder::alignBits() {

  incnt = 0;
}


/*-----------------------------------------------------------------
This is used by the arithmetic decoder to find the byte boundary
after the decoding of a arithmetic coded stream.
-----------------------------------------------------------------*/
void AnimationStreamDecoder::adjustBits() {
  int bitin = 8 - incnt;
  
  if(bitin<=5) {
    //fseek(in_stream, -3L, SEEK_CUR);
    bufptr -= 3;
    incnt=0;
    readBits(bitin+2);
    }
  else {
    //fseek(in_stream, -2L, SEEK_CUR);
    bufptr -= 2;
    incnt=0;
    readBits(bitin-6);
  }
}

/*-----------------------------------------------------------------
Zero bit stuffing to next byte boundary
-----------------------------------------------------------------*/
void AnimationStreamDecoder::alignInputBits()
{
  if(incnt) {
    readBits(incnt); // incnt is reset in this call
  }
}

/*-----------------------------------------------------------------
Reads IEEE floating-point value
-----------------------------------------------------------------*/
float AnimationStreamDecoder :: readFloat ()
{	
#ifdef WIN32	
  unsigned char buf [4]="\0\0\0";
  for(int i=0;i<4;i++)
    buf[3-i]=readBits(8);
  
  return * (float *) buf;
#else
  return 0.0;
#endif
}

float AnimationStreamDecoder :: readGenericFloat (bool useEfficientCoding)
{
  if(useEfficientCoding) return readEfficientFloat();
  return readFloat();
}

float AnimationStreamDecoder :: readEfficientFloat ()
{	
  union
  {	
    float f;
    long l;
  } value;

  int mantissaLength=readBits(4); // mantissaLength
  if(mantissaLength) {
    int exponentLength=readBits(3); // exponentLength
    int mantissaSign=readBit() ; // mantissaSign
    int exponentSign=0;
    int exponent=0;

    int mantissa=readBits(mantissaLength); // mantissa
    /*fprintf(stderr,"mantissaLength=%d\n",mantissaLength);
    fprintf(stderr,"exponentLength=%d\n",exponentLength);
    fprintf(stderr,"mantissaSign=%d\n",mantissaSign);
    fprintf(stderr,"mantissa=%d\n",mantissa);*/
    if(exponentLength) {
      exponentSign=readBit(); // exponentSign
      exponent=readBits(exponentLength); // exponent
      //fprintf(stderr,"exponentSign=%d\n",exponentSign);
      //fprintf(stderr,"exponent=%d\n",exponent);
    }

    long val=mantissaSign<<31;
    unsigned char exp = (1-2*exponentSign)*((1<<(exponentLength-1))+exponent)+127;
    val |= (exp & 0xff)<<23;
    val |= mantissa << (23-mantissaLength);
    value.l=val;
    
    //fprintf(stderr," !!! val=%g\n",value.f);
    return value.f;
  }

  return 0.0;
}

#endif // ANIMATION