/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Executive
FILE:       Executiv.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Impelentation of Executive
-------------------------------------------------------------------------*/

//Modified 2001/05/08 Boris Felts boris.felts@polytechnique.org

#include "defs.h"
#include <assert.h>
#include "Im1defs.h"
#include "Executiv.h"
#include "AL.h"
#include "MemBifs.h"
#include "Presenter.h"
#include "NodeFld.h"
#include "App.h"
#include "Clock.h"
#include "Channel.h"
#include "ODImp.h"
#include "OCIDec.h"
#include "Inline.h"
#include "CrNodes.h"
#include "Route.h"
#include "ScriptImp.h"		//BF Script
#include "..\Animation\Anim.h"				// Animation
#include "SrvCmd.h"
#include "MediaSensor.h"
#include "MediaControl.h"
#include "..\UI\InputSensorImp.h"
//ML - added the next lines
#ifdef PLAYER3D
#include "../OpenGl/OpenGl.h"
#include "../OpenGl/base/OpenGLPresenter.h"
#endif

/*-----------------------------------------------------------------
Executive constructor. Saves the arguments, insantiate linked objects,
and create the Executive thread.
-----------------------------------------------------------------*/
Executive :: Executive (
	Application *pApp,
	HWND hWnd,
	FILE *fTraceFile,
	int flags,
	DeviceFilter *pDeviceFilter	//  CC 010123
) :
	m_pRoot (NULL),			//ML - added this line
	m_fTraceFile (fTraceFile),
	m_flags (flags),
	m_pDeviceFilter (pDeviceFilter)		// CC 010123
{
	m_bSceneInited = m_bActive = false;
	m_pApp = pApp;
	m_hWnd = hWnd;
	m_pPresenter = NULL;

	/*-------------------------------------------------------------
	Load the appropriate DMIF filter
	-------------------------------------------------------------*/
	static ZRegistry registry ("DMIF\\Filter");
	m_hDMIF = NULL;
	m_pDMIF = (DMIF_Filter *) registry.CreateInstance ("DMIFFilter", m_hDMIF);
	if (m_pDMIF == NULL)
	{	Error (DMIF_FILTER_NOT_FOUND);
		return;
	}
    // mp those need to be there because the presenter will overload them later
	OVERLOAD_NODE (InlineImp);
	OVERLOAD_NODE (ScriptImp);		// SCRIPT
    // end mp
	m_pDMIF -> SetSink (this);
#ifdef PLAYER3D
	//ML - added this line
	m_pPresenter = new OpenGLPresenter(this, hWnd);
#else
	m_pPresenter = new Presenter (this, hWnd);
#endif
	m_pPresenter -> Init ();
	SetClocksGranularity (m_pPresenter -> GetIterationInterval ());
	m_nSceneLockedBy = m_nLocks = m_nReserved = 0;

	/*-----------------------------------------------------------------
	The following two lines are required because the implementation of
	Inline and Script is part of the Core.
	-----------------------------------------------------------------*/
	OVERLOAD_NODE (ServerCommandImp);	// ServerCommand
	OVERLOAD_NODE (MediaSensorImp);
	OVERLOAD_NODE (MediaControlImp);
	OVERLOAD_NODE (InputSensorImp);

#ifdef ANIMATION
	OVERLOAD_NODE (AnimationStreamImp);	// Animation
#endif

	/*-------------------------------------------------------------
	Similar overloading is done for OD classes
	-------------------------------------------------------------*/
	OVERLOAD_OD_CLASS (ObjectDescriptorUpdateImp);
	OVERLOAD_OD_CLASS (ObjectDescriptorRemoveImp);
	OVERLOAD_OD_CLASS (ES_DescriptorUpdateImp);
	OVERLOAD_OD_CLASS (ES_DescriptorRemoveImp);
	OVERLOAD_OD_CLASS (IPMP_DescriptorUpdateImp);
	OVERLOAD_OD_CLASS (IPMP_DescriptorRemoveImp);
	OVERLOAD_OD_CLASS (IPMP_DescriptorImp);
	OVERLOAD_OD_CLASS (ObjectDescriptorImp);
	OVERLOAD_OD_CLASS (ES_DescriptorImp);
	OVERLOAD_OD_CLASS (MediaTimeDescriptorImp);
	OVERLOAD_OD_CLASS (ObjectDescriptorExecuteImp);
	CUSTOMIZE_EXECUTIVE_CONSTRUCTOR

}

/*-----------------------------------------------------------------
Executive destructor. Wait till the thread finishes then delete
all linked objects.
-----------------------------------------------------------------*/
Executive :: ~Executive ()
{
	CUSTOMIZE_EXECUTIVE_DESTRUCTOR

	if (m_bActive)
		Stop ();
	m_smpActive.P ();
	delete m_pPresenter;
	delete m_pDMIF;
	if (m_hDMIF)
		FreeLibrary (m_hDMIF);
}

/*-----------------------------------------------------------------
Start() instantiates a service and start a connection.
-----------------------------------------------------------------*/
void Executive :: Start (LPCSTR szAddress)
{
	CUSTOMIZE_EXECUTIVE_START

	if (m_pDMIF == NULL)
	{	Error (DMIF_FILTER_NOT_FOUND);
		return;
	}

	if (m_bActive)
	{	m_pApp -> OnConnect (RESPONSE_UNKNOWN_ERROR);
		return;
	}
	m_pInline = NULL;
	m_pRoot = NULL;
	m_bActive = TRUE;
	m_bPendingChannels = FALSE;
	m_pOD = new ObjectDescriptorImp ();
	m_pInline = new InlineImp ();
	m_pInline -> SetExecutive (this);
	m_pInline -> SetInline (m_pInline);
	m_pInline -> Init ();
	m_pInline -> SetOD (m_pOD);
	m_pOD ->SetInline (m_pInline);
	CreateService (szAddress, m_pOD);
}

/*-----------------------------------------------------------------
Stop() is called when player is closed by the user. Close all the
sessions, abort streams playing, wait till executive is not active,
and shut down everything.
-----------------------------------------------------------------*/
void Executive :: Stop ()
{
	if (m_fTraceFile != NULL)
		fflush (m_fTraceFile);
	if (!m_bActive)
		return;
	m_pPresenter -> Terminate ();
	LockScene ();
	if (m_pInline)
		m_pInline -> CloseStream ();
	m_pInline -> DeleteObject ();
	RemoveSceneObjects ();
	delete m_pOD;
	m_pInline = NULL;
	m_pRoot = NULL;
	m_smpActive.P ();
	UnlockScene ();
	m_bSceneInited = m_bActive = FALSE;
	m_smpActive.V ();
}

/*-----------------------------------------------------------------
Passes a paint request to the compositor
-----------------------------------------------------------------*/
void Executive :: Paint (void *info)
{
	m_pPresenter -> Paint (info);
}

/*-----------------------------------------------------------------
ReserveScene () locks the scene. The function returns the value of
m_nReserved. This variable will be non-zero if the caller is not allowed
to touch the scene, i.e. Unlocked was called with nReserve > 0.
nReserve is non-zero if the scene was unlocked before with nReserve > 0
and now it is reset.
-----------------------------------------------------------------*/
int Executive :: ReserveScene (int nReserve)
{
	assert (m_nReserved <= 1);
	m_mtxScene.P ();
#ifdef _DEBUG
	m_nSceneLockedBy = GetCurrentThreadId ();
#endif
	++m_nLocks;
	return m_nReserved -= nReserve;
}

/*-------------------------------------------------------------
Unlock the scene. If nReserve is >0, other threads are not allowed
to touch the scene even though it has been unlocked. This is used when
there is a need to release a thread which is currently blocking on
LockScene.
-------------------------------------------------------------*/
void Executive :: UnlockScene (int nReserve)
{
	--m_nLocks;
	m_nReserved += nReserve;
	assert (m_nReserved <= 1);

	/*-------------------------------------------------------------
	Delete all objects that were removed from the scene when it was locked
	-------------------------------------------------------------*/
	if (m_nReserved == 0 && m_nLocks <= 0)
		RemoveSceneObjects ();
#ifdef _DEBUG
	if (m_nLocks == 0)
		m_nSceneLockedBy = 0;
#endif
	m_mtxScene.V ();
}

/*-------------------------------------------------------------
LockScene locks the scene, but if there is no access (ReserveScene
returns non-zero value, it unlocks the scene and tries again after 3 ms.
nReserve is non-zero if the scene was unlocked before with nReserve > 0
and now it is reset.
-------------------------------------------------------------*/
void Executive :: LockScene (int nReserve)
{
	while (ReserveScene (nReserve))
	{	UnlockScene (nReserve);
		Sleep (3);
	}
}

/*-----------------------------------------------------------------
Pass user events to the presenter
-----------------------------------------------------------------*/
void Executive :: OnUserEvent  (
	USER_EVENT event,
	int x,
	int y,
	UINT uControlKeysStatus
)
{
	m_pPresenter -> OnUserEvent (event, x, y, uControlKeysStatus);
}

/*-----------------------------------------------------------------
Notify the application on non fatal error
-----------------------------------------------------------------*/
void Executive :: Error (NON_FATAL_ERROR nError, ...)
{
	if (m_fTraceFile != NULL)
		fflush (m_fTraceFile);
	va_list args;
	va_start (args, nError);
	m_pApp -> OnError (nError, args);
	va_end (args);
}

/*-----------------------------------------------------------------
The scene was replaced. If this is the first scene, initialize the
presenter
-----------------------------------------------------------------*/
void Executive :: ReplaceScene ()
{
	m_pRoot = m_pInline -> GetScene ();
	if (!m_bSceneInited && m_pRoot != NULL)
	{	m_pPresenter -> Start ();
		m_bSceneInited = TRUE;
	}
}

/*-----------------------------------------------------------------
Create a secondary session. A session is identified by a URL
-----------------------------------------------------------------*/
void Executive :: CreateService (
	LPCSTR szURL,
	ObjectDescriptorImp *pOD,
	UINT nParentServiceID
)
{
	ALManager *pAL;
	pAL = new ALManager (this, pOD, szURL, nParentServiceID);
	pOD ->SetRemoteAL (pAL);
	pAL -> Connect ();
}

/*-----------------------------------------------------------------
Request all the channels that were set up during the parsing of the
last AU
-----------------------------------------------------------------*/
void Executive :: RequestChannels ()
{
	if (m_bPendingChannels)
	{	m_mtxServices.P ();
		for (int i = m_services.GetLength (); --i >= 0;)
		{	if (i < m_services.GetLength ())
			{	ALManager *pService = m_services [i];
				pService -> InUse ();
				m_mtxServices.V ();
				pService ->Lock ();
				pService -> RequestChannels ();
				pService -> Unlock ();
				m_mtxServices.P ();
			}
		}
		m_bPendingChannels = FALSE;
		m_mtxServices.V ();
	}
}

/*-----------------------------------------------------------------
Set the granularity for all the clocks
-----------------------------------------------------------------*/
void Executive :: SetClocksGranularity (int nMS)
{
	ClockReference :: SetGranularity (nMS);
}

/*-------------------------------------------------------------
Callback functions for Application_Sink
-------------------------------------------------------------*/

/*-------------------------------------------------------------
Add an open channel to the map of channels
-------------------------------------------------------------*/
void Executive :: AddChannel (DataChannel *pChannel)
{
	m_mtxChannels.P ();
	m_channels.InsertAt (pChannel -> GetChannelID (), pChannel);
	m_mtxChannels.V ();
}

/*-------------------------------------------------------------
Remove a channel from the map of channels
-------------------------------------------------------------*/
void Executive :: DeleteChannel(DataChannel *pChannel)
{
	m_mtxChannels.P ();
	m_channels.DeleteFrom (pChannel ->GetChannelID ());
	m_mtxChannels.V ();
}

/*-------------------------------------------------------------
Remove a service from the list
-------------------------------------------------------------*/
void Executive :: DeleteService (ALManager *pService)
{
	m_mtxServices.P ();
	m_services -= pService;
	m_mtxServices.V ();
}

/*-------------------------------------------------------------
Find a service by its DAI id, and lock it
-------------------------------------------------------------*/
ALManager *Executive :: FindService (UINT nServiceId)
{
	ALManager *pService = NULL;
	m_mtxServices.P ();
	for (int i = m_services.GetLength (); --i >= 0;)
		if (nServiceId == m_services [i] -> GetServiceID ())
		{	pService = m_services [i];
			pService -> InUse ();
			break;
		}
	m_mtxServices.V ();
	if (pService != NULL)
		pService -> Lock ();
	return pService;
}

/*-------------------------------------------------------------
Find a channel by its DAI id, and lock it
-------------------------------------------------------------*/
DataChannel *Executive :: FindChannel (UINT nChid)
{
	DataChannel *pChannel;
	m_mtxChannels.P ();
	try
	{	pChannel = m_channels [nChid];
		pChannel -> Lock ();
	}
	catch (ZMap<DataChannel*> ::NotFound)
	{	pChannel = NULL;
	}
	m_mtxChannels.V ();
	return pChannel;
}

/*-------------------------------------------------------------
Handle a ServiceAttach confirmation
-------------------------------------------------------------*/
void Executive :: DAI_ServiceAttach_Cnf (
	const UINT		nTmpId,
	const UINT		nServiceId,
	const BYTE*		uuData,
	const u_short	uuData_Length,
	const RESPONSE	eResponse
)
{
	ALManager *pService = NULL;
	m_mtxServices.P ();
	for (int i = m_services.GetLength (); --i >= 0;)
		if (nTmpId == (UINT) m_services [i])
		{	pService = m_services [i];
			pService -> InUse ();
			break;
		}
	m_mtxServices.V ();
	if (pService != NULL)
	{	pService -> Lock ();
		pService -> OnServiceAttach (nServiceId, uuData, uuData_Length, eResponse);
		pService -> Unlock ();
	}
}

/*-------------------------------------------------------------
Handle a ServiceDetach indication
-------------------------------------------------------------*/
void Executive :: DAI_ServiceDetach_Ind (
	const UINT		nServiceId,
	const REASON	eReason
)
{
	ALManager *pService = FindService (nServiceId);
	if (pService != NULL)
	{	pService -> OnServiceDetach (eReason);
		pService -> Unlock ();
	}
}

/*-------------------------------------------------------------
Handle a ServiceDetach confirmation
-------------------------------------------------------------*/
void Executive :: DAI_ServiceDetach_Cnf (
	const UINT		/* nServiceId */,
	const RESPONSE	eResponse
)
{
	if (eResponse != RESPONSE_OK)
		Error (SERVICE_DETACH_FAILED);
}

/*-------------------------------------------------------------
Handle a ChannelAdd confirmation
-------------------------------------------------------------*/
void Executive :: DAI_ChannelAdd_Cnf(
	const UINT		nServiceId,
	const DAIChannelAddParam& outParam
)
{
	ALManager *pService = FindService (nServiceId);
	if (pService != NULL)
	{	pService -> OnChannelAdd (outParam);
		pService -> Unlock ();
	}
}

/*-------------------------------------------------------------
Handle a ChannelDelete indication
-------------------------------------------------------------*/
void Executive :: DAI_ChannelDelete_Ind (
	const UINT		/* nServiceId */,
	const DAIChannelDeleteParam& inParam
)
{
	for (int i = inParam.GetLength (); --i >= 0;)
	{	DataChannel *pChannel = FindChannel (inParam [i].m_nChid);
		if (pChannel != NULL)
		{	pChannel -> OnChannelDelete (inParam [i].m_eReason);
			pChannel -> Unlock ();
		}
	}
}

/*-------------------------------------------------------------
Handle a ChannelDelete confirmation
-------------------------------------------------------------*/
void Executive :: DAI_ChannelDelete_Cnf(
	const UINT		nServiceId,
	const DAIChannelDeleteParam& outParam
)
{
	ALManager *pService = FindService (nServiceId);
	if (pService != NULL)
	{	pService -> OnChannelDeleteCnf (outParam);
		pService -> Unlock ();
	}
}

/*-------------------------------------------------------------
Handle a UserCommandAck confirmation
-------------------------------------------------------------*/
void Executive :: DAI_UserCommandAck_Cnf(
	const UINT		nServiceId,
	const DAIUserCmdParam& outParam
)
{
	ALManager *pService = FindService (nServiceId);
	if (pService != NULL)
	{	pService -> OnUserCommandCnf (outParam);
		pService -> Unlock ();
	}
	return;
}

/*-------------------------------------------------------------
Handle receiving data
-------------------------------------------------------------*/
void Executive :: DAI_DataReceive_Ind(
	UINT			nChid,
	const BYTE*		pStreamData,
	const int		nStreamDataLength,
	const BYTE*		pAppData,
	const int		nAppDataLength,
	int				nErrorFlag
)
{
	DataChannel *pChannel = FindChannel (nChid);
	if (pChannel)
	{	pChannel ->OnPacketReceived (pStreamData, nStreamDataLength, pAppData, nAppDataLength, nErrorFlag);
		pChannel -> Unlock ();
	}
}

/*-------------------------------------------------------------
Functions to activate DMIF_Filter
-------------------------------------------------------------*/

/*-------------------------------------------------------------
Add the service to the list of services and send a ServiceAttach request
-------------------------------------------------------------*/
void Executive :: ServiceAttach (ALManager *pService, LPCSTR szURL, UINT nParentServiceID)
{
	m_mtxServices.P ();
	m_services += pService;
	m_mtxServices.V ();
	m_pDMIF ->DAI_ServiceAttach_Req ((UINT) pService, nParentServiceID, szURL, NULL, 0);
}

/*-------------------------------------------------------------
Remove the service from the list of services and send a ServiceDetach request
-------------------------------------------------------------*/
void Executive :: ServiceDetach (ALManager *pService, REASON reason)
{
	m_pDMIF ->DAI_ServiceDetach_Req (pService ->GetServiceID (), reason);
}

/*-------------------------------------------------------------
Send a ChannelAdd request
-------------------------------------------------------------*/
void Executive :: ChannelAdd (UINT nServiceID, const DAIChannelAddParam &inParam)
{
	m_pDMIF ->DAI_ChannelAdd_Req (nServiceID, inParam);
}

/*-------------------------------------------------------------
Send a ChannelDelete request
-------------------------------------------------------------*/
void Executive :: ChannelDelete (UINT nServiceID, const DAIChannelDeleteParam &inParam)
{
	m_pDMIF ->DAI_ChannelDelete_Req (nServiceID, inParam);
}

/*-------------------------------------------------------------
Send a SendCommand request
-------------------------------------------------------------*/
void Executive :: SendCommand (UINT nServiceID, const DAIUserCmdParam &inParam)
{
	m_pDMIF ->DAI_UserCommandAck_Req (nServiceID, inParam);
}

/*-------------------------------------------------------------
Send data on back channel
-------------------------------------------------------------*/
void Executive :: DataSend (
	UINT nServiceID,
	UINT nChannelID,
	LPBYTE pData,
	int nLength,
	int nAUnumber
)
{
	static SegmentInfo segInfo = {true, true, true, nAUnumber, false, false, false, 0, 0, nLength, 0, 0, 0};
	m_pDMIF ->DAI_DataSend_Req (
		nServiceID,
		nChannelID,
		pData,
		nLength,
		(const BYTE*) &segInfo,
		sizeof SegmentInfo
	);
}


/*-------------------------------------------------------------
RemoveNode is called when a node is removed from the scene. If the scene is still locked,
the function addes it to a list of removable nodes, otherwise it deletes it. Nodes that were
not deleted now will be deleted when the scene is unlocked. This prevents a crash when a node
is removing itself from the scene.
-------------------------------------------------------------*/
void Executive :: RemoveNode (MediaObject *pNode)
{
	if (m_nLocks <= 0)
		delete pNode;
	else
	{	pNode -> RemoveFields ();
		removedNodes += pNode;
	}
}

/*-------------------------------------------------------------
RemoveNode is called when a route is removed from the scene. If the scene is still locked,
the function addes it to a list of removable routes, otherwise it deletes it. Routes that were
not deleted now will be deleted when the scene is unlocked. This prevents a crash when a node
is removing the Route that activated it from the scene.
-------------------------------------------------------------*/
void Executive :: RemoveRoute (Route *pRoute)
{
	if (m_nLocks <= 0)
		delete pRoute;
	else
		removedRoutes += pRoute;
}

/*-------------------------------------------------------------
Delete all objects that were removed from the scene when it was locked.
Usually this function is called from UnlockScene, but it can also be called directly when
there is a need to delete the objects immediately.
-------------------------------------------------------------*/
void Executive :: RemoveSceneObjects ()
{
	int i, nSaveLocks = m_nLocks;
	m_nLocks = 0;		// to make sure objects will be deleted
	for (i = 0; i < removedNodes.GetLength (); i++)
		delete removedNodes [i];
	removedNodes.Clear ();
	for (i = 0; i < removedRoutes.GetLength (); i++)
		delete removedRoutes [i];
	removedRoutes.Clear ();
	m_nLocks = nSaveLocks;
}

/*-------------------------------------------------------------
Pause all the clocks
-------------------------------------------------------------*/
void Executive::Pause()
{
	m_mtxServices.P ();
	for (int i = m_services.GetLength (); --i >= 0;)
		m_services [i] -> Pause ();
	m_mtxServices.V ();
}

/*-------------------------------------------------------------
Resume all the clocks
-------------------------------------------------------------*/
void Executive::Resume()
{
	m_mtxServices.P ();
	for (int i = m_services.GetLength (); --i >= 0;)
		m_services [i] -> Resume ();
	m_mtxServices.V ();
}
