/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Executive
FILE:       Inline.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of InlineImp, the objects that implements Inline
-------------------------------------------------------------------------*/

#include "defs.h"
#include "Inline.h"
#include "Executiv.h"
#include "BifsDec.h"
#include "ODDec.h"
#include "StrParser.h"
#include "ODImp.h"
#include "Clock.h"
#include "CrNodes.h"
#include "IPMP.h"
#include "Stream.h"
#include "PresBase.h"

/*-------------------------------------------------------------
Create a decoder for an Inline node
-------------------------------------------------------------*/
Decoder *InlineImp :: CreateDecoder (
	DecoderConfigDescriptor *pDecoderConfig,
	QoS_Descriptor *,
	Decoder *
)
{
	GetOD () -> SetInline (this);

	/*-------------------------------------------------------------
	Depending on the stream type, we create either a BIFS decoder
	or an OD decoder
	-------------------------------------------------------------*/
	SysCoder *pCoder;
	switch (pDecoderConfig -> streamType) {
	case ST_SceneDescriptionStream:
		pCoder = new BIFSDecoder (
			GetExecutive (),
			this,
			GetExecutive () -> GetTraceFile (),
			"BIFS",
			GetExecutive () -> GetTraceFlags (),
			pDecoderConfig -> objectTypeIndication == 0xff ? 0 : 
			pDecoderConfig -> objectTypeIndication,
			m_bExternProto
		);
		break;
	case ST_ObjectDescriptorStream:
		if (!m_bExternProto)
		{	pCoder = new ODDecoder (
				GetExecutive (),
				this,
				GetExecutive () -> GetTraceFile (),
				"OD",
				GetExecutive () -> GetTraceFlags (),
				pDecoderConfig -> objectTypeIndication == 0xff ? 0 : 
				pDecoderConfig -> objectTypeIndication
			);
			break;
		}
	default:
		GetExecutive () -> Error (INVALID_STREAM_TYPE);
		return NULL;
	}
	
	DecoderImp *pParser = new StreamParser (GetExecutive (), pCoder, this);
	if (pDecoderConfig -> streamType == ST_SceneDescriptionStream)
		m_pBifsDecoder = pParser;
	return pParser;
}

/*-----------------------------------------------------------------
Render the object by rendering the associated Group object
-----------------------------------------------------------------*/
BOOL InlineImp :: Render (PresenterContext *pContext)
{
	if (m_subScene == NULL_OBJECT)
		return TRUE;
	DWORD dwSaveTime = pContext -> GetTime ();
	pContext ->SetTime (GetClock ());
	BOOL bResult = m_subScene -> Render (pContext);
	pContext ->SetTime (dwSaveTime);
	return bResult;
}

/*-------------------------------------------------------------
Initialize an Inline object after parsing it
-------------------------------------------------------------*/
BOOL InlineImp :: Parse (BIFSDecoder *pBifs)
{
	BOOL bSuccess = StreamConsumer :: Parse (pBifs);
	Init ();
	return bSuccess;
}

/*-----------------------------------------------------------------
Remove the entire sub-scene when the Inline is removed
-----------------------------------------------------------------*/
void InlineImp :: DeleteObject ()
{
	m_mtxStreams.P ();
	m_bActive = false;
	m_mtxStreams.V ();
	m_subScene.SetObject (NULL);
	StreamConsumer ::DeleteObject ();
	GetExecutive () ->GetPresenter () -> DeleteFlexTimeMgr (m_pFlexTimeMgr);
}

/*-------------------------------------------------------------
Initialize an Inline object
-------------------------------------------------------------*/
void InlineImp :: Init ()
{	
	Group *pSubScene = (Group*) m_subScene.CreateNode (1);
	pSubScene -> SetExecutive (GetExecutive ());
	m_subScene.SetObject (pSubScene, this);
	m_pFlexTimeMgr = GetExecutive () ->GetPresenter () -> CreateFlexTimeMgr ();

	/*-------------------------------------------------------------
	Inline node is open as soon as it is created.
	-------------------------------------------------------------*/
	OpenStream ();
}

/*-------------------------------------------------------------
Return a corresponding IPMP_Descriptor, if available
-------------------------------------------------------------*/
IPMP_DescriptorImp * InlineImp :: GetIPMP_Descriptor (int nIPMPid)
{
	int i = m_IPMPDs.Find (nIPMPid);
	if (i >= 0)
		return m_IPMPDs [i].m_pIPMPD;
	return NULL;
}

/*-------------------------------------------------------------
Execute the RemoveIPMPD command. The IPMP_Descriptor is removed from
the list of IPMP descriptors and all associated systems are notified.
-------------------------------------------------------------*/
void InlineImp :: RemoveIPMPD (int nIPMPid)
{
	int i = m_IPMPDs.Find (nIPMPid);
	if (i >= 0)
	{	for (int j = 0; j < m_IPMPDs [i].m_systems.GetLength (); j++)
			m_IPMPDs [i].m_systems [j] -> RemoveDescriptor (nIPMPid);
		m_IPMPDs.DeleteFrom (i);
	}
}

/*-----------------------------------------------------------------
An update IPMPD command was detected. Locate it in the IPMPD list, or
add it if its not there.
Then notify all attached IPMP systems
-----------------------------------------------------------------*/
void InlineImp :: UpdateIPMPD (IPMP_DescriptorImp *pIPMPD)
{
	int nIPMPid = pIPMPD -> IPMP_DescriptorID;
	int i = m_IPMPDs.Find (nIPMPid);
	if (i < 0)
		(m_IPMPDs += nIPMPid).m_pIPMPD = pIPMPD;
	else
	{	delete m_IPMPDs [i].m_pIPMPD;
		m_IPMPDs [i].m_pIPMPD = pIPMPD;
		for (int j = 0; j < m_IPMPDs [i].m_systems.GetLength (); j++)
		{	const BYTE *pData;
			int nDataSize;
			if (pIPMPD -> IPMPS_Type == 0)
			{	pData = pIPMPD -> URLString;
				nDataSize = pIPMPD -> URLString.GetLength ();
			}
			else
			{	pData = pIPMPD -> IPMP_data;
				nDataSize = pIPMPD -> IPMP_data.GetLength ();
			}
			pIPMPD -> Notify (m_IPMPDs [i].m_systems [j]);
		}
	}
}

/*-----------------------------------------------------------------
A remove ESD command was detected
-----------------------------------------------------------------*/
void InlineImp :: RemoveESD (int nODid, int nESDid)
{
	/*-------------------------------------------------------------
	Find the OD and check whether it contains the required ESD
	-------------------------------------------------------------*/
	ObjectDescriptorImp *pOD = NULL;
	ES_DescriptorImp *pESD = NULL;
	m_mtxStreams.P ();
	int i = m_streamAssociation.Find (nODid);
	if (i >= 0)
		pOD = m_streamAssociation [i].m_pOD;
	if (pOD == NULL)
	{	GetExecutive () -> Error (TRYING_TO_UPDATE_UNKNOWN_OD);
		m_mtxStreams.V ();
		return;
	}

	/*-------------------------------------------------------------
	Find the ESD
	-------------------------------------------------------------*/
	pESD = pOD -> FindESD (nESDid);
	if (pESD == NULL)
	{	m_mtxStreams.V ();
		return;
	}

	/*-------------------------------------------------------------
	Tell all the nodes associated to this OD that a the ESD has been
	removed.
	-------------------------------------------------------------*/
	for (int j = m_streamAssociation.GetLength (); --j >= 0;)
	{	if (nODid == m_streamAssociation [j].m_nODid)
		{	StreamConsumer *pNode = m_streamAssociation [j].m_pNode;
			if (pNode != NULL)
				pNode -> RemoveESD (pESD);
		}
	}
	pOD -> RemoveESD (nESDid);
	m_mtxStreams.V ();
	delete pESD;
}

/*-----------------------------------------------------------------
An Update ESD command was detected
-----------------------------------------------------------------*/
void InlineImp :: UpdateESD (int nODid, ES_DescriptorImp *esd)
{
	/*-------------------------------------------------------------
	Find the OD and add the stream to it (ignore the stream if already
	defined)
	-------------------------------------------------------------*/
	ObjectDescriptorImp *pOD = NULL;
	m_mtxStreams.P ();
	int i = m_streamAssociation.Find (nODid);
	if (i >= 0)
		pOD = m_streamAssociation [i].m_pOD;

	if (pOD == NULL)
	{	GetExecutive () -> Error (TRYING_TO_UPDATE_UNKNOWN_OD);
		m_mtxStreams.V ();
		return;
	}
	/*-------------------------------------------------------------
	If this ESD already exists, remove it before entering the new one
	-------------------------------------------------------------*/
	for (i = pOD -> esDescr.GetLength (); --i >= 0;)
	{	if (pOD -> esDescr [i] -> ES_ID == esd -> ES_ID)
		{	RemoveESD (nODid, esd -> ES_ID);
			break;
		}
	}
	if (!pOD -> UpdateESD (esd))
	{	m_mtxStreams.V ();
		return;
	}
	esd ->SetClock (pOD -> GetAL ());
	/*-------------------------------------------------------------
	Tell all the nodes associated to this OD that a new ESD has been
	added
	-------------------------------------------------------------*/
	for (i = m_streamAssociation.GetLength (); --i >= 0;)
	{	if (nODid == m_streamAssociation [i].m_nODid)
		{	StreamConsumer *pNode = m_streamAssociation [i].m_pNode;
			if (pNode != NULL)
				pNode -> AddESD (esd);
		}
	}
	m_mtxStreams.V ();
}

/*-----------------------------------------------------------------
A remove OD command was deteced. Remove the OD from the list
-----------------------------------------------------------------*/
void InlineImp :: RemoveOD (int nODid)
{
	ObjectDescriptorImp *pOD = NULL;
	m_mtxStreams.P ();
	for (int i = m_streamAssociation.GetLength (); --i >= 0;)
	{	if (nODid == m_streamAssociation [i].m_nODid)
		{	pOD = m_streamAssociation [i].m_pOD;
			m_streamAssociation [i].m_pOD = NULL;
			StreamConsumer *pNode = m_streamAssociation [i].m_pNode;
			if (pNode != NULL)
				pNode -> ODRemoved ();
			else
				m_streamAssociation.DeleteFrom (i);
		}
	}
	m_mtxStreams.V ();
	if (pOD == NULL)
		GetExecutive () -> Error (UNKNOWN_ODID);
	delete pOD;
}

/*-----------------------------------------------------------------
An update OD command was detected. Locate it in the OD list, or
add it if its not there. Once both the OD and the node which consumes
it are parsed, the proper channel(s) and decoder(s) are instantiated.
-----------------------------------------------------------------*/
void InlineImp :: UpdateOD (ObjectDescriptorImp *pOD)
{
	int nODid = pOD -> ObjectDescriptorID;
	pOD -> SetAL (GetOD () -> GetAL ());
	pOD -> SetInline (this);
	pOD -> SetClocks ();
	m_mtxStreams.P ();
	int i = m_streamAssociation.Find (nODid);
	if (i < 0)
		(m_streamAssociation += nODid).m_pOD = pOD;
	else
	{	/*-------------------------------------------------------------
		Tell all the nodes associated to this OD that a new ESD has been
		added
		-------------------------------------------------------------*/
		for (i = m_streamAssociation.GetLength (); --i >= 0;)
		{	if (nODid == m_streamAssociation [i].m_nODid)
			{	/*-------------------------------------------------------------
				If an OD with this ID already exist, remove it before entering
				the new one.
				-------------------------------------------------------------*/
				if (m_streamAssociation [i].m_pOD != NULL)
					RemoveOD (nODid);
				m_streamAssociation [i].m_pOD = pOD;
				pOD ->SetUseCount (1);
				StreamConsumer *pNode = m_streamAssociation [i].m_pNode;
				if (pNode)
					pNode -> SetStream (pOD);
			}
		}
	}
	m_mtxStreams.V ();
}

/*-----------------------------------------------------------------
The BIFS decoder parsed a ReplaceScene command
-----------------------------------------------------------------*/
void InlineImp :: ReplaceScene (Root *pScene)
{
	MFGenericNode *pChildren = m_subScene -> GetChildrenField ();
	pChildren -> SetAt (0, pScene, m_subScene, false);
	if (pScene != NULL_OBJECT)
	{	pScene -> ResetTime ();
		GetExecutive () -> ReplaceScene ();
	}
}

/*-----------------------------------------------------------------
A node of type StreamConsumer registers here when it is instantiated
-----------------------------------------------------------------*/
void InlineImp :: RegisterStreamConsumer (StreamConsumer *pNode)
{
	int nODid = pNode -> GetObjectDescriptorID ();
	ObjectDescriptorImp *pOD = NULL;
	int i;

	m_mtxStreams.P ();
	for (i = m_streamAssociation.GetLength (); --i >= 0;)
	{	if (nODid == m_streamAssociation [i].m_nODid)
		{	pOD = m_streamAssociation [i].m_pOD;
			if (m_streamAssociation [i].m_pNode == NULL)
			{	m_streamAssociation [i].m_pNode = pNode;
				break;
			}
		}
	}
	if (i < 0)
	{	i = m_streamAssociation.GetLength ();
		(m_streamAssociation += nODid).m_pNode = pNode;
	}
	if (pOD != NULL)
	{	m_streamAssociation [i].m_pOD = pOD;
		pOD ->SetUseCount (1);
		pOD ->AttachNode (pNode);
	}
	m_mtxStreams.V ();
}

/*-----------------------------------------------------------------
A node of type StreamConsumer unregisters here when it is destroyed
-----------------------------------------------------------------*/
void InlineImp :: UnregisterStreamConsumer  (StreamConsumer *pNode)
{
	m_mtxStreams.P ();
	for (int i = m_streamAssociation.GetLength (); --i >= 0;)
		if (m_streamAssociation [i].m_pNode == pNode)
		{	m_streamAssociation [i].m_pNode = NULL;
			ObjectDescriptorImp *pOD = m_streamAssociation [i].m_pOD;
			if (pOD == NULL || pOD ->SetUseCount (-1) > 0)
				m_streamAssociation.DeleteFrom (i);
		}
	m_mtxStreams.V ();
}

/*-----------------------------------------------------------------
A destructor for a StreamAssociation object
-----------------------------------------------------------------*/
InlineImp :: StreamAssociation :: ~StreamAssociation ()
{
	if (m_pOD && m_pOD ->SetUseCount (0) == 0)
		delete m_pOD;
}

/*-------------------------------------------------------------
Associate an IPMP system to an IPMPDescriptorID
-------------------------------------------------------------*/
void InlineImp :: AssociateIPMPSystem (int nID, IPMPManager * pIPMP)
{
	/*-------------------------------------------------------------
	Find the ID in the list, or create a new list item
	-------------------------------------------------------------*/
	int i = m_IPMPDs.Find (nID);
	if (i >= 0)
		m_IPMPDs [i].m_systems += pIPMP;
	else
		(m_IPMPDs += nID).m_systems += pIPMP;
}

/*-------------------------------------------------------------
The destructor for IPMPDAssociation deletes the descriptor
-------------------------------------------------------------*/
InlineImp :: IPMPDAssociation :: ~IPMPDAssociation ()
{
	delete m_pIPMPD;
}	 

/*-------------------------------------------------------------
Disassociate an IPMP system from a descriptor ID
-------------------------------------------------------------*/
void InlineImp :: RemoveIPMP (int nID, IPMPManager * pIPMP)
{
	int i = m_IPMPDs.Find (nID);
	if (i >= 0)
		m_IPMPDs [i].m_systems -= pIPMP;
}

/*-------------------------------------------------------------
When the OD of an Inline is removed, the entire branch is deleted
-------------------------------------------------------------*/
void InlineImp :: ODRemoved ()
{
	m_subScene.SetObject (NULL);
	GetExecutive () -> RemoveSceneObjects ();
	StreamConsumer ::ODRemoved ();
}		

/*-------------------------------------------------------------
Parse the url of the field (called when this object is used in ExternProto.
Also set the scene pointer of this object to be the root of the Proto object
that called this function.
-------------------------------------------------------------*/
void InlineImp :: ParseUrl (BIFSDecoder *pCoder)
{
	url.Parse (pCoder, NULL);
	Register ();
}	

/*-------------------------------------------------------------
Return a pointer to the scene under this Inline. If it is a special Inline
that implement ExternProto, the scene is the root of this node
-------------------------------------------------------------*/
Root *InlineImp :: GetScene () const
{
	if (m_bExternProto)
		return GetRoot ();
	else
		return m_subScene;
}

/*-------------------------------------------------------------
When a remote OD is connected, delegate it to all nodes that use this OD
-------------------------------------------------------------*/
void InlineImp :: ConnectRemoteOD (ObjectDescriptorImp *pOD)
{
	/*-------------------------------------------------------------
	If the OD association table is empty, this is the IOD of the scene, and the node
	is the current Inline node
	-------------------------------------------------------------*/
	m_mtxStreams.P ();
	if (m_bActive)
	{	if (m_streamAssociation.GetLength () == 0)
			OnConnect (pOD);
		else
		{	int nODid = pOD ->ObjectDescriptorID;
			for (int i = m_streamAssociation.GetLength (); --i >= 0;)
				if (m_streamAssociation [i].m_nODid == nODid &&
					m_streamAssociation [i].m_pNode != NULL
				)
					m_streamAssociation [i].m_pNode -> OnConnect (pOD);
		}
	}
	m_mtxStreams.V ();
	GetExecutive () -> RequestChannels ();
}

/*-------------------------------------------------------------
Get the clock of the BIFS stream
-------------------------------------------------------------*/
DWORD InlineImp :: GetClock () const
{
	if (m_pBifsDecoder)
	{	MediaStreamImp *pStream = (MediaStreamImp*) m_pBifsDecoder ->GetInputStream ();
		if (pStream)
			return pStream ->GetClock ();
	}
	return 0;
}

/*-------------------------------------------------------------
When a parser is removed check if it is the one stored in m_pBifsDecoder
-------------------------------------------------------------*/
void InlineImp :: RemoveDecoder (DecoderImp *pDecoder)
{
	if (pDecoder == m_pBifsDecoder)
		m_pBifsDecoder = NULL;
}

/*-----------------------------------------------------------------
An ObjectDescriptorExecute command was deteced. Find this OD and execute it.
-----------------------------------------------------------------*/
void InlineImp :: ExecuteOD (int nODid)
{
	ObjectDescriptorImp *pOD = NULL;
	m_mtxStreams.P ();
	for (int i = m_streamAssociation.GetLength (); --i >= 0;)
	{	if (nODid == m_streamAssociation [i].m_nODid)
		{	pOD = m_streamAssociation [i].m_pOD;
			if (pOD != NULL)
			{	pOD ->PreRoll ();
				break;
			}
		}
	}
	m_mtxStreams.V ();
	if (pOD == NULL)
		GetExecutive () -> Error (UNKNOWN_ODID);
}
