/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1997.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Decoder/ Compositor / Presenter
FILE:       MediaObj.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of base class for MediaObject objects.
-------------------------------------------------------------------------*/
//$Log: MediaObj.cpp,v $
//Revision 1.2  2000/03/03 18:41:21  Alex Cotarmanac'h
//Version 1 -- Alexandre Cotarmanac_h
//
//Modified 2001/05/08 Boris Felts boris.felts@polytechnique.org

#include "defs.h"
#include <assert.h>
#include "MediaObj.h"
#include "Executiv.h"
#include "BifsDec.h"
#include "Route.h"
#include "Presenter.h"
#include "NodeFld.h"
#include "bifs.h"
#include "Root.h"
#include "Quantize.h"
#include "Nodes.h"
#include "RendObj.h"
#include "Except.h"
#include "Proto.h"
#include "Inline.h"
#include "NodeCode.h"  //BF Script
#include "ScriptImp.h" // BF Script

/*-----------------------------------------------------------------
MediaObject constructor
-----------------------------------------------------------------*/
MediaObject :: MediaObject ()
{
	m_pRoot = NULL;
	m_nUse = 0;
	m_pFirstRoute = NULL;
	m_nID = -1;
	m_pExecutive = NULL;
	m_pRenderingObject = NULL;
	m_bIsRoot = FALSE;
	m_uCloningCounter = 0;

	/*-------------------------------------------------------------
	Grouping initialization is now part of every node
	-------------------------------------------------------------*/
#ifdef PLAYER
	CUSTOMIZE_GROUPNODE_INITIALIZATION
#endif
}

/*-------------------------------------------------------------
MediaObjects are deleted by a call to DeleteObject. This ensures that the rendering
object is deleted before the object itself.
-------------------------------------------------------------*/
void MediaObject :: DeleteObject ()
{
	/*-----------------------------------------------------------------
	Delete all routes which this is their arrival node
	-----------------------------------------------------------------*/
	while (m_pFirstRoute)
		m_pFirstRoute ->Remove ();  // this unlinks the route automatically

	/*-----------------------------------------------------------------
	If this node has an nID unregister it from the root
	-----------------------------------------------------------------*/
	if (m_pRoot && m_nID >= 0)
		m_pRoot -> UnregisterNode (this);

	/*-----------------------------------------------------------------
	Acknowledge all containers that this node does not exist
	-----------------------------------------------------------------*/
	for (int i = m_containers.GetLength (); --i >= 0;)
		if (m_containers [i] != NULL)
			if (m_containers [i].m_bISMF)
				((MFGenericNode *) m_containers [i].m_pContainer) ->
					RemoveNode (this);
			else
				((SFGenericNode *) m_containers [i].m_pContainer) ->
					RemoveNode ();
	if (m_pExecutive == NULL)
		delete this;
	else
		m_pExecutive ->RemoveNode (this);
}

/*-----------------------------------------------------------------
MediaObject destructor.
-----------------------------------------------------------------*/
MediaObject :: ~MediaObject ()
{
	delete m_pRenderingObject;
}

/*-----------------------------------------------------------------
Store the Executive pointer and perform object initialization
Create and attach a rendering object
-----------------------------------------------------------------*/
void MediaObject :: SetExecutive (Executive *pExecutive)
{
	m_pExecutive = pExecutive;
	if (pExecutive)
		m_pPresenter = (Presenter *) pExecutive -> GetPresenter ();
	RenderingObject *pObject = CreateRenderingObject (this);
	if (pObject)
	{	pObject -> Setup (this);
		m_pRenderingObject = pObject;
	}
}

/*-----------------------------------------------------------------
SetRoot stores a pointer to the root object of this object.
If this object has an ID, it registers it with the root.
-----------------------------------------------------------------*/
void MediaObject :: SetRoot (Root *pRoot)
{
	m_pRoot = pRoot;
	if (m_nID >= 0 && pRoot)
		pRoot -> RegisterNode (this);
}

/*-----------------------------------------------------------------
Set the node ID and register it with the node if necessary
-----------------------------------------------------------------*/
void MediaObject :: SetID (int nID)
{
	m_nID = nID;
	if (nID >= 0 && m_pRoot)
		m_pRoot -> RegisterNode (this);
}

/*-----------------------------------------------------------------
Parse a MediaObject by using the field table.
-----------------------------------------------------------------*/
BOOL MediaObject :: Parse (BIFSDecoder *pCoder)
{
	TRACEF ("%s {", GetName());
	TRACENEWLINE;
	TRACEINDENT (1);

	bool bResult = true, bProtoCode = IsProtoCode ();
	int nFieldCount = GetBifsBounds () -> m_nTotalDefs;
	if (bProtoCode)
		nFieldCount = GetBifsBounds () -> m_nTotalAll;

	/*-----------------------------------------------------------------
	A node may be described by a mask description or by a list
	description
	-----------------------------------------------------------------*/
	if (pCoder -> ParseInt (1, "maskAccess"))  // if MaskAccess //BF Conformance
	{	/*-------------------------------------------------------------
		A mask bit tell us whether it is followed by a field description
		If the defintion is part of Proto code the syntax is different because
		every field may be an IS. In this case we index into the entire structure,
		not just the DEFs
		-------------------------------------------------------------*/
		for (int i = 0; i < nFieldCount; i++)
		{
			if (pCoder -> ParseInt (1, "mask")) //BF Conformance
			{	if (!bProtoCode)
					bResult = ParseField (pCoder, GetBifsDefEntry (i));
				else if (pCoder -> ParseInt (1, "isedField"))
					ParseISref (pCoder, i);
				else
					bResult = ParseField (pCoder, GetFieldTable () + i);
			}
		}
	}
	else		// ListNodeDescription
	{
		/*-----------------------------------------------------------------
		Now loop for each field.
		-----------------------------------------------------------------*/
		while (!pCoder -> ParseInt (1, "endFlag"))	// while !endFlag
		{
			if (bProtoCode && pCoder -> ParseInt (1, "isedField"))
				ParseISref (
					pCoder,
					pCoder -> ParseInt (GetBifsBounds () -> m_nBitsPerAll, "def")
				);
			else
			{	int nIndex =
					pCoder -> ParseInt (GetBifsBounds () -> m_nBitsPerDef, "fieldRef"); // mp : conformance
				bResult = ParseField (pCoder, GetBifsDefEntry (nIndex));
			}
		}
	}

	TRACEINDENT (-1);
	TRACEF ("}", 0);
	TRACENEWLINE;
	SetUpdateTime ();
	return TRUE;
}

/*-----------------------------------------------------------------
Add a Route to the node (this is the arrival node)
-----------------------------------------------------------------*/
void MediaObject :: LinkRoute (Route *pRoute)
{
	pRoute -> LinkArrival (m_pFirstRoute);
}

/*-----------------------------------------------------------------
Unlink a Route from the list of routes (this is the arrival node)
-----------------------------------------------------------------*/
void MediaObject :: UnlinkRoute (Route *pRoute)
{
	pRoute -> UnlinkArrival (m_pFirstRoute);
}

/*-----------------------------------------------------------------
A route to this node was activated. If this is a field, update the
field and notify the node. If this is an eventIn, activate the event
handling function.
-----------------------------------------------------------------*/
void MediaObject :: RouteEvent (NodeField *pField, int nFieldIndex)
{
	if (nFieldIndex < 0 || nFieldIndex >= GetFieldCount ())
		throw BifsException (BIFX_PARSING_ERROR);
	GetField (nFieldIndex) -> Copy (*pField, this, false);
	UpdateField (GetFieldTable () + nFieldIndex);
}

/*-----------------------------------------------------------------
Calls Presenter :: GetTime () to retrieve the renderer "time", and
stamp the node with this value.  Used by the renderer to identify new
scene updates.
-----------------------------------------------------------------*/
void MediaObject :: SetUpdateTime ()
{
	if (m_pExecutive)
		m_dwUpdateTime = m_pExecutive -> GetPresenter () -> GetTime ();
}

/*-----------------------------------------------------------------
Called by BIFSDecoder to insert a value into one of the MFxxx fields
of the node.  The field ID, the new value and the exact insertion
position are parsed from the BIFS stream by this function. Returns TRUE
if successful.
-----------------------------------------------------------------*/
BOOL MediaObject :: InsertValue (BIFSDecoder *pCoder)
{
	int nPos;
	BOOL bResult = FALSE;
	int nIndex =
		pCoder -> ParseInt (GetBifsBounds () -> m_nBitsPerIn, "inID");
	const BifsFieldTable *pEntry = GetBifsInEntry (nIndex);
	if (pEntry == NULL)
		return FALSE;
	Quantizer quant (
		pEntry -> m_fMin,
		pEntry -> m_fMax,
		pEntry -> m_nDimension,
		pEntry -> m_nQuantizationType
	);
	TRACEF ("%s ", pEntry -> m_szName);

	MField *pField = (MField *) GetField (pEntry);
	switch (pCoder -> ParseInt (2, "type")) {
	case 0:				// insertion at a specified position
		nPos = pCoder -> ParseInt (16, "position");
		TRACEF ("[%d] ", nPos);
		bResult = pField -> InsertElement (pCoder, this, nPos, &quant);
		break;
	case 2: 			// insertion at the beginning of the field
		TRACEF ("[0] ", 0);
		bResult = pField -> InsertElement (pCoder, this, 0, &quant);
		break;
	case 3: 			// insertion at the end of the field
		TRACEF ("[end] ", 0);
		bResult = pField -> AppendElement (pCoder, this, &quant);
		break;
	}
	if (bResult)
	{	UpdateField (pEntry);
		pField -> OnFieldUpdate ();
	}
	return bResult;
}

/*-----------------------------------------------------------------
Called by BIFSDecoder to delete a value from one of the MFxxx fields
of the node.  The field ID and the element index are parsed from the
BIFS by this function.  Returns TRUE if successful.
-----------------------------------------------------------------*/
BOOL MediaObject :: DeleteValue (BIFSDecoder *pCoder)
{
	int nPos;
	BOOL bResult = FALSE;
	int nIndex =
		pCoder -> ParseInt (GetBifsBounds () -> m_nBitsPerIn, "inID");
	const BifsFieldTable *pEntry = GetBifsInEntry (nIndex);
	if (pEntry == NULL)
		return FALSE;
	TRACEF ("%s ", pEntry -> m_szName);

	MField *pField = (MField *) GetField (pEntry);
	switch (pCoder -> ParseInt (2, "type")) {
	case 0:				// deletion from a specified position
		nPos = pCoder -> ParseInt (16, "position");
		TRACEF ("[%d] ", nPos);
		if (nPos < pField -> GetLength ())
		{	pField -> DeleteFrom (nPos);
			bResult = TRUE;
		}
		break;
	case 2: 			// deletion from the beginning of the field
		TRACEF ("[0] ", 0);
		if (pField -> GetLength () > 0)
		{	pField -> DeleteFrom (0);
			bResult = TRUE;
		}
		break;
	case 3: 			// delete last element of field
		TRACEF ("[end] ", 0);
		if (pField -> GetLength () > 0)
		{	pField -> DeleteFrom (pField -> GetLength () - 1);
			bResult = TRUE;
		}
		break;
	}
	if (bResult)
		UpdateField (pEntry);
	return bResult;
}

/*-----------------------------------------------------------------
Called by BIFSDecoder to replace a value of one of the nodes fields.
The field ID and new value are parsed from the BIFS by this function.
Returns TRUE if successful.
-----------------------------------------------------------------*/
BOOL MediaObject :: ReplaceValue (BIFSDecoder *pCoder)
{
	int nIndex =
		pCoder -> ParseInt (GetBifsBounds () -> m_nBitsPerIn, "inID");
	const BifsFieldTable *pEntry = GetBifsInEntry (nIndex);
	if (pEntry == NULL)
		return FALSE;
	NodeField *pField = ParseField (pCoder, pEntry);
	if (pField)
	{	UpdateField (pEntry);
		pField -> OnFieldUpdate ();
		return TRUE;
	}
	return FALSE;
}

/*-----------------------------------------------------------------
Called by BIFSDecoder to replace a value of one of the elements of
one of the MFxxx fields of the node.  The field ID, the element index
and the new value are parsed from the BIFS by this function.  Returns
TRUE if successful.
-----------------------------------------------------------------*/
BOOL MediaObject :: ReplaceIndexedValue (BIFSDecoder *pCoder)
{
	int nPos;
	BOOL bResult = FALSE;
	int nIndex =
		pCoder -> ParseInt (GetBifsBounds () -> m_nBitsPerIn, "index");
	const BifsFieldTable *pEntry = GetBifsInEntry (nIndex);
	if (pEntry == NULL)
		return FALSE;

	// MBS: init quantizer with current QuantizationParameter node
	Quantizer quant(pCoder->GetCurrentQP(),
			pEntry -> m_fMin,
			pEntry -> m_fMax,
			pEntry -> m_nDimension,
			pEntry -> m_nQuantizationType);
	// end MBS

	TRACEF ("%s ", pEntry -> m_szName);

	MField *pField = (MField *) GetField (pEntry);
	switch (pCoder -> ParseInt (2, "type")) {
	case 0:				// replacement a specified position
		nPos = pCoder -> ParseInt (16, "position");
		TRACEF ("[%d] ", nPos);
		bResult = pField -> ReplaceElement (pCoder, this, nPos, &quant);
		break;
	case 1: 			// replacement at the beginning of the field
		TRACEF ("[0] ", 0);
		bResult = pField -> ReplaceElement (pCoder, this, 0, &quant);
		break;
	case 2: 			// Replacement at the end of the field
		TRACEF ("[end] ", 0);
		if (pField -> GetLength () <= 0)
		{	GetExecutive () -> Error (BIFS_UPDATE_REQUEST_DOES_NOT_MATCH_SCENE);
			return false;
		}
		bResult = pField -> ReplaceElement (
			pCoder,
			this,
			pField -> GetLength () - 1,
			&quant
		);
		break;
	}
	if (bResult)
	{	UpdateField (pEntry);
		pField -> OnFieldUpdate ();
	}
	return bResult;
}

/*-----------------------------------------------------------------
Similar to Parse(), this function is called by BIFSDecoder to insert
one node into the children field of the node.  The information about
the new node and its exact position are parsed from the BIFS stream by
this function.   Returns TRUE if successful.
-----------------------------------------------------------------*/
BOOL MediaObject :: ParseChild (BIFSDecoder *pCoder)
{
	int nPos;
	BOOL bResult = FALSE;
	MFGenericNode *pChildren = GetChildrenField ();
	if (pChildren == NULL)
		return false;

	switch (pCoder -> ParseInt (2, "type")) {
	case 0:				// insertion at a specified position
		nPos = pCoder -> ParseInt (8, "position");
		TRACEF ("[%d] ", nPos);
		bResult = pChildren -> InsertElement (pCoder, this, nPos);
		break;
	case 2: 			// insertion at the beginning of the field
		TRACEF ("[0] ", 0);
		bResult = pChildren -> InsertElement (pCoder, this, 0);
		break;
	case 3: 			// insertion at the end of the field
		TRACEF ("[end] ", 0);
		bResult = pChildren -> AppendElement (pCoder, this);
		break;
	}
	if (bResult)
	{	SetUpdateTime ();
		OnUpdate ();
		pChildren -> OnFieldUpdate ();
	}
	return bResult;
}

/*-----------------------------------------------------------------
Add a new Route whose departure field is one of the fields of this
node.  This departure field ID and arrival node ID are parsed from the
BIFS stream, then BindNode() is called in order to link the new Route
to the proper arrival field.  Returns TRUE if successful.
-----------------------------------------------------------------*/
BOOL MediaObject :: AddRoute (int nRouteID,  BIFSDecoder *pCoder, LPCSTR name)
{
	int nIndex =
		pCoder -> ParseInt (GetBifsBounds () -> m_nBitsPerOut, "outFieldRef"); //mp conformance
	nIndex = GetBifsOutEntry (nIndex);
	if (nIndex < 0)
		return FALSE;
	const BifsFieldTable *pEntry = GetFieldTable () + nIndex;
	TRACEF ("%s TO ", pEntry -> m_szName);
	Route *pRoute = new Route (nRouteID, GetField (pEntry), nIndex, this, name);

	int nNodeID = pCoder -> ParseInt (pCoder -> GetNodeIDbits (), "inNodeID"); //BF Conformance
	MediaObject *pNode = GetRoot () -> FindNode (nNodeID);
	if (pNode == NULL)
		return FALSE;
	TRACEF ("%s.", pNode ->GetNameID ());
	return pNode -> BindRoute (pRoute, pCoder);
}

/*-----------------------------------------------------------------
Having instantiated a new Route, and binding it to a departure field,
AddRoute() calls this function where it requests the node to bind the
Route to one of its fields as an arrival field.  The arrival field index
is parsed from the BIFS stream.  Returns TRUE if successful.
-----------------------------------------------------------------*/
BOOL MediaObject :: BindRoute (Route *pRoute, BIFSDecoder *pCoder)
{
	int nIndex =
		pCoder -> ParseInt (GetBifsBounds () -> m_nBitsPerIn, "inFieldRef"); //BF Conformance
	const BifsFieldTable *pTable = GetFieldTable ();

	for (int i = GetFieldCount (); --i >= 0;)
		if (nIndex == pTable [i].m_inID)
		{	TRACEF ("%s", pTable [i].m_szName);
			pRoute -> Bind (this, i);
			return TRUE;
		}

	return FALSE;
}

/*-----------------------------------------------------------------
This function is called after recalculation of the bounding-box.  The
function calls UpdateBBox() on all the nodes parents.
-----------------------------------------------------------------*/
void MediaObject :: OnBBoxUpdate ()
{
	for (int i = m_parents.GetLength (); --i >= 0;)
		m_parents [i] -> UpdateBBox ();
}

/*-----------------------------------------------------------------
Add the container field to the list of containers and the parent node
to the list of parents.
-----------------------------------------------------------------*/
void MediaObject :: SetContainer (
	NodeField *pContainer,
	BOOL bISMF,
	MediaObject *pParent
)
{
	if (pContainer == NULL)
		return;
	m_containers += Container (pContainer, bISMF);
	SetParent (pParent);
}

/*-----------------------------------------------------------------
Remove the container field from the list of containers and delete
the corresponding rendering object. If the node
is not contained any more in any other node, delete it.
-----------------------------------------------------------------*/
void MediaObject :: RemoveContainer (NodeField
	*pContainer,
	MediaObject *pParent
)
{
	if (!pContainer)
		return;
	int i = m_containers.Find (pContainer);
	if (i >= 0)
	{	RemoveParent (pParent);
		m_containers.DeleteFrom (i);
	}

	if (m_containers.GetLength () == 0)
		Remove ();
}

/*-----------------------------------------------------------------
Returns a pointer to the BIFS table entry for the field whose defID is
nID
-----------------------------------------------------------------*/
const BifsFieldTable * MediaObject :: GetBifsDefEntry (int nID)
{
	const BifsFieldTable *pTable = GetFieldTable ();
	for (int i = GetFieldCount (); --i >= 0;)
		if (nID == pTable [i].m_defID)
			return pTable + i;
	return NULL;
}

/*-----------------------------------------------------------------
Returns a pointer to the BIFS table entry for the field whose inID is
nID
-----------------------------------------------------------------*/
const BifsFieldTable * MediaObject :: GetBifsInEntry (int nID)
{
	const BifsFieldTable *pTable = GetFieldTable ();
	for (int i = GetFieldCount (); --i >= 0;)
		if (nID == pTable [i].m_inID)
			return pTable + i;
	return NULL;
}

/*-----------------------------------------------------------------
Returns a pointer to the BIFS table entry for the field whose outID is
nID
-----------------------------------------------------------------*/
int MediaObject :: GetBifsOutEntry (int nID)
{
	const BifsFieldTable *pTable = GetFieldTable ();
	for (int i = GetFieldCount (); --i >= 0;)
		if (nID == pTable [i].m_outID)
			return i;
	return -1;
}

/*-----------------------------------------------------------------
Returns a pointer to the BIFS table entry for the field whose dynID is
nID
-----------------------------------------------------------------*/
const BifsFieldTable * MediaObject :: GetBifsDynEntry (int nID)
{
	const BifsFieldTable *pTable = GetFieldTable ();
	for (int i = GetFieldCount (); --i >= 0;)
		if (nID == pTable [i].m_dynID)
			return pTable + i;
	return NULL;
}

/*-----------------------------------------------------------------
Parse a node field, when a pointer to the BIFS table entry describing
that field is given
-----------------------------------------------------------------*/
NodeField *MediaObject :: ParseField (
	BIFSDecoder *pCoder,
	const BifsFieldTable *pEntry
)
{
	if (pEntry == NULL)
		return FALSE;

	// MBS: init quantizer with current QuantizationParameter node
	Quantizer quant(pCoder->GetCurrentQP(),
			pEntry -> m_fMin,
			pEntry -> m_fMax,
			pEntry -> m_nDimension,
			pEntry -> m_nQuantizationType);
	// end MBS
//#ifdef PREDICTIVE	//AC 20000227 QP 14  //BF
	if (pEntry->m_nQuantizationType == 14 )
	{
		if (pCoder->isCoordActive()) {
			int k=pCoder->getLengthCoord(); // nb fields
			int nbits=(int) ceil(log(k+1)/log(2)); //so as to be able to store -1.
			// Ceil in the spec is math::ceil(x+1)
			quant.setQuantizationNbBits(nbits);
		}
	}
//#endif	//AC end  //BF

	NodeField *pField = GetField (pEntry);
    if (GetNodeCode () != CODE_Script || pField -> GetType() != MFSTRING)
	{ // "url" field in scripts will be written once the script field decoded
	    TRACEF ("%s ", pEntry -> m_szName);  //BF write the name of the field
    }
    
	if (pField)
	{	pField -> Parse (pCoder, this, &quant);
		pField ->OnFieldUpdate ();
	}
	TRACENEWLINE;
	return pField;
}

/*-----------------------------------------------------------------
A node is replaced by replaceNode BIFS command.
-----------------------------------------------------------------*/
void MediaObject :: ReplaceNode (BIFSDecoder *pCoder)
{
	/*-----------------------------------------------------------------
	Parse a new node.
	-----------------------------------------------------------------*/
	SFWorldNode node;
	node.ParseNodeField (pCoder, this, NULL, FALSE);

	/*-----------------------------------------------------------------
	Go through all the containers and replace
	the pointer to the old node by the new one
	-----------------------------------------------------------------*/
	for (int i = m_containers.GetLength (); --i >= 0;)
		if (m_containers [i] != NULL)
		{	if (m_containers [i].m_bISMF)
				((MFGenericNode *) m_containers [i].m_pContainer) ->
					ReplaceNode (this, node.GetNode ());
			else
				((SFGenericNode *) m_containers [i].m_pContainer) ->
					ReplaceNode (node.GetNode ());
			m_containers [i].m_pContainer = NULL;
		}

	/*-----------------------------------------------------------------
	Call OnUpdate on the new node then delete this (the old) node
	-----------------------------------------------------------------*/
	node.GetNode () -> SetUpdateTime ();
	node.GetNode () -> OnUpdate ();
	Remove ();
}

/*-----------------------------------------------------------------
A node was deleted by a DeleteNode BIFS command
-----------------------------------------------------------------*/
void MediaObject :: Delete ()
{
	/*-----------------------------------------------------------------
	Notify all the containers that the node was deleted
	-----------------------------------------------------------------*/
	for (int i = m_containers.GetLength (); --i >= 0;)
		if (m_containers [i] != NULL)
		{	if (m_containers [i].m_bISMF)
				((MFGenericNode *) m_containers [i].m_pContainer) ->
					RemoveNode (this);
			else
				((SFGenericNode *) m_containers [i].m_pContainer) ->
					RemoveNode ();
			NodeField *pField = m_containers [i].m_pContainer;
			pField -> OnFieldUpdate ();
			m_containers [i].m_pContainer = NULL;
		}

	/*-----------------------------------------------------------------
	Notify all parents that node was deleted
	-----------------------------------------------------------------*/
	for (i = m_parents.GetLength (); --i >= 0;)
	{	m_parents [i] -> SetUpdateTime ();
		m_parents [i] -> OnUpdate ();
	}

	/*-----------------------------------------------------------------
	Delete the node
	-----------------------------------------------------------------*/
	Remove ();
}

/*-----------------------------------------------------------------
A field was updated as a result of a Route or BIFS update
pEntry points to the entry in the BIFS table describing the field
-----------------------------------------------------------------*/
void MediaObject :: UpdateField (const BifsFieldTable *pEntry)
{
	EventHandler function = pEntry -> m_function;
	if (function == NULL)
	{	SetUpdateTime ();
		OnUpdate ();
	}
	else
		(*this.*function) ();
}

/*-------------------------------------------------------------
Parse the reference of an ISed field
-------------------------------------------------------------*/
void MediaObject :: ParseISref (BIFSDecoder *pCoder, int nField)
{
	const BifsFieldTable *pNodeEntry = GetFieldTable () + nField;
	TRACEF ("%s ", pNodeEntry -> m_szName);
	Proto *pProto = (Proto *) GetRoot ();
	int nBits = pProto -> GetBifsBounds () -> m_nBitsPerAll;
	int nArg = pCoder -> ParseInt (nBits, "protoField");
	const BifsFieldTable *pProtoEntry = pProto -> GetFieldTable () + nArg;
	TRACEF ("IS %s", pProtoEntry ->m_szName);
	TRACENEWLINE;

	/*-------------------------------------------------------------
	The field is ISed by linking it to the Proto's field through Routes.
	The Route will go from the Proto to the node if the Proto field is EventIn.
	The Route will go from the node to the Proto if the Proto field is EventOut.
	There will be 2 Routes going both ways if the Proto field is an ExposedField.
	In addition, if this is a Field or an ExposedField the value of the Proto field is copied
	to the node field
	-------------------------------------------------------------*/
	NodeField *pNodeField = GetField (pNodeEntry);
	NodeField *pProtoField = pProto -> GetField (nArg);
	if (pProtoEntry ->m_defID >= 0)
		pNodeField ->Copy(*pProtoField, this);
	if (pProtoEntry ->m_inID >= 0 || pProtoEntry ->m_defID >= 0)
	{	Route *pRoute = new Route (-1, pProtoField, nArg, pProto);
		pRoute ->Bind (this, nField);
	}
	if (pProtoEntry ->m_outID >= 0)
	{	Route *pRoute = new Route (-1, pNodeField, nField, this);
		pRoute -> Bind (pProto, nArg);
	}
}

/*-------------------------------------------------------------
Returns true if the definition of this node is part of Proto code
-------------------------------------------------------------*/
bool MediaObject :: IsProtoCode () const
{
	return m_pRoot && m_pRoot -> IsProto ();
}

/*-------------------------------------------------------------
Clone a node for the instantiation of a Proto
-------------------------------------------------------------*/
MediaObject * MediaObject :: Clone (Root *pScene)
{
	/*-------------------------------------------------------------
	If this is a used node (see explanation in Proto::Instantiate), don't instantiate
	a new node but return a pointer to a copy of the node that has already been instantiated
	-------------------------------------------------------------*/
	if (GetCloningCounter () == GetRoot () -> GetCloningCounter ())
		return m_pClone;
	IncrementCloningCounter ();

	/*-------------------------------------------------------------
	Create a node of same type
	-------------------------------------------------------------*/
	MediaObject *pNode = SFGenericNode ().CreateNode (GetNodeCode ());
	pNode ->SetExecutive (GetExecutive ());
	pNode ->SetRoot (pScene);
	m_pClone = pNode;

	/*-------------------------------------------------------------
	Clone the values of all the fields
	-------------------------------------------------------------*/
	for (int i = 0; i < GetFieldCount (); i++)
		GetField (i) ->Clone (pNode ->GetField (i), pNode);
	pNode -> SetUpdateTime ();
	return pNode;
}

/*-------------------------------------------------------------
Remove all the fields of the node without deleting the node. This is done when
a node is removed from the scene but not deleted phisically
-------------------------------------------------------------*/
void MediaObject  :: RemoveFields ()
{
	for (int i = 0; i < GetFieldCount (); i++)
		GetField (i) -> Remove ();
}


/*-------------------------------------------------------------
Parse the name ID in case CODENAMES is active
-------------------------------------------------------------*/
void MediaObject ::ParseNameID (BIFSDecoder *pCoder)
{
	pCoder ->ParseString (m_name);
}

/*-------------------------------------------------------------
Return the DEF name of the node
-------------------------------------------------------------*/
LPCSTR MediaObject :: GetNameID () const
{
	if (m_name.GetLength () > 0)
		return m_name;
	if (m_nID < 0)
		return NULL;
	static char buf [20];
	sprintf (buf, "ID%d", m_nID);//AC 20000228 Vrml-like
	return buf;
}

/*-------------------------------------------------------------
Delete a node physically
-------------------------------------------------------------*/
void MediaObject :: Remove()
{
	if (m_pRenderingObject)
		m_pRenderingObject ->DeleteObject ();
	DeleteObject ();
}

/*-------------------------------------------------------------
Get time of the scene this node belongs to
-------------------------------------------------------------*/
DWORD MediaObject :: GetClock () const
{
	if (GetRoot ())
		return GetRoot () -> GetSceneTime ();
	return 0;
}