/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1997.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Decoder/ Compositor
FILE:       Route.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of Route
-------------------------------------------------------------------------*/
#include "defs.h"
#include "MediaObj.h"
#include "NodeFld.h"
#include "Route.h"
#include "Root.h"
#include "Nodes.h"
#include "Executiv.h"

int Route :: m_nGlobalCounter = 0;

/*-----------------------------------------------------------------
Route constructor stores the route-id, departure node, and departure
field index, and link itself to the node field
-----------------------------------------------------------------*/
Route :: Route (
	int nRouteID,
	NodeField *pDepartureField,
	int nDepartureIndex,
	MediaObject *pDepartureNode,
	LPCSTR name
)
{
	m_pNext = m_pNextArrival = NULL;
	m_nCurrentCounter = 0;
	m_pArrivalNode = NULL;
	m_nRouteID = nRouteID;
	m_name = name;
	m_pDepartureField = pDepartureField;
	m_nDepartureField = nDepartureIndex;
	m_pDepartureNode = pDepartureNode;
	pDepartureField -> LinkRoute (this);
}

/*-----------------------------------------------------------------
Bind a Route to the arrival field, done by storing the field details
-----------------------------------------------------------------*/
void Route :: Bind (MediaObject *pArrivalNode, int nArrivalField)
{
	m_pArrivalNode = pArrivalNode;
	m_nArrivalField = nArrivalField;
	m_pArrivalNode -> LinkRoute (this);
	if (m_pArrivalNode -> GetRoot () != NULL)
		m_pArrivalNode -> GetRoot () -> RegisterRoute (this);
}

/*-------------------------------------------------------------
Activate a Route. The parameter bIsFirst is used to elimiate loops.
-------------------------------------------------------------*/
void Route :: Activate (bool bIsFirst)
{
	/*-------------------------------------------------------------
	Make sure we are not repeating this Route in a loop
	-------------------------------------------------------------*/
	if (bIsFirst)
		++m_nGlobalCounter;
	if (m_nCurrentCounter == m_nGlobalCounter)
		return;
	m_nCurrentCounter = m_nGlobalCounter;
	if (m_pArrivalNode)
		m_pArrivalNode -> RouteEvent (m_pDepartureField, m_nArrivalField);
}

/*-----------------------------------------------------------------
Links itself to the associated node field
-----------------------------------------------------------------*/
void Route :: Link (Route *&rpListHead)
{
	m_pNext = rpListHead;
	rpListHead = this;
}

/*-----------------------------------------------------------------
Unlinks itself from the associated field
-----------------------------------------------------------------*/
void Route :: Unlink (Route *&rpListHead)
{
	Route **ppPrevious = &rpListHead;
	while (*ppPrevious != this)
	{	if (*ppPrevious == NULL)
			return;
		ppPrevious = &(*ppPrevious) -> m_pNext;
	}
	*ppPrevious = m_pNext;
}

/*-----------------------------------------------------------------
Links itself to the associated arrival node
-----------------------------------------------------------------*/
void Route :: LinkArrival (Route *&rpListHead)
{
	m_pNextArrival = rpListHead;
	rpListHead = this;
}

/*-----------------------------------------------------------------
Unlinks itself from the associated arrival node
-----------------------------------------------------------------*/
void Route :: UnlinkArrival (Route *&rpListHead)
{
	Route **ppPrevious = &rpListHead;
	while (*ppPrevious != this)
	{	if (*ppPrevious == NULL)
			return;
		ppPrevious = &(*ppPrevious) -> m_pNextArrival;
	}
	*ppPrevious = m_pNextArrival;
}

/*-------------------------------------------------------------
Clone all the Routes of a Proto
-------------------------------------------------------------*/
Route * Route :: Clone ()
{
	/*-------------------------------------------------------------
	A Route is cloned by instantiating a new Route object and linking it to
	the cloned node object
	-------------------------------------------------------------*/
	Route *pClone = new Route (
						0,
						m_pDepartureNode -> GetClone () -> GetField (m_nDepartureField),
						m_nDepartureField,
						m_pDepartureNode -> GetClone ()
					);
	pClone -> Bind (m_pArrivalNode -> GetClone (), m_nArrivalField);
	return pClone;
}

/*-------------------------------------------------------------
Return the textual ID of a Route
-------------------------------------------------------------*/
LPCSTR Route :: GetName () const
{
	if (m_name.GetLength () > 0)
		return m_name;
	if (m_nRouteID < 0)
		return NULL;
	static char buf [20];
	sprintf (buf, "%d", m_nRouteID);
	return buf;
}
	
/*-------------------------------------------------------------
Remove a Route from the scene
-------------------------------------------------------------*/
void Route :: Remove ()
{
	if (m_pDepartureField)
		m_pDepartureField -> UnlinkRoute (this);
	if (m_pArrivalNode)
	{	if (m_pArrivalNode -> GetRoot () != NULL)
			m_pArrivalNode -> GetRoot () -> UnregisterRoute (this);
		m_pArrivalNode -> UnlinkRoute (this);
		Executive *pExecutive = m_pArrivalNode ->GetExecutive ();
		if (pExecutive)
		{	pExecutive -> RemoveRoute (this);
			return;
		}
	}
	delete this;
}
