/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Decoder/ Compositor / Presenter
FILE:       StreamCo.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of StreamConsumer
-------------------------------------------------------------------------*/
#include "defs.h"
#include <assert.h>
#include "StreamCo.h"
#include "Executiv.h"
#include "ODImp.h"
#include "Stream.h"
#include "InLine.h"
#include "BifsDec.h"
#include "ODDec.h"
#include "PresBase.h"
#include "AL.h"
#include "NodeCode.h"

/*
#include "Caps.h"
*/

/*-----------------------------------------------------------------
StreamConsumer constructor
-----------------------------------------------------------------*/
StreamConsumer :: StreamConsumer ()
{
	m_pStream = NULL;
	m_pOD = NULL;
	m_nODid = m_nURLIndex = -1;
	m_bStreamOpen = false;
	m_bActive = true;
	m_dwMediaTime = (DWORD) -1;

	CUSTOMIZE_STREAMCONSUMER_CONSTRUCTOR
}

/*-----------------------------------------------------------------
Deletion of StreamConsumer. First flush out everything from the stream
(to make sure no object is waiting on the stream, then delete the
associated decoder and stream.
-----------------------------------------------------------------*/
void StreamConsumer :: DeleteObject ()
{
	CUSTOMIZE_STREAMCONSUMER_DESTRUCTOR

	m_mtxLock.P ();
	m_bActive = false;
	m_mtxLock.V ();
	RemoveStream ();
	if (m_pOD)
		m_pOD -> Detach ();
	if (m_nODid >= 0 && m_pSource != NULL)
		m_pSource -> UnregisterStreamConsumer (this);
	MediaObject :: DeleteObject ();
}

/*-----------------------------------------------------------------
Overload Parse(). After parsing in the usual way, the ODid field
is checked. If has value, register with Executive.
-----------------------------------------------------------------*/
BOOL StreamConsumer :: Parse (BIFSDecoder *pBifs)
{
	BOOL bResult = MediaObject :: Parse (pBifs);
	m_pSource = pBifs -> GetSource ();
	if (GetRoot () && !GetRoot () -> IsProto ())
		Register ();
	return bResult;
}
 
/*-----------------------------------------------------------------
Overload Close(). After parsing in the usual way, the ODid field
is checked. If has value, register with Executive.
-----------------------------------------------------------------*/
MediaObject *StreamConsumer :: Clone (Root *pScene)
{
	StreamConsumer *pObject = (StreamConsumer*) MediaObject :: Clone (pScene);
	pObject -> m_pSource = m_pSource;
	if (pScene && !pScene -> IsProto ())
		pObject -> Register ();
	return pObject;
}

/*-----------------------------------------------------------------
Instantiate the proper decoder and attach it to the object through
streams
-----------------------------------------------------------------*/
void StreamConsumer :: SetStream (ObjectDescriptorImp *pOD)
{
	m_mtxLock.P ();
	m_bActive = true;
	bool bStreamOpen = m_bStreamOpen;
	m_bStreamOpen = false;	// don't open stream till after call to SetStream
	m_pOD = pOD;
	MediaStreamImp *pStream = pOD -> SetStream (this);
	if (pStream != NULL)
		m_pStream = pStream;
	m_mtxLock.V ();
	if (bStreamOpen)		// if necessary open stream now.
		OpenStream ();
}

/*-------------------------------------------------------------
Stop playing from the stream and remove the OD
-------------------------------------------------------------*/
void StreamConsumer :: ODRemoved ()
{
	m_mtxLock.P ();
	m_bActive = false;
	m_mtxLock.V ();
	bool bOpen = m_bStreamOpen;	// if already open, stream remains flagges open although
								// physically closed.
	RemoveStream ();
	m_pOD -> Detach ();
	m_pOD = NULL;
	m_bStreamOpen = bOpen;
}

/*-------------------------------------------------------------
One ESD was added to the OD used by this node.
The consequence may be the attachment of a new stream to the current 
node.
-------------------------------------------------------------*/
void StreamConsumer :: AddESD (ES_DescriptorImp *pESD)
{
	/*-------------------------------------------------------------
	A special treatment for IPMP streams
	-------------------------------------------------------------*/
	if (pESD -> decConfigDescr == NULL_DESC)
		return;
	if (pESD -> decConfigDescr -> streamType == ST_IPMPStream)
	{	Reopen ();
		return;
	}

	/*-------------------------------------------------------------
	Add the stream. This may result in attaching new MediaStream
	to the node
	-------------------------------------------------------------*/
	MediaStreamImp *pStream = pESD -> AddESD (m_pOD, this);
	if (pStream != NULL)
		m_pStream = pStream;
	if (m_bStreamOpen)		// if necessary open stream now.
		pESD -> Start (false);
}

/*-------------------------------------------------------------
One ESD was removed from the OD used by this node.
-------------------------------------------------------------*/
void StreamConsumer :: RemoveESD (ES_DescriptorImp *pESD)
{
	/*-------------------------------------------------------------
	A special treatment for IPMP streams
	-------------------------------------------------------------*/
	if (pESD -> decConfigDescr == NULL_DESC)
		return;
	if (pESD -> decConfigDescr -> streamType == ST_IPMPStream)
	{	Reopen ();
		return;
	}

	/*-------------------------------------------------------------
	Remove the stream. This may result in detaching the MediaStream
	assigned to this node
	-------------------------------------------------------------*/
	if (m_pStream != NULL && pESD -> IsMyStream (m_pStream))
		RemoveStream ();
	else if (m_bStreamOpen)
		pESD -> Stop ();
	pESD -> Detach ();
}

/*-----------------------------------------------------------------
Request transmission of data on the channel
-----------------------------------------------------------------*/
void StreamConsumer :: OpenStream ()
{
	if (!m_bStreamOpen)
	{	m_bStreamOpen = true;
		m_dwMediaTime = (DWORD) -1;
		if (m_pOD)
		{	if (m_pStream)
				m_pStream -> Start ();
			m_pOD -> Start ();
		}
	}
}

/*-----------------------------------------------------------------
Close the stream and notify the executive
-----------------------------------------------------------------*/
void StreamConsumer :: CloseStream ()
{
	if (m_bStreamOpen)
	{	m_bStreamOpen = false;
		if (m_pOD)
		{	if (m_pStream)
				m_pStream -> Terminate ();
			m_pOD -> Stop ();
		}
	}
}

/*-----------------------------------------------------------------
The default implementation of GetCapabilities returns all the 
capabilities of the presenter
-----------------------------------------------------------------*/
void StreamConsumer :: GetCapabilities (Capabilities *pCaps)
{
	GetExecutive () -> GetPresenter () -> GetCapabilities (pCaps);
}

/*-------------------------------------------------------------
A session with remoteOD has been established
-------------------------------------------------------------*/
void StreamConsumer :: OnConnect (ObjectDescriptorImp *pOD)
{
	m_mtxLock.P ();
	if (m_bActive)
		SetStream (pOD);
	m_mtxLock.V ();
}

/*-------------------------------------------------------------
Return the decoder attached to this node
-------------------------------------------------------------*/
Decoder *StreamConsumer :: GetDecoder () const
{
	return m_pOD == NULL ? NULL : m_pOD -> GetDecoder ();
}

/*-------------------------------------------------------------
In case an IPMP stream is added or removed to/from the OD, stop playing
the stream, close the decoder, and reopen all with the new IPMP stream
-------------------------------------------------------------*/
void StreamConsumer :: Reopen ()
{
	bool bStreamOpen = m_bStreamOpen;
	CloseStream ();
	m_pOD -> Detach ();
	m_pStream = NULL;
	m_bStreamOpen = bStreamOpen;
	SetStream (m_pOD);
}

/*-------------------------------------------------------------
Write data to a back channel
-------------------------------------------------------------*/
void StreamConsumer :: WriteToStream (LPBYTE pData, int nLength)
{
	if (m_pOD)
		m_pOD -> DataSend (pData, nLength);
}

/*-------------------------------------------------------------
Register the node as a stream consumer.
-------------------------------------------------------------*/
void StreamConsumer :: Register ()
{
	if (url.GetLength () == 0)
		return;

	/*-------------------------------------------------------------
	If this is an ODid only URL save the ODid, else parse the URL list
	-------------------------------------------------------------*/
	if (url [0] >= 0)
		m_nODid = url [0];
	else
	{	for (int i = 0; i < url.GetLength (); i++)
		{	ZString szURL (url [i].urlValue);
			LPSTR szSegment = strchr (szURL, ':');
			if (szSegment != NULL)
			{	/*-------------------------------------------------------------
				Segments are allowed only in MediaSensor or MediaControl
				-------------------------------------------------------------*/
				if (GetNodeCode () != CODE_MediaSensor &&
					GetNodeCode () != CODE_MediaControl &&
					GetNodeCode () != CODE_TemporalTransform
				)
					GetExecutive () ->Error (ONLY_OD_ID_URL_ALLOWED, (LPCSTR) szURL);
				*szSegment++ = 0;
				szSegment += strspn (szSegment, " ");
				(++m_segments).m_segmentName = szSegment;
			}
			int nODid = atol (szURL);
			if (m_nODid < 0)
				m_nODid = nODid;
			else if (m_nODid != nODid)
				GetExecutive () ->Error (URL_MUST_REFER_TO_A_SINGLE_OD, (LPCSTR) szURL);
		}
	}
	if (m_nODid >= 0)
		m_pSource -> RegisterStreamConsumer (this);
}

/*-------------------------------------------------------------
Get the time stamp of the access unit that is currently ready for retrival from
the composition buffer.
The funtions return -1 if no access unit is ready yet, if the last access unit was fetched
and stream has arrived to end or if the available access unit is outside the bounds of the 
url.
If lStart or lLength (or both) are not -1, the info is retrieved only for the section that is 
inside these bounds, and -1 is returned otherwise
-------------------------------------------------------------*/
DWORD StreamConsumer :: GetMediaPosition (long lStart, long lLength)
{
	int nSize, nStatus = 0;
	DWORD dwTime;
	BOOL bEOU;

	if (m_pStream == NULL)
		return (DWORD) -1;

	/*-------------------------------------------------------------
	Calcualte the time of end of section
	-------------------------------------------------------------*/
	if (lStart > 0 && lLength >= 0)
		lLength += lStart;

	/*-------------------------------------------------------------
	Now get acutal media time and check if within the section bound
	-------------------------------------------------------------*/
	if (m_pStream ->Fetch (nSize, dwTime, bEOU, nStatus, true, false, 0, -1) != NULL)
		m_dwMediaTime = dwTime;

	/*-------------------------------------------------------------
	If the URL defines a segment return -1 if the access unit is outside the segment
	-------------------------------------------------------------*/
	if (m_dStart > 0)
	{	double dTS = dwTime / (double) 1000;
		if (dTS < m_dStart ||
			m_dDuration >= 0 && dTS >= m_dStart + m_dDuration
		)
			return (DWORD) -1;
		if (lStart >= 0)
			lStart += m_dStart * 1000;
	}

	/*-------------------------------------------------------------
	Check if withing the bounds defined by lStart and lLength
	-------------------------------------------------------------*/
	if (lStart >= 0 && m_dwMediaTime < (DWORD) lStart ||
		lLength < 0 && nStatus < 0 ||
		lLength >= 0 &&
		(	m_dwMediaTime > (DWORD) lLength ||
			nStatus != 0 && m_dwMediaTime > m_pStream ->GetClock ()
		)
	)
		return (DWORD) -1;
	return m_dwMediaTime;
}

/*-------------------------------------------------------------
Fetch a frame from the associated MediaStream object, discarding frames that are
outside the active segment.
When a segment reaches its end, switch to the next element of the URL field
-------------------------------------------------------------*/
LPVOID StreamConsumer :: Fetch (
	int &rnSize,
	DWORD &rdwTime,
	BOOL &bEOU,
	int &rnStatus,
	BOOL bSkip,
	BOOL bWait,
	long lDelta,
	int nFlag
)
{
	LPVOID pFrame = NULL;
	if (m_bStreamOpen && m_pStream != NULL)
	{	pFrame =
			m_pStream -> Fetch (rnSize, rdwTime, bEOU, rnStatus, bSkip, bWait, lDelta, nFlag);
		if (pFrame)
		{	/*-------------------------------------------------------------
			Check whether the frame is within the stream interval set by an associated
			MediaControl node
			-------------------------------------------------------------*/
			double dTS =
				rdwTime / (double) 1000 +
				m_pStream ->GetClockReference () ->GetMediaTimeOffset ();
			if (!m_pStream ->GetClockReference () ->IsUnitPlayable (dTS))
			{	Release (FALSE);
				return NULL;
			}
		}
		return pFrame;
	}
	return NULL;
}

/*-------------------------------------------------------------
Release a previously fetched frame
-------------------------------------------------------------*/
void StreamConsumer :: Release (bool bDiscard)
{
	if (m_pStream)
		m_pStream ->Release (bDiscard);
}

/*-------------------------------------------------------------
Make the first URL element active
-------------------------------------------------------------*/
void StreamConsumer :: ResetURL ()
{
	m_dStart = m_dDuration = -1;
	NextURL ();
}

/*-------------------------------------------------------------
Find a match between a segment name from the OD and the URL and if found store
segment attributes
-------------------------------------------------------------*/
void StreamConsumer :: SetSegment (LPCSTR szName, double dStart, double dDuration)
{
	for (int i = m_segments.GetLength (); --i >= 0;)
	{	if (!strcmp (szName, m_segments [i].m_segmentName))
		{	m_segments [i].m_dStart = dStart;
			m_segments [i].m_dDuration = dDuration;
		}
	}
}

/*-------------------------------------------------------------
Switch to the next element in the URL list. Return false if last one was used.
-------------------------------------------------------------*/
bool StreamConsumer :: NextURL ()
{
	double dStart = -1, dDuration = -1;
	m_nURLIndex = -1;
	for (int i = 0; i < m_segments.GetLength (); i++)
		if ((	m_segments [i].m_dStart > m_dStart ||
				m_dDuration >= 0 && m_segments [i].m_dStart == m_dStart &&
				(m_segments [i].m_dDuration < 0 || m_segments [i].m_dDuration > m_dDuration)
			) &&
			(	dStart < 0 || m_segments [i].m_dStart < dStart ||
				m_segments [i].m_dStart == dStart &&
				(	dDuration < 0 ||
					m_segments [i].m_dDuration < dDuration
				)
			)
		)
		{	m_nURLIndex = i;
			dStart = m_segments [i].m_dStart;
			dDuration = m_segments [i].m_dDuration;
		}
	if (dStart >= 0)
	{	m_dStart = dStart;
		m_dDuration = dDuration;
	}
	return m_nURLIndex >= 0;
}

/*-------------------------------------------------------------
A virtual function that is called when the stream attached to the node is removed
-------------------------------------------------------------*/
void StreamConsumer :: RemoveStream ()
{
	CloseStream ();
	m_pStream = NULL;
	m_segments.Clear ();
	m_dStart = m_dDuration = -1;
}
