/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1998.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

#if !defined(_ztype_h)	// Sentry, use file only if it's not already included.
#define _ztype_h

//ML - added the next include line
#include <math.h>

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Compositor
FILE:       ztype.h
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   class definitions for types of node fields
-------------------------------------------------------------------------*/

/*-----------------------------------------------------------------
Bool
-----------------------------------------------------------------*/
class Bool
{
	bool m_b;

public:
	Bool (bool b = false) : m_b (b) {}
	operator bool () const {return m_b;}
	bool operator = (bool b) {return m_b = b;}
	bool operator &= (bool b) {return m_b &= b;}
	bool operator |= (bool b) {return m_b |= b;}
	bool operator ^= (bool b) {return m_b ^= b;}
	bool operator ++ () {return ++m_b;}
};

/*-----------------------------------------------------------------
Time
-----------------------------------------------------------------*/
typedef double Time;

/*-----------------------------------------------------------------
Vec3f
-----------------------------------------------------------------*/
class Vec3f
{
public:
	float x, y, z;

public:
	Vec3f (float xx = 0, float yy = 0, float zz = 0) :
		x (xx), y (yy), z(zz)
	{}
	bool operator == (const Vec3f &vec) const
	{
		return x == vec.x && y == vec.y && z == vec.z;
	}
	// MBS: operator added
	Vec3f operator += (const Vec3f &v) {
	  x += v.x;
	  y += v.y;
	  z += v.z;
	  return *this;
	}

	//ML - added the rest of the functions in the class
	float GetLength()
	{
		return sqrt(x * x + y * y + z * z);
	}

	void Normalize()
	{
		float mag = GetLength();
		if (mag == 0.0f)
			mag = 1.0f;
		x /= mag;
		y /= mag;
		z /= mag;
	}

	Vec3f& Cross (const Vec3f &vec)
	{
		float new_x, new_y, new_z;
		new_x = y * vec.z - z * vec.y;
		new_y = z * vec.x - x * vec.z;
		new_z = x * vec.y - y * vec.x;
		x = new_x;
		y = new_y;
		z = new_z;
		return *this;
	}

	float operator * (const Vec3f &vec) const
	{
		return (x * vec.x +
			y * vec.y +
			z * vec.z);
	}

	Vec3f& operator -= (const Vec3f &vec)
	{
		x -= vec.x;
		y -= vec.y;
		z -= vec.z;
		return *this;
	}

	operator const float* () const
	{
		return &x;
	}

	Vec3f& operator *= (float scale)
	{
		x *= scale;
		y *= scale;
		z *= scale;
		return *this;
	}
};

/*-----------------------------------------------------------------
Vec2f
-----------------------------------------------------------------*/
class Vec2f
{
public:
	float x, y;

public:
	Vec2f (float xx = 0, float yy = 0) :
		x (xx), y (yy)
	{}
	bool operator == (const Vec2f &vec) const
	{
		return x == vec.x && y == vec.y;
	}
	// MBS: operator added
	Vec2f operator += (const Vec2f &v) {
	  x += v.x;
	  y += v.y;
	  return *this;
	}
};

/*-----------------------------------------------------------------
Rotation
-----------------------------------------------------------------*/
class Rotation
{
public:
	float x, y, z, theta;

public:
	Rotation (float xx = 0, float yy = 0, float zz = 0, float th = 0) :
		x (xx), y (yy), z (zz), theta (th)
	{}
	bool operator == (const Rotation &vec) const
	{
		return
			x == vec.x && y == vec.y && z == vec.z && theta == vec.theta;
	}
	void Normalize ();
};

/*-----------------------------------------------------------------
Rotation2D
-----------------------------------------------------------------*/
class Rotation2D
{
public:
	float x, y, theta;

public:
	Rotation2D (float xx = 0, float yy = 0, float th = 0) :
		x (xx), y (yy), theta (th)
	{}
	bool operator == (const Rotation2D &vec) const
	{
		return
			x == vec.x && y == vec.y && theta == vec.theta;
	}
};

/*-----------------------------------------------------------------
Color
-----------------------------------------------------------------*/
class Color3f
{
public:
	float r, g, b;

public:
	Color3f (float rr = 0, float gg = 0, float bb = 0) :
		r (rr), g (gg), b(bb)
	{}
	bool operator == (const Color3f &color) const
	{
		return r == color.r && g == color.g && b == color.b;
	}

	// SB @@@@ cast Color3f to float* for OpenGL >>>>
	operator const float* () const {return &r;}
};

/*-----------------------------------------------------------------
Image
-----------------------------------------------------------------*/
//MBS 110498: added constructor Image(w,h,components)
//Alvar 981223: Added pixel array allocation and refcount stuff
class Image
{
public:
	int width;
	int height;
	int numComponents;
	BYTE *pixels;
	int *m_pRefCnt;  // reference count for pixel array

	Image (int w, int h, int components);      // Mike B-S ctor
	Image (int = 0);                           // Zvi ctor
	Image (const Image &i);                      // copy ctor
	Image& operator= (const Image &i);           // assignment
	~Image ();
};

#endif                                      // _ztype_h sentry.
