/*
 * VsFields.cpp
 *
 *     Implementation of classes:
 *        VsSFInt
 *	  VsFFloat
 *	  VsSFVec3f
 *
 * Subclasses of VsVar that take care of interfacing with VRML
 * fields/eventOuts.
 * 
 * Copyright (C) 1996, Silicon Graphics,  Inc.
 */

/*---------------------------------------------------------------
This module is copied from the VRML Script interpreter owned by
Silicon Graphics, Inc. It is modified by Wenlong Dong, Wei Wu and
Yuval Fisher(Rockwell Science Center) in the course of development
of the MPEG-4 Systems standard(ISO/IEC 14496-1) and added to the 
subsystem: Script Node.
---------------------------------------------------------------*/

#include "defs.h"
#include <assert.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>
#include "vsfields.h"
#include "NodeFld.h"
#include "VSEvents.h"

//#ifdef __sun 
#define _ASSERT(a) assert(a)
//#endif

VsField *
VsField::create(FIELD_TYPE t, Permission p)
{
    VsField *result = NULL;
    switch(t) 
	{
    case SFINT32:
    case SFBOOL:
		result = new VsSFInt(t);
		break;
    case SFFLOAT:
    case SFTIME:
		result = new VsSFFloat(t);
		break;
    case SFVEC2F:
    case SFVEC3F:
    case SFCOLOR:
    case SFROTATION:
		result = new VsSFFloatArray(t);
		break;
    case SFSTRING:
		result = new VsSFString(t);
		break;
    case SFNODE:	// NODEFIELD
		result = new VsSFVoid(t);
		break;
    case MFINT32:
		result = new VsMFInt(t);
		break;
    case MFFLOAT:
    case MFTIME:
		result = new VsMFFloat(t);
		break;
    case MFSTRING:
		result = new VsMFString(t);
		break;
    case MFVEC2F:
    case MFVEC3F:
    case MFCOLOR:
    case MFROTATION:
        result = new VsMFFloatArray(t);
        break;
//  case MFNODE:
//		result = new VsMFVoid(t);
//		break;
    }

    if (result)
        result->permission = p;

    return result;
}


// VsSFInt handles SFInt32 and SFBool fields

VsSFInt::VsSFInt(FIELD_TYPE t)
{
    type = t;
}

VsSFInt::~VsSFInt()
{
} 

VsValue
VsSFInt::evaluate(VsFunctionCall *)
{
    return myValue;
}

VsValue
VsSFInt::assign(const VsValue &v, VsFunctionCall *)
{
    int n = 0;

    switch (v.type) {
    case VsValue::vNUMBER:
		if (type == SFBOOL)
			n = (v.number != 0.0);
		else
			n = (int) v.number;
		break; 
    case VsValue::vSTRING:
		if (type == SFBOOL)
			n = (v.string != NULL && strlen(v.string) != 0);
		else
			n = (int) atof(v.string);
		break;  
	default:
		// ??? Warn if other types ???
		break;
    }

    if (permission == VsField::RW) 
	{
		myValue.number = n;

		if (myField)
		    myField->setHasBeenSet(TRUE);
		*(SF<int> *)(myNodeField) = n;
    }

    return myValue;
}

void
VsSFInt::setField(VSNodeField *f)
{
    assert(f->GetType() == type);

    myField = f;
    setField(f->getNodeField());
}

void VsSFInt::setField(NodeField *f) {
    myNodeField = f;

    myValue.type = VsValue::vNUMBER;

    if (type == SFBOOL) 
	{
		myValue.number = (Bool)*(SF<Bool> *)(myNodeField);
    }
    else 
	{
		myValue.number = (int)*(SF<int> *)(myNodeField);
    }
}


// VsSFFloat handles SFFloat and SFTime fields 

VsSFFloat::VsSFFloat(FIELD_TYPE t)
{
    assert(t == SFFLOAT || t == SFTIME);
    type = t;
}

VsSFFloat::~VsSFFloat()
{
}

VsValue
VsSFFloat::evaluate(VsFunctionCall *)
{
    return myValue;
}

VsValue
VsSFFloat::assign(const VsValue &v, VsFunctionCall *)
{
    if (permission == VsField::RW) 
	{
		switch (v.type) 
		{
		case VsValue::vNUMBER:
			myValue.number = (float)v.number;
			break;
		case VsValue::vSTRING:
			// ??? Use strotod and complain if not completely converted ???
			myValue.number = (float)atof(v.string);
			break;
		default:
			// ??? Warn if other types ???
			break;
		}

        if (myField)
		    myField->setHasBeenSet(TRUE);
		if (type == SFFLOAT) 
		{
			*(SF<float> *)(myNodeField) = (float)myValue.number;
		} 
		else	
		{
			*(SF<Time> *)(myNodeField) = myValue.number;
		}
    }

    return myValue;
}

void
VsSFFloat::setField(VSNodeField *f)
{
    if (!f) return;
    assert(f->GetType() == type);

    myField = f;
    setField(f->getNodeField());
}

void VsSFFloat::setField(NodeField *f) {
    myNodeField = f;

    myValue.type = VsValue::vNUMBER;
    if (type == SFFLOAT)
		myValue.number = (float)*(SF<float> *)(myNodeField);
    else 
		myValue.number = (Time)*(SF<Time> *)(myNodeField);
}


// VsSFFloatArray handles SFVec2f, SFVec3f, SFColor and SFRotation fields 

VsSFFloatArray::VsSFFloatArray(FIELD_TYPE t)
{
    assert(t == SFVEC2F || t == SFVEC3F || t == SFCOLOR || t == SFROTATION);
    type = t;

    myValue.floatArray = NULL;
}

VsSFFloatArray::~VsSFFloatArray()
{
    if (myValue.floatArray != NULL) 
	{
		delete myValue.floatArray;
		myValue.floatArray = NULL;
    }
}

VsValue
VsSFFloatArray::evaluate(VsFunctionCall *)
{
    return myValue;
}

VsValue
VsSFFloatArray::assign(const VsValue &v, VsFunctionCall *)
{
  int i;
  
  if (permission == VsField::RW) 
  {
    switch (v.type) {
    case VsValue::vFLOAT_ARRAY:
		assert(myValue.floatArray->length == v.floatArray->length);
		for (i=0; i<myValue.floatArray->length; i++) 
		{
			myValue.floatArray->array[i] = v.floatArray->array[i];
		}
		break;
      
    default:
		// ??? Warn if other types ???
		break;
    }
    
    if (myField)
		    myField->setHasBeenSet(TRUE);
	float *ary;
    switch (type) {
    case SFVEC2F:
		ary = (float *)myValue.floatArray->array;
		(*(SF<Vec2f> *)(myNodeField)) = Vec2f(ary[0], ary[1]);
		break;
    case SFVEC3F:
		ary = (float *)myValue.floatArray->array;
		(*(SF<Vec3f> *)(myNodeField)) = Vec3f(ary[0], ary[1], ary[2]);
		break;
    case SFCOLOR:
		ary = (float *)myValue.floatArray->array;
		(*(SF<Color3f> *)(myNodeField)) = Color3f(ary[0], ary[1], ary[2]);
		break;
    case SFROTATION:
 		ary = (float *)myValue.floatArray->array;
		(*(SF<Rotation> *)(myNodeField)) = Rotation(ary[0], ary[1], ary[2], ary[3]);
		break;
    default:     
		break;
    }
  }
  
  return myValue;
}

VsValue
VsSFFloatArray::assign(const VsValue &v, int32_t index, VsFunctionCall *)
{
    if (permission == VsField::RW)	
	{
		if (myValue.floatArray->length <= index)	
		{
			return myValue;
		}

		switch (v.type)		
		{
		case VsValue::vNUMBER:
			myValue.floatArray->array[index] = (float)v.number;
			break;
		default:
			// ??? Warn if other types ???
			break;
		}

		if (myField)
		    myField->setHasBeenSet(TRUE);
		float *ary;
		switch (type)	
		{
		case SFVEC2F:
			ary = (float *)myValue.floatArray->array;
			(*(SF<Vec2f> *)(myNodeField)) = Vec2f(ary[0], ary[1]);
		break;
		case SFVEC3F:
			ary = (float *)myValue.floatArray->array;
			(*(SF<Vec3f> *)(myNodeField)) = Vec3f(ary[0], ary[1], ary[2]);
			break;
		case SFCOLOR:
			ary = (float *)myValue.floatArray->array;
			(*(SF<Color3f> *)(myNodeField)) = Color3f(ary[0], ary[1], ary[2]);
			break;
		case SFROTATION:
 			ary = (float *)myValue.floatArray->array;
			(*(SF<Rotation> *)(myNodeField)) = Rotation(ary[0], ary[1], ary[2], ary[3]);
			break;
		default:     
			break;
		}
	}

    return myValue;
}

void
VsSFFloatArray::setField(VSNodeField *f)
{
    assert(f->GetType() == type);

    myField = f;
    setField(f->getNodeField());
}

void VsSFFloatArray::setField(NodeField *f) {
    myNodeField = f;
    myValue.type = VsValue::vFLOAT_ARRAY;

    myValue.floatArray = new FloatArray;

	float *ary;
    switch (type) 
	{
    case SFVEC2F:
		myValue.floatArray->length = 2;
		ary = new float[2];
		ary[0] = ((Vec2f)*(SF<Vec2f> *)myNodeField).x;
		ary[1] = ((Vec2f)*(SF<Vec2f> *)myNodeField).y;
		myValue.floatArray->array = ary;
		break;
	case SFVEC3F:
		myValue.floatArray->length = 3;
		ary = new float[3];
		ary[0] = ((Vec3f)*(SF<Vec3f> *)myNodeField).x;
		ary[1] = ((Vec3f)*(SF<Vec3f> *)myNodeField).y;
		ary[2] = ((Vec3f)*(SF<Vec3f> *)myNodeField).z;
		myValue.floatArray->array = ary;
		break;
    case SFCOLOR:
		myValue.floatArray->length = 3;
		ary = new float[3];
		ary[0] = ((Color3f)*(SF<Color3f> *)myNodeField).r;
		ary[1] = ((Color3f)*(SF<Color3f> *)myNodeField).g;
		ary[2] = ((Color3f)*(SF<Color3f> *)myNodeField).b;
		myValue.floatArray->array = ary;
		break;
    case SFROTATION:
		myValue.floatArray->length = 4;
		ary = new float[4];
		ary[0] = ((Rotation)*(SF<Rotation> *)myNodeField).x;
		ary[1] = ((Rotation)*(SF<Rotation> *)myNodeField).y;
		ary[2] = ((Rotation)*(SF<Rotation> *)myNodeField).z;
		ary[3] = ((Rotation)*(SF<Rotation> *)myNodeField).theta;
		myValue.floatArray->array = ary;
		break;
    default:
		break;
    }
}


// VsSFVoid handles SFNode fields 

VsSFVoid::VsSFVoid(FIELD_TYPE t)
{
    _ASSERT(t == SFNODE);
    type = t;
}

VsSFVoid::~VsSFVoid()
{
}

VsValue
VsSFVoid::evaluate(VsFunctionCall *)
{
    return myValue;
}

VsValue
VsSFVoid::assign(const VsValue &v, VsFunctionCall *)
{
    if (permission == VsField::RW) 
	{
		switch (v.type) 
		{
		case VsValue::vVOID:
			myValue.data = v.data;
			break;

		  default:
			// ??? Warn if other types ???
			break;
		}
		if (myField)
		    myField->setHasBeenSet(TRUE);
	//	if (type == SFNODE) {
	//	  ((VrmlSFNode *)myField)->setValue((VrmlNode *)myValue.data);
	//	} 
    }

    return myValue;
}

void
VsSFVoid::setField(VSNodeField *f)
{
    _ASSERT(f->GetType() == type);

    myField = f;
    setField(f->getNodeField());
}

void VsSFVoid::setField(NodeField *f) {
    myNodeField = f;
    if (type == SFNODE) { // NODEFIELD
	myValue.type = VsValue::vVOID;
	myValue.data = ((SFGenericNode *) myNodeField) -> GetNode ();
//  myValue.data = ((VrmlSFNode *)myField)->getValue();
    }
}

// VsSFString handles SFString fields 

VsSFString::VsSFString(FIELD_TYPE t)
{
    _ASSERT(t == SFSTRING);
    type = t;
}

VsSFString::~VsSFString()
{
}

VsValue
VsSFString::evaluate(VsFunctionCall *)
{
    return myValue;
}

VsValue
VsSFString::assign(const VsValue &v, VsFunctionCall *)
{
    if (permission == VsField::RW) 
	{
		switch (v.type) 
		{
		case VsValue::vSTRING:
			break;
		default:
			//Error("invalid assignement to a string");
			break;
		}

		if (myField)
		    myField->setHasBeenSet(TRUE);
		if (type == SFSTRING) 
		{
			*(SF<ZString> *)myNodeField = v.string;
		} 
    }

    return myValue;
}

void
VsSFString::setField(VSNodeField *f)
{
    _ASSERT(f->GetType() == type);

    myField = f;
    setField(f->getNodeField());
}

void VsSFString::setField(NodeField *f) {
    myNodeField = f;

    if (type == SFSTRING) 
	{
		myValue.type = VsValue::vSTRING;
		myValue.string = (char *)(LPCSTR)(ZString)*(SF<ZString> *)myNodeField;
    }
}


// VsMFInt handles MFInt32 fields

VsMFInt::VsMFInt(FIELD_TYPE t)
{
    _ASSERT(t == MFINT32);
    type = t;

    myValue.int32Array = new Int32Array;
	myValue.int32Array->length = 0;
	myValue.int32Array->array = NULL;
}

VsMFInt::~VsMFInt()
{
    if (myValue.int32Array != NULL) 
	{
		if(myValue.int32Array->array != NULL)
			delete [] myValue.int32Array->array;
		delete myValue.int32Array;
		myValue.int32Array = NULL;
    }
}

VsValue
VsMFInt::evaluate(VsFunctionCall *)
{
    return myValue;
}

VsValue
VsMFInt::assign(const VsValue &v, VsFunctionCall *)
{
    if (permission == VsField::RW) 
	{
		int i, len;
		switch (v.type) 
		{
		case VsValue::vINT32_ARRAY:
			len = v.int32Array->length;
			myValue.int32Array->length = len;
			if(myValue.int32Array->array)
				delete [] myValue.int32Array->array;
			
			myValue.int32Array->array = new int[len];
			(*(MF<int> *)myNodeField).Remove();
			for(i = 0; i < len; i++)
			{
				(*(MF<int> *)myNodeField) += v.int32Array->array[i];
				myValue.int32Array->array[i] = v.int32Array->array[i];
			}

			break;
		default:
			// ??? Warn if other types ???
			break;
		}

		if (myField)
		    myField->setHasBeenSet(TRUE);
    }

    return myValue;
}

VsValue
VsMFInt::assign(const VsValue &v, int32_t index, VsFunctionCall *)
{
    if (permission == VsField::RW) 
	{

		int len, len_old;
		int *ary;
		switch (v.type) 
		{
		case VsValue::vNUMBER:
			len = (*(MF<int> *)myNodeField).GetLength();
			if (index+1 > len) 
			{
				len_old = len;
				myValue.int32Array->length = index+1;
				while(len++ < index)
					(*(MF<int> *)myNodeField) += 0;

				(*(MF<int> *)myNodeField) += v.number;
				ary = myValue.int32Array->array;
				myValue.int32Array->array = new int[index + 1];
				for(int i = 0; i < len_old; i++)
					myValue.int32Array->array[i] = ary[i];
			}
			else
				(*(MF<int> *)myNodeField).SetAt(index, v.number);

			myValue.int32Array->array[index] = v.number;
			break;
		default:
			// ??? Warn if other types ???
			break;
		}

		if (myField)
		    myField->setHasBeenSet(TRUE);
    }

    return myValue;
}

void
VsMFInt::setField(VSNodeField *f)
{
    _ASSERT(f->GetType() == type);


    myField = f;
    setField(f->getNodeField());
}

void VsMFInt::setField(NodeField *f) {
    myNodeField = f;
	MF<int> *mfIntField = (MF<int> *)myNodeField;
    myValue.type = VsValue::vINT32_ARRAY;

	int i, len;
    switch (type) 
	{
    case MFINT32:
		len = mfIntField->GetLength();
		myValue.int32Array->length = len;
		if(myValue.int32Array->array)
			delete [] myValue.int32Array->array;

		if(len != 0)
			myValue.int32Array->array = new int[len];
		for(i = 0; i < len; i++)
			myValue.int32Array->array[i] = (int)((*mfIntField)[i]);
		break;
    default:
		break;
    }
}


// VsMFFloat handles MFFloat fields and MFTime

VsMFFloat::VsMFFloat(FIELD_TYPE t)
{
    _ASSERT(t == MFFLOAT || t == MFTIME);
    type = t;

    myValue.floatArray = new FloatArray;
	myValue.floatArray->length = 0;
	myValue.floatArray->array = NULL;
}

VsMFFloat::~VsMFFloat()
{
    if (myValue.floatArray != NULL) 
	{
		if(myValue.floatArray->array != NULL)
			delete [] myValue.floatArray->array;

		delete myValue.floatArray;
		myValue.floatArray = NULL;
    }
}

VsValue
VsMFFloat::evaluate(VsFunctionCall *)
{
    return myValue;
}

VsValue
VsMFFloat::assign(const VsValue &v, VsFunctionCall *)
{
    if (permission == VsField::RW) 
	{
		int i, len;
		switch (v.type) 
		{
		case VsValue::vFLOAT_ARRAY:
			len = v.floatArray->length;
			myValue.floatArray->length = len;
			if(myValue.floatArray->array)
				delete [] myValue.floatArray->array;

			myValue.floatArray->array = new float[len];
			(*(MF<float> *)myNodeField).Remove();
			for(i = 0; i < len; i++)
			{
				(*(MF<float> *)myNodeField) += v.floatArray->array[i];
				myValue.floatArray->array[i] = v.floatArray->array[i];
			}
			break;
		default:
			// ??? Warn if other types ???
			break;
		}

		if (myField)
		    myField->setHasBeenSet(TRUE);
    }

    return myValue;
}

VsValue
VsMFFloat::assign(const VsValue &v, int32_t index, VsFunctionCall *)
{
    if (permission == VsField::RW) 
	{
		int len, len_old;
		float *ary;
		switch (v.type) 
		{
		case VsValue::vNUMBER:
			len = (*(MF<float> *)myNodeField).GetLength();
			if(index >= len)
			{
				len_old = len;
				myValue.floatArray->length = index+1;
				while(len++ < index)
					(*(MF<float> *)myNodeField) += 0;

				(*(MF<float> *)myNodeField) += v.number; // how about if index>length??
				ary = myValue.floatArray->array;
				myValue.floatArray->array = new float[index + 1];
				for(int i = 0; i < len_old; i++)
					myValue.floatArray->array[i] = ary[i]; 
			}
			else
				(*(MF<int> *)myNodeField).SetAt(index, v.number);

			myValue.floatArray->array[index] = v.number; 
	    break;
	  default:
	    // ??? Warn if other types ???
	    break;
	}

	if (myField)
		    myField->setHasBeenSet(TRUE);
    }

    return myValue;
}

void VsMFFloat::setField(VSNodeField *f)
{
    _ASSERT(f->GetType() == type);


    myField = f;
    setField(f->getNodeField());
}

void VsMFFloat::setField(NodeField *f) {

    myNodeField = f;
    myValue.type = VsValue::vFLOAT_ARRAY;

	int i, len;
    switch (type) 
	{
    case MFFLOAT: {
        MF<float> *mfFloatField = (MF<float> *)myNodeField;
	    len = mfFloatField->GetLength();
	    myValue.floatArray->length = len;
	    if(myValue.floatArray->array)
		    delete [] myValue.floatArray->array;

	    if(len != 0)
		    myValue.floatArray->array = new float[len];
	    for(i = 0; i < len; i++)
		    myValue.floatArray->array[i] = (float)((*mfFloatField)[i]);
        break; }
    case MFTIME: {
        MF<Time>  *mfFloatField = (MF<Time>  *)myNodeField;
	    len = mfFloatField->GetLength();
	    myValue.floatArray->length = len;
	    if(myValue.floatArray->array)
		    delete [] myValue.floatArray->array;

	    if(len != 0)
		    myValue.floatArray->array = new float[len];
	    for(i = 0; i < len; i++)
		    myValue.floatArray->array[i] = (float)((*mfFloatField)[i]);
        break; }
    default:
		return;
    }
}

// VsMFVoid handles MFNode fields

VsMFVoid::VsMFVoid(FIELD_TYPE t)
{
//    _ASSERT(t == MFNODE);

    type = t;

    myValue.voidArray = NULL;
}

VsMFVoid::~VsMFVoid()
{
    if (myValue.voidArray != NULL) 
	{
	delete myValue.voidArray;
	myValue.voidArray = NULL;
    }
}

VsValue
VsMFVoid::evaluate(VsFunctionCall *)
{
    return myValue;
}

VsValue
VsMFVoid::assign(const VsValue &v, VsFunctionCall *)
{
    if (permission == VsField::RW) 
	{
	switch (v.type) 
	{
	case VsValue::vVOID_ARRAY:
		  /*
	    myValue.voidArray->length = v.voidArray->length;
	    ((VrmlMFNode *)myField)->setValues(0, v.voidArray->length, (const VrmlNode **)v.voidArray->array);
	    myValue.voidArray->array = (void **)((VrmlMFNode *)myField)->getValues(0);	 
   	    */
	    break;

	  default:
	    // ??? Warn if other types ???
	    break;
	}

	if (myField)
		    myField->setHasBeenSet(TRUE);
    }

    return myValue;
}

VsValue
VsMFVoid::assign(const VsValue &v, int32_t, VsFunctionCall *)
{
    if (permission == VsField::RW) 
	{
		switch (v.type) 
		{
		case VsValue::vVOID:
			  /*
			  if (index+1 > myValue.voidArray->length) {
			  myValue.voidArray->length = index+1;
			  }
			  ((VrmlMFNode *)myField)->set1Value(index, (VrmlNode *)v.data);
			  myValue.voidArray->array = (void **)((VrmlMFNode *)myField)->getValues(0);	
			  */
			break;

		  default:
			// ??? Warn if other types ???
			break;
		}

		if (myField)
		    myField->setHasBeenSet(TRUE);
    }

    return myValue;
}

void VsMFVoid::setField(NodeField *) {}

void
VsMFVoid::setField(VSNodeField *)
{
	/*
    _ASSERT(f->getType() == type);

    VrmlMFNode* mfNodeField = (VrmlMFNode *)f;

    myField = f;
    myValue.type = VsValue::vVOID_ARRAY;

    myValue.voidArray = new VoidArray;

    switch (type) {
    case MFNODE:
	myValue.voidArray->length = mfNodeField->getNum();
	myValue.voidArray->array = (void **)mfNodeField->getValues(0);
	break;
      default:
	break;
    }
	*/
}


// VsMFString handles MFString fields

VsMFString::VsMFString(FIELD_TYPE t)
{
    _ASSERT(t == MFSTRING);

    type = t;

    myValue.stringArray = new StringArray;
	myValue.stringArray->length = 0;
	myValue.stringArray->array = NULL;
}

VsMFString::~VsMFString()
{
    if (myValue.stringArray != NULL) 
	{
		if(myValue.stringArray->array)
			delete [] myValue.stringArray->array;

		delete myValue.stringArray;
		myValue.stringArray = NULL;
    }
}

VsValue
VsMFString::evaluate(VsFunctionCall *)
{
    return myValue;
}

VsValue
VsMFString::assign(const VsValue &v, VsFunctionCall *)
{
    if (permission == VsField::RW) 
	{

		int i, len;
		switch (v.type) 
		{
		case VsValue::vSTRING_ARRAY:
			len = v.stringArray -> length;
			myValue.stringArray->length = len;
			(*(MF<ZString> *)myNodeField).Remove();
			if(myValue.stringArray->array)
			{
				delete [] myValue.stringArray->array;
				myValue.stringArray->array = new ZString[len];
			}
			for(i = 0; i < len; i++)
			{
				(*(MF<ZString> *)myNodeField) += v.stringArray->array[i];
				myValue.stringArray->array[i] = (LPCSTR)v.stringArray->array[i];
			}

			break;
		default:
			// ??? Warn if other types ???
			break;
		}

		if (myField)
		    myField->setHasBeenSet(TRUE);
    }

    return myValue;
}

VsValue
VsMFString::assign(const VsValue &v, int32_t index, VsFunctionCall *)
{

    if (permission == VsField::RW) 
	{
		int len, len_old;
		switch (v.type) 
		{
		case VsValue::vSTRING:
			len = (*(MF<ZString> *)myNodeField).GetLength();
			if (index >= len)
			{
				len_old = len;
				myValue.stringArray->length = index + 1;
				while(len++ < index)
					(*(MF<ZString> *)myNodeField) += "";

				*(MF<ZString> *)myNodeField += v.string;

				ZString *ary;
				if(index == 0)
					myValue.stringArray->array = new ZString[1];
				else
				{
					ary = myValue.stringArray->array;
					myValue.stringArray->array = new ZString[index + 1];
					for(int i = 0; i < len_old; i++)
						myValue.stringArray->array[i] = ary[i];

					delete [] ary;
				}
			}
			else 
				(*(MF<ZString> *)myNodeField).SetAt(index, ZString(v.string));

			myValue.stringArray->array[index] = v.string;
			break;  
		default:
			// ??? Warn if other types ???
			break;
		}

		if (myField)
		    myField->setHasBeenSet(TRUE);
	}

    return myValue;
}

void
VsMFString::setField(VSNodeField *f)
{
	
    _ASSERT(f->GetType() == type);


    myField = f;
    setField(f->getNodeField());
}

void VsMFString::setField(NodeField *f) {
    
    myNodeField = f;
    MF<ZString> *mfStringField = (MF<ZString> *)myNodeField;
    myValue.type = VsValue::vSTRING_ARRAY;

	int i, len;
    switch (type) 
	{
    case MFSTRING:
		len = mfStringField->GetLength();
		myValue.stringArray->length = len;
		if(!myValue.stringArray->array)
			delete [] myValue.stringArray->array;

		if(len != 0)
			myValue.stringArray->array = new ZString[len];
		for(i = 0; i < len; i++)
			myValue.stringArray->array[i] = (*mfStringField)[i];
		break;
    default:
		break;
    }
	
}

// VsMFFloatArray handles MFVec2f, MFVec3f, MFColor and MFRotation fields 

VsMFFloatArray::VsMFFloatArray(FIELD_TYPE t)
{
    assert(t == MFVEC2F || t == MFVEC3F || t == MFCOLOR || t == MFROTATION);
    type = t;

    myValue.floatArray = new FloatArray;
	myValue.floatArray->length = 0;
	myValue.floatArray->array = NULL;

    switch(t) {
    case MFVEC2F:
 	    vectorSize = 2;
	    break;
    case MFVEC3F:
 	    vectorSize = 3;
	    break;
    case MFCOLOR:
 	    vectorSize = 3;
	    break;
    case MFROTATION:
 	    vectorSize = 4;
	    break;
    }

}

VsMFFloatArray::~VsMFFloatArray()
{
    if (myValue.floatArray != NULL) 
	{
		if(myValue.floatArray->array != NULL)
			delete [] myValue.floatArray->array;

		delete myValue.floatArray;
		myValue.floatArray = NULL;
    }
}

VsValue
VsMFFloatArray::evaluate(VsFunctionCall *)
{
    return myValue;
}

VsValue
VsMFFloatArray::assign(const VsValue &v, VsFunctionCall *)
{
  
  if (permission == VsField::RW) 
  {
    int i, len;
    switch (v.type) {
    case VsValue::vFLOAT_ARRAY:
        len = v.floatArray->length;
        myValue.floatArray->length = len;
        if(myValue.floatArray->array)
	        delete [] myValue.floatArray->array;

	    myValue.floatArray->array = new float[len];
		break;
      
    default:
        return myValue;
		// ??? Warn if other types ???
    }
    
    if (myField)
		    myField->setHasBeenSet(TRUE);
    switch (type) {
    case MFVEC2F:    
	    (*(MF<Vec2f> *)myNodeField).Remove();
	    for(i = 0; i < len / 2; i++)
	    {
		    (*(MF<Vec2f> *)myNodeField) += Vec2f(v.floatArray->array[2*i],v.floatArray->array[2*i+1]);
		    myValue.floatArray->array[2*i] = v.floatArray->array[2*i];
		    myValue.floatArray->array[2*i+1] = v.floatArray->array[2*i+1];
	    }
		break;
    case MFVEC3F:
	    (*(MF<Vec3f> *)myNodeField).Remove();
	    for(i = 0; i < len / 3; i++)
	    {
		    (*(MF<Vec3f> *)myNodeField) += Vec3f(v.floatArray->array[3*i],
                            v.floatArray->array[3*i+1],
                            v.floatArray->array[3*i+2]);
		    myValue.floatArray->array[3*i] = v.floatArray->array[3*i];
		    myValue.floatArray->array[3*i+1] = v.floatArray->array[3*i+1];
		    myValue.floatArray->array[3*i+2] = v.floatArray->array[3*i+2];
	    }
		break;
    case MFCOLOR:
	    (*(MF<Color3f> *)myNodeField).Remove();
	    for(i = 0; i < len / 3; i++)
	    {
		    (*(MF<Color3f> *)myNodeField) += Color3f(v.floatArray->array[3*i],
                            v.floatArray->array[3*i+1],
                            v.floatArray->array[3*i+2]);
		    myValue.floatArray->array[3*i] = v.floatArray->array[3*i];
		    myValue.floatArray->array[3*i+1] = v.floatArray->array[3*i+1];
		    myValue.floatArray->array[3*i+2] = v.floatArray->array[3*i+2];
	    }
		break;
    case MFROTATION:
	    (*(MF<Rotation> *)myNodeField).Remove();
	    for(i = 0; i < len / 4; i++)
	    {
		    (*(MF<Rotation> *)myNodeField) += Rotation(v.floatArray->array[4*i],
                            v.floatArray->array[4*i+1],
                            v.floatArray->array[4*i+2],
                            v.floatArray->array[4*i+3]);
		    myValue.floatArray->array[4*i] = v.floatArray->array[4*i];
		    myValue.floatArray->array[4*i+1] = v.floatArray->array[4*i+1];
		    myValue.floatArray->array[4*i+2] = v.floatArray->array[4*i+2];
		    myValue.floatArray->array[4*i+3] = v.floatArray->array[4*i+3];
	    }
		break;
    default:     
		break;
    }
  }
  
  return myValue;
}

VsValue
VsMFFloatArray::assign(const VsValue &v, int32_t index, VsFunctionCall *)
{
    if (permission == VsField::RW)	
	{
		int len, len_old, j;
		switch (v.type) 
		{
        case VsValue::vFLOAT_ARRAY:
            switch (type) {
		    case MFVEC2F:
 			    len = (*(MF<Vec2f> *)myNodeField).GetLength();
		        break;
		    case MFVEC3F:
 			    len = (*(MF<Vec3f> *)myNodeField).GetLength();
			    break;
		    case MFCOLOR:
 			    len = (*(MF<Color3f> *)myNodeField).GetLength();
			    break;
		    case MFROTATION:
 			    len = (*(MF<Rotation> *)myNodeField).GetLength();
			    break;
            default :
                return myValue;
            }
            if (index >= len) {
			    len_old = len;
				myValue.floatArray->length = index + vectorSize;
                while(len++ < index) {
                    switch (type) {
		            case MFVEC2F:
 			            (*(MF<Vec2f> *)myNodeField) += Vec2f(0,0);
		                break;
		            case MFVEC3F:
 			            (*(MF<Vec3f> *)myNodeField) += Vec3f(0,0,0);
			            break;
		            case MFCOLOR:
 			            (*(MF<Color3f> *)myNodeField) += Color3f(0,0,0);
			            break;
		            case MFROTATION:
 			            (*(MF<Rotation> *)myNodeField) += Rotation(0,0,0,0);
			            break;
                    }
                }

                switch (type) {
		        case MFVEC2F:
 			        (*(MF<Vec2f> *)myNodeField) += Vec2f(v.floatArray->array[0],
                                            v.floatArray->array[1]);
		            break;
		        case MFVEC3F:
 			        (*(MF<Vec3f> *)myNodeField) += Vec3f(v.floatArray->array[0],
                                            v.floatArray->array[1],
                                            v.floatArray->array[2]);
			        break;
		        case MFCOLOR:
 			        (*(MF<Color3f> *)myNodeField) += Color3f(v.floatArray->array[0],
                                            v.floatArray->array[1],
                                            v.floatArray->array[2]);
			        break;
		        case MFROTATION:
 			        (*(MF<Rotation> *)myNodeField) += Rotation(v.floatArray->array[0],
                                            v.floatArray->array[1],
                                            v.floatArray->array[2],
                                            v.floatArray->array[3]);
			        break;
                }
				float * ary = myValue.floatArray->array;
				myValue.floatArray->array = new float[vectorSize*(index + 1)];
				for(int i = 0; i < len_old * vectorSize; i++)
					myValue.floatArray->array[i] = ary[i];
                for (i = len_old * vectorSize; i < (index+1) * vectorSize; i++)
					myValue.floatArray->array[i] = 0;

			}
			else
            switch (type) {
		    case MFVEC2F:
 			    (*(MF<Vec2f> *)myNodeField).SetAt(index, Vec2f(v.floatArray->array[0],
                                            v.floatArray->array[1]));
		        break;
		    case MFVEC3F:
 			    (*(MF<Vec3f> *)myNodeField).SetAt(index, Vec3f(v.floatArray->array[0],
                                            v.floatArray->array[1],
                                            v.floatArray->array[2]));
			    break;
		    case MFCOLOR:
 			    (*(MF<Color3f> *)myNodeField).SetAt(index, Color3f(v.floatArray->array[0],
                                            v.floatArray->array[1],
                                            v.floatArray->array[2]));
			    break;
		    case MFROTATION:
 			    (*(MF<Rotation> *)myNodeField).SetAt(index, Rotation(v.floatArray->array[0],
                                            v.floatArray->array[1],
                                            v.floatArray->array[2],
                                            v.floatArray->array[3]));
			    break;
            }

            for ( j = 0; j < vectorSize; j++) {
                myValue.floatArray->array[index+j] = v.floatArray->array[j]; 
            }
	    break;
    default :
        // error ?
        ERROR("invalid type");
        
    }
	if (myField)
		    myField->setHasBeenSet(TRUE);
    }

    return myValue;
}

void
VsMFFloatArray::setField(VSNodeField *f)
{
    assert(f->GetType() == type);

    myField = f;
    setField(f->getNodeField());
}

void VsMFFloatArray::setField(NodeField *f) {
    myNodeField = f;
    myValue.type = VsValue::vFLOAT_ARRAY;

   	int i, len;

    switch (type) 
	{
    case MFVEC2F: 
        {
        MF<Vec2f> *mfFloatField = (MF<Vec2f> *)myNodeField;
	    len = mfFloatField->GetLength();
	    myValue.floatArray->length = len * 2;
	    if(myValue.floatArray->array)
		    delete [] myValue.floatArray->array;

	    if(len != 0)
		    myValue.floatArray->array = new float[len*2];
        for(i = 0; i < len; i++) {
		    myValue.floatArray->array[2*i] = ((Vec2f)((*mfFloatField)[i])).x;
		    myValue.floatArray->array[2*i+1] = ((Vec2f)((*mfFloatField)[i])).y;
        }
        break; }

	case MFVEC3F:
        {
        MF<Vec3f> *mfFloatField = (MF<Vec3f> *)myNodeField;
	    len = mfFloatField->GetLength();
	    myValue.floatArray->length = len * 3;
	    if(myValue.floatArray->array)
		    delete [] myValue.floatArray->array;

	    if(len != 0)
		    myValue.floatArray->array = new float[len*3];
        for(i = 0; i < len; i++) {
		    myValue.floatArray->array[3*i] = ((Vec3f)((*mfFloatField)[i])).x;
		    myValue.floatArray->array[3*i+1] = ((Vec3f)((*mfFloatField)[i])).y;
		    myValue.floatArray->array[3*i+2] = ((Vec3f)((*mfFloatField)[i])).z;
        }
        break; }

    case MFCOLOR:
        {
        MF<Color3f> *mfFloatField = (MF<Color3f> *)myNodeField;
	    len = mfFloatField->GetLength();
	    myValue.floatArray->length = len * 3;
	    if(myValue.floatArray->array)
		    delete [] myValue.floatArray->array;

	    if(len != 0)
		    myValue.floatArray->array = new float[len*3];
        for(i = 0; i < len; i++) {
		    myValue.floatArray->array[3*i] = ((Color3f)((*mfFloatField)[i])).r;
		    myValue.floatArray->array[3*i+1] = ((Color3f)((*mfFloatField)[i])).g;
		    myValue.floatArray->array[3*i+2] = ((Color3f)((*mfFloatField)[i])).b;
        }
        break; }

    case MFROTATION:
        {
        MF<Rotation> *mfFloatField = (MF<Rotation> *)myNodeField;
	    len = mfFloatField->GetLength();
	    myValue.floatArray->length = len * 4;
	    if(myValue.floatArray->array)
		    delete [] myValue.floatArray->array;

	    if(len != 0)
		    myValue.floatArray->array = new float[len*4];
        for(i = 0; i < len; i++) {
		    myValue.floatArray->array[4*i] = ((Rotation)((*mfFloatField)[i])).x;
		    myValue.floatArray->array[4*i+1] = ((Rotation)((*mfFloatField)[i])).y;
		    myValue.floatArray->array[4*i+2] = ((Rotation)((*mfFloatField)[i])).z;
		    myValue.floatArray->array[4*i+3] = ((Rotation)((*mfFloatField)[i])).theta;
        }
        break; }

    default:
		break;
    }
}

