/*
 * VsFunction.c++
 *
 *     Implementation of classes:
 *        VsFunctionDef
 *        VsFunctionCall
 *	  VsMethodCall
 *	  VsMemberAccess
 *
 * Copyright (C) 1996, Silicon Graphics,  Inc.
 */

/*---------------------------------------------------------------
This module is copied from the VRML Script interpreter owned by
Silicon Graphics, Inc. It is modified by Wenlong Dong, Wei Wu and
Yuval Fisher(Rockwell Science Center) in the course of development
of the MPEG-4 Systems standard(ISO/IEC 14496-1) and added to the 
subsystem: Script Node.
---------------------------------------------------------------*/

#include "defs.h"
#include <assert.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <iostream.h>
#include "vsfunction.h"
#include "vsstatement.h"
#include "jsmath.h"
#include "DecScript.h"
#ifndef BIFSENC	// NODEFIELD
#include "MediaObj.h"
#include "NodeFld.h"
#include "vsfields.h"
#endif

#ifdef __sgi
double JsMath::E       = M_E;
double JsMath::LN10    = M_LN10;
double JsMath::LN2     = M_LN2;
double JsMath::PI      = M_PI;
double JsMath::SQRT1_2 = M_SQRT1_2;
double JsMath::SQRT2   = M_SQRT2;
#else
double JsMath::E       = 2.7182818284590452354;
double JsMath::LN10    = 2.30258509299404568402;
double JsMath::LN2     = 0.69314718055994530942;
double JsMath::PI      = 3.14159265358979323846;
double JsMath::SQRT1_2 = 0.70710678118654752440;
double JsMath::SQRT2   = 1.41421356237309504880;
#endif

#ifdef __sgi
#include <alloca.h> /* For alloca() on SGI */
#endif              /* (alloca() is defined in <malloc.h> on PC) */

// While parsing, function currently being defined:
VsFunctionDef *VsFunctionDef::curFunction = NULL;

VsFunctionDef::VsFunctionDef(const char *_name, DecScript *_parent)
{
    name =_strdup(_name);
    numLocals = numArgs = 0;
    statements = NULL;
    parent = _parent;
    assert(curFunction == NULL);
    curFunction = this;
}

VsFunctionDef::~VsFunctionDef()
{
    free(name);
    for (int i = 0; i < argOrLocalNames.GetLength(); i++) {
	char *n = (char *)argOrLocalNames[i];
	free(n);
    }
    if (statements != NULL)
	delete statements;

#ifdef __sgi
    // Make sure curFunction is set to NULL.  If there was a syntax
    // error inside a function, curFunction may be left in an unknown
    // state.  Make sure it is cleared here.
    curFunction = NULL;
#endif
}

void
VsFunctionDef::endDefinition(VsStatementList *stmnts)
{
    assert(curFunction == this);
    statements = stmnts;
    curFunction = NULL;
}

void
VsFunctionDef::addArg(const char *argName)
{
  assert(curFunction->numLocals == 0);

  ++curFunction->numArgs;
  curFunction->argOrLocalNames += _strdup(argName);
}

VsVar *
VsFunctionDef::findIdentifier(const char *name, bool n_flag) // flag to tell we are in memberAccess
{
    assert(curFunction != NULL);

    if (n_flag && !stricmp(name,"math")) {
        return NULL;
    }

    // Ok, look through local vars and arguments, last-defined first:
    for (int i = curFunction->getNumLocals()-1; i >= 0; i--) 
	{
		const char *n = (const char *)curFunction->argOrLocalNames[i];
		if (strcmp(n, name) == 0) 
		{
		    return new VsArgOrLocal(i);
		}
    }

    VsVar *result = curFunction->parent->getGlobalVar(name);

    if (result == NULL) 
	{
        if (n_flag) {
            fprintf(stdout,"cannot access a member not previously defined : %s\n", name);
            exit(-1);
        }
		// Not found, create a new local variable:
		++curFunction->numLocals;
		curFunction->argOrLocalNames += _strdup(name);
		result = new VsArgOrLocal(curFunction->getNumLocals()-1);
    }

    return result;
}

VsFunctionCall::VsFunctionCall(VsFunctionDef *_func, VsExprList &exprs)
{
    function = _func;

	args.Clear();
    for (int i = 0; i < exprs.GetLength(); i++) 
	{
		args += exprs[i];
		args[i]->ref();
    }

    locals = NULL;
}

VsFunctionCall::~VsFunctionCall()
{
    for (int i = 0; i < args.GetLength(); i++) 
	{
		args[i]->unref();
    }
}

VsValue
VsFunctionCall::evaluate(VsFunctionCall *parent)
{
    returnValue.type = VsValue::vVOID;

    VsValue *callStack;  // Previous args, locals, needed if recursing

    // Gotta allocate new locals[].  Use alloca() for speed.
    callStack = locals;
    int numL = function->getNumLocals();

    if (numL == 0)
		locals = NULL;
    else 
	{
        locals = (VsValue *)malloc(sizeof(VsValue)*function->getNumLocals());

		int i;
		   // Evaluate arguments:
		for (i = 0; i < function->getNumArgs(); i++) 
		{
			locals[i] = args[i]->evaluate(parent);
		}

			// And give other locals/args a VOID value:
		for (i = function->getNumArgs(); i < function->getNumLocals(); i++) 
		{
			locals[i].type = VsValue::vVOID;  // ??? UNINIT ???
		}
	}

	// Execute statements until we get a return:
	VsStatementList *stmnts = function->statements;
	for (int i = 0; stmnts != NULL && i < stmnts->GetLength(); i++) 
	{
		VsStatement *stmt = (*stmnts)[i];
		VsStatement::Status s = stmt->execute(this);
		if (s == VsStatement::RETURN) 
			break;
    }

    // ??? Delete any string-valued args or locals ???

    // Unwind call stack
    locals = callStack;

    return returnValue;
}

void
VsFunctionCall::setReturnValue(const VsValue &val)
{
    returnValue = val;
}

VsValue
VsFunctionCall::getArg(int whichArg)
{
    if (whichArg >= function->getNumLocals()) 
	{
		// Reference to un-passed param, return NULL value
		return VsValue::nothing;
    }

    return locals[whichArg];
}

void
VsFunctionCall::setArgOrLocal(int which, const VsValue &newVal)
{
    if (which < function->getNumLocals()) 
		locals[which] = newVal;
}

int		
VsFunctionCall::isAnUnimplementedFunction(const char* fName)
{
    if (!strcmp(fName, "SFBool")) 
		return 1;
    else if (!strcmp(fName, "SFColor")) 
		return 1;
    else if (!strcmp(fName, "SFFloat")) 
		return 1;
    else if (!strcmp(fName, "SFImage")) 
		return 1;
    else if (!strcmp(fName, "SFInt32")) 
		return 1;
    else if (!strcmp(fName, "SFNode")) 
		return 1;
    else if (!strcmp(fName, "SFRotation")) 
		return 1;
    else if (!strcmp(fName, "SFString")) 
		return 1;
    else if (!strcmp(fName, "SFTime")) 
		return 1;
    else if (!strcmp(fName, "SFVec2f")) 
		return 1;
    else if (!strcmp(fName, "SFVec3f")) 
		return 1;
    else if (!strcmp(fName, "MFColor")) 
		return 1;
    else if (!strcmp(fName, "MFFloat")) 
		return 1;
    else if (!strcmp(fName, "MFInt32")) 
		return 1;
    else if (!strcmp(fName, "MFNode")) 
		return 1;
    else if (!strcmp(fName, "MFRotation")) 
		return 1;
    else if (!strcmp(fName, "MFString")) 
		return 1;
    else if (!strcmp(fName, "MFVec2f")) 
		return 1;
    else if (!strcmp(fName, "MFVec3f")) 
		return 1;
    else if (!strcmp(fName, "parseInt")) 
		return 1;
    else if (!strcmp(fName, "parseFloat")) 
		return 1;

    return 0;
}

VsArgOrLocal::VsArgOrLocal(int _whichArg)
{
    whichArg = _whichArg;
}

VsArgOrLocal::~VsArgOrLocal()
{
}

VsValue
VsArgOrLocal::evaluate(VsFunctionCall *f)
{
    return f->getArg(whichArg);
}

VsValue
VsArgOrLocal::assign(const VsValue &v, VsFunctionCall *f)
{
    f->setArgOrLocal(whichArg, v);
    return v;
}

VsMethodCall::VsMethodCall(VsExpr* _obj,
			     const char* funcName,
			     VsExprList &exprs)
{
    obj = _obj;

    if (obj == NULL) 
	{	    // MATH object is a special case
		if (!strcmp(funcName, "abs")) 
			func = fABS;
		else if (!stricmp(funcName, "acos")) 
			func = fACOS;
		else if (!stricmp(funcName, "asin")) 
			func = fASIN;
		else if (!stricmp(funcName, "atan")) 
			func = fATAN;
		else if (!stricmp(funcName, "ceil")) 
			func = fCEIL;
		else if (!stricmp(funcName, "cos")) 
			func = fCOS;
		else if (!stricmp(funcName, "exp")) 
			func = fEXP;
		else if (!stricmp(funcName, "floor")) 
			func = fFLOOR;
		else if (!stricmp(funcName, "log")) 
			func = fLOG;
		else if (!stricmp(funcName, "max")) 
			func = fMAX;
		else if (!stricmp(funcName, "min")) 
			func = fMIN;
		else if (!stricmp(funcName, "pow")) 
			func = fPOW;
		else if (!stricmp(funcName, "random")) 
			func = fRANDOM;
		else if (!stricmp(funcName, "round")) 
			func = fROUND;
		else if (!stricmp(funcName, "sin")) 
			func = fSIN;
		else if (!stricmp(funcName, "sqrt")) 
			func = fSQRT;
		else if (!stricmp(funcName, "tan")) 
			func = fTAN;
		else 
			func = fUNKNOWN;
    }

	args.Clear();
    for (int i = 0; i < exprs.GetLength(); i++) 
	{
		args += exprs[i];
		args[i]->ref();
    }
}

VsMethodCall::~VsMethodCall()
{
    for (int i = 0; i < args.GetLength(); i++) 
		args[i]->unref();
}

VsValue
VsMethodCall::evaluate(VsFunctionCall *parent)
{
    returnValue.type = VsValue::vVOID;

    if (obj == NULL) 
	{	// MATH
		switch (func) 
		{
		case fABS:
			returnValue.number = JsMath::abs((args[0]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fACOS:
			returnValue.number = JsMath::acos((args[0]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fASIN:
			returnValue.number = JsMath::asin((args[0]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fATAN:
			returnValue.number = JsMath::atan((args[0]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fCEIL:
			returnValue.number = JsMath::ceil((args[0]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fCOS:
			returnValue.number = JsMath::cos((args[0]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fEXP:
			returnValue.number = JsMath::exp((args[0]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fFLOOR:
			returnValue.number = JsMath::floor((args[0]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fLOG:
			returnValue.number = JsMath::log((args[0]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fMAX:
			returnValue.number = JsMath::max((args[0]->evaluate(parent)).number,
							 (args[1]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fMIN:
			returnValue.number = JsMath::min((args[0]->evaluate(parent)).number,
							 (args[1]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fPOW:
			returnValue.number = JsMath::pow((args[0]->evaluate(parent)).number,
							 (args[1]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fRANDOM:
			returnValue.number = JsMath::random();
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fROUND:
			returnValue.number = JsMath::round((args[0]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fSIN:
			returnValue.number = JsMath::sin((args[0]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fSQRT:
			returnValue.number = JsMath::sqrt((args[0]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fTAN:
			returnValue.number = JsMath::tan((args[0]->evaluate(parent)).number);
			returnValue.type   = VsValue::vNUMBER;
			break;
		case fUNKNOWN:
			returnValue.type = VsValue::vVOID;
			break;
		}
    }

    return returnValue;
}


VsMemberAccess::VsMemberAccess(VsExpr* _obj,
			     const char* memName)
{
    obj = _obj;
    m_identifier = memName;

    if (obj == NULL) 
	{	    // MATH object
		if (!stricmp(memName, "E")) 
			member = mE;
		else if (!stricmp(memName, "LN10")) 
			member = mLN10;
		else if (!stricmp(memName, "LN2")) 
			member = mLN2;
		else if (!stricmp(memName, "PI")) 
		    member = mPI;
		else if (!stricmp(memName, "SQRT1_2")) 
		    member = mSQRT1_2;
		else if (!stricmp(memName, "SQRT2")) 
		    member = mSQRT2;             
    } else { // NODEFIELD
	}
}

VsMemberAccess::~VsMemberAccess()
{
 
}

VsValue
VsMemberAccess::evaluate(VsFunctionCall * f)
{
    returnValue.type = VsValue::vVOID;

    if (obj == NULL) 
	{	// MATH object
		switch (member) {
		case mE:
			returnValue.number = JsMath::E;
			returnValue.type   = VsValue::vNUMBER;
			break;
		case mLN10:
			returnValue.number = JsMath::LN10;
			returnValue.type   = VsValue::vNUMBER;
			break;
	    case mLN2:
			returnValue.number = JsMath::LN2;
			returnValue.type   = VsValue::vNUMBER;
			break;
	    case mPI:
			returnValue.number = JsMath::PI;
			returnValue.type   = VsValue::vNUMBER;
			break;
	    case mSQRT1_2:
			returnValue.number = JsMath::SQRT1_2;
			returnValue.type   = VsValue::vNUMBER;
			break;
	    case mSQRT2:
			returnValue.number = JsMath::SQRT2;
			returnValue.type   = VsValue::vNUMBER;
			break;
	    case mUNKNOWN:
			returnValue.type = VsValue::vVOID;
			break;
		}
    } else if (obj->evaluate(f).type == VsValue::vFLOAT_ARRAY) {
		returnValue.type   = VsValue::vNUMBER;
        if (!stricmp(m_identifier,"x")) {
            returnValue.number = obj->evaluate(f).floatArray->array[0];
        } else if (!stricmp(m_identifier,"y")) {
            returnValue.number = obj->evaluate(f).floatArray->array[1];
        } else if (!stricmp(m_identifier,"z")) {
            returnValue.number = obj->evaluate(f).floatArray->array[2];
        } else if (!stricmp(m_identifier,"angle")) {
            returnValue.number = obj->evaluate(f).floatArray->array[3];
        } else {
            returnValue.number = 0;
        }
    } else if (obj->evaluate(f).type == VsValue::vVOID) {
        int nNodeID = ((MediaObject*) ((VsSFVoid*)obj)->GetValue().data)->GetFieldID(m_identifier);
		if (nNodeID == -1) {
			fprintf(stderr, "ERROR: node ID of field %s not found\n", m_identifier);
			exit(-1);
		}     
		MediaObject *pNode = (MediaObject*) ((VsSFVoid*)obj)->GetValue().data;
        NodeField *myfield = pNode->GetField(nNodeID);
		float *ary;
		switch (myfield->GetType()) {
		case SFBOOL:
			returnValue.type = VsValue::vNUMBER;
			returnValue.number = ((SF<Bool> *)myfield)->GetValue();
			break;
		case SFINT32:
			returnValue.type = VsValue::vNUMBER;
			returnValue.number = ((SF<int> *)myfield)->GetValue();
			break;
		case SFFLOAT:
			returnValue.type = VsValue::vNUMBER;
			returnValue.number = ((SF<float> *)myfield)->GetValue();
			break;
		case SFTIME:
			returnValue.type = VsValue::vNUMBER;
			returnValue.number = ((SF<Time> *)myfield)->GetValue();
			break;
		case SFVEC2F:
			returnValue.type = VsValue::vFLOAT_ARRAY;
			returnValue.floatArray = new FloatArray;
			returnValue.floatArray->length = 2;
			ary = new float[2];
			ary[0] = (((SF<Vec2f> *)myfield)->GetValue()).x;
			ary[1] = (((SF<Vec2f> *)myfield)->GetValue()).y;
			returnValue.floatArray->array = ary;
			break;
		case SFVEC3F:
			returnValue.type = VsValue::vFLOAT_ARRAY;
			returnValue.floatArray = new FloatArray;
			returnValue.floatArray->length = 3;
			ary = new float[3];
			ary[0] = (((SF<Vec3f> *)myfield)->GetValue()).x;
			ary[1] = (((SF<Vec3f> *)myfield)->GetValue()).y;
			ary[2] = (((SF<Vec3f> *)myfield)->GetValue()).z;
			returnValue.floatArray->array = ary;
			break;
		case SFCOLOR:
			returnValue.type = VsValue::vFLOAT_ARRAY;
			returnValue.floatArray = new FloatArray;
			returnValue.floatArray->length = 3;
			ary = new float[3];
			ary[0] = (((SF<Color3f> *)myfield)->GetValue()).r;
			ary[1] = (((SF<Color3f> *)myfield)->GetValue()).g;
			ary[2] = (((SF<Color3f> *)myfield)->GetValue()).b;
			returnValue.floatArray->array = ary;
			break;
		case SFROTATION:
			returnValue.type = VsValue::vFLOAT_ARRAY;
			returnValue.floatArray = new FloatArray;
			returnValue.floatArray->length = 4;
			ary = new float[4];
			ary[0] = (((SF<Rotation> *)myfield)->GetValue()).x;
			ary[1] = (((SF<Rotation> *)myfield)->GetValue()).y;
			ary[2] = (((SF<Rotation> *)myfield)->GetValue()).z;
			ary[3] = (((SF<Rotation> *)myfield)->GetValue()).theta;
			returnValue.floatArray->array = ary;
			break;
		case SFSTRING:
			returnValue.type = VsValue::vSTRING;
			returnValue.string = _strdup(((SF<ZString> *)myfield)->GetValue());
			break;
		default:
			fprintf(stderr, "MF vField and Node vField unimplementated\n");
			return VsValue::nothing;
		} 
	}

    return returnValue;
}

VsValue		// NODEFIELD
VsMemberAccess::assign(const VsValue& v, VsFunctionCall * f)
{
	ASSERT(obj != NULL);
	if (obj->evaluate(f).type == VsValue::vVOID) {
        int nNodeID = ((MediaObject*) ((VsSFVoid*)obj)->GetValue().data)->GetFieldID(m_identifier);
		if (nNodeID == -1) {
			fprintf(stderr, "ERROR: node ID of field %s not found\n", m_identifier);
			exit(-1);
		}
        MediaObject *pNode = (MediaObject*) ((VsSFVoid*)obj)->GetValue().data;
	    NodeField *myfield = pNode->GetField(nNodeID);
	    switch (myfield->GetType()) {
	    case SFBOOL:
	    {	ASSERT(v.type == VsValue::vNUMBER);
		    SF<Bool> vField (v.number != 0.0);
		    myfield -> Copy (vField, NULL);
		    break;
	    }
	    case SFINT32:
	    {	ASSERT(v.type == VsValue::vNUMBER);
		    SF<int> vField (v.number);
		    myfield -> Copy (vField, NULL);
		    break;
	    }
	    case SFFLOAT:
	    {	ASSERT(v.type == VsValue::vNUMBER);
		    SF<float> vField (v.number);
		    myfield -> Copy (vField, NULL);
		    break;
	    }
	    case SFTIME:
	    {	ASSERT(v.type == VsValue::vNUMBER);
		    SF<Time> vField (v.number);
		    myfield -> Copy (vField, NULL);
		    break;
	    }
	    case SFVEC2F:
		    ASSERT(v.type == VsValue::vFLOAT_ARRAY);
    //		Vec2f v2f(v.floatArray->array[0], v.floatArray->array[1]);
    //		SF<Vec2f> tmpfield (v2f);
    //		*vField = tmpfield;
		    myfield -> Copy (*(new SF<Vec2f> (* (new Vec2f(v.floatArray->array[0], v.floatArray->array[1])))), NULL);
		    break;
	    case SFVEC3F:
		    ASSERT(v.type == VsValue::vFLOAT_ARRAY);
    //		Vec3f v3f(v.floatArray->array[0], v.floatArray->array[1], v.floatArray->array[2]);
    //		SF<Vec3f> tmpfield(v3f);
    //		*vField = tmpfield;
		    myfield -> Copy (*(new SF<Vec3f> (* (new Vec3f(v.floatArray->array[0], v.floatArray->array[1], v.floatArray->array[2])))), NULL);
		    break;
	    case SFCOLOR:
		    ASSERT(v.type == VsValue::vFLOAT_ARRAY);
    //		Vec3f v3f(v.floatArray->array[0], v.floatArray->array[1], v.floatArray->array[2]);
    //		SF<Color3f> tmpfield (v3f);
    //		*vField = tmpfield;
		    myfield -> Copy (*(new SF<Color3f> (* (new Color3f(v.floatArray->array[0], v.floatArray->array[1], v.floatArray->array[2])))), NULL);

		    break;
	    case SFROTATION:
		    ASSERT(v.type == VsValue::vFLOAT_ARRAY);
    //		Rotation rt(v.floatArray->array[0], v.floatArray->array[1], v.floatArray->array[2], v.floatArray->array[3]);
    //		SF<Rotation> tmpfield (rt);
    //		*vField = tmpfield;
		    myfield -> Copy (*(new SF<Rotation> (* (new Rotation(v.floatArray->array[0], v.floatArray->array[1], v.floatArray->array[2])))), NULL);
		    break;
	    case SFSTRING:
	    {	ASSERT(v.type == VsValue::vSTRING);
		    SF<ZString> vField (v.string);
		    myfield -> Copy (vField, NULL);
		    break;
	    }
	    default:
		    fprintf(stderr, "assign to MF Field and Node Field unimplemented\n");
		    return VsValue::nothing;
	    }
    }
   	if (obj->evaluate(f).type == VsValue::vFLOAT_ARRAY) {
        if (!stricmp(m_identifier,"x")) {
            obj->evaluate(f).floatArray->array[0] = v.number;
        } else if (!stricmp(m_identifier,"y")) {
            obj->evaluate(f).floatArray->array[0] = v.number;
        } else if (!stricmp(m_identifier,"z")) {
            obj->evaluate(f).floatArray->array[0] = v.number;
        } else if (!stricmp(m_identifier,"angle")) {
            obj->evaluate(f).floatArray->array[0] = v.number;
        } else {
        }
    }

	return v;
}


VsNew::VsNew(const char * identifier, VsExprList &exprs)
{
    if (!stricmp(identifier, "SFBool"))
		fType = SFBOOL;
    else if (!stricmp(identifier, "SFColor"))
		fType = SFBOOL;
    else if (!stricmp(identifier, "SFFloat"))
		fType =  SFFLOAT; 
    else if (!stricmp(identifier, "SFImage"))
		fType =  SFIMAGE; 
    else if (!stricmp(identifier, "SFInt32"))
		fType =  SFINT32; 
    else if (!stricmp(identifier, "SFNode"))
		fType =  SFNODE; 
    else if (!stricmp(identifier, "SFRotation"))
		fType =  SFROTATION; 
    else if (!stricmp(identifier, "SFString"))
		fType =  SFSTRING; 
    else if (!stricmp(identifier, "SFTime"))
		fType =  SFTIME; 
    else if (!stricmp(identifier, "SFVec2f"))
		fType =  SFVEC2F; 
    else if (!stricmp(identifier, "SFVec3f"))
		fType =  SFVEC3F; 
    else if (!stricmp(identifier, "MFColor"))
		fType =  MFCOLOR; 
    else if (!stricmp(identifier, "MFFloat"))
		fType =  MFFLOAT; 
    else if (!stricmp(identifier, "MFTime"))
		fType =  MFTIME; 
    else if (!stricmp(identifier, "MFBool"))
		fType =  MFBOOL; 
    else if (!stricmp(identifier, "MFInt32"))
		fType =  MFINT32; 
    else if (!stricmp(identifier, "MFNode"))
		fType =  MFNODE; 
    else if (!stricmp(identifier, "MFRotation"))
		fType =  MFROTATION; 
    else if (!stricmp(identifier, "MFString"))
		fType =  MFSTRING; 
    else if (!stricmp(identifier, "MFVec2f"))
		fType =  MFVEC2F; 
    else if (!stricmp(identifier, "MFVec3f"))
		fType =  MFVEC3F; 
    else {
        char tmp[256];
        sprintf(tmp,"this built in object is not recognized : %s",identifier);
        ERROR(tmp);
        fType = UNKNOWN_TYPE;
    }

	for(int i = 0; i < exprs.GetLength(); i++)
		args += exprs[i];
}

VsValue
VsNew::evaluate(VsFunctionCall *f)
{
    if (fType == UNKNOWN_TYPE) return VsValue(0.0);

	int i, len = args.GetLength();
	VsValue **Valist = new VsValue *[len];
	VsValue result;

    int tmpLen = -1;

	switch(fType)
	{
	/*
	case SFWORLDNODE:
	case SF3DNODE:
	case SF2DNODE:
	case SFSTREAMINGNODE:
	case SFAUDIONODE:
	case SFGEOMETRYNODE:
	case SFAPPEARANCENODE:
	case SFCOLORNODE:
	case SFFONTSTYLENODE:
	case SFTEXTURENODE:
	case SFTEXTURECOORDINATENODE:
	case SFTEXTURETRANSFORMNODE:
	case SFTIMESENSORNODE:
	case SFWORLDINFONODE:
	case SFCOORDINATE2DNODE:
	case SFTOPNODE:
	case SFLINEPROPERTIESNODE:
	case SFMATERIALNODE:
	case SFSHADOWPROPERTIESNODE:
	case SFFACENODE:
	case SFBODYNODE:
	case SFFITNODE:
	case SFFAPNODE:
	case SFFDPNODE:
	case SFFBADEFNODE:
	case SFFBADEFTABLENODE:
	case SFCOORDINATENODE:
	case SFNORMALNODE:
	case SFLAYERNODE:
	*/
	case SFNODE:
        ERROR("new SFNODE not implemented\n");
		return VsValue(0.0); // Not implement right now ??
	case SFBOOL:
	case SFTIME:
    case SFINT32:
    case SFFLOAT:
    {	if(len != 1) {
            ERROR("Incorrect nb args in new");
			return VsValue(0.0);
        }
		VsValue value = args[0]->evaluate(f);
		Valist[0] = &value;

		if(Valist[0]->type == VsValue::vNUMBER)
		{
			result.type = VsValue::vNUMBER;
			result.number = Valist[0]->number;
			return result;
		}
        else {
            ERROR("Incorrect args in new");
			return VsValue(0.0);		
        }
	}
	case SFVEC3F:   
    case SFCOLOR:
        tmpLen = 3;
    case SFROTATION:
        if (tmpLen == -1) { // not already set
            tmpLen = 4;
        }
    case SFVEC2F:
        if (tmpLen == -1) { // not already set
            tmpLen = 2;
        }

        if(len != tmpLen) {
            ERROR("Incorrect nb args in new");
			return VsValue(0.0);
        }

		result.type = VsValue::vFLOAT_ARRAY;
		result.floatArray = new FloatArray;
		result.floatArray->array = new float[len];
		for(i = 0; i < len; i++)
		{	VsValue value = args[i]->evaluate(f);
			Valist[i] = &value;		

			if(Valist[i]->type == VsValue::vNUMBER)
			{
				result.floatArray->array[i] = Valist[i]->number;
			}
            else {
                ERROR("Incorrect args in new");
				return VsValue(0.0);
            }
		}
		return result;
	case SFSTRING:  
    {	if(len != 1) {
            ERROR("Incorrect nb args in new SFString");
			return VsValue("");
        }
		VsValue value = args[0]->evaluate(f);
		Valist[0] = &value;

		if(Valist[0]->type == VsValue::vSTRING)
		{
			result.type = VsValue::vSTRING;
			result.string = Valist[0]->string;
			return result;
		}
        else {
            ERROR("Incorrect args in new SFSTRING");
			return VsValue("");		
        }
    }
	case MFINT32:
	case MFFLOAT:
    case MFTIME:
    case MFBOOL:
		result.type = VsValue::vFLOAT_ARRAY;
		result.floatArray = new FloatArray;
		result.floatArray->array = new float[len];
		for(i = 0; i < len; i++)
		{	VsValue value = args[i]->evaluate(f);
			Valist[i] = &value;		

			if(Valist[i]->type == VsValue::vNUMBER)
			{
				result.floatArray->array[i] = Valist[i]->number;
			}
            else {
                ERROR("Incorrect args in new");
				return VsValue(0.0);
            }
		}
		return result;
	case MFCOLOR:
	case MFROTATION:  		
    case MFSTRING:
	case MFVEC2F:     
	case MFVEC3F:
	default:
        ERROR("type in new not implemented yet");
		return VsValue(0.0);
	}

}

