/*
 * VsFunction.h
 *
 *      Declaration of classes:
 *        VsFunctionDef
 *        VsFunctionCall
 *	  VsArgOrLocal
 *
 * Function support
 * 
 * Copyright (C) 1996, Silicon Graphics,  Inc.
 */

/*---------------------------------------------------------------
This module is copied from the VRML Script interpreter owned by
Silicon Graphics, Inc. It is modified by Wenlong Dong, Wei Wu and
Yuval Fisher(Rockwell Science Center) in the course of development
of the MPEG-4 Systems standard(ISO/IEC 14496-1) and added to the 
subsystem: Script Node.
---------------------------------------------------------------*/

#ifndef _VS_FUNCTION_
#define _VS_FUNCTION_

#include "vsexpr.h"
#include "FldTypes.h"

class DecScript;

//
// Functions are split into two classes:  function definitions, which
// can appear only at the top-level of a file, and function calls,
// which are expressions and can appear wherever an expression
// appears.
//
// Function definition class.
// 
class VsFunctionDef 
{
public:
    // FunctionDef constructor.  Called during parsing; gotta keep
    // track of current function so that we know if identifiers found
    // inside the function are function arguments, local variables, or
    // global variables.  The VsParseTree argument is needed because
    // global variables are kept by the ParseTree class.
    VsFunctionDef(const char *name, DecScript *parent);

    // Again, used during parsing, adds arguments.  It is assumed that
    // this is called BEFORE findIdentifier() is called.
    static void		addArg(const char *argName);

    // Returns the local variable, argument, or global variable with
    // the given name.  A new local variable is automatically created
    // and returned if the name isn't already known in this function.
    static VsVar *	findIdentifier(const char *name, bool n_flag = false); // nflag is for member access

    // And when function is done, this cleans everything up and
    // remembers what statements are part of this function definition:
    void		endDefinition(VsStatementList *stmnts);

    // Returns number of local vars and args.  Determines how many
    // VsValues need to be pushed onto stack at a function call:
    int			getNumLocals() { return numArgs+numLocals; }

    // Returns number of arguments.  Event in functions need to
    // have two arguments, so we need to test for this.
    int			getNumArgs() { return numArgs; }

    // Returns name of this function definition
    const char *	getName() { return name; }
    
    // Statements that make up this function.  Used by VsFunctionCall.
    VsStatementList *statements;

    virtual ~VsFunctionDef();

private:

    // When parsing, function being defined:
    static VsFunctionDef *curFunction;

    char *name;			// This function's name
    int numArgs;		// Number of arguments
    int numLocals;		// Number of local variables
//    VbPList argOrLocalNames;	// Names of arguments and local vars
	ZArray<char *> argOrLocalNames;
//    VsParseTree *parent;	// Script we're defined in
	DecScript *parent;
};

class VsFunctionCall : public VsExpr 
{
public:
    
    VsFunctionCall(VsFunctionDef *, VsExprList &exprs);
	void SetFuncDef(VsFunctionDef *_func) { function = _func;}

    // Evaluate method.
    virtual VsValue		evaluate(VsFunctionCall *parent);

    // The enclosing VsFunctionCall is passed to all expressions and
    // statements so that a return statement can turn around and call
    // this:
    void		setReturnValue(const VsValue &val);

    // Argument getting/setting routines, used by VsArgOrLocal class.
    VsValue		getArg(int whichArg);
    void		setArgOrLocal(int whichArg, const VsValue &newVal);    
    static int		isAnUnimplementedFunction(const char* fName);

protected:

    virtual ~VsFunctionCall();

private:

    VsFunctionDef *function;	// Function definition contains statements
    VsExprList args;		// Argument expressions
    VsValue *locals;		// Arguments and local variables
    VsValue returnValue;
}; 

class VsMethodCall : public VsExpr 
{
public:
	enum Function 
	{
	    // Math library functions
	    fABS, 
	    fACOS, 
	    fASIN, 
	    fATAN, 
	    fCEIL, 
	    fCOS, 
	    fEXP, 
	    fFLOOR, 
	    fLOG, 
	    fMAX, 
	    fMIN, 
	    fPOW, 
	    fRANDOM, 
	    fROUND, 
	    fSIN, 
	    fSQRT, 
	    fTAN, 
	    fUNKNOWN, 
	};

	VsMethodCall(VsExpr* object,  const char* funcName,  VsExprList &exprs);
	   
    // Evaluate method.
    virtual VsValue		evaluate(VsFunctionCall *);

protected:
	virtual ~VsMethodCall();
	
private:
	VsExpr*   obj;
	Function func;
	VsExprList args;
	
	VsValue returnValue;
};

class VsMemberAccess : public VsVar 
{
public:
	enum Member {
	    // Math library members
	    mE, 
	    mLN10, 
	    mLN2, 
	    mPI, 
	    mSQRT1_2, 
	    mSQRT2, 
	    mUNKNOWN, 
	};

	VsMemberAccess(VsExpr* obj,  const char* memberName);
	   
    // Evaluate method.
    virtual VsValue		evaluate(VsFunctionCall *);
//  virtual VsValue		assign(const VsValue &, VsFunctionCall *) { return 0.0; }
    virtual VsValue		assign(const VsValue &, VsFunctionCall *); // NODEFIELD
protected:
	virtual ~VsMemberAccess();
	
private:
	VsExpr* obj;
	Member member;
    const char * m_identifier;	
	VsValue returnValue;
};

//
// Function argument or local variable (they're both implemented in
// exactly the same way-- as an integer that refers to a VsValue
// created at a VsFunctionCall).
//
class VsArgOrLocal : public VsVar 
{
public:
    VsArgOrLocal(int whichArg);

    virtual VsValue		evaluate(VsFunctionCall *);
    virtual VsValue		assign(const VsValue &, VsFunctionCall *);

protected:
    virtual ~VsArgOrLocal();

private:
    int whichArg;
};

// 
// New method expression:
//
class VsNew: public VsExpr 
{
public:

	VsNew(const char * identifier, VsExprList &exprs);
	virtual VsValue evaluate(VsFunctionCall *);
protected:

	virtual ~VsNew(){}
private:

	FIELD_TYPE fType;
    VsExprList args;
};

#endif /* _VS_FUNCTION_ */
