/***********************************************************************
This software module was originally developed by Zvi Lifshitz
(Triton R&D Ltd.) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. Triton R&D Ltd. retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  1997.
************************************************************************

For more information or to receive an up to date version of this module,
you may call Zvi Lifshitz at zvil@csi.com.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  BIFS Decoder/ Compositor
FILE:       NodeFld.cpp
AUTHOR:     Zvi Lifshitz (Triton R&D Ltd.)
OVERVIEW:   Implementation of NodeField and derivations.
-------------------------------------------------------------------------*/
#include "defs.h"
#include "Nodes.h"
#include "NodeFld.h"
#include "MediaObj.h"
#include "BifsDec.h"
#include "Route.h"
#include "Root.h"
#include "Quantize.h"
#include "NodeCode.h"

/*-----------------------------------------------------------------
NodeField destructor. Delete all routes
-----------------------------------------------------------------*/
NodeField :: ~NodeField ()
{
	while (m_pFirstRoute)
		delete m_pFirstRoute;  // this unlinks the route automatically
}

/*-----------------------------------------------------------------
Add a Route to the nodeField
-----------------------------------------------------------------*/
void NodeField :: LinkRoute (Route *pRoute)
{
	pRoute -> Link (m_pFirstRoute);
}

/*-----------------------------------------------------------------
Unlink a Route from the list of routes
-----------------------------------------------------------------*/
void NodeField :: UnlinkRoute (Route *pRoute)
{
	pRoute -> Unlink (m_pFirstRoute);
}

/*-----------------------------------------------------------------
Activate attached routes on nodeField update
-----------------------------------------------------------------*/
void NodeField :: OnFieldUpdate ()
{
	for (Route *pRoute = m_pFirstRoute;
		pRoute != NULL;
		pRoute = pRoute -> GetNext ()
	)
		pRoute -> Activate ();
#ifdef BIFSCAR
	m_bNeedsEncoding = true;
#endif
}

/*-----------------------------------------------------------------
Parse an SFInt
-----------------------------------------------------------------*/
BOOL SF<int> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject*, 
	Quantizer *pQuant
)
{
	fieldType = SFINT32;	// SCRIPT

	if(!pQuant || !pQuant -> IsActive ()) 
		m_value = pCoder -> ParseInt (32, "intValue");
	else 
		m_value = pQuant -> ParseInt (pCoder);

	TRACEF ("%d", m_value);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFBool
-----------------------------------------------------------------*/
BOOL SF<Bool> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	fieldType = SFBOOL;	// SCRIPT
	m_value = pCoder -> ParseInt (1, "boolValue") != 0;
	TRACEF ("%s", m_value ? "TRUE" : "FALSE");
	return TRUE;
}

/*-----------------------------------------------------------------
Parse and SFFloat
MBS: modified to take into account quantization parameters if any
-----------------------------------------------------------------*/
BOOL SF<float> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * pQuant
)
{
	fieldType = SFFLOAT;	// SCRIPT
	if(!pQuant || !pQuant -> IsActive ()) 
		m_value = pCoder -> ParseFloat ("floatValue");
	else 
		m_value = pQuant->ParseFloat(pCoder);

	TRACEF ("%.2f", m_value);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFTime
-----------------------------------------------------------------*/
BOOL SF<Time> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	fieldType = SFTIME;	// SCRIPT
	m_value = pCoder -> ParseDouble("doubleValue");
	TRACEF ("%.2lf", m_value);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFVec3f
MBS: modified to take into account quantization parameters if any
-----------------------------------------------------------------*/
BOOL SF<Vec3f> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * pQuant
)
{
	fieldType = SFVEC3F;	// SCRIPT
	if(!pQuant || !pQuant->IsActive())
	{	m_value.x = pCoder -> ParseFloat("x");
		m_value.y = pCoder -> ParseFloat("y");
		m_value.z = pCoder -> ParseFloat("z");
	}
	else
		m_value = pQuant -> ParseVec3f(pCoder);

	TRACEF3 ("%.2f %.2f %.2f", m_value.x, m_value.y, m_value.z);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFVec2f
MBS: modified to take into account quantization parameters if any
-----------------------------------------------------------------*/
BOOL SF<Vec2f> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * pQuant
)
{
	fieldType = SFVEC2F;	// SCRIPT
	if(!pQuant || !pQuant->IsActive())
	{	m_value.x = pCoder -> ParseFloat("x");
		m_value.y = pCoder -> ParseFloat("y");
	}
	else
		m_value = pQuant -> ParseVec2f(pCoder);

	TRACEF2 ("%.2f %.2f", m_value.x, m_value.y);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFRotation
MBS: modified to take into account quantization parameters if any
-----------------------------------------------------------------*/
BOOL SF<Rotation> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * pQuant
)
{
	fieldType = SFROTATION;	// SCRIPT
	if(!pQuant || !pQuant->IsActive())
	{	m_value.x = pCoder -> ParseFloat("x");
		m_value.y = pCoder -> ParseFloat("y");
		m_value.z = pCoder -> ParseFloat("z");
		m_value.theta = pCoder -> ParseFloat("theta");
	}
	else
		m_value = pQuant -> ParseRotation(pCoder);

	TRACEF4 ("%.2f %.2f %.2f %.2f", m_value.x, m_value.y, m_value.z, m_value.theta);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFRotation2D
MBS: modified to take into account quantization parameters if any
-----------------------------------------------------------------*/
BOOL SF<Rotation2D> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * pQuant
)
{
  if(!pQuant || !pQuant->IsActive()) {
    m_value.x = pCoder -> ParseFloat("x");
    m_value.y = pCoder -> ParseFloat("y");
    m_value.theta = pCoder -> ParseFloat("theta");
  }
  else
    m_value = pQuant -> ParseRotation2D(pCoder);

  TRACEF3 ("%2.f %.2f %.2f", m_value.x, m_value.y, m_value.theta);
  return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFColor
MBS: modified to take into account quantization parameters if any
-----------------------------------------------------------------*/
BOOL SF<Color3f> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * pQuant
)
{
	fieldType = SFCOLOR;	// SCRIPT
	if(!pQuant || !pQuant->IsActive())
	{	m_value.r = pCoder -> ParseFloat("r");
		m_value.g = pCoder -> ParseFloat("g");
		m_value.b = pCoder -> ParseFloat("b");
	}
	else
		m_value = pQuant -> ParseColor3f(pCoder);

	TRACEF3 ("%.2f %.2f %.2f", m_value.r, m_value.g, m_value.b);
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an SFObjectID
-----------------------------------------------------------------*/
BOOL SF<Url> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	
	if (pCoder -> ParseInt (1, "isOD"))
	{	m_value.ODid = pCoder -> ParseInt (10, "ODid");
		TRACEF ("%d", m_value.ODid);
	}
	else
	{	m_value.ODid = -1;
		m_value.urlValue.Parse (pCoder, NULL);
	}
	return TRUE;
}

/*-----------------------------------------------------------------
Parse SFString
MBS 110698: updated for UTF-8 String
Zvi 021799: redo MBS change. The string length is given in bytes, not
in characters. The handling of UTF-8 coding shall be done by the compositor -
it is not the matter of the parser.
-----------------------------------------------------------------*/
BOOL SF<ZString> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	fieldType = SFSTRING;	// SCRIPT
	m_value.Cut (0);
	int nBits = pCoder -> ParseInt (5, "nbits");
	int nSize = pCoder -> ParseInt (nBits, "stringLength");

	while (--nSize >= 0)
		m_value += pCoder -> ParseInt (8, "value");
	TRACEF ("%s", LPCSTR (m_value));
	return TRUE;
}

/*-----------------------------------------------------------------
Parse SFImage
MBS 110498: this is a new type
-----------------------------------------------------------------*/
BOOL SF<Image> :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject * /*pNode*/, 
	Quantizer * /*pQuant*/
)
{
	m_value.width = pCoder -> ParseInt (12, "width");
	m_value.height = pCoder -> ParseInt (12, "height");
	m_value.numComponents = pCoder -> ParseInt (2, "numComponents");

	//Alvar 981223: Added pixel array allocation and refcount stuff
	int nPixels = m_value.numComponents * m_value.width * m_value.height;
	BYTE *p = m_value.pixels = new BYTE [nPixels];
	m_value.m_pRefCnt = new int;
	*(m_value.m_pRefCnt) = 1;  
	for (int i = 0; i < nPixels; i++ )
		*p++ = pCoder -> ParseInt (8, "pixel");
	// end of Alvar changes

	TRACEF ("Image %d ", m_value.width);
	TRACEF ("Image %d ", m_value.height);
	TRACEF ("Image %d ", m_value.numComponents);
	return TRUE;
}

/*-----------------------------------------------------------------
A dummy function
-----------------------------------------------------------------*/
BOOL SF<MediaObject *> :: Parse (
	BIFSDecoder *,
	MediaObject *, 
	Quantizer *
)
{
	return FALSE;
}

/*-----------------------------------------------------------------
SFGenericNode Destructor.  Deletes the MediaObject, unless it is defined
by other node and only USEd here.
-----------------------------------------------------------------*/
SFGenericNode :: ~SFGenericNode ()
{
	if (m_pNode)
		m_pNode -> RemoveContainer (this);
}

/*-----------------------------------------------------------------
Assign a value into a generic-node. If the old value is not a copy 
of another nodeField, delete the old node
-----------------------------------------------------------------*/
MediaObject *SFGenericNode :: operator = (MediaObject *node)
{
	if (m_pNode)
		m_pNode -> RemoveContainer (this);
	m_pNode = node;
//	if (node)
//		node -> SetContainer (this);
	OnFieldUpdate ();
	#ifdef BIFSCAR
	if (node == NULL_OBJECT)
		SetNeedsEncoding (false);
	#endif
	return node;
}

/*-----------------------------------------------------------------
Parse a node field which is a sub-node
-----------------------------------------------------------------*/
BOOL SFGenericNode :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject *pNode, 
	Quantizer * /*pQuant*/
)
{
	m_pParent = pNode;
	return ParseNodeField (pCoder, pNode, this, FALSE);
}

/*-------------------------------------------------------------
This function does the actual work of node parsing when the node
is either a scalar field in a parent node (in which case the function
is called from SFGenericNode :: Parse), an element in a vector field,
or when it is parsed in ReplaceNode command. The later two cases ParseNode
is called directly, not through Parse().
-------------------------------------------------------------*/
BOOL SFGenericNode :: ParseNodeField (
	BIFSDecoder *pCoder, 
	MediaObject *pNode,			// the node this field belongs to
	NodeField *pContainer,		// a pointer to the container field
								// (this is not always 'this' because
								// a temporary NodeField object can be
								// used to parse the node, and then the
								// node is assigned or inserted in another
								// NodeField object.
	BOOL bISMF					// TRUE when the field is an MF field
)
{
	if (pCoder -> ParseInt (1, "isReused"))	// if isReused
	{	int nID = pCoder -> ParseInt (pCoder -> GetNodeIDbits (), "NodeID");
		TRACEF ("USE %d\n", nID);
		if (pNode &&
			(m_pNode = pNode -> GetRoot () -> FindNode (nID)) != NULL
		)
			m_pNode -> SetContainer (pContainer, bISMF, pNode);
		else
			return FALSE;
	}
	else
	{	
		/*-----------------------------------------------------------------
		Get object type, instantiate the object, and let it parse itself
		-----------------------------------------------------------------*/
		int nID = -1;
//		pCoder -> ParseInt (1);		// predefined assumed to be zero
									// deleted in the FDIS
		int nNodeType =
			pCoder -> ParseInt (BitFieldWidth (GetNodeTableSize () + 1), "nNodeType");
											// +1 added for FDIS
		if (pCoder -> ParseInt (1, "isUpdatable")) // if isUpdatable
		{	
			nID = pCoder -> ParseInt (pCoder -> GetNodeIDbits (), "NodeID");
			TRACEF ("DEF %d ", nID);
		}
	
		/*-------------------------------------------------------------
		m_pNode may already point to another node if the function is called
		during a ReplaceField command.
		-------------------------------------------------------------*/
		if (m_pNode)
			m_pNode -> RemoveContainer (pContainer, pNode);

		/*-------------------------------------------------------------
		Create a new node
		-------------------------------------------------------------*/
		m_pNode = CreateNode (nNodeType);
		if (!m_pNode)
			return FALSE;

		// SB 99/03: SetExecutive AFTER SetRoot
		/*** >>>>
		m_pNode -> SetExecutive (pCoder -> GetExecutive ());
		<<<< ***/
		m_pNode -> SetID (nID);
		m_pNode -> SetContainer (pContainer, bISMF, pNode);

		/*-----------------------------------------------------------------
		The root of the new object is the same as of the container 
		object. If the container object is itslef a root, the root is
		the container object.
		-----------------------------------------------------------------*/
		if (pNode)
			m_pNode -> SetRoot (pNode -> GetRoot ());
		else if (m_pNode -> IsRoot ())
			m_pNode -> SetRoot ((Root *) m_pNode);

		// SB 99/03: see above
		m_pNode -> SetExecutive (pCoder -> GetExecutive ());

		/*-------------------------------------------------------------
		Parse the node
		-------------------------------------------------------------*/
		m_pNode -> Parse (pCoder);

		// MBS: sets current QP in decoder if we have a QP node
		if(m_pNode->GetNodeCode() == QuantizationParameter :: GetCode()) 
		  pCoder->SetCurrentQP(m_pNode);
		
		// end MBS
	}
	return TRUE;
}

/*-----------------------------------------------------------------
Parse an MF nodeField
-----------------------------------------------------------------*/
BOOL MField :: Parse (
	BIFSDecoder *pCoder, 
	MediaObject *pNode, 
	Quantizer *pQuant
)
{
	TRACEF ("[", 0);
	TRACENEWLINE;
	TRACEINDENT (1);

	Clear ();
	/*-----------------------------------------------------------------
	MF nodeField may be coded as ListMFDescription or VectorMFDescription
	-----------------------------------------------------------------*/
        // MBS: variables for quantization
        BOOL localQP=FALSE;
        MediaObject *qpCurrent, *qpOld = NULL;
        // end MBS
	if (!pCoder -> ParseInt (1, "reserved"))		// reserved, added in FDIS
	{	if (pCoder -> ParseInt (1, "isListDescription"))	// isListDescription
			while (pCoder -> ParseInt (1, "end") == 0)	// changed for  FDIS
			{
				// MBS: saves old QP
				if(!localQP) qpOld=pCoder->GetCurrentQP();
				// end MBS

				if (!AppendElement (pCoder, pNode, pQuant))
					return FALSE;
		
				// MBS: restores old QP or saves current QP
				qpCurrent=pCoder->GetCurrentQP();
				if(pCoder->hasLocalQP()) {
					if(localQP) {
						localQP=FALSE;
						pCoder->SetCurrentQP(qpOld);
					}
					else {
						localQP=TRUE;
					}
				}
				// end MBS

				TRACENEWLINE;
			}
		else
			for (int i = pCoder -> ParseInt (pCoder -> ParseInt (5, "nBits"), "size"); --i >= 0;)
			{
					// MBS: saves old QP
					if(!localQP) qpOld=pCoder->GetCurrentQP();
					// end MBS

				if (!AppendElement (pCoder, pNode, pQuant))
					return FALSE;

				// MBS: restores old PQ or saves current QP
				qpCurrent=pCoder->GetCurrentQP();
				if(pCoder->hasLocalQP()) {
				  if(localQP) {
					localQP=FALSE;
					pCoder->SetCurrentQP(qpOld);
				  }
				  else {
					localQP=TRUE;
				  }
				}
				// end MBS

				TRACENEWLINE;
			}
	}

	TRACEINDENT (-1);
	TRACEF ("]", 0);
	TRACENEWLINE;
		
	return TRUE;
}

/*-----------------------------------------------------------------
Deletes all nodes
-----------------------------------------------------------------*/
void MFGenericNode :: RemoveAll ()
{
	for (int i = GetLength (); --i >= 0;)
		if (m_array [i])
			m_array [i] -> RemoveContainer (this, m_pParent);
}

/*-----------------------------------------------------------------
Manipulation of MFGenericNode arrays
-----------------------------------------------------------------*/
void MFGenericNode :: InsertAt (int index, MediaObject *value, BOOL bRoute)
{
	m_array.InsertAt (index, value);
	if (bRoute)
		OnFieldUpdate ();
}

void MFGenericNode :: Append (MediaObject *value, BOOL bRoute)
{
	if (value -> GetNodeCode () != CODE_Script)	// SCRIPT
		m_array += value;
	if (bRoute)
		OnFieldUpdate ();
}

void MFGenericNode :: SetAt (int index, MediaObject *value, BOOL bRoute)
{
	if (index < GetLength () && m_array [index] != NULL)
		m_array [index] -> RemoveContainer (this, m_pParent);
	m_array.SetAt (index, value);
	if (bRoute)
		OnFieldUpdate ();
}

void MFGenericNode :: DeleteFrom (int index, BOOL bRoute)
{
	if (index < GetLength () && m_array [index] != NULL)
		m_array [index] -> RemoveContainer (this, m_pParent);
	m_array.DeleteFrom (index);
	if (bRoute)
		OnFieldUpdate ();
	#ifdef BIFSCAR
	if (m_array.GetLength () == 0)
		SetNeedsEncoding (false);
	#endif
}

/*-----------------------------------------------------------------
Assigment operator activated in case of Routes. We assume that
the nodeField argument is always of the same type as this, therefor
allow for deliberate downcasting. This assumption is dangerous.
Commercial application should implement safe casting or proper
conversion mechanism.
-----------------------------------------------------------------*/
void MFGenericNode :: operator = (NodeField &nodeField)
{
	MFGenericNode &fld = (MFGenericNode &) nodeField;
	for (int i = GetLength (); --i >= 0;)
		DeleteFrom (i, FALSE);		
	for (i = 0; i < fld.GetLength (); i++)
		Append (fld [i], FALSE);
	OnFieldUpdate ();
}

/*-------------------------------------------------------------
The node in this field has been replaced by ReplaceNode command
-------------------------------------------------------------*/
void SFGenericNode :: ReplaceNode (MediaObject *pNew)
{
	m_pNode = pNew;
	pNew -> SetContainer (this, FALSE, m_pParent);
}

/*-------------------------------------------------------------
Node pOld in this MFxxxNode field has been replaced by a 
ReplaceNode command.
-------------------------------------------------------------*/
void MFGenericNode :: ReplaceNode (MediaObject *pOld, MediaObject *pNew)
{
	for (int i = m_array.GetLength (); --i >= 0; i)
		if (m_array [i] == pOld)
		{	m_array [i] = pNew;
			pNew -> SetContainer (this, TRUE, m_pParent);
			return;
		}
}
