/* bapdf.C */


/****************** deformations of body animation parts ******************/

#define EXTERN
#include <bddef.h>
#include <math.h>

/* external */

void skip_comments (ifstream & inf);                     /* bapio.C */

/****************************************************************************/


/*------------------ utility functions -------------------------*/


/* magnitude of a vector of any type of a given dimension; result is float */

template <class T> float magnitude_vector (T v[], int dim)

{
  int i;
  float mag = 0;

  for (i = 0; i < dim; i++)
    mag += (v[i]*v[i]);
  mag = sqrt (mag);
  return mag;
}


/* normalize a vector of given dimension */

template <class T> void normalize_vector (T v[], int dim)

{
  float mag = magnitude_vector (v, dim);
  int i;

  for (i = 0; i < dim; i++)
    v[i] /= mag;
}


/* dot product between two bap combinations of a given dimension */

template <class T> float dotp_vectors (T v1[], T v2[], int dim)

{
  int i;
  float dotp = 0;

  for (i = 0; i < dim; i++)
    dotp += (v1[i]*v2[i]);
  return dotp;
}


/* create vector floating pt from two points: p1 to p2 (p1->p2) */

template <class T> void create_float_vector(float v[], T p1[], T p2[], int dim)

{
  for (int i = 0; i < dim; i++)
    v[i] = p2[i] - p1[i];
}


/* template for distance between two points of any type of
   a given dimension; result is float */

template <class T> float dist_bet_points (T bc1[], T bc2[], int dim)

{
  int i;
  float dist = 0;
  float f;

  for (i = 0; i < dim; i++) {
    f = bc1[i]-bc2[i];
    dist += (f*f);
  }
  dist = sqrt ((float) dist);
  return dist;
}


/* template for bubble sort: (adapted from C++ prog. lang. Strostrup p.271);
 * while values are sorted, an associative index array is also rearranged 
 * accordingly;
 * sorting in increasing order;
 */

template <class T> void associative_bubble_sort (int n, T v[], int assoc[])

{
  for (int i = 0; i < n-1; i++)
    for (int j = n-1; i < j; j--)
      if (v[j] < v[j-1]) { /* swap v[j] and v[j-1] */
	T temp = v[j];
	v[j] = v[j-1];
	v[j-1] = temp;
	int a_temp = assoc[j];
	assoc[j] = assoc[j-1];
	assoc[j-1] = a_temp;
      }
}


/*-------------------- derived quantities --------------------*/

/* set centre of gravoty for bap combination points */

void Bat::derive_bap_comb_cg ()

{
  int i, j, start;

  /* get storage for bap combination cg point */
  bap_comb_cg = x_alloc (Bap_comb_t, N_bap_ids);
  /* reset */
  for (j = 0; j < N_bap_ids; j++) bap_comb_cg[j] = 0;
  /* add each of N_bap_ids components from all N_bap_combs combinations */
  for (i = 0; i < N_bap_combs; i++) {
    start = i*N_bap_ids;
    for (j = 0; j < N_bap_ids; j++) 
      bap_comb_cg[j] += bap_comb_entries[start+j];
  }
  //cout << "Total of Bap combinations:\n";
  /*for (j = 0; j < N_bap_ids; j++)
    cout << bap_comb_cg[j] << ' ';
    cout << '\n';*/
  /* get center of gravity by taking average */
  for (j = 0; j < N_bap_ids; j++)
    bap_comb_cg[j] = bap_comb_cg[j]/(float)N_bap_combs;
  //cout << "Center of gravity of Bap combinations:\n";
  /*for (j = 0; j < N_bap_ids; j++)
    cout << bap_comb_cg[j] << ' ';
  cout << '\n';*/
}


/* derive bap combination vectors from BAP key points */

void Bat::derive_bap_comb_vectors ()

{
  int i, j, start;

  bap_comb_v = x_alloc (Bap_vector_t *, N_bap_combs);
  for (i = 0; i < N_bap_combs; i++)
    bap_comb_v[i] = x_alloc (Bap_vector_t, N_bap_ids);
  /* get vectors w.r.t cg from bap combination vals */
  for (i = 0; i < N_bap_combs; i++) {
    start = i*N_bap_ids;
    //cout << "Bap vector " << i+1 << '\n';
    for (j = 0; j < N_bap_ids; j++) {
      bap_comb_v[i][j] = bap_comb_entries[start+j]; 
      //cout << bap_comb_v[i][j] << "   ";
    }
    //cout << '\n';
  }
}

/*-------------------- constructor/destructor --------------------*/

/* Bat constructor */

Bat::Bat () 

{
  N_bap_ids = N_vertex_ids = N_bap_combs = N_displacements = 0;
  bap_ids = vertex_ids = NULL; bap_comb_entries = NULL;
  displacements = NULL;
  next = NULL;
}


/* Bat destructor */

Bat::~Bat()

{
  if (N_bap_ids > 0) {
    /* BAT table exists; delete allocated memories */
    delete bap_ids; delete vertex_ids;
    delete bap_comb_entries;
    delete displacements;
    delete bap_comb_cg;
    for (int i = 0; i < N_bap_combs; i++)
      delete bap_comb_v[i];
    delete bap_comb_v;
  }
  if (next != NULL) {
    /* recursively delete next BAT table */
    cout << "Deleting Next Table\n";
    delete next;
  }
}


/* Bap constructor */

Bap::Bap ()

{
  N_vertices = 0; N_polygons = 0;
  normal_displacements = NULL;
}



/* Bap destructor */

Bap::~Bap ()

{
  delete normal_displacements;
  delete total_displacements;
  delete vertex_connectivity;
}

/*----------------------------------------------------------------*/


/* find 'n' closest BAP points to a given point; n <= N_bap_combs;
 * 'v' array of indices into bap points in ascending order of distance;
 */

void Bat::closest_points_to_bap_point (Bap_comb_t p[], 
				       int & n, int v[], float dist[])

{
  int i, j;
  Bap_vector_t *p_local = new Bap_vector_t[N_bap_ids]; /* type converted p */
  float avg_dist;

  if (n > N_bap_combs) n = N_bap_combs; /* limit error check */
  /* get p in Bap_vector_t type */
  for (j = 0; j < N_bap_ids; j++) {
    /* make non-active dimensional vals 0 (probably already 0) */
    p_local[j] = p[j]; 
  }
  
  /* get distances of all bap points from p_local */
  for (i = 0; i < N_bap_combs; i++) {
    dist[i] = dist_bet_points (bap_comb_v[i], p_local, N_bap_ids);
    v[i] = i; /* indices of BAP keys */
  }
  /* sort values in 'dist' and rearrange indices in 'v' accordingly */
  associative_bubble_sort (N_bap_combs, dist, v);
  /* consider only those closest to p */
  for (i = 2; i < n; i++) {
    /* get avg distnace of i closest points */
    avg_dist = 0;
    for (j = 0; j < i; j++) {
      avg_dist += dist[j];
    }
    avg_dist /= i;
    /* consider next key only if it is within average distance 
       (multiplied by a factor) of previous keys */
    if (dist[i] > 1.5*avg_dist) {
      /* this key is too far; limit upto i-1 */
      n = i;
      break;
    }
  }
  delete p_local;
}


/* generate 'total displacement' of body part due to deformation by
   adding interpolated displacement (bat_disp)
   from BAT tables to normal (default)
   displacement in BAP */

void Bat::add_bat_displacements (int n, float total_disp[], float bat_disp[])

{
  int     vertex_id, bap_vertex_loc, i, j;

  /* consider displacements of vertices from vertex_ids list in BAT */
  j = 0;
  for (i = 0; i < N_vertex_ids; i++) {
    vertex_id = vertex_ids[i];
    /* get location of vertex in BAP normal displacement array */
    bap_vertex_loc = 3*vertex_id; 
    /* add normal displacement of vertex_id to bat displacement */
    total_disp[bap_vertex_loc]   += bat_disp[j];
    total_disp[bap_vertex_loc+1] += bat_disp[j+1];
    total_disp[bap_vertex_loc+2] += bat_disp[j+2];
    j += 3;
    if ( (bap_vertex_loc+2) >= n) {
      cout << "ERROR (add_bat_displacements): index out of range\n";
      cout.flush();
    }
  }
}


/* generate 'total displacement' of body part due to deformation by
   adding interpolated displacement (bat_disp)
   from BAT tables to normal (default)
   displacement in BAP */

void Bap::generate_total_bap_displacement_UNUSED (float bat_disp[])

{
  int     vertex_id, bap_vertex_loc, i, j;

  /* transfer normal displacements into total (as default) */
  for (i = 0; i < 3*N_vertices; i++) 
    total_displacements[i] = normal_displacements[i];

  /* consider displacements of vertices from vertex_ids list in BAT */
  j = 0;
  for (i = 0; i < N_vertex_ids; i++) {
    vertex_id = vertex_ids[i];
    /* get location of vertex in BAP normal displacement array */
    bap_vertex_loc = 3*vertex_id; 
    /* add normal displacement of vertex_id to bat displacement */
    total_displacements[bap_vertex_loc]   += bat_disp[j];
    total_displacements[bap_vertex_loc+1] += bat_disp[j+1];
    total_displacements[bap_vertex_loc+2] += bat_disp[j+2];
    j += 3;
  }
}


/* produce deformation for given bap combination point for a given 
 * BAT table; The BAT table specifies deformations at 'key' BAP
 * points. By considering keys closest to the given point deformations
 * are obtsained through linear interpolation
 */

void Bat::deform_bap_using_bat (int n_total, float total_disp[], 
				Bap_comb_t p[])

{
  int    *v = new int[N_bap_combs];
  int     tab_length = 3*N_vertex_ids;
  float  *disps = new float[tab_length];
  float  *dist = new float[N_bap_combs];
  int     i, j, n, tab_no;
  Bap_vector_t *p_conv = new Bap_vector_t[N_bap_ids];
  float **bv;
  float   cumulative_dist;

  /* minimum no of keys for forming N-simplex volume element */
  n = N_bap_ids + 1;
  if (n < 2) {
    cout << "ERROR (deform_bap_using_bat): needs at least 2 keys\n";
    cout.flush();
    return;
  }
  if (n > N_bap_combs)
    /* not enough keys; interpolate with existing ones;
       (interpolation takes place inaccurately at lower dimensions) */
    n = N_bap_combs;

  cout << "No of Keys: " << n << '\n';

  /* assign space for BAP vectors derived from keys */
  bv = x_alloc (float *,  n);
  for (i = 0; i < n; i++)
    bv[i] = new float[N_bap_ids];
  /* get n closest bap combination key points to 'p' */
  closest_points_to_bap_point (p, n, v, dist);
  cout << "No of Keys after eliminating far off ones: " << n << '\n';

  float  *disp_factor = new float[n];   /* displacement factor */
  float  *distance = new float[n];      /* distances from n keys */
  int    *disp_indices = new int[n];    /* indices into diaplcement table */

  /* print point and other vectors closest */
  cout << "BAP POINT:   ";
  for (j = 0; j < N_bap_ids; j++) 
    cout << p[j] << "   ";
  cout << '\n';
  /* print closest BAP combinations */
  cout << "CLOSEST COMBINATIONS\n";
  for (i = 0; i < n; i++) {
    tab_no = v[i];   /* table no */
    for (j = 0; j < N_bap_ids; j++)
      cout << bap_comb_v[tab_no][j] << "   ";
    cout << '\n';
  }
  //cout << '\n';
  
  /* interpolate displacement at 'p' by considering key displacements
     at closest BAP combimations keys */
  if (N_bap_ids > 0) {
    /* BAT table exists for this BAP */
    cout << "Deforming the body part " << bodySceneGraphNode << '\n';
    /* convert bap point p to bap-vector form */
    for (j = 0; j < N_bap_ids; j++)
      p_conv[j] = p[j];
    /* consider closest two combinations */
    /* vectors from closest keys to bap point: bv */
    for (i = 0; i < n; i++)
      create_float_vector (bv[i], bap_comb_v[v[i]], p_conv, N_bap_ids);
    /* get magnitudes */
    for (i = 0; i < n; i++)
      distance[i] = magnitude_vector (bv[i], N_bap_ids);
    /* get cumulative distance */
    cumulative_dist = 0;
    for (j = 0; j < n; j++)
      cumulative_dist += distance[j];
    if (cumulative_dist > 0) {
      /* if keys are located at non-zero distance displacements will be > 0
	 (zero distance is a possibility due to float precison problem) */
      for (j = 0; j < n; j++) {
	/* every combination contributes to displacement 
	 * which is inversely proportional to distance from given BAP pt
	 */
	disp_factor[j] = distance[j]/cumulative_dist; 
	/* make inverse relation w.r.t distance:
       * working out gives: summation (disp_factor[j]) = 1.0
       */
	disp_factor[j] = (1 - disp_factor[j])/(n-1);
      }
    }
    else {
      /* then there is a single key OR all keys along with 
	 the point are too close; in latter case considering a single
	 key (1st one) is enough: i.e. make 1st factor 1 and others 0 */
      for (j = 0; j < n; j++) disp_factor[j] = 0;
      disp_factor[0] = 1;
    }
    /* compute indices of deformation entries in the table for key BAPs */
    for (j = 0; j < n; j++)
      disp_indices[j] = v[j]*tab_length;
    for (j = 0; j < tab_length; j++) {
      /* interpolate displacement vector as an addition of
       * key displacements multiplied by respective factors
       */
      disps[j] = 0; /* reset displacements */
      for (i = 0; i < n; i++)
	disps[j] += (displacements[disp_indices[i]+j]*disp_factor[i]);
      disps[j] *= 1000;
    }
  }
  else
    cout << "No BAT table for " << bodySceneGraphNode << '\n';
  /* add displacements contributed by this BAT to total displacement */
  add_bat_displacements (n_total, total_disp, disps);
  
  /* release space */
  for (i = 0; i < n; i++)
    delete bv[i];
  free (bv);
  delete disps;
  delete disp_factor;
  delete disp_indices;
  delete distance;
  delete dist;
  delete p_conv;
  delete v;
}


/* deform body part using all its BAT tables at given BAP point and
 * create total displacement 
 */

void  Bat::deform_bap_using_all_bats (int n, float total_disp[], 
				      Bap_comb_t p[])

{
  if (N_bap_ids > 0)
    deform_bap_using_bat (n, total_disp, p);
  else
    cout << "No BAT table for " << bodySceneGraphNode << '\n';
  /* recursively compute displacements for other BATs of this body part */
  if (next != NULL)
    next->deform_bap_using_all_bats (n, total_disp, p);
}


/* produce deformation for given bap combination point: not all
 * dimensions are available;
 */

void Bap::deform_bap_UNUSED (Bap_comb_t p[])

{
  int    *v = new int[N_bap_combs];
  int     n = 2; 
  int     tab_length = 3*N_vertex_ids;
  float  *disps = new float[tab_length];
  float  *dist = new float[N_bap_combs];
  int     i, j, tab_no, disp_ix1, disp_ix2;
  Bap_vector_t *p_conv = new Bap_vector_t[N_bap_ids];
  float  *bv1 = new float[N_bap_ids], *bv2 = new float[N_bap_ids];
  float   mag, dp, t;

  /* get half closest bap combination key points to 'p' */
  closest_points_to_bap_point (p, n, v, dist);
  /* print point and other vectors closest */
  cout << "BAP POINT:   ";
  for (j = 0; j < N_bap_ids; j++) 
    cout << p[j] << "   ";
  cout << '\n';
  /* print closest BAP combinations */
  cout << "CLOSEST COMBINATIONS\n";
  for (i = 0; i < n; i++) {
    tab_no = v[i];   /* table no */
    for (j = 0; j < N_bap_ids; j++)
      cout << bap_comb_v[tab_no][j] << "   ";
    cout << '\n';
  }
  //cout << '\n';
  for (j = 0; j < tab_length; j++) disps[j] = 0; /* reset displacements */
  /* interpolate displacement at 'p' by considering key displacements
     at two closest BAP combimations keys */
  if (N_bap_ids > 0) {
    /* BAT table exists for this BAP */
    cout << "Deforming the body part " << bodySceneGraphNode << '\n';
    /* convert bap point p to bap-vector form */
    for (j = 0; j < N_bap_ids; j++)
      p_conv[j] = p[j];
    /* consider closest two combinations */
    /* vector from 1st closest key to bap point: bv1 */
    create_float_vector (bv1, bap_comb_v[v[0]], p_conv, N_bap_ids);
    /* vector from 1st closest key to 2nd closest key: bv2 */
    create_float_vector (bv2, bap_comb_v[v[0]], bap_comb_v[v[1]], N_bap_ids);
    mag = magnitude_vector (bv2, N_bap_ids);
    normalize_vector (bv2, N_bap_ids); /* normalize for following dot-product */
    /* magnitude of projection of bv1 onto bv2 */
    dp = dotp_vectors (bv1, bv2, N_bap_ids);
    t = dp/mag; /* projection val as proportion of dist bet bap keys */
    /* interpolate displacement vector from first key to second using t */
    disp_ix1 = v[0]*tab_length; disp_ix2 = v[1]*tab_length;
    for (j = 0; j < tab_length; j++) {
      disps[j] = displacements[disp_ix1+j] + t*(displacements[disp_ix2+j] - displacements[disp_ix1+j]);
      disps[j] *= 1000;
    }
  }
  else
    cout << "No BAT table for " << bodySceneGraphNode << '\n';
  /****** do something with displacements ********/
  generate_total_bap_displacement_UNUSED (disps);
  delete disps;
  delete dist;
  delete p_conv;
  delete bv1; delete bv2;
  delete v;
}


/* animate body parts using active bap ids 
   indicated by flags and corresponding values;
   bap-ids are counted from 1 in the input, but this program counts bap ids
   from 0; subtract 1 to get correct vals;
*/

void Bap::bap_deformation (unsigned char active_bap_flags[], 
			   Bap_comb_t bap_values[])

{
  Bap_comb_t bap_point[MAX_BAPIDS];
  int i;

  //cout << "Animating Bap\n";
  /* create bap combination from flags and values */
  for (i = 0; i < N_bap_ids; i++) {
    if (active_bap_flags[bap_ids[i]-1]) {
      /* bap id active: component modified in thos frame */
      bap_point[i] = bap_values[bap_ids[i]-1];
    }
    else
      /* take component from previous frame */
      bap_point[i] = Prv_frame_bap_points[bap_ids[i]-1];
  }
  /* deformation */
  // OLD version deform_bap (frameno, bap_point);
  /* transfer normal displacements into total (as default) */
  for (i = 0; i < 3*N_vertices; i++) 
    total_displacements[i] = normal_displacements[i];
  deform_bap_using_all_bats (3*N_vertices, total_displacements, bap_point);
}


/* standard output of bap */

void Bap::output_bap ()

{
  //cout << "Outputting " <<  bodySceneGraphNode << '\n';
}



ostream & operator<<(ostream & os, Bap & bap)

{
  bap.output_bap ();
  return os;
}


ofstream & operator<<(ofstream & ouf, Bap & bap)

{
  bap.output_bap_iv (ouf);
  return ouf;
}



