/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* The University of Rochester retains full right to use the code for */
/* their own purposes, assign or donate the code to a third party and */
/* to inhibit third parties from using the code for non MPEG-4 Visual */
/* (ISO/IEC 14496-2) conforming products.                             */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1996-1999                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         delaunay.cpp
   AUTHOR/DATE  Peter van Beek - University of Rochester
   DATE         Mon Aug 17 14:39:38 1998
   DESCRIPTION  This module contains routines for constrained Delaunay 
                triangulation of point sets with polygonal boundary. 
                The constraining boundary can be non-convex but must be
                simple (non-selfintersecting). These routines allow Delaunay 
                triangulation using the following steps:
                a. triangulate the polygonal boundary (non-Delaunay), i.e.
                   create a triangulation of the boundary points that
                   necessarily contains the boundary segments as triangle 
                   edges and contains no triangles outside the boundary;
                b. turn that triangulation into a Delaunay triangulation
                   by iterative optimization of local topology, i.e. edge
                   flipping;
                c. iteratively insert the interior mesh points one by one;
                   create triangles as necessary and apply iterative local 
                   optimization to keep the entire mesh Delaunay. 
                Thus the mesh construction algorithm is partially one-shot 
                and partially incremental. 

   SYNOPSIS     The following routines are provided by this module:
                Construct a triangular mesh by triangulating a simple polygon:
                void Tri2dTess::triangulate(Poly2d &poly);
                Flip edges in a given triangular mesh until it is Delaunay:
                int Tri2dTess::makeDelaunay ();
                Insert new point in a Delaunay mesh:
                int Tri2dTess::insertDelaunaySite(Point2d &p);

   ASSUMPTIONS  The origin is at the top-left, i.e. the x-axis points to
                the right and the y-axis points down.
                Vertices of triangles are ordered in clockwise fashion,
                and so are the pointers to adjacent triangles. Adjacent
                triangles share an edge between them. Adjacent triangle
                i is connected to the edge between vertices i and i+1.
                Input vertex coordinates have limited accuracy: here they 
                all have 0.5 accuracy. 

   NOTES        No special precautions were taken in the code implementing
                the geometric primitives to ensure robustness. This code
                is not intended to be used in a general setting. However,
                the vertices are not assumed to be in general position;
                cocircularities are handled consistently so that the same
                Delaunay triangulation is generated no matter what the
                order of point insertion is. 
                
   LITERATURE   M. de Berg, M. van Kreveld, M. Overmars and O. Schwarzkopf:
                Computational Geometry - Algorithms and Applications
                Springer, 1997

                L. Guibas and J. Stolfi: Primitives for the manipulation of
                general subdivisions and the computation of Voronoi diagrams
                ACM Transactions on Graphics, vol. 4, no. 2, pp.74-123, 1985
                
                L.J. Guibas, D.E. Knuth and M. Shahir: Randomized incremental
                construction of Delaunay and Voronoi Diagrams
                Algorithmica, vol. 7, pp.381-413, 1992

***************************************************************************-*/

#include "general.h"
#include "pointClass.h"					// Point2d
#include "polyClass.h"					// Poly2d/Tri2d
#include "triRefClass.h"				// Tri2dRef
#include "triTessClass.h"				// Tri2dTess



//////////////////////////////////////////////////////////////////////////////

// #defined in pointClass.h: EPSILON = 1.0e-5 



//////////////////////////////////////////////////////////////////////////////


// Given four points in the plane a, b, c and d, determine whether point d
// lies in the interior of the unique circle passing through points a, b 
// and c. Points a, b and c must be oriented clockwise. Returns 1 if d is 
// in the interior, -1 if d is in the exterior, 0 if d is on the circle. 
// See de Berg et al., 1997, p. 206. 
int inCircle(Point2d &pa, Point2d &pb, Point2d &pc, Point2d &pd)
{
	Point2d aDiff = pa-pd;
	Point2d bDiff = pb-pd;
	Point2d cDiff = pc-pd;

	double aLen = aDiff.normSqr();
	double bLen = bDiff.normSqr();
	double cLen = cDiff.normSqr();

	double det = 
		aLen * (bDiff^cDiff) + bLen * (cDiff^aDiff) + cLen * (aDiff^bDiff);

	return ((det > 0.0) ? 1 : ((det < 0.0) ? -1 : 0));
}

// Returns 1 if the points a, b, c are in clockwise order; 0 otherwise.
inline int cw(Point2d& pa, Point2d& pb, Point2d& pc)
{
	double det = 
		(pb.X()-pa.X())*(pc.Y()-pa.Y()) - (pb.Y()-pa.Y())*(pc.X()-pa.X());
	return (det > 0.0);
}

// Returns signed area of triangle a, b, c times two
inline double triArea(Point2d& pa, Point2d& pb, Point2d& pc)
{
	return (pb.X()-pa.X())*(pc.Y()-pa.Y()) - (pb.Y()-pa.Y())*(pc.X()-pa.X());
}

// Returns 1 if point p lies inside the triangle a, b, c, or on one
// of the edges; return 0 otherwise.
int onTriangle(Point2d &p, Point2d &a, Point2d &b, Point2d &c)
{
	if (cw(a, b, c)) {
		if ((triArea(p, a, b) >= 0.0) && (triArea(p, b, c) >= 0.0) && (triArea(p, c, a) >= 0.0))
			return 1;
	}
	else {
		if ((triArea(p, a, b) <= 0.0) && (triArea(p, b, c) <= 0.0) && (triArea(p, c, a) <= 0.0))
			return 1;
	}
	return 0;
}

// Compute perpendicular distance (squared) from p to the line a, b
inline double perpDistSqr(Point2d &p, Point2d &a, Point2d &b)
{
  Point2d line = a - b;
  Point2d point = p - b;
  return (point - ((point*line)/line.normSqr()) * line).normSqr();
}

// Find the vertex with minimum X value; in case of ties choose vertex with
// minimum Y value. Returns vertex index. Assumes there is at least 1 
// vertex in the polygon. 
int findMinVertex(Poly2d &poly)
{
	const int size = poly.getSize();
	int index = 0;
	double minX = poly[0].X();
	double minY = poly[0].Y();

	for (int i = 1; i < size; i++) {
		double valX = poly[i].X();
		if ((valX < minX) || (valX == minX && poly[i].Y() < minY)) {
			index = i;
			minX = valX;
			minY = poly[i].Y();
		}
	}
	return index;
}

// Find the vertex of a polygon inside or on the triangle given by vertices 
// n-1,n,n+1 that is farthest from the diagonal given by vertices n-1, n+1;
// if there is no such vertex, return -1; otherwise return the vertex index.
// Assumes there are at least 4 vertices in the polygon.
int findFarthestVertex (Poly2d &poly, int n)
{
	const int size = poly.getSize();
	int index = -1;
	double maxDist = -1.0;
	Point2d &a = poly[Mod(n-1, size)];
	Point2d &b = poly[Mod(n, size)];
	Point2d &c = poly[Mod(n+1, size)];

	for (int i = 0; i < size-3; i++) {
		Point2d &p = poly[Mod(n+2+i, size)];
		if (onTriangle(p, a, b, c)) {
			double dist = perpDistSqr(p, a, c);
			if (dist >= maxDist) {
				maxDist = dist;
				index = Mod(n+2+i, size);
			}
		}
	}
	return index;
}

// Create a mesh by triangulating a simple polygon using a simple 
// recursive O(n^2) algorithm; From de Berg et al., 1997, p. 47. 
void ptriangulate(Tri2dTess &mesh, Poly2d &poly)
{
	const int size = poly.getSize();
	if (size < 3) {
		doError("triangulate: invalid input polygon...\n");
		return;
	}
	else if (size == 3) {
		// if polygon is a triangle we are done
		Tri2d tri(poly[0], poly[1], poly[2]);
		mesh.append(tri);
	}
	else {
		// find polygon vertex with minimum x-coordinate (min y when tied)
		int n = findMinVertex(poly);
		// find vertex inside n-1,n,n+1 farthest from the diagonal n-1,n+1
		int m = findFarthestVertex(poly, n);
		if (m == -1) {
			// if no such vertex, then n-1,n+1 is a valid diagonal
			Tri2d tri(poly[n-1], poly[n], poly[n+1]);
			mesh.append(tri);
			// triangulate the other part of the polygon
			Poly2d newpoly(poly, n+1, n-1);
			ptriangulate(mesh, newpoly);
		}
		else {
			// cut the polygon through diagonal n,m
			Poly2d firstpoly(poly, n, m);
			Poly2d secondpoly(poly, m, n);
			// triangulate both parts of the polygon
			ptriangulate(mesh, firstpoly);
			ptriangulate(mesh, secondpoly);
		}
	}
}

// Create a mesh by triangulating a simple polygon using the
// above algorithm. First destroy old mesh if there was one.
void Tri2dTess::triangulate(Poly2d &poly)
{
	// First, destruct the old tesselation if there is one
	destroy();
	// Call recursive triangulation routine
	ptriangulate(*this, poly);
}


// Return 1 if pa has smaller y-value than pb, using x-value to break ties;
// otherwise return 0
inline int minPoint(Point2d &pa, Point2d &pb)
{
	return ((pa.Y() < pb.Y()) || (pa.Y() == pb.Y() && pa.X() < pb.X()));
}

// Return 1 if pa has greater x-value than pb, using y-value to break ties;
// otherwise return 0
inline int maxPoint(Point2d &pa, Point2d &pb)
{
	return ((pa.X() > pb.X()) || (pa.X() == pb.X() && pa.Y() > pb.Y()));
}

// Determine if the edge 'e1' between triangle 'tri1' and 'tri2' is 
// locally a valid Delaunay edge. It is locally valid if the point
// opposite this edge on 'tri2' passes the in-circle test, i.e. if it
// lies outside the circumcircle of 'tri1'. If that point is inside,
// the edge is invalid. E.g. if point D below is outside the circumcircle
// of A,B,C, then e1 (== e2 of tri2) is a Delaunay edge. 
//		   A x-------x D
//			/ \		/  
//		t1 / e1\e2 / t2
//		  /		\ /	   
//	   C x-------x B   
// A special case occurs if the point is exactly on the circumcircle.
// We need a rule to break this kind of tie consistently and produce
// a canonical Delaunay triangulation. The rule is that the point with
// maximum x-value is treated as if it were perturbed to lie just outside
// the circle. It follows that it must lie opposite the suspect edge.
// E.g. if point D above is exactly on the circumcircle of A,B,C, we treat
// the point with maximum x-value (also D) as if it is symbolically 
// perturbed to be outside, so e1/e2 is considered Delaunay. However, 
// if B would have been the point with maximum x-value, the same edge 
// would not be considered Delaunay.
int Tri2dTess::delaunayEdge(Tri2dRef *tri1, int e1, Tri2dRef *tri2, int e2)
{
	// get the data
//	Tri2dRef *tri2 = tri1->neighbor[e1];
//	int e2 = tri1->nedge[e1];
	Point2d pa = tri1->corner[e1]->getPoint();
	Point2d pb = tri1->corner[(e1+1)%3]->getPoint();
	Point2d pc = tri1->corner[(e1+2)%3]->getPoint();
	Point2d pd = tri2->corner[(e2+2)%3]->getPoint();
	
	// do in-circle test
	int valid = inCircle (pa, pb, pc, pd);
	switch (valid) {
	  case 1:				// inside the circle: invalid
		return 0;
		break;
	  case -1:				// outside the circle: valid
		return 1;
		break;
	  case 0:				// on circle: special case
		// test if either of the opposite points has minimum y-value
		if ((maxPoint(pc, pa) && maxPoint(pc, pb) && maxPoint(pc, pd)) ||
			(maxPoint(pd, pa) && maxPoint(pd, pb) && maxPoint(pd, pc)))
			return 1;
		else
			return 0;
		break;
	  default:
		return 0;
	}
}


// Flips the common edge between two triangles t1 and t2; this edge
// is edge number 'e1' in t1 and edge number 'e2' in t2.
//			A		  D			A		  D	  
//			 x-------x			 x-------x
//			/ \		/			  \		/ \
//		t1 / e1\e2 / t2	   =>	t1 \   /   \ t2
//		  /		\ /					\ /		\
//		 x-------x					 x-------x
//		C		 B					C		  B
//
void Tri2dTess::flipEdge(Tri2dRef *t1, Tri2dRef *t2, int e1, int e2)
{
	int e1next = ((e1 == 0)? 1 : ((e1 == 1)? 2 : 0));
	int e2next = ((e2 == 0)? 1 : ((e2 == 1)? 2 : 0));

	// swap pointers to vertices
	t1->corner[e1next] = t2->corner[((e2 == 0)? 2 : ((e2 == 1)? 0 : 1))];
	t2->corner[e2next] = t1->corner[((e1 == 0)? 2 : ((e1 == 1)? 0 : 1))];

	// swap pointers to neighboring triangles
	Tri2dRef *tmpRef = t2->neighbor[e2next];
	int tmpEdge = t2->nedge[e2next];
	t1->connect(tmpRef, e1, tmpEdge);

	tmpRef = t1->neighbor[e1next];
	tmpEdge = t1->nedge[e1next];
	t2->connect(tmpRef, e2, tmpEdge);

	// swap pointers to each other
	t1->neighbor[e1next] = t2;
	t1->nedge[e1next] = e2next;
	t2->neighbor[e2next] = t1;
	t2->nedge[e2next] = e1next;
}


// Given a triangular mesh, iteratively flip two opposite triangles
// that are not locally Delaunay until no such opposite triangles can be
// found. Note that this is an O(n^3) algorithm: the mesh is scanned each
// iteration for an invalid edge and the number of flips required could
// be O(n^2) in theory (although mostly lower in practice). Therefore,
// this routine is not practical for meshes with a significant number
// of triangles. 
int Tri2dTess::makeDelaunay ()
{
	if (getSize() < 1 || getSizePoints() < 3) {
		doError("makeDelaunay: invalid input mesh, aborting...\n");
		return 0;
	}
	int maxIter = getSizePoints()*getSizePoints()+1;
	int countFlips = 0;

	// Limit the number of possible iterations
	while (--maxIter) {
		int flip = 0;
		// unmark all triangles
		markRefs(0);
		// scan the mesh to find an invalid Delaunay edge
		Tri2dRef *tri1 = nextTriangle(1);
		for (int i=0; i<sizeTess; i++) {
			// check each edge of this triangle
			for (int e1 = 0; e1 < 3; e1++) {
				Tri2dRef *tri2 = tri1->neighbor[e1];
				if (tri2 != NULL && tri2->mark == 0) {
					int e2 = tri1->nedge[e1];
					if (delaunayEdge(tri1, e1, tri2, e2) == 0) {
						// this edge is invalid, so flip it
						flipEdge(tri1, tri2, e1, e2);
						flip = 1;
						countFlips++;
						break;
					}
				}
			}
			if (flip)
				// break from search loop and start over
				break;
			else {
				// mark this triangle (all edges were checked) and search on
				tri1->mark = 1;
				tri1 = nextTriangle();
			}
		}
		// stop if no flips occurred; otherwise continue
		if (flip == 0)
			break;
	}
//	cout << "makeDelaunay: flipped " << countFlips << " edges" << endl;

	if (maxIter == 0) {
		doError("makeDelaunay: too many iterations, aborting...\n");
		return 0;
	}
	return 1;
}


// Locate a site 'p' in the triangulation using brute force method
int Tri2dTess::locatePoint(Point2d &p, Tri2dRef *&tri, int &edge, int &onEdge)
{
	// scan the mesh to find a triangle containing p
	tri = nextTriangle(1);
	if (tri == NULL) {
		doError("locatePoint: cannot locate point in empty mesh!\n");
		return 0;
	}
	for (int i=0; i<sizeTess; i++) {
		Point2d a = tri->corner[0]->getPoint();
		Point2d b = tri->corner[1]->getPoint();
		Point2d c = tri->corner[2]->getPoint();
		if (onTriangle(p, a, b, c)) {
			// p is inside triangle a, b, c or on one of its edges!
			// now find edge closest to p
			double d[3];
			d[0] = perpDistSqr(p, a, b);
			d[1] = perpDistSqr(p, b, c);
			d[2] = perpDistSqr(p, c, a);
			edge = (d[0]<d[1] ? (d[0]<d[2] ? 0 : 2) : (d[1]<d[2] ? 1 : 2));
			// check if p is on this edge
			onEdge = (d[edge] < EPSILON*EPSILON);
			return 1;
		}
		tri = nextTriangle();
	}
	// error: could not find the triangle
	doError("locatePoint: failed! - point not located in mesh?\n");
	return 0;
}


// Insert a new site 'p' into a Delaunay triangulation and update the
// triangulation so it remains Delaunay everywhere. First locate the
// triangle containing p and check if p isn't already in the point list.
// Add new triangles having p as a vertex. Then, iteratively flip edges
// between triangles around p that are not Delaunay. 
// See de Berg et al., 1997 and Guibas et al., 1985 and 1992 for 
// descriptions of this basic incremental construction algorithm.
int Tri2dTess::insertDelaunaySite(Point2d &p)
{
	Tri2dRef *tri = tess;				// pointers to triangle
	Tri2dRef *triOp, *tStart;
	Tri2dRef *new1, *new2;				// new triangles
	Pt2dLink *pLink = NULL;
	int e0, e1, e2, eOp0, eOp2, edge;	// edge numbers
	int onEdge=0;

	// locate the triangle (tri) that this new site (p) is in or on;
	// return the edge (e0) closest to p; and check if it is on that edge.
	if (this->locatePoint(p, tri, e0, onEdge) == 0) {
		doError("insertDelaunaySite: failed!\n");
		return 1;
	}

	// add this site to the point list - if not already there
	if (points.grow(p, pLink) == 0) {
		doError("insertDelaunaySite: new site already existed...\n");
		return 1;
	}

	// now we know the site didn't already exist, so it must lie in 
	// the interior of a single triangle or in the interior of an edge 
	// shared by two triangles
	e1 = ((e0==0)? 1: ((e0==1)? 2: 0));
	e2 = ((e0==0)? 2: ((e0==1)? 0: 1));
	
	if (onEdge) {
		// the new site is on the edge of an existing triangle

		// check if this triangle edge is a mesh boundary edge
		triOp = tri->neighbor[e0];
		if (triOp == NULL) {
			// we do not allow this in our application!
			doError("insertDelaunaySite: cannot insert site on mesh boundary...\n");
			return 1;
		}
		// so there is a triangle 'triOp' opposite to triangle 'tri';
		// replace the two triangles (tri and triOp) that share this edge 
		// by four triangles surrounding the new site 
		eOp0 = tri->nedge[e0];
		eOp2 = ((eOp0==0)? 2: ((eOp0==1)? 0: 1));

		// create two new triangles 
		new1 = new Tri2dRef();
		new2 = new Tri2dRef();

		// set all the vertex pointers right ...
		new1->corner[0] = pLink;
		new1->corner[1] = tri->corner[e1];
		new1->corner[2] = tri->corner[e2];
		new2->corner[0] = triOp->corner[eOp0];
		new2->corner[1] = pLink;
		new2->corner[2] = triOp->corner[eOp2];
		tri->corner[e1] = pLink;
		triOp->corner[eOp0] = pLink;
		
		// set all the neighbor pointers right (in the right order!)
		new1->connect(new2, 0, 0);
		new1->connect(tri->neighbor[e1], 1, tri->nedge[e1]);
		new2->connect(triOp->neighbor[eOp2], 2, triOp->nedge[eOp2]);
		new1->connect(tri, 2, e1);
		new2->connect(triOp, 1, eOp2);

		tStart = tri;
		edge = e2;				// first suspect edge
	}
	else {
		// the new site is strictly inside an existing triangle;
		// replace this triangle by three triangles around the new site

		// create two new triangles 
		new1 = new Tri2dRef();
		new2 = new Tri2dRef();

		// set all the vertex pointers right ...
		new1->corner[0] = pLink;
		new1->corner[1] = tri->corner[e1];
		new1->corner[2] = tri->corner[e2];
		new2->corner[0] = tri->corner[e0];
		new2->corner[1] = pLink;
		new2->corner[2] = tri->corner[e2];
		tri->corner[e2] = pLink;

		// set all the neighbor pointers right (in the right order!)
		new1->connect(new2, 2, 1);
		new1->connect(tri->neighbor[e1], 1, tri->nedge[e1]);
		new2->connect(tri->neighbor[e2], 2, tri->nedge[e2]);
		new1->connect(tri, 0, e1);
		new2->connect(tri, 0, e2);
			
		tStart = tri;
		edge = e0;				// first suspect edge
	}
	// add the two new triangles to the mesh triangle list
	new1->next = new2;
	new2->next = tess;
	new1->prev = tess->prev;
	new2->prev = new1;
	tess->prev->next = new1;
	tess->prev = new2;
	sizeTess+=2;

	// go clockwise around the new site iteratively; inspect triangle 
	// edges opposite the new site; flip edges if necessary, until all 
	// of them are locally Delaunay
	tri = tStart;
	int progress;
	do {
		progress = 1;
		triOp = tri->neighbor[edge];
		if (triOp != NULL) {
			eOp0 = tri->nedge[edge];
			if (delaunayEdge(tri, edge, triOp, eOp0) == 0) {
				// this edge is invalid, so flip it clockwise
				flipEdge(tri, triOp, edge, eOp0);
				// note that 'tri' and 'edge' are still the triangle and
				// edge to be inspected next
				progress = 0;
			}
		}
		if (progress) {
			// go to next clockwise triangle and next suspect edge
			edge = ((edge==0)? 1: ((edge==1)? 2: 0));
			eOp0 = tri->nedge[edge];
			tri = tri->neighbor[edge];
			edge = ((eOp0==0)? 1: ((eOp0==1)? 2: 0));
		}

	} while (tri != tStart || progress == 0);

	// all done; the mesh is Delaunay again!
	return 0;
}


// end of file

