/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* The University of Rochester retains full right to use the code for */
/* their own purposes, assign or donate the code to a third party and */
/* to inhibit third parties from using the code for non MPEG-4 Visual */
/* (ISO/IEC 14496-2) conforming products.                             */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1997-1999                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         mesh_decode.cpp
   AUTHOR/DATE  Peter van Beek, Wed Mar  3 17:19:07 1999

   DESCRIPTION  ISO/IEC 14496 (MPEG-4) 2D Mesh object decoding.

                In general, the decoding program parses a bitstream and 
                generates a sequence of 2-D triangular meshes. Each mesh 
                describes the geometry of a 2-D mesh at a certain point in 
                time. This mesh data may be used by other systems to
                render animated video object planes. 
                Each mesh consists of a number of triangles, and the 
                vertices of the triangles are referred to as the node 
                points of the mesh. The node points are animated by the
                motion vectors coded in the bitstream. Note that all the 
                nodes in the mesh are present from the first to the last 
                time instant (i.e. no nodes are ever added or deleted). 

                Some overhead is present around the core syntax. These 
                are "mesh object" start codes, "mesh object" end codes, 
                and "mesh object plane" start codes.

                This code can be used both integrated into the MPEG-4
                Systems player or as a stand-alone Visual Object decoder. 

                In the latter case, each mesh object plane is saved in a 
                data-file having the extension ".tri" (for triangulation). 
                The place where the mesh is to be placed for composition
                is given as arguments on the command-line, in pixels wrt
                the origin of the output space (assumed at top-left). 
                In the former case, the Systems should use the MeshDecoder
                class API to access the mesh data. 
                
   NOTES        Note that in this code, one assumes a coordinate system
                with origin at the top-left, such that the x-axis points
                to the right and the y-axis points down. Definitions of
                above/below, clockwise/counterclockwise etc. all depend
                on this assumption.

                This code assumes that all mesh node positions and motion 
                vectors have 0.5 pixel accuracy with respect to the image/
                texture map that will be used for rendering. Multiplying 
                these positions and motion vectors by 2.0 transforms them 
                to integer accuracy, after which the integer values are coded
                losslessly. 

                The code assumes that 'int'-types are at least 32 bits, 
                so that start-codes fit into an 'int'.

*********************************************************************-*/

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <math.h>

#if defined(WIN32)
#include <io.h>
#include <time.h>
#include <sys/timeb.h>
#include <float.h>
#define MAXDOUBLE DBL_MAX
#else
#include <values.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

// define to get elaborate debug messages
//#define DEBUGGING

#include "general.h"			// various definitions
#include "linkClass.h"			// Pt2dLink
#include "pointClass.h"			// Point2d
#include "polyClass.h"			// Poly2d
#include "triRefClass.h"		// Tri2dRef
#include "triTessClass.h"		// Tri2dTess
#include "queueClass.h"			// TriangleQueue

#include "basic.h"				// definitions
#include "entropy.h"			// entropy coding
#include "huffman.h"			// huffman coding 
#include "bitstrm.h"			// bitstream functions

#include "mesh_decode.h"		// MeshDecoder



////////// COMPILER SWITCHES /////////////////////////////////////////////////

// define VISUAL_DECODER to run the decoder on a stand-alone bitstream
// conforming to the Visual part of the MPEG-4 standard;
// do not define in case of integration into "Systems" compositor/player
#if defined(VISUAL_DECODER)
// Define TIMER to get running time message
#define TIMER
#endif


// define DPRINTING to get some basic useful messages
#if defined(DPRINTING) || defined(_DEBUG)
// Define printing macro which allows variable number of arguments
#define DPRINT(vargs)				((void) printf vargs)
#else
#define DPRINT(vargs)
#endif



////////// CONSTANTS /////////////////////////////////////////////////////////

// Mesh Object (MO) overhead
#define SC_PREFIX					1			// start-code prefix
#define NUMBITS_SC_PREFIX			24
#define MO_START_CODE				188			// 0xbc
#define MO_END_CODE					177			// 0xb1 (visual object endcode)
#define NUMBITS_MO_START			8
#define NUMBITS_MO_END				8

// Mesh Object Plane (MOP) overhead
#define MOP_START_CODE				189			// 0xbd
#define NUMBITS_MOP_START			8

// Geometry and motion coding
#define NUMBITS_MESH_TYPE			2
#define NUMBITS_NUMMESHNODES		16
#define NUMBITS_NUMBOUNNODES		10
#define NUMBITS_FIRSTNODEPOS		12
#define NUMBITS_NUMNODE_UNIF		10
#define NUMBITS_RECTSIZE_UNIF		8
#define NUMBITS_SPLIT_UNIF			2
#define NUMBITS_MOP_FCODE			3

// Maximum length of coordinate difference codeword
#define MAX_COORD_DIFF_CODE_LEN		14

// Node location and motion vector accuracy
#define COORD_FACTOR				2.0
#define MOTVEC_FACTOR				2.0

// Default frame-rate of Mesh Object
#define DEF_FRAME_RATE				10



////////// CLASS MEMBER FUNCTIONS ////////////////////////////////////////////

// constructor
MeshDecoder::MeshDecoder()
{
	m_startStrm = m_endStrm = 0;
	m_xmin = m_xmax = m_ymin = m_ymax = 0.0;
	m_fCode = m_scaleFactor = 1;
	m_motionRange = 32;
	// creates input stream not attached to any input buffer
	m_bitStrmIn	   = new CInBitStream ();
	// attach input stream to entropy decoders
	m_entrDecSet   = new CEntropyDecoderSet (*m_bitStrmIn);
	m_initTri = NULL;
	m_initEdge = -1;
}


// constructor with given input buffer and its size
MeshDecoder::MeshDecoder(BYTE *pInBuffer, const int inSize)
{
	m_startStrm = m_endStrm = 0;
	m_xmin = m_xmax = m_ymin = m_ymax = 0.0;
	m_fCode = m_scaleFactor = 1;
	m_motionRange = 32;
	// creates input stream attached to input buffer
	m_bitStrmIn	   = new CInBitStream (pInBuffer, inSize);
	// attach input stream to entropy decoders
	m_entrDecSet   = new CEntropyDecoderSet (*m_bitStrmIn);
	m_initTri = NULL;
	m_initEdge = -1;
}


// (re-)attach a given input stream-buffer
void MeshDecoder::attachStream(BYTE *pInBuffer, const int inSize, const int bitPos)
{
	// check input
	if (pInBuffer == NULL || inSize <= 0)
		doError("MeshDecoder: invalid input stream buffer\n");

	// (re-)attach input buffer to stream
	m_bitStrmIn->attach (pInBuffer, inSize);
	// read some bits away if necessary
	if (bitPos > 0)
		m_bitStrmIn->getBits(bitPos);
}


// destructor
MeshDecoder::~MeshDecoder()
{
	delete m_entrDecSet;
	delete m_bitStrmIn;
}


// return x- and y-coordinates of vertex i; return -1 if error
int MeshDecoder::getVertexCoordinates(int i, double &x, double &y)
{
	if (i < 0 || i >= meshCurr.getSizePoints()) {
		doError("MeshDecoder: invalid vertex index\n");
		return 1;
	}
	else {
		Point2d vert = meshCurr.point(i);
		x = vert.X();
		y = vert.Y();
		return 0;
	}
}

// return vertex indices of triangle i; return -1 if error
int MeshDecoder::getTriangleIndices(int i, int &v0, int &v1, int &v2)
{
	if (i < 0 || i >= meshCurr.getSize()) {
		doError("MeshDecoder: invalid triangle index\n");
		return 1;
	}
	else {
		Tri2dRef* pTri = meshCurr.nextTriangle(i == 0);
		v0 = meshCurr.where(pTri->getLink(0));
		v1 = meshCurr.where(pTri->getLink(1));
		v2 = meshCurr.where(pTri->getLink(2));
		if (v0 == -1  ||  v1 == -1  ||  v2 == -1) {
			doError("MeshDecoder: cannot find triangle vertex\n");
			return 1;
		}
		else
			return 0;
	}
}

// return pointer to current location in stream
BYTE *MeshDecoder::getStreamPtr()	
{
	return m_bitStrmIn->getStreamPtr(); 
}

// return current number of bits in buffer
uint MeshDecoder::getBitsLeft()
{ 
	return m_bitStrmIn->getNumBitsInBuf(); 
}


/*+==========================================================================
	NAME		MeshDecoder::decodeMesh
	VERSION		2.0
	PURPOSE		Decode one Mesh Object Plane (MOP) from a mesh stream.
	METHOD		Generally, either an I-MOP is decoded or a P-MOP. An
				I-MOP is an MOP with the initial geometry of a mesh; a
				P-MOP is an MOP with motion of a mesh with respect to the
				previous instance.
	IN			isIntra		  1: Intra plane (I-MOP); 0: Inter plane (P-MOP)
	RETURNS		0: OK
				1: End of stream detected or error detected
	NOTES		This routine depends on the VISUAL_DECODER compiler condition.
				If defined (standalone Visual stream decoder), the Mesh Object
				Plane header is read here, including the is_intra bit and
				any timing information. In this case, the isIntra parameter
				is ignored.
				If VISUAL_DECODER is not defined (code is integrated into
				Systems player), the Mesh Object Plane header is read by the 
				Animation Stream decoder routines. In this case, the isIntra
				parameter is an input parameter.
==========================================================================-*/
int MeshDecoder::decodeMesh(int &isIntra)
{
	uint meshMask = 1;					// mesh mask flag
	int	 useStartCode = 1;				// flags use of start code 
	double timeStamp = 0.0;				// timestamp of mesh object plane
	int code;							// a return code or bitstream code

	// check to see if mesh object has already ended
	if (m_endStrm == 1) {
		doWarning("MeshDecoder: stream has ended\n");
		return 1;
	}

	// check to see if mesh object has already started
	if (m_startStrm == 0) {
#if defined(VISUAL_DECODER)
		// decode mesh object start-code
		code = m_bitStrmIn->getBits(NUMBITS_SC_PREFIX + NUMBITS_MO_START);
		if (code != ((SC_PREFIX << NUMBITS_MO_START) + MO_START_CODE)) {
			doError("MeshDecoder: no mesh-object-start-code\n");
			m_endStrm = 1;
			return 1;
		}
#endif
		m_startStrm = 1;
	}

#if defined(VISUAL_DECODER)
	// decode mesh object plane header 
	if (decodeMopHeader(isIntra, meshMask, timeStamp, useStartCode)) {
		m_endStrm = 1;
		return 1;
	}
#endif

	// decode mesh object plane data
	DPRINT(("------- Decoding mesh %c-plane --------------------\n", 
			(isIntra? 'I': 'P')));

	if (meshMask) {
		if (isIntra) {
			// decode the mesh type code; this flag is '01' or '10'
			code = (int) m_bitStrmIn->getBits(NUMBITS_MESH_TYPE);
			if (code == 1)
				// decode uniform mesh geometry 
				decUniformMesh();
			else if (code == 2)
				// decode object-based (Delaunay) mesh geometry 
				decDelaunayMesh();
			else {
				doError("MeshDecoder: invalid mesh type code\n");
				m_endStrm = 1;
				return 1;
			}
			// compute mesh bounding box attributes
			setBBox();
			// find the initial triangle and edge for motion decoding
			setTopLeftTriangle();
		}
		else {
			// decode motion from previous mesh to current mesh
			decMeshMotion();
		}
	}
	else {
		DBUG((">> meshMask == 0!\n"));
	}

	// go to the next byte-aligned position in the bitstream
	if (useStartCode)
		if (nextByteAlign()) {
			m_endStrm = 1;
			return 1;
		}

#if defined(VISUAL_DECODER)
	// print presentation time stamp
	DPRINT(("\tTime stamp = %f msec\n", timeStamp));
#endif

	// return OK
	return 0;

} // decodeMesh



/*+==========================================================================
	NAME		MeshDecoder::decodeMopHeader 
	VERSION		1.1, Peter van Beek, Wed Apr 28 14:56:53 1999
	PURPOSE		Decode Mesh Object Plane header and compute time stamp
	METHOD		Start-codes are optional, so perform a check first.
				Then decode other header bits.
	INPUT		none
	OUTPUT		isIntra		  1: Intra plane; 0: Inter plane
				meshMask	  1: Normal; 0: Skip this frame
				timeStamp	  Plane timestamp in milliseconds
				useStartCode  1: Detected use of MOP start code
	RETURNS		0: OK;	1: NOT_OK - stream ended
	ERROR HANDLING
	NOTES		
==========================================================================-*/
int MeshDecoder::decodeMopHeader(int &isIntra, uint &meshMask, double &timeStamp, int &useStartCode)
{
	static uint frameRateBase = DEF_FRAME_RATE;
	static uint frameRateFrac = 0;
	static uint frameRateOffset = 0;
	static double nextTimeStamp = 0.0;
	uint isTimeCode = 0, timeHours = 0, timeMinutes = 0, timeSeconds = 0;
	uint frameSkip = 0;


	// peek at the next 24 bits and check if it is a start-code-prefix
	int code = m_bitStrmIn->peekBits (NUMBITS_SC_PREFIX);
	if (code == EOF) {
		doError("MeshDecoder: MopHeader: end of stream ...\n");
		return 1;
	}
	else if (code == SC_PREFIX) {
		// decode full start/end code 
		code = m_bitStrmIn->getBits(NUMBITS_SC_PREFIX + NUMBITS_MOP_START);
		// if not a "mesh_object_plane_start_code", stream has ended
		if (code != ((SC_PREFIX << NUMBITS_MOP_START) + MOP_START_CODE)) {
			DPRINT(("\tMeshDecoder: end of stream detected ...\n"));
			return 1;
		}
		else
			useStartCode = 1;
	}
	else
		useStartCode = 0;

	DBUG((">> useStartCode = %d\n", useStartCode));

	// decode the "is_intra" bit and the "mesh_mask" bit
	isIntra	 = m_bitStrmIn->getBits(1);
	meshMask = m_bitStrmIn->getBits(1);

	if (isIntra) {
		// decode "is_frame_rate" bit and frame rate data
		code = (int) m_bitStrmIn->getBits(1);
		if (code) {
			frameRateBase = m_bitStrmIn->getBits(8);
			frameRateFrac = m_bitStrmIn->getBits(4);
			frameRateOffset = m_bitStrmIn->getBits(1);
			if (frameRateBase == 0 && frameRateFrac == 0) {
				doError("MeshDecoder: invalid frame rate ...\n");
				return 1;
			}
		}
		// decode "is_time_code" bit and time code data
		isTimeCode = m_bitStrmIn->getBits(1);
		if (isTimeCode) {
			timeHours = m_bitStrmIn->getBits(5);
			timeMinutes = m_bitStrmIn->getBits(6);
			decodeMarkerBit();
			timeSeconds = m_bitStrmIn->getBits(6);
			if (timeHours < 0 || timeMinutes < 0 || timeSeconds < 0 ||
				timeHours > 23 || timeMinutes > 59 || timeSeconds > 59) {
				doError("MeshDecoder: invalid time code ...\n");
				return 1;
			}
		}
	}
	// decode "skip_frames" bit and skip frame data
	code = (int) m_bitStrmIn->getBits(1);
	if (code) {
		do {
			code = m_bitStrmIn->getBits(4);
			frameSkip += code;
		} while (code == 15);
	}

	// compute presentation time stamp in milliseconds
	if (isTimeCode)
		timeStamp = (3600*timeHours + 60*timeMinutes + timeSeconds)*1000;
	else
		timeStamp = nextTimeStamp;

	// compute next presentation time stamp in milliseconds
	nextTimeStamp = timeStamp + 
		(double) ((frameSkip+1)*((1000 + frameRateOffset) << 4)) /
		(double) ((frameRateBase << 4) + frameRateFrac);

	return 0;

} // decodeMopHeader


/*+==========================================================================
	NAME		MeshDecoder::decDelaunayMesh
	VERSION		4.1
	AUTHOR		Peter van Beek, Wed Apr 28 16:42:12 1999
	PURPOSE		Decode initial geometry and topology of an object-based
				mesh with Delaunay topology.
	METHOD		First decode the number of nodes and boundary nodes. Then 
				decode the boundary nodes and keep them in a polygon. Apply
				Delaunay triangulation to this polygon to create a mesh.
				Then, decode all the interior node points and add them to
				the mesh one by one, each time updating the triangulation to
				keep it Delaunay. So the mesh is created incrementally.
	PRE			meshCurr		empty mesh
				m_bitStrmIn		pointer to input stream
				m_entrDecSet	pointer to entropy decoders
	POST		meshCurr		decoded mesh
	RETURNS		none
	NOTES		Assumes there are no two nodes with the same coordinates.
==========================================================================-*/
void MeshDecoder::decDelaunayMesh()
{
	int n;

	// decode the number of nodes in the mesh and a marker bit
	int numMeshNodes = (int) m_bitStrmIn->getBits(NUMBITS_NUMMESHNODES);
	decodeMarkerBit();
	// decode the number of nodes on the boundary and a marker bit
	int numBounNodes = (int) m_bitStrmIn->getBits(NUMBITS_NUMBOUNNODES);
	decodeMarkerBit();
	if (numMeshNodes < 3  ||  numBounNodes < 3) {
		doError("MeshDecoder: decoded number of (boundary) nodes invalid\n");
		return;
	}

	// decode the first (boundary) node point coordinates
	Point2d pt = decodeFirstNode();

	// initialize boundary polygon with first node point
	Poly2d boundary(numBounNodes);
	boundary[0] = pt;
	// decode other boundary node points 
	for (n = 1; n < numBounNodes; n++) {
		// decode coordinates of next node and store it in boundary polygon
		pt += decodeNextNode();
		boundary[n] = pt;
	}
	// create constrained Delaunay triangulation of boundary polygon
	meshCurr.triangulate(boundary);
	meshCurr.makeDelaunay();

	// decode interior node points and insert them into the triangulation
	// one by one; constructing the Delaunay triangulation incrementally
	for (n = numBounNodes; n < numMeshNodes; n++) {
		// decode coordinates of next node 
		pt += decodeNextNode();
		// insert the new node into Delaunay triangulation
		meshCurr.insertDelaunaySite(pt);
	}

} // decDelaunayMesh



/*+==========================================================================
	NAME		MeshDecoder::decUniformMesh
	VERSION		3.1
	AUTHOR		Peter van Beek, Wed Apr 28 16:42:44 1999
	PURPOSE		Decode initial geometry and topology of a uniform mesh.
	METHOD		A uniform mesh is coded by only 5 parameters. Based on the
				parameters, one can construct the mesh simply by enumerating
				all triangles. 
	PRE			meshCurr		empty mesh
				m_bitStrmIn		pointer to input stream
	POST		meshCurr		decoded mesh
	RETURNS		none
	NOTES		
==========================================================================-*/
void MeshDecoder::decUniformMesh()
{
	int	   code;						// value decoded from bitstream
	int	   nhori, nverti;				// number of hor/vert nodes 
	double xsize, ysize;				// rectangle sizes

	// decode the number of nodes in the mesh (horizontal/vertical)
	nhori = (int) m_bitStrmIn->getBits(NUMBITS_NUMNODE_UNIF);
	nverti = (int) m_bitStrmIn->getBits(NUMBITS_NUMNODE_UNIF);
	if (nhori < 2  ||  nverti < 2) {
		doError("MeshDecoder: number of nodes uniform mesh invalid\n");
		return;
	}
	decodeMarkerBit();

	// decode the sizes of the rectangles in the mesh
	code = (int) m_bitStrmIn->getBits(NUMBITS_RECTSIZE_UNIF);
	xsize = ((double) code) / COORD_FACTOR;
	code = (int) m_bitStrmIn->getBits(NUMBITS_RECTSIZE_UNIF);
	ysize = ((double) code) / COORD_FACTOR;
	if (xsize <= 0.0 || ysize <= 0.0) {
		doError("MeshDecoder: size values uniform mesh invalid\n");
		return;
	}
	// decode the triangle split code
	code = (int) m_bitStrmIn->getBits(NUMBITS_SPLIT_UNIF);

	// create the mesh from top-left to bottom-right, rectangle by rectangle;
	// append two triangles for each rectangle; the orientation of these
	// triangles is determined by the split code.
	meshCurr.makeUniform(nhori, nverti, xsize, ysize, code);

} // decUniformMesh



/*+==========================================================================
	NAME		MeshDecoder::decMeshMotion
	VERSION		5.1
	AUTHOR		Peter van Beek, Wed Apr 28 16:46:52 1999
	PURPOSE		Decoding of motion vectors of nodes in a triangular mesh. 
				Thereby computing the state of the mesh at the current time.
				A motion vector is a vector between a node in a mesh and 
				the corresponding node in another mesh.
	METHOD		The motion vectors are encoded into the bitstream in a
				certain order, defined by a breadth-first traversal of 
				the triangles of a mesh (see encoder). This order is
				reconstructed by the decoder given the previous mesh, and
				each decoded motion vector is added to a node of the
				triangle being visited according to that order.
				Breadth-first traversal of triangles uses a fifo queue of 
				triangles. Visited triangles and nodes are marked, such
				that each node is processed exactly once.
	PRE			meshCurr		Mesh in the previous state
				m_bitStrmIn		pointer to input stream
				m_entrDecSet	pointer to entropy decoders
	POST		meshCurr		Mesh in the current state (+ motion vectors)
	RETURNS		none
	NOTES		Meshes are assumed to triangular and the two meshes
				are assumed to be in corresponcence, i.e. nodes/triangles 
				with same index correspond.
				Triangle vertices are assumed to be ordered clockwise, as
				are the pointers to their neibhoring triangles.
==========================================================================-*/
void MeshDecoder::decMeshMotion()
{
	Tri2dRef *currTri, *leftTri, *rightTri; // triangle pointers
	int edge, rightEdge, leftEdge;			// edge numbers
	Pt2dLink  *pb1, *pb2, *paa;				// node pointers
	Point2d mvPred;							// prediction motion vector
	int code;								// read from bitstream


	// decode motion range parameters
	if (decodeMotionRange()) return;

	// initialize node and polygon markers
	meshCurr.markLinks(0);
	meshCurr.markRefs(0);

	// get the initial triangle and node on the boundary of the mesh
	if (getTopLeftTriangle(currTri, edge)) return;
	pb1 = (Pt2dLink *) currTri->getLink(edge);
	pb2 = (Pt2dLink *) currTri->getLink((edge+1)%3);

	// decode motion vectors of nodes b1 and b2
	code = (int) m_bitStrmIn->getOneBit();
	if (code == 0) {				// nonzero motion vector
		mvPred = Point2d(0.,0.);	// prediction vector x 2
		pb1->motion = decodeMV(mvPred);
	}
	else							// zero motion vector
		pb1->motion = 0.0;

	code = (int) m_bitStrmIn->getOneBit();
	if (code == 0) {					// nonzero motion vector
		mvPred = MOTVEC_FACTOR*(pb1->motion); // prediction vector x 2
		pb2->motion = decodeMV(mvPred);
	}
	else							// zero motion vector
		pb2->motion = 0.0;

	// mark initial triangle and nodes
	currTri->mark = 1;
	pb1->mark = pb2->mark = 1;

	// initialize the triangle queue 
	TriangleQueue queue;
	queue.append (currTri, edge);

	// now start the actual breadth-first traversal of the mesh
	while (queue.empty() == 0) {
		// get and remove first triangle from the queue
		queue.remove(currTri, edge);
		leftEdge = ((edge==0)? 1: ((edge==1)? 2: 0));
		rightEdge = ((edge==0)? 2: ((edge==1)? 0: 1));
		pb1 = currTri->getLink(edge);
		pb2 = currTri->getLink(leftEdge);
		paa = currTri->getLink(rightEdge);

		// process node aa if it is not already marked processed
		if (paa->mark == 0) {
			code = (int) m_bitStrmIn->getOneBit();
			if (code == 0) {			// nonzero motion vector
				// compute mv prediction and round to half-pixel accuracy
				mvPred = MOTVEC_FACTOR*(pb1->motion + pb2->motion)/2.0;
				mvPred.roundToInt();
				// decode motion vector and store it
				paa->motion = decodeMV(mvPred);
			}
			else						// zero motion vector
				paa->motion = 0.0;
			
			// mark this node processed
			paa->mark = 1;
		}

		// if there is a neighboring unmarked triangle at the right
		// of the current triangle, mark it and queue it
		rightTri = currTri->getNeighborTri(rightEdge);
		if (rightTri != NULL && rightTri->mark == 0) {
			rightTri->mark = 1;
			queue.append(rightTri, currTri->getNeighborEdge(rightEdge));
		}
			
		// if there is a neighboring unmarked triangle at the left
		// of the current triangle, mark it and queue it
		leftTri = currTri->getNeighborTri(leftEdge);
		if (leftTri != NULL && leftTri->mark == 0) {
			leftTri->mark = 1;
			queue.append(leftTri, currTri->getNeighborEdge(leftEdge));
		}

	} // end-while

	// now apply motion vectors to node locations
	for (int i=0; i<meshCurr.getSizePoints(); i++) {
		paa = (Pt2dLink *) meshCurr.addressC(i);
		paa->setPoint(paa->getPoint() + paa->motion);
	}

} // decMeshMotion


// Decode coordinates of first mesh node point; return them.
Point2d MeshDecoder::decodeFirstNode()
{
	int coordX, coordY;

	// x-coordinate of first node (signed integer coded as 2s-complement)
	if ((m_bitStrmIn->getOneBit()) == 0)
		coordX = (int) m_bitStrmIn->getBits(NUMBITS_FIRSTNODEPOS);
	else
		coordX = (int) m_bitStrmIn->getBits(NUMBITS_FIRSTNODEPOS) - (1 << NUMBITS_FIRSTNODEPOS);

	decodeMarkerBit();

	// y-coordinate of first node (signed integer coded as 2s-complement)
	if ((m_bitStrmIn->getOneBit()) == 0)
		coordY = (int) m_bitStrmIn->getBits(NUMBITS_FIRSTNODEPOS);
	else
		coordY = (int) m_bitStrmIn->getBits(NUMBITS_FIRSTNODEPOS) - (1 << NUMBITS_FIRSTNODEPOS);
	m_bitStrmIn->getOneBit();

	// return the node point
	return Point2d((double)coordX/COORD_FACTOR, (double)coordY/COORD_FACTOR);
}


// Decode coordinate-differences for other node points than the first; 
// return them. Coordinate-differences are coded using VLC's.
Point2d MeshDecoder::decodeNextNode()
{
	int diffX, diffY;

	// decode x-coordinate-difference
	long lengthSymbol = m_entrDecSet->m_pentrdecWrpPnt->decodeSymbol();
	if (lengthSymbol < 0 || lengthSymbol > MAX_COORD_DIFF_CODE_LEN) {
		doError("MeshDecoder: invalid length symbol VLC\n");
		return Point2d(0,0);
	}
	int iSign = m_bitStrmIn->peekBits(1);	// peek at the sign
	if (iSign == 1) 
		diffX = m_bitStrmIn->getBits(lengthSymbol);
	else {
		// first get the magnitude
		int iAbsDiff = ~(m_bitStrmIn->getBits(lengthSymbol));
		// mask out the irrelevant bits and apply sign
		int iMask = ~(0xFFFFFFFF << lengthSymbol);	
		diffX = -1 * (iAbsDiff & iMask);
	}

	// decode y-coordinate-difference
	lengthSymbol = m_entrDecSet->m_pentrdecWrpPnt->decodeSymbol();
	if (lengthSymbol < 0 || lengthSymbol > MAX_COORD_DIFF_CODE_LEN) {
		doError("MeshDecoder: invalid length symbol VLC\n");
		return Point2d(0,0);
	}
	iSign = m_bitStrmIn->peekBits(1);	// peek at the sign
	if (iSign == 1) 
		diffY = m_bitStrmIn->getBits(lengthSymbol);
	else {
		// first get the magnitude
		int iAbsDiff = ~(m_bitStrmIn->getBits(lengthSymbol));
		// mask out the irrelevant bits and apply sign
		int iMask = ~(0xFFFFFFFF << lengthSymbol);	
		diffY = -1 * (iAbsDiff & iMask);
	}

	// return the (difference) node point
	return Point2d((double)diffX/COORD_FACTOR, (double)diffY/COORD_FACTOR);
}


// Decode motion range parameters
int MeshDecoder::decodeMotionRange()
{
	// decode motion-range-code
	m_fCode = (uint) m_bitStrmIn->getBits (NUMBITS_MOP_FCODE);
	if (m_fCode <= 0  ||  m_fCode >= 8) {
		doError("MeshDecoder: invalid motion range code\n");
		return 1;
	}
	// compute motion vector scale factor and motion range in half-pel units
	m_scaleFactor = 1 << (m_fCode - 1);
	m_motionRange = 32 << (m_fCode - 1);
	return 0;
}


// Descale a difference motion vector component for decoding
int deScaleMV(const int iVLC, const uint uiResidual, const uint uiScaleFactor)
{
	if (iVLC == 0 && uiResidual == 0)
		return 0;
	else if (uiScaleFactor == 1)
		return (iVLC);
	else {
		uint uiAbsDiffMVcomponent = 
			abs(iVLC)*uiScaleFactor + uiResidual - uiScaleFactor + 1;
		return (sign (iVLC) * uiAbsDiffMVcomponent);
	}
}


// Decode a motion vector (given a prediction vector) and return it;
// mvPred is the prediction vector x 2
Point2d MeshDecoder::decodeMV(Point2d &mvPred)
{
	double mvX, mvY;
	uint uiResidual;

	// get difference code and residual from bitstream
	int iVLC = m_entrDecSet->m_pentrdecMV->decodeSymbol(); 
	if (iVLC < 0) {
		doError("MeshDecoder: invalid motion vector VLC\n");
		return Point2d(0,0);
	}
	iVLC -= 32;
	if (m_fCode > 1 && iVLC != 0)	
		uiResidual = m_bitStrmIn->getBits(m_fCode - 1);
	else
		uiResidual = 0;

	// compute motion vector component in integer units
	mvX = mvPred.X() + (double) deScaleMV(iVLC, uiResidual, m_scaleFactor);
	// fit in the appropriate range [-range, range)
	if (mvX < -m_motionRange)
		mvX += (m_motionRange << 1);
	else if (mvX >= m_motionRange)
		mvX -= (m_motionRange << 1);

	// get difference code and residual from bitstream
	iVLC = m_entrDecSet->m_pentrdecMV->decodeSymbol();
	if (iVLC < 0) {
		doError("MeshDecoder: invalid motion vector VLC\n");
		return Point2d(0,0);
	}
	iVLC -= 32;
	if (m_fCode > 1 && iVLC != 0)
		uiResidual = m_bitStrmIn->getBits(m_fCode - 1);
	else
		uiResidual = 0;

	// compute motion vector component in integer units
	mvY = mvPred.Y() + (double) deScaleMV (iVLC, uiResidual, m_scaleFactor);
	// fit in the appropriate range [-range, range)
	if (mvY < -m_motionRange)
		mvY += (m_motionRange << 1);
	else if (mvY >= m_motionRange)
		mvY -= (m_motionRange << 1);

	// return motion vector at full pixel scale
	return (Point2d(mvX / MOTVEC_FACTOR, mvY / MOTVEC_FACTOR));
}


// Decode marker bit and check it has value 1
void MeshDecoder::decodeMarkerBit()
{
	int bit = m_bitStrmIn->getOneBit();
	if (bit == 0)
		doError("MeshDecoder: marker-bit != 1 ...\n");
}


// Search for the next start-code in the bitstream, or return EOF
int MeshDecoder::findStartCode()
{
	// ensure byte alignment by flushing current byte
	m_bitStrmIn->flush ();

	if (m_bitStrmIn->eof()==EOF)
		return EOF;

	// read bytes until start-code found or EOF occurs
	while (m_bitStrmIn->peekBits(NUMBITS_SC_PREFIX) != SC_PREFIX)
	{
		m_bitStrmIn->getBits(8);
		if (m_bitStrmIn->eof()==EOF)
			return EOF;
	}
	return 0;
}


// Go to next byte-aligned position in the bitstream;
// note: reads at least one 0; then some 1's if necessary
int MeshDecoder::nextByteAlign()
{
	register int align = (m_bitStrmIn->getCounter())%8;

	if (align != 0)
		// ensure byte alignment by just flushing current byte
		m_bitStrmIn->flush ();
	else {
		// check if end of stream is reached
		if (m_bitStrmIn->eof() == EOF) {
			doError("MeshDecoder: nextByteAlign: end of stream ...\n");
			return EOF;
		}
		// go to next byte aligned position by reading a whole byte
		register int code = (int) m_bitStrmIn->getBits(8);
		if (code != (int) 0x007F) {
			doError("MeshDecoder: nextByteAlign: byte != 0x007F ...\n");
			return 1;
		}
	}
	return 0;
}


// Compute mesh bounding box attibutes
int MeshDecoder::setBBox()
{
	if (meshCurr.getSizePoints() > 0) {
		// initialize attributes using first node point
		Point2d &first = meshCurr.point(0);
		m_xmin =  floor(first.X());
		m_xmax =  ceil(first.X());
		m_ymin =  floor(first.Y());
		m_ymax =  ceil(first.Y());

		// for all the other vertices in the mesh do ...
		for (int i=1; i<meshCurr.getSizePoints(); i++) {
			// get i'th vertex and check the coordinates
			Point2d &p = meshCurr.point(i);
			double val = floor(p.X());
			if (val < m_xmin)
				m_xmin = val;
			val = ceil(p.X());
			if (val > m_xmax)
				m_xmax = val;
			val = floor(p.Y());
			if (val < m_ymin)
				m_ymin = val;
			val = ceil(p.Y());
			if (val > m_ymax)
				m_ymax = val;
		}
	}
	else {
		doError("MeshDecoder: setBBox: current mesh has no nodes\n");
		m_xmin = m_xmax = m_ymin = m_ymax = 0.0;
		return 1;
	}
	DPRINT(("\txmin=%.2f, ymin=%.2f, xmax=%.2f, ymax=%.2f\n", 
			m_xmin, m_ymin, m_xmax, m_ymax));
	return 0;
}


// Get the top left node on the boundary of the mesh and the triangle
// that contains it 
int MeshDecoder::getTopLeftTriangle(Tri2dRef *&tref, int &edge)
{
	if (m_initTri == NULL  ||  m_initEdge == -1) {
		doError("MeshDecoder: top left triangle not defined...\n");
		return 1;
	}
	tref = m_initTri;
	edge = m_initEdge;
	return 0;
}

// Find the top left node on the boundary of the mesh and the triangle
// that contains it by searching for the triangle vertex (node) with the
// minimum value of x+y (using y value to break ties)
int MeshDecoder::setTopLeftTriangle()
{
	if (meshCurr.getSize() < 1) {
		doError("MeshDecoder: current mesh has no triangles\n");
		return 1;
	}
	Tri2dRef *tmpRef = meshCurr.nextTriangle(1);
	double pxymin = MAXDOUBLE;
	double pymin  = MAXDOUBLE; 
	// check each triangle
	for (int i=0; i<meshCurr.getSize(); i++) {
		// check each edge/node point
		for (int j=0; j<3; j++) {
			if (tmpRef->getNeighborTri(j) == NULL) {
				// edge j,j+1 this is a boundary edge 
				Point2d pt = tmpRef->getLink(j)->getPoint();
				double pxy = pt.X()+pt.Y();
				// find node with minimum x+y; y value breaks ties
				if ((pxy < pxymin) ||	
					(pxy == pxymin	&&	pt.Y() < pymin)) {	
					pxymin = pxy;
					pymin = pt.Y();
					// store pointer to triangle and edge index
					m_initTri = tmpRef;
					m_initEdge = j;
				}
			}
		} // endfor edges
		tmpRef = meshCurr.nextTriangle();
	} // endfor triangles
	return 0;
}




////////// MAIN FUNCTION TO RUN STAND-ALONE //////////////////////////////////

#if defined(VISUAL_DECODER)

void usage (char *name)
{
	fprintf(stderr, "ISO/IEC 14496 (MPEG-4) 2D Mesh Object decoder\n");
	fprintf(stderr, "Usage: %s <bitstream> <basename> <firstframe> <frameincr> [xo] [yo]\n", name);
	fprintf(stderr, "bitstream:	 input mesh bitstream to be decoded\n");
	fprintf(stderr, "basename:	 basename of output mesh-files\n");
	fprintf(stderr, "firstframe: number of first frame\n");
	fprintf(stderr, "frameinc:	 frame increment value\n");
	fprintf(stderr, "xo,yo:		 bounding-box origin coordinates\n");
}


// Stand-alone Visual decoder main program
int main (int argc, char *argv[])
{
	// Basename of filename with output mesh data, without numbers or extension
	char meshDatFile[200];
	// Full filename of input bitstream file
	char bitStreamFile[200];
	// Full filename of output mesh file
	char meshFile[200];
	// Pointer to buffer holding file-input
	BYTE *inBuffer;
	// Flags first decoded frame
	int first = 1;
	// MOP is_intra bit (I or P coding type)
	int isIntra;
	// Frame number being decoded, number of first frame and frame increment
	int frame = 0;
	int firstFrame = 0;
	int frameInc = 1;
	// Mesh bounding box origin coordinates for composition
	double xOrig = 0.0;
	double yOrig = 0.0;


	// check number of arguments and read arguments from command line
	if (argc < 5) {
		usage(argv[0]);
		return 1;
	}
	sprintf (bitStreamFile, "%s", argv [1]);
	sprintf (meshDatFile, "%s", argv [2]);
	firstFrame = atoi(argv[3]);
	frameInc   = atoi(argv[4]);
	xOrig = yOrig = 0.0;
	if (argc == 7) {
		xOrig = atof(argv[5]);
		yOrig = atof(argv[6]);
	}

	DPRINT(("========================================================\n"));
	DPRINT(("ISO/IEC 14496 (MPEG-4) 2D Mesh Object decoder ---- v.5.0\n"));
	DPRINT(("========================================================\n\n"));
	DPRINT(("\tBitstreamfile       : %s\n", bitStreamFile));
	DPRINT(("\tBasename meshfiles  : %s\n", meshDatFile));
	DPRINT(("\tFirst frame no.     : %d\n", firstFrame));
	DPRINT(("\tFrame increment     : %d\n", frameInc));
	DPRINT(("\tBounding-box offset : %.2f,%.2f\n\n", xOrig, yOrig));

	// open bitstream file and determine its size
	struct stat fstatus;
	if (stat (bitStreamFile, &fstatus) == -1) {
		fprintf(stderr, "Cannot open bitstream file\n");
		usage(argv[0]);
		return 1;
	}

	// read complete file into BYTE buffer
	// assume this file is not enormously big
	uint finSize = (int) fstatus.st_size;
	inBuffer = new BYTE [finSize];
	int fd = open(bitStreamFile, O_RDONLY|O_BINARY);
	read(fd, inBuffer, finSize);
	close(fd);

	// declare mesh decoder and attach it to the input buffer
	MeshDecoder decoder(inBuffer, finSize);


	DPRINT(("========================================================\n\n"));

#if defined(TIMER)
	// Run time logging
#if defined(WIN32)
	struct _timeb tstart, tstop;
	_ftime(&tstart);
#else
	struct timeval tstart,tstop;
	gettimeofday(&tstart, (struct timezone *)NULL);
#endif
#endif

	frame = firstFrame;
	
	/////////////// MESH OBJECT PLANE LOOP ////////////////////////////
	
	// until end of bitstream file do
	while (decoder.getEndStrm() == 0) {

		// decode a mesh by decoding a I- or P-MOP
		int ret = decoder.decodeMesh(isIntra);
		if (ret != 0)
			break;

		if (first) {
			// move the mesh by specified offset
			decoder.moveMesh(xOrig, yOrig);
			first = 0;
		}

		// write mesh data of current frame
		sprintf(meshFile, "%s%d.tri", meshDatFile, frame);
		decoder.dumpMesh(meshFile);
		frame+=frameInc;

	} // endwhile (frame loop)

	DPRINT(("\n========================================================\n\n"));

#if defined(TIMER)
	// compute and print running time in hundredths of a second
#if defined(WIN32)
	_ftime(&tstop);
	int runtime = 100*(tstop.time-tstart.time) +
		(tstop.millitm-tstart.millitm)/10;
#else
	gettimeofday(&tstop, (struct timezone *)NULL);
	int runtime = 100*(tstop.tv_sec-tstart.tv_sec) +
		(tstop.tv_usec-tstart.tv_usec)/10000;
#endif
	if (runtime <= 0) runtime = 1;
	frame = (frame-firstFrame)/frameInc;
	printf("\tTiming info		 : %d.%02d seconds, %d frames, %d.%02d fps\n",
		   runtime/100, runtime%100,
		   frame, ((10000*frame+runtime/2)/runtime)/100,
		   ((10000*frame+runtime/2)/runtime)%100);
#endif


	////////// cleanup my junk and exit ///////////////////////////////////
	delete [] inBuffer;
	return 0;

} // main


#endif // VISUAL_DECODER


// end of file
