/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* The University of Rochester retains full right to use the code for */
/* their own purposes, assign or donate the code to a third party and */
/* to inhibit third parties from using the code for non MPEG-4 Visual */
/* (ISO/IEC 14496-2) conforming products.                             */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1997-1999                                            */
/*                                                                    */
/**********************************************************************/

/*+***********************************************************-*-C++-*-*
   NAME         mesh_decode.h
   AUTHOR/DATE  Peter Van Beek, Fri Aug 21 17:51:21 1998
   DESCRIPTION  Mesh decoder class definitions.
   SYNOPSIS     
*********************************************************************-*/

#ifndef	MESH_DECODER_DEF
#define	MESH_DECODER_DEF

#include "pointClass.h"					// Point2d
#include "triRefClass.h"				// Tri2dRef
#include "triTessClass.h"				// Tri2dTess


/////////////// DEFINITIONS /////////////////////////////////////////////////

// Define byte
#ifndef BYTE
#define BYTE 	unsigned char
#endif


/////////////// CLASS DEFINITIONS ///////////////////////////////////////////

// forward class definitions
class CInBitStream;
class CEntropyDecoderSet;


// MeshDecoder class definition
#ifndef WIN32
class MeshDecoder
#else
class __declspec(dllexport) MeshDecoder
#endif
{
  public:
	// Constructor 
	MeshDecoder();

	// Constructor that attaches given input buffer to the decoder
	MeshDecoder(BYTE *pInBuffer, const int inSize);

	// Destructor
	~MeshDecoder();

	// Attach a given input buffer with given size and bit-position
	void attachStream(BYTE *pInBuffer, const int inSize, const int bitPos);

	// Decode one Mesh Object Plane (MOP) from a bit stream 
	int decodeMesh (int &isIntra);

	// Get attributes
	int getStartStrm()     { return m_startStrm; }
	int getEndStrm()       { return m_endStrm; }
	BYTE *getStreamPtr(); 
	uint getBitsLeft(); 

	// Access to mesh data
	int getNumVertices() { return meshCurr.getSizePoints(); }
	int getNumTriangles() { return meshCurr.getSize(); }
	int getVertexCoordinates(int i, double &x, double &y);
	int getTriangleIndices(int i, int &v0, int &v1, int &v2);

	// Mesh bounding box attributes
	double getMinX() { return m_xmin; }
	double getMinY() { return m_ymin; }
	double getMaxX() { return m_xmax; }
	double getMaxY() { return m_ymax; }

	// Move (translate) mesh data
	void moveMesh(double dx, double dy) { meshCurr.move(dx, dy); }
	// Dump mesh data to a file
	void dumpMesh(char *fileName) { meshCurr.write(fileName); }
	
  private:
	// Decode Mesh Object Plane Header
	int decodeMopHeader (int &isIntra, uint &meshMask, double &timeStamp, int &useStartCode);
	// Decode geometry and boundary of Delaunay-type mesh (I-MOP)
	void decDelaunayMesh ();
	// Decode geometry and boundary of Uniform-type mesh (I-MOP)
	void decUniformMesh ();
	// Decode motion of previous to current mesh (P-MOP)
	void decMeshMotion ();

	// Decode mesh node point coordinate (difference)
	Point2d decodeFirstNode();
	Point2d decodeNextNode();
	// Decode motion range parameters
	int decodeMotionRange();
	// Decode a mesh node motion vector
	Point2d decodeMV (Point2d &mv);
	// Decode marker bit
	void decodeMarkerBit();
	// Skip stuffing bits until next start code
	int findStartCode();
	// Skip stuffing bits until byte-aligned
	int nextByteAlign();
	// Compute mesh bounding box attributes
	int setBBox();
	// Find initial triangle and edge for breadth-first traversal 
	int getTopLeftTriangle(Tri2dRef *&tref, int &edge);
	int setTopLeftTriangle();

	// The current (decoded) mesh
	Tri2dTess meshCurr;

	// Boolean to flag if stream has started
	int m_startStrm;
	// Boolean to flag if stream has ended
	int m_endStrm;

	// Bitstream and Huffman entropy decoding pointers
	CInBitStream* m_bitStrmIn;
	CEntropyDecoderSet* m_entrDecSet;

	// Motion vector range code (Fcode): 1, 2, ..., 7
	uint m_fCode; 
	// Motion vector scale factor
	uint m_scaleFactor;
	// Motion range in half pixel units: 32, 64, ..., 2048
	int  m_motionRange;

	// Mesh bounding box attributes
	double m_xmin, m_xmax, m_ymin, m_ymax;

	// Pointer to initial triangle for breadth-first-traversal
	Tri2dRef *m_initTri;
	// Number of initial vertex/edge on initial triangle
	int m_initEdge;
};


#endif // MESH_DECODER_DEF

// End of file
