/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* The University of Rochester retains full right to use the code for */
/* their own purposes, assign or donate the code to a third party and */
/* to inhibit third parties from using the code for non MPEG-4 Visual */
/* (ISO/IEC 14496-2) conforming products.                             */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1996 - 1999                                          */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         mesh_encode.cpp
   AUTHOR/DATE  Peter van Beek, Wed Mar  3 17:19:07 1999

   DESCRIPTION  ISO/IEC 14496 (MPEG-4) 2D Mesh object encoding.

                In general, the encoding program reads tracking data
                of a moving video-object from a set of files to generate 
                a bitstream. 
                The tracking data is in the form of a sequence of 2-D 
                triangular meshes, each describing the geometry of a 2-D 
                mesh at a certain point in time. These files have extension
                ".tri" (for triangulation). 
                Note that, all the nodes in the mesh are present from
                the first to the last time instant (i.e. no nodes are ever
                added or deleted). 

                Some overhead is necessary around the core syntax. These
                bits are counted separately. These are "mesh object" start 
                codes, "mesh object" end codes, and "mesh object plane" 
                start codes.

                This encoder can generate two types of bitstreams. The
                first would be used as a standalone bitstream, which can
                be part of a Visual Object bitstream (see ISO/IEC 14496-2).
                The second would be used as part of a BIFS-animation 
                stream, which can be re-played on a MPEG-4 Systems player
                (see ISO/IEC 14496-1). The two types differ only in their
                headers (start-codes etc.), the core syntax is identical.

                A parameter file is used to communicate all necessary
                parameters to the program, instead of using command-line
                arguments.

   NOTES        Note that in this code, one assumes a coordinate system
                with origin at the top-left, such that the x-axis points
                to the right and the y-axis points down. Definitions of
                above/below, clockwise/counterclockwise etc. all depend
                on this assumption. Other assumptions on the mesh data-
                structure are many. 

                This code assumes that all mesh node positions and motion 
                vectors have 0.5 pixel accuracy with respect to the image/
                texture map that will be used for rendering. Therefore,
                after decoding these positions and motion vectors in
                integer values, they are divided by 2.0 to transform them 
                into 0.5 accurate values.

                Motion vector data coding is similar to that of MPEG-4 
                Video (prediction of motion vectors is not the same, but
                coding of prediction error is the same).
                The mesh node motion vectors are expected to lie within
                a certain range (which can be specified to be several 
                values). Since the motion of the input meshes was estimated
                externally, one should take care in specifying a correct
                motion vector range value. 


*********************************************************************-*/

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <assert.h>
#include <math.h>

#if defined(WIN32)
#include <float.h>
#define MAXDOUBLE DBL_MAX
#include <time.h>
#include <sys/timeb.h>
#else
#include <values.h>
#include <sys/time.h>
#endif

#include "general.h"			// various definitions
#include "linkClass.h"			// Pt2dLink
#include "pointClass.h"			// Point2d
#include "polyClass.h"			// Poly2d
#include "triRefClass.h"		// Tri2dRef
#include "triTessClass.h"		// Tri2dTess
#include "queueClass.h"			// TriangleQueue

#include "basic.h"				// definitions
#include "entropy.h"			// entropy coding
#include "huffman.h"			// huffman coding 
#include "bitstrm.h"			// bitstream functions

#include "mesh_encode.h"		// MeshEncoder



////////// COMPILER SWITCHES /////////////////////////////////////////////////

// Define to get some basic useful messages
//#define DPRINTING
// Define printing macro which allows variable number of arguments
#if defined(DPRINTING) || defined(_DEBUG)
#define DPRINT(vargs)		((void) printf vargs)
#else
#define DPRINT(vargs)
#endif

// Define to get running-time message
#define TIMER

// Define to check if any motion vector is out of range
#define CHECK_MV


// Define VISUAL_ENCODER in Makefile if encoder must produce a stand-alone 
// bitstream conforming to Visual part of MPEG-4 standard
#if !defined(VISUAL_ENCODER)
// Define COMPOSITION if encoder is used in systems/composition framework
// - this will produce a BIFS animation-stream containing mesh-data!
#define COMPOSITION
#endif

// Define only for generating conformance test bitstreams
//#define CONFORMANCE_TEST
#if defined(CONFORMANCE_TEST)
#include "conformance.h"
//#define TWEAK_STARTCODE
//#define TWEAK_HEADER
//#define TWEAK_MASK
#endif


////////// CONSTANTS /////////////////////////////////////////////////////////

// Mesh Object (MO) overhead
#define START_CODE					1
#define NUMBITS_START_CODE			24
#define MO_START_CODE				188			// 0xbc
#define MO_END_CODE					177			// 0xb1
#define NUMBITS_MO_START			8
#define NUMBITS_MO_END				8

// Mesh Object Plane (MOP) overhead
#define MOP_START_CODE				189			// 0xbd
#define NUMBITS_MOP_START			8

// Geometry and motion coding
#define NUMBITS_MESH_TYPE			2
#define NUMBITS_NUMMESHNODES		16
#define NUMBITS_NUMBOUNNODES		10
#define NUMBITS_FIRSTNODEPOS		13
#define NUMBITS_NUMNODE_UNIF		10
#define NUMBITS_RECTSIZE_UNIF		8
#define NUMBITS_SPLIT_UNIF			2
#define NUMBITS_MOP_FCODE			3

// Maximum difference in coordinate units between two points
#define MAX_COORD_DIFF				16383
// Maximum length of coordinate difference codeword
#define MAX_COORD_DIFF_CODE_LEN		14

// Node location and motion vector accuracy
#define COORD_FACTOR				2.0
#define MOTVEC_FACTOR				2.0

// Default frame-rate of Mesh Object
#define DEF_FRAME_RATE				10
// Default time (in hours) 
#define DEF_TIME_HOUR				0
// Default frame skip value 
#define DEF_FRAME_SKIP				0

#if defined(COMPOSITION)
// BIFS-anim start / end codes
#define ANIM_START_CODE				396			// 0x18c
#define ANIM_END_CODE				397			// 0x18d
#endif



////////// CLASS MEMBER FUNCTIONS ////////////////////////////////////////////


// Mesh Encoder constructor
MeshEncoder::MeshEncoder(const EncParam &par)
{
	// open bitstream file
	m_outStrm.open(par.bitStreamFile, ios::binary | ios::out | ios::trunc);
	// create output bitstream attached to file output stream
	m_bitStrmOut = new COutBitStream (m_outStrm, 0, NULL);
	// attach output bitstream to entropy encoders
	m_entrEncSet = new CEntropyEncoderSet (*m_bitStrmOut);

	m_meshType = par.meshType;
	m_xmin = m_xmax = m_ymin = m_ymax = 0.0;
	m_initTri = m_initEdge = -1;

	// set motion range related coding parameters
	if (par.motionRange <= 16)
		m_fCode = 1;
	else if (par.motionRange <= 32)
		m_fCode = 2;
	else if (par.motionRange <= 64)
		m_fCode = 3;
	else if (par.motionRange <= 128)
		m_fCode = 4;
	else if (par.motionRange <= 256)
		m_fCode = 5;
	else if (par.motionRange <= 512)
		m_fCode = 6;
	else if (par.motionRange <= 1024)
		m_fCode = 7;
	else {
		cerr << "Motion range specified is too big ..." << endl;
		assert (FALSE);
	}
	m_motionRange = 16 << m_fCode;
	m_scaleFactor = 1 << (m_fCode - 1);

	// set timing header related coding parameters
	hInfo.isRate = par.isRate;
	hInfo.frameRateBase = par.frameRateBase;
	hInfo.frameRateFrac = par.frameRateFrac;
	hInfo.frameRateOffset = par.frameRateOffset;
}


// Mesh Encoder destructor
MeshEncoder::~MeshEncoder()
{
	delete m_entrEncSet;
	delete m_bitStrmOut;
}


// Increase motion vector range
void MeshEncoder::incrementMotionRange() 
{
	if (m_fCode == 7) {
		doError("Motion range cannot be increased further ...\n");
		assert (FALSE);
	}
	else {
		m_fCode++;
		cout << "Increasing motion range code to " << m_fCode << endl;
		m_motionRange = 16 << m_fCode;
		m_scaleFactor = 1 << (m_fCode - 1);
	}
}


// Encode Mesh Object start-code; return number of bits spent.
uint MeshEncoder::encodeMoStart()
{
	m_bitStrmOut->putBits(START_CODE, NUMBITS_START_CODE, "start_code");
#if defined(COMPOSITION)	// bifs-anim
	m_bitStrmOut->putBits(ANIM_START_CODE, NUMBITS_MO_START, "MO_start");
#else						// visual 
	m_bitStrmOut->putBits(MO_START_CODE, NUMBITS_MO_START, "MO_start");
#endif
	return (NUMBITS_START_CODE + NUMBITS_MO_START);
}


// Encode Mesh Object Plane start-code; return number of bits spent.
uint MeshEncoder::encodeMopStart()
{
	m_bitStrmOut->putBits(START_CODE, NUMBITS_START_CODE, "start_code");
#if defined(COMPOSITION)
	// encode bifs-anim "mesh object plane" start code
	m_bitStrmOut->putBits(ANIM_START_CODE, NUMBITS_MOP_START, "MOP_start");
#else
	// encode visual "mesh object plane" start code
	m_bitStrmOut->putBits(MOP_START_CODE, NUMBITS_MOP_START, "MOP_start");
#endif
	return (NUMBITS_START_CODE + NUMBITS_MOP_START);
}


// Encode Mesh Object end-code; return number of bits spent.
// Note: in the FCD/DIS of ISO/IEC 14496-2 the Mesh Object end code doesn't
// exist anymore; rather an end-code is sent only at the end of a visual 
// object sequence. We leave it here for purposes of bitstream exchange
// and verification etc. Peter Van Beek, Thu Aug 20 12:04:52 1998.
uint MeshEncoder::encodeMoEnd()
{
	m_bitStrmOut->putBits (START_CODE, NUMBITS_START_CODE, "Start_Code");
#if defined(COMPOSITION)	// bifs-anim
	m_bitStrmOut->putBits (ANIM_END_CODE, NUMBITS_MO_START, "MO_End");
#else						// visual
	m_bitStrmOut->putBits (MO_END_CODE, NUMBITS_MO_END, "MO_End");
#endif
	m_bitStrmOut->flush();
	return (NUMBITS_START_CODE + NUMBITS_MO_END);
}


/*+==========================================================================
	NAME		MeshEncoder::encodeMopHeader 
	VERSION		3.0, Peter van Beek, Wed Mar 31 09:21:59 1999
	PURPOSE		Encode Mesh Object Plane header (except start code)
	METHOD		Simple. Optionally inserts some timing information. 
	INPUT		isIntra			1: Intra plane; 0: Inter plane
				meshMask		1: Coded plane; 0: Not coded
	OUTPUT		none
	RETURNS		Number of bits used
	ERROR HANDLING
	NOTES		
==========================================================================-*/
uint MeshEncoder::encodeMopHeader(const int meshMask, const int isIntra)
{
	uint nbits = 0;


	// encode "is_intra" bit
	m_bitStrmOut -> putBits (isIntra, 1, "is_intra");
	nbits += 1;

	// encode "mesh_mask" bit
	m_bitStrmOut -> putBits (meshMask, 1, "mesh_mask");
	nbits += 1;

	if (isIntra) {
		// encode "is_frame_rate" bit and frame rate data
		if (hInfo.isRate) {
			m_bitStrmOut -> putBits (1, 1, "is_frame_rate");
			m_bitStrmOut -> putBits (hInfo.frameRateBase, 8, "rate_base");
			m_bitStrmOut -> putBits (hInfo.frameRateFrac, 4, "rate_frac");
			m_bitStrmOut -> putBits (hInfo.frameRateOffset, 1, "rate_offset");
			nbits += 14;
		}
		else {
			m_bitStrmOut -> putBits (0, 1, "is_frame_rate");
			nbits += 1;
		}
		// encode "is_time_code" bit and time code data
		if (hInfo.isTime) {
			m_bitStrmOut -> putBits (1, 1, "is_time_code");
			m_bitStrmOut -> putBits (hInfo.timeHours, 5, "time_hours");
			m_bitStrmOut -> putBits (hInfo.timeMinutes, 6, "time_minutes");
			m_bitStrmOut -> putBits (1, 1, "marker_bit");
			m_bitStrmOut -> putBits (hInfo.timeSeconds, 6, "time_seconds");
			nbits += 19;
		}
		else {
			m_bitStrmOut -> putBits (0, 1, "is_time_code");
			nbits += 1;
		}
	}
	// encode "skip_frames" bit and skip frame data
	if (hInfo.isSkip) {
		int frameSkip = hInfo.frameSkip;
		m_bitStrmOut -> putBits (1, 1, "skip_frames");
		nbits += 1;
		do {
			if (frameSkip < 15) 
				m_bitStrmOut -> putBits (frameSkip, 4, "frame_skip_val");
			else
				m_bitStrmOut -> putBits (15, 4, "frame_skip_val");

			frameSkip -= 15;
			nbits += 4;
		} while (frameSkip >= 0);
	}
	else {
		m_bitStrmOut -> putBits (0, 1, "skip_frames");
		nbits += 1;
	}
	return nbits;
}


/*+==========================================================================
	NAME		MeshEncoder::encodeMopData 
	VERSION		1.0, Peter van Beek, Wed Mar 31 09:21:59 1999
	PURPOSE		Encode Mesh Object Plane data
	METHOD		Simple. 
	INPUT		isIntra			1: Intra plane; 0: Inter plane
				meshMask		1: Coded plane; 0: Not coded
	OUTPUT		none
	RETURNS		Number of bits used
	ERROR HANDLING
	NOTES		
==========================================================================-*/
uint MeshEncoder::encodeMopData(const int meshMask, const int isIntra)
{
	uint nbits = 0;

	if (meshMask) {
		if (isIntra) {
			// encode mesh geometry (intra MOP)
			if (m_meshType == 0) {
				// encode uniform mesh geometry 
				nbits += encUniformMesh();
			}
			else if (m_meshType == 1) {
				// encode object-based (Delaunay) mesh geometry 
				nbits += encDelaunayMesh();
			}
		}
		else {
			// encode mesh motion (inter MOP)
			nbits += encMeshMotion ();
		}
	}
	return nbits;
}


/*+==========================================================================
	NAME		encUniformMesh
	VERSION		2.0, Peter Van Beek, Wed Aug 19 12:02:39 1998
	PURPOSE		Encode initial uniform mesh geometry by 5 parameters.
	METHOD		Simple
	RETURNS		Number of bits used to code the mesh geometry. 
	NOTES		Assumes current mesh is uniform and that the triangle
				dimensions are multiples of 0.5 pixel.
				Assumes that the orientations of all triangles is as one
				of four options defined in ISO/IEC 14496-2.
==========================================================================-*/
uint MeshEncoder::encUniformMesh ()
{
	int n, nhori, nverti;
	Point2d ptmin, ptmax;
	Tri2dRef *triTopLeft, *triSecond, *triThird;
	int edge, edgePrev, edgeNext;
	uint nBits = 0;

	DPRINT(("\tEncoding uniform mesh ...\n"));

	// check current input mesh
	int numMeshNodes = meshCurr.getSizePoints();
	if (numMeshNodes < 4) {
		doError("MeshEncoder: invalid input mesh ...\n");
		return 0;
	}

	// code the mesh_type_code; this flag is '01'
	m_bitStrmOut -> putBits (1, NUMBITS_MESH_TYPE, "mesh_type_code");
	nBits += 2;

	// get the most top-left and bottom right mesh nodes
	getTopLeftTriangle(meshCurr, triTopLeft, edge);
	ptmin = triTopLeft->getLink(edge)->getPoint();
	getUniformBottomRight(meshCurr, ptmax);

	// count all nodes with same x, resp. y coordinate as top-left pt
	nhori = 0; nverti = 0;
	for (n=0; n<meshCurr.getSizePoints(); n++) {
		Point2d &pt = meshCurr.point(n);
		if (pt.X() - ptmin.X() < EPSILON) 
			nverti++;
		if (pt.Y() - ptmin.Y() < EPSILON) 
			nhori++;
	}
		
	DPRINT(("\tMesh dimensions: %d (%d x %d) nodes\n", numMeshNodes, nhori, nverti));
	m_bitStrmOut -> putBits (nhori, NUMBITS_NUMNODE_UNIF);
	m_bitStrmOut -> putBits (nverti, NUMBITS_NUMNODE_UNIF);
	m_bitStrmOut->putBits(1, 1, "marker_bit");
	nBits += (2*NUMBITS_NUMNODE_UNIF + 1);

	// compute the triangle dimensions (must be multiple of 0.5!)
	double xsize = (ptmax.X()-ptmin.X())/(nhori-1);
	double ysize = (ptmax.Y()-ptmin.Y())/(nverti-1);
	int ixsize = (int) (COORD_FACTOR*xsize);
	int iysize = (int) (COORD_FACTOR*ysize);
	DPRINT(("\tMesh rectangle size : %.2f x %.2f (%d x %d )\n", xsize, ysize, ixsize, iysize));

	m_bitStrmOut -> putBits (ixsize, NUMBITS_RECTSIZE_UNIF);
	m_bitStrmOut -> putBits (iysize, NUMBITS_RECTSIZE_UNIF);
	nBits += 2*NUMBITS_RECTSIZE_UNIF;

	// now find out what the triangle orientations are
	edgePrev = ((edge==0)? 2: ((edge==1)? 0: 1));
	edgeNext = ((edge==0)? 1: ((edge==1)? 2: 0));
	if (triTopLeft->getNeighborTri(edgePrev) == NULL) {
		// orientation code is either 1 or 3 - look at third triangle
		triSecond = triTopLeft->getNeighborTri(edgeNext);
		edge = triTopLeft->getNeighborEdge(edgeNext);
		edgeNext = ((edge==0)? 1: ((edge==1)? 2: 0));
		triThird = triSecond->getNeighborTri(edgeNext);
		if (triThird == NULL) {
			// there is no third triangle; pick 1 arbitrarily
			DPRINT(("\tTriangle split code : 01\n"));
			m_bitStrmOut -> putBits ((int)1, (uint)NUMBITS_SPLIT_UNIF);
		}
		else {
			edge = triSecond->getNeighborEdge(edgeNext);
			edgeNext = ((edge==0)? 1: ((edge==1)? 2: 0));
			if (triThird->getNeighborTri(edgeNext) == NULL) {
				DPRINT(("\tTriangle split code : 01\n"));
				m_bitStrmOut -> putBits ((int)1, (uint)NUMBITS_SPLIT_UNIF);
			}
			else {
				DPRINT(("\tTriangle split code : 11\n"));
				m_bitStrmOut -> putBits ((int)3, (uint)NUMBITS_SPLIT_UNIF);
			}
		}
	}
	else {
		// orientation code is either 0 or 2 - look at third triangle
		triThird = triTopLeft->getNeighborTri(edgeNext);
		if (triThird == NULL) {
			// there is no third triangle; pick 2 arbitrarily
			DPRINT(("\tTriangle split code : 10\n"));
			m_bitStrmOut -> putBits ((int)2, (uint)NUMBITS_SPLIT_UNIF);
		}
		else {
			edge = triTopLeft->getNeighborEdge(edgeNext);
			edgeNext = ((edge==0)? 1: ((edge==1)? 2: 0));
			if (triThird->getNeighborTri(edgeNext) == NULL) {
				DPRINT(("\tTriangle split code : 10\n"));
				m_bitStrmOut -> putBits ((int)2, (uint)NUMBITS_SPLIT_UNIF);
			}
			else {
				DPRINT(("\tTriangle split code : 00\n"));
				m_bitStrmOut -> putBits ((int)0, (uint)NUMBITS_SPLIT_UNIF);
			}
		}
	}
	nBits += 2;

	DPRINT(("\tEncoded uniform mesh using: %d bits\n", nBits)); 
	return nBits;
}


/*+==========================================================================
	NAME		MeshEncoder::encDelaunayMesh
	VERSION		4.0, Peter Van Beek, Wed Aug 19 16:27:41 1998
	PURPOSE		Encode initial mesh geometry of an object-based mesh with
				Delaunay topology by encoding node positions.
	METHOD		First visits all boundary nodes, then all interior nodes.
				Applies greedy (nearest neighbor) algorithm to find a 
				tour of all interior nodes. Node positions are encoded
				by computing the difference in the coordinates wrt to the
				previously encoded node coordinates. This difference is
				VLC encoded.
	RETURNS		Number of bits used to encode the mesh geometry. 
	NOTES		Assumes 0.5 pixel accuracy of initial node coordinates.
				Assumes there are no two nodes with the same coordinates.
==========================================================================-*/
uint MeshEncoder::encDelaunayMesh ()
{
	uint nb, nBits = 0;					// numbers of bits
	Point2d pt, ptlast;					// current and previous point
	Point2d ptDiff;						// difference between 2 points
	int n, diff;		
	Tri2dRef *triInit, *tri;			// triangle pointers
	int edge, edgeInit;					// edge indices

	DPRINT(("\tEncoding Delaunay mesh ...\n")); 

	// check current input mesh
	int numMeshNodes = meshCurr.getSizePoints();
	if (numMeshNodes < 3) {
		doError("MeshEncoder: invalid input mesh ...\n");
		return 0;
	}
	DPRINT(("\tNumber of mesh nodes    : %d \n", numMeshNodes));

	// get the triangle/edge with the most top left node (on the boundary)
	getTopLeftTriangle(meshCurr, triInit, edgeInit);

	// walk around the boundary, count the number of boundary nodes of
	// the current input mesh and compute bounding box parameters
	int numBounNodes = getBounData(triInit, edgeInit);
	DPRINT(("\tNumber of boundary nodes: %d \n", numBounNodes));

	// code the mesh_type_code; this flag is '10'
	m_bitStrmOut -> putBits (2, NUMBITS_MESH_TYPE, "mesh_type_code");

	// code the number of nodes in the mesh with fixed length encoding
	m_bitStrmOut->putBits(numMeshNodes, NUMBITS_NUMMESHNODES, "n_mesh_nodes");
	m_bitStrmOut->putBits(1, 1, "Marker_Bit");
	// code the number of nodes on the boundary with fixed length encoding
	m_bitStrmOut->putBits(numBounNodes, NUMBITS_NUMBOUNNODES, "n_boun_nodes");
	m_bitStrmOut->putBits(1, 1, "Marker_Bit");
	// count coded bits
	nBits += NUMBITS_MESH_TYPE+NUMBITS_NUMMESHNODES+NUMBITS_NUMBOUNNODES+2;

	// initialize node markers
	meshCurr.markLinks(0);

	// encode the first node using a FLC for each coordinate
	pt = triInit->getLink(edgeInit)->getPoint();
#if defined(COMPOSITION)
	// coordinates are coded with respect to the origin of an integer- 
	// accurate bounding box around the mesh
	diff = (int) (COORD_FACTOR * (pt.X() - floor(m_xmin)));
#else
	// coordinates are coded with respect to image origin at top-left;
	// this is mainly to ease bitstream exchange and testing
	diff = (int) (COORD_FACTOR * pt.X());
#endif
	if (diff >= 0)
		m_bitStrmOut->putBits(diff, NUMBITS_FIRSTNODEPOS, "node0_x");
	else
		m_bitStrmOut->putBits(diff & 0x01FFF, NUMBITS_FIRSTNODEPOS, "node0_x");
	m_bitStrmOut->putBits(1, 1, "Marker_Bit");

#if defined(COMPOSITION)
	// coordinates are coded with respect to the origin of an integer-
	// accurate bounding box around the mesh
	diff = (int) (COORD_FACTOR * (pt.Y() - floor(m_ymin)));
#else
	// coordinates are coded with respect to image origin at top-left;
	// this is mainly to ease bitstream exchange and testing
	diff = (int) (COORD_FACTOR * pt.Y());
#endif
	if (diff >= 0)
		m_bitStrmOut->putBits(diff, NUMBITS_FIRSTNODEPOS, "node0_y");
	else
		m_bitStrmOut->putBits(diff & 0x01FFF, NUMBITS_FIRSTNODEPOS, "node0_y");

	m_bitStrmOut->putBits(1, 1, "marker_bit");
	nBits += (2*NUMBITS_FIRSTNODEPOS + 2);

	// mark this node
	triInit->getLink(edgeInit)->mark = 1;

	// code the other boundary nodes predictively with VLC+FLC
	tri = triInit;
	edge = edgeInit;
	for (n = 1; n < numBounNodes; n++) {
		ptlast = pt;
		// get the next boundary point in counterclockwise order
		getNextBounTriangle(tri, edge);
		pt = tri->getLink(edge)->getPoint();
		// compute and code coordinate differences and mark the node
		ptDiff = pt - ptlast;
		nb = encodeCoordinate (ptDiff);
		nBits += nb;
		tri->getLink(edge)->mark = 1;
	}

	// code the other mesh nodes predictively with VLC+FLC
	for (n = numBounNodes; n < numMeshNodes; n++) {
		ptlast = pt;
		// get (ptr to) the "closest" not already marked node
		Pt2dLink *link = getNearestNode(meshCurr, ptlast);
		// compute and code coordinate differences and mark the node
		pt = link->getPoint();
		ptDiff = pt - ptlast;
		nb = encodeCoordinate (ptDiff);
		nBits += nb;
		link->mark = 1;
	} 

	// return number of bits spent
	DPRINT(("\tEncoded %d mesh nodes using: %d bits\n", numMeshNodes, nBits)); 
	return nBits;
}


// Given a triangle on the mesh boundary and the number of the boundary
// edge, find the triangle on the boundary with the next ccw boundary edge.
// Assumes triangle vertices and neighbors are stored in clockwise manner.
void MeshEncoder::getNextBounTriangle(Tri2dRef *&tri, int &edge)
{
	int edgePrev = ((edge==0)? 2: ((edge==1)? 0: 1));
	// a NULL pointer corresponds to a boundary edge
	while (tri->getNeighborTri(edgePrev) != NULL) {
		Tri2dRef *triNext = tri->getNeighborTri(edgePrev);
		edge = tri->getNeighborEdge(edgePrev);
		tri = triNext;
		edgePrev = ((edge==0)? 2: ((edge==1)? 0: 1));
	}
	// this is the next counter-clockwise boundary edge
	edge = edgePrev;
}


// Given a triangle on the mesh boundary and the number of the boundary
// edge, walk around the boundary and count the number of boundary node 
// points and return it; also computes the mesh shape bounding box. 
int MeshEncoder::getBounData(Tri2dRef *triInit, const int edgeInit)
{
	// walk around the boundary - start from initial triangle/edge
	Tri2dRef *tri = triInit;
	int edge = edgeInit;
	int nBounNodes = 0;
	m_xmin = m_ymin = MAXDOUBLE;
	do {
		// get point, count it and update bounding box parameters
		Point2d pt = tri->getLink(edge)->getPoint();
		nBounNodes++;
		if (pt.X() < m_xmin)
			m_xmin = pt.X();
		if (pt.Y() < m_ymin)
			m_ymin = pt.Y();

		// go to the next boundary triangle/edge
		getNextBounTriangle(tri, edge);

	} while (tri != triInit || edge != edgeInit);

	// return number of boundary nodes
	return nBounNodes;
}


// Find the not already marked node nearest to 'ptlast'; distance measure
// is defined as |deltax|+|deltay|, to decide ties in this value first look
// at y coordinate of node, then at x coordinate. Assumes 'ptlast' is 
// already marked.
Pt2dLink *MeshEncoder::getNearestNode(Tri2dTess &mesh, Point2d &ptlast)
{
	Pt2dLink *retLink = NULL;
	double xy, xymin, ymin, xmin;
	xymin = ymin = xmin = MAXDOUBLE;
	// look at all nodes of this mesh
	for (int i = 0; i < mesh.getSizePoints(); i++) {
		Pt2dLink *link = mesh.addressC(i);
		if (link->mark == 0) {
			// node is not yet marked
			Point2d pt = link->getPoint();
			xy = fabs(pt.X()-ptlast.X())+fabs(pt.Y()-ptlast.Y()); 
			if ((xy < xymin) || (xy == xymin  &&  pt.Y() < ymin) || 
				(xy == xymin && pt.Y() == ymin && pt.X() < xmin)) { 
				// better point found
				retLink = link;
				xymin = xy;
				ymin = pt.Y();
				xmin = pt.X();
			}
		}
	}
	return retLink;
}


// Encode coordinates (differences) of a node with VLC + FLC;
// returns number of bits used.
uint MeshEncoder::encodeCoordinate (Point2d& vct)
{
	uint nBits = 0;
	int diff;

	for (int XorY = 0; XorY < 2; XorY++) {
		if (XorY == 0)	diff = (int) (COORD_FACTOR * vct.X());
		else			diff = (int) (COORD_FACTOR * vct.Y());
		assert (diff >= (-MAX_COORD_DIFF)  &&  diff <= MAX_COORD_DIFF);

		// compute length of difference value codeword
		uint uiAbsDiff = (uint) abs (diff);
		long len = 0;
		for (int k = MAX_COORD_DIFF_CODE_LEN; k > 0; k--)
			if (uiAbsDiff & (1 << (k - 1))) {
				len = k;
				break;
			}
		// send length symbol with a VLC
		nBits += m_entrEncSet->m_pentrencWrpPnt->encodeSymbol(len, "VecLen");
		// send difference value with a FLC
		if (diff > 0)
			m_bitStrmOut -> putBits (diff, len, "VecVal");
		else if (diff < 0)
			m_bitStrmOut -> putBits (~abs(diff), len, "VecVal");
		// don't encode difference if diff=0
		nBits += len;
	}
	// return number of bits used
	return nBits;
}


/*+==========================================================================
	NAME		MeshEncoder::encMeshMotion
	VERSION		5.0, Peter Van Beek, Thu Aug 20 10:52:58 1998
	PURPOSE		Differential encoding of motion vectors of nodes in a
				triangular mesh. A motion vector is a vector between a 
				node in a mesh and the corresponding node in another mesh.
				Uses breadth-first traversal of the triangles of a mesh. 
				This traversal is fixed throughout a sequence of IPPPP...
				planes and depends on the triangular topology of the I plane.
				Starting from a boundary triangle, we look for its neighboring
				triangles. Then for each of those triangles, we, in turn, look
				for their neighboring triangles, etc. Each time we go to another
				neighboring triangle, we may encode the motion vector of the
				node opposite the edge common to the two neighboring triangles.
				The traversal is such, that for every node that is processed,
				the other two nodes in the triangle are already processed
				before.
	METHOD		Breadth-first traversal of triangles uses a fifo queue of 
				triangles. Visited triangles are marked so as not to be 
				visited again. Nodes are marked as well, so that each node 
				is processed exactly once.
	RETURNS		The number of bits used to code the mesh motion.
	NOTES		Meshes are assumed to triangular and the two meshes
				are assumed to be in corresponcence, i.e. nodes/triangles 
				with same index correspond.
				Triangle vertices are assumed to be ordered clockwise, as
				are the pointers to their neibhoring triangles.
==========================================================================-*/
uint MeshEncoder::encMeshMotion ()
{
	TriangleQueue queue;					// FIFO triangle queue
	Tri2dRef *currTri, *leftTri, *rightTri; // triangle pointers
	int edge, rightEdge, leftEdge;			// edge numbers
	int nmn, nmp;							// number of marked nodes/polygons
	Pt2dLink  *pb1, *pb2, *paa;				// node pointers
	Point2d mv, mv1, mv2, mvPred;			// motion vectors
	uint nbits = 0;							// number of bits spent
	uint nb = 0;							// number of bits spent
	uint count = 0;							// number of nonzero vectors


	// check sizes
	int nn = meshPrev.getSizePoints();		// number of nodes
	int np = meshPrev.getSize();			// number of polygons (triangles)
	if (nn == 0	 ||	 np == 0) {
		return 0;
	}
	if (nn != meshCurr.getSizePoints()	||	np != meshCurr.getSize()) {
		doError("Meshes do not correspond ...");
		return 0;
	}

	// first store all motion vectors to nodes of previous mesh
	for (int i=0; i<nn; i++) {
		pb1 = (Pt2dLink *) meshPrev.addressC(i);
		pb2 = (Pt2dLink *) meshCurr.addressC(i);
		pb1->motion = pb2->getPoint() - pb1->getPoint();
#if defined(CHECK_MV)
		// check motion vector range and increase if necessary
		checkMV(pb1->motion);
#endif
	}
	// encode the motion range parameter
	m_bitStrmOut->putBits(m_fCode, NUMBITS_MOP_FCODE, "Fcode");
	nbits += NUMBITS_MOP_FCODE;

	// initialize node and polygon markers of the previous mesh
	meshPrev.markRefs(0);
	meshPrev.markLinks(0);

	// get the most top-left node on the boundary of the previous mesh
	getTopLeftTriangle(meshPrev, currTri, edge);
	pb1 = (Pt2dLink *) currTri->getLink(edge);
	pb2 = (Pt2dLink *) currTri->getLink((edge+1)%3);

	// encode motion vectors of nodes b1 and b2
	mv1 = pb1->motion;
	nb = 1;
	if (mv1 == (double)0.0)			// zero motion vector
		m_bitStrmOut->putBits (1, 1, "mv_flag");
	else {
		m_bitStrmOut->putBits (0, 1, "mv_flag");
		nb += encodeMV (mv1);		// send motion vector 
		count++;
	}
	nbits += nb;

	mv2 = pb2->motion;
	nb = 1;
	if (mv2 == (double)0.0)			// zero motion vector
		m_bitStrmOut->putBits (1, 1, "mv_flag");
	else {
		m_bitStrmOut->putBits (0, 1, "mv_flag");
		mv2 -= mv1;					// prediction by mv1
		nb += encodeMV (mv2);		// send prediction error
		count++;
	}
	nbits += nb;

	// initialize marking of triangles and nodes
	currTri->mark = 1;
	pb1->mark = pb2->mark = 1;
	nmp = 1;
	nmn = 2;

	// initialize the triangle queue 
	queue.append (currTri, edge);

	// now start the actual breadth-first traversal of the mesh nodes
	while (queue.empty() == 0) {
		// get and remove first triangle from the queue
		queue.remove(currTri, edge);
		leftEdge = ((edge==0)? 1: ((edge==1)? 2: 0));
		rightEdge = ((edge==0)? 2: ((edge==1)? 0: 1));
		pb1 = currTri->getLink(edge);
		pb2 = currTri->getLink(leftEdge);
		paa = currTri->getLink(rightEdge);

		// process node aa if it is not already marked processed
		if (paa->mark == 0) {
			mv = paa->motion;
			nb = 1;
			if (mv == (double)0.0)		// zero motion vector
				m_bitStrmOut->putBits (1, 1, "mv_flag");
			else {						// nonzero motion vector
				m_bitStrmOut->putBits (0, 1, "mv_flag");
				// compute mv prediction and round to half-pixel accuracy
				mvPred = (pb1->motion + pb2->motion)/2.0;
				mvPred *= MOTVEC_FACTOR;
				mvPred.roundToInt();
				mvPred /= MOTVEC_FACTOR;
				mv -= mvPred;
				nb += encodeMV(mv);		// send prediction error
				count++;
			}
			nbits += nb;
			paa->mark = 1;
			nmn++;
		}

		// if there is a neighboring unmarked triangle at the right
		// of the current triangle, mark it and queue it
		rightTri = currTri->getNeighborTri(rightEdge);
		if (rightTri != NULL && rightTri->mark == 0) {
			rightTri->mark = 1;
			queue.append(rightTri, currTri->getNeighborEdge(rightEdge));
			nmp++;
		}

		// if there is a neighboring unmarked triangle at the left
		// of the current triangle, mark it and queue it
		leftTri = currTri->getNeighborTri(leftEdge);
		if (leftTri != NULL && leftTri->mark == 0) {
			leftTri->mark = 1;
			queue.append(leftTri, currTri->getNeighborEdge(leftEdge));
			nmp++;
		}
	} // end-while

	// check for errors
	if (nmn != nn  ||  nmp != np) {
		doError("Not all nodes or triangles were processed!\n");
		return 0;
	}

	// return number of bits spent
	DPRINT(("\tEncoded %d motion vectors (%d nonzero) using: %d bits\n", nn, count, nbits));
	return nbits;
}


// check the magnitude of motion vector components -
// adapt motion vector coding range if necessary
void MeshEncoder::checkMV (Point2d &mv)
{
	Point2d mvHalfPel = MOTVEC_FACTOR*mv;
	int valid = 0;
	do {
		if (mvHalfPel.X() >= m_motionRange || mvHalfPel.X() < -m_motionRange ||
			mvHalfPel.Y() >= m_motionRange || mvHalfPel.Y() < -m_motionRange) {
			cerr << "Motion vector " << mv << " out of range ..." << endl;
			incrementMotionRange();
		}
		else
			valid = 1;
	} while (valid == 0);
}


// scale a difference motion vector component and compute residual
void MeshEncoder::scaleMV (int &iVLC, uint &uiResidual, int iDiffMVcomponent)
{
	assert (iDiffMVcomponent >= (-3 * (int) m_motionRange));
	assert (iDiffMVcomponent < 3 * (int) m_motionRange);
	if (iDiffMVcomponent < (-1 * (int) m_motionRange))
		iDiffMVcomponent += 2 * m_motionRange;
	else if (iDiffMVcomponent > (int) (m_motionRange - 1))
		iDiffMVcomponent -= 2 * m_motionRange;

	if (iDiffMVcomponent == 0) {
		iVLC = 0;				// nothing to do
		uiResidual = 0;
	}
	else {
		if (m_scaleFactor == 1) {
			iVLC = iDiffMVcomponent; // simple no-scale case
			uiResidual = 0;
		}
		else {					// scaling
			uint uiAbsDiffMVcomponent = abs (iDiffMVcomponent);
			uiResidual = (uiAbsDiffMVcomponent - 1) % m_scaleFactor;
			uint uiVLCmagnitude = 
				(uiAbsDiffMVcomponent - uiResidual + (m_scaleFactor - 1)) / m_scaleFactor;
			iVLC = uiVLCmagnitude * sign (iDiffMVcomponent);
		}
	}
}


// send a difference motion vector
uint MeshEncoder::encodeMV (Point2d &vctDiffMV)
{
	int iVLC;
	uint uiResidual;
	uint nBits = 0;

	// send the bits for the x component
	int iDiffMv = (int) (vctDiffMV.X() * MOTVEC_FACTOR);
	scaleMV (iVLC, uiResidual, iDiffMv);
	long lsymbol = iVLC + 32;
	nBits += m_entrEncSet->m_pentrencMV->encodeSymbol (lsymbol, "MV_Mag");
	if (iVLC != 0) {
		m_bitStrmOut->putBits ((char) uiResidual, m_fCode - 1, "MV_Res");
		nBits += m_fCode - 1;
	}
	else
		assert (uiResidual == 0);

	// send the bits for the y component
	iDiffMv = (int) (vctDiffMV.Y() * MOTVEC_FACTOR);
	scaleMV (iVLC, uiResidual, iDiffMv);
	lsymbol = iVLC + 32;
	nBits += m_entrEncSet->m_pentrencMV->encodeSymbol(lsymbol, "MV_Mag");
	if (iVLC != 0) {
		m_bitStrmOut->putBits ((char) uiResidual, m_fCode - 1, "MV_Res");
		nBits += m_fCode - 1;
	}
	else
		assert (uiResidual == 0);

	// return number of bits used
	return nBits;
}


// Get the top left node on the boundary of a mesh and the triangle
// that contains it
void MeshEncoder::getTopLeftTriangle(Tri2dTess &mesh, Tri2dRef *&tref, int &edge)
{
	assert(m_initTri != -1	&&	m_initEdge != -1);
	tref = mesh.addressR(m_initTri);
	edge = m_initEdge;
}

// Find the top left node on the boundary of a mesh and the triangle
// that contains it by searching for the triangle vertex (node) with the
// minimum value of x+y (using y value to break ties)
void MeshEncoder::setTopLeftTriangle(Tri2dTess &mesh)
{
	Tri2dRef *tmpRef = mesh.nextTriangle(1);
	double pxymin = MAXDOUBLE;
	double pymin  = MAXDOUBLE; 
	// check each triangle
	for (int i=0; i<mesh.getSize(); i++) {
		// check each edge/node point
		for (int j=0; j<3; j++) {
			if (tmpRef->getNeighborTri(j) == NULL) {
				// edge j,j+1 this is a boundary edge 
				Point2d pt = tmpRef->getLink(j)->getPoint();
				double pxy = pt.X()+pt.Y();
				// find node with minimum x+y; y value breaks ties
				if ((pxy < pxymin) ||	
					(pxy == pxymin	&&	pt.Y() < pymin)) {	
					pxymin = pxy;
					pymin = pt.Y();
					// store index of triangle and edge
					m_initTri = i;
					m_initEdge = j;
				}
			}
		} // endfor edges
		tmpRef = mesh.nextTriangle();
	} // endfor triangles
}


// Find the bottom right node of a uniform rectangular mesh by searching
// for the node with maximum x+y value
void MeshEncoder::getUniformBottomRight(Tri2dTess &mesh, Point2d &ptBr)
{
	double xymax = -(MAXDOUBLE);
	double xy;
	for (int n=0; n<mesh.getSizePoints(); n++) {
		Point2d pt = mesh.point(n);
		if ((xy = pt.X()+pt.Y()) > xymax) {
			xymax = xy;
			ptBr = pt;
		}
	}
}


// put some stuffing bits into the bitstream until we are byte-aligned;
// note: puts at least one 0; then some 1's if necessary
uint MeshEncoder::nextStartCode()
{
	register uint nbits = m_bitStrmOut->getCounter();
	register uint emptybits = 8-nbits%8;

	if (emptybits != 8) {
		m_bitStrmOut -> putBits ((int)0, (uint)1);
		if (emptybits > 1)
			m_bitStrmOut -> putBits ((int) 0xFFFF, emptybits - 1);
		return emptybits;
	}
	else {				// emptybits == 8
		m_bitStrmOut -> putBits ((int) 0x007F, 8);
		return (uint) 8;
	}
}


// Header information (timing) class constructor
TimeHeader::TimeHeader()
{
	useStartCode = 1;
	isRate = isTime = isSkip = 0;
	frameRateBase = DEF_FRAME_RATE;
	frameRateFrac = frameRateOffset = 0;
	timeHours = DEF_TIME_HOUR;
	timeMinutes = timeSeconds = 0;
	frameSkip = DEF_FRAME_SKIP;
	timeStamp = nextTimeStamp = 0.0;
}

// Header information (timing) class destructor
TimeHeader::~TimeHeader()
{
}

// Update current and next time-stamps 
void TimeHeader::updateTime(int isIntra)
{
	// compute presentation time stamp in milliseconds
	if (isIntra && isTime)
		timeStamp = (3600*timeHours + 60*timeMinutes + timeSeconds)*1000.0;
	else
		timeStamp = nextTimeStamp;

	// compute next presentation time stamp in milliseconds
	nextTimeStamp = timeStamp + 
		(double)((frameSkip+1)*((1000 + frameRateOffset) << 4)) /
		(double)((frameRateBase << 4) + frameRateFrac);
}

// Print header info
void TimeHeader::print()
{
	cout << "useStartCode = " << useStartCode << endl;
	cout << "isRate = " << isRate << endl;
	cout << "isTime = " << isTime << endl;
	cout << "isSkip = " << isSkip << endl;
	cout << "frameRateBase = " << frameRateBase << endl;
	cout << "frameRateFrac = " << frameRateFrac << endl;
	cout << "frameRateOffset = " << frameRateOffset << endl;
	cout << "timeHours = " << timeHours << endl;
	cout << "timeMinutes = " << timeMinutes << endl;
	cout << "timeSeconds = " << timeSeconds << endl;
	cout << "frameSkip = " << frameSkip << endl;
	cout << "timeStamp = " << timeStamp << endl;
	cout << "nextTimeStamp = " << nextTimeStamp << endl;
}

#if defined(CONFORMANCE_TEST)
// Randomize the value of useStartCode; used only for generating
// conformance test bitstreams!
void TimeHeader::randomizeUseStartCode()
{
	// assume seed == 1; test last bit of random number
	if (rand() & 1) {
		useStartCode = 1;
		cout << "\tUsing start-code ..." << endl;
	}
	else
		useStartCode = 0;
}

// Tweak or randomize MOP-header coding parameters; used only for
// generating conformance test bitstreams; and only in intra MOPs!
void TimeHeader::tweakTimeParams(int t)
{
	if (t < 0  ||  t >= scheduleLength) {
		cerr << "Event number too high for schedule ..." << endl;
		return;
	}
	isTime = isTimeSchedule[t];
	isRate = isRateSchedule[t];
	isSkip = isSkipSchedule[t];
	timeHours = tHoursSchedule[t];
	timeMinutes = tMinutesSchedule[t];
	timeSeconds = tSecondsSchedule[t];
	frameSkip = fSkipSchedule[t];
	frameRateBase = frBaseSchedule[t];
	frameRateFrac = frFracSchedule[t];
	frameRateOffset = frOffsetSchedule[t];
}

// Tweak mask to skip frames periodically (to lower bit rate)
void TimeHeader::tweakMask(int &meshMask, int f)
{
	if (f >= 4 && (((f-4)%6 == 0) || ((f-5)%6 == 0)))
		meshMask = 0;
	else
		meshMask = 1;
}
#endif


// Encoder parameter class constructor
EncParam::EncParam()
{
	meshType = 0;
	firstFrame = lastFrame = 0;
	nFrame = 1;
	frameInc = 1;
	nGop = 0;
	gopSize = NULL;
	nLoop = 1;
	motionRange = 16;
	isRate = 0;
	frameRateBase = DEF_FRAME_RATE;
	frameRateFrac = frameRateOffset = 0;
	pline = 0;
}

// EncParam destructor
EncParam::~EncParam()
{
	if (gopSize != NULL)
		delete [] gopSize;
}

// Read the parameter file with the name 'fname' and set coding
// parameters according to the user inputfile
void EncParam::readParams(char *fname)
{
	char  fileName[200];				// file name
	char  dirName[200];					// directory name


	if ((fpara = fopen (fname, "r")) == NULL ){
		cerr << "parameter file not found... " << endl;
		exit (1);
	}
	DPRINT(("Reading parameter file...\n"));

	// original mappings input file name
	nextLine(); fscanf (fpara, "%s", fileName);
	// original input file directory
	nextLine(); fscanf (fpara, "%100s", dirName);

	sprintf(meshDatFile, "%s/%s", dirName, fileName);
	DPRINT(("Basename mesh-files    : %s\n", meshDatFile));

	// mesh type (uniform or object-based)
	nextLine(); fscanf (fpara, "%u", &meshType);
	assert (meshType == 0 || meshType == 1);

	// first and last frame number, and temporal sampling rate
	nextLine(); fscanf (fpara, "%u", &firstFrame);
	nextLine(); fscanf (fpara, "%u", &nFrame);
	nextLine(); fscanf (fpara, "%d", &frameInc);
	assert (firstFrame >= 0 && nFrame >= 1 && frameInc >= 1);

	lastFrame = firstFrame + (nFrame-1)*frameInc;

	DPRINT(("First --- last frame   : %d --- %d\n", firstFrame, lastFrame));
	DPRINT(("Frame increment value  : %d\n", frameInc));

	// number of group-of-mesh-object-planes
	nextLine(); fscanf (fpara, "%u", &nGop);
	assert (nGop >= 1);
	DPRINT(("Number of GOPs         : %d\n", nGop));
	gopSize = new int [nGop];
	// number of planes in each group-of-mesh-object-planes
	nextLine(); 
	for (int g = 0; g < nGop; g++) {
		int size;
		fscanf (fpara, "%d", &size);
		assert (size >= 1);
		gopSize[g] = size;
	}
		
	// number of (forward/backward) loops over the sequence
	nextLine(); fscanf (fpara, "%d", &nLoop);
	assert (nLoop >= 1);

	// output file name
	nextLine(); fscanf (fpara, "%s", fileName);
	// output file directory
	nextLine(); fscanf (fpara, "%100s", dirName);

	sprintf(bitStreamFile, "%s/%s.bits", dirName, fileName);
	DPRINT(("Output bitstream file  : %s\n", bitStreamFile));

	// motion vector range
	nextLine(); fscanf (fpara, "%u", &motionRange);
	assert (motionRange >= 1);
	DPRINT(("Motion range           : %d\n", motionRange));

	// frame rate information
	nextLine(); fscanf (fpara, "%u", &isRate);
	assert (isRate == 0 || isRate == 1);
	if (isRate == 1) {
		fscanf (fpara, "%u", &frameRateBase);
		assert (frameRateBase > 0);
		fscanf (fpara, "%u", &frameRateFrac);
		assert (frameRateFrac >= 0);
		fscanf (fpara, "%u", &frameRateOffset);
		assert (frameRateOffset >= 0);
	}

	// that's all
	DPRINT(("\n\n"));
	fclose(fpara);
}

// Reads comment lines until the next non-comment-line in a file and
// keeps track of the line number
void EncParam::nextLine()	
{
	uchar firstchar;
	char line[200];

	fgets (line, 200, fpara);		// get the next line 
	pline++;
									// skip comment lines
	while (feof (fpara) == 0 && (firstchar = fgetc(fpara)) == '%') {
		fgets (line, 200, fpara);	// get the next line 
		pline++;
	}
	ungetc (firstchar, fpara);		// put back first char
}


////////// MAIN FUNCTION /////////////////////////////////////////////////////


// This is it, folks!
int main (int argc, char *argv[])
{
	// Fullname of input file with mesh data
	char fileName[200];
	// Overall number of bits spent on mesh geometry and motion
	uint nMeshBits = 0;
	// Overall number of overhead bits spent on start/end codes, etc
	uint nOverhead = 0;
	// Number of frames encoded
	int frameCount = 0;
	// Input frame number and loop counters
	int	 fnum, f, l, g;
	// Encoding parameters
	EncParam par;
	// Mask value;
	int meshMask = 1;
	// Timing value
	double startTime;


	// check parameter file argument
	if (argc < 2) {
		cerr << "ISO/IEC 14496 (MPEG-4) 2D Mesh Object encoder" << endl;
		cerr << "Usage: " << argv[0] << " <parameter-file>" << endl;
		exit (1);
	}

	time_t	 timeval = time(NULL);
	printf("========================================================\n");
	printf("ISO/IEC 14496 (MPEG-4) 2D Mesh Object encoder ---- v.5.0\n");
	printf("%s\n", ctime(&timeval));
	printf("========================================================\n\n\n");

	// read encoding parameters from ascii file
	par.readParams(argv[1]);

	// create and initialize encoder object
	MeshEncoder encoder(par);

#if defined(TIMER)
	// run-time logging
#if defined(WIN32)
	struct _timeb tstart, tstop;
	_ftime(&tstart);
#else
	struct timeval tstart,tstop;
	gettimeofday(&tstart, (struct timezone *)NULL);
#endif
#endif

	printf("========================================================\n");

	// encode Mesh Object Start code
	nOverhead += encoder.encodeMoStart();

	/////////////// GOP LOOP //////////////////////////////////////////

//	fnum = par.firstFrame;
	int gopStartFrame = par.firstFrame;

	for (g = 0; g < par.nGop; g++) {

		printf("\n------- GOP %3d --------------------------------\n", g);

#if defined(CONFORMANCE_TEST) && defined(TWEAK_HEADER)
		fnum = par.firstFrame;			// reset
#else
		fnum = gopStartFrame;
#endif
		// read mesh data of intra (reference) frame
		sprintf(fileName, "%s%d.tri", par.meshDatFile, fnum);
		encoder.meshCurr.read(fileName);
		// quantize the mesh node positions to half pixel accuracy
		encoder.meshCurr.quantize(0.5);

		// find the indices of the top-left triangle and edge
		encoder.setTopLeftTriangle(encoder.meshCurr);

		printf("------- Encoding mesh geometry of frame %3d ----\n", fnum);

		// update time information and frame count
#if defined(CONFORMANCE_TEST) && defined(TWEAK_HEADER)
		encoder.hInfo.tweakTimeParams(g);
#endif
		encoder.hInfo.updateTime(1);
		if (g == 0) startTime = encoder.hInfo.timeStamp;
		frameCount++;

		// encode Mesh Object Plane Header (intra MOP)
		nOverhead += encoder.encodeMopStart();
		nOverhead += encoder.encodeMopHeader(1, 1);

		// encode Mesh Object Plane Data (intra MOP)
		nMeshBits += encoder.encodeMopData(1, 1);

		// put some stuffing bits and byte-align for the next start-code
		nOverhead += encoder.nextStartCode();

		/////////////// INTER FRAME LOOP //////////////////////////////////

		// loop over the group-of-planes a number of times
		for (l = 0; l < par.nLoop; l++) 
			// from the first to the last plane of this group
			for (f = 1; f < par.gopSize[g]; f++) {

				// copy previous mesh 
				encoder.meshPrev = encoder.meshCurr;

				// compute input frame number
				if (l%2 == 0)				// forward loop
					fnum += par.frameInc;
				else						// backward loop
					fnum -= par.frameInc;

				// read mesh data of current prediction frame
				sprintf(fileName, "%s%d.tri", par.meshDatFile, fnum);
				encoder.meshCurr.read(fileName);
				// quantize the mesh node positions to half pixel accuracy
				encoder.meshCurr.quantize(0.5);

				printf("\n------- Encoding motion to frame %3d -------------------\n", fnum);

				// update time information and frame count
#if defined(CONFORMANCE_TEST)
#if defined(TWEAK_STARTCODE)
				encoder.hInfo.randomizeUseStartCode();
#else if defined(TWEAK_MASK)
				encoder.hInfo.tweakMask(meshMask, f);
#endif
#endif
				encoder.hInfo.updateTime(0);
				frameCount++;

				// encode Mesh Object Plane Header (inter MOP)
				if (encoder.hInfo.useStartCode)
					nOverhead += encoder.encodeMopStart();
				nOverhead += encoder.encodeMopHeader(meshMask, 0);

				// encode Mesh Object Plane Data (inter MOP) 
				nMeshBits += encoder.encodeMopData(meshMask, 0);

				// put some stuffing bits and byte-align
				if (encoder.hInfo.useStartCode)
					nOverhead += encoder.nextStartCode();

				// print number of bits spent
				DPRINT(("\tTotal bits used: %d ...\n", nMeshBits+nOverhead));

			} // endfor (inter frame loop)

		gopStartFrame += par.gopSize[g];

	} // endfor (group-of-plane loop)

	// encode Mesh Object end-code
	nOverhead += encoder.encodeMoEnd();

	////////// report results and return ///////////////////////////////////

	uint nTotalBits = nMeshBits+nOverhead;
	encoder.hInfo.updateTime(0);
	double timeElapsed = encoder.hInfo.timeStamp - startTime;

	printf("--------------------------------------------------------\n\n");
	printf("\tMesh bits     : %d\n", nMeshBits);
	printf("\tOverhead bits : %d\n", nOverhead);
	printf("\tTotal bits    : %d\n", nTotalBits);
	printf("\tTime elapsed  : %.2f msec\n", timeElapsed);
	printf("\tBit rate      ~ %.2f kbit/sec \n", nTotalBits/timeElapsed);

#if defined(TIMER)
	// compute and print running time in hundredths of a second
#if defined(WIN32)
	_ftime(&tstop);
	int runtime = 100*(tstop.time-tstart.time) +
		(tstop.millitm-tstart.millitm)/10;
#else
	gettimeofday(&tstop, (struct timezone *)NULL);
	int runtime = 100*(tstop.tv_sec-tstart.tv_sec)
		+ (tstop.tv_usec-tstart.tv_usec)/10000;
#endif
	if (runtime <= 0) runtime = 1;
	printf("\n\tRun-time info : %d.%02d seconds, %d frames, %d.%02d fps\n",
		   runtime/100, runtime%100, frameCount, 
		   ((10000*frameCount+runtime/2)/runtime)/100,
		   ((10000*frameCount+runtime/2)/runtime)%100);
#endif
	printf("\n========================================================\n\n");
	return 0;

} // main



// end of file
