/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* The University of Rochester retains full right to use the code for */
/* their own purposes, assign or donate the code to a third party and */
/* to inhibit third parties from using the code for non MPEG-4 Visual */
/* (ISO/IEC 14496-2) conforming products.                             */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1997-1999                                            */
/*                                                                    */
/**********************************************************************/

/*+***********************************************************-*-C++-*-*
   NAME         mesh_encode.h
   AUTHOR/DATE  Peter van Beek, Wed Aug 19 13:03:22 1998
   DESCRIPTION  Mesh encoder definitions.
   SYNOPSIS     
*********************************************************************-*/

#ifndef	MESH_ENCODER_DEF
#define	MESH_ENCODER_DEF

#include <iostream.h>
#include <fstream.h>
#include "pointClass.h"					// Point2d
#include "linkClass.h"					// Pt2dLink
#include "triRefClass.h"				// Tri2dRef
#include "triTessClass.h"				// Tri2dTess


/////////////// CLASS DEFINITIONS ///////////////////////////////////////////

// forward class definitions
class EncParam;
class COutBitStream;
class CEntropyEncoderSet;


// Header information (timing) class definition
class TimeHeader
{
  public:
	// Constructor
	TimeHeader();
	// Desctructor
	~TimeHeader();

	// Boolean that flags use of mesh object plane start code
	int useStartCode;
	// Booleans for switching on/off timing info
	int isRate, isTime, isSkip;
	// Frame rate values
	uint frameRateBase, frameRateFrac, frameRateOffset;
	// Hour part of time code value
	uint timeHours, timeMinutes, timeSeconds;
	// Skip frames value
	int frameSkip;
	// Time stamps
	double timeStamp, nextTimeStamp;

	// Update current and next time stamp
	void updateTime(int isIntra);
	// Print info
	void print();

	// For generating conformance test bitstreams
	void randomizeUseStartCode();
	void tweakTimeParams(int t);
	void tweakMask(int &meshMask, int f);
};


// MeshEncoder class definition
#ifndef WIN32
class MeshEncoder
#else
class __declspec(dllexport) MeshEncoder
#endif
{
  public:
	// Constructor 
	MeshEncoder(const EncParam &par);

	// Destructor
	~MeshEncoder();

	// Increase motion vector range
	void incrementMotionRange();
	// Encode Mesh Object start-code
	uint encodeMoStart();
	// Encode Mesh Object Plane start-code
	uint encodeMopStart();
	// Encode Mesh Object end-code
	uint encodeMoEnd();
	// Encode Mesh Object Plane Header
	uint encodeMopHeader (const int meshMask, const int isIntra);
	// Encode Mesh Object Plane Data
	uint encodeMopData(const int meshMask, const int isIntra);
	// Encode geometry and boundary of Uniform-type mesh (I-MOP)
	uint encUniformMesh ();
	// Encode geometry and boundary of Delaunay-type mesh (I-MOP)
	uint encDelaunayMesh ();
	// Encode motion of previous to current mesh (P-MOP)
	uint encMeshMotion ();
	// Encode stuffing bits 
	uint nextStartCode();
	// Find indices of initial triangle and edge with top-left node
	void setTopLeftTriangle(Tri2dTess &mesh);
	// Get indices of initial triangle and edge with top-left node
	void getTopLeftTriangle(Tri2dTess &mesh, Tri2dRef *&tref, int &edge);

	// The previous and current mesh
	Tri2dTess meshPrev, meshCurr;
	// Header (timing) information
	TimeHeader hInfo;

  private:

	// Encode a mesh node point coordinate (difference)
	uint encodeCoordinate(Point2d &pt);
	// Check the magnitude of a motion vector and adapt coding range
	void checkMV (Point2d &mv);
	// Scale a difference motion vector component
	void scaleMV (int &iVLC, uint &uiResidual, int iDiffMVcomponent);
	// Encode a mesh node motion vector (difference)
	uint encodeMV (Point2d &mv);
	// Get bottom-right node of uniform mesh
	void getUniformBottomRight(Tri2dTess &mesh, Point2d &ptBr);
	// Find nearest not marked node
	Pt2dLink *getNearestNode(Tri2dTess &mesh, Point2d &ptlast);
	// Get next boundary triangle and edge
	void getNextBounTriangle(Tri2dRef *&tri, int &edge);
	// Compute mesh boundary attributes
	int getBounData(Tri2dRef *tref, const int edge);

	// Output file, bitstream and Huffman entropy encoders
	ofstream m_outStrm;
	COutBitStream* m_bitStrmOut;
	CEntropyEncoderSet* m_entrEncSet;

	// Type of mesh : 0 for uniform ; 1 for object-based (Delaunay)
	uint m_meshType;
	// Motion vector range code (Fcode): 1, 2, ..., 7
	uint m_fCode; 
	// Motion vector scale factor
	uint m_scaleFactor;
	// Motion range in half pixel units: 32, 64, ..., 2048
	int  m_motionRange;

	// Mesh bounding box attributes
	double m_xmin, m_xmax, m_ymin, m_ymax;

	// index of initial triangle for breadth-first-traversal
	int m_initTri;
	// number of initial vertex/edge on initial triangle
	int m_initEdge;
};


// Encoder parameters class definition
class EncParam
{
  public:
	// Constructor 
	EncParam();
	// Destructor
	~EncParam();
	// Read encoder parameters from an ascii file
	void readParams(char *fileName);
	void nextLine();

	// Base of filename with mesh data, without numbers or extension
	char meshDatFile[200];
	// Full filename of output bitstream file
	char bitStreamFile[200];
	// Type of initial mesh : 0 for uniform ; 1 for object-based (Delaunay)
	int meshType;
	// First and last frames of the sequence to be processed
	int firstFrame, lastFrame;
	// Total number of mesh object planes
	int nFrame;
	// Frame increment determining temporal frame rate
	int frameInc;
	// Number of group-of-mesh-object-planes
	int nGop;
	// Number of planes in each group-of-mesh-object-plane
	int *gopSize;
	// Number of (forward/backward) loops over the sequence of frames
	int nLoop;
	// Motion vector range in pixel accuracy (up to 1024) 
	int motionRange;
	// Booleans for switching on/off timing info
	int isRate;
	// Frame rate values
	uint frameRateBase, frameRateFrac, frameRateOffset;

  private:
	// parameter file
	FILE *fpara;
	// line number in parameter file
	uint  pline;
};


#endif // MESH_ENCODER_DEF

// End of file
