/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* A. Tanju Erdem (Eastman Kodak Company)                             */
/* Candemir Toklu (University of Rochester)                           */
/*                                                                    */
/* and edited by                                                      */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* Eastman Kodak Company and the University of Rochester retain full  */
/* right to use the code for their own purposes, assign or donate the */
/* code to a third party and to inhibit third parties from using the  */
/* code for non MPEG-4 Visual (ISO/IEC 14496-2) conforming products.  */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1994-1998                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         pointClass.h
   AUTHOR       A. Tanju Erdem - Eastman Kodak Company and Candemir Toklu - 
                University of Rochester.
   DATE         September 1994
   UPDATE       14 August 1996 by Candemir Toklu
                - general
                Peter Van Beek, Thu Aug 29 18:16:25 1996
                - Added stuff to set/access u/v/z values
                Peter Van Beek, Tue Oct 29 09:22:39 1996
                - Added moveto() functions
                Peter Van Beek, Fri Jun 12 12:01:53 1998
                - inlined constructors 
   DESCRIPTION  The Point2d class forms the basis for the polygon and mesh 
                tools. The class has five private members two for the 2-D 
                location, and three for possible values associated with the 
                2-D location which could be Cr, Cb, Y; or R, B, G.
   SYNOPSIS     Include files: 
                #include <iostream.h>
                #include <fstream.h>
                #include <math.h>
                #include <time.h>
*********************************************************************-*/

#include <iostream.h>
#include <fstream.h>
#include <math.h>
#include <time.h>

#ifndef POINT_CLASS
#define POINT_CLASS

class Poly2d;
class Quad2d;

#ifndef PI
#define PI 3.14159265
#endif

#define	 EPSILON 1.0e-5
#define	 ONEPLUSEPS 1.00001
#define	 ONEMINUSEPS 0.99999

// A point (i.e., Point2d) has 2 coordinates and up to 3 sample
// values associated with the location of the point.
#ifndef WIN32
class Point2d {
#else
class __declspec(dllexport) Point2d {
#endif
	friend Point2d operator*( const Point2d&, double);	// Scaling
	friend Point2d operator*( double, const Point2d&);	// Scaling
	friend Point2d operator/( const Point2d&, double);	// Scaling
	friend ostream& operator<<( ostream&, const Point2d&);
	friend istream& operator>>( istream&, Point2d&);
  public:
	inline Point2d(); // Initialization
	inline Point2d( double, double); // Initialization
	inline Point2d( const Point2d&); // Initialization
//	  ~Point2d();				  // Default destructor
	inline Point2d& operator=( const Point2d&);
	inline Point2d& operator=( double);
	inline Point2d& operator+=( const Point2d&);
	inline Point2d& operator-=( const Point2d&);
	inline Point2d& operator*=( double);  // Scaling
	inline Point2d& operator/=( double);  // Scaling
	inline Point2d operator+( const Point2d&);	// Addition
	inline Point2d operator-( const Point2d&);	//Subtraction
	inline double operator*( const Point2d&);  // Dot product
	inline double operator%( const Point2d&);  // Normalized cross prod. 
											   // (can be neg.)
	inline double operator^( const Point2d&);  // Cross product value 
											   // (can be negative)
	inline int operator||( const Point2d&);	 // Check parallel.
	inline int operator==( const Point2d&);
	inline int operator==( double);
	inline int operator!=( const Point2d&);
	inline int operator!=( double);
	inline double normSqr();  // The magnitude-squared of a point
	inline double norm();  // The magnitude of a point
	Point2d unit();	 // Return a unit vector in the direction of the point
	double distanceSqrTo( Point2d);	 // Distance-squared
	double distanceTo( Point2d);  // Distance
	int isInside( Poly2d);	// Inside a polygon?
	inline Point2d& move( double, double);
	inline Point2d& move( const Point2d&);
	inline Point2d& scale( double, double);
	inline Point2d& scale( double);
	inline Point2d& roundToInt();
	inline double X();								// return x
	inline double Y();								// return y
	inline double U();
	inline double V();
	inline double Z();
  private:
	double x, y;	 // 2-D location
	double u, v, z;	 // Possible values associated with the 2-D location;
					 // they could be Cr, Cb, Y; or R, B, G.
};

//------------------------------------------------------------------------
// Constructors for class Point2d

Point2d::Point2d()	// Initialization
{
	x = y = 0.0;
	u = v = z = 0.0;
}

Point2d::Point2d( double a, double b)  // Initialization
{
	x = a;
	y = b;
	u = 0.0;
	v = 0.0;
	z = 0.0;
}

Point2d::Point2d( const Point2d& p)	 // Initialization
{
	x = p.x;
	y = p.y;
	u = 0.0;
	v = 0.0;
	z = 0.0;
}

Point2d& Point2d::operator=( const Point2d& p)
{
	x = p.x;
	y = p.y;
	u = p.u;
	v = p.v;
	z = p.z;
	return *this;
}

Point2d& Point2d::operator=( double d)
{
	x = d;
	y = d;
	u = 0.0;
	v = 0.0;
	z = 0.0;
	return *this;
}

//------------------------------------------------------------------------
// Arithmetic operators for class Point2d

Point2d& Point2d::operator+=( const Point2d& p)
{
	x += p.x;
	y += p.y;
	return *this;
}

Point2d& Point2d::operator-=( const Point2d& p)
{
	x -= p.x;
	y -= p.y;
	return *this;
}

Point2d& Point2d::operator*=( double a)
{
	x *= a;
	y *= a;
	return *this;
}

Point2d& Point2d::operator/=( double a)
{
	x /= a;
	y /= a;
	return *this;
}

Point2d Point2d::operator+( const Point2d& p)  // Addition
{
	return Point2d( x + p.x, y + p.y);
}

Point2d Point2d::operator-( const Point2d& p)  // Subtraction
{
	return Point2d( x - p.x, y - p.y);
}

double Point2d::operator*( const Point2d& p)  // Dot product
{
	return ( x * p.x + y * p.y);
}

// Normalized cross product (can be negative)
double Point2d::operator%( const Point2d& p)
{
	double a = x * p.y, b = y * p.x;
	return ( (a - b) / (fabs(a) + fabs(b) + 1.0));
}

// Cross product value (can be negative)
double Point2d::operator^( const Point2d& p)
{
	return ( x * p.y - y * p.x);
}

// Check parallel
int Point2d::operator||( const Point2d& p)	
{
	if(fabs(*this^p) < EPSILON)
		return 1;
	else
		return 0;
}

int Point2d::operator==( const Point2d& p)
{
	return( fabs(x - p.x) < EPSILON && fabs(y - p.y) < EPSILON);
}

int Point2d::operator==( double a)
{
	return( fabs(x - a) < EPSILON && fabs(y - a) < EPSILON);
}

int Point2d::operator!=( const Point2d& p)
{
	return( fabs(x - p.x) >= EPSILON || fabs(y - p.y) >= EPSILON);
}

int Point2d::operator!=( double a)
{
	return( fabs(x - a) >= EPSILON || fabs(y - a) >= EPSILON);
}

//--------------------------------------------------------------------------
// Norm-squared of vector point
//--------------------------------------------------------------------------
double Point2d::normSqr()
{
	return (double) (x*x + y*y);
}

//--------------------------------------------------------------------------
// Norm of a vector point
//--------------------------------------------------------------------------
double Point2d::norm()
{
	return sqrt( normSqr());
}

Point2d& Point2d::move( double a, double b)
{
	x += a;
	y += b;
	return *this;
}

Point2d& Point2d::move( const Point2d& p)
{
	x += p.x;
	y += p.y;
	return *this;
}

Point2d& Point2d::scale( double a, double b)
{
	x *= a;
	y *= b;
	return *this;
}

Point2d& Point2d::scale( double a)
{
	x *= a;
	y *= a;
	return *this;
}

Point2d& Point2d::roundToInt()
{
	x = floor( x + 0.5);
	y = floor( y + 0.5);
	return *this;
}

double Point2d::X()
{
	return x;
}

double Point2d::Y()
{
	return y;
}

double Point2d::U()
{
	return u;
}

double Point2d::V()
{
	return v;
}

double Point2d::Z()
{
	return z;
}

#endif
