/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* A. Tanju Erdem (Eastman Kodak Company)                             */
/* Candemir Toklu (University of Rochester)                           */
/*                                                                    */
/* and edited by                                                      */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* Eastman Kodak Company and the University of Rochester retain full  */
/* right to use the code for their own purposes, assign or donate the */
/* code to a third party and to inhibit third parties from using the  */
/* code for non MPEG-4 Visual (ISO/IEC 14496-2) conforming products.  */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1994-1998                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         polyClass.cpp
   AUTHOR       A. Tanju Erdem - Eastman Kodak Company and Candemir Toklu - 
                University of Rochester.
   DATE         September 1994
   UPDATE       14 August 1996 by Candemir Toklu
                - various
                Peter Van Beek, Fri May  1 16:08:26 1998 
                - added quantize method
   DESCRIPTION  Member functions for Poly2d class are defined in this
                file.
   SYNOPSIS     Include files: 
                #include "general.h"
                #include "polyClass.h"
*********************************************************************-*/
#include "general.h"
#include "polyClass.h"


//------------------------------------------------------------------------

Poly2d::Poly2d()  // Empty polygon
{
	poly = (Point2d *) NULL;
	size = 0;
}

Poly2d::Poly2d( int s)	// A polygon with "s" sides. All points are set to 0.
{
	poly = new Point2d[s];
	size = s;
	for( int i = 0; i < size; i++) poly[i] = point(0,0);
}

Poly2d::Poly2d(const Poly2d& from)	// Initialization with another polygon
{
	size = from.size;
	if( size == 0) {
		poly = (Point2d *) NULL;
	}
	else {
		poly = new Point2d[ size];
		for( int i = 0; i < size; i++) poly[i] = from.poly[i];
	}
}

// Initialization with another sub-polygon
Poly2d::Poly2d(const Poly2d& from, int n, int m)  
{
	//	size = Mod(m-n+from.size+1, from.size);
	size = (m-n+1)%from.size;
	if (size <= 0) size+=from.size;
	if (size == 0)
		poly = (Point2d *) NULL;
	else {
		poly = new Point2d[size];
		for (int i = 0; i < size; i++) 
			poly[i] = from.poly[Mod(n+i+from.size, from.size)];
	}
}

Poly2d::Poly2d( const Point2d& p)  // A polygon with one side. Vertex set to p.
{
	poly = new Point2d[1];
	size = 1;
	poly[0] = p;
}

Poly2d::~Poly2d()
{
	delete[] poly;
}

Point2d& Poly2d::operator[]( int indx)	// Indexing operator
{
	if( size == 0) {
		cout << "\nError: Cannot index an empty polygon." << endl;
		exit(1);
	}
	else {
		return poly[ Mod( indx, size)];	 // Indexing is modular.
	}
}

Poly2d& Poly2d::operator=(const Poly2d& from)  // Setting equal to a polygon
{
	if( this == &from) {
		cout << "\n	 Warning: Copy request rejected. "
			 << "Cannot copy a polygon onto itself" << endl;
	}
	else {
		if( size != from.size) {
			delete[] poly; // Destruct the old polygon
			size = from.size;
			poly = new Point2d[ size];  // Construct a new polygon
		}
		for( int i = 0; i < size; i++) poly[i] = from.poly[i];
	}
	return *this;
}

Poly2d& Poly2d::empty()
{
	if( size != 0) {
		delete[] poly;
		poly = (Point2d *) NULL;
		size = 0;
	}
	return *this;
}

Poly2d& Poly2d::replace( int indx, const Point2d& p)
{
	if( size == 0) {
		cout << "\nWarning: Cannot replace vertex in empty polygon." << endl;
	}
	else {
		poly[ Mod( indx, size)] = p;
	}
	return *this;
}

Poly2d& Poly2d::add( int indx, const Point2d& p)  // Add a new vertex at indx
{
	if( indx < 0 || indx > size) {
		cout << "\nWarning: Cannot add to the polygon at the specified vertex." << endl;
		return *this;
	}

	int tmpSize = size + 1;
	Point2d *tmpPoly;
	int i;

	// Construct a new polygon that has one more point.
	tmpPoly = new Point2d[ tmpSize];
	for( i = 0; i < indx; i++) tmpPoly[i] = poly[i];
	tmpPoly[ indx] = p;
	for( i = indx; i < size; i++) tmpPoly[ i + 1] = poly[i];

	if( size != 0) {
		// Destruct the old polygon
		delete[] poly;
	}

	poly = tmpPoly;
	size = tmpSize;

	return *this;
}

Poly2d& Poly2d::insert( const Point2d& p)
{
	add( 0, p);
	return *this;
}

Poly2d& Poly2d::append( const Point2d& p)
{
	add( size, p);
	return *this;
}

Poly2d& Poly2d::move( double a, double b)
{
	for( int i = 0; i < size; i++) {
		poly[i].move( a, b);
	}
	return *this;
}

Poly2d& Poly2d::move( const Point2d& p)
{
	return *this += p;
}

Poly2d& Poly2d::scale( double a, double b)
{
	for( int i = 0; i < size; i++) {
		poly[i].scale(a,b);
	}
	return *this;
}

Poly2d& Poly2d::scale( double a)
{
	for( int i = 0; i < size; i++) {
		poly[i].scale(a);
	}
	return *this;
}

Poly2d& Poly2d::roundToInt()
{
	for( int i = 0; i < size; i++) {
		poly[i].roundToInt();
	}
	return *this;
}

Poly2d& Poly2d::quantize(double quant)
{
	Point2d pt;

	for( int i = 0; i < size; i++) {
		// get i'th point; divide, round and multiply ...
		pt = poly[i] / quant;
		pt.roundToInt();
		poly[i] = pt * quant;
	}
	return *this;
}

Poly2d& Poly2d::operator*=( double a)
{
	for( int i = 0; i < size; i++) {
		poly[i] *= a;
	}
	return *this;
}

Poly2d& Poly2d::operator/=( double a)
{
	if( a == 0) {
		cout << "\nCannot divide by 0! Did not alter the polygon." << endl;
	}
	else {
		for( int i = 0; i < size; i++) {
			poly[i] /= a;
		}
	}
	return *this;
}

Poly2d& Poly2d::operator+=( const Point2d& p)
{
	for( int i = 0; i < size; i++) {
		poly[i] += p;
	}
	return *this;
}

Poly2d& Poly2d::operator-=( const Point2d& p)
{
	for( int i = 0; i < size; i++) {
		poly[i] -= p;
	}
	return *this;
}

int Poly2d::operator==( Poly2d& p)
{
	int i, j;

	if( size != p.size ) return 0;
	else {
		for( i = 0; i < size; i++)
			if( poly[0] == p[i]) break;
		if( i == size) return 0;
		for( j = 1; j < size; j++)
			if( poly[j] != p[j+i]) break;
		return ( j == size);
	}
}

int Poly2d::operator!=( Poly2d& p)
{
	int i, j;

	if( size != p.size) return 1;
	else {
		for( i = 0; i < size; i++)
			if( poly[0] == p[i]) break;
		if( i == size) return 1;
		for( j = 1; j < size; j++)
			if( poly[j] != p[j+i]) break;
		return ( !(j == size));
	}
}

int Poly2d::findIndex( const Point2d& vertex)
{
	int i;

	if( !poly) {
		cout << "\nWarning(Poly2d::findIndex): Polygon is empty." 
			 << "\n<-1> is returned" << endl;
		return -1;
	}
	else {
		for( i = 0; i < size; i++) {
			if( poly[i] == vertex) break;
		}
		if( i == size) {
			cout << "\nWarning(Poly2d::findIndex): Given point is not a vertex of the polygon."
				 << "\n<-1> is returned" << endl;
			return -1;
		}
		else return i;
	}
}


//--------------------------------------------------------------------------
// Proper if no two consecutive sides are parallel and no sides have 0 length.
//--------------------------------------------------------------------------
int Poly2d::isProper() const
{
	if( size <= 1) return 1;
	if( size == 2)
		return (poly[0] != poly[1]);
	for( int i = 0; i < size; i++) {
		if( (poly[Mod(i-1,size)]-poly[i]) || (poly[(i+1)%size]-poly[i]) )
			return 0;
	}
	return 1;
}


//--------------------------------------------------------------------------
// All triangles are convex. If a polygon is convex then it is valid.
// If the polygon is convex, then angles between the consecutive 
// vectors will either all be >= 180 or <= 180. We determine whether an	 
// angle is less or greater than or equal to 180 by using the cross-
// product operation. Note that this routine does not assume either clock-
// wise or counterclockwise ordering of the vertices of the polygon.
//--------------------------------------------------------------------------
int Poly2d::isConvex() const
{

	if( size <= 3) return 1;


	int i, i0;
	int newSign;
	int initialSign = 0;
	Point2d p1, p2;
	for( i0 = 0; i0 < size; i0++) {
		p1 = poly[Mod(i0-1,size)] - poly[i0];
		if( p1 != 0.0) break;
	}

	for( i = i0; ; ) {
		do {
			if( i == size) return 1;
			p2 = poly[(i+1)%size] - poly[i];
			i++;
		} while( p2 == 0.0);
		if( ( newSign = SignE( p1 ^ p2)) !=0 ) {
			if( initialSign == 0) initialSign = newSign;
			if( newSign != initialSign) break;
		}	  
		p1 = -1.0 * p2;
	}
	return 0;
}


//--------------------------------------------------------------------------
// Is the polygon a rectangle?
//--------------------------------------------------------------------------
int Poly2d::isRectangle()
{
	if( size != 4) return 0;

	if( poly[0].X() == poly[1].X()) {
		if( poly[1].Y() != poly[2].Y()) return 0;
		if( poly[2].X() != poly[3].X()) return 0;
		if( poly[3].Y() != poly[0].Y()) return 0;
		return 1;
	}
	else if( poly[0].Y() == poly[1].Y()) {
		if( poly[1].X() != poly[2].X()) return 0;
		if( poly[2].Y() != poly[3].Y()) return 0;
		if( poly[3].X() != poly[0].X()) return 0;
		return 1;
	}

	return 0;
}


//--------------------------------------------------------------------------
// Return 1 if ordering of vertices in polygon is counter-clockwise; 
// return -1 if ordering is clockwise; return 0 if polygon is empty.
// Assumes origin is at top left (x-axis points right; y-axis points down).
//--------------------------------------------------------------------------
int Poly2d::isCCW() const
{
	if (size < 3) { 
		return 0;
	}

	// compute zero'th moment of polygon using cross products
	double sum = 0.0;
	for (int i = 1; i < size-1; i++) {
		sum += (poly[i]-poly[0]) ^ (poly[i+1]-poly[0]);
	}
	if (sum > 0.0)
		return -1;
	else
		return 1;
}


//--------------------------------------------------------------------------
// Same as isCCW() but takes into account first n vertices of polygon only
//--------------------------------------------------------------------------
int Poly2d::isCCW(int n) const
{
	if (n > size	||	n < 3) { 
		return 0;
	}

	// compute zero'th moment of polygon using cross products
	double sum = 0.0;
	for (int i = 1; i < n-1; i++) {
		sum += (poly[i]-poly[0]) ^ (poly[i+1]-poly[0]);
	}
	if (sum > 0.0)
		return -1;
	else
		return 1;
}


double Poly2d::perimeter() const
{
	int i;
	double peri = 0.0;
	if( size < 2) return (0.0);
	if( size == 2) return sqrt(SQR(poly[0].X()-poly[1].X())+SQR(poly[0].Y()-poly[1].Y()));
	for( i = 0; i < size-1; i++){
		peri += sqrt(SQR(poly[i+1].X()-poly[i].X())+SQR(poly[i+1].Y()-poly[i].Y()));
	}
	peri += sqrt(SQR(poly[i].X()-poly[0].X())+SQR(poly[i].Y()-poly[0].Y()));
	return peri;
}

double Poly2d::area() const
{
	if( size < 3) {
		// A polygon with less than 3 sides has zero area
		return 0.0;
	}

	double two_times_area_pg = 0.0;

	// Use cross products to compute the area
	for( int i = 1; i < size-1; i++) {
		two_times_area_pg += (poly[i]-poly[0]) ^ (poly[i+1]-poly[0]);
	}

	return fabs( two_times_area_pg / 2.0);
}


int Poly2d::minM()
{
	if( size == 0) {
		return 0;
	}
	else {
		int valInt;
		int minVal = (int) floor( poly[0].X() + 0.5);
		for( int i = 1; i < size; i++)
			{
				valInt = (int) floor( poly[i].X() + 0.5);
				if( valInt < minVal) minVal = valInt;
			}
		return minVal;
	}
}

int Poly2d::maxM()
{
	if( size == 0) {
		return 0;
	}
	else {
		int valInt;
		int maxVal = (int) ceil( poly[0].X() - 0.5);
		for( int i = 1; i < size; i++) {
			valInt = (int) ceil( poly[i].X() - 0.5);
			if( valInt > maxVal) maxVal = valInt;
		}
		return maxVal;
	}
}

int Poly2d::minN()
{
	if( size == 0) {
		return 0;
	}
	else {
		int valInt;
		int minVal = (int) floor( poly[0].Y() + 0.5);
		for( int i = 1; i < size; i++) {
			valInt = (int) floor( poly[i].Y() + 0.5);
			if( valInt < minVal) minVal = valInt;
		}
		return minVal;
	}
}

int Poly2d::maxN()
{
	if( size == 0) {
		return 0;
	}
	else {
		int valInt;
		int maxVal = (int) ceil( poly[0].Y() - 0.5);
		for( int i = 1; i < size; i++) {
			valInt = (int) ceil( poly[i].Y() - 0.5);
			if( valInt > maxVal) maxVal = valInt;
		}
		return maxVal;
	}
}


int Poly2d::read(char filein[])
{
	ifstream fp;
	fp.open(filein, ios::in);
	if( ! fp) {
		cout << "Could not open file for input!" << endl;
		return 0;
	}

	double tmpX, tmpY;
	int tmpSize;

	fp>>tmpSize;
	if( tmpSize != size) {
		if (poly!=NULL)
			delete[] poly;
		poly = new Point2d[ size = tmpSize];
	}

	if(!fp) return 0;
	for( int i = 0; i < size; i++) {
		fp>>tmpX;
		if(!fp) return 0;
		fp>>tmpY;
		if(!fp) return 0;
		poly[i] = point( tmpX, tmpY);
	}

	fp.close();
	return 1;
}

int Poly2d::write(char fileout[])
{
	ofstream fp;
	fp.open(fileout, ios::out);
	if( ! fp) {
		cout << "Error: Could not open file for output!" << endl;
		return 0;
	}

	double tmpX, tmpY;

	if(!fp) return 0;
	fp<<size<<"\n";
	for( int i = 0; i < size; i++) {
		if(!fp) return 0;
		tmpX = poly[i].X();
		fp<<tmpX<<" ";
		if(!fp) return 0;
		tmpY = poly[i].Y();
		fp<<tmpY<<"\n";
	}
	fp.close();
	return 1;
}


void Poly2d::print()
{
	if( size == 0) 
		cout << "\n Polygon is empty!";
	for( int i = 0; i < size; i++) {
		cout << "\n	 Vertex #" << i+1 << " (x, y) = " << poly[i];
	}
	cout << endl;
}


//------------------------------------------------------------------------

// Default constructor: All points are set to 0.
Tri2d::Tri2d()	
: Poly2d(3)
{}

// Copy constructor: Initialization with another triangle
Tri2d::Tri2d( const Tri2d& from)
: Poly2d( from)
{}

// Initialization with three points.
Tri2d::Tri2d( const Point2d& p0, const Point2d& p1, const Point2d& p2)
{
	poly = new Point2d[3];
	size = 3;
	poly[0] = p0;
	poly[1] = p1;
	poly[2] = p2;
}

// Destructor
Tri2d::~Tri2d()
{
	//delete[] poly;
}

// Assignment operator: setting equal to another triangle
Tri2d& Tri2d::operator=(const Tri2d& from)
{
	Poly2d::operator=( from);
	return *this;
}

// Assignment operator: setting equal to another triangle (polygon)
int Tri2d::operator=(Poly2d& from)
{
	if( from.getSize() == 3) {
		Poly2d::operator=( from);
		return 1;
	}
	else {
		return 0;
	}
}

// Checks counter-clockwise-ness of vertices (see Poly2d::isCCW())
int Tri2d::isCCW() const
{
	int val = 0;
	if (size != 3)
		cout << "\nError: Tri2d::isCCW(): invalid triangle!" << endl;
	else
		val = ((((poly[1]-poly[0]) ^ (poly[2]-poly[0])) > 0.0) ? -1 : 1);

	return val;
}



//------------------------------------------------------------------------

Quad2d::Quad2d()  // All points are set to 0.
: Poly2d(4)
{}

Quad2d::Quad2d( const Quad2d& from)	 // Initialization with another quad.
: Poly2d( from)
{}

// Initialization with four points.
Quad2d::Quad2d( const Point2d& p0, const Point2d& p1, 
				const Point2d& p2, const Point2d& p3)
: Poly2d(4)
{
	poly[0] = p0;
	poly[1] = p1;
	poly[2] = p2;
	poly[3] = p3;
}

Quad2d::~Quad2d()
{
	//delete[] poly;
}

Quad2d& Quad2d::operator=(const Quad2d& from)  // Setting equal to a quad.
{
	Poly2d::operator=( from);
	return *this;
}

int Quad2d::operator=(Poly2d& from)	 // Setting equal to a triangle
{
	if( from.getSize() == 4) {
		Poly2d::operator=( from);
		return 1;
	}
	else {
		return 0;
	}
}

