/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* A. Tanju Erdem (Eastman Kodak Company)                             */
/* Candemir Toklu (University of Rochester)                           */
/*                                                                    */
/* and edited by                                                      */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* Eastman Kodak Company and the University of Rochester retain full  */
/* right to use the code for their own purposes, assign or donate the */
/* code to a third party and to inhibit third parties from using the  */
/* code for non MPEG-4 Visual (ISO/IEC 14496-2) conforming products.  */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1994-1998                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         polyClass.h
   AUTHOR       A. Tanju Erdem - Eastman Kodak Company and Candemir Toklu - 
                University of Rochester.
   DATE         September 1994
   UPDATE       14 August 1996 by Candemir Toklu
                - various
                Peter Van Beek, Fri May  1 16:08:26 1998 
                - added quantize method
   DESCRIPTION  The Poly2d class defines a 2-D polygon. Its private members 
                are an array of Point2d and the size of the polygon. The
                size of the polygon can dynamically change during time. 
                Classes Tri2d and Quad2d which simply represent a triangle 
                and a quadrilateral are derived from this class.

   SYNOPSIS     Include files : 
                                
                #include <iostream.h>
                #include <fstream.h>
                #include <math.h>
                #include <time.h>
                #include "pointClass.h"

*********************************************************************-*/
#include <iostream.h>
#include <fstream.h>
#include <math.h>
#include <time.h>
#include "pointClass.h"

#ifndef POLY_CLASS
#define POLY_CLASS

//--------------------------------------------------------------------------
// Classes Tri2d and Quad2d are derived from class Poly2d.
// One can change the size of a polygon after it is defined.
class Poly2d {
	friend class Tri2dTess;
  public:
	Poly2d();
	Poly2d( int);  // Initializes all points to zero
	Poly2d( const Poly2d&);
	Poly2d( const Poly2d&, int m, int n);
	Poly2d( const Point2d&);
	~Poly2d();
	Poly2d& operator=( const Poly2d&);
	Point2d& operator[]( int);
	Poly2d& empty();  // Remove all the vertices of the polygon (polygon = NULL)
	Poly2d& replace( int, const Point2d&);	// Replace a side
	Poly2d& add( int indx, const Point2d&);	 // Add a new vertex at indx
	Poly2d& insert( const Point2d&);  // Insert at the beginning
	Poly2d& append( const Point2d&);  // Append to the end
	Poly2d& move( double, double);
	Poly2d& move( const Point2d&);
	Poly2d& scale( double, double);
	Poly2d& scale( double);
	Poly2d& roundToInt();
	Poly2d& quantize(double);
	Poly2d& operator*=( double);
	Poly2d& operator/=( double);
	Poly2d& operator+=( const Point2d&);
	Poly2d& operator-=( const Point2d&);
	int operator==( Poly2d&);
	int operator!=( Poly2d&);
	int findIndex( const Point2d&); //Sends the index of a vertex of the polygon
	// returns -1 if the vertex does not exist on the polygon.
	int isProper() const;  // Proper if no two consecutive sides are 
						   // parallel and no sides have 0 length
	int isConvex() const;  // Is triangle convex?
	int isRectangle();	// Is the polygon a rectangle?
	int isCCW() const;	   // Are vertices ordered in counter-clockwise fashion?
	int isCCW(int) const;  // Are first n vertices ordered counter-clockwise?
	double perimeter() const;	 // Perimeter of a polygon
	double area() const;		 // Area of a polygon
	inline int getSize() const;	 // Return number of vertices
	int minM();
	int minN();
	int maxM();
	int maxN();
	int read( char []);
	int write( char[]);
	void print();
  protected:
	Point2d* poly;
	int size;
};

int Poly2d::getSize() const
{
	return size;
}

//-------------------------------------------------------------------------
// A Triangle is a Polygon with 3 sides.
class Tri2d: public Poly2d {
  public:
	Tri2d();  // Initialize all points with (0,0)
	Tri2d( const Point2d&, const Point2d&, const Point2d&);
	Tri2d( const Tri2d&);
	~Tri2d();
	Tri2d& operator=( const Tri2d&);
	int operator=( Poly2d&);  // Returns 1 if assignment is legal 0 otherwise
	int isCCW() const;
};

//-------------------------------------------------------------------------
// A Quadrilateral is a Polygon with 4 sides.
class Quad2d: public Poly2d {
  public:
	Quad2d();  // Initialize all points with (0,0)
	Quad2d( const Point2d&, const Point2d&, const Point2d&, const Point2d&);
	Quad2d( const Quad2d&);
	~Quad2d();
	Quad2d& operator=( const Quad2d&);
	int operator=( Poly2d&);  // Returns 1 if assignment is legal 0 otherwise
};

#endif
