/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* The University of Rochester retains full right to use the code for */
/* their own purposes, assign or donate the code to a third party and */
/* to inhibit third parties from using the code for non MPEG-4 Visual */
/* (ISO/IEC 14496-2) conforming products.                             */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1997-1998                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         queueClass.cpp
   AUTHOR       Peter Van Beek, Sat Mar  1 14:31:35 1997
   DESCRIPTION  Queue of triangle elements implemented by singly linked
                list. Queue elements are handled on a first-in-first-out 
                basis. Each element contains a pointer to a triangle
                as well as an edge number.
   SYNOPSIS     Include files:
                #include "general.h"
                #include "triRefClass.h"
                #include "queueClass.h"
*********************************************************************-*/

#include "general.h"
#include "queueClass.h"
#include "triRefClass.h"


// constructor
TriangleQueue::TriangleQueue()
{	
	first = NULL; last = NULL;
}

// destructor
TriangleQueue::~TriangleQueue()
{	
	TriangleQueueElem *elem = first;

	while (elem != NULL) {
		elem = first->next;
		delete (first);
		first = elem;
	}
	first = NULL; last = NULL;
}

// copy constructor
TriangleQueue::TriangleQueue (const TriangleQueue& queue)
{
	if (queue.first == NULL	 ||	 queue.last == NULL) {
		first = NULL; last = NULL;
	}
	else {
		TriangleQueueElem *elem = queue.first;
		TriangleQueueElem *newelem = new TriangleQueueElem;

		// first copy the first element
		newelem->tri = elem->tri;
		newelem->edge = elem->edge;
		newelem->next = NULL;
		first = last = newelem;

		while (elem->next != NULL) {
			elem = elem->next;
			newelem = new TriangleQueueElem;
			newelem->tri = elem->tri;
			newelem->edge = elem->edge;
			newelem->next = NULL;
			last->next = newelem;
			last = newelem;
		}
	}
}

// assignment operator
TriangleQueue& TriangleQueue::operator= (const TriangleQueue& queue)
{
	// protect against copying to itself
	if (this == &queue) return (*this);

	TriangleQueueElem *elem = first;

	// destruct old queue first
	while (elem != NULL) {
		elem = first->next;
		delete (first);
		first = elem;
	}
	first = NULL; last = NULL;

	// now copy other queue
	if (queue.first != NULL	 &&	 queue.last != NULL) {
		elem = queue.first;
		TriangleQueueElem *newelem = new TriangleQueueElem;

		// first copy the first element
		newelem->tri = elem->tri;
		newelem->edge = elem->edge;
		newelem->next = NULL;
		first = last = newelem;

		while (elem->next != NULL) {
			elem = elem->next;
			newelem = new TriangleQueueElem;
			newelem->tri = elem->tri;
			newelem->edge = elem->edge;
			newelem->next = NULL;
			last->next = newelem;
			last = newelem;
		}
	}

	return (*this);
}

// append an element at the end of the queue
TriangleQueue& TriangleQueue::append (Tri2dRef *tri, int edge)
{
	TriangleQueueElem *newelem = new TriangleQueueElem;

	// fill new element
	newelem->tri = tri;
	newelem->edge = edge;
	newelem->next = NULL;

	if (first == NULL  ||  last == NULL) {
		// list was empty
		first = last = newelem;
	}
	else {
		// add to the linked list
		last->next = newelem;
		last = newelem;
	}
	return (*this);
}

// remove and return the element at the front of the queue
TriangleQueue& TriangleQueue::remove(TriangleQueueElem& elem)
{
	if (first == NULL  ||  last == NULL) {
		cout << "TriangleQueue: queue is empty!" << endl;
		exit(1);
	}
	else {
		TriangleQueueElem *oldelem = first;
		elem.tri = oldelem->tri;
		elem.edge = oldelem->edge;
		elem.next = NULL;
		first = first->next;
		if (first == NULL) last = NULL;
		delete oldelem;
	}
	return (*this);
}

// remove and return the element at the front of the queue
TriangleQueue& TriangleQueue::remove(Tri2dRef *&tri, int &edge)
{
	if (first == NULL  ||  last == NULL) {
		cout << "TriangleQueue: queue is empty!" << endl;
		exit(1);
	}
	else {
		TriangleQueueElem *oldelem = first;
		tri = (Tri2dRef *) oldelem->tri;
		edge = (int) oldelem->edge;
		first = first->next;
		if (first == NULL) last = NULL;
		delete oldelem;
	}
	return (*this);
}

// return the data of the element at the front of the queue
TriangleQueue& TriangleQueue::front(Tri2dRef *&tri, int &edge)
{
	if (first == NULL  ||  last == NULL) {
		cout << "TriangleQueue: queue is empty!" << endl;
		exit(1);
	}
	else {
		tri = (Tri2dRef *) first->tri;
		edge = (int) first->edge;
	}
	return (*this);
}

// check if queue is empty or not
int TriangleQueue::empty ()
{
	if (first == NULL  ||  last == NULL) 
		return (1);
	else
		return (0);
}

// print contents of elements of the queue
void TriangleQueue::print()
{
	TriangleQueueElem *elem = first;
	while (elem != NULL) {
		cout << elem->tri << " " << elem->edge << endl;
		elem = elem->next;
	}
}

// end of file
