/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* A. Tanju Erdem (Eastman Kodak Company)                             */
/* Candemir Toklu (University of Rochester)                           */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* Eastman Kodak Company and the University of Rochester retain full  */
/* right to use the code for their own purposes, assign or donate the */
/* code to a third party and to inhibit third parties from using the  */
/* code for non MPEG-4 Visual (ISO/IEC 14496-2) conforming products.  */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1994-1998                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         triRefClass.h
   AUTHOR       Peter van Beek - University of Rochester.
   DATE         Fri Aug  7 11:37:54 1998
   DESCRIPTION  The Tri2dRef class defines a triangle structure. It has
                pointers to its three vertices as well as pointers to
                its neighboring triangles. A Tri2dRef can be part of a
                linked list.
   SYNOPSIS     Include files: 
                #include <iostream.h>
                #include <fstream.h>
                #include <math.h>
                #include "polyClass.h"
                #include "linkClass.h"
*********************************************************************-*/

#include <iostream.h>
#include <fstream.h>
#include <math.h>
#include "polyClass.h"
#include "linkClass.h"

class Tri2d;
class Tri2dTess;

#ifndef TRI_LINK_CLASS
#define TRI_LINK_CLASS

// A reference-to-a-triangle represents a triangle. A triangle contains
// pointers to the corner points of the triangle and to its neighboring 
// triangles. Two triangles are said to be neighbors if they share a side.
// The indexing of the corner points is clockwise; further, as in the 
// figure below, the neighboring triangle on side [i,i+1] is accessed with
// the index i. Note that i+1 is computed modulo 3. 
//		   0
//		   +
//		   |\				 +'s indicate the corner points.
//	 2 <-- | \ --> 0	   -->'s point to the neighboring triangles.
//		   |  \
//		 2 +---+ 1
//			 |
//			 V
//			 1
//
#ifndef WIN32
class Tri2dRef {
#else
class __declspec(dllexport) Tri2dRef {
#endif
	friend class Tri2dTess;
  public:
	Tri2dRef();
	~Tri2dRef();
	Tri2d	  triangle();				// return a copy of the triangle
	inline Pt2dLink *getLink(int);		// return ptr to vertex
	inline Tri2dRef *getNeighborTri(int); // return ptr to neighboring triangle
	inline int		getNeighborEdge(int); // return number of neighboring edge
	char	  mark;						// marker
  private:
	Pt2dLink* corner[3];				// pointers to the vertices
	Tri2dRef* neighbor[3];				// pointer to the neighbors
	char	  nedge[3];					// number of neighboring edge
	Tri2dRef *next, *prev;				// pointers to previous and next 
										// links in a linked list
	inline void connect(Tri2dRef*, int, int); // connect with other triangle
};


// return the n'th vertex of the triangle
Pt2dLink* Tri2dRef::getLink(int n)
{
	Pt2dLink* ptr = NULL;
	if (n<0 || n>2)
		cout << "Tri2dRef::getLink: index out of range" << endl;
	else
		ptr = corner[n];
	return ptr;
}

// return ptr to n'th neighboring triangle
Tri2dRef *Tri2dRef::getNeighborTri(int n)
{
	Tri2dRef *tri = NULL;
	if (n<0 || n>2)
		cout << "Tri2dRef::getNeighborTri: index out of range" << endl;
	else
		tri = neighbor[n];
	return tri;
}

// return number of the shared edge in the n'th neighboring triangle
int Tri2dRef::getNeighborEdge(int n)
{
	int edge = -1;
	if (n<0 || n>2)
		cout << "Tri2dRef::getNeighborEdge:index out of range" << endl;
	else 
		edge = nedge[n];
	return edge;
}

// connect this triangle with another triangle 'triOther' along edge 'e'; 
// 'eOther' is the edge of triOther that should be connected with
void Tri2dRef::connect(Tri2dRef *triOther, int e, int eOther)
{
	this->neighbor[e] = triOther;
	this->nedge[e] = eOther;
	if (triOther != (Tri2dRef *)NULL) {
		triOther->neighbor[eOther] = this;
		triOther->nedge[eOther] = e;
	}
}

#endif
