/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* A. Tanju Erdem (Eastman Kodak Company)                             */
/* Candemir Toklu (University of Rochester)                           */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* Eastman Kodak Company and the University of Rochester retain full  */
/* right to use the code for their own purposes, assign or donate the */
/* code to a third party and to inhibit third parties from using the  */
/* code for non MPEG-4 Visual (ISO/IEC 14496-2) conforming products.  */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1994-1998                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         triTessClass.cpp
   AUTHOR       Peter van Beek - University of Rochester.
   DATE         Fri Aug  7 12:15:35 1998
   DESCRIPTION  Member functions for Tri2dTess class.
   SYNOPSIS     Include files: 
                #include "general.h"
                #include "triTessClass.h"
*********************************************************************-*/

#include "general.h"
#include "triTessClass.h"


//------------------------------------------------------------------------
// Constructor: empty tesselation
Tri2dTess::Tri2dTess()
{
	sizeTess = 0;
	tess = NULL;
}

//------------------------------------------------------------------------
// Constructor: a tesselation with a single triangle
Tri2dTess::Tri2dTess(const Tri2d& p)
{
	if (p.size != 3) { // Invalid triangle
		sizeTess = 0;
		tess = NULL;
	}
	else {
		// Check orientation
		int ccw = (p.isCCW() == 1);

		sizeTess = 1;
		tess = new Tri2dRef();
		Pt2dLink *tmpLink;
		for(int i = 0; i < 3; i++) {
			// Add vertex to the list of points
			if (ccw)
				tmpLink = points.add(p.poly[2-i]);
			else
				tmpLink = points.add(p.poly[i]);
			tess->corner[i] = tmpLink;
		}
		tess->next = tess->prev = tess;	 // Tesselation is closed

		// A single triangle has no neighbors
	}
}

//------------------------------------------------------------------------
// Copy constructor: initialization with another tesselation
Tri2dTess::Tri2dTess(const Tri2dTess& from)
{
	sizeTess = 0;
	tess = NULL;
	*this = from;
}

//------------------------------------------------------------------------
// Destructor
Tri2dTess::~Tri2dTess()
{
	Tri2dRef *tmpRef;
	for (int i = 0; i < sizeTess; i++) {
		tmpRef = tess->prev;
		tmpRef->prev->next = tmpRef->next;
		tmpRef->next->prev = tmpRef->prev;
		delete tmpRef;	// Remove the current last reference from the list
	}
	// The destructor for "points" is called automatically after this one.
}

//------------------------------------------------------------------------
// Destroyer
Tri2dTess& Tri2dTess::destroy()
{
	Pt2dLink* tmpLink;
	Tri2dRef* tmpRef;
	int i;

	for (i = 0; i < sizeTess; i++) {
		tmpRef = tess->prev;
		tmpRef->prev->next = tmpRef->next;
		tmpRef->next->prev = tmpRef->prev;
		delete tmpRef;	// Remove the current last reference from the list
	}

	for (i = 0; i < points.size; i++) {
		tmpLink = points.list->prev;
		tmpLink->prev->next = tmpLink->next;
		tmpLink->next->prev = tmpLink->prev;
		delete tmpLink;	 // Remove the current last point-link from the list
	}
	tess = NULL;
	sizeTess = 0;
	points.list = NULL;
	points.plast = NULL;
	points.ilast = 0;
	points.size = 0;
	return *this;
}

//------------------------------------------------------------------------
// Assignment operator: set equal to another tesselation
Tri2dTess& Tri2dTess::operator=(const Tri2dTess& from) 
{
	Tri2dRef *tmpRef, *newRef, *adjRef;
	int i, j;

	// Cannot copy a tesselation onto itself
	if (this == &from) return *this;  

	// Destruct the old tesselation
	for (i = 0; i < sizeTess; i++) {
		tmpRef = tess->prev;
		tmpRef->prev->next = tmpRef->next;
		tmpRef->next->prev = tmpRef->prev;
		delete tmpRef;	// Remove the current last reference from the list
	}

	// Copy the new list of points
	points = from.points;  

	// Then, construct a new tesselation that is identical to the given one
	if (from.sizeTess == 0) {
		tess = NULL;
		sizeTess = 0;
	}
	else {
		// Create the list of references
		tmpRef = from.tess;
		tess = new Tri2dRef();	// Construct the 1st ref. in the list
		tess->next = tess->prev = tess;	 // Close the list
		for (i = 1; i < from.sizeTess; i++) {
			tmpRef = tmpRef->next;
			newRef = new Tri2dRef();
			newRef->prev = tess->prev;
			newRef->next = tess;
			tess->prev->next = newRef;
			tess->prev = newRef;
		}
		sizeTess = from.sizeTess;

		// Fill in the references
		tmpRef = from.tess;
		newRef = tess;
		Pt2dLink *tmpLink;
		for (i = 0; i < from.sizeTess; i++, tmpRef = tmpRef->next, newRef = newRef->next) {
			for (j = 0; j < 3; j++) {
				// Connect triangle to its corners and neighbors
				tmpLink = addressC(from.where(tmpRef->corner[j]));
				newRef->corner[j] = tmpLink;
				adjRef = tmpRef->neighbor[j];
				if (adjRef == NULL)
					newRef->neighbor[j] = NULL;
				else
					newRef->neighbor[j] = addressR(from.where(adjRef));
				newRef->nedge[j] = tmpRef->nedge[j];
			}
		}
	}
	return *this;
}

//--------------------------------------------------------------------------
// Append a single triangle to existing tesselation
Tri2dTess& Tri2dTess::append(Tri2d& p)
{
	// Check triangle
	if (p.size != 3) {
		cout << "Tri2dTess::append: invalid triangle..." << endl;
		return *this;
	}
	// Check orientation
	int ccw = (p.isCCW() == 1);
	// Initialize new triangle, with corners = neighbors = NULL
	Tri2dRef *newRef = new Tri2dRef();	
	Pt2dLink *tmpLink;
	// Connect triangle to its corner vertices in clockwise direction
	for (int j = 0; j < 3; j++) {
		// Add vertex to the point list if not already in
		if (ccw)
			tmpLink = points.add(p[2-j]);
		else
			tmpLink = points.add(p[j]);

		newRef->corner[j] = tmpLink;
	}
	
	// Connect triangle to its neighboring triangles
	for (int i = 0; i < 3; i++) {
		// is there another triangle with vertices i and i+1 ?
		Tri2dRef *tmpRef;
		int eIndex;
		// this is very inefficient :-( ...
		int found = findEdge(newRef->corner[(i+1)%3], newRef->corner[i], tmpRef, eIndex);
		if (found) {
			// connect it
			newRef->neighbor[i] = tmpRef;
			newRef->nedge[i] = eIndex;
			tmpRef->neighbor[eIndex] = newRef;
			tmpRef->nedge[eIndex] = i;
		}
	}

	if (tess) {	 // Add the new reference to the existing tesselation
		newRef->next = tess;
		newRef->prev = tess->prev;
		tess->prev->next = newRef;
		tess->prev = newRef;
	}
	else {	// Tesselation was empty
		tess = newRef;
		tess->next = tess->prev = tess;
	}
	sizeTess++;
	return *this;
}

//--------------------------------------------------------------------------
Tri2dTess& Tri2dTess::move(double dx, double dy)
{
	for( int i = 0; i < points.size; i++) {
		points[i].move(dx,dy);
	}
	return *this;
}

//--------------------------------------------------------------------------
Tri2dTess& Tri2dTess::move(const Point2d& p)
{
	for( int i = 0; i < points.size; i++) {
		points[i]+=p;
	}
	return *this;
}

//--------------------------------------------------------------------------
Tri2dTess& Tri2dTess::scale(double dx, double dy)
{
	for( int i = 0; i < points.size; i++) {
		points[i].scale(dx,dy);
	}
	return *this;
}

//--------------------------------------------------------------------------
// Quantize point coordinates with interval 'quant'
Tri2dTess& Tri2dTess::quantize(double quant)
{
	Point2d pt;

	for (int i = 0; i < points.size; i++) {
		// get i'th point; divide, round and multiply ...
		pt = points[i] / quant;
		pt.roundToInt();
		points[i] = pt * quant;
	}
	return *this;
}


//--------------------------------------------------------------------------
// Return a copy of indexed triangle 
Tri2d Tri2dTess::triangle(int indx)
{
	if(indx < 0 || indx >= sizeTess) {
		cout << "Tri2dTess::triangle: Illegal polygon index..." << endl;
		exit(1);
	}

	int i;

	// Move to the requested triangle
	Tri2dRef *tmp = tess;
	for (i = 0; i < indx; i++)
		tmp = tmp->next;

	// Build a copy of the triangle
	Tri2d p;
	for (i = 0; i < 3; i++)
		p[i] = tmp->corner[i]->here;

	return p;
}


//--------------------------------------------------------------------------
int Tri2dTess::where(Pt2dLink* pLink) const
{
	return points.where( pLink);
}

//--------------------------------------------------------------------------
int Tri2dTess::where(Tri2dRef* pRef) const
{
	if (sizeTess == 0) {
		return -1;	// The tesselation is empty
	}
	else {
		int indx = 0;

		Tri2dRef* tmpRef = tess;
		while (tmpRef != pRef && indx < sizeTess) {
			indx++;
			tmpRef = tmpRef->next;
		}
		if (indx == sizeTess) {	 // Triangle reference not found
			return -1;
		}
		else {
			return indx;   // Return the location of the triangle reference
		}
	}
}

//--------------------------------------------------------------------------
int Tri2dTess::where(Point2d& p) const
{
	return points.where( p);
}

//--------------------------------------------------------------------------
Pt2dLink* Tri2dTess::addressC(int indx)
{
	return points.address( indx);
}

//--------------------------------------------------------------------------
Pt2dLink* Tri2dTess::addressC(const Point2d p)
{
	return points.address( p);
}

//--------------------------------------------------------------------------
Tri2dRef* Tri2dTess::addressR(int indx)
{
	if(indx < 0 || indx >= sizeTess) {
		cout << "Tri2dTess::addressR: Index value is out of range!" << endl;
		exit( 1);
	}
	else {
		int i;
		Tri2dRef *tmpRef = tess;
		for (i = 0; i < indx; i++) tmpRef = tmpRef->next;
		return tmpRef;
	}
	return(NULL);
}

//--------------------------------------------------------------------------
Tri2dRef *Tri2dTess::nextTriangle(int first)
{
	static Tri2dRef *triPtr = tess;
	
	if (first) {
		if (tess != NULL)
			triPtr = tess;
		else
			cout << "Tri2dTess::nextTriangle: tess is NULL pointer!" << endl;
	}
	else {
		if (triPtr != NULL)
			triPtr = triPtr->next;
		else 
			cout << "Tri2dTess::nextTriangle: NULL pointer!" << endl;
	}
	return triPtr;
}

//--------------------------------------------------------------------------
// Find triangle and edge with these vertices
int Tri2dTess::findEdge(Pt2dLink* pt1, Pt2dLink* pt2, Tri2dRef* &triPtr, int &e)
{
	triPtr = tess;
	for (int i = 0; i < sizeTess; i++, triPtr = triPtr->next) {
		if (triPtr->corner[0] == pt1  &&  triPtr->corner[1] == pt2) {
			e = 0;
			return 1;
		}
		else if (triPtr->corner[1] == pt1  &&  triPtr->corner[2] == pt2) {
			e = 1;
			return 1;
		}
		else if (triPtr->corner[2] == pt1  &&  triPtr->corner[0] == pt2) {
			e = 2;
			return 1;
		}
	}
	return 0;
}


//--------------------------------------------------------------------------
void Tri2dTess::markRefs(const char m)
{
	if (sizeTess != 0) {
		Tri2dRef* tmpRef = tess;
		do {
			tmpRef->mark = m;
			tmpRef = tmpRef->next;
		} while (tmpRef != tess);
	}
}


//--------------------------------------------------------------------------
void Tri2dTess::markLinks(const char m)
{
	points.markLinks(m);
}


//--------------------------------------------------------------------------
int Tri2dTess::read(ifstream& fp)
{
	Pt2dLink *tmpLink;
	Tri2dRef *tmpRef, *newRef;
	double x,y;
	int i, j, index;
  
	// First, destruct the old tesselation. 
	destroy();

	// Read number of points and number of triangles
	if(!fp) return 0;
	fp>>points.size;
	if(!fp) return 0;
	fp>>sizeTess;

	// Create the triangle list
	tess = (Tri2dRef*) new Tri2dRef();
	tess->next = tess->prev = tess;
	for (i = 1; i < sizeTess; i++) {
		newRef = (Tri2dRef*) new Tri2dRef();
		newRef->prev = tess->prev;
		newRef->next = tess;
		tess->prev->next = newRef;
		tess->prev = newRef;
	}

	// Create the point list
	Pt2dLink *newLink;
	points.list = (Pt2dLink*) new Pt2dLink();
	points.list->next = points.list->prev = points.list;
	for (i = 1; i < points.size; i++) {
		newLink = (Pt2dLink*) new Pt2dLink();
		newLink->prev = points.list->prev;
		newLink->next = points.list;
		points.list->prev->next = newLink;
		points.list->prev = newLink;
	}
	points.plast = points.list->prev;
	points.ilast = points.size - 1;

	// Read the point coordinates
	tmpLink = points.list;
	for (i = 0; i < points.size; i++, tmpLink = tmpLink->next) {
		fp>>x;
		fp>>y;
		tmpLink->here = ::point( x,y);
		if (!fp) return 0;
	}

	// Read the triangle vertex indices
	tmpRef = tess;
	for(i = 0; i < sizeTess; i++, tmpRef = tmpRef->next) {
		if(!fp) return 0;
		for (j = 0; j < 3; j++) {
			fp>>index;
			tmpRef->corner[j] = addressC(index); 
		}
	}

	// Now link triangles to their neighors
	tmpRef = tess;
	for (i = 0; i < sizeTess; i++, tmpRef = tmpRef->next) {
		// for each triangle edge ...
		for (j = 0; j < 3; j++) {
			// is there another triangle with vertices j and j+1 ?
			// this is very inefficient :-( ...
			int found = findEdge(tmpRef->corner[(j+1)%3], tmpRef->corner[j], newRef, index);
			if (found) {
				// connect it
				tmpRef->neighbor[j] = newRef;
				tmpRef->nedge[j] = index;
				newRef->neighbor[index] = tmpRef;
				newRef->nedge[index] = j;
			}
		}
	}
	// Done
	return 1;
}

//--------------------------------------------------------------------------
// Write the tesselation into a file using a simple file format
int Tri2dTess::write(ofstream& fp)
{
	fp.precision(18);
	if(!fp) return 0;
	fp<<points.size<<"\n";
	if(!fp) return 0;
	fp<<sizeTess<<"\n\n"; 
  
	Pt2dLink* tmpLink = points.list;
	do {
		if(!fp) return 0;
		fp<<tmpLink->here.X()<<" "<<tmpLink->here.Y()<<"\n";
		tmpLink = tmpLink->next;
	} while (tmpLink != points.list);

	if(!fp) return 0;
	fp<<"\n";

	int j;
	Tri2dRef* tmpRef = tess;
	do {
		if(!fp) return 0;
		for (j = 0; j < 3; j++) {
			fp<<where(tmpRef->corner[j])<<" ";
		}
		fp<<"\n";
		if(!fp) return 0;
		tmpRef = tmpRef->next;
	} while (tmpRef != tess);

	return 1;
}


//--------------------------------------------------------------------------
// read in the tesselation from fin
int Tri2dTess::read(char fin[])
{
	ifstream ifs;

	ifs.open(fin, ios::in);
	if( ! ifs) {
		cout << "\n	 Could not open file " << fin 
			 << " for input!" << endl;
		return 0;
	}
	if( ! read( ifs)) {
		cout << "\n	 Error in reading file" << endl;
		return 0;
	}
	ifs.close();
	return 1;
}

//--------------------------------------------------------------------------
// save the tesselation to fout
int Tri2dTess::write(char fout[])
{
	ofstream ofs;

	ofs.open(fout, ios::out);
	if( ! ofs) {
		cout << "\n	 Error: Could not open file " << fout 
			 << " for output!" << endl;
		return 0;
	}
	if( ! write( ofs)) {
		cout << "\n	 Error in writing file" << endl;
		return 0;
	}
	ofs.close();
	return 1;
}

//--------------------------------------------------------------------------
void Tri2dTess::print()
{
	int i, j;
	Pt2dLink *tmpLink = points.list;
	i = 0;
	cout << "\nThe list of points of the mesh is as follows." << endl;
	do {
		cout << i++ << ": " << tmpLink->here << endl;
		tmpLink = tmpLink->next;
	} while (tmpLink != points.list);

	Tri2dRef *tmpRef = tess;
	i = 0;
	cout << "\nThe list of triangles and their vertices is as follows." << endl;
	do {
		cout << i++ << ": [" ;
		for (j = 0; j < 3; j++) {
			cout << tmpRef->corner[j]->here;
			if (j != 2) cout << ", ";
			else cout << "] (";
		}
		for (j = 0; j < 3; j++) {
			cout << where(tmpRef->corner[j]);
			if (j != 2) cout << ", ";
			else cout << ")	 - ";
		}
		cout << "Adjacent: ";
		for (j = 0; j < 3; j++) {
			if (tmpRef->neighbor[j] != NULL)
				cout << where(tmpRef->neighbor[j]);
			else
				cout << "nil";
			if (j != 2) cout << ", ";
			else cout << endl;
		}
		tmpRef = tmpRef->next;

	} while(tmpRef != tess);
}


// Construct a uniform triangular mesh - only four types of uniform meshes
// are allowed with different triangle orientations
// nhori and nvert are the number of vertices in the horizontal, resp. 
// vertical direction; hsize and vsize are the dimensions of each basic
// rectangle; splitcode defines how each rectangle is split into two
// triangles. 
// Note: triangle orientations are clockwise, assuming the origin is at the
// top-left (x-axis points right, y-axis points down).
int Tri2dTess::makeUniform(int nhori, int nvert, double hsize, double vsize, int splitcode) 
{
	int i, n, nh = 0, nv, split = 0;
	double hpos, vpos;
	int ntotal = nhori*nvert;
	Pt2dLink *arrLink;
	Tri2dRef *arrRef, *tmpRef, *nextRef;

	if (nhori < 2 || nvert < 2 || hsize <= 0.0 || vsize <= 0.0) {
		cout << "\nError: illegal parameter in makeUniform " << endl;
		return 0;
	}

	// First, destruct the old tesselation if there is one
	destroy();

	// Construct the point list from an array of Pt2dLinks
	arrLink = new Pt2dLink [ntotal];
	points.size	 = ntotal;
	points.list	 = points.plast = arrLink;
	points.ilast = 0;
	// Fill the values of each Pt2dLink
	vpos = 0.0;
	for (nv = 0, n = 0; nv < nvert; nv++, vpos+=vsize) {
		hpos = 0.0;
		for (nh = 0; nh < nhori; nh++, n++, hpos+=hsize) {
			arrLink[n].here = Point2d(hpos, vpos);
			arrLink[n].next = &arrLink[(n+1)%ntotal];
			arrLink[n].prev = &arrLink[(n-1+ntotal)%ntotal];
		}
	}

	// Construct the polygon list from an array of Tri2dRefs
	sizeTess = 2*(nvert-1)*(nhori-1);
	arrRef = new Tri2dRef [sizeTess];
	tess = tmpRef = &arrRef[0];
	tess->next = tess->prev = tess;			// close the list
	for (i = 1; i < sizeTess; i++) {
		tmpRef++;
		tmpRef->prev = tess->prev;
		tmpRef->next = tess;
		tess->prev->next = tmpRef;
		tess->prev = tmpRef;
	}

	// Fill in the references to points and link to neighbors
	tmpRef = tess;
	for (nv = 1; nv < nvert; nv++) {
		for (nh = 1; nh < nhori; nh++) {
			// decide local splitting direction in this rectangle
			switch (splitcode) {
			  case 0:
				split = 0; break;
			  case 1:
				split = 1; break;
			  case 2:
				split = ((nh & 1)? 0 : 1); break;
			  case 3:
				split = ((nh & 1)? 1 : 0); break;
			}
			// specify the two triangles lying inside one rectangle
			// put vertices in clockwise order!
			nextRef = tmpRef->next;			// the next triangle
			switch (split) {
			  case 0:
				tmpRef->corner[0]  = &arrLink[(nv-1)*nhori+nh-1];
				tmpRef->corner[1]  = &arrLink[nv*nhori+nh];
				tmpRef->corner[2]  = &arrLink[nv*nhori+nh-1];
				nextRef->corner[0] = &arrLink[(nv-1)*nhori+nh-1];
				nextRef->corner[1] = &arrLink[(nv-1)*nhori+nh];
				nextRef->corner[2] = &arrLink[nv*nhori+nh];
				break;
			  case 1:
				tmpRef->corner[0]  = &arrLink[(nv-1)*nhori+nh-1];
				tmpRef->corner[1]  = &arrLink[(nv-1)*nhori+nh];
				tmpRef->corner[2]  = &arrLink[nv*nhori+nh-1];
				nextRef->corner[0] = &arrLink[nv*nhori+nh-1];
				nextRef->corner[1] = &arrLink[(nv-1)*nhori+nh];
				nextRef->corner[2] = &arrLink[nv*nhori+nh];
				break;
			}
			// connect each of these triangles to its neighbors
			switch (split) {
			  case 0:
				tmpRef->neighbor[0] = nextRef;
				tmpRef->nedge[0] = 2;
				if (nv < nvert-1) {
					tmpRef->neighbor[1] = nextRef+(nhori-1)*2;
					tmpRef->nedge[1] = 0;
				}
				if (nh > 1) {
					tmpRef->neighbor[2] = tmpRef-1;
					tmpRef->nedge[2] = 1;
				}
				if (nv > 1) {
					nextRef->neighbor[0] = tmpRef-(nhori-1)*2;
					nextRef->nedge[0] = 1;
				}
				if (nh < nhori-1) {
					nextRef->neighbor[1] = nextRef+1;
					nextRef->nedge[1] = 2;
				}
				nextRef->neighbor[2] = tmpRef;
				nextRef->nedge[2] = 0;
				break;
			  case 1:
				if (nv > 1) {
					tmpRef->neighbor[0] = nextRef-(nhori-1)*2;
					tmpRef->nedge[0] = 2;
				}
				tmpRef->neighbor[1] = nextRef;
				tmpRef->nedge[1] = 0;
				if (nh > 1) {
					tmpRef->neighbor[2] = tmpRef-1;
					tmpRef->nedge[2] = 1;
				}
				nextRef->neighbor[0] = tmpRef;
				nextRef->nedge[0] = 1;
				if (nh < nhori-1) {
					nextRef->neighbor[1] = nextRef+1;
					nextRef->nedge[1] = 2;
				}
				if (nv < nvert-1) {
					nextRef->neighbor[2] = tmpRef+(nhori-1)*2;;
					nextRef->nedge[2] = 0;
				}
				break;
			}

			tmpRef = nextRef->next;		// two triangles further
		}
	}

	// I think that is all - so return OK
	return 1;
}



// end of file
