/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* A. Tanju Erdem (Eastman Kodak Company)                             */
/* Candemir Toklu (University of Rochester)                           */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* Eastman Kodak Company and the University of Rochester retain full  */
/* right to use the code for their own purposes, assign or donate the */
/* code to a third party and to inhibit third parties from using the  */
/* code for non MPEG-4 Visual (ISO/IEC 14496-2) conforming products.  */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1994-1998                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         triTessClass.h
   AUTHOR       Peter van Beek - University of Rochester.
   DATE         Fri Aug  7 12:06:54 1998
   DESCRIPTION  Tri2dTess defines a tesselation, i.e. triangular mesh. 
                A mesh is formed from a list of points (Pt2dList) and a 
                linked list of triangles (Tri2dRefs). Pt2dList holds the 
                locations of the nodes of the mesh in a linked fashion.
                The member variable tess is a pointer to the first Tri2dRef
                in the patch list. Triangles of the mesh are linked to 
                their neighbors by pointers and also contain pointers to 
                their vertices. The vertices of triangles are always
                ordered in clockwise fashion and so are the pointers to
                the neighbors (see triRefClass.h).
                The class also has member functions to create a uniform
                mesh or create and update a Delaunay triangulated mesh. 
    SYNOPSIS    Include files: 
                #include <iostream.h>
                #include <fstream.h>
                #include <math.h>
                #include <time.h>
                #include "polyClass.h"
                #include "pointClass.h"
                #include "triRefClass.h"
                #include "listClass.h"
*********************************************************************-*/

#include <iostream.h>
#include <fstream.h>
#include <math.h>
#include <time.h>
#include "polyClass.h"
#include "pointClass.h"
#include "triRefClass.h"
#include "listClass.h"

#ifndef TRI_TESS_CLASS
#define TRI_TESS_CLASS


//------------------------------------------------------------------------
// Tesselation of triangles
#ifndef WIN32
class Tri2dTess {
#else
class __declspec(dllexport) Tri2dTess {
#endif
  public:
    Tri2dTess();
    Tri2dTess( const Tri2d&);
    Tri2dTess( const Tri2dTess&);
    ~Tri2dTess();
    Tri2dTess& destroy();
    Tri2dTess& operator=( const Tri2dTess&);
    Tri2dTess& append( Tri2d&);			// Add triangle to the mesh
	inline Point2d& point(int);			// Get i'th node point
    Tri2d triangle(int);				// Get i'th triangle
    Tri2dTess& move( double, double);
    Tri2dTess& move( const Point2d&);
    Tri2dTess& scale( double, double);
	Tri2dTess& quantize( double);		// Quantize coordinates
	void markRefs(const char);
	void markLinks(const char);
    int where( Pt2dLink*) const;  // Index of a point whose address is given
    int where( Tri2dRef*) const;  // Index of a triangle whose address is given
    int where( Point2d&) const;	  // Index of a point in the point list.
    Pt2dLink* addressC( const Point2d);  // Address of a point if it exits.
    Pt2dLink* addressC( int);  // Address of a point whose index is given
	Tri2dRef* addressR( int);  // Address of a triangle whose index is given
	Tri2dRef *nextTriangle(int first = 0); // Address of next triangle in list
    inline int getSizePoints();
    inline int getSize();
    int read( ifstream&);
    int write( ofstream&);
    int read(char []);
    int write(char []);
    void print();
	// Construct a triangular mesh with uniform topology
	int makeUniform(int nhori, int nvert, double hsize, double vsize, int splitcode); 
	// Construct a triangular mesh by triangulating a simple polygon
	void triangulate(Poly2d &poly);
	// Flip edges in a given triangular mesh until it is Delaunay
	int makeDelaunay ();
	// Locate a point in a mesh
	int locatePoint(Point2d &p, Tri2dRef *&tri, int &edge, int &onEdge);
	// Insert new point in a Delaunay mesh
	int insertDelaunaySite(Point2d &p);
  private:
	// private member functions
	int findEdge(Pt2dLink*, Pt2dLink*, Tri2dRef*&, int&);
	int delaunayEdge(Tri2dRef *tri1, int e1, Tri2dRef *tri2, int e2);
	void flipEdge(Tri2dRef *t1, Tri2dRef *t2, int e1, int e2);
  private:
	// private member variables
    int sizeTess;	  // Number of triangles in the mesh
    Tri2dRef* tess;   // Pointer to first triangle
    Pt2dList points;  // Contains list of mesh points
};

//--------------------------------------------------------------------------
int Tri2dTess::getSizePoints()
{
	return points.getSize();
}

//--------------------------------------------------------------------------
int Tri2dTess::getSize()
{
	return sizeTess;
}

//--------------------------------------------------------------------------
Point2d& Tri2dTess::point( int indx)
{
	return points[indx];
}

#endif
