/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* A. Patti (University of Rochester)                                 */
/* J. Riek (University of Rochester)                                  */
/*                                                                    */
/* and edited by                                                      */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* The University of Rochester retain full right to use the code for  */
/* their own purposes, assign or donate the code to a third party and */
/* to inhibit third parties from using the code for non MPEG-4 Visual */
/* (ISO/IEC 14496-2) conforming products.                             */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1993-1998                                            */
/*                                                                    */
/**********************************************************************/

/*+/////////////////////////////////////////////////////////////////////
   NAME         imageClass.cpp
   AUTHORS      A. Patti and J. Riek, University of Rochester
   DATE         July 7, 1993
   UPDATE       Peter Van Beek, Wed Nov  4 17:15:14 1998
   DESCRIPTION  Implementation of the image class defined in image.h.
                Contains basic image manipulation operators and
                various image processing operators.
                Image class has float pixel values. Images can be either
                greylevel or RGB color. 

   SYNOPSIS     See imageClass.h


///////////////////////////////////////////////////////////////////////-*/

#include "imageClass.h"


///////////////////////////////////////////////////////////////////////////
// Implementation of BaseObject 
///////////////////////////////////////////////////////////////////////////


// Toggle the bits (which) in flags.
void BaseObject::toggle_flags(long which)
{
	flags^=which;
}

// Set the bits (which) in flags to 0.
void BaseObject::unset_flags(long which)
{
	flags&=~which;
}

// Set the bits (which) in flags
void BaseObject::set_flags(long which)
{
	flags|=which;
}

// Return TRUE (1) if all the bits in which are set in flags.
int BaseObject::flags_set(long which)
{
	if((flags&which)==which) return TRUE;
	else return FALSE;
}



///////////////////////////////////////////////////////////////////////////
// Implementation of BaseImage 
///////////////////////////////////////////////////////////////////////////

// Constructor #1
BaseImage::BaseImage()
{
	height=width=0;
}


// Constructor #2
BaseImage::BaseImage(int w,int h)
{
	height=h;
	width=w;
}



///////////////////////////////////////////////////////////////////////////
// Implementation of image 
///////////////////////////////////////////////////////////////////////////

// Destructor routine to delete all dynamically allocated memeory.
image::~image()
{
	if(data!=NULL) { delete data; data=NULL;}
		
	if(cmap[0]!=NULL) { delete cmap[0]; cmap[0]=NULL;}
	if(cmap[1]!=NULL) { delete cmap[1]; cmap[1]=NULL; }
	if(cmap[2]!=NULL) { delete cmap[2]; cmap[2]=NULL; }
}


// Constructor #1
image::image() : BaseImage()
{
	depth=color=0;
	data=last=NULL;
	cmap[0]=cmap[1]=cmap[2]=NULL;
}

// Constructor #2
image::image(int w,int h) : BaseImage(w,h)
{
	depth=color=0;
	data=last=NULL;
	cmap[0]=cmap[1]=cmap[2]=NULL;
	if (!getmem(h*w)) die("Not enough memory in image constructor");
}

// Constructor #3
image::image(int w,int h, int d, int c) : BaseImage(w,h)
{
	depth=d; 
	color=c;
	data=last=NULL;
	cmap[0]=cmap[1]=cmap[2]=NULL;
	if (!getmem(h*w)) die("Not enough memory in image constructor");
}

// Copy Constructor
image::image(image& from)
{
	depth=color=0;
	data=last=NULL;
	cmap[0]=cmap[1]=cmap[2]=NULL;
	*this=from;
}


// Here is a function that will get memory of specific size if 
// it is available and return 1, otherwise it will return 0.
int image::getmem(int size)
{
	if(color) size*=3;
	if (data==NULL) {
		data=new float[size];
		if(data!=NULL) {last=data+size;return 1;}
		else {
			cerr<<"Couldn't allocate the "<<size*sizeof(float)<< " bytes requested.\n"<<flush;
			return (0);
		}
	}
	else if(last-data<size) {
		delete data; data=NULL;
		
		data=new float[size];
		if (data!=NULL) {last=data+size;return 1;}
		else {
			cerr<<"Couldn't allocate the "<<size*sizeof(float)<< " bytes requested.\n"<<flush;
			return (0);
		}
	}
	else return 1;
}


// Get memory for image with default size
int image::getmem()
{
	int size=width*height;
	if(color) size*=3;
	if (data==NULL) {
		data=new float[size];
		if(data!=NULL) {last=data+size;return 1;}
		else {
			cerr<<"Couldn't allocate the "<<size*sizeof(float)<< " bytes requested.\n"<<flush;
			return (0);
		}
	}
	else if(last-data<size) {
		delete data; data=NULL;
		
		data=new float[size];
		if (data!=NULL) {last=data+size;return 1;}
		else {
			cerr<<"Couldn't allocate the "<<size*sizeof(float)<< " bytes requested.\n"<<flush;
			return (0);
		}
	}
	else return 1;
}


// Get memory for image with specific attributes
int image::getmem(int w, int h, int d, int c)
{
	width  = w;
	height = h;
	depth  = d;
	color  = c;
	if (getmem() == 0)
		return 0;
	else
		return 1;
}


// Operator to access image data using x,y coordinates
float& image::operator()(int x,int y)
{
	if(x<0 || x>=width)  die("Gone beyond the edge of the image");
	if(y<0 || y>=height) die("Gone beyond the top/bottom of the image");
	return data[x+y*width];
}


// Operator to access image data using index
float& image::operator[] (int indx)
{
	if(last-data<=indx) die("Gone beyond the size of the image");
	if(indx<0)  die("Negative indices are not allowed for an image");
	return data[indx];
}


// The assignment operator is defined for assigning one image 
// to another.	We must declare this function since we have dynamically
// allocated memory. It has a return value for multiple assignments.
image& image::operator=(image& from)
{
	int size;
	width=from.width;
	height=from.height;
	color=from.color;
	depth=from.depth;
	size=width*height;
	if (!getmem()) die("Not enough memory in image=image");
	if(color) size*=3;
	for(int i=0;i<size;i++) data[i]=from.data[i];
	return *this;
}


image& image::operator-=(image& from)
{
	int size;
	if((width!=from.width) || (height!=from.height) || (color!=from.color)){
		die("Tried to use -= operator on images not of the same size");
	} 
	size=width*height*(1+2*color);
	for(int i=0;i<size;i++) data[i]-=from.data[i];
	return *this;
}

image& image::operator+=(image& from)
{
	int size;
	if((width!=from.width) || (height!=from.height) || (color!=from.color)){
		die("Tried to use += operator on images not of the same size");
	}
	size=width*height*(1+2*color);
	for(int i=0;i<size;i++) data[i]+=from.data[i];
	return *this;
}

image& image::operator=(float a)
{
	if (!getmem()) exit(1);
	for(int i=0; i<width*height; i++) data[i]=a;
	return *this;
}

image& image::operator+=(float a)
{
	for(int i=0; i<width*height; i++) data[i]+=a;
	return *this;
}

image& image::operator-=(float a)
{
	for(int i=0; i<width*height*(2*color+1); i++) data[i]-=a;
	return *this;
}

image& image::operator*=(float a)
{
	for(int i=0; i<width*height*(2*color+1); i++) data[i]*=a;
	return *this;
}

image& image::operator/=(float a)
{
	for(int i=0; i<width*height*(2*color+1); i++) data[i]/=a;
	return *this;
}


image operator+(image& a1, image& a2)
{
	if(a1.width!=a2.width || a1.height!=a2.height){
		die("can't add two images of different size, aborting...");	}
	image temp;
	temp=a1;
	temp+=a2;
	return temp;
}


image operator-(image& a1, image& a2)
{
	if(a1.width!=a2.width || a1.height!=a2.height){
		die("can't subtract two images of different size, aborting..." );	  }
	image temp;
	temp=a1;
	temp-=a2;
	return temp;
}


///////////////////////////////////////////////////////////////////////////
// Greylevel/colorlevel functions

// Quantize() linearly quantizes the values of this->data into
// the range [0,255].  It then sets this->depth=8.
void image::quantize(int ask)
{
	int i,size=width*height*(2*color+1);
	char reply[10];
	float min=data[0];
	float max=min;
	for(i=0;i<size;i++)
		if(data[i]<min) min=data[i];
		else if(data[i]>max) max=data[i];
	if(ask) {
		cout<<"Minimum value = "<<min<<", Maximum value = "<<max<<".\n";
		cout<<"Use these values? (y/n) ";
		cin>>reply;
		if(*reply!='y' && *reply!='Y') {
			cout<<"Enter new values for the minimum and maximum => ";
			cin>>min>>max;
		}
	}
	float fact=0;
	if(max!=min)  fact=255.0f/(max-min);
	for(i=0;i<size;i++) data[i]=(data[i]-min)*fact;
	depth=8*(2*color+1);
	return;
}


// Gamma correction
void image::gamma(float g)
{
	int i;
	float min,max;

	if(depth!=8 && depth!=32) {
		min=max=data[0];
		for(i=1;i<(1+2*color)*width*height;i++) {
			if(data[i]<min) min=data[i];
			else if(data[i]>max) max=data[i];
		}
	}
	else {min=0;max=255;}

	for(i=0;i<(1+2*color)*width*height;i++) {
		data[i]=min+max*(float)pow((data[i]-min)/(max-min),1.0/g);
	}
}


// Contrast stretch the image from its original [0, 255] range;
// Note that clipping occurs at low and high end of gray scale!
void image::stretch(float low, float upp)
{
	// scale factor
	upp = (upp - low) / 255.0f;

	int size = width*height*(color? 3: 1);
	for (int i=0; i<size; i++) {
		float val = low + upp * data[i];
		if (val < 0.0)
			data[i] = 0.0;
		else if (val > 255.0)
			data[i] = 255.0;
		else
			data[i] = val;
	}
	return;
} 


///////////////////////////////////////////////////////////////////////////
// Resizing functions

void image::resize(int x, int y)
{
	int i,j,il,jl,k;
	float scalex,scaley,tx,ty;
	float *tar,*tdata,*t2data;
	image ar;

	scalex=(float)width/x;
	scaley=(float)height/y;
	ar.color=color;
	ar.width=x;
	ar.height=y;
	ar.depth=depth;
	if(!ar.getmem(x*y)) die("Not enough memory in image::resize()");
	for(k=0;k<1+2*color;k++)
	{
		tar=(float *)ar+k*x*y;
		t2data=data+k*width*height;
		for(i=0;i<y;i++)
		{
			ty=scaley*i;
			il=(int)floor(ty);
			ty-=il;
			tdata=t2data+il*width;
			if(il<height-1){
				for(j=0;j<x;j++,tar++)
				{
					tx=scalex*j;
					jl=(int)floor(tx);
					tx-=jl;
					if(jl<width-1) *tar=tdata[jl]*(1-tx)*(1-ty)+tdata[jl+1]*tx*(1-ty)
						+tdata[width+jl]*(1-tx)*ty+tdata[width+jl+1]*tx*ty;
					else *tar=tdata[jl];
				}
			}
			else{
				for(j=0;j<x;j++,tar++) 
				{
					jl=(int)floor(scalex*j);
					*tar=tdata[jl];
				}
			}
		}
	}
	*this=ar;
	return;
}


///////////////////////////////////////////////////////////////////////////
// Functions related to maximum, minimum, magnitude

// Return the maximum pixel value
float image::maximum()
{
	int i,size=width*height*(2*color+1);
	float max=data[0];

	for(i=1;i<size;i++)
		if(data[i]>max) max=data[i];

	return(max);
}

// Return the minimum pixel value
float image::minimum()
{
	int i,size=width*height*(2*color+1);
	float min=data[0];

	for(i=1;i<size;i++)
		if(data[i]<min) min=data[i];

	return(min);
}

// Set pixels to absolute value (magnitude)
void image::mag()
{
	for(int i=0;i<height*width;i++) data[i] = (float) fabs(data[i]);
}


///////////////////////////////////////////////////////////////////////////
// Some basic functions
///////////////////////////////////////////////////////////////////////////


// Print a message and kill current process
void die(char *message)
{
	if(message)
		cerr<<message<<endl;
	else
		cerr<<"Something bad happened : Aborting"<<endl;

	// kill(getpid(),SIGSEGV);
	// Peter van Beek, Mon Mar  8 16:56:48 1999
	exit(1);
}


// Prompt user with question and return answern
double prompt(char *phrase, double def)
{
	char dummy[50];
	double rv;
	int num;

	num=cin.rdbuf()->in_avail();
	if(num>0) cin.ignore(1,'\n');
	cout<<phrase<<" ("<<def<<") => ";
	cin.get(dummy,20);
	if(strlen(dummy)==0) rv=def;
	else rv=atof(dummy);
	return rv;
}


// Prompt user with yes/no question and return boolean
int yes(char *question)
{
	char response[20];

	for(;;) {
		cout<<question<<" (y/n)? => ";
		cin>>response;
		if(*response=='Y' || *response=='y') return 1;
		else if(*response=='N' || *response=='n') return 0;
	}
}


// end of file
