/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* A. Patti (University of Rochester)                                 */
/* J. Riek (University of Rochester)                                  */
/*                                                                    */
/* and edited by                                                      */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* The University of Rochester retain full right to use the code for  */
/* their own purposes, assign or donate the code to a third party and */
/* to inhibit third parties from using the code for non MPEG-4 Visual */
/* (ISO/IEC 14496-2) conforming products.                             */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1993-1998                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         imageClass.h
   DESCRIPTION  Image class with float pixel values. Images can be either
                greylevel or RGB color. 
   AUTHORS      A. Patti and J. Riek, University of Rochester
   DATE         July 7, 1993
   UPDATE       Peter Van Beek, Wed Nov  4 17:15:23 1998
   DESCRIPTION  Header file for basic image class (see also imageClass.cpp).
***********************************************************************-*/

#ifndef IMAGE_CLASS
#define IMAGE_CLASS

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <fstream.h>
#include <iostream.h>
#include <signal.h>
#ifndef WIN32
#include <unistd.h>
#include <stream.h>
#include <sys/time.h>
#endif
 

// Forward class declarations
class BaseImage;
class image;


///////////////////////////////////////////////////////////////////////////
// General definitions
///////////////////////////////////////////////////////////////////////////

#ifndef NULL
#define NULL	0
#endif

#ifndef TRUE
#define TRUE  1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#ifndef OK
#define OK				1
#define NOT_OK	0
#endif
#ifndef MAXPIXVAL
#define MAXPIXVAL 255.0
#endif	MAXPIXVAL
 
#ifndef MINPIXVAL
#define MINPIXVAL	0.0
#endif	MINPIXVAL
 
#ifndef BYTE
#define BYTE	unsigned char
#endif
 
#define PIXEL		float		 // basic pixel type

#define MAX(a,b) ((a)<(b)?(b):(a))	// Define the maximum function
#define MIN(a,b)  ((a)>(b)?(b):(a)) // Define the minimum function
#define SGN(a) ((a)<0?-1:1) // Define the sgn function 
#define SQR(a) ((a)*(a))	// Define the square function
#define ABS(a) (a<0?-(a):(a))	// Redefine abs to be type insensitive
#define HYP(a,b) sqrt(SQR(a)+SQR(b)) // Hypotonuse of triangle of sides a,b.
#define FLIP_SWAP(a,b) {temp=a;a=b;b=temp;}


///////////////////////////////////////////////////////////////////////////
// Some typedefs
///////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////
// The class BaseObject is the class every other class is derived
// from. Having a BaseObject is useful when creating linked lists.
///////////////////////////////////////////////////////////////////////////

class BaseObject
{
  public:
	long flags;							// binary flags
  public:
	BaseObject() {flags=0;}
	~BaseObject() {}
	void set_flags(long which);			// Set bits in flags.
	void toggle_flags(long which);		// Toggle bits in flags
	void unset_flags(long which);		// Set bits to 0 in flags.
	int flags_set(long which);			// Are these bits set?
};



///////////////////////////////////////////////////////////////////////////
// BaseImage is the class which all other images are based upon.
// You should never declare anything of type BaseImage.
///////////////////////////////////////////////////////////////////////////

class BaseImage : public BaseObject
{
  public:
	int height,width;	 // height and width of image

	// This is the constructor.	 It is called whenever anything of type
	// BaseImage is declared. (e.g. BaseImage myimage;)
	BaseImage();

	// This is another constructor.	 It is called whenever anything of type
	// BaseImage is declared. (e.g. BaseImage myimage(256,256); )
	BaseImage(int w ,int h );
};



///////////////////////////////////////////////////////////////////////////
// Class image is derived from class BaseImage. It will have all the
// fields of BaseImage plus all the fields we declare here.
///////////////////////////////////////////////////////////////////////////

class image : public BaseImage
{
  public:								// used to be private
	// Colormap for i/o.
	unsigned char *(cmap[3]);
	// Data array pointer and memory management pointer.
	float *data,*last;

	// depth is the number of bits per pixel
	// color is 1 for a color image, 0 for grayscale.
	int depth,color;	

	// Used for memory allocation.	If you are synthesizing an image, 
	// either declare the image using  image myimage(256,256);
	// or declare it as image myimage; then set width and height 
	// and then call getmem e.g. as follows:
	// if(!myimage.getmem(256*256)) exit(1);
	int getmem(int size);
	int getmem();
	// Get memory for image with specific attributes:
	// image.getmem(width, height, depth, color);
	int getmem(int w, int h, int d, int c);

  public:
	// Returns number of pixels
	inline int size();

	// Constructor called when something declared of type image. BaseImage
	// constructor called first. (e.g. image myimage; )
	image();

	// Constructor called when something declared of type image. BaseImage
	// constructor called first. (e.g. image myimage(256,256); )
	// This constructor sets width and height, and allocates memory.
	image(int w, int h);

	// Constructor that sets width, height, depth and color and
	// allocates memory
	image(int w,int h, int d, int c);

	// This is the copy constructor; called when an image 
	// is returned from a function. i.e., as opposed
	// to using our equality operator.
	image(image&);

	// Destructor; called when a variable of type image goes out of scope.
	// Deallocates the memory used by the image.
	~image();

	// Redefine the [] operator to return an element of data.
	// (i.e. myimage[i] is the same as myimage.data[i])
	// Since data is protected, it cannot be directly accessed.
	float& operator[](int i);

	// Redefine the () operator to return an element of data.
	// (i.e. myimage(x,y) is the same as myimage.data[x+y*myimage.width])
	// Since data is protected, it cannot be directly accessed.
	float& operator()(int x,int y);

	// The assignment operator is redefined to equate two images.
	image& operator=(image&);
	
	image& operator-=(image&);
	image& operator+=(image&);
	friend image operator+(image& a1, image& a2);
	friend image operator-(image& a1, image& a2);

	// The assignment operator is redefined to equate all pixels of an image
	// to a constant value.
	image& operator=(float);

	image& operator+=(float);
	image& operator-=(float);
	image& operator*=(float);
	image& operator/=(float);

	// The equality operator is redefined for a floating point array.
	// e.g. image myimage; float ar[65536]; myimage=ar;
	// If you can avoid this, don't use it.	 It doesn't update the memory
	// allocation pointer, and it doesn't change the width and height.
	image& operator=(float *);

	// Redefine the typecast operator (float *) to return the image data.
	// Anytime you want a pointer to the data, do this:
	// float *temp; temp=myimage;  Or, if you have a subroutine
	// which expects a float *, then you can also pass in myimage
	// and the program will actually pass myimage.data.
	operator float*() {return data;}

	// Rescales the image to be in the range [0,255], and sets depth to 8.
	void quantize(int ask=1);

	// Contrast stretch the image from its original [0, 255] range;
	// Note that clipping occurs at low and high end of gray scale!
	void stretch(float low, float upp);

	// Correct for a screen with a gamma of g.
	void gamma(float g);

	// Resize an image to width w and height h - linearly interpolates
	// between points.
	void resize(int w,int h);

	// Takes the absolute value of the image.
	void mag();

	// find the maximum or minimum value in the image.
	float maximum();
	float minimum();

}; // class image



///////////////////////////////////////////////////////////////////////////
// inline functions
///////////////////////////////////////////////////////////////////////////


// Returns number of pixels in the image
int image::size()
{
	return ((color?3:1) * width * height);
}


///////////////////////////////////////////////////////////////////////////
// Prototypes of basic functions
///////////////////////////////////////////////////////////////////////////

void die(char *message);
double prompt(char*,double);
int yes(char*);


#endif // IMAGE_CLASS
