/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* The University of Rochester retains full right to use the code for */
/* their own purposes, assign or donate the code to a third party and */
/* to inhibit third parties from using the code for non MPEG-4 Visual */
/* (ISO/IEC 14496-2) conforming products.                             */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1996-1998                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         imageFunc.cpp
   AUTHOR       Peter van Beek, University of Rochester
   DATE         Wed Apr 17 10:32:32 1996
   UPDATE       Peter Van Beek, Wed Nov  4 18:05:50 1998
   DESCRIPTION  Functions for reading and writing single component video 
                or three-component (rgb or yuv) video data from/to raw files.
                Routines for drawing and clipping lines in a 2D array.
                Routine for morphological padding of image.

   SYNOPSIS     Functions:
                int read_im(image &im, char * fname, int frame)
                int write_im(image &im, char * fname, int frame)
                int read_yuv(image &y, image &u, image &v, char * fname, 
                    int frame)
                int write_yuv(image &y, image &u, image &v, char * fname, 
                    int frame)
                void im_draw_clip_line(image &im, int b, int x1, int y1, 
                    int x2, int y2, float colour)
                int im_padding(image &imi, image &mask, image &imo, int conn, 
                    int maxdist)
                int im_extract_alpha (image &imalpha, int index, int mode)
                double im_comp_psnr (image &yCurr, image &aCurr, image &yOrig, 
                    image &aOrig, double &totSSD, long &totPix)


*********************************************************************-*/

#include	"imageClass.h"		// UoR C++ image class
#include	"imageFunc.h"
#include	<malloc.h>




			/* ===== local function declarations ===== */

BYTE *read_component(FILE *fp, int s1, int s2, int s3, int frame, int type);
int write_component(BYTE *data, FILE *fp, int s1, int s2, int s3, int frame, int type);

void drawline (float *dp, int w, int h, int o, int x1, int y1, int x2, int y2, float value);
int clipline2box(float xmax, float ymax, float *x1, float *y1, float *x2, float *y2);
int cliptest(float p, float q, float *u1, float *u2);



			/* ===== function definitions ===== */


/*+==========================================================================
	NAME		read_im
	VERSION		1.0
	PURPOSE		Read single image component from raw file into an image. 
	I/O			im			image to read into
	INPUT		fname		filename
				frame		framenumber
	OUTPUT		none
	RETURNS		OK or NOT_OK
	NOTES		Sufficient memory in 'im' should be allocated already
==========================================================================-*/
int read_im(image &im, char * fname, int frame)
{
	FILE	*fp=NULL;
	BYTE	*dp=NULL;
	int		size, i;


	// determine frame size
	size = (im.width)*(im.height);

	// open file and read data bytes
	if((fp=fopen(fname, "rb")) == NULL) {
		(void) fprintf(stderr, "read_im: Can't open file\n");
		return(NOT_OK);
	}
	dp=read_component(fp, size, 0, 0, frame, T_S);
	if (dp == NULL) {
		(void) fprintf(stderr, "read_im: Error reading component\n");
		return(NOT_OK);
	}

	// copy and convert data to image
	for (i=0; i<size; i++)
		im[i] = (PIXEL) dp[i];

	// cleanup and return
	(void) free(dp);
	(void) fclose (fp);
	return (OK);

} // end read_im


/*+==========================================================================
	NAME		read_yuv
	VERSION		1.0
	PURPOSE		Read yuv image components from raw file into images. 
	I/O			y, u, v			images in which yuv components are put
	INPUT		fname			filename
				frame			framenumber
	RETURNS		OK or NOT_OK
	NOTES		
==========================================================================-*/
int read_yuv(image &y, image &u, image &v, char * fname, int frame)
{
	FILE	*fp;
	BYTE	*dpy=NULL, *dpu=NULL, *dpv=NULL;
	int		sizey, sizeuv, i;


	// determine data block size
	sizey  = (y.width)*(y.height);
	sizeuv = (u.width)*(u.height);

	// open file and read data bytes
	if((fp=fopen(fname, "rb")) == NULL) {
		(void) fprintf(stderr, "read_yuv: Can't open file\n");
		return(NOT_OK);
	}
	dpy=read_component(fp, sizey, sizeuv, sizeuv, frame, T_Y);
	dpu=read_component(fp, sizey, sizeuv, sizeuv, frame, T_U);
	dpv=read_component(fp, sizey, sizeuv, sizeuv, frame, T_V);
	if ((dpy == NULL) || (dpu == NULL) || (dpv == NULL)) {
		(void) fprintf(stderr, "read_yuv: Error reading components\n");
		return(NOT_OK);
	}

	// copy and convert data to image
	for (i=0; i<sizey; i++)
		y[i] = (PIXEL) dpy[i];
	for (i=0; i<sizeuv; i++)
		u[i] = (PIXEL) dpu[i];
	for (i=0; i<sizeuv; i++)
		v[i] = (PIXEL) dpv[i];

	// cleanup and return
	(void) free(dpy);
	(void) free(dpu);
	(void) free(dpv);
	(void) fclose (fp);
	return (OK);

} // end read_yuv 



/*+==========================================================================
	NAME		write_im
	VERSION		1.0
	PURPOSE		Write single component image into raw file. 
	INPUT		image		image to write
				fname		filename
				frame		framenumber (if < 0 append to end)
	OUTPUT		none
	RETURNS		OK or NOT_OK
	NOTES		
==========================================================================-*/
int write_im(image &im, char * fname, int frame)
{
	FILE	*fp;
	BYTE	*dp=NULL;
	int		size, i;
	register PIXEL	p;


	// determine data block size and allocate temporary memory space
	size  = (im.width)*(im.height);
	dp = (BYTE *) malloc ((unsigned) size);
	if (dp == NULL) {
		(void) fprintf(stderr, "write_im: Not enough memory...\n");
		return(NOT_OK);
	}
	// copy and convert data to image
	for (i=0; i<size; i++)
		dp[i] = (BYTE) ((p=im[i]) > (PIXEL)0.0 ? p+0.5 : p-0.5);

	// open file and write data bytes
	if (frame < 0)
		fp=fopen(fname, "ab");	// append
	else
		fp=fopen(fname, "wb");	// write
	if (fp==NULL) {
		(void) fprintf(stderr, "write_im: Can't open file\n");
		return(NOT_OK);
	}
	if (write_component(dp, fp, size, 0, 0, frame, T_S) != OK) {
		(void) fprintf(stderr, "write_im: Error writing data\n");
		return(NOT_OK);
	}

	// cleanup and return
	(void) free(dp);
	(void) fclose (fp);
	return (OK);

} // end write_im 



/*+==========================================================================
	NAME		write_yuv
	VERSION		1.0
	PURPOSE		Write yuv image components from images into raw file. 
	INPUT		y, u, v		images to write
				fname		filename
				frame		framenumber (if < 0 append to end)
	OUTPUT		none
	RETURNS		OK or NOT_OK
	NOTES		
==========================================================================-*/
int write_yuv(image &y, image &u, image &v, char * fname, int frame)
{
	FILE	*fp;
	BYTE	*dpy=NULL, *dpu=NULL, *dpv=NULL;
	int		sizey, sizeuv, i;
	register PIXEL	p;


	// determine data block size
	sizey  = (y.width)*(y.height);
	sizeuv = (u.width)*(u.height);

	// allocate temporary memory space
	dpy = (BYTE *) malloc ((unsigned) sizey);
	dpu = (BYTE *) malloc ((unsigned) sizeuv);
	dpv = (BYTE *) malloc ((unsigned) sizeuv);
	if ((dpy == NULL) || (dpu == NULL) || (dpv == NULL)) {
		(void) fprintf(stderr, "write_yuv: Not enough memory...\n");
		return(NOT_OK);
	}
	// copy and convert data to image
	for (i=0; i<sizey; i++)
		dpy[i] = (BYTE) ((p=y[i]) > (PIXEL)0.0 ? p+0.5 : p-0.5);
	for (i=0; i<sizeuv; i++)
		dpu[i] = (BYTE) ((p=u[i]) > (PIXEL)0.0 ? p+0.5 : p-0.5);
	for (i=0; i<sizeuv; i++)
		dpv[i] = (BYTE) ((p=v[i]) > (PIXEL)0.0 ? p+0.5 : p-0.5);

	// open file and write data bytes
	if (frame < 0)
		fp=fopen(fname, "ab");	// append
	else
		fp=fopen(fname, "wb");	// write
	if (fp==NULL) {
		(void) fprintf(stderr, "write_yuv: Can't open file\n");
		return(NOT_OK);
	}
	write_component(dpy, fp, sizey, sizeuv, sizeuv, frame, T_Y);
	write_component(dpu, fp, sizey, sizeuv, sizeuv, frame, T_U);
	write_component(dpv, fp, sizey, sizeuv, sizeuv, frame, T_V);

	// cleanup and return
	(void) free(dpy);
	(void) free(dpu);
	(void) free(dpv);
	(void) fclose (fp);
	return (OK);

} // end write_yuv 



/*+==========================================================================
	NAME		read_component
	PURPOSE		Read a block of image bytes from a file, which stores up to
				three components of image data per frame (e.g. yuv or rgb).
	METHOD		Allocate space, seek to correct frame and read.
	INPUT		fp			file pointer
				s1,s2,s3	size of first, second and third component
				frame		frame number
				type		which component
	OUTPUT		none
	RETURNS		a pointer to the newly read image data
	NOTES		
==========================================================================-*/
BYTE *read_component(FILE *fp, int s1, int s2, int s3, int frame, int type)
{
	BYTE	*data=NULL;			// data pointer
	long	offset;				// offset address in file
	int		num;				// number of bytes read
	int		size;				// sizes of data blocks


	// determine data block size and offset address
	switch(type) {
	case 1:				// first component
		offset = (long) OFFSET_HEADER + (s1+s2+s3)*frame;
		size = s1;
		break;
	case 2:				// second component
		offset = (long) OFFSET_HEADER + (s1+s2+s3)*frame+s1;
		size = s2;
		break;
	case 3:				// third component
		offset = (long) OFFSET_HEADER + (s1+s2+s3)*frame+s1+s2;
		size = s3;
		break;
	default:
		(void) fprintf(stderr, "read_component: Invalid data type\n");
		return(NULL);
	}

	// allocate memory
	if ((data = (BYTE *) malloc ((unsigned) size)) == NULL) {
		fprintf (stderr, "read_component: Not enough memory\n");
		return (NULL);
	}

	// read bytes from file
	if (fseek (fp, offset, SEEK_SET) != 0) {
		fprintf (stderr, "read_component: Error seeking file\n");
		return (NULL);
	}
	if ((num = fread ((char *)data, sizeof(BYTE), size, fp)) != size) {
		fprintf (stderr, "read_component: Error reading file\n");
		fprintf (stderr, "--> %d bytes read of %d\n", num, size);
		return (NULL);
	}

	// return data pointer
	return ((BYTE *) data);

} // end read_component 



/*+==========================================================================
	NAME		write_component
	PURPOSE		Write a block of image bytes into a file, which stores up to
				three components of image data per frame (e.g. yuv or rgb).
	METHOD		Seek correct position (if frame >= 0), then write.
	INPUT		data		pointer to image data
				fp			file pointer
				s1,s2,s3	size of first, second and third component
				frame		framenumber (if < 0 append to end)
				type		which component
	OUTPUT		none
	RETURNS		OK or NOT_OK
	NOTES		
==========================================================================-*/
int write_component(BYTE *data, FILE *fp, int s1, int s2, int s3, int frame, int type)
{
	long	offset;			// offset address in file
	int		num;			// number of bytes written
	int		size;			// sizes of data blocks


	// check data pointer
	if (data == (BYTE *)NULL) {
		fprintf (stderr, "write_component: Data lost...\n");
		return (NOT_OK);
	}

	// determine data block size and offset address
	switch(type) {
	case 1:	
		offset = (long) OFFSET_HEADER + (s1+s2+s3)*frame;
		size = s1;
		break;
	case 2:	
		offset = (long) OFFSET_HEADER + (s1+s2+s3)*frame+s1;
		size = s2;
		break;
	case 3:	
		offset = (long) OFFSET_HEADER + (s1+s2+s3)*frame+s1+s2;
		size = s3;
		break;
	default:
		(void) fprintf(stderr, "write_component: Invalid data type\n");
		return(NOT_OK);
	}

	// write bytes to file
	if (frame >= 0) {
		if (fseek (fp, offset, SEEK_SET) != 0)
			{
				fprintf (stderr, "write_component: Error seeking!\n");
				return (NOT_OK);
			}
	}
	if ((num = fwrite ((char *)data, sizeof(BYTE), size, fp)) != size) {
		fprintf (stderr, "write_component: Error writing to file\n");
		fprintf (stderr, "--> %d bytes written of %d\n", num, size);
		return (NOT_OK);
	}
	return (OK);

} // end write_component 



/*+==========================================================================
	NAME		im_draw_clip_line
	PURPOSE		Draw a straight line in an image. Clip line first.
	METHOD		
	INPUT		im				image
				b				band number
				x1,y1,x2,y2		line end points
				colour			drawing colour
	OUTPUT		im				image
	RETURNS		NONE
==========================================================================-*/
void im_draw_clip_line(image &im, int b, int x1, int y1, int x2, int y2, float colour)
{
	float i1, i2, j1, j2;
	int	  nx1, ny1, nx2, ny2, w, h;

	w = im.width;
	h = im.height;

	// clip and quantize the colour value if necessary
	if (colour > MAXPIXVAL)	 colour = MAXPIXVAL;
	if (colour < MINPIXVAL)	 colour = MINPIXVAL;
	switch (im.depth) {
	  case 1:  colour = (float) floor(colour/MAXPIXVAL+0.5); break;
	  case 8:  colour = (float) floor(colour+0.5); break;
	  default: break;
	}

	// clip the line to a bounding box defined by the image borders
	i1=(float) x1; i2=(float) y1;
	j1=(float) x2; j2=(float) y2;
	if (clipline2box((float)(w-1), (float)(h-1), &i1, &i2, &j1, &j2)) {
		// convert points back and do actual drawing
		nx1=(int)(i1+0.5); ny1=(int)(i2+0.5);
		nx2=(int)(j1+0.5); ny2=(int)(j2+0.5);
		drawline(im.data, w, h, b*w*h, nx1, ny1, nx2, ny2, colour);
	}

} // im_draw_clip_line


/*+==========================================================================
	NAME		drawline
	PURPOSE		Draw a straight line in an array.
	METHOD		Uses midpoint line algorithm, see Computer Graphics, Foley,
				van Dam, Feiner & Hughes, Addison Wesley, 1990
	INPUT		dp				data-pointer
				w,h				array dimensions
				o				array offset
				x1,y1,x2,y2		line end points
	OUTPUT		dp				data pointer
	RETURNS		NONE
==========================================================================-*/
void drawline (float *dp, int w, int h, int o, int x1, int y1, int x2, int y2, float value)
{
	register int	xi, yi, adx, ady, r1;
	int		hor, dx, dy, r2, r3;


	// initialize 
	hor = 1;

	// calculate horizontal and vertical displacements 
	dx = x2-x1;
	dy = y2-y1;
	adx = abs(dx);
	ady = abs(dy);

	// test for zero length 
	if((dx == 0) && (dy == 0)) {
		dp[o+y1*w+x1] = value;
		return;
	}

	// horizontal or vertical line ? 
	if(ady > adx)
		hor = 0;

	// calculate stepsizes and lengths of string 
	if(dx != 0)
		adx = adx/dx;
	if(dy != 0)
		ady = ady/dy;

	dx = abs(dx);
	dy = abs(dy);

	if(hor) {
		yi = y1;
		r1 = 2*dy-dx;
		r2 = 2*(dy-dx);
		r3 = 2*dy;

		if(adx > 0)
			for(xi=x1; xi <= x2; xi++) {
				dp[o+yi*w+xi] = value;

				if(r1 >= 0) {
					yi += ady;
					r1 += r2;
				}
				else 
					r1 += r3;
			}
		else
			for(xi=x1; xi >= x2; xi--) {
				dp[o+yi*w+xi] = value;

				if(r1 >= 0) {
					yi += ady;
					r1 += r2;
				}
				else
					r1 += r3;
			}
	}
	else {			// vertical line 
		xi = x1;
		r1 = 2*dx-dy;
		r2 = 2*(dx-dy);
		r3 = 2*dx;

		if(ady > 0)
			for(yi=y1; yi <= y2; yi++) {
				dp[o+yi*w+xi] = value;

				if(r1 >= 0) {
					xi += adx;
					r1 += r2;
				}
				else
					r1 += r3;
			}
		else
			for(yi=y1; yi >= y2; yi--) {
				dp[o+yi*w+xi] = value;

				if(r1 >= 0) {
					xi += adx;
					r1 += r2;
				}
				else
					r1 += r3;
			}
	} 
}


/*+==========================================================================
	NAME		clipline2box
	PURPOSE		Calculates new end points with respect to a bounding box.
	METHOD		The Liang-Barsky algorithm is used, see Computer Graphics, 
				Hearn & Baker, Prentice Hall, page 134. 
	INPUT		xmax, ymax		define bounding box (0,0)-(xmax,ymax)
				x1,y1,x2,y2		end points of line
	OUTPUT		x1,y1,x2,y2		recalculated end points
	RETURNS		0,				if the line defined by the end points is
								entirely outside the box
				1,				otherwise
==========================================================================-*/
int clipline2box(float xmax, float ymax, 
				 float *x1, float *y1, float *x2, float *y2)
{
	float	oldx1, oldx2, oldy1, oldy2;
	float	u1, u2, dx, dy;
	int outside=0;				// outside: not totally inside 


	oldx1= *x1; oldy1 = *y1;
	oldx2= *x2; oldy2 = *y2;
	
	outside = ((*x1 < 0) || (*x2 < 0) || (*y1 < 0) || (*y2 < 0) ||
			   (*x1 > xmax) || (*x2 > xmax) || (*y1 > ymax) || (*y2 > ymax));

	if (!outside)
		return(1);

	u1 = 0.0; u2 = 1.0;
	dx = (*x2) - (*x1);

	if (cliptest(-dx, *x1, &u1, &u2))
		if (cliptest(dx, xmax - *x1, &u1, &u2)) {
			dy = *y2 - *y1;

			if (cliptest(-dy, *y1, &u1, &u2))
				if (cliptest(dy, ymax - *y1, &u1, &u2)) {
					if (u2 < 1.0) {
						*x2 = *x1+u2*dx;
						*y2 = *y1+u2*dy;
					}
					if (u1 > 0.0) {
						*x1 = *x1+u1*dx;
						*y1 = *y1+u1*dy;
					}
				}
		}

	if (outside && (oldx1 == *x1) && (oldx2 == *x2) &&
	   (oldy1 == *y1) && (oldy2 == *y2))
		return(0);
	else
		return(1);
}


/*+==========================================================================
	NAME		cliptest
	PURPOSE		Test for rejection of line/edge intersection and compute
				new parameters if necessary. Used by the Liang-Barsky 
				clipping algorithm.
	INPUT		q, p			numerator/denominator
	OUTPUT		u1, u2			recalculated line parameters
	RETURNS		0,				if the line segment is rejected
				1				otherwise
==========================================================================-*/
int cliptest(float p, float q, float *u1, float *u2)
{
	float	r;
	int result=TRUE;

	if (p < 0.0) {
		r=q/p;
		if (r > *u2)
			result=FALSE;
		else if (r > *u1)
			*u1=r;
	}
	else if (p > 0.0) {
		r=q/p;
		if (r < *u1)
			result=FALSE;
		else if (r < *u2)
			*u2=r;
	}
	else	
		// p = 0, so line is parallel to this clipping edge
		if (q < 0.0)
			// line is outside this clipping edge
			result=FALSE;
	
	return(result);
}


/*+==========================================================================
	NAME		im_padding
	VERSION		1.0
	PURPOSE		Propagate the grey-levels of pixels, using a distance
				transform and averaged pixel dilations in a 3x3
				4-connected or 8-connected neighborhood.
	METHOD		First compute distance transform. Then, for each
				discrete distance, dilate the pixels from a lower
				distance (using averaging).
	INPUT		imi				seed input image
				mask			mask input image
				conn			connectivity (4 or 8)
				maxdist			propagated distance
	OUTPUT		imo				output image
	RETURNS		OK or NOT_OK
	NOTES		
==========================================================================-*/
int im_padding(image &imi, image &mask, image &imo, int conn, int maxdist)
{
	int		w, h, m, mw, n, index, i, j, k, kmax, iter, numval; 
	PIXEL	val, sumval, minval;


	// arrays containing row and column offset for neighborhood pixels
	int moff[8] = {1, 0, -1, 0, 1, 1, -1, -1};
	int noff[8] = {0, 1, 0, -1, 1, -1, -1, 1};

	// check input and initialize
	w  = (int) (imi.width);
	h  = (int) (imi.height);
	if ((w != (imo.width)) || (h != (imo.height)) ||
		(w != (mask.width)) || (h != (mask.height))) {
		fprintf(stderr, "im_padding: image sizes not equal!\n");
		return(NOT_OK);
	}
	switch (conn) {
	case 4: kmax = 4; break;
	case 8: kmax = 8; break;
	default:
		fprintf(stderr, "im_padding: connectivity must be 4 or 8!\n");
		return(NOT_OK);
	}

	image imdt(w,h,8,0);				// distance transform image

	// initialize
	for (index=0; index<w*h; index++) {
		if (mask[index] == (PIXEL) MINPIXVAL) {
			imdt[index] = (PIXEL) BIGNUMBER;
			imo[index] = (PIXEL) 0;
		}
		else {
			imdt[index] = (PIXEL) 0;
			imo[index] = imi[index];
		}
	}

	// compute distance transform in two passes
	for (m=0, mw = 0; m<h; m++, mw+=w)
		for (n=0; n<w; n++) {
			index = mw+n;
			minval = imdt[index];
			if (m > 0) {
				val = imdt[index-w] + DT_A;
				if (val < minval) minval = val;
				if (n > 0) {
					val = imdt[index-w-1] + DT_B;
					if (val < minval) minval = val;
				}
				if (n < w-1) {
					val = imdt[index-w+1] + DT_B;
					if (val < minval) minval = val;
				}
			}
			if (n > 0) {
				val = imdt[index-1] + DT_A;
				if (val < minval) minval = val;
			}
			imdt[index] = minval;
		} // end distance transform forward pass 

	for (m=h-1, mw = w*(h-1); m>=0; m--, mw-=w)
		for (n=w-1; n>=0; n--) {
			index = mw+n;
			minval = imdt[index];
			if (m < h-1) {
				val = imdt[index+w] + DT_A;
				if (val < minval) minval = val;
				if (n < w-1) {
					val = imdt[index+w+1] + DT_B;
					if (val < minval) minval = val;
				}
				if (n > 0) {
					val = imdt[index+w-1] + DT_B;
					if (val < minval) minval = val;
				}
			}
			if (n < w-1) {
				val = imdt[index+1] + DT_A;
				if (val < minval) minval = val;
			}
			imdt[index] = minval;
		} // end distance transform backward pass 

	// now dilate greylevel image layer by layer
	for (iter=1; iter <= maxdist; iter++) {
		index = 0;
		for (m=0; m<h; m++)
			for (n=0; n<w; n++, index++) {
				// only process pixels in current distance band
				if ((int)imdt[index] == iter) {
					// average lower distance pixels in a neighborhood 
					sumval = (PIXEL)0;
					numval = 0;
					for (k=0; k<kmax; k++) {
						i = m+moff[k];
						j = n+noff[k];
						if (i<0 || i>=h || j<0 || j>=w)
							continue;
						if (imdt(j,i) < iter) {
							numval++;
							sumval += (PIXEL) imo(j,i);
						}
					} // end for neighborhood loop
					if (numval != 0)
						imo[index] = (PIXEL) (sumval/numval);
				}
			} // end image loop
	} // end dilation loop

	return(OK);

} // im_padding



// extracts the alpha plane with label 'index' from an image and
// returns the number of pixels with that label
// if mode is 1, label is ignored: any non-zero pixel means alpha=1
int im_extract_alpha (image &imalpha, int index, int mode)
{
	int voppixels = 0;					// number of pixels inside VOP

	if (mode == 0) {
		for (int i=0; i<imalpha.width*imalpha.height; i++) {
			if (imalpha[i] == index) {
				imalpha[i] = 255.0;
				voppixels++;
			}
			else
				imalpha[i] = 0.0;
		}
	}
	else {
		for (int i=0; i<imalpha.width*imalpha.height; i++) {
			if (imalpha[i] >= ALPHA_TH) {
				imalpha[i] = 255.0;
				voppixels++;
			}
			else
				imalpha[i] = 0.0;
		}
	}
	return voppixels;
}


// Compute PSNR between two images on pixels inside both alpha planes;
// the PSNR is returned; variables for total SSD and total number of
// pixels are updated as well.
double im_comp_psnr (image &yCurr, image &aCurr, image &yOrig, image &aOrig, double &totSSD, long &totPix)
{
	int	   n = 0;
	double ssd = 0.0;
	int	   npix = yCurr.width*yCurr.height;

	// sum square errors between current and original pixels
	for (int i=0; i<npix; i++) {
		if (aCurr[i] >= ALPHA_TH  &&  aOrig[i] >= ALPHA_TH) {
			double diff = (double) (yCurr[i] - yOrig[i]);
			ssd += (diff*diff);
			n++;
		}
	}
	// update total values
	totSSD += (double) ssd;
	totPix += (long) n;

	// compute and return psnr
	if (n == 0	||	ssd == 0.0)
		return (99.99);
	else
		return (log10 ((double)n * 255. * 255. / ssd) * 10.0);
}


// end of file 
