/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* A. Tanju Erdem (Eastman Kodak Company)                             */
/* Candemir Toklu (University of Rochester)                           */
/*                                                                    */
/* and edited by                                                      */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* Eastman Kodak Company and the University of Rochester retain full  */
/* right to use the code for their own purposes, assign or donate the */
/* code to a third party and to inhibit third parties from using the  */
/* code for non MPEG-4 Visual (ISO/IEC 14496-2) conforming products.  */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1994-1998                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         mappingFunc.cpp
   AUTHOR       A. Tanju Erdem - Eastman Kodak Company and 
                Candemir Toklu - University of Rochester.
   DATE         September 1994
   UPDATE       14 August 1996 by Candemir Toklu
                Peter Van Beek, Sun Sep 22 13:05:08 1996
                - added mapTessImage routine
                Peter Van Beek, Thu Nov  5 10:49:28 1998
                - changed to new mesh datastructure
   DESCRIPTION  In this file there are several functions for warping 
                intensity inside one polygon to another polygon using
                affine motion.

   FUNCTIONS    int mapTessImage(image&, image&, Tri2dTess&, Tri2dTess&, 
                  HexSearchPar)
                int mapRegion( ParAffine, image&, image&, Poly2d, Poly2d, int,
                  mapDirection, interpMethod, double, double);
                int mapRegion( ParAffine, image&, image&, Poly2d, int,
                  mapDirection, interpMethod, double, double);
                void interpNearestN( ParAffine, image&, image&, int, int, int,
                  skip, double, double);
                void interpBilinear( ParAffine, image&, image&, int i, int, 
                  int, int, double, double);
                void interpAreawise( ParAffine, image&, image&, int, int, 
                  int, int, Poly2d&, double, double);

   SYNOPSIS     Include files : 
                #include "warpingProto.h"
                              
   ASSUMPTIONS  The polygons should be convex.

*********************************************************************-*/

#include "warpingProto.h"


//------------------------------------------------------------------------
// Global variables
int BORDERS_AREAWISE = 0;   // If 1, borders of a polygon are interpolated
                            // areawise (useful when bilinear or nearest-
                            // neighbor interpolation is used).
int BACKGROUND_COLOR[3] = { 0, 0, 0};  // Background color in mapRegion()
int USE_MASK = 0;           // Masking out image regions during hex search
char* MASK = NULL;			// The mask bytes

int CHECK_CONVEX = 1;       // If 1, the polygon is tested for convexity.
int CHECK_PROPER = 1;       // If 1, the polygon is tested for properness.
int CHECK_VALID = 1;        // If 1, the polygon is tested for validness.
int CHECK_SIZE = 1;         // If 1, the polygons are tested for matching size



//------------------------------------------------------------------------
// mapTessImage
//------------------------------------------------------------------------
// This function maps all image regions in (triangular) polygons in
// the source tesselation (mesh) to their corresponding regions in the
// target image, specified by the target tesselation (mesh). 
// This function in effect motion compensates the regions in the image
// inside the tesselation/mesh. Uses 'mapRegion' to do the mapping.
// Assumes that the source and target mesh have the same number of points
// and triangles and their structure corresponds.
// Note that if the global variable USE_MASK is set, the data that MASK
// (another global variable) points to will be used as a mask, specifying
// which pixels should actually be updated in the target image.
//------------------------------------------------------------------------
// Parameters:	Type:  Comment:
// frSource		I	   source image, pixels will be mapped toward target
// frTarget		O	   target image, pixels will be updated from source
// sourceTess	I	   source triangular mesh
// targetTess	I	   target triangular mesh
// intMet		I	   interpolation method - see parClass
//------------------------------------------------------------------------
// Returns:		0/1	   status
//------------------------------------------------------------------------
// Update:		Peter Van Beek, Thu Nov	 5 10:50:53 1998
//------------------------------------------------------------------------
int mapTessImage(image& frSource, image& frTarget, Tri2dTess& sourceTess, 
				 Tri2dTess& targetTess, interpMethod intMet)
{
	// mapping direction - only backward implemented
	mapDirection mapDir = MBACKWARD;
	// skip should always be 1 here
	int skip = 1;
	// no intensity scaling functionality
	double base = 0.0;
	double mult = 1.0;
	// source and target polygons in the source/target tesselations
	Poly2d polySource, polyTarget;
	// mapping parameers
	ParAffine par;

	int n = sourceTess.getSize();
	if (n != targetTess.getSize()) {
		cout << "\nmapTessImage: tesselations have different sizes!\n";
		return 0;
	}
	for (int i=0; i < n; i++) {
		// get source and target polygons (triangles)
		polySource = sourceTess.triangle(i);
		polyTarget = targetTess.triangle(i);
		if (CHECK_PROPER){
			if( !(polySource.isProper()) ) {
				cout << "Warning (mapTessImage): skipping improper source triangle..." << endl;
				continue;
			}
			if ( !(polyTarget.isProper()) ) {
				cout << "Warning (mapTessImage): skipping target triangle..." << endl;
				continue;
			}
		}
		// map source image region to target image region
		if (mapRegion(par, frSource, frTarget, polySource, polyTarget, skip,
					  mapDir, intMet, base, mult) == 0)
			cout << "Warning (mapTessImage): mapRegion failed!" << endl;

	}
	// return OK
	return 1;

} // mapTessImage



//------------------------------------------------------------------------
// Map the region polySource in frSource into polyTarget in frTarget.
// If mapDir is MFORWARD, then the motion parameters are from frSource to  
// frTarget; otherwise, i.e., if mapDir is MBACKWARD, then the motion  
// parameters are from frTarget to frSource. The motion type only is 
// passed in by the pointer par; the motion parameters are returned in par.
//------------------------------------------------------------------------
int mapRegion( ParAffine par, image& frSource, image& frTarget, 
			   Poly2d polySource, Poly2d polyTarget, int skip,
			   mapDirection mapDir, interpMethod intMet,
			   double base, double mult)
{

	Poly2d polyFrom;
	switch( mapDir) {
	case MFORWARD:
		// Find motion parameters from Source to Target polygon
		if( ! par.getPars( polySource, polyTarget)) {
			cout << "\nCould not find the Forward mapping parameters.";
			par.printType();
			cout << "\nSource and Target polygons are as follows:";
			polySource.print();
			polyTarget.print();
			return 0;
		}
		polyFrom = polySource;
		break;
	case MBACKWARD:
		// Find motion parameters from Target to Source polygon
		if( ! par.getPars( polyTarget, polySource)) {
			cout << "\nCould not find the Backward mapping parameters.";
			par.printType();
			cout << "\nSource and Target polygons are as follows:";
			polySource.print();
			polyTarget.print();
			return 0;
		}
		polyFrom = polyTarget;
		break;
	default:
		cout << "\nMapping direction not allowed!. No mapping done!\n";
		return 0;
	}
	return mapRegion( par, frSource, frTarget, polyFrom, 
					  skip, mapDir, intMet, base, mult);
}

//------------------------------------------------------------------------
// Fill in a region in frTarget using the image data provided by frSource. 
// The parameter mapDir determines the direction of the motion parameters 
// and whether the region specified by polyFrom is in frSource or frTarget.
// If mapDir is MFORWARD, then polyFrom is in frSource and the motion	
// parameters are from frSource to frTarget; otherwise, i.e., if mapDir is 
// MBACKWARD, polyFrom is in frTarget and the motion parameters are from 
// frTarget to frSource. The motion type and parameters are passed in by 
// the pointer par. Note that by the preceding description, one always uses 
// par.tform(polyFrom) to find the corresponding polygon. Once again, 
// this function always fills in a region in frTarget.
//------------------------------------------------------------------------
int mapRegion( ParAffine par, image& frSource, image& frTarget, 
			   Poly2d polyFrom, int skip,
			   mapDirection mapDir, interpMethod intMet,
			   double base, double mult)
{
	Poly2d polySource;
	Poly2d polyTarget;
	switch( mapDir) {
	case MFORWARD:
		// Find Target polygon
		polySource = polyFrom;
		polyTarget = par.tform( polySource);
		break;
	case MBACKWARD:
		// Find Source polygon
		polyTarget = polyFrom;
		polySource = par.tform( polyTarget);
		break;
	default:
		cout << "\nMapping direction not allowed!. No mapping done!\n";
		return 0;
	}

	int minMtarget, minNtarget, maxMtarget, maxNtarget;

	// Find the smallest rectangle covering the target polygon
	smallestRect( polyTarget, minMtarget, minNtarget, maxMtarget, maxNtarget,
				  frTarget.width, frTarget.height);

	int tmpminM = minMtarget;
	int tmpmaxM = maxMtarget;
	int tmpminmaxM, tmpmaxminM;
	int targetIsRectangle = polyTarget.isRectangle();
	if( targetIsRectangle && (tmpminM % skip)) {
		tmpminM = skip * (1 + (tmpminM / skip));
	}

	if( mapDir == MBACKWARD ) {
		// Warping of pixels in the source image frSource starts here

		int i;
		switch( intMet) {
		case NEAREST:
			// Nearest neighbor interpolation
			if( (minNtarget % skip)) {
				minNtarget = skip * (1 + (minNtarget / skip));
			}
			for( i = minNtarget; i <= maxNtarget; i += skip) {
				if( ! targetIsRectangle) {
					tmpminM = minMtarget;
					tmpmaxM = maxMtarget;
					if( skip == 1 && BORDERS_AREAWISE) {
						decide(tmpminM, tmpmaxM, tmpminmaxM, tmpmaxminM, polyTarget, i);
						interpAreawise( par, frSource, frTarget, i, tmpminM, tmpminmaxM,
										1, polyTarget, base, mult);
						interpAreawise( par, frSource, frTarget, i, tmpmaxminM, tmpmaxM,
										1, polyTarget, base, mult);
						tmpminM = tmpminmaxM + 1;
						tmpmaxM = tmpmaxminM - 1;
					}
					else if( ! decide(tmpminM, tmpmaxM, polyTarget, skip, i)) 
						continue;
				}
				interpNearestN( par, frSource, frTarget, i, tmpminM, tmpmaxM,
								skip, base, mult);
			}
			break;

		case BILINEAR:
			// Bilinear interpolation
			if( (minNtarget % skip)) {
				minNtarget = skip * (1 + (minNtarget / skip));
			}
			for( i = minNtarget; i <= maxNtarget; i += skip) {
				if( ! targetIsRectangle) {
					tmpminM = minMtarget;
					tmpmaxM = maxMtarget;
					if( skip == 1 && BORDERS_AREAWISE) {
						decide(tmpminM, tmpmaxM, tmpminmaxM, tmpmaxminM, polyTarget, i);
						interpAreawise( par, frSource, frTarget, i, tmpminM, tmpminmaxM,
										1, polyTarget, base, mult);
						interpAreawise( par, frSource, frTarget, i, tmpmaxminM, tmpmaxM,
										1, polyTarget, base, mult);
						tmpminM = tmpminmaxM + 1;
						tmpmaxM = tmpmaxminM - 1;
					}
					else if( ! decide(tmpminM, tmpmaxM, polyTarget, skip, i)) 
						continue;
				}
				interpBilinear( par, frSource, frTarget, i, tmpminM, tmpmaxM,
								skip, base, mult);
			}
			break;

		case AREAWISE:
			// Area-weighted interpolation
			if( (minNtarget % skip)) {
				minNtarget = skip * (1 + (minNtarget / skip));
			}
			for( i = minNtarget; i <= maxNtarget; i += skip) {
				if( ! targetIsRectangle) {
					tmpminM = minMtarget;
					tmpmaxM = maxMtarget;
					if( skip == 1)
						decide( tmpminM, tmpmaxM, tmpminmaxM, tmpmaxminM, polyTarget, i);
					if( ! decide(tmpminM, tmpmaxM, polyTarget, skip, i)) 
						continue;
				}
				interpAreawise( par, frSource, frTarget, i, tmpminM, tmpmaxM,
								skip, polyTarget, base, mult);
			}
			break;

		} // switch
		return 1;
	}
	else if( mapDir == MFORWARD) {
		cout << "\nError: MFORWARD mapping has not been implemented yet!" 
			 << endl;	   
		exit(0);
	}
	return 0;
}

//------------------------------------------------------------------------
//
//------------------------------------------------------------------------
void interpNearestN( ParAffine par, image& frSource, image& frTarget, 
					 int i, int tmpminM, int tmpmaxM,
					 int skip, double base, double mult)
{
	int sizeBandTarget = frTarget.width * frTarget.height;
	int sizeBandSource = frSource.width * frSource.height;
	Point2d pointTarget;
	Point2d pointSource;
	int j, m, n, band, loc;

	for( j = tmpminM; j <= tmpmaxM; j += skip) {
		pointTarget = point( j, i);
		pointSource = par.tform( pointTarget);  // Map to the source
		m = (int) floor( pointSource.X() + 0.5);
		n = (int) floor( pointSource.Y() + 0.5);
		if( m < 0 || n < 0 || m >= frSource.width || n >= frSource.height) {
			// Outside the source image area
			for( band = 0; band < ((frSource.color)?3:1); band++) {
				frTarget.data[ i*frTarget.width + j + band*sizeBandTarget] 
					= (float)BACKGROUND_COLOR[band];
			}
		}
		else {
			loc = n*frSource.width + m;
			if( USE_MASK) {
				if(!MASK[loc]) continue;
			}
			for( band = 0; band < ((frSource.color)?3:1); band++) {
				frTarget.data[ i*frTarget.width + j + band*sizeBandTarget]
					= (float) (base + mult * 
					frSource.data[ loc + band*sizeBandSource]);
			}
		}
	}
}

//------------------------------------------------------------------------
//	  A +---------+ B
// (m,n)|		  |(m+1,n)
//		|  + I	  |
//		| (x,y)	  |
//		|		  |
//	  C +---------+ D
//	(m,n+1)	   (m+1,n+1)
//
//	if( m == frSource.width-1+(int)(1-xMm)) 
//      m = frSource.width-2+(int)(1-xMm);
//	if( n == frSource.height-1+(int)(1-yMn)) 
//      n = frSource.height-2+(int)(1-yMn);
//------------------------------------------------------------------------
void interpBilinear( ParAffine par, image& frSource, image& frTarget, 
					 int i, int tmpminM, int tmpmaxM,
					 int skip, double base, double mult)
{
	int sizeBandTarget = frTarget.width * frTarget.height;
	int sizeBandSource = frSource.width * frSource.height;
	Point2d pointTarget;
	Point2d pointSource;
	int j, m, n, band, loc, loc0;
	double xMm, yMn;
	double xI, yI;
	double valA, valBmA, valCmA, valI;

	for( j = tmpminM; j <= tmpmaxM; j += skip) {
		pointTarget = point( j, i);
		pointSource = par.tform( pointTarget);  // Map to the source
		xI = pointSource.X();
		yI = pointSource.Y();
		xMm = xI - ( m = (int) floor(xI) );
		yMn = yI - ( n = (int) floor(yI) );
		if( xI < -0.5 || yI < -0.5 || xI > frSource.width-0.5
			|| yI > frSource.height-0.5) {
			for( band = 0; band < ((frSource.color)?3:1); band++) {
				frTarget.data[ i*frTarget.width + j + band*sizeBandTarget] 
					= (float) BACKGROUND_COLOR[band];
			}
		}
		else {
			if( m == -1) { m = 0; xMm = 0;}
			else if( m == frSource.width-1) xMm = 0;
			if( n == -1) { n = 0; yMn = 0;}
			else if( n == frSource.height-1) yMn = 0;
			loc0 = frSource.width * n + m;
			if( USE_MASK) {
				if( (!MASK[loc0]) && (xMm < 0.5) && (yMn < 0.5)) 
					continue;
				if( (!MASK[loc0+frSource.width]) && (xMm < 0.5) && (yMn > 0.5))
					continue;
				if( (!MASK[loc0+1]) && (xMm > 0.5) && (yMn < 0.5)) 
					continue;
				if( (!MASK[loc0+frSource.width+1]) && (xMm > 0.5)
					&& (yMn > 0.5)) 
					continue;
			}
			for( band = 0; band < ((frSource.color)?3:1); band++) {
				loc = band * sizeBandSource + loc0;
				// Initial value of interpolation
				valI = (valA = frSource.data[loc]);
				if( xMm != 0.0) {
					valI += xMm * ( valBmA = frSource.data[loc+1] - valA);
				}
				if( yMn != 0.0) {
					valI += yMn * ( valCmA = frSource.data[loc+frSource.width] - valA);
					if( xMm != 0.0) {
						valI += xMm * yMn * ( frSource.data[loc+frSource.width+1] 
											  - valA - valBmA - valCmA);
					}
				}
				frTarget.data[ band*sizeBandTarget + frTarget.width*i + j] 
					= (float) (base + mult * valI);
			}
		}
	}
}

//------------------------------------------------------------------------
// Note that USE_MASK is not incorporated.
//------------------------------------------------------------------------
void interpAreawise( ParAffine par, image& frSource, image& frTarget, 
					 int i, int tmpminM, int tmpmaxM,
					 int skip, Poly2d& polyTarget, double base, double mult)
{
	int sizeBandTarget = frTarget.width * frTarget.height;
	int sizeBandSource = frSource.width * frSource.height;
	Poly2d cellSource;
	Poly2d cellTarget;
	Poly2d warpedCellTarget;

	double sum[3];
	double areaCell;
	double weight;
	int j, m, n, band, loc;

	int minMsource, minNsource, maxMsource, maxNsource;

	for( j = tmpminM; j <= tmpmaxM; j += skip) {
		cellTarget = intersection( cellAroundPixel( j, i), polyTarget, 0);
		if( (areaCell = cellTarget.area()) != 0) {
			// Map target cell to the source
			cellTarget = par.tform( cellTarget);  
			// Find the smallest rectangle covering mapped polygon in the source image
			smallestRect( cellTarget, minMsource, minNsource, maxMsource, 
						  maxNsource, frSource.width, frSource.height);

			for( band = 0; band < ((frSource.color)?3:1); band++) {
				sum[band] = 0.0;
			}
			for( n = minNsource; n <= maxNsource; n++) {
				for( m = minMsource; m <= maxMsource; m++) {
					cellSource = cellAroundPixel( m, n);
					weight = intersection( cellSource, cellTarget, 0).area();
					if (weight != 0) {
						for( band = 0; band < ((frSource.color)?3:1); band++) {
							sum[band] += weight * frSource.data[ n*frSource.width + m + band*sizeBandSource];
						}
					}
					//else
					//{
					//	for( band = 0; band < ((frSource.color)?3:1); band++)
					//	{
					//	  sum[band] = BACKGROUND_COLOR[band] * cellTarget.area();
					//	}
					//}
				}
			}
			for( band = 0; band < ((frSource.color)?3:1); band++) {
				loc = i*frTarget.width + j + band*sizeBandTarget;
				if( areaCell == 1) {
					frTarget.data[loc] = (float)(base + mult * sum[band]/cellTarget.area());
				}
				else {
					frTarget.data[loc] = (float)
						(( 1 - areaCell)*frTarget.data[loc] + areaCell*
						 ( base + mult * sum[band]/cellTarget.area())); 
				}
			}
		}
	}
}


//------------------------------------------------------------------------
int decide(int& minMtarget, int& maxMtarget, Poly2d& polyTarget, int skip,
		   int k)
{
	Point2d pt1, pt2;
	int intnum;

	intnum = getEndPoints( double(k), polyTarget, pt1, pt2);
	if( intnum == 2) {
		if( pt1.X() > pt2.X()) {
			minMtarget = (int) ceil(pt2.X()-EPSILON);
			maxMtarget = (int) floor(pt1.X()+EPSILON);
		}
		else {
			minMtarget = (int) ceil(pt1.X()-EPSILON);
			maxMtarget = (int) floor(pt2.X()+EPSILON);
		}	  
	}
	else if( intnum == 1) { 
		if( fabs(pt1.X() - floor(pt1.X()+.5)) < EPSILON)
			{ minMtarget = maxMtarget = (int) floor(pt1.X()+.5);}
		else { minMtarget = (int) ceil(pt1.X()); maxMtarget = minMtarget - 1;}
	}

	if( (minMtarget % skip)) 
		minMtarget = skip * (1 + (minMtarget / skip));

	return intnum;
}

//------------------------------------------------------------------------
// Decide version 2
//------------------------------------------------------------------------
int decide(int& minMtarget, int& maxMtarget, 
		   int& minmaxMtarget, int& maxminMtarget, 
		   Poly2d& polyTarget, int k)
{
	Point2d pt1Top, pt2Top;
	Point2d pt1Bot, pt2Bot;
	Point2d tmpPt;
	double xMin, xMax, xMinmax, xMaxmin;
	int intnumTop;
	int intnumBot;

	intnumTop = getEndPoints( k-0.5, polyTarget, pt1Top, pt2Top);
	if( intnumTop == 2) {
		if( pt2Top.X() < pt1Top.X()) {
			tmpPt = pt1Top;
			pt1Top = pt2Top;
			pt2Top = tmpPt;
		}
	}
	else if( intnumTop == 1) {
		pt2Top = pt1Top;
	}

	intnumBot = getEndPoints( k+0.5, polyTarget, pt1Bot, pt2Bot);
	if( intnumBot == 2) {
		if( pt2Bot.X() < pt1Bot.X()) {
			tmpPt = pt1Bot;
			pt1Bot = pt2Bot;
			pt2Bot = tmpPt;
		}
	}
	else if( intnumBot == 1) {
		pt2Bot = pt1Bot;
	}

	if( intnumTop + intnumBot > 2) {
		if( pt1Top.X() < pt1Bot.X()) { 
			xMin = pt1Top.X(); 
			xMinmax = pt1Bot.X();
		}
		else { 
			xMin = pt1Bot.X(); 
			xMinmax = pt1Top.X();
		}
		if( pt2Top.X() > pt2Bot.X()) { 
			xMax = pt2Top.X(); 
			xMaxmin = pt2Bot.X();
		}
		else { 
			xMax = pt2Bot.X(); 
			xMaxmin = pt2Top.X();
		}
	}
	else if( intnumTop == 2 && intnumBot == 0) {
		xMin = pt1Top.X();
		xMax = pt2Top.X();
		xMinmax = xMaxmin = xMax;
	}
	else if( intnumTop == 0 && intnumBot == 2) {
		xMin = pt1Bot.X();
		xMax = pt2Bot.X();
		xMinmax = xMaxmin = xMax;
	}
	else {
		// Peter van Beek, Mon Mar  8 11:13:31 1999
		cerr << "I don't think I should reach this point!" << endl;
		xMin = xMax = xMinmax = xMaxmin = 0.0;
	}

	if( intnumTop == 2 || intnumBot == 2) {
		minMtarget = (int) floor(xMin + 0.5 + EPSILON);
		maxMtarget = (int) ceil(xMax - 0.5 - EPSILON);
		minmaxMtarget = (int) ceil(xMinmax - 0.5 - EPSILON);
		maxminMtarget = (int) floor(xMaxmin + 0.5 + EPSILON);
		if( maxminMtarget <= minmaxMtarget) 
			maxminMtarget = minmaxMtarget+1;
	}
	else {
		minmaxMtarget = maxMtarget;
		maxminMtarget = maxMtarget + 1;
	}

	return( intnumTop == 2 || intnumBot == 2);
}



// end of file
