/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* The University of Rochester retains full right to use the code for */
/* their own purposes, assign or donate the code to a third party and */
/* to inhibit third parties from using the code for non MPEG-4 Visual */
/* (ISO/IEC 14496-2) conforming products.                             */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1996-1999                                            */
/*                                                                    */
/**********************************************************************/

/*+***************************************************************************
   NAME         mesh_cmp.cpp
   AUTHOR/DATE  Peter van Beek, Tue Mar  2 15:53:30 1999

   DESCRIPTION  Comparison of two sets of files with 2D mesh sequence
                data. Intended to compare one original mesh sequence 
                with another mesh sequence, e.g. a decoded version with
                a reference version. 
                The meshes of the both sets must be in ".tri" file 
                format (see the README file). The filenames of the
                mesh sequence is assumed to contain the frame number
                at the end of the base-part, e.g. akiyo15.tri. The
                program assumes the numbering of the files in the two
                sequences is the same.

   NOTES        This program takes into account the degenerate but
                possible case of two node points in a single mesh
                having the same location (e.g. due to bad tracking).
                This program does not take into account the case
                of two triangles in the same mesh having nodes with 
                the exact same location.

***************************************************************************-*/

#include "general.h"
#include "pointClass.h"
#include "polyClass.h"
#include "triTessClass.h"
#include <string.h>


////////// LOCAL FUNCTIONS ///////////////////////////////////////////////////

static void usage(char *name);
static int findTriangle(Tri2d &tri, Tri2dTess &mesh);
static int findTriangleNew(Tri2dRef *tri, Tri2dTess &mesh);


////////// MAIN FUNCTION /////////////////////////////////////////////////////


int main (int argc, char *argv[])
{
	// Mesh file basenames, e.g. "akiyo", "bream"
	char baseName1[100], baseName2[100];
	// Full filename
	char fileName[256];
	// Original mesh and other mesh
	Tri2dTess mesh1, mesh2;
	// First frame, last frame and frame increment
	int	 firstFrame, lastFrame, frameInc;
	// Frame number being decoded
	int	 frame1;
	// Offset to apply to second mesh
	double xoff, yoff;
	// Number of nodes/triangles
	int nNode, nTri;
	// A point and a triangle
	Point2d node;
	Tri2d tri;
	// Some arrays
	int *nodeCheck = NULL;
	int *triCheck = NULL;
	// Option flags
	int offsetOption, uniqOption, verboseOption;
	// Various
	int n, index, counter;


	// check number of command-line arguments and read them in
	if (argc < 6) {
		usage(argv[0]);
		return 1;
	}
	sprintf (baseName1, "%s", argv [1]);
	sprintf (baseName2, "%s", argv [2]);
	firstFrame	= atoi(argv[3]);
	lastFrame	= atoi(argv[4]);
	frameInc	= atoi(argv[5]);

	// other -optional- arguments
	xoff = yoff = 0;
	offsetOption = uniqOption = verboseOption = 0;
	n = 6;
	while (n < argc) {
		if (!strcmp(argv[n], "-o")) {
			offsetOption = 1;
			xoff = atof(argv[n+1]);
			yoff = atof(argv[n+2]);
			n+=3;
		}
		else if (!strcmp(argv[n], "-u")) {
			uniqOption = 1;
			n+=1;
		}
		else if (!strcmp(argv[n], "-v")) {
			verboseOption = 1;
			n+=1;
		}
		else {
			usage(argv[0]);
			return 1;
		}
	}

	
	/////////////// FRAME LOOP ////////////////////////////////////////
	
	for (frame1=firstFrame; frame1 <= lastFrame; frame1+=frameInc) {

		// read mesh data of first (reference) set and second (test) set
		sprintf(fileName, "%s%d.tri", baseName1, frame1);
		mesh1.read(fileName);
		cout << "----- Comparing " << fileName;
		sprintf(fileName, "%s%d.tri", baseName2, frame1);
		mesh2.read(fileName);
		cout << " with " << fileName << " -----" << endl;

		// check number of nodes/triangles
		if (mesh1.getSizePoints() != mesh2.getSizePoints()) {
			cout << ">> Number of nodes not equal!" << endl;
			cout << ">> Number of nodes in reference mesh is ";
			cout << mesh1.getSizePoints() << endl;
			cout << ">> Number of nodes in test mesh is ";
			cout << mesh2.getSizePoints() << endl;
			return 1;
		}
		if (mesh1.getSize() != mesh2.getSize()) {
			cout << ">> Number of triangles not equal!" << endl;
			cout << ">> Number of triangles in reference mesh is ";
			cout << mesh1.getSize() << endl;
			cout << ">> Number of triangles in test mesh is ";
			cout << mesh2.getSize() << endl;
			return 1;
		}

		nNode = mesh1.getSizePoints();
		nTri  = mesh1.getSize();

		// allocate data for flags
		if (frame1 == firstFrame) {
			nodeCheck = new int [nNode];
			triCheck = new int[nTri];
		}

		if (offsetOption) {
			// apply offset to second mesh
			mesh2.move(xoff, yoff);
		}
		if (uniqOption) {
			// check for multiple nodes in test mesh with same location
			counter = 0;
			for (n=0; n<nNode-1; n++) {
				node = mesh2.point(n);
				// search this node point in same mesh
				for (index = n+1; index < nNode; index++) {
					if (node == mesh2.point(index)) {
						cout << ">> Node " << n << " and node " << index;
						cout << " in test mesh have the same location at ";
						cout << node.X() << "," << node.Y() << " !" << endl;
						counter++;
					}						
				}
			}
			if (counter > 0) {
				cout << ">> Counted " << counter << " uniqueness violation(s)";
				cout << endl;
			}
		}

		// check if for each node in mesh 2, there is a node in mesh 1
		for (n=0; n<nNode; n++)
			nodeCheck[n] = 0;
		counter = 0;
		for (n=0; n<nNode; n++) {
			node = mesh2.point(n);
			// search this node point in first mesh
			for (index = 0; index < nNode; index++) {
				if (node == mesh1.point(index))
					// beware of multiple node points with same location;
					// this shouldn't happen but is possible
					if (nodeCheck[index] == 0) { 
						nodeCheck[index] = 1;
						break;
					}
			}
			if (index == nNode) {
				if (verboseOption) {
					cout << "Node in second but not in first mesh: ";
					cout << node << endl;
				}
				counter++;
			}
		}
		if (counter > 0) {
			cout << ">> Counted " << counter << " bad node(s) in test mesh";
			cout << endl;
		}

		// check if for each node in mesh 1, there is a node in mesh 2
		counter = 0;
		for (n=0; n<nNode; n++) {
			if (nodeCheck[n] == 0) {
				if (verboseOption) {
					node = mesh1.point(n);
					cout << "Node in first but not in second mesh: ";
					cout << node << endl;
				}
				counter++;
			}
		}
		if (counter > 0) {
			cout << ">> Counted " << counter << " missing node(s)";
			cout << endl;
		}

		// check if for each triangle in mesh 2, there is a triangle in mesh 1
		for (n=0; n<nTri; n++)
			triCheck[n] = 0;
		counter = 0;
		Tri2dRef *triPtr = mesh2.nextTriangle(1);
		for (n=0; n<nTri; n++) {
//			tri = mesh2.triangle(n);
//			index = findTriangle(tri, mesh1);
			index = findTriangleNew(triPtr, mesh1);
			if (index != -1)
				triCheck[index] = 1;
			else {
				if (verboseOption) {
					cout << "Triangle in second but not in first mesh: ";
					tri.print();
				}
				counter++;
			}
			triPtr = mesh2.nextTriangle(0);
		}
		if (counter > 0) {
			cout << ">> Counted " << counter << " bad triangle(s) in test mesh";
			cout << endl;
		}

		// check if for each triangle in mesh 1, there is a triangle in mesh 2
		counter = 0;
		for (n=0; n<nTri; n++) {
			if (triCheck[n] == 0) {
				if (verboseOption) {
					tri = mesh1.triangle(n);
					cout << "Triangle in first but not in second mesh: ";
					tri.print();
				}
				counter++;
			}
		}
		if (counter > 0) {
			cout << ">> Counted " << counter << " missing triangle(s)";
			cout << endl;
		}

	} // endfor (frame loop)

	// return OK
	if (nodeCheck != NULL)
		delete [] nodeCheck;
	if (triCheck != NULL)
		delete [] triCheck;
	return 0;

} // main



// Try to find a matching triangle in a triangular mesh;
// Return index of matching triangle or -1 if none found.
int findTriangle(Tri2d &tri, Tri2dTess &mesh)
{
	for (int m = 0; m < mesh.getSize(); m++) {
		Tri2d meshTri = mesh.triangle(m);
		Point2d p = meshTri[0];
		if (tri[0] == p) {
			if ((tri[1] == meshTri[1] && tri[2] == meshTri[2]) ||
				(tri[2] == meshTri[1] && tri[1] == meshTri[2]))
				return m;
		} else if (tri[1] == p) {
			if ((tri[2] == meshTri[1] && tri[0] == meshTri[2]) ||
				(tri[0] == meshTri[1] && tri[2] == meshTri[2]))
				return m;
		} else if (tri[2] == p) {
			if ((tri[0] == meshTri[1] && tri[1] == meshTri[2]) ||
				(tri[1] == meshTri[1] && tri[0] == meshTri[2]))
				return m;
		}
	}
	return -1;
}

// Try to find a matching triangle in a triangular mesh;
// Return index of matching triangle or -1 if none found.
int findTriangleNew(Tri2dRef *tri, Tri2dTess &mesh)
{
	Point2d p0 = tri->getLink(0)->getPoint();
	Point2d p1 = tri->getLink(1)->getPoint();
	Point2d p2 = tri->getLink(2)->getPoint();
	Tri2dRef *meshTri = mesh.nextTriangle(1);
	for (int m = 0; m < mesh.getSize(); m++) {
		Point2d p = meshTri->getLink(0)->getPoint();
		if (p0 == p) {
			if ((p1 == meshTri->getLink(1)->getPoint() && 
				 p2 == meshTri->getLink(2)->getPoint()) ||
				(p2 == meshTri->getLink(1)->getPoint() && 
				 p1 == meshTri->getLink(2)->getPoint()))
				return m;
		} else if (p1 == p) {
			if ((p2 == meshTri->getLink(1)->getPoint() && 
				 p0 == meshTri->getLink(2)->getPoint()) ||
				(p0 == meshTri->getLink(1)->getPoint() && 
				 p2 == meshTri->getLink(2)->getPoint()))
				return m;
		} else if (p2 == p) {
			if ((p0 == meshTri->getLink(1)->getPoint() && 
				 p1 == meshTri->getLink(2)->getPoint()) ||
				(p1 == meshTri->getLink(1)->getPoint() && 
				 p0 == meshTri->getLink(2)->getPoint()))
				return m;
		}
		meshTri = mesh.nextTriangle(0);
	}
	return -1;
}


// Print program usage and arguments
void usage(char *name)
{
	cerr << "Comparison of 2D meshes with reference 2D meshes" << endl;
	cerr << "Usage: " << name << " <basename1> <basename2> <firstframe> <lastframe> <frame-inc> [-o <xo> <yo>] [-u] [-v]" << endl;
	cerr << "basename1:   basename of reference mesh-files" << endl;
	cerr << "basename2:   basename of test mesh-files" << endl;
	cerr << "firstframe:  number of first frame of mesh sequence" << endl;
	cerr << "lastframe:   number of last frame of mesh sequence" << endl;
	cerr << "frame-inc:   frame increment of mesh sequence" << endl;
	cerr << "-o xo,yo:    offset to be applied to test meshes" << endl;
	cerr << "-u:          force uniqueness of mesh object plane vertices" << endl;
	cerr << "-v:          verbose output" << endl;
}


// end of file
