/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* The University of Rochester retains full right to use the code for */
/* their own purposes, assign or donate the code to a third party and */
/* to inhibit third parties from using the code for non MPEG-4 Visual */
/* (ISO/IEC 14496-2) conforming products.                             */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1996-1999                                            */
/*                                                                    */
/**********************************************************************/

/*+***************************************************************************
   NAME         mesh_render.cpp
   AUTHOR       Peter van Beek
   UPDATE       Peter van Beek, Wed Mar  3 17:18:47 1999

   DESCRIPTION  Rendering a moving video-object (VO) over a number
                of frames, from given texture and mesh-mappings.

                A video object plane (VOP) refers to the texture and 
                alpha planes of a VO at one particular time instant. 

                A parameter file is used to communicate all necessary
                parameters to the program.

                Types of rendering:
                0 = self-mapping: mapping the first frame of the original
                    sequence to all other frames of the same sequence;
                1 = transfiguration: mapping a replacement object to all
                    frames of the original sequence (replacing the original);
                2 = augmentation: mapping an augmentation object over all
                    frames of the original sequence (overlaid on original);

                In general, the rendering programs uses tracking data
                of a moving video-object to generate (an approximation
                to that video-object) from its first frame/VOP only. This
                is called self-mapping. This first frame is called the
                reference frame in this file. 
                The tracking data is in the form of a sequence of 2-D 
                triangular meshes, each describing the geometry of a 2-D 
                mesh at a certain point in time.
                At the moment, all the nodes in the mesh are present from
                the first to the last time instant (i.e. no nodes are ever
                added or deleted). 
                Optionally, the rendering can be augmented with another
                image/VOP, such that it inherits the (deformable) motion
                of the original video object. This is called augmentation
                and is performed by first using other tracking data that
                describes the mapping from the augmentation VO to the 
                reference frame/VOP of the original VO. Then, again the
                reference frame is rendered for the other time instants.
                We call transfiguration the act of replacing the entire
                original VO with the rendering of another image/VOP.

                PSNR of the rendered luma frames can be computed, in case
                of self-mapping. This requires one to have to full original
                data (not just the first frame used in texture mapping).

   NOTES        Assumes 4:2:0 sampling of yuv texture input.

                It is possible to overlay the meshes on the output
                images by setting the appropriate parameter in the
                parameter file. Currently, only luminance is overlaid.

   LITERATURE   C. Toklu, A.T. Erdem, M.I. Sezan and A.M. Tekalp, Tracking
                motion and intensity variations using hierarchical 2-D mesh
                modeling. Graphical Models and Image Processing, vol. 58,
                no. 6, pp. 553-573, November 1996

                C. Toklu, A.T. Erdem, M.I. Sezan and A.M. Tekalp, 2-D Mesh 
                tracking for synthetic transfiguration. Proceedings Int. 
                Conference on Image Processing, 1995

***************************************************************************-*/

#include "imageClass.h"					// UoR C++ image class
#include "warpingProto.h"				// UoR C++ mesh classes
#include "imageFunc.h"					// some image functions

#include <stdio.h>
#include <iostream.h>
#include <fstream.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>


////////// COMPILER SWITCHES /////////////////////////////////////////////////

// to get some basic useful messages
#define DEBUGGING

// Define debug printing macro which allows variable number of arguments
#ifdef DEBUGGING
#define DPRINT(vargs)				((void) printf vargs)
#else
#define DPRINT(vargs)
#endif

// Define to let output sequence have same frame rate as original sequence
#define ORIG_RATE

// Define to compute the PSNR of self-rendered luma frames (except first)
//#define DO_PSNR



////////// CONSTANT PARAMETERS ///////////////////////////////////////////////




////////// GLOBAL VARIABLE INPUT PARAMETERS //////////////////////////////////

// Full filename of original yuv and alpha file and mesh mappings
char yuvFile[200], alphaFile[200], meshDatFile[200];
// Width & height of original input frames
int w, h;
// First and last frames of the sequence to be processed
int firstFrame, lastFrame;
// Frame increment determining temporal frame rate
int frameInc;
// Index (label) of original video object in the segmentation file
int indexVO;

// Full filename of new (augmentation) yuv and alpha file and mesh mappings
char yuvNewFile[200], alphaNewFile[200], meshNewFile[200];
// Width & height of new (augmentation) input frames
int wNew, hNew;
// Index (label) of new (augmentation) video object in the segmentation file
int indexNewVO;

// Video object extraction mode
int alphaMode = 0;

// Full filename of output (rendered) yuv texture and alpha file
char outImFile[200], outAlphaFile[200];

// Type of rendering
// 0: self-mapping,	 1: transfiguration,  2: augmentation
int renderType = 0;

// Draw mesh overlay on output
int drawMesh = 0;



////////// GLOBAL VARIABLES //////////////////////////////////////////////////

// Number of frames in texture planes file and in alpha planes file
int	 yuvFrames=1, alphaFrames=1;

// Reference mesh
Tri2dTess meshRef;
// New (augmentation/transfiguration) meshes
Tri2dTess meshNew, meshNewTo;
// Current mesh
Tri2dTess meshCurr;
// Next mesh
Tri2dTess meshNext;

// Pixel interpolation method
interpMethod intMet = BILINEAR;

// Meshes for chrominance components
Tri2dTess meshRefChrom;
Tri2dTess meshNewChrom, meshNewToChrom;
Tri2dTess meshCurrChrom;
Tri2dTess meshNextChrom;

#ifdef DO_PSNR
// Signal-to-Noise ratio (running-average) and number of frames processed
double currSNR = 0;
int	 numframe = 0;
// Total Sum-of-Squared-Differences and number of pixels counted
double totSSD = 0.0;
long   totPix = 0;
#endif


////////// EXTERNAL VARIABLES ////////////////////////////////////////////////

// Parameters used in mapping
extern int USE_MASK;
extern char* MASK;


////////// LOCAL FUNCTIONS ///////////////////////////////////////////////////

static void nextvalidline (FILE *fpara, uint &pline);
static void read_para_file (char *fname);
static void im_draw_mesh (image &im, Tri2dTess &m, float *color);



////////// MAIN FUNCTION /////////////////////////////////////////////////////

// This is it, folks!
int main (int argc, char *argv[])
{
	int	 frame;							// frame number being processed
	int	 stat = 1;						// return status
	int	 visibleVO;						// VOP visible or not?
	char fileName[200];
	int	 i, j;


	// check parameter file argument
	if (argc < 2) {
		fprintf(stderr, "2-D mesh-based video-object rendering\n");
		fprintf(stderr, "Usage: %s <parameter-file>\n", argv[0]);
		exit (1);
	}

	time_t	 timeval = time(NULL);
	printf("======================================================\n");
	printf("2-D mesh-based video-object rendering ==== version 2.0\n");
	printf("%s\n", ctime(&timeval));
	printf("======================================================\n\n\n");

	// read parameter file and set parameters
	read_para_file (argv[1]);

	printf("======================================================\n\n");

	// declare and allocate memory for reference image data
	image yRefVOP (w, h, 8, 0);			// reference luminance y
	image uRefVOP (w/2, h/2, 8, 0);		// chrominance u
	image vRefVOP (w/2, h/2, 8, 0);		// chrominance v
	image aRefVOP (w, h, 8, 0);			// luminance alpha plane
	// declare and allocate memory for current image data
	image yCurrVOP (w, h, 8, 0);		// current luminance y
	image uCurrVOP (w/2, h/2, 8, 0);	// chrominance u
	image vCurrVOP (w/2, h/2, 8, 0);	// chrominance v
	image aCurrVOP (w, h, 8, 0);		// luminance alpha plane
#ifdef DO_PSNR
	// declare and allocate memory for original image data
	image yOrigVOP (w, h, 8, 0);		// original luminance y
	image uOrigVOP (w/2, h/2, 8, 0);	// chrominance u
	image vOrigVOP (w/2, h/2, 8, 0);	// chrominance v
	image aOrigVOP (w, h, 8, 0);		// luminance alpha plane
#endif
	// declare and allocate memory for new (augmentation) image data
	image yNewVOP (wNew, hNew, 8, 0);		// luminance y
	image uNewVOP (wNew/2, hNew/2, 8, 0);	// chrominance u
	image vNewVOP (wNew/2, hNew/2, 8, 0);	// chrominance v
	image aNewVOP (wNew, hNew, 8, 0);		// luminance alpha plane

	/////////////// READ INITIAL IMAGE AND MESH DATA //////////////////

	if (renderType == 0	 || renderType == 2) {
		// self-mapping or augmentation of original with new object
		DPRINT(("Reading reference image data ...\n"));

		// read first (reference) original yuv and alpha planes
		if (read_yuv(yRefVOP, uRefVOP, vRefVOP, yuvFile, (firstFrame%yuvFrames)) != OK) {
			fprintf(stderr, "unable to read y,u,v planes ... \n");
			assert (FALSE);
		}
		if (read_im(aRefVOP, alphaFile, (firstFrame%alphaFrames)) != OK) {
			fprintf(stderr, "unable to read alpha plane ... \n");
			assert (FALSE);
		}
		// extract reference video object plane alpha map
		visibleVO = im_extract_alpha (aRefVOP, indexVO, alphaMode);
		if (visibleVO == 0) {
			fprintf(stderr, "no video object in this frame ... \n");
			assert (FALSE);
		}
		// pad pixels of reference luma frame which have zero alpha value
		im_padding(yRefVOP, aRefVOP, yCurrVOP, 8, 8);
		yRefVOP = yCurrVOP;
	}

	// read mesh data of first (reference) frame
	sprintf(fileName, "%s%d.tri", meshDatFile, firstFrame);
	DPRINT(("Reading reference mesh from %s ...\n", fileName));
	meshRef.read(fileName);
	// change to pixel-centered coordinate system
	meshRef.move(-0.5, -0.5);

	// create mesh for chrominance components of first (reference) frame
	meshRefChrom = meshRef;
	meshRefChrom.move(-0.5, -0.5);
	meshRefChrom.scale(0.5, 0.5);

	if (renderType == 1	 || renderType == 2) {
		// transfiguration, augmentation or animation mode
		DPRINT(("Reading augmentation image data ...\n"));

		// read new (augmentation) yuv and alpha planes
		if (read_yuv(yNewVOP, uNewVOP, vNewVOP, yuvNewFile, 0) != OK) {
			fprintf(stderr, "unable to read y,u,v planes ... \n");
			assert (FALSE);
		}
		if (read_im(aNewVOP, alphaNewFile, 0) != OK) {
			fprintf(stderr, "unable to read alpha plane ... \n");
			assert (FALSE);
		}
		// extract video object plane
		visibleVO = im_extract_alpha (aNewVOP, indexNewVO, alphaMode);
		if (visibleVO == 0) {
			fprintf(stderr, "no video object in this frame ... \n");
			assert (FALSE);
		}

		// read mesh data defining augmentation mapping
		sprintf(fileName, "%s_from.tri", meshNewFile);
		DPRINT(("Reading object mesh from %s ...\n", fileName));
		meshNew.read(fileName);
		sprintf(fileName, "%s_to.tri", meshNewFile);
		DPRINT(("Reading object mesh from %s ...\n", fileName));
		meshNewTo.read(fileName);

		// create meshes for chrominance components 
		meshNewChrom = meshNew;
		meshNewChrom.move(-0.5, -0.5);
		meshNewChrom.scale(0.5, 0.5);
		meshNewToChrom = meshNewTo;
		meshNewToChrom.move (-0.5, -0.5);
		meshNewToChrom.scale(0.5, 0.5);
	}

	/////////////// PRE-MAPPING ///////////////////////////////////////

	if (renderType == 1	 || renderType == 2) {
		// map new (augmentation/transfiguration) image to reference image
		printf("\nMapping augmentation VOP to reference VOP %d ...\n\n", firstFrame);

		if (renderType == 1) {
			// transfiguration: initialize reference yuv image with zeros
			yRefVOP = 0.0;
			uRefVOP = 128.0;
			vRefVOP = 128.0;
			aRefVOP = 0.0;
		}

		// here we have to use a mask to map augmentation object only
		USE_MASK = 1;
		MASK = new char [wNew*hNew];
		for (i=0; i<wNew*hNew; i++)
			MASK[i] = (char) aNewVOP[i];

		// forward warping of replacement y and a
		stat &= mapTessImage(yNewVOP, yRefVOP, meshNew, meshNewTo, intMet);
		if (renderType == 1)
			stat &= mapTessImage(aNewVOP, aRefVOP, meshNew, meshNewTo, intMet);

		// create quarter sized mask - this is tricky
		delete [] MASK;
		MASK = new char [(wNew/2)*(hNew/2)];
		
		for (i=0; i<wNew*hNew/4; i++)
			MASK[i] = (char) 0;
		for (i=0; i<hNew; i++)
			for (j=0; j<wNew; j++) {
				if (aNewVOP[i*wNew+j] == 255.0)
					MASK[(i/2)*(wNew/2)+j/2] = (char) 255;
			}

		// forward warping of replacement u and v
		stat &= mapTessImage(uNewVOP, uRefVOP, meshNewChrom, meshNewToChrom, intMet);
		stat &= mapTessImage(vNewVOP, vRefVOP, meshNewChrom, meshNewToChrom, intMet);

		// don't use mask hereafter
		USE_MASK = 0;
		delete [] MASK;
	}

	/////////////// FRAME LOOP ////////////////////////////////////////

	// from the first to the last frame do
	for (frame = firstFrame; frame <= lastFrame; frame += frameInc)
	{
		printf("\n--------------------------------------------------------\n");

		// read mesh data of current frame
		sprintf(fileName, "%s%d.tri", meshDatFile, frame);
		DPRINT(("Reading object mesh from %s ...\n", fileName));
		meshCurr.read(fileName);
		// change to pixel-centered coordinate system
		meshCurr.move(-0.5, -0.5);

		// create mesh for chrominance components of current frame
		meshCurrChrom = meshCurr;
		meshCurrChrom.move(-0.5, -0.5);
		meshCurrChrom.scale(0.5, 0.5);
		
		// initialize output yuv image with black or grey background
		yCurrVOP = 128.0;
//		yCurrVOP = 0.0;
		uCurrVOP = 128.0;
		vCurrVOP = 128.0;
		// initialize output alpha plane with zeros
		aCurrVOP = 0.0;

		// forward warping of reference y, u and v
		printf("Mapping reference VOP to VOP %d ...\n", frame);
		stat &= mapTessImage(yRefVOP, yCurrVOP, meshRef, meshCurr, intMet);
		stat &= mapTessImage(uRefVOP, uCurrVOP, meshRefChrom, meshCurrChrom, intMet);
		stat &= mapTessImage(vRefVOP, vCurrVOP, meshRefChrom, meshCurrChrom, intMet);
		stat &= mapTessImage(aRefVOP, aCurrVOP, meshRef, meshCurr, intMet);
		assert (stat == 1);

#ifdef DO_PSNR
		// PSNR calculation - skip first frame
		if (renderType == 0	 &&	 frame > firstFrame) {
			// read original yuv and alpha planes
			read_yuv(yOrigVOP, uOrigVOP, vOrigVOP, yuvFile, frame);
			read_im(aOrigVOP, alphaFile, frame%alphaFrames);
			visibleVO = im_extract_alpha (aOrigVOP, indexVO, alphaMode);
			if (visibleVO == 0)
				fprintf(stderr, "No video object in this frame ... \n");

			// compute PSNR of current luma and update running average
			numframe++;
			double SNR = im_comp_psnr (yCurrVOP, aCurrVOP, yOrigVOP, aOrigVOP, totSSD, totPix);
			DPRINT(("PSNR of current frame		 : %lf\n", SNR));
			currSNR = ((double)(numframe-1)*currSNR + SNR)/(double)numframe;
			DPRINT(("Current average PSNR		 : %lf\n", currSNR));
			double tmp = 10.0 * log10((255.*255.*(double)totPix) / totSSD);
			DPRINT(("Current PSNR over all frames: %lf\n", tmp));
		}
#endif

		if (drawMesh) {
			// draw mesh in contrast-stretched luminance image
			DPRINT(("Drawing mesh over rendered y image ...\n"));
			float  fcol = 250.0;
			yCurrVOP.stretch (0.0, 208.0);
			im_draw_mesh (yCurrVOP, meshCurr, &fcol);
		}

		// append y, u and v and alpha planes to output image file
		DPRINT(("Saving rendered yuv and seg images of VOP %d ...\n", frame));
		if (frame==firstFrame) {		// create new file
			write_yuv(yCurrVOP, uCurrVOP, vCurrVOP, outImFile, 0);
			write_im(aCurrVOP, outAlphaFile, 0);
		}
		else {							// append
			write_yuv(yCurrVOP, uCurrVOP, vCurrVOP, outImFile, -1);
			write_im(aCurrVOP, outAlphaFile, -1);
		}
#ifdef ORIG_RATE
		// append more frames to get to original frame rate
		for (int n=1; n<frameInc; n++) {
			write_yuv(yCurrVOP, uCurrVOP, vCurrVOP, outImFile, -1);
			write_im(aCurrVOP, outAlphaFile, -1);
		}
#endif
	} // endfor (frame loop)

	printf("\n\n======================================================\n\n");

	////////// cleanup any junk and exit //////////////////////////////////
	return 0;

} // main



// read the parameter file with the name 'fname' and sets all global
// parameter variables according to the input
void read_para_file (char *fname)
{
	FILE *fpara;						// parameter file
	uint  pline = 0;					// line number in parameter file
	char  texturefile[200];				// original texture planes filename
	char  alphafile[200];				// original alpha planes filename
	char  mapfile[200];					// mappings of original VO
	char  indir[200];					// input file directory
	char  outfile[200];					// output file name 
	char  outdir[200];					// output file directory
	struct stat stat_buf;
	int	  fsize;
	int	  idummy;


	if ((fpara = fopen (fname, "r")) == NULL ){
		fprintf (stderr, "parameter file not found... \n");
		exit (1);
	}
	DPRINT(("Reading parameter file...\n"));

	// original texture input file name
	nextvalidline (fpara, pline);
	fscanf (fpara, "%s", texturefile);
	// original alpha input file name
	nextvalidline (fpara, pline);
	fscanf (fpara, "%s", alphafile);
	// original mappings input file name
	nextvalidline (fpara, pline);
	fscanf (fpara, "%s", mapfile);
	// original input file directory
	nextvalidline (fpara, pline);
	fscanf (fpara, "%100s", indir);

	sprintf(yuvFile, "%s/%s.yuv", indir, texturefile);
	sprintf(alphaFile, "%s/%s.seg", indir, alphafile);
	sprintf(meshDatFile, "%s/%s", indir, mapfile);
	DPRINT(("Original yuv-file      : %s\n", yuvFile));
	DPRINT(("Original seg-file      : %s\n", alphaFile));
	DPRINT(("Original mesh-file     : %s\n", meshDatFile));

	// frame sizes original VO
	nextvalidline (fpara, pline);
	if (fscanf (fpara, "%u", &w) != 1)	{
		fprintf(stderr, "wrong parameter file format on line %d\n", pline);
		assert (FALSE);
	}
	nextvalidline (fpara, pline);
	if (fscanf (fpara, "%u", &h) != 1)	{
		fprintf(stderr, "wrong parameter file format on line %d\n", pline);
		assert (FALSE);
	}
	DPRINT(("Frame width x height   : %d x %d\n", w, h));

	// video object index/label
	nextvalidline (fpara, pline);
	if (fscanf (fpara, "%u", &indexVO) != 1)	{
		fprintf(stderr, "wrong parameter file format on line %d\n", pline);
		assert (FALSE);
	}
	assert (indexVO >= 0);
	DPRINT(("Video object index     : %d\n", indexVO));

	// first and last frame number
	nextvalidline (fpara, pline);
	if (fscanf (fpara, "%u", &firstFrame) != 1) {
		fprintf(stderr, "wrong parameter file format on line %d\n", pline);
		assert (FALSE);
	}
	assert (firstFrame >= 0);
	nextvalidline (fpara, pline);
	if (fscanf (fpara, "%u", &idummy) != 1) {
		fprintf(stderr, "wrong parameter file format on line %d\n", pline);
		assert (FALSE);
	}
	assert (idummy >= 2);

	// temporal sampling rate
	nextvalidline (fpara, pline);
	if (fscanf (fpara, "%d", &frameInc) != 1)	{
		fprintf(stderr, "wrong parameter file format on line %d\n", pline);
		assert (FALSE);
	}
	assert (frameInc >= 1);

	lastFrame = firstFrame + (idummy-1)*frameInc;

	DPRINT(("First --- last frame   : %d --- %d\n", firstFrame, lastFrame));
	DPRINT(("Frame increment value  : %d\n", frameInc));


	// augmentation texture input file name
	nextvalidline (fpara, pline);
	fscanf (fpara, "%s", texturefile);
	// augmentation alpha input file name
	nextvalidline (fpara, pline);
	fscanf (fpara, "%s", alphafile);
	// augmentation -> original mapping input file name
	nextvalidline (fpara, pline);
	fscanf (fpara, "%s", mapfile);
	// directory of augmentation file 
	nextvalidline (fpara, pline);
	fscanf (fpara, "%100s", indir);

	sprintf(yuvNewFile, "%s/%s.yuv", indir, texturefile);
	sprintf(alphaNewFile, "%s/%s.seg", indir, alphafile);
	sprintf(meshNewFile, "%s/%s", indir, mapfile);
	DPRINT(("Augmentation yuv-file  : %s\n", yuvNewFile));
	DPRINT(("Augmentation seg-file  : %s\n", alphaNewFile));
	DPRINT(("Augmentation mesh-file : %s\n", meshNewFile));

	// frame sizes augmentation VO
	nextvalidline (fpara, pline);
	if (fscanf (fpara, "%u", &wNew) != 1)	{
		fprintf(stderr, "wrong parameter file format on line %d\n", pline);
		assert (FALSE);
	}
	nextvalidline (fpara, pline);
	if (fscanf (fpara, "%u", &hNew) != 1)	{
		fprintf(stderr, "wrong parameter file format on line %d\n", pline);
		assert (FALSE);
	}
	DPRINT(("Frame width x height   : %d x %d\n", wNew, hNew));

	// video object index/label
	nextvalidline (fpara, pline);
	if (fscanf (fpara, "%u", &indexNewVO) != 1) {
		fprintf(stderr, "wrong parameter file format on line %d\n", pline);
		assert (FALSE);
	}
	assert (indexNewVO >= 0);
	DPRINT(("Video object index     : %d\n", indexNewVO));


	// output file name
	nextvalidline (fpara, pline);
	fscanf (fpara, "%s", outfile);
	// output file directory
	nextvalidline (fpara, pline);
	fscanf (fpara, "%100s", outdir);

	sprintf(outImFile, "%s/%s.yuv", outdir, outfile);
	sprintf(outAlphaFile, "%s/%s.seg", outdir, outfile);
	DPRINT(("File rendered images   : %s\n", outImFile));
	DPRINT(("File rendered alphas   : %s\n", outAlphaFile));


	// rendering type
	nextvalidline (fpara, pline);
	if (fscanf (fpara, "%u", &renderType) != 1) {
		fprintf(stderr, "wrong parameter file format on line %d\n", pline);
		assert (FALSE);
	}
	assert (renderType >= 0);
	switch (renderType) {
	  case 0:
		DPRINT(("Rendering type         : self-mapping\n"));
		break;
	  case 1:
		DPRINT(("Rendering type         : transfiguration\n"));
		break;
	  case 2:
		DPRINT(("Rendering type         : augmentation\n"));
		break;
	  default:
		fprintf(stderr, "Rendering type incorrect ...\n");
		exit (1);
	}

	if (renderType == 0	 || renderType == 2) {
		// self-mapping or augmentation

		// check if texture planes input file exist
		if (stat (yuvFile, &stat_buf) == -1) {
			fprintf(stderr, "cannot find file %s\n", yuvFile);
			exit (1);
		}
		// find number of frames in texture planes file
		else {
			fsize = (int) stat_buf.st_size;
			yuvFrames = (int) (fsize/(1.5*w*h)); // assume 4:2:0 yuv sampling
		}
		// check if alpha planes input file exist
		if (stat (alphaFile, &stat_buf) == -1) {
			fprintf(stderr, "cannot find file %s\n", alphaFile);
			exit (1);
		}
		// find number of frames in alpha planes file
		else {
			fsize = (int) stat_buf.st_size;
			alphaFrames = (int) (fsize/(w*h));
		}
	}
	if (renderType == 1	 || renderType == 2) {
		// transfiguration or augmentation

		// check if texture plane input file exist
		if (stat (yuvNewFile, &stat_buf) == -1) {
			fprintf(stderr, "cannot find file %s\n", yuvNewFile);
			exit (1);
		}
		// check if alpha plane input file exist
		if (stat (alphaNewFile, &stat_buf) == -1) {
			fprintf(stderr, "cannot find file %s\n", alphaNewFile);
			exit (1);
		}
	}

	// draw mesh boolean
	nextvalidline (fpara, pline);
	if (fscanf (fpara, "%u", &drawMesh) != 1)	{
		fprintf(stderr, "wrong parameter file format on line %d\n", pline);
		assert (FALSE);
	}
	assert (drawMesh >= 0);

	// alpha plane extraction mode
	nextvalidline (fpara, pline);
	if (fscanf (fpara, "%u", &alphaMode) != 1)	{
		fprintf(stderr, "wrong parameter file format on line %d\n", pline);
		assert (FALSE);
	}
	assert (alphaMode >= 0);

	// that's all
	DPRINT(("\n\n"));
	fclose(fpara);
}



// reads comment lines until the next non-comment-line in a file and
// keeps track of the line number
void nextvalidline (FILE *fpara, uint &pline)	
{
	uchar firstchar;
	char line[200];

	fgets (line, 200, fpara);			// get the next line 
	pline++;
										// skip comment lines
	while (feof (fpara) == 0 && (firstchar = fgetc(fpara)) == '%') {
		fgets (line, 200, fpara);		// get the next line 
		pline++;
	}
	ungetc (firstchar, fpara);			// put back first char
}


// draw a mesh (triangulation) into an image
void im_draw_mesh (image &im, Tri2dTess &m, float *color)
{
	Tri2dRef *p;
	int	   x0, y0, x1, y1;
	int	   i, j, b;

	// for all triangles do ...
	for (i=0; i<m.getSize(); i++) {
		// get i'th triangle and for all edges do ...
		p = m.nextTriangle(i==0);
		for (j=0; j<3; j++) {
			// get the vertex coordinates
			Point2d p0 = p->getLink(j)->getPoint();
			Point2d p1 = p->getLink((j+1)%3)->getPoint();
			x0 = (int)(floor(p0.X()) + 0.5);
			y0 = (int)(floor(p0.Y()) + 0.5);
			x1 = (int)(floor(p1.X()) + 0.5);
			y1 = (int)(floor(p1.Y()) + 0.5);
			for (b=0; b<(im.color?3:1); b++)
				im_draw_clip_line(im, b, x0, y0, x1, y1, color[b]);
		}

	}

} // im_draw_mesh


// end of file
