/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* A. Tanju Erdem (Eastman Kodak Company)                             */
/* Candemir Toklu (University of Rochester)                           */
/*                                                                    */
/* and edited by                                                      */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* Eastman Kodak Company and the University of Rochester retain full  */
/* right to use the code for their own purposes, assign or donate the */
/* code to a third party and to inhibit third parties from using the  */
/* code for non MPEG-4 Visual (ISO/IEC 14496-2) conforming products.  */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1994-1998                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         parClass.cpp
   AUTHOR       A. Tanju Erdem - Eastman Kodak Company and Candemir Toklu - 
                University of Rochester.
   DATE         September 1994
   UPDATE       14 August 1996 by Candemir Toklu
   DESCRIPTION  Member functions for ParAffine class are defined in this
                file.

   SYNOPSIS     Include files : 

                #include "general.h"
                #include "pointClass.h"
                #include "polyClass.h"
                #include "matrixClass.h"
                #include "vectorClass.h"
                #include "parClass.h"
                               
*********************************************************************-*/

#include "general.h"
#include "pointClass.h"
#include "polyClass.h"
#include "matrixClass.h"
#include "vectorClass.h"
#include "parClass.h"

extern int CHECK_CONVEX;
extern int CHECK_PROPER;
extern int CHECK_VALID;
extern int CHECK_SIZE;

//----------------------------------------------------------------------------

ParAffine::ParAffine()
{
	a11 = 1; a12 = 0;
	a21 = 0; a22 = 1;
	a31 = 0; a32 = 0;
}

ParAffine::ParAffine( double i11, double i12, 
					  double i21, double i22,
					  double i31, double i32)
{
	a11 = i11; a12 = i12;
	a21 = i21; a22 = i22;
	a31 = i31; a32 = i32;
}

ParAffine::ParAffine( const ParAffine& from)
{
	a11 = from.a11; a12 = from.a12;
	a21 = from.a21; a22 = from.a22;
	a31 = from.a31; a32 = from.a32;
}

ParAffine& ParAffine::operator=( const ParAffine& from)
{
	if( this == &from) {
		//	Copying onto itself. Do nothing.
	}
	else {
		a11 = from.a11; a12 = from.a12;
		a21 = from.a21; a22 = from.a22;
		a31 = from.a31; a32 = from.a32;
	}
	return *this;
}

void ParAffine::toMatrix( Matrix& m)	
{
	Matrix t(3,3);

	t(0,0) = a11;	   t(0,1) = a12;	t(0,2) = 0.0;
	t(1,0) = a21;	   t(1,1) = a22;	t(1,2) = 0.0;
	t(2,0) = a31;	   t(2,1) = a32;	t(2,2) = 1.0;

	m = t;
}

void ParAffine::fromMatrix( Matrix& m)
{
	if( m.getRow() == 3 && m.getColumn() == 3) {
		a11 = m(0,0);	 a12 = m(0,1);
		a21 = m(1,0);	 a22 = m(1,1);
		a31 = m(2,0);	 a32 = m(2,1);
	}
	else {
		cout << "\nWarning: The input matrix has to be 3x3."
			 << "\nDid not modify the transformation parameters." << endl;
	}
}

ParAffine& ParAffine::getParUtoT( Tri2d t)
{
	Point2d d1=t[0];
	Point2d d2=t[1] - t[0];
	Point2d d3=t[2] - t[0];

	a11=d2.X(); a21=d3.X(); a31=d1.X();
	a12=d2.Y(); a22=d3.Y(); a32=d1.Y();

	return *this;
}

ParAffine& ParAffine::getParTtoUInverse( Tri2d t)
{

	//First, find the unit triangle to triangle mapping
	getParUtoT(t);

	Matrix m(3,3);
	toMatrix(m);

	// Invert to find triangle to unit mapping
	m.inverse();
	fromMatrix(m);

	return *this;
}

ParAffine& ParAffine::getParTtoUDirect( Tri2d t)
{
	double t0x=t[0].X(), t0y=t[0].Y();
	double t1x=t[1].X(), t1y=t[1].Y();
	double t2x=t[2].X(), t2y=t[2].Y();
	double t1t0x=t1x-t0x, t1t0y=t1y-t0y;
	double t2t0x=t2x-t0x, t2t0y=t2y-t0y;
	double det=t1t0x*t2t0y - t2t0x*t1t0y;
	a11 = t2t0y/det; a21 = -t2t0x/det; a31 = ((t2x*t0y)-(t2y*t0x))/det;
	a12 = -t1t0y/det; a22 = t1t0x/det; a32 = ((t0x*t1y)-(t0y*t1x))/det;

	/*getParUtoT(t);

	  ParAffine parinverse;

	  double	det=a11*a22-a12*a21;

	  parinverse.a11=a22/det;					parinverse.a12=-a12/det; 
	  parinverse.a21=-a21/det;				parinverse.a22=a11/det;
	  parinverse.a31=(a21*a32-a22*a31)/det;	parinverse.a32=(a12*a31-a11*a32)/det;

	  *this=parinverse; */
	return *this;
}


ParAffine& ParAffine::cascade( ParAffine& t1, ParAffine& t2)
{
	Matrix m1(3,3);
	Matrix m2(3,3);

	t1.toMatrix(m1);
	t2.toMatrix(m2);

	m1*=m2;

	fromMatrix(m1);  // Combined mapping

	return *this;
}


int ParAffine::getPars( Poly2d poly1 , Poly2d poly2)
{
	if( CHECK_SIZE) {
		if( poly1.getSize() != poly2.getSize()) {
			cout << "\nWarning (ParAffine::getPars): Polygon Size Mismatch"
				 << "\nCould not find the affine mapping parameters." << endl;
			return 0;
		}
	}
	if( CHECK_PROPER) {
		if( !(poly1.isProper() && poly2.isProper())) {
			cout << "\nWarning (ParAffine::getPars): Polygons are not Proper"
				 << "\nCould not find the affine mapping parameters." << endl;
			return 0;
		}
	}
	if( CHECK_CONVEX) {
		if( !(poly1.isConvex() && poly2.isConvex())) {
			cout << "\nWarning (ParAffine::getPars): Polygons are not Convex"
				 << "\nCould not find the affine mapping parameters." << endl;
			return 0;
		}
	}

	if( poly1.getSize() < 3) {
		cout << "\nWarning: Sorry number of equations are not enough"
			 << "\nNothing has done, Original parameters are send back" << endl;
		return 0;
	}
	else if( poly1.getSize() == 3) {
		Tri2d tri1( poly1[0], poly1[1], poly1[2]);
		Tri2d tri2( poly2[0], poly2[1], poly2[2]);

		ParAffine UtoT;
		ParAffine TtoU;

		UtoT.getParUtoT( tri2);
		TtoU.getParTtoUDirect( tri1);
		//TtoU.getParTtoUInverse( tri1);

		cascade( TtoU, UtoT); 
	
		return 1;
	}
	else {
		//form the matrix first.
		int i,j,index;
		int limit;
		limit=2*poly1.getSize();
		Matrix system(limit , 6);
		Vector sol(limit);
		for ( i=0; i<limit; i++) {
			index=i/2;
			if( i%2) {
				sol[i]=poly2[index].X();
				for ( j=0; j<6; j++) { 
					switch (j%6) {
					case 0: 
						system(i,j)=poly1[index].X();
						break;
					case 2:
						system(i,j)=poly1[index].Y();
						break;
					case 4: 
						system(i,j)=1.0;
						break;
					default:
						system(i,j)=0.0;
						break;
					}
				}
			} 
			else {
				sol[i]=poly2[index].Y();
				for ( j=0; j<6; j++) { 
					switch (j%6) {
					case 1: 
						system(i,j)=poly1[index].X();
						break;
					case 3:
						system(i,j)=poly1[index].Y();
						break;
					case 5: 
						system(i,j)=1.0;
						break;
					default:
						system(i,j)=0.0;
						break;
					}  //switch
				}	  //for j
			}  //else
		} //for i
		system.leastSquares(sol);
		a11 = sol[0]; a12 = sol[1];
		a21 = sol[2]; a22 = sol[3]; 
		a31 = sol[4]; a32 = sol[5];
		return 1;	 
	}	 //else
}	//main


Poly2d ParAffine::tform( Poly2d& tmp)
{
	Poly2d p( tmp.getSize());
	for( int i = 0; i < tmp.getSize(); i++) {
		p[i] = tform( tmp[i]);
	}
	return p;
}

Point2d ParAffine::tform( Point2d& tmp)
{
	double x, y;
	x = tmp.X() * a11 + tmp.Y() * a21 + a31;
	y = tmp.X() * a12 + tmp.Y() * a22 + a32;
	return point(x, y);
}

void ParAffine::print()
{
	cout << "\n  a11 = " << a11 << "	a12 = " << a12
		 << "\n  a21 = " << a21 << "	a22 = " << a22
		 << "\n  a31 = " << a31 << "	a32 = " << a32 << endl;
}

int ParAffine::read(istream& fp)
{
	fp>>a11; if(!fp) return 0;
	fp>>a12; if(!fp) return 0;
	fp>>a21; if(!fp) return 0;
	fp>>a22; if(!fp) return 0;
	fp>>a31; if(!fp) return 0;
	fp>>a32; if(!fp) return 0;
	return 1;
}

int ParAffine::write(ostream& fp)
{
	if(!fp) return 0; fp<<a11<<" ";
	if(!fp) return 0; fp<<a12<<"\n";
	if(!fp) return 0; fp<<a21<<" ";
	if(!fp) return 0; fp<<a22<<"\n";
	if(!fp) return 0; fp<<a31<<" ";
	if(!fp) return 0; fp<<a32<<"\n";
	return 1;
}

mapMethod ParAffine::type()
{
	return AFFINE;
}

void ParAffine::printType()
{
	cout << "\nMapping type is Affine." << endl;
}

int ParAffine::sizePar()
{
	return 6;
}

