/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* A. Tanju Erdem (Eastman Kodak Company)                             */
/* Candemir Toklu (University of Rochester)                           */
/*                                                                    */
/* and edited by                                                      */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* Eastman Kodak Company and the University of Rochester retain full  */
/* right to use the code for their own purposes, assign or donate the */
/* code to a third party and to inhibit third parties from using the  */
/* code for non MPEG-4 Visual (ISO/IEC 14496-2) conforming products.  */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1994-1998                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         polygonFunc.cpp
   AUTHOR       A. Tanju Erdem - Eastman Kodak Company and 
                Candemir Toklu - University of Rochester.
   DATE         September 1994
   UPDATE       14 August 1996 by Candemir Toklu
                Peter Van Beek, Thu Aug 22 09:24:19 1996
                - added more comments
   DESCRIPTION  In this file there are several polygon intersection or
                creation utilities which are used in intensity warping
                from polygons to transformed polygons, i.e. from the
                reference patch to the moved patch.     

   FUNCTIONS    
                Quad2d squareAroundPoint( Point2d&, double);
                Poly2d cellAroundPixel( int, int);
                void smallestRect( Poly2d&, int&, int&, int&, int&, int, int);
                int intersection( Vect2d, Vect2d, Point2d&);
                Poly2d intersection(Poly2d, Poly2d, int);
                int intersectHoriz( double, Vect2d, Point2d&, Point2d&);
                int getEndPoints( double, Poly2d&, Point2d&, Point2d&);
                int getEndPoints( double, int, int, Poly2d&, Point2d&, 
                                  Point2d&);

   SYNOPSIS     Include files : 
                #include "warpingProto.h"
                              
   ASSUMPTIONS  

*********************************************************************-*/

#include "warpingProto.h"



//--------------------------------------------------------------------------
// Create a square centered at a point
//--------------------------------------------------------------------------
Quad2d squareAroundPoint( Point2d& p, double t)
{
	Quad2d q;

	q[0] = point( p.X() - t/2, p.Y() - t/2);
	q[1] = point( p.X() + t/2, p.Y() - t/2);
	q[2] = point( p.X() + t/2, p.Y() + t/2);
	q[3] = point( p.X() - t/2, p.Y() + t/2);

	return q;
}

//--------------------------------------------------------------------------
// Create a unit square around a pixel
//--------------------------------------------------------------------------
Poly2d cellAroundPixel( int j, int i)
{
	Poly2d cell(4);
	cell[0] = point( j-0.5, i-0.5);
	cell[1] = point( j+0.5, i-0.5);
	cell[2] = point( j+0.5, i+0.5);
	cell[3] = point( j-0.5, i+0.5);
	return cell;
}

//--------------------------------------------------------------------------
// Define the smallest rectangle around a polygon that fits the image 
// the polygon belongs to.
//--------------------------------------------------------------------------
void smallestRect( Poly2d& pg, int& minMpg, int& minNpg, 
				   int& maxMpg, int& maxNpg, int width, int height)
{
	minMpg = pg.minM(); if( minMpg < 0) minMpg = 0;
	minNpg = pg.minN(); if( minNpg < 0) minNpg = 0;
	maxMpg = pg.maxM(); if( maxMpg >= width) maxMpg = width-1;
	maxNpg = pg.maxN(); if( maxNpg >= height) maxNpg = height-1;
}


//-------------------------------------------------------------------------
// A routine to sort numbers in ascending numerical order.
//-------------------------------------------------------------------------
void sortNumbers( double* arr, int* krr, int n)
{
	int i, j;
	double a; 
	int k;

	for( j = 1; j < n; j++) {
		a = arr[j];
		k = krr[j];
		i = j - 1;
		while( i >= 0 && arr[i] > a) {
			arr[i+1] = arr[i];
			krr[i+1] = krr[i];
			i--;
		}
		arr[i+1] = a;
		krr[i+1] = k;
	}
}


//--------------------------------------------------------------------------
// Find the intersection point sr of two line seqments s and r.
// If there are 2 intersection points function returns 2 instead of 1;
// the 2nd intersection point is returned in sr2.
//--------------------------------------------------------------------------
int intersection( Vect2d s, Vect2d r, Point2d& sr)
{
	sr = point( 0.0, 0.0);  // Return value when there is no intersection

	double a11 = s.tail.X() - s.head.X();
	double a12 = r.head.X() - r.tail.X();
	double a21 = s.tail.Y() - s.head.Y();
	double a22 = r.head.Y() - r.tail.Y();
	double b1 = r.head.X() - s.head.X();
	double b2 = r.head.Y() - s.head.Y();
	double determinant = a11 * a22 - a12 * a21;
	if( fabs(determinant) > EPSILON*EPSILON) {
		// lines are not parallel
		double alpha = (a22 * b1 - a12 * b2 ) / determinant;
		double beta = (-a21 * b1 + a11 * b2 ) / determinant;
		if (alpha >= 0.0 && alpha <= 1.0 && beta >= 0.0 && beta <= 1.0) {
			sr = point( alpha * s.tail.X() + (1 - alpha) * s.head.X(),
						alpha * s.tail.Y() + (1 - alpha) * s.head.Y());
			return 1;
		}
	}
	return 0;
}


//--------------------------------------------------------------------------
// Find the polygon that is the intersection of two CONVEX polygons
//--------------------------------------------------------------------------
Poly2d intersection( Poly2d s, Poly2d r, int checkIfBothConvex)
{
	if( checkIfBothConvex) {
		if( ! s.isConvex() || ! r.isConvex()) {
			cout << "\nWarning: Cannot find the intersection of the two polygons;"
				 << "\n		  They must both be convex. Returned empty polygon"
				 << "intersection." << endl;
			return Poly2d();	// s and r must both be convex
		}
	}

	Pt2dList points;
	Point2d newPt;
	int s_side, r_side;
	int i, k;
	int size_s = s.getSize();
	int size_r = r.getSize();

	// Find the vertices of the intersection polygon.

	// First, find all points that are inside the given polygons.
	// If one polygon is completely inside the other one, return 
	// the polygon that is inside.

	k = 0;
	for( s_side = 0; s_side < size_s; s_side++) {
		if( s[s_side].isInside(r)) {
			points.append(s[s_side]);
			k++;
		}
	}
	if( k == size_s) return s;  // s is completely inside r

	k = 0;
	for( r_side = 0; r_side < size_r; r_side++) {
		if( r[r_side].isInside(s)) {
			points.append(r[r_side]);
			k++;
		}
	}
	if( k == size_r) return r;  // r is completely inside s

	// Next, find all points of intersection on the sides of the two polygons.
	// Note that a polygon side can have at most 2 intersections with another 
	// convex polgon

	int* s_numInt = new int[size_s];
	int* r_numInt = new int[size_r];
	Pt2dLink *dummy;
	for( i = 0; i < size_s; i++) {
		s_numInt[i] = 0;  // Number of intersections on i-th side of s
	}
	for( i = 0; i < size_r; i++) {
		r_numInt[i] = 0;  // Number of intersections on i-th side of r
	}

	for( s_side = 0; s_side < size_s; s_side++) {
		for( r_side = 0; r_side < size_r; r_side++) {
			if( s_numInt[s_side] == 2 || r_numInt[r_side] == 2) {
				continue;  // At least one of the two sides already has 2 intersections
			}
			else {
				if( intersection( vector2d( s[s_side], s[s_side+1]), 
								  vector2d( r[r_side], r[r_side+1]), newPt)) {
					if( points.grow(newPt, dummy)) {
						// newPt is indeed new to the list
						s_numInt[s_side] += 1;
						r_numInt[r_side] += 1;
					}
				}
			}
		}
	}

	delete[] s_numInt;
	delete[] r_numInt;

	// Initialization of the intersection polygon

	int size = points.getSize();
	Poly2d sr(size);	// Intersection polygon

	if( size <= 2) {
		for( i = 0; i < size; i++) {
			sr[i] = points[i];
		}
	}
	else {
		// Re-order the list of points to make a convex polygon

		// Find a point inside the intersection polygon
		int i;
		Point2d center;
		for( i = 0; i < size; i++) {
			center += points[i] / (double) size;
		}

		Point2d tmpPt;
		double* numbers = new double[size];
		int* newOrder = new int[size];
		for( i = 0; i < size; i++) {
			// Unit vectors centered at the origin
			tmpPt = (points[i] - center).unit();
			// Numbers in (-2, 2]
			numbers[i] = Sign1at0(tmpPt.Y()) * ( 1 + tmpPt.X());
			// Counter-clockwise order
			newOrder[i] = i;
		}

		// Sort numbers in ascending order
		sortNumbers( numbers, newOrder, size);

		for( i = 0; i < size; i++) {
			sr[size - 1 - i] = points[newOrder[i]];  // Clockwise order
		}

		delete[] numbers;
		delete[] newOrder;
	}

	return sr;
}


int intersectHoriz( double y, Vect2d vect, Point2d& pt1, Point2d& pt2)
{
	double y0, y1, x0, x1, x;
	y1 = vect.head.Y();
	y0 = vect.tail.Y();
	x1 = vect.head.X();
	x0 = vect.tail.X();
	if( (y-y0)*(y-y1)<=0.0) {
		// y is in (y0,y1).
		if( y1==y && y0==y) {
			pt1 = vect.tail; pt2 = vect.head;
			return 2;
		} 
		/*
		  tmpMinM	   x0	   x1	  tmpMaxM
		  X--------X-------X---------X
		  i.e. x0 and x1 are in between or fabs(x?-tmp?M)<0.5
		  */
		else if( y1 == y) {
			pt1 = vect.head;
			return 1;
		}		
		else if( y0 == y) {
			pt1 = vect.tail;
			return 1;
		}
		else {
			x = x0 + (x1-x0)*(y-y0)/(y1-y0);
			if( x1 == x0) {
				pt1 = point(x,y);
				return 1;
			}
			else if( x1 > x0) {
				if( x<=x1 && x>=x0) {
					pt1 = point(x,y);
					return 1;
				}
			}
			else { 
				//if( x1 < x0)
				if( x>=x1 && x<=x0) {
					pt1 = point(x,y);
					return 1;
				}
			}
		}
	}
	return 0;
}

//--------------------------------------------------------------------------
// Find the intersection points a polygon and a line segment
// If there are 1 intersection point function returns 1 instead of 2;
// the 1st intersection point is returned in pt1 and 2nd one is
// returned in pt2. The polygon has to be convex.
//--------------------------------------------------------------------------
int getEndPoints( double y, Poly2d& p, Point2d& pt1, Point2d& pt2)
{
	int size = p.getSize();

	if( size == 0) {
		cout << "\nWarning: Reference polygon for intersection is empty."
			 << "\nNo intersection point is returned." << endl;
		return 0;
	}
	else if( size == 1) {
		if( y == p[1].Y()) {
			pt1 = p[1];
			return 1;
		}
		else
			return 0;
	}
	else if( size == 2) {
		Vect2d vect;
		vect.head = p[1];
		vect.tail = p[0];
		return intersectHoriz( y, vect, pt1, pt2);		
	}
	else {
		int count = 0;
		Point2d tmp1, tmp2;
		int intNum;
		Vect2d vect;

		for( int i=0; i< size; i++) {
			vect.head=p[i];
			vect.tail=p[i+1];
			intNum = intersectHoriz( y, vect, tmp1, tmp2);	   
			if( intNum ) {
				if( intNum == 1) {
					if( !count) {
						pt1 = tmp1;
						count++;
					}
					else {
						if( tmp1 != pt1) {
							pt2 = tmp1;
							return 2;
						}
					}
				}
				else {
					// if( intersectHoriz( y, vect, tmp1, tmp2) == 2)
					pt1 = tmp1;
					pt2 = tmp2;
					return 2;
				}
			}
		}
		return count;
	}
}

//--------------------------------------------------------------------------
int getEndPoints( double y, int tmpminM, int tmpmaxM, Poly2d& p, Point2d& pt1,
				  Point2d& pt2)
{
	Vect2d line, s;
	int count = 0;
	Point2d	temp;
	int size = p.getSize();

	if( p.getSize() == 0) {
		cout << "\nWarning: Reference polygon for intersection is empty."
			 << "\nNo intersection point is returned." << endl;
		return 0;
	}
	else if( p.getSize() == 2) {
		line.head = p[0]; line.tail = p[1];
		if( (y == line.tail.Y()) && (y == line.head.Y())) {
			pt1 = line.head; 
			pt2 = line.tail;
			return 2;	  
		}
		else {
			count += intersection( line, s, pt1);
			return count;
		}
	}
	else {
		s.head = point( tmpminM - 0.5, y);
		s.tail = point( tmpmaxM + 0.5, y);
		for( int i=0; i< size; i++) {
			line.head = p[i]; line.tail = p[(i+1)%size];
			if( fabs(y - line.tail.Y()) < EPSILON 
				&& fabs(y - line.head.Y()) < EPSILON) {
				pt1 = line.head; 
				pt2 = line.tail; 
				return 2;
			}
			else {
				if( count == 0)  
					count += intersection( line, s, pt1);
				else {
					if( intersection( line, s, temp)) {
						if( temp != pt1) {
							pt2 = temp;
							return 2;
						}
					}
				}
			}
		}
		return count;
	}
}



// end of file
