/**********************************************************************/
/*                                                                    */
/* This software module was originally developed by                   */
/*                                                                    */
/* Candemir Toklu (University of Rochester)                           */
/* A. Tanju Erdem (Eastman Kodak Company)                             */
/*                                                                    */
/* and edited by                                                      */
/*                                                                    */
/* Peter van Beek (University of Rochester)                           */
/*                                                                    */
/* in the course of development of MPEG-4 Visual (ISO/IEC 14496-2).   */
/* This software module is an implementation of a part of one or      */
/* more MPEG-4 Visual (ISO/IEC 14496-2) tools as specified by the     */
/* MPEG-4 Visual (ISO/IEC 14496-2). ISO/IEC gives users of MPEG-4     */
/* Visual (ISO/IEC 14496-2) free license to this software module or   */
/* modifications thereof for use in hardware or software products     */
/* claiming conformance to MPEG-4 Visual (ISO/IEC 14496-2). Those     */
/* intending to use this software module in hardware or software      */
/* products are advised that its use may infringe existing patents.   */
/* The original developer of this software module and his/her         */
/* company, the subsequent editors and their companies, and ISO/IEC   */
/* have no liability for use of this software module or modifications */ 
/* thereof in an implementation. Copyright is not released for non    */
/* MPEG-4 Visual (ISO/IEC 14496-2) conforming products.               */
/* Eastman Kodak Company and the University of Rochester retain full  */
/* right to use the code for their own purposes, assign or donate the */
/* code to a third party and to inhibit third parties from using the  */
/* code for non MPEG-4 Visual (ISO/IEC 14496-2) conforming products.  */
/* This copyright notice must be included in all copies or derivative */
/* works.                                                             */
/*                                                                    */
/* Copyright (c) 1994-1998                                            */
/*                                                                    */
/**********************************************************************/

/*+*********************************************************************
   NAME         vectorClass.cpp
   AUTHOR       Candemir Toklu - University of Rochester and
                A. Tanju Erdem - Eastman Kodak Company.
   DATE         September 1994
   UPDATE       14 August 1996 by Candemir Toklu
   DESCRIPTION  Member functions for Vector class are defined in this
                file.

   SYNOPSIS     Include files : 
                                
                #include "vectorClass.h"
                #include "matrixClass.h"
                #include "pointClass.h"
                              
*********************************************************************-*/

#include "vectorClass.h"
#include "matrixClass.h"
#include "pointClass.h"


//----------------------------------------------------------------------------

Vector::Vector()
{
	column = NULL;
	n = 0;
}

Vector::Vector( int m)
{
	if( m <= 0) {
		cout << "\nWarning: Size must be positive! "
			 << "Defined an empty vector." << endl;
		column = NULL;
		n = 0;
	}
	else {
		n = m;
		column = (double*) new double[m];
		for( int i = 0; i < m; i++) {
			column[i] = 0.0;
		}
	}
}

Vector::Vector( const Vector& m)
{
	column = NULL;
	*this = m;
}

Vector::Vector( double f)
{
	column = new double[1];
	column[0] = f;
	n = 1;
}

Vector::~Vector()
{
	if( n) {
		delete[] column;
	}
}

Vector& Vector::empty()
{
	if( n) {
		delete[] column;
	}
	column = NULL;
	n = 0;
	return *this;
}

Vector& Vector::setSize( int newSize)
{
	//this->~Vector();
	empty();
	if( newSize >0) {
		n = newSize;
		column = (double*) new double[ n];
		for( int i = 0; i < n; i++) column[i] = 0.0;
	}
	else {
		n = 0;
		column = NULL;
	}
	return *this;
}

Vector& Vector::operator=( const Vector& m)
{
	if(n) {
		// First, delete the old column if it is not empty
		delete[] column;
	}

	if(m.column) {
		// Construct the new column
		n = m.n;
		column = (double*) new double[n];
		for( int i = 0; i < n; i++) {
			column[i] = m.column[i];
		}
	}
	else {
		column = NULL;
		n = 0;
	}
	return *this;
}

Vector& Vector::operator=( double a)
{
	if(column) {
		for( int i = 0; i < n; i++) {
			column[i] = a;
		}
	}
	return *this;
}

Vector& Vector::operator*=( double a)
{
	if(column) {
		for( int i = 0; i < n; i++) {
			column[i] *= a;
		}
	}
	return *this;	  
}

Vector& Vector::operator/=( double a)
{
	if( a == 0.0) {
		cout << "\nWarning: Division by zero is attempted. Division is cancelled.";
		return *this;
	}
	if(column) {
		for( int i = 0; i < n; i++) {
			column[i] /= a;
		}
	}
	return *this;	  
}

double& Vector::operator()( int i)
{
	if( ! column || i < 0 || i >= n ) {
		cout << "\nError: The index is out of range or the vector is empty."
			 << endl;
		exit(1);
	}
	else {
		return column[i];
	}
}

double& Vector::operator[]( int i)
{
  if( !column || i < 0 || i >= n) {
	cout << "\nError: The index is OUT of range or the vector is empty."
		 << endl;
	exit(1);
  }
  else {
	return column[i];
  }
}

Vector& Vector::operator+=( const Vector& m)
{
	if( n == m.n ) {
		for( int i = 0; i < n; i++) {
			column[i] += m.column[i];
		}
	}
	else {
		cout << "\nWarning: The matrices are not compatible for addition."
			 << "\nThe first column is left unchanged" << endl;
	}

	return *this;
}

Vector& Vector::operator-=( const Vector& m)
{
	if( n == m.n) {
		for( int i = 0; i < n; i++) {
			column[i] -= m.column[i];
		}
	}
	else {
		cout << "\nWarning: The matrices are not compatible for addition."
			 << "\nThe first column is left unchanged" << endl;
	}

	return *this;
}

int Vector::setN( int m)
{
	if( n != m) {
		// Delete old vector
		delete[] column;
		// Define new vector
		if( m <= 0) {
			column = NULL;
			n = 0;
			if( m < 0) {
				cout << "\nWarning: Size must be non-negative! "
					 << "Defined an empty vector." << endl;
				return 0;
			}
		}
		else {
			column = (double*) new double[m];
			n = m;
		}
	}
	return 1;
}

Vector& Vector::append( double a)
{
	if( n){
		int tn = n+1;
		double *tmpColumn = new double[tn];
		for( int i = 0; i < n; i++) tmpColumn[i] = column[i];
		tmpColumn[n] = a;
		delete[] column;
		column = tmpColumn;
		n = tn;
	}
	else{
		column = (double*) new double[1];
		column[0] = a;
		n = 1;
	}	 
	return (*this);
}

Vector& Vector::remove( int m)
{
	if( !n){
		cout << "\nWarning: Empty vector cannot delete an element!" << endl;
	}
	else{
		if( (m<0) || (m>=n)){
			cout << "\nError: Index out of range!" << endl;
			exit(0);
		}
		else{
			if( n == 1){
				delete[] column; n--; column = NULL;
			}
			else{
				double *tmpCol = new double[n-1];
				int i;
				for( i = 0; i < m; i++) tmpCol[i] = column[i];
				for( i = m+1; i < n; i++) tmpCol[i-1] = column[i];
				delete[] column;
				column = tmpCol;
				n--;
			}		
		}
	}	 
	return (*this);
}

int Vector::where( double d)
{
	if(!n) return -1;
	int i;
	for( i = 0; i < n; i++) if( fabs(column[i]-d)<EPSILON) break;
	if( i == n) return -1;
	else return i;
}

Vector& Vector::leftMult( Matrix& m)
{
	if( column) {
		if( n != m.getRow()) {
			cout << "\nError:Vector and Matrix are not compatible for multiplication"
				 << "Vector has unchanged." << endl;
		}  
		else {
			Vector result(m.getColumn());
			for( int i=0; i < m.getColumn(); i++) {
				for (int j=0; j < n ; j++) {
					result.column[i] += column[i]*m(j,i);
				}
			}
			(*this) = result;
		}
	}	  
	else {
		cout << "\nWarning:Vector is empty. Vector has unchanged."
			 << endl;
	}
	return (*this);
}


Vector& Vector::rightMult( Matrix& m)
{
	if( column) {
		if( n != m.getColumn()) {
			cout << "\nError:Vector and Matrix are not compatible for multiplication"
				 << "Vector has unchanged." << endl;
		}  
		else {
			Vector result(m.getRow());
			for( int i=0; i < m.getRow(); i++) {
				for (int j=0; j < n ; j++) {
					result.column[i] += column[j]*m(i,j);
				}
			}
			(*this) = result;
		}
	}	  
	else {
		cout << "\nWarning:Vector is empty. Vector has unchanged."
			 << endl;
	}
	return (*this);
}

Matrix Vector::matrixRow()	// Make a row matrix from a vector
{
	Matrix m(n,1);
	for( int i = 0; i < n; i++)
		m[i] = column[i];
	return m;
}

Matrix Vector::matrixCol()	// Make a column matrix from a vector
{
	Matrix m(1,n);
	for( int i = 0; i < n; i++)
		m[i] = column[i];
	return m;
}

void Vector::print()
{
	cout << "\nVector(" << n << ") = \n";
	for( int i = 0; i < n; i++) {
		cout << "  [";
		cout.width(10);
		cout << column[i] << "]" << endl;
	}
}

int Vector::read(ifstream& fp)
{
	int tmpN;

	fp>>tmpN;
	if( tmpN != n) {
		delete[] column;
		column = new double[ n = tmpN];
	}

	for( int i = 0; i < n; i++) {
		if(!fp) return 0;
		fp >> column[i];
	}

	return 1;
}

int Vector::write(ofstream& fp)
{
	if(!fp) return 0;
	fp << n << "\n";
	for( int i = 0; i < n; i++) {
		if(!fp) return 0;
		fp.width(10);
		fp << column[i] << "\n";
	}
	return 1;
}

// end of file
