// =====================================================
//
// File:   matrix.cc
// Class:  Matrix
// 
// Author: Jrgen Ahlberg <ahlberg@isy.liu.se>
//         Dept. of Electrical Engineering
//         Linkping University
//         SE-581 83 Linkping, SWEDEN
// Date:   1998-01-12
//
// This program is distributed in the hope that it will
// be useful, but WITHOUT ANY WARRANTY.
//
// =====================================================


#include "matrix.hh"


// =====================================================
//
// Copy
//
// =====================================================


Matrix& Matrix::
copy(Matrix& src) {
  resize(src.nrows,src.ncols);
  for(int i=0;i<nrows*ncols;i++) data[i]=src.data[i];
  return *this;
}

// =====================================================

void Matrix::
copyrow(Matrix& srcmat, int srcrow, int destrow) {
  for(int col = 0; col < ncols; col++)
    set(destrow, col, srcmat.get(srcrow,col));
}

// =====================================================

void Matrix::
copycolumn(Matrix& srcmat, int srccol, int destcol) {
  for(int row = 0; row < nrows; row++)
    set(row, destcol, srcmat.get(row,srccol));
}

// =====================================================

void Matrix::
subtractrow(Matrix& srcmat, int srcrow, int destrow) {
  for(int col = 0; col < ncols; col++)
    set(destrow, col, get(destrow, col) - srcmat.get(srcrow,col));
}


// =====================================================
//
// Arithmetics
//
// =====================================================

float Matrix::
norm(void) {
  float sum=0;
  for(int i=0;i<nrows*ncols;i++) sum+=(data[i]*data[i]);
  sum=sqrt(sum);
}

// =====================================================

float Matrix::
maxincol(int col) {
  float cmax = get(0,col);
  for(int row = 1; row < nrows; row++) cmax = max(cmax,get(row,col));
  return cmax;
}

float Matrix::
minincol(int col) {
  float cmin = get(0,col);
  for(int row = 1; row < nrows; row++) cmin = min(cmin,get(row,col));
  return cmin;
}

// =====================================================

Matrix& Matrix::
transpose(void) {

  int row, col;
  Matrix *m;
  
  m = new Matrix(ncols,nrows);
  for(row=0;row<nrows;row++)
    for(col=0;col<ncols;col++)
      m->set(col,row,get(row,col));

  return *m;
}

// =====================================================


Mask& Matrix::
remove_rows(Mask& mask) {

  int row, col, newnrows, newrow;
  float *newdata;

  newnrows = mask.sum();
  newdata = new float[newnrows*ncols];

  newrow = 0;
  for(row = 0; row < nrows; row++) {
    if(mask.bit(row)) {
      for(col = 0; col < ncols; col++)
	newdata[newrow*ncols+col] = data[row*ncols+col];
      newrow++;
    }
  }
  renew(newnrows, ncols, newdata);

  return mask;
}


// =====================================================


Mask& Matrix::
remove_cols(Mask& mask) {

  int row, col, newncols, newcol;
  float *newdata;

  if(ncols != mask.length()) {
    fprintf(stderr, "Error: (Matrix::remove_cols()) Mask has wrong size!\n");
    exit(0);
  }

  newncols = mask.sum();
  newdata = new float[nrows*newncols];

  newcol = 0;
  for(col = 0; col < ncols; col++) {
    if(mask.bit(col)) {
      for(row = 0; row < nrows; row++)
	newdata[row*newncols+newcol] = data[row*ncols+col];
      newcol++;
    }
  }
  renew(nrows, newncols, newdata);

  return mask;
}


// =====================================================


Mask& Matrix::
createEmptyColMask(Mask& colmask) {

  float *arrf, tmpf;
  int r,c;

  arrf = new float[ncols];
  colmask.resize(ncols);

  for(c = 0; c < ncols; c++) {
    arrf[c] = 0;
    for(r = 0; r < nrows; r++) {
      tmpf = get(r,c);
      arrf[c] += max(tmpf,-tmpf);
    }
    colmask.set(c, arrf[c]>0);
  }
  delete arrf;
  return colmask;
}


// =====================================================


Mask& Matrix::
reduce(Mask& rowmask) {

  Mask colmask;
  remove_rows(rowmask);
  return remove_cols(createEmptyColMask(colmask));
}


// =====================================================


Matrix& Matrix::
operator * (Matrix& X) {

  Matrix *Y;
  int row, col, i;
  float tmpf;

  if(ncols != X.nrows) {
    printf("Matrix::operator *: Inner matrix dimensions must agree!\n");
    printf("Tried to multiply %dx%d-matrix with %dx%d-matrix.\n",
	   nrows, ncols, X.nrows, X.ncols);
    exit(0);
  }

  Y = new Matrix(nrows, X.ncols);
  for(row = 0; row < Y->nrows; row++) {
    for(col = 0; col < Y->ncols; col++) {
      tmpf = 0;
      for(i = 0; i < ncols; i++) {
	tmpf += get(row,i) * X.get(i,col);
      }
      Y->set(row,col,tmpf);
    }
  }

  return *Y;
}


// =====================================================


Matrix& Matrix::
operator - (Matrix& X) {

  Matrix *Y;
  int row, col, i;
  float tmpf;

  if(ncols != X.ncols || nrows != X.nrows) {
    printf("Matrix::operator -: Matrix dimensions must agree!\n");
    printf("Tried to subtract %dx%d-matrix from %dx%d-matrix.\n",
	   X.nrows, X.ncols, nrows, ncols);
    exit(0);
  }

  Y = new Matrix(nrows, ncols);
  for(i = 0; i < nrows*ncols; i++) {
    Y->data[i] = data[i] - X.data[i];
  }

  return *Y;
}


// =====================================================


Matrix& Matrix::
square(void) {

  Matrix *Y;
  int i;

  Y = new Matrix(nrows,ncols);
  for(i = 0; i < nrows*ncols; i++) {
    Y->data[i] = data[i] * data[i];
  }

  return *Y;
}


// =====================================================
//
// Quantization
//
// =====================================================

void Matrix::quantize(float *qstep) {
  for(int row = 0; row < nrows; row++)
    quantrow(qstep, row);
}

// =====================================================

void Matrix::quantize(float qstep) {
  for(int row = 0; row < nrows; row++)
    quantrow(qstep, row);
}

// =====================================================

void Matrix::quantrow(float qstep, int row) {
  for(int col = 0; col < ncols; col++)
    set(row, col, quantize(qstep, get(row, col)));
}

// =====================================================

void Matrix::quantrow(float *qstep, int row) {
  for(int col = 0; col < ncols; col++)
    set(row, col, quantize(qstep[col], get(row, col)));
}

// =====================================================

void Matrix::dequantize(float *qstep) {
  for(int row = 0; row < nrows; row++)
    dequantrow(qstep, row);
}

// =====================================================

void Matrix::dequantize(float qstep) {
  for(int row = 0; row < nrows; row++)
    dequantrow(qstep, row);
}

// =====================================================

void Matrix::dequantrow(float qstep, int row) {
  for(int col = 0; col < ncols; col++)
    set(row, col, dequantize(qstep, get(row, col)));
}

// =====================================================

void Matrix::dequantrow(float *qstep, int row) {
  for(int col = 0; col < ncols; col++)
    set(row, col, dequantize(qstep[col], get(row, col)));
}
