// =====================================================
//
// File:   matrix.hh
// Class:  Matrix
// Uses:   Mask
//         LAPACK
// 
// Author: Jrgen Ahlberg <ahlberg@isy.liu.se>
//         Dept. of Electrical Engineering
//         Linkping University
//         SE-581 83 Linkping, SWEDEN
// Date:   1998-
//
// This program is distributed in the hope that it will
// be useful, but WITHOUT ANY WARRANTY.
//
// =====================================================


#ifndef MATRIX_HH
#define MATRIX_HH

#include <math.h>
#include "../codec.hh"
#include "../sequence/mask.hh"


// =====================================================


class Matrix {

public:

  // Attributes

  int nrows, ncols;
  float *data;

  // Constructors

  Matrix(void) {nrows = 0; ncols = 0; data = 0;}
  Matrix(int r, int c) {nrows=r; ncols=c; data=new float[r*c];}
  Matrix(int r, int c, float *d) {nrows = r; ncols = c; data = d;}

  // Primitives

  inline Matrix& resize(int r, int c) {renew(r,c,new float[r*c]); return *this;}
  inline Matrix& renew(int r, int c, float *d) {nrows = r; ncols = c; if(data) delete data; data = d; return *this;}

  inline float set(int r, int c, float f) {return(data[r*ncols + c] = f);}
  inline int set(int r, int c, int i) {return (int)set(r,c,(float)i);}
  inline float get(int r, int c) {return data[r*ncols + c];}
  inline float* getp(int r, int c) {return &(data[r*ncols + c]);}

  inline void empty(void) {if(data) for(int i=0;i<nrows*ncols;i++) data[i]=0.0;}

  // Copy

  Matrix& copy(Matrix&);
  void copyrow(Matrix&, int, int);
  void copycolumn(Matrix&, int, int);
  void subtractrow(Matrix&, int, int);

  // Arithmetics

  float norm(void);
  float maxincol(int);
  float minincol(int);
  Matrix& transpose(void);
  Matrix& least_square_fit(Matrix&, float, float);
  Matrix& least_square_fit(Matrix&);
  Mask& remove_rows(Mask&);
  Mask& remove_cols(Mask&);
  Mask& reduce(Mask&);
  Mask& createEmptyColMask(Mask&);

  Matrix& operator * (Matrix&);
  Matrix& operator - (Matrix&);
  Matrix& square(void);

  inline int round(float f) {
    return int(f+(f>0?0.5:-0.5));
  }

  // Quantization

  inline int quantize(float q, float f) {return round(max(FBF_MIN,min(FBF_MAX,f))/q);}
  inline float dequantize(float q, float f) {return f*q;}

  void quantize(float*);
  void quantize(float);
  void quantrow(float, int);
  void quantrow(float*, int);

  void dequantize(float*);
  void dequantize(float);
  void dequantrow(float, int);
  void dequantrow(float*, int);

  // Input

  int read_ascii(FILE *);
  int read_ascii(char *);
  float read_element_ascii(FILE*, int, int);
  int readln_ascii(FILE*, int);

  // Output

  void write_ascii(FILE*);
  void writerow_ascii(FILE *, int);

  void write_int_ascii(FILE*);
  void writerow_int_ascii(FILE *, int);

  void report(void);

};

#endif
