// =====================================================
//
// File:   faps.cc
// Class:  FAPs
// 
// Author: Jrgen Ahlberg <ahlberg@isy.liu.se>
//         Dept. of Electrical Engineering
//         Linkping University
//         SE-581 83 Linkping, SWEDEN
// Date:   1998-
//
// This program is distributed in the hope that it will
// be useful, but WITHOUT ANY WARRANTY.
//
// =====================================================


#include "faps.hh"


// =====================================================
//
// Constructors
//
// =====================================================

FAPs::FAPs(void) :
  Sequence() {
    mask.resize(NFAPS);
    fileinfo.copy(CURRENT_FAP_VERSION, "Empty FAP-file", 0, 0);
}    

FAPs::FAPs(int frames) :
  Sequence(frames, NFAPS) {
    mask.resize(NFAPS);
    fileinfo.copy(CURRENT_FAP_VERSION, "Empty FAP-file", 0, frames);
}    

FAPs::FAPs(int frames, Mask& newmask) :
  Sequence(frames, newmask) {
    fileinfo.copy(CURRENT_FAP_VERSION, "Empty FAP-file", 0, frames);
}    

FAPs::FAPs(Mask& newmask, FileInfo& newinfo) :
  Sequence(newmask, newinfo) {
}    


// =====================================================
//
// psnr
//
// =====================================================

float FAPs::
psnr(FAPs& noisy) {

  float peak2peak, mincol, maxcol, SSE, PSNR;
  int col, row;
  Matrix SE;
  Mask mask;

  // Mask and squared peak-to-peak signal
  mask.resize(ncols);
  peak2peak = 0.0;
  for(col = 0; col < ncols; col++) {
    mincol = get(0,col);
    maxcol = mincol;
    for(row = 1; row < nrows; row++) {
      mincol = min(mincol, get(row,col));
      maxcol = max(maxcol, get(row,col));
    }
    peak2peak += (maxcol - mincol) * (maxcol - mincol);
    mask.set(col, (maxcol - mincol)>0?1:0);
  }


  // Squared error matrix
  SE = (*this - noisy).square();

  // SSE in each row
  PSNR = 0.0;
  for(row = 0; row < nrows; row++) {    
    SSE = 0.0;
    for(col = 0; col < ncols; col++) {
      if(mask.bit(col)) {
	SSE += SE.get(row, col);
      }
    }
    PSNR += min(100000,peak2peak/SSE);
  }

  PSNR /= (float)nrows;
  PSNR = 10.0*log10(PSNR);

  return PSNR;
}


// =======================================================
//
// read_ascii
//
// =======================================================

int FAPs::
read_ascii(char *filename) {

  FILE *fp;

  if((fp = fopen(filename, "r")) == 0) {
    fprintf(stderr, "File %s cannot be opened!\n", filename);
    return(0);
  }

  read_ascii(fp);
  fclose(fp);

}

// =======================================================

int FAPs::
read_ascii(FILE* fp) {

  int i, tmpi, frm, fap;
  Mask current_mask(NFAPS);

  nextValidLine(fp);
  fileinfo.read_ascii(fp);

  mask.resize(NFAPS);

  if(version() <= 1.5) for(i=0; i<NGROUPS; i++) fscanf(fp,"%d",&tmpi);

  mask.read_ascii(fp);
  Matrix::resize(nframes(), usedFAPs());
  current_mask.copy(mask);

  if(version() > 1.5) {
    fscanf(fp,"%d",&tmpi);
    for(fap = 0; fap < usedFAPs(); fap++) {
      read_element_ascii(fp, 0, fap);
    }
    for(frm = 1; frm < nframes(); frm++) {
      current_mask.read_ascii(fp);
      fscanf(fp,"%d",&tmpi);
      i = 0;
      for(fap = 0; fap < NFAPS; fap++) {
	if(current_mask.bit(fap)) {
	  fscanf(fp,"%d",&tmpi);
	  if(mask.bit(fap))
	    set(frm, i++, tmpi);
	} else
	  if(mask.bit(fap))
	    set(frm, i++, 0);
      }
    }
  } else {
    float tmpf;
    for(frm = 0; frm < nframes(); frm++) {
      fscanf(fp,"%d",&tmpi);
      readln_ascii(fp, frm);
    }
  }

}


// =======================================================
//
// write_ascii
//
// =======================================================

int FAPs::
write_ascii(char *filename) {

  FILE *fp;

  if((fp = fopen(filename, "w")) == 0) {
    printf("File %s cannot be opened!\n", filename);
    return(0);
  }
  write_ascii(fp);
  fclose(fp);
  return(1);
}

// =======================================================

int FAPs::
write_ascii(FILE *fp) {

  int frm;
  fileinfo.version = 2.0;
  fileinfo.write_ascii(fp);
  for(frm = 0; frm < nframes(); frm++) {
    mask.write_ascii(fp);
    fprintf(fp,"%d ",frm);
    writerow_int_ascii(fp, frm);
  }
}


