// =====================================================
//
// File:   sequence.cc
// Class:  Sequence
// 
// Author: Jrgen Ahlberg <ahlberg@isy.liu.se>
//         Dept. of Electrical Engineering
//         Linkping University
//         SE-581 83 Linkping, SWEDEN
// Date:   1998-01-12
//
// This program is distributed in the hope that it will
// be useful, but WITHOUT ANY WARRANTY.
//
// =====================================================


#include "sequence.hh"


// ============================================
//
// Constructors
//
// ============================================

Sequence::Sequence(void) :
  Matrix() {
    quant = 0;
}

Sequence::Sequence(int frames, int cols) :
  Matrix(frames, cols) {
    quant = 0;
    coding_method = 0;
}

Sequence::Sequence(int frames, Mask& newmask) :
  Matrix(frames, newmask.sum()) {
    quant = 0;
    coding_method = 0;
    setMask(newmask);
}

Sequence::Sequence(Mask& newmask, FileInfo& newinfo) :
  Matrix(newinfo.nframes, newmask.sum()) {
    quant = 0;
    coding_method = 0;
    setMask(newmask);
    setInfo(newinfo);
}


// ============================================
//
// getindex
//
// ============================================

int Sequence::
getindex(int index) {
  int i, newindex=0;
  if(index<1 || index>mask.length()) {
    return -1;
  } else {
    for(i=0;i<index-1;i++) {
      newindex += mask.bit(i);
    }
    return newindex;
  }
}


// ============================================
//
// remove_params
//
// ============================================

void Sequence::
remove_params(Mask& newmask) {
  if(newmask.length() == mask.length()) {
    Mask remmask;
    remmask.resize(ncols);
    int col = 0;
    for(int maskcol=0;maskcol<mask.length();maskcol++)
      if(mask.bit(maskcol)) {
	 remmask.set(col++, newmask.bit(maskcol));
      }
    remove_cols(remmask);
  } else {
    remove_cols(newmask);
  }
  setMask(newmask);
}


// ============================================
//
// insert
//
// ============================================


void Sequence::
insert(Sequence& src) {
  int destcol, srccol, col;

  if(mask.length() != src.mask.length()) {
    fprintf(stderr,"Error: (Sequence::insert()) Sequences with different mask-length!\n");
    exit(0);
  }

  destcol = 0; srccol = 0;
  for(col=0;col<mask.length();col++) {
    if(mask.bit(col) && src.mask.bit(col)) {
      copycolumn(src,srccol,destcol);
    }
    if(src.mask.bit(col)) srccol++;
    if(mask.bit(col)) destcol++;
  }
}


// ============================================
//
// encode / decode
//
// ============================================

void Sequence::
encode(byte _method, byte _quant) {

  float quantvector[] = FBFQUANT;
  float predcoeffs[] = PREDICTOR;
  int predorder = PREDORDER;

  float *newquantvec;

  coding_method = _method;
  quant = _quant;

  newquantvec = construct_quantvector(quantvector, qstep());

  switch(coding_method) {
  case CODING_DCT:
    //fprintf(stderr,"Encoding in DCT-mode.\n");
    encode_dct(newquantvec);
    break;
  case CODING_PRED:
    //fprintf(stderr,"Encoding in predictive mode.\n");
    encode_predictive(newquantvec, predcoeffs, predorder);
    break;
  case CODING_NONE:
    //fprintf(stderr,"Quantizing only.\n");
    quantize(newquantvec);
  }

  delete newquantvec;
}

// ============================================

void Sequence::
decode(void) {

  float quantvector[] = FBFQUANT;
  float predcoeffs[] = PREDICTOR;
  int predorder = PREDORDER;

  float *newquantvec;

  newquantvec = construct_quantvector(quantvector, qstep());

  switch(coding_method) {
  case CODING_DCT:
    //fprintf(stderr,"Decoding in DCT-mode.\n");
    decode_dct(newquantvec);
    break;
  case CODING_PRED:
    //fprintf(stderr,"Decoding in predictive mode.\n");
    decode_predictive(newquantvec, predcoeffs, predorder);
    break;
  case CODING_NONE:
    //fprintf(stderr,"Dequantizing only.\n");
    dequantize(newquantvec);
  }

  delete newquantvec;
}

// ============================================

float* Sequence::
construct_quantvector(float* qv, float q) {

  int col, i;
  float *newqv;

  newqv = new float[ncols];
  col = 0;
  for(i = 0; i < mask.size(); i++) {
    if(mask.bit(i)) {
      newqv[col++] = qv[i]*q;
      if(coding_method == CODING_DCT) {
	newqv[col-1] /= DCT_QP;
      }
    }
  }
  return newqv;
}

// ============================================
//
// encode_predictive
//
// ============================================

void Sequence::
encode_predictive(float *quantvector, float *predcoeffs, int predorder) {

  Matrix newseq, predmem;
  int row, col, frame, i;
  float tmpf;

  newseq.resize(nrows, ncols);
  predmem.resize(predorder, ncols);
  predmem.empty();

  for(frame = 0; frame < nrows; frame++) {
    newseq.copyrow(*this, frame, frame);
    for(col = 0; col < ncols; col++) {
      tmpf = 0.0;
      for(row = predorder-1; row > 0; row--) {
	tmpf += predmem.get(row, col) * predcoeffs[row];
	predmem.set(row, col, predmem.get(row-1, col));
      }
      predmem.set(0, col, tmpf + predmem.get(0, col) * predcoeffs[0]);
    }
    newseq.subtractrow(predmem, 0, frame);
    newseq.quantrow(quantvector, frame);
    for(col = 0; col < ncols; col++) {
      predmem.set(0, col, predmem.get(0, col) + Matrix::dequantize(quantvector[col], newseq.get(frame, col)));
    }
  }

  copy(newseq);

}


// ============================================
//
// decode_predictive
//
// ============================================

void Sequence::
decode_predictive(float *quantvector, float *predcoeffs, int predorder) {

  int frame, col, i;

  for(frame = 0; frame < nframes(); frame++) {
    dequantrow(quantvector, frame);
    for(i = 1; i <= predorder; i++) {
      if(frame-i >= 0) {
	for(col = 0; col < ncols; col++) {
	  set(frame, col,
	      get(frame-i, col) * predcoeffs[i-1]
	      + get(frame, col));
	}
      }
    }
  }

}


// ============================================
//
// encode_dct
//
// ============================================

void Sequence::
encode_dct(float *quantvector) {

  int row, col, newblksz;
  DCT *dct;

  dct = new DCT(DCT_BLOCKSIZE);
  for(row = 0; row < nrows-DCT_BLOCKSIZE; row += DCT_BLOCKSIZE) {
    for(col = 0; col < ncols; col ++)
      dct->encode(getp(row,col), ncols);
  }
  delete dct;

  newblksz = DCT_BLOCKSIZE;
  while(row != nrows) {
    while(nrows - row < newblksz) {newblksz /= 2;}
    dct = new DCT(newblksz);
    for(col = 0; col < ncols; col ++)
      dct->encode(getp(row,col), ncols);
    delete dct;
    row +=newblksz;
  }

  quantize(quantvector);

}

// ============================================
//
// decode_dct
//
// ============================================

void Sequence::
decode_dct(float *quantvector) {

  int row, col, newblksz;
  DCT *dct;

  dequantize(quantvector);

  dct = new DCT(DCT_BLOCKSIZE);
  for(row = 0; row < nrows-DCT_BLOCKSIZE; row += DCT_BLOCKSIZE) {
    for(col = 0; col < ncols; col ++)
      dct->decode(getp(row,col), ncols);
  }
  delete dct;

  newblksz = DCT_BLOCKSIZE;
  while(row != nrows) {
    while(nrows - row < newblksz) {newblksz /= 2;}
    dct = new DCT(newblksz);
    for(col = 0; col < ncols; col ++)
      dct->decode(getp(row,col), ncols);
    delete dct;
    row +=newblksz;
  }

}


// ==========================================
//
// nextValidLine
//
// ==========================================

void Sequence::
nextValidLine(FILE *f) {
  char c;
  c = fgetc(f);
  while(c=='\n' || c=='#') {
    c = fgetc(f);
    while(c != '\n') {
      c = fgetc(f);
    }
    c = fgetc(f);
  }
  ungetc(c,f);
}


// ==========================================
//
// report
//
// ==========================================

void Sequence::
report(void) {
  printf("FileInfo: ");
  fileinfo.write_ascii(stdout);
  Matrix::report();
}
