//-------------------------------------------------------------------------
// 
// Copyright (C) IBM Corporation & Samsung Electronics Co. LTD., 1996, 1997, 1998, 1999
// Licensed Materials, Program Property of IBM & Samsung.
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  IBM
// Corp. & Samsung Electronics Co. LTD retains full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//  AUTHORS:
//    Gabriel Taubin
//    Andre Gueziec
//    Claudio Silva
//    Ioana Martin
//  
//  IBM T. J. Watson Research Center
//  P.O.Box 704
//  Yorktown Heights, New York 10598
//  
//  BASED ON EARLIER IMPLEMENTATION OF
//  VRML COMPRESSED BINARY FORMAT PROPOSAL BY:
//    Gabriel Taubin
//    Bill Horn
//    Francis Lazarus
//
//  Functionality : Error Resilience
//
//  Sungjin Kim  (sunjin@sait.samsung.co.kr)
//  Mahn-Jin Han  (manjini@sait.samsung.co.kr)
//  Mun-Sup Song (mssong@sait.samsung.co.kr)  
//
//  Signal Processing Lab. Samsung A.I.T. 
//-------------------------------------------------------------------------

//#define ADD_VISUAL_HEADER	
#ifdef ADD_VISUAL_HEADER
const char visual_object_sequence_start_code	= 0xB0;
const char visual_object_sequence_end_code		= 0xB1;
const char visual_object_start_code				= 0xB5;
#endif //ADD_VISUAL_HEADER

#pragma warning(disable:4786)

#include <stdio.h>
#include <stdlib.h>
#include "MPG_IfsQ_E.h"
#include "MPG_QFac.h"
#include "MPG_Global.h"
#include "MPG_VGraph.h"
#include "MPG_SPolygon.h"
#include "MPG_Stitcher.h"
#include "MPG_IfsMesh.h"
#include "MPG_MeshFs.h"

#ifndef M_PI
#define M_PI    3.14159265
#endif

#ifndef MPGELIB
extern char* baseFname;
extern char* inExt;
extern char* inDir;
extern char* outDir;
extern char* inFname;
extern char* outFname;
extern char* logFname;
#endif // MPGELIB

//---------------------------------------------------------------------------
//
//   MPG_IfsQ_E::expand
//
//---------------------------------------------------------------------------

bool MPG_IfsQ_E::expand(MPG_InputStream& istrm, bool saveWrlLevels)
{
  MPG_QFpm::_shift = 16;
	
  bool success = false;

#ifdef MPGELIB
  bool newIFS = TRUE;
#endif
	
  clear();
	
  // arrays to store quantized property values
	
  // AG: in 3D_Mesh_Object_Base_Layer
  // these arrays are global for per vertex properties
  // (containing all the values decoded so far),
  // and local otherwise
  // (containing only values corresponding to the current component)
	
  // GT: in 3D_Mesh_Object_Forest_Split
  // all these arrays are global
	
  vector <int> qCoord;
  vector <int> qNormal;
  vector <int> qColor;
  vector <int> qTexCoord;
	
  // create proper headers and quantizers
  // if not yet created by calling method
  MPG_Property pCoord(PT_COORD,PB_NONE);
  MPG_BBoxQ* coordQuantizer = getCoordQuantizer();
  if((MPG_BBoxQ*)0==coordQuantizer)
	{
		coordQuantizer    = new MPG_BBoxQ(3);
		setCoordQuantizer(coordQuantizer);
	}
  pCoord.setBBoxQ(coordQuantizer);
	
  MPG_Property pNormal(PT_NORMAL,PB_NONE);
  MPG_NormalQ* normalQuantizer = getNormalQuantizer();
  if((MPG_NormalQ*)0==normalQuantizer)
	{
		normalQuantizer   = new MPG_NormalQ();
		setNormalQuantizer(normalQuantizer);
	}
  pNormal.setNormalQ(normalQuantizer);
	
  MPG_Property pColor(PT_COLOR,PB_NONE);
  MPG_BBoxQ* colorQuantizer = getColorQuantizer();
  if((MPG_BBoxQ*)0==colorQuantizer)
	{
		colorQuantizer    = new MPG_BBoxQ(3);
		setColorQuantizer(colorQuantizer);
	}
  pColor.setBBoxQ(colorQuantizer);
	
  MPG_Property pTexCoord(PT_TEXCOORD,PB_NONE);
  MPG_BBoxQ* texCoordQuantizer = getTexCoordQuantizer();
  if((MPG_BBoxQ*)0==texCoordQuantizer)
	{
		texCoordQuantizer = new MPG_BBoxQ(2);
		setTexCoordQuantizer(texCoordQuantizer);
	}
  pTexCoord.setBBoxQ(texCoordQuantizer);
	
  // quantization parameters
  int bitsPerCoord    = 0;
  int bitsPerNormal   = 0;
  int bitsPerColor    = 0;
  int bitsPerTexCoord = 0;
  
  // input bitstream parsing --------------------------------------------------
	
  MPG_Stitcher stitcher;
  vector<int>  v_father;
  MPG_Stitches v_stitches;
  vector<int>  stk;  // stitch encoding stack
	
  vector<int>  v_visited;
  vector<int>  v_cluster;
  int          nV0    = 0; // number of vertices so far without current cc
  int          nV1    = 0; // number of vertices so far with current cc
  int          nVcls0 = 0; // AG: number of vertices after clustering
  int          nVcls1 = 0; // AG: number of vertices after clustering
	
	// open a bit stream
	MPG_InputBitStream ibstrm(istrm);
	
#ifdef ADD_VISUAL_HEADER
  {
		if(istrm.getVisualStartCode(visual_object_sequence_start_code)==false)
			return false;
		
		int profile_and_level_indication = 0;
		int is_visual_object_identifier = 0;
		int visual_object_verid = 0;
		int visual_object_priority = 0;
		int visual_object_type = 0;

		if(ibstrm.getMBitInt(8,profile_and_level_indication)==false) return false;
		//if(profile_and_level_indication != 0xFF) return false;	//MJH 062600 : 0xF1-->0xFF
		
		if(istrm.getVisualStartCode(visual_object_start_code)==false)
			return false;
		
		if(ibstrm.getMBitInt(1,is_visual_object_identifier)==false) return false;
		//if(is_visual_object_identifier != 0x01) return false;		

		if(is_visual_object_identifier == 0x01) //MJH 062600 :
		{
			if(ibstrm.getMBitInt(4,visual_object_verid)==false) return false;
			//if(visual_object_verid != 0x0F) return false;	
			
			if(ibstrm.getMBitInt(3,visual_object_priority)==false) return false;
			//if(visual_object_priority != 0x01) return false;		
		}

		if(ibstrm.getMBitInt(4,visual_object_type)==false) return false;
		//if(visual_object_type != 0x05) return false;		
  }
#endif //ADD_VISUAL_HEADER
	
  // decode 3D_Mesh_Object() 
  //
  // 3D_Mesh_Object() 
  //   3D_MO_start_code
  //   3D_Mesh_Object_Header()
  //   do 
  //     3D_Mesh_Object_Layer()
  //   while(nextbits_bytealigned()==3D_MOL_start_code)

  // read 3D_MO_start_code
  if(ibstrm.getMOStartCodeBitInt(MPG_Global::MO_start_code)==false)
    return false;
	
  // decode 3D_Mesh_Object_Header()
  {
    // 3D_Mesh_Object_Header() 
    //   ccw
    //   convex
    //   solid
    //   creaseAngle
    //   coord_header()
    //   normal_header()
    //   color_header()
    //   texCoord_header()
    //   ce_SNHC
    //   if(ce_SNHC=='1')
    //     ce_SNHC_header()
		
	if(ibstrm.getMOBit(_ccw               )==false) return false;
    if(ibstrm.getMOBit(_convex            )==false) return false;
    if(ibstrm.getMOBit(_solid             )==false) return false;
    int cAng = 0;
	if(ibstrm.getMOBitInt(6,cAng)==false) return false;
    if(cAng<0 || cAng>=64) return false;
    _creaseAngle = M_PI*((float)cAng)/32;
		
    // decode coord_header()
    {
      // coord_header() 
      //   coord_binding       <--- must be == 1 (PER_VERTEX)
			
      int b = 0;
      if(ibstrm.getMBitInt(2,b)==false) return false;
      MPG_Binding binding =
        (b==0)?PB_NONE:(b==1)?PB_PER_VERTEX:(b==2)?PB_PER_FACE:PB_PER_CORNER;
      if(binding!=PB_NONE)
			{
				// coord can only be bound PER_VERTEX
				if(binding!=PB_PER_VERTEX) return false;
				pCoord.setBinding(binding);
				
				//   coord_bbox
				//   if(coord_bbox) 
				//     coord_xmin
				//     coord_ymin
				//     coord_zmin
				//     coord_size
				//   coord_quant
				//   coord_pred_type
				//   if (coord_pred_type=="tree_prediction" ||
				//       coord_pred_type=='"parallelogram_prediction") 
				//     coord_nlambda
				//     for(i=1;i<coord_nlambda;i++)
				//       coord_lambda[i]
				
				success = pCoord.decode(ibstrm);
				if(success==false) return false;
			}
    }
		
    // decode normal_header()
    {
      // normal_header() 
      //   normal_binding
      //   if(normal_binding!="not_bound") 
			
      int b = 0;
      if(ibstrm.getMBitInt(2,b)==false) return false;
      MPG_Binding binding =
        (b==0)?PB_NONE:(b==1)?PB_PER_VERTEX:(b==2)?PB_PER_FACE:PB_PER_CORNER;
      if(binding!=PB_NONE)
			{
				pNormal.setBinding(binding);
				
				//     normal_bbox       <--- must be == 0
				//     normal_quant
				//     normal_pred_type
				//     if (normal_pred_type=="tree_prediction"||
				//         normal_pred_type=="parallelogram_prediction") 
				//       normal_nlambda
				//       for(i=1;i<normal_nlambda;i++)
				//         normal_lambda[i]
				
				success = pNormal.decode(ibstrm);
				if(success==false) return false;
			}
      _normalPerVertex = (binding==PB_PER_FACE)?false:true;
    }
		
    // decode color_header()
    {
      // color_header()
      //   color_binding
      //   if(color_binding!="not_bound")
			
      int b = 0;
      if(ibstrm.getMBitInt(2,b)==false) return false;
      MPG_Binding binding =
        (b==0)?PB_NONE:(b==1)?PB_PER_VERTEX:(b==2)?PB_PER_FACE:PB_PER_CORNER;
      if(binding!=PB_NONE)
			{
				pColor.setBinding(binding);
				
				//     color_bbox
				//     if(color_bbox == '1') 
				//       color_rmin
				//       color_gmin
				//       color_bmin
				//       color_size
				//     color_quant
				//     color_pred_type
				//     if (color_pred_type=="tree_prediction" ||
				//         color_pred_type=='"parallelogram_prediction")
				//       color_nlambda
				//       for(i=1;i<color_nlambda;i++)
				//         color_lambda[i]
				
				success = pColor.decode(ibstrm);
				if(success==false) return false;
			}
      _colorPerVertex = (binding==PB_PER_FACE)?false:true;
    }
		
    // decode texture_header()
    {
      // texCoord_header() 
      //   texCoord_binding     <--- must be != 2 (PER_FACE)
      //   if(texCoord_binding!="not_bound")
			
      int b = 0;
      if(ibstrm.getMBitInt(2,b)==false) return false;
      MPG_Binding binding =
        (b==0)?PB_NONE:(b==1)?PB_PER_VERTEX:(b==2)?PB_PER_FACE:PB_PER_CORNER;
      if(binding!=PB_NONE)
			{
				// TexCoord cannot be bound per_face
				if(binding==PB_PER_FACE) return false;
				pTexCoord.setBinding(binding);
				
				//     texCoord_bbox
				//     if(texCoord_bbox=='1')
				//       texCoord_umin
				//       texCoord_vmin
				//       texCoord_size
				//     texCoord_quant
				//     texCoord_pred_type
				//     if (texCoord_pred_type=="tree_prediction" ||
				//         texCoord_pred_type=="parallelogram_prediction") 
				//       texCoord_nlambda
				//       for(i=1;i<texCoord_nlambda;i++)
				//         texCoord_lambda[i]
				
				success = pTexCoord.decode(ibstrm);
				if(success==false) return false;
			}
    }
		
    // decode cgd_data
    {
      _cgd_data = false;
      if(ibstrm.getBit(_cgd_data)==false) return false;
			
      if(_cgd_data==true)
			{
				// decode ce_SNHC_header()
				//
				//	ce_SNHC_n_proj_surface_spheres
				//	if (ce_SNHC_n_proj_surface_spheres != 0) 
				//		ce_SNHC_x_coord_center_point
				//		ce_SNHC_y_coord_center_point
				//		ce_SNHC_z_coord_center_point
				//		ce_SNHC_normalized_screen_distance_factor
				//		for (i=0; i< ce_SNHC_n_proj_surface_spheres; i++) 
				//			ce_SNHC_radius
				//			ce_SNHC_min_proj_surface
				//			ce_SNHC_n_proj_points
				//			for (j=0; j< ce_SNHC _n_proj_points; j++) 
				//				ce_SNHC_sphere_point_coord
				//				ce_SNHC_proj_surface
				//
				// for now : just parse and ignore values
				// later   : create a new class to store values
				//
				
				int N_Proj_Surface_Spheres = 0;
				if(ibstrm.getMBitInt(4,N_Proj_Surface_Spheres)==false) return false;
				if(N_Proj_Surface_Spheres>0)
				{
					float x_coord_Center_Point = 0.0;
					if(ibstrm.get32BitFloat(x_coord_Center_Point)==false)
						return false;
					
					float y_coord_Center_Point = 0.0;
					if(ibstrm.get32BitFloat(y_coord_Center_Point)==false)
						return false;
					
					float z_coord_Center_Point = 0.0;
					if(ibstrm.get32BitFloat(z_coord_Center_Point)==false)
						return false;
					
					int Normalized_Screen_Distance_Factor = 0;
					if(ibstrm.getMBitInt(8,Normalized_Screen_Distance_Factor)==false)
						return false;
					
					int i;
					for(i=0;i<N_Proj_Surface_Spheres;i++) {
						
						float Radius = 0.0;
						if(ibstrm.get32BitFloat(Radius)==false)
							return false;
						
						float Min_Proj_Surface = 0.0;
						if(ibstrm.get32BitFloat(Min_Proj_Surface)==false)
							return false;
						
						int N_Proj_Points = 0;
						if(ibstrm.getMBitInt(8,N_Proj_Points)==false)
							return false;
						
						int j;
						for(j=0;j<N_Proj_Points;j++)
						{
							
							int Sphere_Point_Coord = 0;
							if(ibstrm.getMBitInt(11,Sphere_Point_Coord)==false)
								return false;
							
							float Proj_Surface = 0.0;
							if(ibstrm.get32BitFloat(Proj_Surface)==false)
								return false;
							
						}
					}
				}
			}
    }
		
    // flush input buffer
    ibstrm.flush();
    
  } // end decode 3D_Mesh_Object_Header()
	
  bitsPerCoord    =  pCoord.getBBoxQ()->getBitsPerCoordinate();
  bitsPerNormal   =  pNormal.getNormalQ()->getBitsPerNormal();
  bitsPerColor    =  pColor.getBBoxQ()->getBitsPerCoordinate();
  bitsPerTexCoord =  pTexCoord.getBBoxQ()->getBitsPerCoordinate();
	
  // decode one or more 3D_Mesh_Object_Layer()
  int mol_id = 0;
  do {
    // decode 3D_Mesh_Object_Layer()
		
    // 3D_Mesh_Object_Layer()
		//   3D_MOL_start_code
    //   mol_id
    //   if(ce_SNHC =='1')
    //     ce_SNHC_n_vertices
    //     ce_SNHC_n_triangles
    //     ce_SNHC_n_edges
    //   if(mol_id==0)
    //     3D_Mesh_Object_Base_Layer()
    //   else
    //     3D_Mesh_Object_Refinement_Layer()
		
    // read 3D_MOL_start_code
    if(istrm.getStartCode(MPG_Global::MOL_start_code)==false)
      return false;
		
    {
      // open a bit stream
      MPG_InputBitStream ibstrm(istrm);
			
      // read mol_id
      if(ibstrm.getMBitInt(8,mol_id)==false) return false;
			
      // read cgd data
      if(_cgd_data==true)
			{
				int N_Vertices = 0;
				if(ibstrm.getMBitInt(24,N_Vertices)==false) return false;
				
				int N_Triangles = 0;
				if(ibstrm.getMBitInt(24,N_Triangles)==false) return false;
				
				int N_Edges = 0;
				if(ibstrm.getMBitInt(24,N_Edges)==false) return false;
			}
      
      ibstrm.flush();
    }
		
    if(mol_id==0)
		{
			//
			// decode 3D_Mesh_Object_Base_Layer()
			//
			
			// 3D_Mesh_Object_Base_Layer() 
			//	do 
			//    3D_MOBL_start_code
			//    mobl_id
			//		while (!bytealigned())
			//			one_bit
			//		qf_start()
			//		if (3D_MOBL_start_code == "partition_type_0") 
			//			...
			//		else if (3D_MOBL_start_code == "partition_type_1") 
			//			...
			//		else if (3D_MOBL_start_code == "partition_type_2") 
			//			...
			//	while (nextbits_bytealigned() == 3D_MOBL_start_code)
			
			int previous_mobl_id = -1; // Index of previous MOBL; used to check correct sequencing
			int vg_number = 0;
			int nVcc;	
			
			typedef vector<int> vec_array;
			vector<vec_array> v_loop_cc;
			vector<int> nVtotal;	// for pt1
			MPG_SPolygon sp;	//MJH 001006 : moved from below
			
			do {  // 3D_Mesh_Object_Base_Layer()
				// read 3D_MOBL_start_code
				// get mobl start code and determine the partition type
				
synchronize:
			int partition_type;
			unsigned char mobl_sc = istrm.getMOBLStartCode(); 
			if (mobl_sc == MPG_Global::MOBL0_start_code)
				partition_type = 0;
			else if (mobl_sc == MPG_Global::MOBL1_start_code)
				partition_type = 1;
			else if (mobl_sc == MPG_Global::MOBL2_start_code)
				partition_type = 2;
			else
				break;
			
			// open a bit stream
			MPG_InputBitStream ibstrm(istrm);     
			
			// read mobl_id
			int mobl_id = 0;
			if(ibstrm.getMBitInt(8,mobl_id)==false) 
			{
				fprintf(stderr, "Error in mobl_id\n");
				goto synchronize;	
			}
//			else
//				printf("decoding mobl : %d\n", mobl_id);
			
			int correct_sequencing = (mobl_id == ((previous_mobl_id + 1) & 0xff));
			//Check for partition loss error
			if (!correct_sequencing) {
				printf("Warning: partition number sequence incomplete. ");
				printf("(%d, %d)\n", previous_mobl_id, mobl_id);
			}
			previous_mobl_id = mobl_id;
			
			ibstrm.flush();
			
			//  start the arithmetic codec
			MPG_QFac codec(istrm);
			codec.start_decode();
			MPG_Global::reset();
			
			///////////////////////////////////////////////////////////////
			// partition type 0 ( one or more cc's)
			//	if (3D_MOBL_start_code == "partition_type_0") 
			//		do 
			//			connected_component()
			//			qf_decode(last_component, last_component_context)
			//		while (last_component == '0')
			//
			if (partition_type == 0) {	
				vg_number=0;	
				
				do {
					// decode connected_component()
					
					// connected_component() 
					//   vertex_graph()
					//   qf_decode(has_stitches, has_stitches_context)
					//   if(has_stitches=='1')
					//     stitches()
					//   triangle_tree()
					//   triangle_data()
					// 
					
					// decode vertex_graph() and construct vertex loop
					nV0 = nV1;	
					v_loop_cc.clear(); 
					vec_array v_loop_cc0;
					v_loop_cc.push_back(v_loop_cc0);
					{
						// decode vertex_graph()
						MPG_VGraph vg;
						if(vg.decode(codec) == false) return false;
						
						nVcc = vg.getNumberOfVertices();
						nV1 = nV0 + nVcc;
						while(v_visited.size()<nV1)
							v_visited.push_back(false);
						
						// convert to vertex loop look-up table
						MPG_Loops loop;
						vg.toLoop(loop);
						loop.getLoop(0, v_loop_cc[0]);
						
						if(vg.getFather(v_father)==false || v_father.size()!=nV1)
							return false;
					}
					
					v_stitches.decode(nV0,nV1,stk,codec);
					{
						// use stitches nV0<=i<nV1 to partially recover v_cluster
						
						if(stitcher.use(nV0,v_father,v_stitches,v_cluster)==false)
							return false;
					}
					
					// we should have here :
					// v_cluster.size()==nV1
					
					if(v_cluster.size()!=nV1)
						return false;
					
					// for i=0,..,nV1-1
					//   v_cluster[i]>=0 && v_cluster[i]<nV1
					// v_father.size()==nV1 
					// for i=0,..,nV1-1
					//   v_father[i]<=i
					// v_stitches.size()==nV1
					
					// decode triangle_tree()
//					MPG_SPolygon sp;	//MJH 001006 : moved to top
					sp.clear();	//MJH 001006
					
					sp.codap_left_bloop_idx = 1;	
					sp.codap_right_bloop_idx = 0; 
					sp.codap_bdry_pred = 1;
					
					if(sp.decodeTriangleTree(codec,partition_type,v_loop_cc[0].size()) == false)	
						goto synchronize;
					
					// decode triangle_data()
					vector<int>   coordIndex;
					if(sp.decodeTriangleData(codec,
						nVcc,
						nV0,
						v_loop_cc[0],
						v_visited,
						v_cluster,
						coordIndex,
						// property headers and quantizers
						pCoord,
						pNormal,
						pColor,
						pTexCoord,
						// quantized values
						qCoord,
						qNormal,
						qColor,
						qTexCoord,
#ifdef MPGELIB
						newIFS,	//MJH 082100
#endif
						partition_type	
						)==false) 
						goto synchronize;	
					
					// transfer coordIndex to Ifs
					{
						int j0,j1,j;
						for(j0=j1=0;j1<coordIndex.size();j1++)
							if(coordIndex[j1]==-1)	
							{
								int nf = j1-j0;
//								bool skip_poly = 0;		//MJH 000111 : skip_poly related not needed for pt0
								int* f = &(coordIndex[j0]);
								for(j=0;j<nf;j++)
								{
//									if (f[j] == -2)
//										skip_poly = 1;
//									if (skip_poly)
//										_coordIndex.push_back(-2);
//									else
										_coordIndex.push_back(f[j]);
								}
								_coordIndex.push_back(-1);
//								skip_poly = 0;
								j0 = j1+1;
							}
					}
					
			} while(codec.decode(MPG_Global::last_component) == false);	
			
			nVtotal.clear();
			nVtotal.push_back(nV0);
			nVtotal.push_back(nV1);	
		}
		
		///////////////////////////////////////////////////////////////
		// partition type 1 ( one or more vg's)
		//	else if (3D_MOBL_start_code == "partition_type_1") 
		//		vg_number=0
		//		do 
		//			vertex_graph()
		//			vg_number++
		//			qf_decode(has_stitches, has_stitches_context)
		//			if (has_stitches == '1')
		//				stitches()
		//			qf_decode(codap_last_vg, codap_last_vg_context)
		//		while (codap_last_vg == '0')
		//		
		else if (partition_type == 1) {	
			vg_number=0;	
			v_loop_cc.clear(); 
			nVtotal.clear();
			nVtotal.push_back(nV1);
			int codap_last_vg;
			
			do {
				nV0 = nV1;	
				
				vec_array v_loop_cc1;
				v_loop_cc.push_back(v_loop_cc1);
				
				// decode vertex_graph()
				MPG_VGraph vg;
				if(vg.decode(codec) == false) return false;
				
				nVcc = vg.getNumberOfVertices();
				nV1 = nV0 + nVcc;
				
				nVtotal.push_back(nV1);
				
				while(v_visited.size()<nV1)
					v_visited.push_back(false);
				
				// convert to vertex loop look-up table
				MPG_Loops loop;
				vg.toLoop(loop);
				loop.getLoop(0, v_loop_cc[vg_number]);
				
				if(vg.getFather(v_father)==false || v_father.size()!=nV1)
				{
					return false;
				}
				
				v_stitches.decode(nV0,nV1,stk,codec);
				{
					// use stitches nV0<=i<nV1 to partially recover v_cluster
					
					if(stitcher.use(nV0,v_father,v_stitches,v_cluster)==false)
						return false;
				}
				
				// we should have here :
				// v_cluster.size()==nV1
				
				if(v_cluster.size()!=nV1)
					return false;
				
				// for i=0,..,nV1-1
				//   v_cluster[i]>=0 && v_cluster[i]<nV1
				// v_father.size()==nV1 
				// for i=0,..,nV1-1
				//   v_father[i]<=i
				// v_stitches.size()==nV1
				
				vg_number++;	
				codap_last_vg = codec.decode(MPG_Global::codap_last_vg_context);
				
			} while (!codap_last_vg);
			
		} // if (partition_type == 1)
		
		///////////////////////////////////////////////////////////////
		// partition type 2 (one tt/td pair)
		//	else if (3D_MOBL_start_code == "partition_type_2") 
		//		if(vg_number > 1)
		//			qf_decode(codap_vg_id)
		//		qf_decode(codap_left_bloop_idx)
		//		qf_decode(codap_right_bloop_idx)
		//		qf_decode(codap_bdry_pred)
		//		triangle_tree()
		//		triangle_data()
		//		
		else { // if (partition_type == 2)
			
			// decode triangle_tree()
//			MPG_SPolygon sp;	//MJH 001006 : moved to top
			sp.clear();	//MJH 001006
			
			int l,nbits;
			int codap_vg_id = 0;
			if (vg_number>1) {
				nbits = slog2(vg_number);
				for (l=0; l<nbits; l++)
					codap_vg_id += codap_vg_id + codec.decode();
			}
			nbits = slog2(v_loop_cc[codap_vg_id].size());
			sp.codap_left_bloop_idx = 0;	
			for (l=0; l<nbits; l++)
				sp.codap_left_bloop_idx += sp.codap_left_bloop_idx + codec.decode();
			sp.codap_right_bloop_idx = 0; 
			for (l=0; l<nbits; l++)
				sp.codap_right_bloop_idx += sp.codap_right_bloop_idx + codec.decode();
			sp.codap_bdry_pred = codec.decode();
			
			// decode triangle_tree()
			if(sp.decodeTriangleTree(codec,partition_type,v_loop_cc[codap_vg_id].size()) == false) //return false;
				goto synchronize;
			
			nV0 = nVtotal[codap_vg_id];
			nVcc = nVtotal[codap_vg_id+1]-nVtotal[codap_vg_id];
			
			// decode triangle_data()
			vector<int>   coordIndex;
			if(sp.decodeTriangleData(codec,
				nVcc,
				nV0,
				v_loop_cc[codap_vg_id],
				v_visited,
				v_cluster,
				coordIndex,
				// property headers and quantizers
				pCoord,
				pNormal,
				pColor,
				pTexCoord,
				// quantized values
				qCoord,
				qNormal,
				qColor,
				qTexCoord,
#ifdef MPGELIB
				newIFS,	//MJH 082100
#endif
				partition_type	
				)==false) 
				goto synchronize;	
			
			// transfer coordIndex to Ifs
			{
				int j0,j1,j;
				for(j0=j1=0;j1<coordIndex.size();j1++)
					if(coordIndex[j1]==-1)	
					{
						int nf = j1-j0;
						bool skip_poly = 0;		
						int* f = &(coordIndex[j0]);
						for(j=0;j<nf;j++)
						{
							if (f[j] == -2)
								skip_poly = 1;
							_coordIndex.push_back(f[j]);
						}
						if (skip_poly)
						{
							int corner_cnt;
							//MJH 052500 : _coordIndex.size()>0 added in case the first triangle in _coordIndex has missing vertex
							for (corner_cnt=0; _coordIndex.back() != -1 && _coordIndex.size()>0 
								; corner_cnt++)
								_coordIndex.pop_back();
							for ( ; corner_cnt>0; corner_cnt--)
								_coordIndex.push_back(-2);
						}
						_coordIndex.push_back(-1);
						skip_poly = 0;
						j0 = j1+1;
					}
			}
				  
		} // if (partition_type == 2)
		///////////////////////////////////////////////////////////////
		
		codec.finish_decode();
		
	} while (1);	// 3D_Mesh_Object_Base_Layer()
	
	
	int i0 = 0; 
	nV0 = 0;	
	
	// determine the new number of vertices after clustering
	{
		int i;
		for (i=nV0; i < nV1; i++)
			if (v_cluster[i] > nVcls1) nVcls1 = v_cluster[i];
	}
	int nVcls11 = nVcls1+1;
	
	// unquantize vertices between nVcls0 and nVcls1
	
	// coord properties
	vector<float> coord((size_t) (3*(nVcls11-nVcls0)),0.0);
	MPG_BBoxQ*   coordQuantizer    = pCoord.getBBoxQ();
	coordQuantizer->unQuantize(qCoord.begin()+3*nVcls0,
		coord.begin(),
		nVcls11-nVcls0);
	
	//  normal properties
	vector<float> normal;
	MPG_NormalQ* normalQuantizer   = pNormal.getNormalQ();
	if(pNormal.getBinding() != PB_NONE)
	{
		// special case for per-vertex binding,
		// the qNormal coordinates are global to allow reuse
		// of some vertices by components decoded afterwards
		if(pNormal.getBinding() == PB_PER_VERTEX)
		{
			// take only the trailing part of qNormal
			vector <int> qNormal1;
			int i;
			for (i=nVcls0;i<nVcls11;i++)
				qNormal1.push_back(qNormal[i]);
			
			normalQuantizer->unQuantize(qNormal1, normal);
		}
		else normalQuantizer->unQuantize(qNormal, normal);
	}
	//  color properties
	vector<float> color;
	MPG_BBoxQ*   colorQuantizer    = pColor.getBBoxQ();
	if(pColor.getBinding() != PB_NONE)
	{
		if(pColor.getBinding() == PB_PER_VERTEX)
		{
			color.erase(color.begin(),color.end());
			color.insert(color.end(),(size_t)(3*(nVcls11-nVcls0)),0.0);
			colorQuantizer->unQuantize(qColor.begin()+3*nVcls0,
				color.begin(),
				nVcls11-nVcls0);
		}
		else colorQuantizer->unQuantize(qColor, color);
	}
	//  texture properties
	vector<float> texCoord;
	MPG_BBoxQ*   texCoordQuantizer = pTexCoord.getBBoxQ();
	if(pTexCoord.getBinding() != PB_NONE)
	{
		if(pTexCoord.getBinding() == PB_PER_VERTEX)
		{
			texCoord.erase(texCoord.begin(),texCoord.end());
			texCoord.insert(texCoord.end(),
				(size_t)(2*(nVcls11-nVcls0)),0.0);
			texCoordQuantizer->unQuantize(qTexCoord.begin()+2*nVcls0,
				texCoord.begin(),
				nVcls11-nVcls0);
		}
		else
			texCoordQuantizer->unQuantize(qTexCoord, texCoord);
	}
	
	// transfer coord to Ifs
	{
		int i;
		for(i=0;i<coord.size();i++) _coord.push_back(coord[i]);
	}
	
	if(pNormal.getBinding()==PB_PER_CORNER)
	{
		// adjust _normalIndex
		// normal.size() should be equal to
		// the number of new corners
		int i = i0;
		int j = _normal.size()/3;
		for(;i<_coordIndex.size();i++)
			if(_coordIndex[i]!=-1)
			{ _normalIndex.push_back(j); j++; }
			else	
			{ 
				if (_normalIndex.back()==-2)
				{
					while (_normalIndex.back() != -1)
						_normalIndex.pop_back();
				}
				else
					_normalIndex.push_back(-1); 
				  }
	}
	
	// transfer normal to Ifs
	{
		int i;
		for(i=0;i<normal.size();i++)
			_normal.push_back(normal[i]);
	}
	
	if(pColor.getBinding()==PB_PER_CORNER)
	{
		// adjust _colorIndex
		// color.size() should be equal to
		// the number of new corners
		int i = i0;
		int j = _color.size()/3;
		for(;i<_coordIndex.size();i++)
			if(_coordIndex[i]!=-1)
			{ _colorIndex.push_back(j); j++; }
			else
			{ 
				if (_colorIndex.back()==-2)
				{
					while (_colorIndex.back() != -1)
						_colorIndex.pop_back();
				}
				else
					_colorIndex.push_back(-1); 
			}
	}
	
	// transfer color to Ifs
	{
		int i;
		for(i=0;i<color.size();i++)
			_color.push_back(color[i]);
	}
	
	if(pTexCoord.getBinding()==PB_PER_CORNER)
	{
		// adjust _texCoordIndex
		// texCoord.size() should be equal to
		// the number of new corners
		int i = i0;
		int j = _texCoord.size()/2;
		for(;i<_coordIndex.size();i++)
			if(_coordIndex[i]!=-1)
			{ _texCoordIndex.push_back(j); j++; }
			else
			{ 
				if (_texCoordIndex.back()==-2)
				{
					while (_texCoordIndex.back() != -1)
						_texCoordIndex.pop_back();
				}
				else
					_texCoordIndex.push_back(-1); 
			}
	}
	
	// transfer texCoord to Ifs
	{
		int i;
		for(i=0;i<texCoord.size();i++)
			_texCoord.push_back(texCoord[i]);
	}
	
	nVcls0 = nVcls11;
	nV0 = nV1;
    }
		
    else // if(mol_id!=0)
    {
			//
			//
			// decode 3D_Mesh_Object_Refinement_Layer()
			//
			//
			
			// progressive mode only defined for manifold base mesh
			{
				MPG_IfsMesh mesh(*this);
				if(mesh.isOrientedManifold()==false)
					return false;
			}
			
			// reset the per_vertex property bindings, if any,
			// but save original setings;
			// to prevent MPG_SPolygon::decodeTriangleData()
			// to decode per vertex properties
			MPG_Binding bCoord    = pCoord.getBinding();
			if(bCoord == PB_PER_VERTEX)    pCoord.setBinding(PB_NONE);
			MPG_Binding bNormal   = pNormal.getBinding();
			if(bNormal == PB_PER_VERTEX)   pNormal.setBinding(PB_NONE);
			MPG_Binding bColor    = pColor.getBinding();
			if(bColor == PB_PER_VERTEX)    pColor.setBinding(PB_NONE);
			MPG_Binding bTexCoord = pTexCoord.getBinding();
			if(bTexCoord == PB_PER_VERTEX) pTexCoord.setBinding(PB_NONE);
			
			// quantize properties
			// from now on, these arrays are global
			if(bCoord!=PB_NONE &&
				coordQuantizer->quantize(_coord,qCoord)==false)
				return false;
			if(bNormal!=PB_NONE &&
				normalQuantizer->quantize(_normal,qNormal)==false)
				return false;
			if(bColor!=PB_NONE &&
				colorQuantizer->quantize(_color,qColor)==false)
				return false;
			if(bTexCoord!=PB_NONE &&
				texCoordQuantizer->quantize(_texCoord,qTexCoord)==false)
				return false;
			
			// 3D_Mesh_Object_Refinement_Layer() 
			//   do 
			//     3D_MORL_start_code
			//     morl_id
			//
			//     connectivity_update
			//
			//     pre_smoothing
			//     if(pre_smoothing)
			//       pre_smoothing_parameters()
			//     post_smoothing
			//     if(post_smoothing)
			//       post_smoothing_parameters()
			//
			//     start_qcoder()
			//
			//     if(connectivity_update==FOREST_SPLIT)
			//       fs_pre_update()
			//
			//     if(pre_smoothing || post_smoothing)
			//       smoothing_constraints()
			//
			//     if(pre_smoothing)
			//       apply pre_smoothing as predictor
			//       to vertex coordinates and re-quantize
			//
			//     if(connectivity_update==FOREST_SPLIT)
			//       fs_post_update()
			//
			//     if(connectivity_update!=NOT_UPDATED)
			//       other_update
			//
			//     if(connectivity_update==NOT_UPDATED || other_update)
			//       other_property_update()
			//
			//     if(post_smoothing)
			//       apply post_smoothing as predictor
			//       to vertex coordinates and re-quantize
			//
			//     stop_qcoder()
			//
			//   while(nextbits_bytealigned()==3D_MORL_start_code)
			// 
			
			do {
				// read 3D_MORL_start_code
				if(istrm.getStartCode(MPG_Global::MORL_start_code)==false)
					return false;
				
				int   morl_id               = 0;
				
				int   connectivity_update   = 0;
				
				bool  pre_smoothing         = false;
				int   pre_smoothing_n       = 0;
				float pre_smoothing_lambda  = 0;
				float pre_smoothing_mu      = 0;
				
				bool  post_smoothing        = false;
				int   post_smoothing_n      = 0;
				float post_smoothing_lambda = 0;
				float post_smoothing_mu     = 0;
				
				{
					// open a bit stream
					MPG_InputBitStream ibstrm(istrm);       
					
					// read morl_id
					if(ibstrm.getMBitInt(8,morl_id)==false) return false;
					
					if(ibstrm.getMBitInt(2,connectivity_update)==false) return false;
					
					if(ibstrm.getBit(pre_smoothing)==false) return false;
					if(pre_smoothing==true)
					{
						if(ibstrm.getMBitInt(8,pre_smoothing_n)==false)
							return false;
						if(ibstrm.get32BitFloat(pre_smoothing_lambda)==false)
							return false;
						if(ibstrm.get32BitFloat(pre_smoothing_mu)==false)
							return false;
					}
					
					if(ibstrm.getBit(post_smoothing)==false) return false;
					if(post_smoothing==true)
					{
						if(ibstrm.getMBitInt(8,post_smoothing_n)==false)
							return false;
						if(ibstrm.get32BitFloat(post_smoothing_lambda)==false)
							return false;
						if(ibstrm.get32BitFloat(post_smoothing_mu)==false)
							return false;
					}
					
					// close bitstream
					ibstrm.flush();
				}
				
				//     start_qcoder()
				MPG_QFac codec(istrm);
				codec.start_decode();
				MPG_Global::reset();
				
				// build mesh with no spanning trees
				// MPG_IfsMesh mesh(*this);
				MPG_MeshFs mesh(getCoordIndex());
				
				int nV = mesh.getNumberOfVertices();
				int nF = mesh.getNumberOfFaces();
				int nE = mesh.getNumberOfEdges();
				
				// variables used to construct and represent forest
				vector<bool>        is_forest_edge((size_t)nE,false);
				vector<MPG_pVecInt> vp_coordIndex;
				vector<int>         nTrees;
				
				// fs_pre_update() 
				//   for each connected component 
				//     forest()
				//     for each tree in forest 
				//       triangle_tree()
				//       /* for each tree loop vertex set visited=1 */
				//       triangle_data()
				
				// for each connected component 
				int nCC = mesh.getNumberOfConnectedComponents();
				for(mesh.ccBegin();mesh.ccEnd()==false;mesh.ccNext())
				{
					int icc      = mesh.getIndexCC();
					
					int nVcc     = mesh.getNumberOfVerticesCC();
					int nFcc     = mesh.getNumberOfFacesCC();
					int nEcc     = mesh.getNumberOfEdgesCC();
					int nTreesCC = 0;
					
					// decoding forest()
					{
						// decode part of forest contained in
						// current connected component
						MPG_Partition p(nVcc);
						MPG_Graph::pEdge e; 
						for(e=mesh.getFirstEdgeCC();
						e!=(MPG_Graph::pEdge)0;
						e=mesh.getNextEdgeCC())
						{
							int i0cc = mesh.getInvCoordIndexCC(e->getVertex(0));
							int i1cc = mesh.getInvCoordIndexCC(e->getVertex(1));
							if(p.find(i0cc)!=p.find(i1cc) &&
								codec.decode(MPG_Global::pfs_forest_edge)==1)
							{
								is_forest_edge[e->getIndex()] = true;
								p.join(i0cc,i1cc);
							}
						}
						
						// count number of trees in forest
						if(p.parts()<nVcc)
						{
							p.makeParts();
							
							int i;
							for(i=0;i<nVcc;i++)
								if(p.size(i)>1)
									nTreesCC++;
						}
					}
					
					nTrees.push_back(nTreesCC);
					
					//   for each tree in forest
					//     decode one simple polygon with no
					//     associated property data per vertex;
					//     but with per face and per corner
					//     property data if present
					
					//   for each tree in forest 
					//     triangle_tree()
					//     for each vertex in tree loop
					//       visited = 1
					//     triangle_data()
					
					int i;
					for(i=0;i<nTreesCC;i++)
					{
						MPG_SPolygon sp;
						if(sp.decodeTriangleTree(codec) == false) return false;
						int nVi = sp.getNumberOfVertices();
						
						// decode triangle_data()
						
						// since all the per vertex property binding
						// have been reset, the vectors
						// v_loop, v_visited and v_cluster are ignored
						// (dummy variables)
						
						vector<int> v_visited;
						vector<int> v_loop;
						vector<int> v_cluster;
						
						// arrays to contain the quantized
						// per face and per corner
						// decoded for this simple polygon
						vector<int> i_qCoord;
						vector<int> i_qNormal;
						vector<int> i_qColor;
						vector<int> i_qTexCoord;
						
						// array to contain the simple polygon faces
						// on the boundary loop indices
						
						// coordIndex array is saved
						// to be passed to mesh.decodeForestSplit
						// after exiting this loop
						
						MPG_pVecInt coordIndex = new vector<int>();
						vp_coordIndex.push_back(coordIndex);
						
						if(sp.decodeTriangleData(codec,
							nVi,
							0,
							v_loop,
							v_visited,
							v_cluster,
							*coordIndex,
							pCoord,
							pNormal,
							pColor,
							pTexCoord,
							i_qCoord,
							i_qNormal,
							i_qColor,
							i_qTexCoord
#ifdef MPGELIB
							,newIFS	//MJH 082100
#endif
							)==false) return false;
						
						// since no per vertex properties are decoded here
						// i_qCoord should be empty
						
						// but per face properties should have been collected in
						// i_qNormal, i_qColor, and i_qTexCoord;
						// transfer them to the global quantized arrays
						
						if(i_qNormal.size()>0)
							qNormal.insert(qNormal.end(),
							i_qNormal.begin(),i_qNormal.end());
						
						if(i_qColor.size()>0)
							qColor.insert(qColor.end(),
							i_qColor.begin(),i_qColor.end());
						
						if(i_qTexCoord.size()>0)
							qTexCoord.insert(qTexCoord.end(),
							i_qTexCoord.begin(),i_qTexCoord.end());
						
					} // for(i=0;i<nTreesCC;i++)
					
            } // for(mesh.ccBegin();mesh.ccEnd()==false;mesh.ccNext())
						
						// here the mesh has not changed yet, but
						// we have the forest edges marked,
						// each simple polygon expanded into one coordIndex array,
						// and all the quantized per face and per corner properties
						// corresponding to the new faces 
						
						// determine dimensions before pfs connectivity update
						int nV_before = mesh.getNumberOfVertices();
						int nF_before = mesh.getNumberOfFaces();
						int nE_before = mesh.getNumberOfEdges();
						int nC_before = mesh.getNumberOfCorners();
						
						// save size of _coordIndex
						// needed later to update per corner properties
						int i0 = _coordIndex.size();
						
						// apply the connectivity forest split operation in place
						// properties remain unchanged
						// after the mesh is updated if v_cluster is applied to
						// the resulting mesh we obtain the original mesh
						vector<int>         v_cluster;
						vector<MPG_pVecInt> vp_vcf_loop;
						success = mesh.decodeForestSplit(is_forest_edge,
							vp_coordIndex,
							v_cluster,
							vp_vcf_loop);
						
						// vp_coordIndex is no longer needed
						{
							int i;
							for(i=0;i<vp_coordIndex.size();i++)
              { delete vp_coordIndex[i]; vp_coordIndex[i] = NULL; }
							vp_coordIndex.erase(vp_coordIndex.begin(),vp_coordIndex.end());
						}
						
						// determine dimensions after pfs connectivity update
						int nV_after = mesh.getNumberOfVertices();
						int nF_after = mesh.getNumberOfFaces();
						int nE_after = mesh.getNumberOfEdges();
						int nC_after = mesh.getNumberOfCorners();
						
						// quantized per vertex property array dimensions
						// do not match mesh sizes here;
						// duplicate per new vertex coordinates and
						
						{
							int i,j,k;
							if(bCoord==PB_PER_VERTEX)
								for(i=nV_before;i<nV_after;i++)
                {
                  j = v_cluster[i];
                  for(k=0;k<3;k++)
									{
										float qCjk = qCoord[3*j+k];
										qCoord.push_back(qCjk);
									}
                }
								
								if(bNormal==PB_PER_VERTEX)
									for(i=nV_before;i<nV_after;i++)
									{
										j = v_cluster[i];
										for(k=0;k<3;k++)
											qNormal.push_back(qNormal[3*j+k]);
									}
									
									if(bColor==PB_PER_VERTEX)
										for(i=nV_before;i<nV_after;i++)
										{
											j = v_cluster[i];
											for(k=0;k<3;k++)
												qColor.push_back(qColor[3*j+k]);
										}
										
										if(bTexCoord==PB_PER_VERTEX)
											for(i=nV_before;i<nV_after;i++)
											{
												j = v_cluster[i];
												for(k=0;k<2;k++)
													qTexCoord.push_back(qTexCoord[2*j+k]);
											}
											
						}          // we should have v_cluster.size()==nV_after here
						
						// unquantize and transfer new values
						if(qCoord.size()>_coord.size())
            {
              int  n = _coord.size();
              int dn = qCoord.size()-_coord.size();
              _coord.insert(_coord.end(),(size_t)dn,0.0);
							
							
              coordQuantizer->unQuantize(qCoord.begin()+n,
								_coord.begin()+n,
								dn/3);
            }
						
						if(qNormal.size()>_normal.size())
            {
              int  n = _normal.size();
              int dn = qNormal.size()-_normal.size();
              _normal.insert(_normal.end(),(size_t)dn,0.0);
              normalQuantizer->unQuantize(qNormal.begin()+n,
								_normal.begin()+n,
								dn);
            }
						
						if(qColor.size()>_color.size())
            {
              int  n = _color.size();
              int dn = qColor.size()-_color.size();
              _color.insert(_color.end(),(size_t)dn,0.0);
              colorQuantizer->unQuantize(qColor.begin()+n,
								_color.begin()+n,
								dn/3);
            }
						
						if(qTexCoord.size()>_texCoord.size())
            {
              int  n = _texCoord.size();
              int dn = qTexCoord.size()-_texCoord.size();
              _texCoord.insert(_texCoord.end(),(size_t)dn,0.0);
              texCoordQuantizer->unQuantize(qTexCoord.begin()+n,
								_texCoord.begin()+n,
								dn/2);
            }
						
						// fix property index arrays for per corner properties
						if(pNormal.getBinding()==PB_PER_CORNER)
            {
              // adjust _normalIndex
              // normal.size() should be equal to
              // the number of new corners
              int i = i0;
              int j = _normal.size();
              for(;i<_coordIndex.size();i++)
                if(_coordIndex[i]>=0)
								{ _normalIndex.push_back(j); j++; }
                else
								{ _normalIndex.push_back(-1); }
            }
						if(pColor.getBinding()==PB_PER_CORNER)
            {
              // adjust _colorIndex
              // color.size() should be equal to
              // the number of new corners
              int i = i0;
              int j = _color.size();
              for(;i<_coordIndex.size();i++)
                if(_coordIndex[i]>=0)
								{ _colorIndex.push_back(j); j++; }
                else
								{ _colorIndex.push_back(-1); }
            }
						
						if(pTexCoord.getBinding()==PB_PER_CORNER)
            {
              // adjust _texCoordIndex
              // texCoord.size() should be equal to
              // the number of new corners
              int i = i0;
              int j = _texCoord.size();
							
              for(;i<_coordIndex.size();i++)
                if(_coordIndex[i]>=0)
								{ _texCoordIndex.push_back(j); j++; }
                else
								{ _texCoordIndex.push_back(-1); }
            }
						
						// allocate arrays of sharp edge and fixed vertex marks
						vector<bool> edge_is_sharp(nE_after,false);
						vector<bool> vertex_is_fixed(nV_after,false);
						
						// decode if necessary
						if(pre_smoothing==true || post_smoothing==true)
            {
              // smoothing_constraints() 
              //   sharp_edges
              //   if(sharp_edges)
              //     edge_marks()
              //   fixed_vertices
              //   if(fixed_vertices)
              //     vertex_marks()
              
              if(codec.decode(MPG_Global::pfs_sharp_edges) == 1)
							{
								int i;
								for(i=0;i<nE_after;i++)
									edge_is_sharp[i] =
									(codec.decode(MPG_Global::pfs_sharp_edges)==1)?true:false;
							}
              
              if(codec.decode(MPG_Global::pfs_fixed_vertices) == 1)
							{
								int i;
								for(i=0;i<nV_after;i++)
									vertex_is_fixed[i] =
									(codec.decode(MPG_Global::pfs_fixed_vertices))?true:false;
							}
            }
						
						// if(pre_smoothing)
						//   apply pre_smoothing as a predictor
						//   to vertex coordinates and requantize
						if(pre_smoothing==true)
            {
              if(mesh.smooth(_coord,
								_coord,
								pre_smoothing_lambda,
								pre_smoothing_mu,
								pre_smoothing_n,
								&edge_is_sharp,
								&vertex_is_fixed)==false) return false;
              if(coordQuantizer->quantize(_coord,qCoord)==false)
                return false;
            }
						
						// decode loop vertex displacements,
						// loop face displacements, and loop corner displacements
						// for faces and corners of the original mesh
						// incident to the loops;
						// the indices of these faces and corners have been
						// collected in MPG_MeshFs::decodeForestSplit()
						
						// if(connectivity_update==FS_UPDATE)
						//   fs_post_update()
						if(connectivity_update==1)
            {
              // fs_post_update() 
              //   for each connected component 
              //     for each tree in forest 
              //       tree_loop_property_update()
							
              // remember that these are global quantized property arrays here
              // qCoord
              // qNormal
              // qColor
              // qTexCoord
							
              bool coord_updated    = false;
              bool normal_updated   = false;
              bool color_updated    = false;
              bool texCoord_updated = false;
							
              int itree = 0;
              int icc;
              for(icc=0;icc<nCC;icc++)
							{
								// for each connected component
								int nTreesCC = nTrees[icc];
								
								int iTreesCC;
								for(iTreesCC=0;iTreesCC<nTreesCC;iTreesCC++,itree++)
								{
									//   for each tree in forest
									
									// tree_loop_update() 
									//   loop_coord_update()
									//   loop_normal_update()
									//   loop_color_update()
									//   loop_texCoord_update()
									
									// arrays of displacement indices
									// v_loop and vcf_loop have been
									// collected in MPG_MeshFs::decodeForestSplit()
									vector<int>& vcf_loop = *(vp_vcf_loop[itree]);
									vector<int>  v_loop;
									vector<int>  f_loop;
									vector<int>  c_loop;
									{
										int i;
										int v  = -1;
										int c  = -1;
										int f  = -1;
										int cf = -1;
										for(i=0;3*i+2<vcf_loop.size();i++)
										{
											int vi  = vcf_loop[3*i  ];
											int ci  = vcf_loop[3*i+1];
											int fi  = vcf_loop[3*i+2];
											int cfi = mesh.getFaceCornerIndex(fi,ci);
											if(vi != v) { v = vi; v_loop.push_back(v); }
											if(fi != f) { f = fi; f_loop.push_back(f); }
											if(cfi!=cf) { cf = cfi; c_loop.push_back(cf); }
										}
										if(v_loop.size()>0 && v_loop.back()==v_loop.front())
											v_loop.pop_back();
										if(f_loop.size()>0 && f_loop.back()==f_loop.front())
											f_loop.pop_back();
										if(c_loop.size()>0 && c_loop.back()==c_loop.front())
											c_loop.pop_back();
									}
									
									
									// loop_coord_update()
									//   for each tree loop vertex 
									//     coord_sample()
									
									if(bCoord!=PB_NONE)
									{
										vector<int>& loop = v_loop;
										int i,j,k,qj[3];
										
										for(j=0;j<loop.size();j++)
										{
											k = loop[j];
											MPG_Sample::
												decodeCoordSample(codec,bitsPerCoord,qj);
											for(i=0;i<3;i++)
												qCoord[3*k+i] += qj[i];
											
										}
										coord_updated = true;
									}
									
									// loop_normal_update() 
									//   if(normal_binding==PER_VERTEX) 
									//     for each tree loop vertex 
									//       normal_sample()
									//   else if(normal_binding==PER_FACE) 
									//     for each tree loop face 
									//       normal_sample()
									//   else if(normal_binding==PER_CORNER) 
									//     for each tree loop corner 
									//       normal_sample()
									
									if(bNormal!=PB_NONE)
									{
										vector<int>& loop =
											(bNormal==PB_PER_VERTEX)?v_loop:
										(bNormal==PB_PER_FACE)?  f_loop:c_loop;
										
										int j,k,qj;
										for(j=0;j<loop.size();j++)
										{
											k = loop[j];
											MPG_Sample::
												decodeNormalSample(codec,bitsPerNormal,&qj);
											qNormal[k] += qj; // ??? 
										}
										
										normal_updated = true;
									}
									
									// loop_color_update() 
									//   if(color_binding==PER_VERTEX) 
									//     for each tree loop vertex 
									//       color_sample()
									//   else if(color_binding==PER_FACE) 
									//     for each tree loop face 
									//       color_sample()
									//   else if(color_binding==PER_CORNER) 
									//     for each tree loop corner 
									//       color_sample()
									
									if(bColor!=PB_NONE)
									{
										vector<int>& loop =
											(bColor==PB_PER_VERTEX)?v_loop:
										(bColor==PB_PER_FACE)?  f_loop:c_loop;
										
										int i,j,k,qj[3];
										for(j=0;j<loop.size();j++)
										{
											k = loop[j];
											MPG_Sample::
												decodeColorSample(codec,bitsPerColor,qj);
											for(i=0;i<3;i++)
												qColor[3*k+i] += qj[i];
										}
										color_updated = true;
									}
									
									// loop_texCoord_update() 
									//   if(texCoord_binding==PER_VERTEX) 
									//     for each tree loop vertex 
									//       texCoord_sample()
									//   else if(texCoord_binding==PER_CORNER) 
									//     for each tree loop corner 
									//       texCoord_sample()
									
									if(bTexCoord!=PB_NONE)
									{
										vector<int>& loop =
											(bTexCoord==PB_PER_VERTEX)?v_loop:c_loop;
										
										int i,j,k,qj[2];
										for(j=0;j<loop.size();j++)
										{
											k = loop[j];
											MPG_Sample::
												decodeTexCoordSample(codec,bitsPerTexCoord,qj);
											for(i=0;i<2;i++)
												qTexCoord[2*k+i] += qj[i];
										}
										texCoord_updated = true;
									}
                    }
                }
								
            }
						
						// if(connectivity_update!=NOT_UPDATED)
						//   other_update
						int other_update = 0;
						if(connectivity_update!=0)
							other_update = (codec.decode()==false)?0:1;
						
						// if(connectivity_update==NOT_UPDATED || other_update==true)
						//   other_property_update()
						
						//   decode property updates for all property values not
						//   updated already
						//   note that all the properties can be updated here if
						//   the forest is empty, i.e., the there is no connectivity
						//   update
						
            // if(connectivity_update==NOT_UPDATED || other_update==true)
            //   other_property_update()
            if(connectivity_update==0 || other_update==1)
						{
							// other_property_update() 
							//   other_coord_update()
							//   other_normal_update()
							//   other_color_update()
							//   other_texCoord_update()
							
							// other_coord_update() 
							//   for each vertex in mesh 
							//     if(not a tree loop vertex)
							//       coord_sample()
							
							// other_normal_update() 
							//   if(normal_binding==PER_VERTEX) 
							//     for each vertex in mesh 
							//       if(not a tree loop vertex)
							//         normal_sample()
							//   else if(normal_binding==PER_FACE) 
							//     for each face in mesh 
							//       if(not a tree loop face)
							//         normal_sample()
							//   else if(normal_binding==PER_CORNER) 
							//     for each corner in mesh 
							//       if(not a tree loop corner)
							//         normal_sample()
							
							// other_color_update() 
							//   if(color_binding==PER_VERTEX) 
							//     for each vertex in mesh 
							//       if(not a tree loop vertex)
							//         color_sample()
							//   else if(color_binding==PER_FACE) 
							//     for each face in mesh 
							//       if(not a tree loop face)
							//         color_sample()
							//   else if(color_binding==PER_CORNER) 
							//     for each corner in mesh 
							//       if(not a tree loop corner)
							//         color_sample()
							
							// other_texCoord_update() 
							//   if(texCoord_binding==PER_VERTEX) 
							//     for each vertex in mesh 
							//       if(not a tree loop vertex)
							//         texCoord_sample()
							//   else if(texCoord_binding==PER_CORNER) 
							//     for each corner in mesh 
							//       if(not a tree loop corner)
							//         texCoord_sample()
						}
						
						// unQuantized updated properties
						// (should unQuantized ONLY updated properties
						//  but we doing all !)
						
						if(qCoord.size()>0 && 
							coordQuantizer->unQuantize(qCoord,_coord)==false)
							return false;
						
						if(qNormal.size()>0 && 
							normalQuantizer->unQuantize(qNormal,_normal)==false)
							return false;
						
						if(qColor.size()>0 && 
							colorQuantizer->unQuantize(qColor,_color)==false)
							return false;
						
						if(qTexCoord.size()>0 && 
							texCoordQuantizer->unQuantize(qTexCoord,_texCoord)==false)
							return false;
						
						// if(post_smoothing)
						//   apply post_smoothing
						//   to vertex coordinates and requantize
						
						if(post_smoothing==true)
            {
              if(mesh.smooth(_coord,
								_coord,
								post_smoothing_lambda,
								post_smoothing_mu,
								post_smoothing_n,
								&edge_is_sharp,
								&vertex_is_fixed)==false) return false;
              if(coordQuantizer->quantize(_coord,qCoord)==false)
                return false;
            }
						
						// stop_qcoder()
						codec.finish_decode();
						
						// advance to next byte boundary and read ahead to determine
						// if anoter 3D_Mesh_Object_Refinement_Layer() folllows
						
        } while(istrm.lookingAtStartCode(MPG_Global::MORL_start_code)==true);
				
        // restore original property bindings
        pCoord.setBinding(bCoord);
        pNormal.setBinding(bNormal);
        pColor.setBinding(bColor);
        pTexCoord.setBinding(bTexCoord);
        
      } // else if(mol_id!=0)
			
			// advance to next byte boundary and read ahead to determine
			// if anoter 3D_Mesh_Object_Layer() folllows
			
#ifndef MPGELIB
			if(saveWrlLevels==true)
      {
        int n = strlen(outDir)+strlen(baseFname)+16;
        outFname = new char[n];
        if(!strcmp(outDir,"."))
          sprintf(outFname,"%s-E%02d.wrl",baseFname,mol_id);
        else
          sprintf(outFname,"%s/%s-E%02d.wrl",outDir,baseFname,mol_id);
        bool success = write(outFname);
        delete [] outFname;
        outFname = NULL;
      }
#endif // MPGELIB    
  } while(istrm.lookingAtStartCode(MPG_Global::MOL_start_code)==true);
  
#ifdef ADD_VISUAL_HEADER
  if(istrm.getVisualStartCode(visual_object_sequence_end_code)==false)
		return false;
#endif //ADD_VISUAL_HEADER

  qCoord.clear();
  qNormal.clear();
  qColor.clear();
  qTexCoord.clear();
  
  success = true;
  return success;
}

#ifdef _SINGLE_RESOLUTION_PROGRESSIVE
// these are methods to be used for progressive decoding of a 
// single resolution .m3d stream
// the decoder resolution can be set to a number n of triangles at a time,
// one connected component at a time, or a two-step decoding process (header + rest)

//------------------------------------------------------
// expandHeader
//------------------------------------------------------
bool MPG_IfsQ_E::expandHeader(MPG_InputStream& istrm, MPG_State& state, 
                              MPG_Property& pCoord, MPG_Property& pNormal,
                              MPG_Property& pColor, MPG_Property& pTexCoord,
                              vector<int>& qCoord, vector<int>& qNormal,
                              vector<int>& qColor, vector<int>& qTexCoord,
                              vector<int>& v_visited, vector<int>& v_cluster,
                              MPG_Stitcher& stitcher, vector<int>& v_father,
                              MPG_Stitches& v_stitches, vector<int>& stk)
{
  MPG_QFpm::_shift = 16;
	
  bool success = false;
	
  clear();

  int nV0, nV1, nVcls0, nVcls1;
  state.GetNVertices(nV0, nV1, nVcls0, nVcls1);
	
  // arrays to store quantized property values
	
  // AG: in 3D_Mesh_Object_Base_Layer
  // these arrays are global for per vertex properties
  // (containing all the values decoded so far),
  // and local otherwise
  // (containing only values corresponding to the current component)
	
  // GT: in 3D_Mesh_Object_Forest_Split
  // all these arrays are global
	
  //vector <int> qCoord;
  //vector <int> qNormal;
  //vector <int> qColor;
  //vector <int> qTexCoord;
	
  // create proper headers and quantizers
  // if not yet created by calling method
  //MPG_Property pCoord(PT_COORD,PB_NONE);
  MPG_BBoxQ* coordQuantizer = getCoordQuantizer();
  if((MPG_BBoxQ*)0==coordQuantizer)
    {
      coordQuantizer    = new MPG_BBoxQ(3);
      setCoordQuantizer(coordQuantizer);
    }
  pCoord.setBBoxQ(coordQuantizer);
	
  //MPG_Property pNormal(PT_NORMAL,PB_NONE);
  MPG_NormalQ* normalQuantizer = getNormalQuantizer();
  if((MPG_NormalQ*)0==normalQuantizer)
    {
      normalQuantizer   = new MPG_NormalQ();
      setNormalQuantizer(normalQuantizer);
    }
  pNormal.setNormalQ(normalQuantizer);
	
  //MPG_Property pColor(PT_COLOR,PB_NONE);
  MPG_BBoxQ* colorQuantizer = getColorQuantizer();
  if((MPG_BBoxQ*)0==colorQuantizer)
    {
      colorQuantizer    = new MPG_BBoxQ(3);
      setColorQuantizer(colorQuantizer);
    }
  pColor.setBBoxQ(colorQuantizer);
	
  //MPG_Property pTexCoord(PT_TEXCOORD,PB_NONE);
  MPG_BBoxQ* texCoordQuantizer = getTexCoordQuantizer();
  if((MPG_BBoxQ*)0==texCoordQuantizer)
    {
      texCoordQuantizer = new MPG_BBoxQ(2);
      setTexCoordQuantizer(texCoordQuantizer);
    }
  pTexCoord.setBBoxQ(texCoordQuantizer);
	
  // quantization parameters
  int bitsPerCoord    = 0;
  int bitsPerNormal   = 0;
  int bitsPerColor    = 0;
  int bitsPerTexCoord = 0;
  
  // input bitstream parsing --------------------------------------------------
	
  //MPG_Stitcher stitcher;
  //vector<int>  v_father;
  //MPG_Stitches v_stitches;
  //vector<int>  stk;  // stitch encoding stack
	
  //vector<int>  v_visited;
  //vector<int>  v_cluster;
  //int          nV0    = 0; // number of vertices so far without current cc
  //int          nV1    = 0; // number of vertices so far with current cc
  //int          nVcls0 = 0; // AG: number of vertices after clustering
  //int          nVcls1 = 0; // AG: number of vertices after clustering
	
  // decode 3D_Mesh_Object() {
  //
  // 3D_Mesh_Object() {
  //   3D_MO_start_code
  //   3D_Mesh_Object_Header()
  //   do {
  //     3D_Mesh_Object_Layer()
  //   } while(nextbits_bytealigned()==3D_MOL_start_code)
  //
	
  // read 3D_MO_start_code
  if(istrm.getStartCode(MPG_Global::MO_start_code)==false)
    return false;
	
  // decode 3D_Mesh_Object_Header()
  {
    // open a bit stream
    MPG_InputBitStream ibstrm(istrm);
		
    // 3D_Mesh_Object_Header() {
    //   ccw
    //   convex
    //   solid
    //   creaseAngle
    //   coord_header()
    //   normal_header()
    //   color_header()
    //   texCoord_header()
    //   cgd_data
    //   if(cgd_data==1)
    //     cgd_header()
    // }
		
    if(ibstrm.getBit(_ccw               )==false) return false;
    if(ibstrm.getBit(_convex            )==false) return false;
    if(ibstrm.getBit(_solid             )==false) return false;
		
    int cAng = 0;
    if(ibstrm.getMBitInt(6,cAng)==false) return false;
    if(cAng<0 || cAng>=64) return false;
    _creaseAngle = M_PI*((float)cAng)/32;
		
    // decode coord_header()
    {
      // coord_header() {
      //   coord_binding       <--- must be == 1 (PER_VERTEX)
			
      int b = 0;
      if(ibstrm.getMBitInt(2,b)==false) return false;
      MPG_Binding binding =
        (b==0)?PB_NONE:(b==1)?PB_PER_VERTEX:(b==2)?PB_PER_FACE:PB_PER_CORNER;
      if(binding!=PB_NONE)
        {
				// coord can only be bound PER_VERTEX
          if(binding!=PB_PER_VERTEX) return false;
          pCoord.setBinding(binding);
				
				//   coord_bbox
				//   if(coord_bbox) {
				//     coord_xmin
				//     coord_ymin
				//     coord_zmin
				//     coord_size
				//   }
				//   coord_quant
				//   coord_pred_type
				//   if(coord_pred_type==2) {
				//     coord_nlambda
				//     for(i=1;i<coord_nlambda;i++)
				//       coord_lambda[i]
				//   }
				
          success = pCoord.decode(ibstrm);
          if(success==false) return false;
        }
			
      // }
    }
		
    // decode normal_header()
    {
      // normal_header() {
      //   normal_binding
      //   if(normal_binding!=0) {
			
      int b = 0;
      if(ibstrm.getMBitInt(2,b)==false) return false;
      MPG_Binding binding =
        (b==0)?PB_NONE:(b==1)?PB_PER_VERTEX:(b==2)?PB_PER_FACE:PB_PER_CORNER;
      if(binding!=PB_NONE)
        {
          pNormal.setBinding(binding);
				
				//     normal_bbox       <--- must be == 0
				//     normal_quant
				//     normal_pred_type
				//     if(normal_pred_type==2) {
				//       normal_nlambda
				//       for(i=1;i<normal_nlambda;i++)
				//         normal_lambda[i]
				//     }
				
          success = pNormal.decode(ibstrm);
          if(success==false) return false;
				
        }
			
      //   }
      // }
			
      _normalPerVertex = (binding==PB_PER_FACE)?false:true;
    }
		
    // decode color_header()
    {
      // color_header() {
      //   color_binding
      //   if(color_binding!=0) {
			
      int b = 0;
      if(ibstrm.getMBitInt(2,b)==false) return false;
      MPG_Binding binding =
        (b==0)?PB_NONE:(b==1)?PB_PER_VERTEX:(b==2)?PB_PER_FACE:PB_PER_CORNER;
      if(binding!=PB_NONE)
        {
          pColor.setBinding(binding);
				
				//     color_bbox
				//     if(color_bbox) {
				//       color_rmin
				//       color_gmin
				//       color_bmin
				//       color_size
				//     }
				//     color_quant
				//     color_pred_type
				//     if(color_pred_type==2) {
				//       color_nlambda
				//       for(i=1;i<color_nlambda;i++)
				//         color_lambda[i]
				//     }
				
          success = pColor.decode(ibstrm);
          if(success==false) return false;
        }
			
      //   }
      // }
			
      _colorPerVertex = (binding==PB_PER_FACE)?false:true;
    }
		
    // decode texture_header()
    {
      // texCoord_header() {
      //   texCoord_binding     <--- must be != 2 (PER_FACE)
      //   if(texCoord_binding!=0) {
			
      int b = 0;
      if(ibstrm.getMBitInt(2,b)==false) return false;
      MPG_Binding binding =
        (b==0)?PB_NONE:(b==1)?PB_PER_VERTEX:(b==2)?PB_PER_FACE:PB_PER_CORNER;
      if(binding!=PB_NONE)
        {
				// TexCoord cannot be bound per_face
          if(binding==PB_PER_FACE) return false;
          pTexCoord.setBinding(binding);
				
				//     texCoord_bbox
				//     if(texCoord_bbox) {
				//       texCoord_umin
				//       texCoord_vmin
				//       texCoord_size
				//     }
				//     texCoord_quant
				//     texCoord_pred_type
				//     if(texCoord_pred_type==2) {
				//       texCoord_nlambda
				//       for(i=1;i<texCoord_nlambda;i++)
				//         texCoord_lambda[i]
				//     }
				
          success = pTexCoord.decode(ibstrm);
          if(success==false) return false;
        }
			
      //   }
      // }
    }
		
    // decode cgd_data
    {
      _cgd_data = false;
      if(ibstrm.getBit(_cgd_data)==false) return false;
			
      if(_cgd_data==true)
        {
				// decode cgd_header()
				//
				// for now : just parse and ignore values
				// later   : create a new class to store values
				//
          int N_Proj_Surface_Spheres = 0;
          if(ibstrm.getMBitInt(4,N_Proj_Surface_Spheres)==false) return false;
          if(N_Proj_Surface_Spheres>0)
            {
              float x_coord_Center_Point = 0.0;
              if(ibstrm.get32BitFloat(x_coord_Center_Point)==false)
                return false;
					
              float y_coord_Center_Point = 0.0;
              if(ibstrm.get32BitFloat(y_coord_Center_Point)==false)
                return false;
					
              float z_coord_Center_Point = 0.0;
              if(ibstrm.get32BitFloat(z_coord_Center_Point)==false)
                return false;
					
              int Normalized_Screen_Distance_Factor = 0;
              if(ibstrm.getMBitInt(8,Normalized_Screen_Distance_Factor)==false)
                return false;
					
              int i;
              for(i=0;i<N_Proj_Surface_Spheres;i++) {
						
                float Radius = 0.0;
                if(ibstrm.get32BitFloat(Radius)==false)
                  return false;
						
                float Min_Proj_Surface = 0.0;
                if(ibstrm.get32BitFloat(Min_Proj_Surface)==false)
                  return false;
						
                int N_Proj_Points = 0;
                if(ibstrm.getMBitInt(8,N_Proj_Points)==false)
                  return false;
						
                int j;
                for(j=0;j<N_Proj_Points;j++)
                  {
							
                    int Sphere_Point_Coord = 0;
                    if(ibstrm.getMBitInt(11,Sphere_Point_Coord)==false)
                      return false;
							
                    float Proj_Surface = 0.0;
                    if(ibstrm.get32BitFloat(Proj_Surface)==false)
                      return false;
							
                  }
              }
            }
        }
    }
		
    // flush input buffer
    ibstrm.flush();
		
  } // end decode 3D_Mesh_Object_Header()

  state.SetNVertices(nV0, nV1, nVcls0, nVcls1);
  success = true;
  return success;
}

//------------------------------------------------------
// expandRest
//------------------------------------------------------
bool MPG_IfsQ_E::expandRest(MPG_InputStream& istrm, 
                            MPG_Property& pCoord, MPG_Property& pNormal,
                            MPG_Property& pColor, MPG_Property& pTexCoord,
                            vector<int>& qCoord, vector<int>& qNormal,
                            vector<int>& qColor, vector<int>& qTexCoord,
                            vector<int>& v_visited, vector<int>& v_cluster, 
                            MPG_Stitcher& stitcher, vector<int>& v_father,
                            MPG_Stitches& v_stitches, vector<int>& stk,
                            int& nV0, int& nV1, int& nVcls0, int& nVcls1)
{
  int bitsPerCoord    =  pCoord.getBBoxQ()->getBitsPerCoordinate();
  int bitsPerNormal   =  pNormal.getNormalQ()->getBitsPerNormal();
  int bitsPerColor    =  pColor.getBBoxQ()->getBitsPerCoordinate();
  int bitsPerTexCoord =  pTexCoord.getBBoxQ()->getBitsPerCoordinate();
	
  bool success = false;

  // decode one or more 3D_Mesh_Object_Layer()
  int mol_id = 0;
  do {
    // decode 3D_Mesh_Object_Layer()
		
    // 3D_Mesh_Object_Layer() {
    //   3D_MOL_start_code
    //   mol_id
    //   if(cgd_data==1) {
    //     N_Vertices
    //     N_Triangles
    //     N_Edges
    //   }
    //   if(mol_id==0)
    //     3D_Mesh_Object_Base_Layer()
    //   else
    //     3D_Mesh_Object_Refinement_Layer()
    // }
		
    // read 3D_MOL_start_code
    if(istrm.getStartCode(MPG_Global::MOL_start_code)==false)
      return false;
		
    {
      // open a bit stream
      MPG_InputBitStream ibstrm(istrm);
			
      // read mol_id
      if(ibstrm.getMBitInt(8,mol_id)==false) return false;
			
      // read cgd data
      if(_cgd_data==true)
        {
          int N_Vertices = 0;
          if(ibstrm.getMBitInt(24,N_Vertices)==false) return false;
				
          int N_Triangles = 0;
          if(ibstrm.getMBitInt(24,N_Triangles)==false) return false;
				
          int N_Edges = 0;
          if(ibstrm.getMBitInt(24,N_Edges)==false) return false;
        }
			
      ibstrm.flush();
    }
		
    if(mol_id==0)
      {
        //
        // decode 3D_Mesh_Object_Base_Layer()
        //
			
        // GT Wed Jun 09 21:13:23 1999 TODO:
        // codap support ?
			
        // 3D_Mesh_Object_Base_Layer() {
        //   do {
        //     3D_MOBL_start_code
        //     mobl_id
        //     start_qcoder()
        //     do {
        //       connected_component()
        //       last_component
        //     } while(last_component==false)
        //     stop_qcoder()
        //   } while(nextbits_bytealigned()==3D_MOBL_start_code)
        // }
			
        do {
				
				// read 3D_MOBL_start_code
          if(istrm.getStartCode(MPG_Global::MOBL_start_code)==false)
            return false;
				
          {
            // open a bit stream
            MPG_InputBitStream ibstrm(istrm);     
					
            // read mobl_id
            int mobl_id = 0;
            if(ibstrm.getMBitInt(8,mobl_id)==false) return false;
					
            ibstrm.flush();
          }
				
				//  start the arithmetic codec
          MPG_QFac codec(istrm);
          codec.start_decode();
          MPG_Global::reset();
				
          do {
            // decode connected_component()
					
            // connected_component() {
            //   vertex_graph()
            //   has_stitches
            //   if(has_stitches==1)
            //     stitches()l
            //   triangle_tree()
            //   triangle_data()
            // }
					
            // decode vertex_graph() and construct vertex loop
            int nVcc = 0;
            vector<int> v_loop_cc;
            {
              // decode vertex_graph()
              MPG_VGraph vg;
              if(vg.decode(codec) == false) return false;
						
              nVcc = vg.getNumberOfVertices();
              nV1 = nV0 + nVcc;
              while(v_visited.size()<nV1)
                v_visited.push_back(false);
						
              // convert to vertex loop look-up table
              MPG_Loops loop;
              vg.toLoop(loop);
              loop.getLoop(0, v_loop_cc);
						
              if(vg.getFather(v_father)==false || v_father.size()!=nV1)
                {
                  return false;
                }
            }
					
            v_stitches.decode(nV0,nV1,stk,codec);
            {
              // use stitches nV0<=i<nV1 to partially recover v_cluster
						
              if(stitcher.use(nV0,v_father,v_stitches,v_cluster)==false)
                return false;
            }
					
            // we should have here :
            // v_cluster.size()==nV1
					
            if(v_cluster.size()!=nV1)
                return false;
					
            // for i=0,..,nV1-1
            //   v_cluster[i]>=0 && v_cluster[i]<nV1
            // v_father.size()==nV1 
            // for i=0,..,nV1-1
            //   v_father[i]<=i
            // v_stitches.size()==nV1
					
            // decode triangle_tree()
            MPG_SPolygon sp;
            if(sp.decodeTriangleTree(codec) == false) return false;
					
            // decode triangle_data()
            vector<int>   coordIndex;
            if(sp.decodeTriangleData(codec,
                                     nVcc,
                                     nV0,
                                     v_loop_cc,
                                     v_visited,
                                     v_cluster,
                                     coordIndex,
                                     // property headers and quantizers
                                     pCoord,
                                     pNormal,
                                     pColor,
                                     pTexCoord,
                                     // quantized values
                                     qCoord,
                                     qNormal,
                                     qColor,
                                     qTexCoord)==false) return false;
					
            // save size of _coordIndex
            // used to update per corner properties
            int i0 = _coordIndex.size();
					
            // transfer coordIndex to Ifs
            {
              int j0,j1,j;
              for(j0=j1=0;j1<coordIndex.size();j1++)
                if(coordIndex[j1]<0)
                  {
                    int* f = &(coordIndex[j0]);
                    int nf = j1-j0;
                    for(j=0;j<nf;j++)
                      _coordIndex.push_back(v_cluster[nV0+v_loop_cc[f[j]]]);
                    _coordIndex.push_back(-1);
                    j0 = j1+1;
                  }
            }
					
            // 2/12/99 determine the new number of vertices after clustering
            {
              int i;
              for (i=nV0; i < nV1; i++)
                if (v_cluster[i] > nVcls1) nVcls1 = v_cluster[i];
            }
            int nVcls11 = nVcls1+1;
					
            // unquantize vertices between nVcls0 and nVcls1
					
            // coord properties
            vector<float> coord((size_t) (3*(nVcls11-nVcls0)),0.0);
            MPG_BBoxQ*   coordQuantizer    = pCoord.getBBoxQ();
            coordQuantizer->unQuantize(qCoord.begin()+3*nVcls0,
                                       coord.begin(),
                                       nVcls11-nVcls0);
					
            //  normal properties
            vector<float> normal;
            MPG_NormalQ* normalQuantizer   = pNormal.getNormalQ();
            if(pNormal.getBinding() != PB_NONE)
              {
                // special case for per-vertex binding,
                // the qNormal coordinates are global to allow reuse
                // of some vertices by components decoded afterwards
                if(pNormal.getBinding() == PB_PER_VERTEX)
                  {
                    // take only the trailing part of qNormal
                    vector <int> qNormal1;
                    int i;
                    for (i=nVcls0;i<nVcls11;i++)
                      qNormal1.push_back(qNormal[i]);
							
                    normalQuantizer->unQuantize(qNormal1, normal);
                  }
                else normalQuantizer->unQuantize(qNormal, normal);
              }
            //  color properties
            vector<float> color;
            MPG_BBoxQ*   colorQuantizer    = pColor.getBBoxQ();
            if(pColor.getBinding() != PB_NONE)
              {
                if(pColor.getBinding() == PB_PER_VERTEX)
                  {
                    color.erase(color.begin(),color.end());
                    color.insert(color.end(),(size_t)(3*(nVcls11-nVcls0)),0.0);
                    colorQuantizer->unQuantize(qColor.begin()+3*nVcls0,
                                               color.begin(),
                                               nVcls11-nVcls0);
                  }
                else colorQuantizer->unQuantize(qColor, color);
              }
            //  texture properties
            vector<float> texCoord;
            MPG_BBoxQ*   texCoordQuantizer = pTexCoord.getBBoxQ();
            if(pTexCoord.getBinding() != PB_NONE)
              {
                if(pTexCoord.getBinding() == PB_PER_VERTEX)
                  {
                    texCoord.erase(texCoord.begin(),texCoord.end());
                    texCoord.insert(texCoord.end(),
                                    (size_t)(2*(nVcls11-nVcls0)),0.0);
                    texCoordQuantizer->unQuantize(qTexCoord.begin()+2*nVcls0,
                                                  texCoord.begin(),
                                                  nVcls11-nVcls0);
                  }
                else
                  texCoordQuantizer->unQuantize(qTexCoord, texCoord);
              }
            // transfer coord to Ifs
            {
              int i;
              for(i=0;i<coord.size();i++) _coord.push_back(coord[i]);
            }
					
            if(pNormal.getBinding()==PB_PER_CORNER)
              {
                // adjust _normalIndex
                // normal.size() should be equal to
                // the number of new corners
                int i = i0;
                int j = _normal.size()/3;
                for(;i<_coordIndex.size();i++)
                  if(_coordIndex[i]>=0)
                    { _normalIndex.push_back(j); j++; }
                  else
                    { _normalIndex.push_back(-1); }
              }
					
            // transfer normal to Ifs
            {
              int i;
              for(i=0;i<normal.size();i++)
                _normal.push_back(normal[i]);
            }
					
            if(pColor.getBinding()==PB_PER_CORNER)
              {
                // adjust _colorIndex
                // color.size() should be equal to
                // the number of new corners
                int i = i0;
                int j = _color.size()/3;
                for(;i<_coordIndex.size();i++)
                  if(_coordIndex[i]>=0)
                    { _colorIndex.push_back(j); j++; }
                  else
                    { _colorIndex.push_back(-1); }
              }
					
            // transfer color to Ifs
            {
              int i;
              for(i=0;i<color.size();i++)
                _color.push_back(color[i]);
            }
					
            if(pTexCoord.getBinding()==PB_PER_CORNER)
              {
                // adjust _texCoordIndex
                // texCoord.size() should be equal to
                // the number of new corners
                int i = i0;
                int j = _texCoord.size()/2;
                for(;i<_coordIndex.size();i++)
                  if(_coordIndex[i]>=0)
                    { _texCoordIndex.push_back(j); j++; }
                  else
                    { _texCoordIndex.push_back(-1); }
              }
					
            // transfer texCoord to Ifs
            {
              int i;
              for(i=0;i<texCoord.size();i++)
                _texCoord.push_back(texCoord[i]);
            }
					
            nVcls0 = nVcls11;
            nV0 = nV1;
					
          } while(codec.decode(MPG_Global::last_component) == false);
		  
          // stop_qcoder()
          codec.finish_decode();
          
          // advance to next byte boundary and read ahead to determine
          // if another 3D_Mesh_Object_Base_Layer() follows
		  
        } while(istrm.lookingAtStartCode(MPG_Global::MOBL_start_code)==true);
		
      }

    else // if(mol_id!=0)
      {
        //
        //
        // progressive version not implemented
        //
        //
      } // else if(mol_id!=0)
	  
    // advance to next byte boundary and read ahead to determine
    // if anoter 3D_Mesh_Object_Layer() folllows
	  
	  
    /*    if(saveWrlLevels==true)
          {
	  int n = strlen(outDir)+strlen(baseFname)+16;
	  outFname = new char[n];
	  if(!strcmp(outDir,"."))
	  sprintf(outFname,"%s-E%02d.wrl",baseFname,mol_id);
	  else
	  sprintf(outFname,"%s/%s-E%02d.wrl",outDir,baseFname,mol_id);
	  bool success = write(outFname);
	  delete [] outFname;
	  outFname = NULL;
          }
    */    
  } while(istrm.lookingAtStartCode(MPG_Global::MOL_start_code)==true);

  success = true;
  return success;
}

//------------------------------------------------------
// expandCC
//------------------------------------------------------
bool MPG_IfsQ_E::expandCC(MPG_InputStream& istrm, 
                          MPG_Property& pCoord, MPG_Property& pNormal,
                          MPG_Property& pColor, MPG_Property& pTexCoord,
                          vector<int>& qCoord, vector<int>& qNormal,
                          vector<int>& qColor, vector<int>& qTexCoord,
                          vector<int>& v_visited, vector<int>& v_cluster,
                          MPG_Stitcher& stitcher, vector<int>& v_father,
                          MPG_Stitches& v_stitches, vector<int>& stk,
                          int& nV0, int& nV1, int& nVcls0, int& nVcls1,
                          bool& newMol, bool& newMobl, bool& newCC,
                          int& mol_id, MPG_QFac& codec, 
				//vector<int>&   coordIndex,
				//vector<float>& coord,
				//vector<float>& normal,
				//vector<float>& color,
				//vector<float>& texCoord,
                          bool& expandDone)
{
  bool success = false;
  bool wasNewMobl = false;

  // decode one or more 3D_Mesh_Object_Layer()

  if(newMol)
    {
      // decode 3D_Mesh_Object_Layer()
		
      // 3D_Mesh_Object_Layer() {
      //   3D_MOL_start_code
      //   mol_id
      //   if(cgd_data==1) {
      //     N_Vertices
      //     N_Triangles
      //     N_Edges
      //   }
      //   if(mol_id==0)
      //     3D_Mesh_Object_Base_Layer()
      //   else
      //     3D_Mesh_Object_Refinement_Layer()
      // }
		
      // read 3D_MOL_start_code
      if(istrm.getStartCode(MPG_Global::MOL_start_code)==false)
        return false;
		
      mol_id = 0;
      {
        // open a bit stream
        MPG_InputBitStream ibstrm(istrm);
			
        // read mol_id
        if(ibstrm.getMBitInt(8,mol_id)==false) return false;
			
        // read cgd data
        if(_cgd_data==true)
          {
            int N_Vertices = 0;
            if(ibstrm.getMBitInt(24,N_Vertices)==false) return false;
				
            int N_Triangles = 0;
            if(ibstrm.getMBitInt(24,N_Triangles)==false) return false;
				
            int N_Edges = 0;
            if(ibstrm.getMBitInt(24,N_Edges)==false) return false;
          }
			
        ibstrm.flush();
      }
		
      newMol = false;
    }

  if(mol_id==0)
    {
      //
      // decode 3D_Mesh_Object_Base_Layer()
      //
		
      // GT Wed Jun 09 21:13:23 1999 TODO:
      // codap support ?
		
      // 3D_Mesh_Object_Base_Layer() {
      //   do {
      //     3D_MOBL_start_code
      //     mobl_id
      //     start_qcoder()
      //     do {
      //       connected_component()
      //       last_component
      //     } while(last_component==false)
      //     stop_qcoder()
      //   } while(nextbits_bytealigned()==3D_MOBL_start_code)
      // }
		
      if(newMobl) 
        {	
          // read 3D_MOBL_start_code
          if(istrm.getStartCode(MPG_Global::MOBL_start_code)==false)
            return false;
			
          {
				// open a bit stream
            MPG_InputBitStream ibstrm(istrm);     
				
				// read mobl_id
            int mobl_id = 0;
            if(ibstrm.getMBitInt(8,mobl_id)==false) return false;
				
            ibstrm.flush();
          }
			
          //  start the arithmetic codec
          // MPG_QFac codec(istrm);
          codec.start_decode();

          // reset probability model
          MPG_Global::reset();

          newMobl = false;
          wasNewMobl = true;
        }
	
      if(newCC)
        {
          // decode connected_component()
			
          // connected_component() {
          //   vertex_graph()
          //   has_stitches
          //   if(has_stitches==1)
          //     stitches()l
          //   triangle_tree()
          //   triangle_data()
          // }
			
          // decode vertex_graph() and construct vertex loop
          int nVcc = 0;
          vector<int> v_loop_cc;
          {
				// decode vertex_graph()
            MPG_VGraph vg;
            if(vg.decode(codec) == false) return false;
				
            nVcc = vg.getNumberOfVertices();
            nV1 = nV0 + nVcc;
            while(v_visited.size()<nV1)
              v_visited.push_back(false);
				
				// convert to vertex loop look-up table
            MPG_Loops loop;
            vg.toLoop(loop);
            loop.getLoop(0, v_loop_cc);
				
            if(vg.getFather(v_father)==false || v_father.size()!=nV1)
                return false;
          }
					
          v_stitches.decode(nV0,nV1,stk,codec);
          {
				// use stitches nV0<=i<nV1 to partially recover v_cluster
				
            if(stitcher.use(nV0,v_father,v_stitches,v_cluster)==false)
              return false;
				
          }
					
          // we should have here :
          // v_cluster.size()==nV1
			
          if(v_cluster.size()!=nV1)
              return false;
			
          // for i=0,..,nV1-1
          //   v_cluster[i]>=0 && v_cluster[i]<nV1
          // v_father.size()==nV1 
          // for i=0,..,nV1-1
          //   v_father[i]<=i
          // v_stitches.size()==nV1
			
          // decode triangle_tree()
          MPG_SPolygon sp;
          if(sp.decodeTriangleTree(codec) == false) return false;
					
          // decode triangle_data()
          vector<int>   coordIndex;
          if(sp.decodeTriangleData(codec,
                                   nVcc,
                                   nV0,
                                   v_loop_cc,
                                   v_visited,
                                   v_cluster,
                                   coordIndex,
				// property headers and quantizers
                                   pCoord,
                                   pNormal,
                                   pColor,
                                   pTexCoord,
				// quantized values
                                   qCoord,
                                   qNormal,
                                   qColor,
                                   qTexCoord)==false) return false;
					
				// save size of _coordIndex
				// used to update per corner properties
          int i0 = _coordIndex.size();
				
				// transfer coordIndex to Ifs
          {
            int j0,j1,j;
            for(j0=j1=0;j1<coordIndex.size();j1++)
              if(coordIndex[j1]<0)
                {
                  int* f = &(coordIndex[j0]);
                  int nf = j1-j0;
                  for(j=0;j<nf;j++)
                    _coordIndex.push_back(v_cluster[nV0+v_loop_cc[f[j]]]);
                  _coordIndex.push_back(-1);
                  j0 = j1+1;
                }
          }
					
			// 2/12/99 determine the new number of vertices after clustering
          {
            int i;
            for (i=nV0; i < nV1; i++)
              if (v_cluster[i] > nVcls1) nVcls1 = v_cluster[i];
          }
          int nVcls11 = nVcls1+1;
				// unquantize vertices between nVcls0 and nVcls1
				
				// coord properties
          vector<float> coord((size_t) (3*(nVcls11-nVcls0)),0.0);
          MPG_BBoxQ*   coordQuantizer    = pCoord.getBBoxQ();
          coordQuantizer->unQuantize(qCoord.begin()+3*nVcls0,
                                     coord.begin(),
                                     nVcls11-nVcls0);
				
				//  normal properties
          vector<float> normal;
          MPG_NormalQ* normalQuantizer   = pNormal.getNormalQ();
          if(pNormal.getBinding() != PB_NONE)
            {
              // special case for per-vertex binding,
              // the qNormal coordinates are global to allow reuse
              // of some vertices by components decoded afterwards
              if(pNormal.getBinding() == PB_PER_VERTEX)
                {
                  // take only the trailing part of qNormal
                  vector <int> qNormal1;
                  int i;
                  for (i=nVcls0;i<nVcls11;i++)
                    qNormal1.push_back(qNormal[i]);
						
                  normalQuantizer->unQuantize(qNormal1, normal);
                }
              else normalQuantizer->unQuantize(qNormal, normal);
            }
				//  color properties
          vector<float> color;
          MPG_BBoxQ*   colorQuantizer    = pColor.getBBoxQ();
          if(pColor.getBinding() != PB_NONE)
            {
              if(pColor.getBinding() == PB_PER_VERTEX)
                {
                  color.erase(color.begin(),color.end());
                  color.insert(color.end(),(size_t)(3*(nVcls11-nVcls0)),0.0);
                  colorQuantizer->unQuantize(qColor.begin()+3*nVcls0,
                                             color.begin(),
                                             nVcls11-nVcls0);
                }
              else colorQuantizer->unQuantize(qColor, color);
            }
				//  texture properties
          vector<float> texCoord;
          MPG_BBoxQ*   texCoordQuantizer = pTexCoord.getBBoxQ();
          if(pTexCoord.getBinding() != PB_NONE)
            {
              if(pTexCoord.getBinding() == PB_PER_VERTEX)
                {
                  texCoord.erase(texCoord.begin(),texCoord.end());
                  texCoord.insert(texCoord.end(),
                                  (size_t)(2*(nVcls11-nVcls0)),0.0);
                  texCoordQuantizer->unQuantize(qTexCoord.begin()+2*nVcls0,
                                                texCoord.begin(),
                                                nVcls11-nVcls0);
                }
              else
                texCoordQuantizer->unQuantize(qTexCoord, texCoord);
            }
				// transfer coord to Ifs
          {
            int i;
            for(i=0;i<coord.size();i++) _coord.push_back(coord[i]);
          }
				
          if(pNormal.getBinding()==PB_PER_CORNER)
            {
              // adjust _normalIndex
              // normal.size() should be equal to
              // the number of new corners
              int i = i0;
              int j = _normal.size()/3;
              for(;i<_coordIndex.size();i++)
                if(_coordIndex[i]>=0)
                  { _normalIndex.push_back(j); j++; }
                else
                  { _normalIndex.push_back(-1); }
            }
				
				// transfer normal to Ifs
          {
            int i;
            for(i=0;i<normal.size();i++)
              _normal.push_back(normal[i]);
          }
				
          if(pColor.getBinding()==PB_PER_CORNER)
            {
              // adjust _colorIndex
              // color.size() should be equal to
              // the number of new corners
              int i = i0;
              int j = _color.size()/3;
              for(;i<_coordIndex.size();i++)
                if(_coordIndex[i]>=0)
                  { _colorIndex.push_back(j); j++; }
                else
                  { _colorIndex.push_back(-1); }
            }
				
				// transfer color to Ifs
          {
            int i;
            for(i=0;i<color.size();i++)
              _color.push_back(color[i]);
          }
				
          if(pTexCoord.getBinding()==PB_PER_CORNER)
            {
              // adjust _texCoordIndex
              // texCoord.size() should be equal to
              // the number of new corners
              int i = i0;
              int j = _texCoord.size()/2;
              for(;i<_coordIndex.size();i++)
                if(_coordIndex[i]>=0)
                  { _texCoordIndex.push_back(j); j++; }
                else
                  { _texCoordIndex.push_back(-1); }
            }
				
				// transfer texCoord to Ifs
          {
            int i;
            for(i=0;i<texCoord.size();i++)
              _texCoord.push_back(texCoord[i]);
          }
				
          nVcls0 = nVcls11;
          nV0 = nV1;

          newCC = false;
					
        }
      //} while(codec.decode(MPG_Global::last_component) == false);
      if(codec.decode(MPG_Global::last_component) == false)
        newCC = true;
      if(!newCC && wasNewMobl)
        {
          // stop_qcoder()
          codec.finish_decode();
        }
		  
      // advance to next byte boundary and read ahead to determine
      // if another 3D_Mesh_Object_Base_Layer() follows
      //} while(istrm.lookingAtStartCode(MPG_Global::MOBL_start_code)==true);
      if(!newCC && (istrm.lookingAtStartCode(MPG_Global::MOBL_start_code)==true))
        newMobl = true;

    }
  else
    {
      // decode 3D_Mesh_Object_Forest_Split()
    }

  // advance to next byte boundary and read ahead to determine
  // if anoter 3D_Mesh_Object_Layer() folllows
  //} while(istrm.lookingAtStartCode(MPG_Global::MOL_start_code)==true);
  if(!newCC && !newMobl && (istrm.lookingAtStartCode(MPG_Global::MOL_start_code)==true))
    newMol = true;

  if(!newCC && !newMobl && !newMol)
    expandDone = true;

  success = true;
  return success;
}

//------------------------------------------------------
// expandTriangles
//------------------------------------------------------
bool MPG_IfsQ_E::expandTriangles(MPG_InputStream& istrm,
                                 MPG_State& state, 
                                 MPG_Property& pCoord, MPG_Property& pNormal,
                                 MPG_Property& pColor, MPG_Property& pTexCoord,
                                 vector<int>& qCoord, vector<int>& qNormal,
                                 vector<int>& qColor, vector<int>& qTexCoord,
                                 vector<int>& v_visited, vector<int>& v_cluster,
                                 MPG_Stitcher& stitcher, vector<int>& v_father,
                                 MPG_Stitches& v_stitches, 
                                 MPG_QFac& codec, 
                                 vector<int>&   coordIndex,
                                 vector<float>& coord,
                                 vector<float>& normal,
                                 vector<float>& color,
                                 vector<float>& texCoord,
                                 vector<int>& f_father,
                                 vector<int>& c_traversal,
                                 vector<int>& c_father,
                                 vector<int>& run_father,
                                 vector<int>& run_length,
                                 vector<bool>& run_ends_in_leaf,
                                 vector<int>& stk, 
                                 vector<int>& faceStack, 
                                 MPG_SPolygon& sp, MPG_VGraph& vg,
                                 vector<int>& v_loop_cc, 
                                 MPG_Loops& loop, MPG_Loops& c_loop,
                                 bool& expandDone)
{
  bool success = false;
  bool wasNewMobl = false;

  int nTriangles = 0;
  int nV0 = 0, nV1 = 0, nVcls0 = 0, nVcls1 = 0;
  bool newMol, newMobl, newCC, newTri;
  int mol_id, lastTriangle;
  int prev_marching_pattern, prev_polygon_edge, triangulated;
  int nPE, nF, nC, nCoord, nNormal, nColor, nTexCoord;
  int iP, iR, iL, iO, iF, iFp, iC;
  int cRp, cR, cLp, cL, cO, nVcc, iRun;

  state.GetState(nTriangles, nV0, nV1, nVcls0, nVcls1,
                 newMol, newMobl, newCC, newTri, mol_id, lastTriangle, prev_marching_pattern, 
                 prev_polygon_edge, triangulated, nPE, nF, nC, nCoord, nNormal, nColor, 
                 nTexCoord, iP, iR, iL, iO, iF, iFp, iC, cRp, cR, cLp, cL, cO, nVcc, iRun);

  // decode one or more 3D_Mesh_Object_Layer()

  if(newMol)
    {
      // decode 3D_Mesh_Object_Layer()
		
      // 3D_Mesh_Object_Layer() {
      //   3D_MOL_start_code
      //   mol_id
      //   if(cgd_data==1) {
      //     N_Vertices
      //     N_Triangles
      //     N_Edges
      //   }
      //   if(mol_id==0)
      //     3D_Mesh_Object_Base_Layer()
      //   else
      //     3D_Mesh_Object_Refinement_Layer()
      // }
		
      // read 3D_MOL_start_code
      if(istrm.getStartCode(MPG_Global::MOL_start_code)==false)
        return false;
		
      mol_id = 0;
      {
        // open a bit stream
        MPG_InputBitStream ibstrm(istrm);
			
        // read mol_id
        if(ibstrm.getMBitInt(8,mol_id)==false) return false;
			
        // read cgd data
        if(_cgd_data==true)
          {
            int N_Vertices = 0;
            if(ibstrm.getMBitInt(24,N_Vertices)==false) return false;
				
            int N_Triangles = 0;
            if(ibstrm.getMBitInt(24,N_Triangles)==false) return false;
				
            int N_Edges = 0;
            if(ibstrm.getMBitInt(24,N_Edges)==false) return false;
          }
			
        ibstrm.flush();
      }
		
      newMol = false;
    }

  if(mol_id==0)
    {
      //
      // decode 3D_Mesh_Object_Base_Layer()
      //
		
      // GT Wed Jun 09 21:13:23 1999 TODO:
      // codap support ?
		
      // 3D_Mesh_Object_Base_Layer() {
      //   do {
      //     3D_MOBL_start_code
      //     mobl_id
      //     start_qcoder()
      //     do {
      //       connected_component()
      //       last_component
      //     } while(last_component==false)
      //     stop_qcoder()
      //   } while(nextbits_bytealigned()==3D_MOBL_start_code)
      // }
		
      if(newMobl) 
        {	
          // read 3D_MOBL_start_code
          if(istrm.getStartCode(MPG_Global::MOBL_start_code)==false)
            return false;
			
          {
				// open a bit stream
            MPG_InputBitStream ibstrm(istrm);     
				
				// read mobl_id
            int mobl_id = 0;
            if(ibstrm.getMBitInt(8,mobl_id)==false) return false;
				
            ibstrm.flush();
          }
			
          //  start the arithmetic codec
          // MPG_QFac codec(istrm);
          codec.start_decode();

          // reset probability model
          MPG_Global::reset();

          newMobl = false;
          wasNewMobl = true;
        }

      if(newCC)
        {
          // decode connected_component()
			
          // connected_component() {
          //   vertex_graph()
          //   has_stitches
          //   if(has_stitches==1)
          //     stitches()l
          //   triangle_tree()
          //   triangle_data()
          // }
			
          // decode vertex_graph() and construct vertex loop
          int nVcc = 0;
          //vector<int> v_loop_cc;
          v_loop_cc.erase(v_loop_cc.begin(), v_loop_cc.end());
          {
				// decode vertex_graph()
				//MPG_VGraph vg;
            vg.clear();
            if(vg.decode(codec) == false) return false;
				
            nVcc = vg.getNumberOfVertices();
            nV1 = nV0 + nVcc;
            while(v_visited.size()<nV1)
              v_visited.push_back(false);
				
				// convert to vertex loop look-up table
            MPG_Loops loop;
            vg.toLoop(loop);
            loop.getLoop(0, v_loop_cc);
				
            if(vg.getFather(v_father)==false || v_father.size()!=nV1)
                return false;
          }
					
          v_stitches.decode(nV0,nV1,stk,codec);
          {
				// use stitches nV0<=i<nV1 to partially recover v_cluster
				
            if(stitcher.use(nV0,v_father,v_stitches,v_cluster)==false)
              return false;
				
          }
					
          // we should have here :
          // v_cluster.size()==nV1
			
          if(v_cluster.size()!=nV1)
            {
              return false;
            }

          // for i=0,..,nV1-1
          //   v_cluster[i]>=0 && v_cluster[i]<nV1
          // v_father.size()==nV1 
          // for i=0,..,nV1-1
          //   v_father[i]<=i
          // v_stitches.size()==nV1
			
          // decode triangle_tree()
          //MPG_SPolygon sp;
          sp.clear();
          if(sp.decodeTriangleTree(codec) == false) return false;
					
          // decode triangle_data()
          //vector<int>   coordIndex;
          coordIndex.erase(coordIndex.begin(), coordIndex.end());
          coord.erase(coord.begin(), coord.end());
          normal.erase(normal.begin(), normal.end());
          texCoord.erase(texCoord.begin(), texCoord.end());

          prev_marching_pattern = 0; 
          prev_polygon_edge = 1;
          triangulated = 0;

          // face tree array
          f_father.erase(f_father.begin(), f_father.end());

          // corner tree arrays
          c_traversal.erase(c_traversal.begin(), c_traversal.end());
          c_father.erase(c_father.begin(), c_father.end());

          run_father.erase(run_father.begin(), run_father.end());
          run_length.erase(run_length.begin(), run_length.end());
          run_ends_in_leaf.erase(run_ends_in_leaf.begin(), run_ends_in_leaf.end());

          // traversal stack
          stk.erase(stk.begin(), stk.end());
          stk.insert(stk.begin(), (size_t)1, -1);

          nPE    = 0;    //      number of {non_root polygon_edge==0}
          nF     = 0;   // - nPE
          nC     = 0; // - 2*nPE

          nCoord = 0;
          nNormal = 0;
          nColor = 0;
          nTexCoord = 0;

          iP = -1;
          iR =  0;
          iL =  1;
          iO = -1;

          faceStack.erase(faceStack.begin(), faceStack.end());

          iF  = f_father.size();	// == 0
          iFp = iF;

          iC   = 0;
          cRp  = 0;
          cR   = 0;
          cLp  = 1;
          cL   = 1;
          cO   = -1;

          loop.clear();
          c_loop.clear();
	
          sp.preprocessTriangleData(codec, nVcc,
                                    nV0, 
                                    v_loop_cc,
                                    v_visited,
                                    v_cluster,
                                    coordIndex,
                                    pCoord, // property headers and quantizers
                                    pNormal,
                                    pColor,
                                    pTexCoord,
                                    qCoord, // quantized values
                                    qNormal,
                                    qColor,
                                    qTexCoord,
                                    prev_marching_pattern, 
                                    prev_polygon_edge, 
                                    triangulated,
                                    f_father,
                                    c_traversal,
                                    c_father,
                                    run_father,
                                    run_length,
                                    run_ends_in_leaf,
                                    stk, nPE, nF, nC,
                                    nCoord, nNormal, nColor, nTexCoord,
                                    iP, iR, iL, iO,
                                    faceStack, iF, iFp, iC,
                                    cRp, cR, cLp, cL, cO);

          newCC = false;
          iRun = 0;
          lastTriangle = 0;
          //	nVcls0 = nV0; 
          nVcls1 = 0;
        }

      if(newTri)
        {
          int nT = sp.getNumberOfTriangles();
          int nVloop = v_loop_cc.size();
          for(int iT=lastTriangle;(iT<lastTriangle+nTriangles) && (iT<nT);iT++)
            //for(int iT=0;iT<nT;iT++)
            {
              int iR1 = 0, iL1 = 0, iO1 = 0;
              sp.decodeSingleTriangle(iT, iRun, codec, nVcc,
                                      nV0, nVloop,
                                      loop, c_loop,
                                      v_loop_cc,
                                      v_visited,
                                      v_cluster,
                                      coordIndex,
                                      pCoord, // property headers and quantizers
                                      pNormal,
                                      pColor,
                                      pTexCoord,
                                      qCoord, // quantized values
                                      qNormal,
                                      qColor,
                                      qTexCoord,
                                      prev_marching_pattern, 
                                      prev_polygon_edge, 
                                      triangulated,
                                      f_father,
                                      c_traversal,
                                      c_father,
                                      run_father,
                                      run_length,
                                      run_ends_in_leaf,
                                      stk, nPE, nF, nC,
                                      nCoord, nNormal, nColor, nTexCoord,
                                      iP, iR, iL, iO, iR1, iL1, iO1,
                                      faceStack, iF, iFp, iC,
                                      cRp, cR, cLp, cL, cO);

              _coordIndex.push_back(v_cluster[nV0+v_loop_cc[iR1]]);
				//if(v_cluster[nV0+v_loop_cc[iR1]] > nVcls1) 
              //nVcls1 = v_cluster[nV0+v_loop_cc[iR1]];
              _coordIndex.push_back(v_cluster[nV0+v_loop_cc[iL1]]);
				//if(v_cluster[nV0+v_loop_cc[iL1]] > nVcls1) 
              //nVcls1 = v_cluster[nV0+v_loop_cc[iL1]];
              _coordIndex.push_back(v_cluster[nV0+v_loop_cc[iO1]]);
				//if(v_cluster[nV0+v_loop_cc[iO1]] > nVcls1) 
              //nVcls1 = v_cluster[nV0+v_loop_cc[iO1]];
              _coordIndex.push_back(-1);

				//TRACE("\n\t triangle (%d, %d, %d)",
				//	v_cluster[nV0+v_loop_cc[iR1]], v_cluster[nV0+v_loop_cc[iL1]], v_cluster[nV0+v_loop_cc[iO1]]);
				//TRACE(" or \t triangle (%d, %d, %d)",
				//	nV0+v_loop_cc[iR1], nV0+v_loop_cc[iL1], nV0+v_loop_cc[iO1]);

				// write out statistics
				/*{
                                  int v[3];
                                  v[0] = nV0+v_loop_cc[iR1];
                                  v[1] = nV0+v_loop_cc[iL1];
                                  v[2] = nV0+v_loop_cc[iO1];

                                  int k = 1;
                                  while(k)
                                  {
                                  k = 0;
                                  if(v[0] > v[1])
                                  {
                                  int a = v[0]; v[0] = v[1];
                                  v[1] = a; k = 1;
                                  }
                                  if(v[1] > v[2])
                                  {
                                  int a = v[1]; v[1] = v[2];
                                  v[2] = a; k = 1;
                                  }
                                  }

                                  freq[v[1] - v[0]] ++;
                                  freq[v[2] - v[1]] ++;
                                  }*/
            }

          lastTriangle += nTriangles;

          // 2/12/99 determine the new number of vertices after clustering
          /*		{
                        int i;
                        for (i=nV0+lastTriangle*3; (i < nV0+(lastTriangle+nTriangles)*3) && (i<nV1); i++)
                        if (v_cluster[i] > nVcls1) 
                        nVcls1 = v_cluster[i];
			}*/
          int nVcls11 = nVcls1+1;

          // unquantize vertices between nVcls0 and nVcls1

          // coord properties
          MPG_BBoxQ*   coordQuantizer    = pCoord.getBBoxQ();
          coord.erase(coord.begin(), coord.end());
          /*if(nVcls11-nVcls0)
            {
            coord.insert(coord.end(),(size_t) (3 * (nVcls11-nVcls0)),(float)0.0);
            coordQuantizer->unQuantize(qCoord.begin()+3*nVcls0, coord.begin(), nVcls11-nVcls0);
            }*/
          // this could be optimized, as it is it unquantizes all the coordinates
          // decoded so far every time
          coordQuantizer->unQuantize(qCoord, coord);
			
          /*		
			// progresive decoding of properties, not yet implemented
			
			//  normal properties
			MPG_NormalQ* normalQuantizer   = pNormal.getNormalQ();
			if(pNormal.getBinding() != PB_NONE)
			{
				// special case for per-vertex binding,
				// the qNormal coordinates are global to allow reuse
				// of some vertices by components decoded afterwards
				if(pNormal.getBinding() == PB_PER_VERTEX)
				{
                                // take only the trailing part of qNormal
                                vector <int> qNormal1;
                                int i;
                                for (i=nVcls0;i<nVcls11;i++)
                                qNormal1.push_back(qNormal[i]);

                                normalQuantizer->unQuantize(qNormal1, normal);
				}
				else normalQuantizer->unQuantize(qNormal, normal);
                                }

                                //  color properties
                                MPG_BBoxQ*   colorQuantizer    = pColor.getBBoxQ();
                                if(pColor.getBinding() != PB_NONE)
                                {
				if(pColor.getBinding() == PB_PER_VERTEX)
				{
                                color.erase(color.begin(),color.end());
                                color.insert(color.end(),(size_t) (3 * (nVcls11-nVcls0)),(float)0.0);
                                colorQuantizer->unQuantize(qColor.begin()+3*nVcls0, color.begin(), nVcls11-nVcls0);
				}
				else colorQuantizer->unQuantize(qColor, color);
                                }

                                //  texture properties
                                MPG_BBoxQ*   texCoordQuantizer = pTexCoord.getBBoxQ();
                                if(pTexCoord.getBinding() != PB_NONE)
                                {
				if(pTexCoord.getBinding() != PB_PER_VERTEX)
				{
                                texCoord.erase(texCoord.begin(),texCoord.end());
                                texCoord.insert(texCoord.end(),(size_t) (2 * (nVcls11-nVcls0)),(float)0.0);
                                texCoordQuantizer->unQuantize(qTexCoord.begin()+2*nVcls0, texCoord.begin(), nVcls11-nVcls0);
				}
				else
				texCoordQuantizer->unQuantize(qTexCoord, texCoord);
                                }
          */
          //if(lastTriangle - nTriangles)
          //{
          //	while(_coord.size() > nV0*3)
          //		_coord.pop_back();
          //}

          // move decoded coordinates into _coord
          _coord.erase(_coord.begin(), _coord.end());
          for(int i=0;i<coord.size();i++)
            _coord.push_back(coord[i]);

          /*
            // transfer coord to Ifs
            {
            int i;
            for(i=0;i<coord.size();i++) _coord.push_back(coord[i]);
            }

            if(pNormal.getBinding()==PB_PER_CORNER)
            {
				// adjust _normalIndex
				// normal.size() should be equal to
				// the number of new corners
				int i = i0;
				int j = _normal.size();
				for(;i<_coordIndex.size();i++)
                                if(_coordIndex[i]>=0)
                                { _normalIndex.push_back(j); j++; }
                                else
                                { _normalIndex.push_back(-1); }
                                }

                                // transfer normal to Ifs
                                {
				int i;
				for(i=0;i<normal.size();i++)
                                _normal.push_back(normal[i]);
                                }

                                if(pColor.getBinding()==PB_PER_CORNER)
                                {
				// adjust _colorIndex
				// color.size() should be equal to
				// the number of new corners
				int i = i0;
				int j = _color.size();
				for(;i<_coordIndex.size();i++)
                                if(_coordIndex[i]>=0)
                                { _colorIndex.push_back(j); j++; }
                                else
                                { _colorIndex.push_back(-1); }
                                }

                                // transfer color to Ifs
                                {
				int i;
				for(i=0;i<color.size();i++)
				_color.push_back(color[i]);
                                }

                                if(pTexCoord.getBinding()==PB_PER_CORNER)
                                {
				// adjust _texCoordIndex
				// texCoord.size() should be equal to
				// the number of new corners
				int i = i0;
				int j = _texCoord.size();
				for(;i<_coordIndex.size();i++)
                                if(_coordIndex[i]>=0)
                                { _texCoordIndex.push_back(j); j++; }
                                else
                                { _texCoordIndex.push_back(-1); }
                                }

                                // transfer texCoord to Ifs
                                {
				int i;
				for(i=0;i<texCoord.size();i++)
				_texCoord.push_back(texCoord[i]);
                                }
          */
          //nVcls0 = nVcls11;

          // AG 1/27/99 bug: nV0 should be increased:
          //nV0 = nV1;

          if(lastTriangle < nT)
            newTri = true;
          else
            newTri = false;
        }
      //} while(codec.decode(MPG_Global::last_component) == false);
      if(!newTri)
        if(codec.decode(MPG_Global::last_component) == false)
          {
            newCC = newTri = true;
            nV0 = nV1;
          }

      if(!newTri && !newCC && wasNewMobl)
        {
          // stop_qcoder()
          codec.finish_decode();
        }
		  
      // advance to next byte boundary and read ahead to determine
      // if another 3D_Mesh_Object_Base_Layer() follows
      //} while(istrm.lookingAtStartCode(MPG_Global::MOBL_start_code)==true);
      if(!newTri && !newCC)
        if((istrm.lookingAtStartCode(MPG_Global::MOBL_start_code)==true))
          newMobl = true;

    }
  else
    {
      // decode 3D_Mesh_Object_Forest_Split()
    }

  // advance to next byte boundary and read ahead to determine
  // if anoter 3D_Mesh_Object_Layer() folllows
  //} while(istrm.lookingAtStartCode(MPG_Global::MOL_start_code)==true);
  if(!newTri && !newCC && !newMobl)
    if((istrm.lookingAtStartCode(MPG_Global::MOL_start_code)==true))
      newMol = true;

  if(!newTri && !newCC && !newMobl && !newMol)
    expandDone = true;

  state.SetState(nTriangles, nV0, nV1, nVcls0, nVcls1,
                 newMol, newMobl, newCC, newTri, mol_id, lastTriangle, prev_marching_pattern, 
                 prev_polygon_edge, triangulated, nPE, nF, nC, nCoord, nNormal, nColor, 
                 nTexCoord, iP, iR, iL, iO, iF, iFp, iC, cRp, cR, cLp, cL, cO, nVcc, iRun);

  success = true;
  return success;
}

#endif // _SINGLE_RESOLUTION_PROGRESSIVE
