/************************************************************************

This software module was originally developed by 
	Pat Mulroy (pmulroy@visual.bt.co.uk) BT Labs
	Wei Wu (weiwu@stallion.risc.rockwell.com),  Rockwell Science Center
	Homer Chen (homer@jupiter.risc.rockwell.com),  Rockwell Science Center
	Karl Olav Lillevold (Karl.Lillevold@nta.no) Telenor
	Robert Danielsen (Robert.Danielsen@nta.no) Telenor
and modified by:
        Tolga K. Capin (capin@lig.di.epfl.ch), EPFL-LIG,
        Joaquim Esmerado (jle@lig.di.epfl.ch), EPFL-LIG.

in the course of development of the MPEG-4 Video (ISO/IEC 14496-2). 
This software module is an implementation of a part of one or more MPEG-4 Video
tools as specified by the MPEG-4 Video. ISO/IEC gives users of the MPEG-4 Video
free license to this software module or modifications thereof for use in 
hardware or software products claiming conformance to the MPEG-4 Video. Those 
intending to use this software module in hardware or software products are advised 
that its use may infringe existing patents. The original developer of this software 
module and his/her company, the subsequent editors and their companies, and ISO/IEC 
have no liability for use of this software module or modifications thereof in an 
implementation. Copyright is not released for non MPEG-4 Video conforming 
products. BT Labs, Rockwell, and Telenor retains full right to use the 
code for his/her own purpose, assign or donate the code to a third party and to
inhibit third parties from using the code for non <MPEG standard> conforming 
products. This copyright notice must be included in all copies or derivative 
works. Copyright (c) 1997.
 
*************************************************************************/


/************************************************************************
 *
 *  sac.c, part of tmn (TMN encoder)
 *  Copyright (C) 1996  Telenor R&D, Norway
 *        Karl Olav Lillevold <Karl.Lillevold@nta.no>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Karl Olav Lillevold               <Karl.Lillevold@nta.no>
 *  Telenor Research and Development
 *  P.O.Box 83                        tel.:   +47 63 84 84 00
 *  N-2007 Kjeller, Norway            fax.:   +47 63 81 00 76
 *
 *  Robert Danielsen                  e-mail: Robert.Danielsen@nta.no
 *  Telenor Research and Development  www:    http://www.nta.no/brukere/DVC/
 *  P.O.Box 83                        tel.:   +47 63 84 84 00
 *  N-2007 Kjeller, Norway            fax.:   +47 63 81 00 76
 *  
 ************************************************************************/
/*********************************************************************
 *
 * SAC Encoder Module
 * Algorithm as specified in H263 (Annex E)
 *         (c) BT Labs 1995
 *
 * Author: Pat Mulroy <pmulroy@visual.bt.co.uk>
 *
 *********************************************************************/
/*********************************************************************
 *
 * Code modified for adaptive arithmetic coding
 *
 * Author: Wei Wu (weiwu@stallion.risc.rockwell.com)
 *
 *********************************************************************/

#include "sim.h"

#define   q1    16384
#define   q2    32768
#define   q3    49152
#define   top   65535

#define	Max_frequency	16383


static long   low=0, high=top, opposite_bits=0, length=0, zerorun=0;

/*********************************************************************
 *
 *      Name:           AR_Encode
 *
 *      Description:    Encodes a symbol using syntax based arithmetic
 *        coding. Algorithm specified in H.263 (Annex E).
 *
 *      Input:          Array holding cumulative frequency data.
 *        Index into specific cumulative frequency array.
 *                      Static data for encoding endpoints.
 *
 *      Returns:        Number of bits used while encoding symbol.
 *
 *      Side Effects:   Modifies low, high, length and opposite_bits
 *        variables.
 *
 *      Author:         pmulroy@visual.bt.co.uk
 *
 *********************************************************************/

int AR_Encode(int index, int cumul_freq[ ])
{
  int bitcount=0;

  if (index<0) 
    return -1; /* Escape Code */

  length = high - low + 1;
  high = low - 1 + (length * cumul_freq[index]) / cumul_freq[0];
  low += (length * cumul_freq[index+1]) / cumul_freq[0];

  for ( ; ; ) {
    if (high < q2) {
      bitcount+=bit_opp_bits(0);
    }
    else if (low >= q2) {
      bitcount+=bit_opp_bits(1);	
      low -= q2; 
      high -= q2;
    }
    else if (low >= q1 && high < q3) {
      opposite_bits += 1; 
      low -= q1; 
      high -= q1;
    }
    else break;
 
    low *= 2; 
    high = 2*high+1;
  }
  return bitcount;
}

int bit_opp_bits(int bit) /* Output a bit and the following opposite bits */              
{                                   
  int bitcount=0;

  bitcount = bit_in_psc_layer(bit);

  while(opposite_bits > 0){
    bitcount += bit_in_psc_layer(!bit);
    opposite_bits--;
  }
  return bitcount;
}

/*********************************************************************
 *
 *      Name:           encoder_flush
 *
 *      Description:    Completes arithmetic coding stream before any
 *        fixed length codes are transmitted.
 *
 *      Input:          None
 *
 *      Returns:        Number of bits used.
 *
 *      Side Effects:   Resets low, high, zerorun and opposite_bits 
 *        variables.
 *
 *      Author:         pmulroy@visual.bt.co.uk
 *
 *********************************************************************/

int encoder_flush()
{
  int bitcount = 0;

  opposite_bits++;
  if (low < q1) {
    bitcount+=bit_opp_bits(0);
  }
  else {
    bitcount+=bit_opp_bits(1);
  }
  low = 0; 
  high = top;

  zerorun=0;

  return bitcount;
}

/*********************************************************************
 *
 *      Name:           bit_in_psc_layer
 *
 *      Description:    Inserts a bit into output bitstream and avoids
 *        picture start code emulation by stuffing a one
 *        bit.
 *
 *      Input:          Bit to be output.
 *
 *      Returns:        Nothing
 *
 *      Side Effects:   Updates zerorun variable.
 *
 *      Author:         pmulroy@visual.bt.co.uk
 *
 *********************************************************************/

int bit_in_psc_layer(int bit)
{
  int bitcount = 0;

  if (zerorun == 22) {
    bitcount += putbits (1, 1);
    zerorun = 0;
  }

  bitcount += putbits (1, bit);

  if (bit)
    zerorun = 0;
  else
    zerorun++;

  return bitcount;
}

/*********************************************************************
 *
 *      Name:           indexfn
 *
 *      Description:    Translates between symbol value and symbol
 *        index.
 *
 *      Input:          Symbol value, index table, max number of
 *        values.
 *
 *      Returns:        Index into cumulative frequency tables or
 *        escape code.
 *
 *      Side Effects:   none
 *
 *      Author:         pmulroy@visual.bt.co.uk
 *
 *********************************************************************/

int indexfn(int value, int table[], int max)
{
  int n=0;

  while(1) {
    if (table[n++]==value) return n-1;
    if (n>max) return -1;
  }

}


/*
 * update module for adaptive arithmetic coding
 *
 * Author: Wei Wu (weiwu@stallion.risc.rockwell.com)
 * Date:   1/14/97
 *
 */

void
update(int symbol, int index) 
{
  int n;
  int *cumul_tmp, *freq_tmp;

  if(fop.is_intra) {
    n = fop.fo.Irange_max[index]-fop.fo.Irange_min[index]+1;
    cumul_tmp = fop.fo.Icumul_freq[index];
    freq_tmp = fop.fo.Ifreq[index];
  } else {
    n = fop.fo.range_max[index]-fop.fo.range_min[index]+1;
    cumul_tmp = fop.fo.cumul_freq[index];
    freq_tmp = fop.fo.freq[index];
  }
  if (cumul_tmp[0] == Max_frequency) {
    int i, cum;
    cum = 0;
    for(i=n-1; i>=0; i--) {
      freq_tmp[i] = (freq_tmp[i]+1)/2;
      cum += freq_tmp[i];
      cumul_tmp[i] = cum;
    }
    cumul_tmp[n] = 0;
  }
  freq_tmp[symbol] += 1;
  cumul_tmp[symbol] += 1;
  while(symbol>0) {
    symbol -= 1;
    cumul_tmp[symbol] += 1;
  }
}


void
bap_update(int symbol, int index) 
{
  int n;
  int *cumul_tmp, *freq_tmp;

  if(fop.is_intra) {
    n = fop.bo.Irange_max[index]-fop.bo.Irange_min[index]+1;
    cumul_tmp = fop.bo.Icumul_freq[index];
    freq_tmp = fop.bo.Ifreq[index];
  } else {
    n = fop.bo.range_max[index]-fop.bo.range_min[index]+1;
    cumul_tmp = fop.bo.cumul_freq[index];
    freq_tmp = fop.bo.freq[index];
  }
  if (cumul_tmp[0] == Max_frequency) {
    int i, cum;
    cum = 0;
    for(i=n-1; i>=0; i--) {
      freq_tmp[i] = (freq_tmp[i]+1)/2;
      cum += freq_tmp[i];
      cumul_tmp[i] = cum;
    }
    cumul_tmp[n] = 0;
  }
  freq_tmp[symbol] += 1;
  cumul_tmp[symbol] += 1;
  while(symbol>0) {
    symbol -= 1;
    cumul_tmp[symbol] += 1;
  }
}

void
update_viseme_expression(int symbol, int index) 
{
  int n;
  int *cumul_tmp, *freq_tmp;

  if(fop.is_intra) {
    n = fop.fo.Irange_ve_max[index] - fop.fo.Irange_ve_min[index]+1;
    cumul_tmp = fop.fo.Icumul_ve[index];
    freq_tmp = fop.fo.Ifreq_ve[index];
  } else {
    n = fop.fo.range_ve_max[index] - fop.fo.range_ve_min[index]+1;
    cumul_tmp = fop.fo.cumul_ve[index];
    freq_tmp = fop.fo.freq_ve[index];
  }
  if (cumul_tmp[0] == Max_frequency) {
    int i, cum;
    cum = 0;
    for(i=n-1; i>=0; i--) {
      freq_tmp[i] = (freq_tmp[i]+1)/2;
      cum += freq_tmp[i];
      cumul_tmp[i] = cum;
    }
    cumul_tmp[n] = 0;
  }
  freq_tmp[symbol] += 1;
  cumul_tmp[symbol] += 1;
  while(symbol>0) {
    symbol -= 1;
    cumul_tmp[symbol] += 1;
  }
 
}


void 
init_arith() {

  int i, j, n;

  if(fop.fo.fap_mask[0]) {
	free(fop.fo.Ifreq_ve[0]);
	free(fop.fo.Icumul_ve[0]);
	free(fop.fo.Ifreq_ve[1]);
	free(fop.fo.Icumul_ve[1]);
	free(fop.fo.Ifreq_ve[2]);
	free(fop.fo.Icumul_ve[2]);
	n = fop.fo.Irange_ve_max[0] - fop.fo.Irange_ve_min[0] + 1;
	fop.fo.Ifreq_ve[0] = (int*) malloc(n*sizeof(int));
	fop.fo.Icumul_ve[0] = (int*) malloc((n+1)*sizeof(int));
	for(j=0; j<n; j++) {
		fop.fo.Ifreq_ve[0][j] = 1;
		fop.fo.Icumul_ve[0][j] = n-j;
	}
	fop.fo.Icumul_ve[0][n] = 0;

	n = fop.fo.Irange_ve_max[1] - fop.fo.Irange_ve_min[1] + 1;
	fop.fo.Ifreq_ve[1] = (int*) malloc(n*sizeof(int));
	fop.fo.Icumul_ve[1] = (int*) malloc((n+1)*sizeof(int));
	for(j=0; j<n; j++) {
		fop.fo.Ifreq_ve[1][j] = 1;
		fop.fo.Icumul_ve[1][j] = n-j;
	}
	fop.fo.Icumul_ve[1][n] = 0;

	n = fop.fo.Irange_ve_max[2] - fop.fo.Irange_ve_min[2] + 1;
	fop.fo.Ifreq_ve[2] = (int*) malloc(n*sizeof(int));
	fop.fo.Icumul_ve[2] = (int*) malloc((n+1)*sizeof(int));
	for(j=0; j<n; j++) {
		fop.fo.Ifreq_ve[2][j] = 1;
		fop.fo.Icumul_ve[2][j] = n-j;
	}
	fop.fo.Icumul_ve[2][n] = 0;
  }

  if(fop.fo.fap_mask[1]) {
	free(fop.fo.Ifreq_ve[3]);
	free(fop.fo.Icumul_ve[3]);
	free(fop.fo.Ifreq_ve[4]);
	free(fop.fo.Icumul_ve[4]);
	free(fop.fo.Ifreq_ve[5]);
	free(fop.fo.Icumul_ve[5]);
	free(fop.fo.Ifreq_ve[6]);
	free(fop.fo.Icumul_ve[6]);

        n = fop.fo.Irange_ve_max[3] - fop.fo.Irange_ve_min[3] + 1;
        fop.fo.Ifreq_ve[3] = (int*) malloc(n*sizeof(int));
        fop.fo.Icumul_ve[3] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.Ifreq_ve[3][j] = 1;
                fop.fo.Icumul_ve[3][j] = n-j;
        }
        fop.fo.Icumul_ve[3][n] = 0;
 
	n = fop.fo.Irange_ve_max[4] - fop.fo.Irange_ve_min[4] + 1;
	fop.fo.Ifreq_ve[4] = (int*) malloc(n*sizeof(int));
	fop.fo.Icumul_ve[4] = (int*) malloc((n+1)*sizeof(int));
	for(j=0; j<n; j++) {
		fop.fo.Ifreq_ve[4][j] = 1;
		fop.fo.Icumul_ve[4][j] = n-j;
	}
	fop.fo.Icumul_ve[4][n] = 0;

        n = fop.fo.Irange_ve_max[5] - fop.fo.Irange_ve_min[5] + 1;
        fop.fo.Ifreq_ve[5] = (int*) malloc(n*sizeof(int));
        fop.fo.Icumul_ve[5] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.Ifreq_ve[5][j] = 1;
                fop.fo.Icumul_ve[5][j] = n-j;
        }
        fop.fo.Icumul_ve[5][n] = 0;
 
	n = fop.fo.Irange_ve_max[6] - fop.fo.Irange_ve_min[6] + 1;
	fop.fo.Ifreq_ve[6] = (int*) malloc(n*sizeof(int));
	fop.fo.Icumul_ve[6] = (int*) malloc((n+1)*sizeof(int));
	for(j=0; j<n; j++) {
		fop.fo.Ifreq_ve[6][j] = 1;
		fop.fo.Icumul_ve[6][j] = n-j;
	}
	fop.fo.Icumul_ve[6][n] = 0;
  }

  for(i=2; i<NFAP; i++) {
	if(fop.fo.fap_mask[i]) {
		free(fop.fo.Ifreq[i]);
		free(fop.fo.Icumul_freq[i]);
		n = fop.fo.Irange_max[i]-fop.fo.Irange_min[i]+1;
		fop.fo.Ifreq[i] = (int *) malloc(n*sizeof(int));
		fop.fo.Icumul_freq[i] = (int *) malloc((n+1)*sizeof(int));
		for(j=0; j<n; j++) {
			fop.fo.Ifreq[i][j] = 1;
			fop.fo.Icumul_freq[i][j] = n-j;
		}
		fop.fo.Icumul_freq[i][n] = 0;
	}
  }

  if(fop.fo.fap_mask[0]) {
	free(fop.fo.freq_ve[0]);
	free(fop.fo.cumul_ve[0]);
	free(fop.fo.freq_ve[1]);
	free(fop.fo.cumul_ve[1]);
	free(fop.fo.freq_ve[2]);
	free(fop.fo.cumul_ve[2]);
        n = fop.fo.range_ve_max[0] - fop.fo.range_ve_min[0] + 1;
        fop.fo.freq_ve[0] = (int*) malloc(n*sizeof(int));
        fop.fo.cumul_ve[0] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.freq_ve[0][j] = 1;
                fop.fo.cumul_ve[0][j] = n-j;
        }
        fop.fo.cumul_ve[0][n] = 0;
 
        n = fop.fo.range_ve_max[1] - fop.fo.range_ve_min[1] + 1;
        fop.fo.freq_ve[1] = (int*) malloc(n*sizeof(int));
        fop.fo.cumul_ve[1] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.freq_ve[1][j] = 1;
                fop.fo.cumul_ve[1][j] = n-j;
        }
        fop.fo.cumul_ve[1][n] = 0;
 
	n = fop.fo.range_ve_max[2] - fop.fo.range_ve_min[2] + 1;
	fop.fo.freq_ve[2] = (int*) malloc(n*sizeof(int));
	fop.fo.cumul_ve[2] = (int*) malloc((n+1)*sizeof(int));
	for(j=0; j<n; j++) {
		fop.fo.freq_ve[2][j] = 1;
		fop.fo.cumul_ve[2][j] = n-j;
	}
	fop.fo.cumul_ve[2][n] = 0;
  }

  if(fop.fo.fap_mask[1]) {
	free(fop.fo.freq_ve[3]);
	free(fop.fo.cumul_ve[3]);
	free(fop.fo.freq_ve[4]);
	free(fop.fo.cumul_ve[4]);
	free(fop.fo.freq_ve[5]);
	free(fop.fo.cumul_ve[5]);
	free(fop.fo.freq_ve[6]);
	free(fop.fo.cumul_ve[6]);
        n = fop.fo.range_ve_max[3] - fop.fo.range_ve_min[3] + 1;
        fop.fo.freq_ve[3] = (int*) malloc(n*sizeof(int));
        fop.fo.cumul_ve[3] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.freq_ve[3][j] = 1;
                fop.fo.cumul_ve[3][j] = n-j;
        }
        fop.fo.cumul_ve[3][n] = 0;
 
	n = fop.fo.range_ve_max[4] - fop.fo.range_ve_min[4] + 1;
	fop.fo.freq_ve[4] = (int*) malloc(n*sizeof(int));
	fop.fo.cumul_ve[4] = (int*) malloc((n+1)*sizeof(int));
	for(j=0; j<n; j++) {
		fop.fo.freq_ve[4][j] = 1;
		fop.fo.cumul_ve[4][j] = n-j;
	}
	fop.fo.cumul_ve[4][n] = 0;

        n = fop.fo.range_ve_max[5] - fop.fo.range_ve_min[5] + 1;
        fop.fo.freq_ve[5] = (int*) malloc(n*sizeof(int));
        fop.fo.cumul_ve[5] = (int*) malloc((n+1)*sizeof(int));
        for(j=0; j<n; j++) {
                fop.fo.freq_ve[5][j] = 1;
                fop.fo.cumul_ve[5][j] = n-j;
        }
        fop.fo.cumul_ve[5][n] = 0;

	n = fop.fo.range_ve_max[6] - fop.fo.range_ve_min[6] + 1;
	fop.fo.freq_ve[6] = (int*) malloc(n*sizeof(int));
	fop.fo.cumul_ve[6] = (int*) malloc((n+1)*sizeof(int));
	for(j=0; j<n; j++) {
		fop.fo.freq_ve[6][j] = 1;
		fop.fo.cumul_ve[6][j] = n-j;
	}
	fop.fo.cumul_ve[6][n] = 0;
  }

  for(i=2; i<NFAP; i++) {
	if(fop.fo.fap_mask[i]) {
		free(fop.fo.freq[i]);
		free(fop.fo.cumul_freq[i]);
		n = fop.fo.range_max[i]-fop.fo.range_min[i]+1;
		fop.fo.freq[i] = (int *) malloc(n*sizeof(int));
		fop.fo.cumul_freq[i] = (int *) malloc((n+1)*sizeof(int));
		for(j=0; j<n; j++) {
			fop.fo.freq[i][j] = 1;
			fop.fo.cumul_freq[i][j] = n-j;
		}
		fop.fo.cumul_freq[i][n] = 0;
	}
  }

}


void 
bap_init_arith() {

  int i, j, n;

  for(i=0; i<NUM_BAPS; i++) {

	if(fop.bo.bap_mask[i]) {
	  free(fop.bo.Ifreq[i]);
	  free(fop.bo.Icumul_freq[i]);
	  n = fop.bo.Irange_max[i]-fop.bo.Irange_min[i]+1;
	  fop.bo.Ifreq[i] = (int *) malloc(n*sizeof(int));
	  fop.bo.Icumul_freq[i] = (int *) malloc((n+1)*sizeof(int));
	  for(j=0; j<n; j++) {
	    fop.bo.Ifreq[i][j] = 1;
	    fop.bo.Icumul_freq[i][j] = n-j;
	  }
	  fop.bo.Icumul_freq[i][n] = 0;
	}
  }

  for(i=0; i<NUM_BAPS; i++) {
    if(fop.bo.bap_mask[i]) {
      free(fop.bo.freq[i]);
      free(fop.bo.cumul_freq[i]);
      n = fop.bo.range_max[i]-fop.bo.range_min[i]+1;
      fop.bo.freq[i] = (int *) malloc(n*sizeof(int));
      fop.bo.cumul_freq[i] = (int *) malloc((n+1)*sizeof(int));
      for(j=0; j<n; j++) {
	fop.bo.freq[i][j] = 1;
	fop.bo.cumul_freq[i][j] = n-j;
      }
      fop.bo.cumul_freq[i][n] = 0;
    }
  }

}
