/*************************************************************************

This software module was originally developed by 

	Ming-Chieh Lee (mingcl@microsoft.com), Microsoft Corporation
	Wei-ge Chen (wchen@microsoft.com), Microsoft Corporation
	Bruce Lin (blin@microsoft.com), Microsoft Corporation
	Chuang Gu (chuanggu@microsoft.com), Microsoft Corporation
	Simon Winder (swinder@microsoft.com), Microsoft Corporation

in the course of development of the MPEG-4 Video (ISO/IEC 14496-2). 
This software module is an implementation of a part of one or more MPEG-4 Video tools 
as specified by the MPEG-4 Video. 
ISO/IEC gives users of the MPEG-4 Video free license to this software module or modifications 
thereof for use in hardware or software products claiming conformance to the MPEG-4 Video. 
Those intending to use this software module in hardware or software products are advised that its use may infringe existing patents. 
The original developer of this software module and his/her company, 
the subsequent editors and their companies, 
and ISO/IEC have no liability for use of this software module or modifications thereof in an implementation. 
Copyright is not released for non MPEG-4 Video conforming products. 
Microsoft retains full right to use the code for his/her own purpose, 
assign or donate the code to a third party and to inhibit third parties from using the code for non <MPEG standard> conforming products. 
This copyright notice must be included in all copies or derivative works. 

Copyright (c) 1996, 1997.


Module Name:

	wvtdecomp.cpp

Abstract:

	decomposition tree

Revision History:

*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "typeapi.h"
#include "wavelet.h"

#ifdef __MFC_
#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

#define new DEBUG_NEW				   
#endif // __MFC_

CWvtDecompTree::CWvtDecompTree(Int iWidth,Int iHeight,Int iDecompLevels,
							   FilterType type)
{
	assert((iWidth&1 | iHeight&1)==0);

	iWidth>>=1;
	iHeight>>=1;
	m_rct = CRct(0,0,iWidth,iHeight);
	m_filterType=type;
	m_pParent=NULL;
	m_pChild=NULL;
	m_pfiData[WVT_LL] = new CFloatImage(m_rct); // mostly partial dc results
	m_pfiData[WVT_LH] = new CFloatImage(m_rct);
	m_pfiData[WVT_HL] = new CFloatImage(m_rct);
	m_pfiData[WVT_HH] = new CFloatImage(m_rct);
	if(iDecompLevels>1)
	{
		m_pChild = new CWvtDecompTree(iWidth,iHeight,iDecompLevels-1,type);
		m_pChild->m_pParent = this;
	}
}

CWvtDecompTree::~CWvtDecompTree()
{
	if(data(WVT_LL)!=NULL)
		delete data(WVT_LL);
	if(data(WVT_LH)!=NULL)
		delete data(WVT_LH);
	if(data(WVT_HL)!=NULL)
		delete data(WVT_HL);
	if(data(WVT_HH)!=NULL)
		delete data(WVT_HH);
	if(m_pChild!=NULL)
		delete m_pChild;
}

CFloatImage *CWvtDecompTree::dcBand()
{
	if(m_pChild==NULL)
		return data(WVT_LL);
	else
		return m_pChild->dcBand();
}

void CWvtDecompTree::forwardTransform(CIntImage *piiSrc)
{
	// decomp source into a tree recursively
	CWvtFilterUnit filterUnit(m_filterType);
	CFloatImage fiCopy(*piiSrc);
	filterUnit.forwardTransform(&fiCopy,this);
	if(m_pChild!=NULL)
		m_pChild->forwardTransform(data(WVT_LL));
}

void CWvtDecompTree::forwardTransform(CFloatImage *pfiSrc)
{
	// decomp source into a tree recursively
	CWvtFilterUnit filterUnit(m_filterType);
	filterUnit.forwardTransform(pfiSrc,this);
	if(m_pChild!=NULL)
		m_pChild->forwardTransform(data(WVT_LL));
}

void CWvtDecompTree::roundNearestInt(FilterBank band)
{
	Int i;
	PixelF *ppxlf = (PixelF *)data(band)->pixels();
	Float fVal;

	for(i=m_rct.area();i;i--,ppxlf++)
	{
		fVal = *ppxlf;
		if(fVal>=0)
			*ppxlf = (PixelF)floor(fVal+0.5);
		else
			*ppxlf = (PixelF)ceil(fVal-0.5);
	}

	if(m_pChild!=NULL)
		m_pChild->roundNearestInt(band);
}

void CWvtDecompTree::inverseTransform(CFloatImage *pfiDst)
{
	// reconstruct tree into dst recursively 

	if(m_pChild!=NULL)
		m_pChild->inverseTransform(data(WVT_LL));
	CWvtFilterUnit filterUnit(m_filterType);
	filterUnit.inverseTransform(this,pfiDst);
}

void CWvtDecompTree::inverseTransform(CIntImage *piiDst)
{
	// reconstruct tree into dst recursively 

	if(m_pChild!=NULL)
		m_pChild->inverseTransform(data(WVT_LL));
	CWvtFilterUnit filterUnit(m_filterType);
	CFloatImage fiCopy(piiDst->where());
	filterUnit.inverseTransform(this,&fiCopy);
	piiDst->overlay(fiCopy);
}

void CWvtDecompTree::printDecompTree()
{
	CoordI iX,iY;

	printf("\nLH Band\n");
	for(iY=0;iY<where().height();iY++)
	{
		for(iX=0;iX<where().width;iX++)
			printf("%.5g ",m_pfiData[WVT_LH]->pixel(iX,iY));
		printf("\n");
	}
	printf("\nHL Band\n");
	for(iY=0;iY<where().height();iY++)
	{
		for(iX=0;iX<where().width;iX++)
			printf("%.5g ",m_pfiData[WVT_HL]->pixel(iX,iY));
		printf("\n");
	}
	printf("\nHH Band\n");
	for(iY=0;iY<where().height();iY++)
	{
		for(iX=0;iX<where().width;iX++)
			printf("%.5g ",m_pfiData[WVT_HH]->pixel(iX,iY));
		printf("\n");
	}
	if(m_pChild!=NULL)
		m_pChild->printDecompTree();
	else
	{
		printf("\nLL Band\n");
		for(iY=0;iY<where().height();iY++)
		{
			for(iX=0;iX<where().width;iX++)
				printf("%.5g ",m_pfiData[WVT_LL]->pixel(iX,iY));
			printf("\n");
		}
	}
}
