/*************************************************************************

This software module was originally developed by 

	Ming-Chieh Lee (mingcl@microsoft.com), Microsoft Corporation
	Wei-ge Chen (wchen@microsoft.com), Microsoft Corporation
	Bruce Lin (blin@microsoft.com), Microsoft Corporation
	Chuang Gu (chuanggu@microsoft.com), Microsoft Corporation
	Simon Winder (swinder@microsoft.com), Microsoft Corporation

in the course of development of the MPEG-4 Video (ISO/IEC 14496-2). 
This software module is an implementation of a part of one or more MPEG-4 Video tools 
as specified by the MPEG-4 Video. 
ISO/IEC gives users of the MPEG-4 Video free license to this software module or modifications 
thereof for use in hardware or software products claiming conformance to the MPEG-4 Video. 
Those intending to use this software module in hardware or software products are advised that its use may infringe existing patents. 
The original developer of this software module and his/her company, 
the subsequent editors and their companies, 
and ISO/IEC have no liability for use of this software module or modifications thereof in an implementation. 
Copyright is not released for non MPEG-4 Video conforming products. 
Microsoft retains full right to use the code for his/her own purpose, 
assign or donate the code to a third party and to inhibit third parties from using the code for non <MPEG standard> conforming products. 
This copyright notice must be included in all copies or derivative works. 

Copyright (c) 1996, 1997.


Module Name:

	wvtqtree.cpp

Abstract:

	quantisation and coding tree

Revision History:

*************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "typeapi.h"
#include "wavelet.h"

#define ZT_Z    0
#define ZT_ZTR	1
#define ZT_VZTR 2
#define ZT_VAL	3
#define ZT_IGN	4
#define ZT_OLD	8
#define ZT_XXX  5

#define NSYMS       3

#define ACSYM_ZTR	2
#define ACSYM_VZTR	1
#define ACSYM_VAL	0

#define NSYMSBL     5

#define ACSYMBL_IZ  0
#define ACSYMBL_DEC 0
#define ACSYMBL_ZTR 1
#define ACSYMBL_INC 1
#define ACSYMBL_POS 2
#define ACSYMBL_NEG 3
#define ACSYMBL_CNT 3
#define ACSYMBL_EOF 4

static Int rgiFreqTableCoef[] = { 1,1,1,1,0 };
static Int rgiFreqTableLeaf[] = { 1,0,1,1,0 };
static Int rgiFreqTableRefine[] = { 1,1,0,0,0 };
static Int rgiFreqTableEnd[] = { 0,0,0,4,1 };

#ifdef __MFC_
#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

#define new DEBUG_NEW				   
#endif // __MFC_

CWvtQuantTree::CWvtQuantTree(CWvtDecompTree *pwvtDecomp)
{
	CRct rct = pwvtDecomp->where();
	Int iSize = rct.area();
	m_pParent=m_pChild=NULL;
	m_pwvtDecomp=pwvtDecomp;
	m_piiQuant[WVT_LL] = NULL;
	m_piiQuant[WVT_LH] = new CIntImage(rct);
	m_piiQuant[WVT_HL] = new CIntImage(rct);
	m_piiQuant[WVT_HH] = new CIntImage(rct);
	m_piiResidual[WVT_LL] = NULL;
	m_piiResidual[WVT_LH] = new CIntImage(rct);
	m_piiResidual[WVT_HL] = new CIntImage(rct);
	m_piiResidual[WVT_HH] = new CIntImage(rct);
	m_pucMap[WVT_LL] = NULL;
	m_pucMap[WVT_LH] = new U8 [iSize];
	m_pucMap[WVT_HL] = new U8 [iSize];
	m_pucMap[WVT_HH] = new U8 [iSize];
	m_piiOrigLL = new CIntImage(rct);
		
	if(pwvtDecomp->getChild()!=NULL)
	{
		m_pChild = new CWvtQuantTree(pwvtDecomp->getChild());
		m_pChild->m_pParent = this;
	}
}

CWvtQuantTree::~CWvtQuantTree()
{
	if(m_piiQuant[WVT_LL]!=NULL)
		delete m_piiQuant[WVT_LL];
	if(m_piiQuant[WVT_LH]!=NULL)
		delete m_piiQuant[WVT_LH];
	if(m_piiQuant[WVT_HL]!=NULL)
		delete m_piiQuant[WVT_HL];
	if(m_piiQuant[WVT_HH]!=NULL)
		delete m_piiQuant[WVT_HH];
	if(m_piiResidual[WVT_LL]!=NULL)
		delete m_piiResidual[WVT_LL];
	if(m_piiResidual[WVT_LH]!=NULL)
		delete m_piiResidual[WVT_LH];
	if(m_piiResidual[WVT_HL]!=NULL)
		delete m_piiResidual[WVT_HL];
	if(m_piiResidual[WVT_HH]!=NULL)
		delete m_piiResidual[WVT_HH];
	if(m_pucMap[WVT_LL]!=NULL)
		delete m_pucMap[WVT_LL];
	if(m_pucMap[WVT_LH]!=NULL)
		delete m_pucMap[WVT_LH];
	if(m_pucMap[WVT_HL]!=NULL)
		delete m_pucMap[WVT_HL];
	if(m_pucMap[WVT_HH]!=NULL)
		delete m_pucMap[WVT_HH];
	if(m_piiOrigLL!=NULL)
		delete m_piiOrigLL;
	if(m_pChild!=NULL)
		delete m_pChild;
}

CWvtQuantTree *CWvtQuantTree::getRoot()
{
	if(getChild()!=NULL)
		return getChild()->getRoot();
	else
		return this;
}

// load residual from wavelet decomp tree
void CWvtQuantTree::load(FilterBank band)
{
	Int i;

	PixelI *ppxliDst = (PixelI *)m_piiResidual[band]->pixels();
	const PixelF *ppxlfSrc = m_pwvtDecomp->data(band)->pixels();
	for(i=getArea();i;i--)
		*ppxliDst++ = (PixelI)*ppxlfSrc++;
}

void CWvtQuantTree::saveOriginal()
{
	Int i;
	Float fVal;

	PixelI *ppxliDst = (PixelI *)m_piiOrigLL->pixels();
	const PixelF *ppxlfSrc = m_pwvtDecomp->data(WVT_LL)->pixels();

	// copy and round
	for(i=getArea();i;i--,ppxliDst++)
	{
		fVal = *ppxlfSrc++;
		if(fVal>=0.0)
			*ppxliDst = (Int)(fVal+0.5);
		else
			*ppxliDst = (Int)(fVal-0.5);
	}

	// recurse
	if(getChild()!=NULL)
		getChild()->saveOriginal();
}

// quantise the residual into quant
void CWvtQuantTree::quant(Int iQuant,FilterBank band)
{
	Int i,iVal;

	const PixelI *ppxliSrc = m_piiResidual[band]->pixels();
	PixelI *ppxliDst = (PixelI *)m_piiQuant[band]->pixels();
	for(i=getArea();i;i--,ppxliSrc++,ppxliDst++)
	{ 
		iVal = ACQuant(*ppxliSrc,iQuant);
		*ppxliDst = ACReorder(iVal);
	}
}

// inv quantise quant, subtract from residual and possibly store or add
// iq back into decomp tree
void CWvtQuantTree::invQuant(Int iQuant,FilterBank band,Bool bReconFlag,
							 Bool bFirstLayerFlag)
{
	Int i,iVal;

	const PixelI *ppxliSrc = m_piiQuant[band]->pixels();
	PixelI *ppxliDst = (PixelI *)m_piiResidual[band]->pixels();
	if(bReconFlag==TRUE)
	{
		PixelF *ppxlfRec = (PixelF *)m_pwvtDecomp->data(band)->pixels();
		if(bFirstLayerFlag==TRUE)
			for(i=getArea();i;i--,
				ppxliSrc++,ppxliDst++,ppxlfRec++)
			{
				iVal=*ppxliSrc;
				iVal=InvACReorder(iVal);
				iVal=InvACQuant(iVal,iQuant);
				*ppxliDst -= iVal;
				*ppxlfRec = (PixelF)iVal;  // store in wavelet decomp
			}
		else
			for(i=getArea();i;i--,
				ppxliSrc++,ppxliDst++,ppxlfRec++)
			{
				iVal=*ppxliSrc;
				iVal=InvACReorder(iVal);
				iVal=InvACQuant(iVal,iQuant);
				*ppxliDst -= iVal;
				*ppxlfRec += (PixelF)iVal; // add to wavelet decomp
			}
	}
	else  // ignore wavelet decomp
		for(i=getArea();i;i--,ppxliSrc++,ppxliDst++)
		{
			iVal=*ppxliSrc;
			iVal=InvACReorder(iVal);
			iVal=InvACQuant(iVal,iQuant);
			*ppxliDst -= iVal;
		}
}

// load residual from wavelet decomp tree,
// quantise current residual, inverse quant, subtract from residual and
// possibly also store iq back into decomp tree
void CWvtQuantTree::makeSNRLayer(Int iQuant,Bool bReconFlag,Bool bFirstLayerFlag)
{
	if(bFirstLayerFlag==TRUE)
	{
		load(WVT_HL);
		load(WVT_LH);
		load(WVT_HH);
	}

	quant(iQuant,WVT_HL);
	quant(iQuant,WVT_LH);
	quant(iQuant,WVT_HH);

	invQuant(iQuant,WVT_HL,bReconFlag,bFirstLayerFlag);
	invQuant(iQuant,WVT_LH,bReconFlag,bFirstLayerFlag);
	invQuant(iQuant,WVT_HH,bReconFlag,bFirstLayerFlag);

	// recurse
	if(getChild()!=NULL)
		getChild()->makeSNRLayer(iQuant,bReconFlag,bFirstLayerFlag);
}

// add residual back into original decomp tree
void CWvtQuantTree::restoreDecompTree(FilterBank band)
{
	Int i;
	
	const PixelI *ppxliSrc = m_piiResidual[band]->pixels();
	PixelF *ppxlfDst = (PixelF *) m_pwvtDecomp->data(band)->pixels();
	for(i=getArea();i;i--)
		*ppxlfDst++ += (PixelF)*ppxliSrc++;

	// recurse
	if(getChild()!=NULL)
		getChild()->restoreDecompTree(band);
}

void CWvtQuantTree::makeZeroTree(FilterBank band,Bool bInitMapFlag,
								 Int *piSymsValz,Int *piSymsValnz)
{
	U8 *pucMap = m_pucMap[band];
	const PixelI *ppxliQuant = m_piiQuant[band]->pixels();

	if(getParent()==NULL)
	{
		// leaf
		Int i;
		for(i=getArea();i;i--,pucMap++,ppxliQuant++)
		{
			// will transmit no type, just value with Valz table
			if(*ppxliQuant==0)
				*pucMap=ZT_ZTR;
			else
				*pucMap=ZT_VZTR;
			if(*ppxliQuant>=*piSymsValz)
				*piSymsValz=*ppxliQuant+1;
		}
	}
	else
	{
		U8 *pucMapPrev = getParent()->m_pucMap[band];
		Int iWidth = getWidth();
		Int iHeight = getHeight();
		Int iWidthPrev = iWidth<<1;
		Int iX,iY;
		for(iY=0;iY<iHeight;iY++,pucMapPrev+=iWidthPrev)
			for(iX=0;iX<iWidth;iX++,pucMap++,ppxliQuant++,pucMapPrev+=2)
			{
				if(bInitMapFlag==FALSE && (*pucMap==ZT_VAL || *pucMap==ZT_OLD))
				{
					// will transmit no type, just value with Valz table
					*pucMap=ZT_OLD;
					if(*ppxliQuant>=*piSymsValz)
						*piSymsValz=*ppxliQuant+1;
				}
				else if(pucMapPrev[0]==ZT_ZTR && pucMapPrev[1]==ZT_ZTR
					&& pucMapPrev[iWidthPrev]==ZT_ZTR
					&& pucMapPrev[iWidthPrev+1]==ZT_ZTR)
				{
					// mark previous layer for ignoring
					pucMapPrev[0]=pucMapPrev[1]=ZT_IGN;
					pucMapPrev[iWidthPrev]=pucMapPrev[iWidthPrev+1]=ZT_IGN;
					if(*ppxliQuant==0)
						// will transmit ZT_ZTR and no value
						*pucMap=ZT_ZTR;
					else
					{
						// will transmit ZT_VZTR and value with Valnz table
						*pucMap=ZT_VZTR;
						if(*ppxliQuant>*piSymsValnz)
							*piSymsValnz=*ppxliQuant;
					}
				}
				else
				{
					// will transmit ZT_VAL and value with Valz table
					*pucMap=ZT_VAL;
					if(*ppxliQuant>=*piSymsValz)
						*piSymsValz=*ppxliQuant+1;
				}
			}
	}

	if(getChild()!=NULL)
		getChild()->makeZeroTree(band,bInitMapFlag,piSymsValz,piSymsValnz);
}

Int CWvtQuantTree::codeZeroTreeNode(WvtArithCoderSet *pacSet,Int iMapVal,
									 Int iQuantVal)
{
	CWvtArithCoder *pacCoder = pacSet->pacCoder;
	Int iSymsCoded = 0;

	if(iMapVal==ZT_ZTR)
	{
		pacCoder->encodeSymbol(*(pacSet->pacModelType),ACSYM_ZTR);
		iSymsCoded=1;
	}
	else if(iMapVal==ZT_VZTR)
	{
		pacCoder->encodeSymbol(*(pacSet->pacModelType),ACSYM_VZTR);
		pacCoder->encodeSymbol(*(pacSet->pacModelValnz),iQuantVal-1);
		iSymsCoded=2;
	}
	else if(iMapVal==ZT_OLD)
	{
		pacCoder->encodeSymbol(*(pacSet->pacModelValz),iQuantVal);
		iSymsCoded=1;
	}
	else if(iMapVal==ZT_VAL)
	{
		pacCoder->encodeSymbol(*(pacSet->pacModelType),ACSYM_VAL);
		pacCoder->encodeSymbol(*(pacSet->pacModelValz),iQuantVal);
		iSymsCoded=2;
	}
	return iSymsCoded;
}

Int CWvtQuantTree::codeZeroTreeVert(WvtArithCoderSet *pacSet,FilterBank band,
									 Int iX,Int iY)
{
	Int iOffset = iX + iY*getWidth();
	U8 *pucMap = m_pucMap[band] + iOffset;
	const PixelI *ppxliQuant = m_piiQuant[band]->pixels() + iOffset;
	CWvtQuantTree *pwvtqtParent = getParent();
	CWvtArithCoder *pacCoder = pacSet->pacCoder;
	Int iSymsCoded = 0;

	if(pwvtqtParent==NULL)
	{
		// leaf
		pacCoder->encodeSymbol(*(pacSet->pacModelValz),*ppxliQuant);
		iSymsCoded++;
	}
	else
	{
		iSymsCoded += codeZeroTreeNode(pacSet,(Int)*pucMap,(Int)*ppxliQuant);

		if(*pucMap==ZT_VAL || *pucMap==ZT_OLD)
		{
			Int iXX = iX<<1;
			Int iYY = iY<<1;
			iSymsCoded += pwvtqtParent->codeZeroTreeVert(pacSet,band,iXX,iYY);
			iSymsCoded += pwvtqtParent->codeZeroTreeVert(pacSet,band,iXX+1,iYY);
			iSymsCoded += pwvtqtParent->codeZeroTreeVert(pacSet,band,iXX,iYY+1);
			iSymsCoded += pwvtqtParent->codeZeroTreeVert(pacSet,band,iXX+1,iYY+1);
		}
	}
	return iSymsCoded;
}

void CWvtQuantTree::codeSingleQ(CWvtCoderChannel *pCoderChan,Int iSpatialLayer,
								Int iSNRLayer,CWvtCodedStream *pwvtCodedStream,
								CIntImage *piiSave)
{
	// zero tree is coded depth first coefficient by coefficient
	// including all layers valid for the current spatial scalability level
#if(DEBUG_OPS)
	printf("  Single Quantiser, spatial level = %d\n",iSpatialLayer);
#endif

	Int iQuant = pCoderChan->getSNRACQuant(iSpatialLayer,iSNRLayer);
	Bool bReconFlag = (pCoderChan->getSpatialSave()==iSpatialLayer)?TRUE:FALSE;

	// set scalability limit
	CWvtQuantTree *pwvtqtParentSave = m_pParent;
	m_pParent=NULL;

	// quantise and put into wvt decomp tree if recon flag set
	makeSNRLayer(iQuant,bReconFlag,TRUE);

	if(bReconFlag)
	{
		// inverse transform the quantised wavelet tree

		//m_pwvtDecomp->printDecompTree();
		m_pwvtDecomp->inverseTransform(piiSave);
		
		// restore original non quantised tree
		restoreDecompTree(WVT_HL);
		restoreDecompTree(WVT_LH);
		restoreDecompTree(WVT_HH);
	}

	Int iSymsValz = 2;
	Int iSymsValnz = 2;
	makeZeroTree(WVT_HL,TRUE,&iSymsValz,&iSymsValnz);
	makeZeroTree(WVT_LH,TRUE,&iSymsValz,&iSymsValnz);
	makeZeroTree(WVT_HH,TRUE,&iSymsValz,&iSymsValnz);

#if(DEBUG_OPS)
	printf("    Symbol table sizes: type = %d, valz = %d, valnz = %d\n",
		NSYMS,iSymsValz,iSymsValnz);
#endif

	CWvtArithCoderModel acModelType(AC_ADAPTIVE,NSYMS); // 0=VAL, 1=VZTR, 2=ZTR
	CWvtArithCoderModel acModelValz(AC_ADAPTIVE,iSymsValz); // 0=0, 1=1 etc
	CWvtArithCoderModel acModelValnz(AC_ADAPTIVE,iSymsValnz); // 0=1, 1=2 etc
	
	CWvtArithCoder *pacCoder = new CWvtArithCoder();

	WvtArithCoderSet acSet;
	acSet.pacCoder = pacCoder;
	acSet.pacModelType = &acModelType;
	acSet.pacModelValz = &acModelValz;
	acSet.pacModelValnz = &acModelValnz;

	Int iBitOverhead = 0;
	pacCoder->encodeStart();
	iBitOverhead += pacCoder->encodeVLCValue(iSymsValz,NUMBITS_AC_NUMSYMS);
	iBitOverhead += pacCoder->encodeVLCValue(iSymsValnz,NUMBITS_AC_NUMSYMS);

	CWvtQuantTree *pwvtqtRoot = getRoot();
	Int iWidth = pwvtqtRoot->getWidth();
	Int iHeight = pwvtqtRoot->getHeight();
	Int iX,iY;
	Int iSymsCoded = 0;

	for(iY=0;iY<iHeight;iY++)
		for(iX=0;iX<iWidth;iX++)
		{
			iSymsCoded += pwvtqtRoot->codeZeroTreeVert(&acSet,WVT_HL,iX,iY);
			iSymsCoded += pwvtqtRoot->codeZeroTreeVert(&acSet,WVT_LH,iX,iY);
			iSymsCoded += pwvtqtRoot->codeZeroTreeVert(&acSet,WVT_HH,iX,iY);
		}

	pacCoder->encodeEnd();

	m_pParent=pwvtqtParentSave;

#if(DEBUG_OPS)
	printf("    Total symbols = %d, Total bits = %d (overhead = %d)\n",
		iSymsCoded,pacCoder->getBitCount(),iBitOverhead);
#endif

	pwvtCodedStream->attachACStream(iSpatialLayer,pacCoder);
}


Int CWvtQuantTree::codeZeroTreeHoriz(WvtArithCoderSet *pacSet)
{
	Int i;

	U8 *pucMapHL = m_pucMap[WVT_HL];
	U8 *pucMapLH = m_pucMap[WVT_LH];
	U8 *pucMapHH = m_pucMap[WVT_HH];
	const PixelI *ppxliQuantHL = m_piiQuant[WVT_HL]->pixels();
	const PixelI *ppxliQuantLH = m_piiQuant[WVT_LH]->pixels();
	const PixelI *ppxliQuantHH = m_piiQuant[WVT_HH]->pixels();
	CWvtArithCoder *pacCoder = pacSet->pacCoder;
	Int iSymsCoded = 0;

	if(getParent()==NULL)
	{
		// leaf
		for(i=getArea();i;i--,ppxliQuantHL++,ppxliQuantLH++,ppxliQuantHH++,
			pucMapHL++,pucMapLH++,pucMapHH++)
		{
			if(*pucMapHL!=ZT_IGN)
			{
				pacCoder->encodeSymbol(*(pacSet->pacModelValz),*ppxliQuantHL);
				iSymsCoded++;
			}
			if(*pucMapLH!=ZT_IGN)
			{
				pacCoder->encodeSymbol(*(pacSet->pacModelValz),*ppxliQuantLH);
				iSymsCoded++;
			}
			if(*pucMapHH!=ZT_IGN)
			{
				pacCoder->encodeSymbol(*(pacSet->pacModelValz),*ppxliQuantHH);
				iSymsCoded++;
			}
		}
	}
	else
	{
		for(i=getArea();i;i--,ppxliQuantHL++,ppxliQuantLH++,ppxliQuantHH++,
			pucMapHL++,pucMapLH++,pucMapHH++)
		{
			iSymsCoded += codeZeroTreeNode(pacSet,(Int)*pucMapHL,(Int)*ppxliQuantHL);
			iSymsCoded += codeZeroTreeNode(pacSet,(Int)*pucMapLH,(Int)*ppxliQuantLH);
			iSymsCoded += codeZeroTreeNode(pacSet,(Int)*pucMapHH,(Int)*ppxliQuantHH);
		}

		iSymsCoded += getParent()->codeZeroTreeHoriz(pacSet);
	}
	return iSymsCoded;
}

void CWvtQuantTree::codeMultipleQ(CWvtCoderChannel *pCoderChan,Int iSpatialLayer,
								  CWvtCodedStream *pwvtCodedStream,
								  CIntImage *piiSave)
{
	// zero tree is coded spatial layer by spatial layer starting with
	// the low freq bands, all relevant decomp bands are sent for each spatial
	// scalability level

	Int iSnr;
	Bool bInitMapFlag;
	Int iSymsValz;
	Int iSymsValnz;
	WvtArithCoderSet acSet;
	Int iSNRLevels = pCoderChan->getSNRLevels(iSpatialLayer);
	Bool bReconFlag = (pCoderChan->getSpatialSave()==iSpatialLayer)?TRUE:FALSE;
	Int iSnrSave = iSNRLevels - pCoderChan->getSNRSave() - 1;

	// set scalability limit
	CWvtQuantTree *pwvtqtParentSave = m_pParent;
	m_pParent=NULL;

	bInitMapFlag=TRUE;
	for(iSnr=0;iSnr<iSNRLevels;iSnr++)
	{
#if(DEBUG_OPS)
		printf("  Multiple Quantiser, spatial level = %d, snr level = %d\n",
			iSpatialLayer,iSnr);
#endif

		// quantise and put into wvt decomp tree if recon flag set
		makeSNRLayer(pCoderChan->getSNRACQuant(iSpatialLayer,iSnr),bReconFlag,
			bInitMapFlag);
		
		// if saving, inverse transform the quantised wavelet tree
		if(bReconFlag && (iSnr==iSnrSave))
			m_pwvtDecomp->inverseTransform(piiSave);

		// make zero trees and work out how big the codebook needs to be
		iSymsValz = 2;
		iSymsValnz = 2;
		makeZeroTree(WVT_HL,bInitMapFlag,&iSymsValz,&iSymsValnz);
		makeZeroTree(WVT_LH,bInitMapFlag,&iSymsValz,&iSymsValnz);
		makeZeroTree(WVT_HH,bInitMapFlag,&iSymsValz,&iSymsValnz);

#if(DEBUG_OPS)
		printf("    Symbol table sizes: type = %d, valz = %d, valnz = %d\n",
			NSYMS,iSymsValz,iSymsValnz);
#endif

		CWvtArithCoderModel acModelType(AC_ADAPTIVE,NSYMS); // 0=VAL, 1=VZTR, 2=ZTR
		CWvtArithCoderModel acModelValz(AC_ADAPTIVE,iSymsValz); // 0=0, 1=1 etc
		CWvtArithCoderModel acModelValnz(AC_ADAPTIVE,iSymsValnz); // 0=1, 1=2 etc

		acSet.pacCoder = new CWvtArithCoder();
		acSet.pacModelType = &acModelType;
		acSet.pacModelValz = &acModelValz;
		acSet.pacModelValnz = &acModelValnz;

		// encode the alphabet size
		Int iBitOverhead = 0;
		acSet.pacCoder->encodeStart();
		iBitOverhead += acSet.pacCoder->encodeVLCValue(iSymsValz,NUMBITS_AC_NUMSYMS);
		iBitOverhead += acSet.pacCoder->encodeVLCValue(iSymsValnz,NUMBITS_AC_NUMSYMS);

		// encode the tree
		Int iSymsCoded = getRoot()->codeZeroTreeHoriz(&acSet);

		acSet.pacCoder->encodeEnd();

		pwvtCodedStream->attachACStream(iSpatialLayer,acSet.pacCoder);

#if(DEBUG_OPS)
		printf("    Total symbols = %d, Total bits = %d (overhead = %d)\n",
			iSymsCoded,acSet.pacCoder->getBitCount(),iBitOverhead);
#endif

		bInitMapFlag=FALSE;
	}

	if(bReconFlag)
	{
		// restore original non quantised tree
		restoreDecompTree(WVT_HL);
		restoreDecompTree(WVT_LH);
		restoreDecompTree(WVT_HH);
	}

	m_pParent=pwvtqtParentSave;
}

Int CWvtQuantTree::loadGetMax()
{
	Int i,iMax=0;
	const PixelI *ppxliHL = m_piiResidual[WVT_HL]->pixels();
	const PixelI *ppxliLH = m_piiResidual[WVT_LH]->pixels();
	const PixelI *ppxliHH = m_piiResidual[WVT_HH]->pixels();
	
	load(WVT_LH);
	load(WVT_HL);
	load(WVT_HH);

	for(i=getArea();i;i--,ppxliHL++,ppxliLH++,ppxliHH++)
	{
		if(*ppxliHL>iMax)
			iMax = *ppxliHL;
		if(-*ppxliHL>iMax)
			iMax = -*ppxliHL;
		if(*ppxliLH>iMax)
			iMax = *ppxliLH;
		if(-*ppxliLH>iMax)
			iMax = -*ppxliLH;
		if(*ppxliHH>iMax)
			iMax = *ppxliHH;
		if(-*ppxliHH>iMax)
			iMax = -*ppxliHH;
	}
	
	if(getChild()!=NULL)
	{
		Int iChildVal = getChild()->loadGetMax();
		if(iChildVal>iMax)
			iMax=iChildVal;
	}
	return iMax;
}

void CWvtQuantTree::threshBiLevel(FilterBank band,Int iThresh,Bool bInitMapFlag)
{
	Int i;
	const PixelI *ppxli = m_piiResidual[band]->pixels();
	U8 *pucMap = m_pucMap[band];
	Int iCount=0;

	if(bInitMapFlag==FALSE)
	{
		// mark as VAL if newly above threshold
		for(i=getArea();i;i--,ppxli++,pucMap++)
			if(*pucMap!=ZT_OLD)
			{
				if(*ppxli>=iThresh || (-*ppxli)>=iThresh)
				{*pucMap=ZT_VAL;iCount++;}
				else
					*pucMap=ZT_ZTR;
			}
	}
	else
	{
		PixelI *ppxliQuant = (PixelI *)m_piiQuant[band]->pixels();

		// mark as VAL if above threshold
		for(i=getArea();i;i--,ppxli++,pucMap++,ppxliQuant++)
		{
			*ppxliQuant=0;
			if(*ppxli>=iThresh || (-*ppxli)>=iThresh)
			{*pucMap=ZT_VAL;iCount++;}
			else
				*pucMap=ZT_ZTR;
		}
	}

	if(getChild()!=NULL)
		getChild()->threshBiLevel(band,iThresh,bInitMapFlag);
}

void CWvtQuantTree::makeZeroTreeBiLevel(FilterBank band)
{
	U8 *pucMap = m_pucMap[band];
	Int iWidth = getWidth();
	Int iHeight = getHeight();
	Int iX,iY;

	if(getParent()!=NULL)
	{
		U8 *pucPrevMap = getParent()->m_pucMap[band];

		Int iPrevWidth = iWidth<<1;

		// not leaf so check for zero trees
		for(iY=0;iY<iHeight;iY++,pucPrevMap+=iPrevWidth)
			for(iX=0;iX<iWidth;iX++,pucMap++,pucPrevMap+=2)
				if(*pucMap==ZT_ZTR) // not a value
				{
					if( (pucPrevMap[0]==ZT_ZTR ||
							pucPrevMap[0]==ZT_OLD) &&
						(pucPrevMap[1]==ZT_ZTR ||
							pucPrevMap[1]==ZT_OLD) &&
						(pucPrevMap[iPrevWidth]==ZT_ZTR ||
							pucPrevMap[iPrevWidth]==ZT_OLD) &&
						(pucPrevMap[iPrevWidth+1]==ZT_ZTR ||
							pucPrevMap[iPrevWidth+1]==ZT_OLD) )
					{ // zero tree
						*pucMap=ZT_ZTR;
						// ignore lower zero trees
						if(pucPrevMap[0]==ZT_ZTR)
							pucPrevMap[0]=ZT_IGN;
						if(pucPrevMap[1]==ZT_ZTR)
							pucPrevMap[1]=ZT_IGN;
						if(pucPrevMap[iPrevWidth]==ZT_ZTR)
							pucPrevMap[iPrevWidth]=ZT_IGN;
						if(pucPrevMap[iPrevWidth+1]==ZT_ZTR)
							pucPrevMap[iPrevWidth+1]=ZT_IGN;
					}
					else
						*pucMap=ZT_Z; // isolated zero
				}
	}

	if(getChild()!=NULL)
		getChild()->makeZeroTreeBiLevel(band);
}

Int CWvtQuantTree::codeZeroTreeBiLevel(Int iThresh,WvtBiLevelACSet *pacSet)
{
	Int iSymsCoded = 0;

	if(getChild()==NULL)		// root
		pacSet->iCurrCoder=0;	// initialise first time
	else if(pacSet->iNumCoders>1)
		pacSet->iCurrCoder++;	// increment if multi spatial

	if(getParent()==NULL)
		pacSet->pacModel->loadModel(rgiFreqTableLeaf);	// leaf
	else if(getChild()==NULL || pacSet->iNumCoders>1)
		pacSet->pacModel->loadModel(rgiFreqTableCoef);	// root or multi spatial

	iSymsCoded += codeZeroTreeBandBiLevel(WVT_LH,iThresh,pacSet);
	iSymsCoded += codeZeroTreeBandBiLevel(WVT_HL,iThresh,pacSet);
	iSymsCoded += codeZeroTreeBandBiLevel(WVT_HH,iThresh,pacSet);

	if(getParent()!=NULL)
		iSymsCoded += getParent()->codeZeroTreeBiLevel(iThresh,pacSet);

	return iSymsCoded;
}

Int CWvtQuantTree::codeZeroTreeBandBiLevel(FilterBank band,Int iThresh,
												WvtBiLevelACSet *pacSet)
{
	Int i;
	const PixelI *ppxliInput = m_piiResidual[band]->pixels();
	U8 *pucMap = m_pucMap[band];
	PixelI *ppxliQuant = (PixelI *)m_piiQuant[band]->pixels();
	CWvtArithCoderModel *pacModel = pacSet->pacModel;
	CWvtArithCoder *pacCoder = pacSet->ppacCoder[pacSet->iCurrCoder];
	Int iSymsCoded = 0;

	for(i=getArea();i;i--,ppxliInput++,pucMap++,ppxliQuant++)
	{
		if(*pucMap==ZT_VAL)
		{
			if(*ppxliInput>0)
			{
				// code POS
				*ppxliQuant+=(iThresh+(iThresh>>1));
				pacCoder->encodeSymbol(*pacModel,ACSYMBL_POS);
				iSymsCoded++;
			}
			else
			{
				// code NEG
				*ppxliQuant-=(iThresh+(iThresh>>1));
				pacCoder->encodeSymbol(*pacModel,ACSYMBL_NEG);
				iSymsCoded++;
			}
			*pucMap=ZT_OLD;
		}
		else if(*pucMap==ZT_ZTR)
		{
			// code ZTR
			pacCoder->encodeSymbol(*pacModel,ACSYMBL_ZTR);
			iSymsCoded++;
		}
		else if(*pucMap==ZT_Z)
		{
			// code IZ
			pacCoder->encodeSymbol(*pacModel,ACSYMBL_IZ);
			iSymsCoded++;
		}
	}

	return iSymsCoded;
}

Int CWvtQuantTree::codeZeroTreeRefineBiLevel(Int iThresh,WvtBiLevelACSet *pacSet)
{
	Int iSymsCoded = 0;

	if(getChild()==NULL)		// root
		pacSet->iCurrCoder=0;	// initialise first time
	else if(pacSet->iNumCoders>1)
		pacSet->iCurrCoder++;	// increment if multi spatial
	
	if(getChild()==NULL || pacSet->iNumCoders>1) // root or multi spatial
		pacSet->pacModel->loadModel(rgiFreqTableRefine);

	iSymsCoded += codeZeroTreeRefineBandBiLevel(WVT_LH,iThresh,pacSet);
	iSymsCoded += codeZeroTreeRefineBandBiLevel(WVT_HL,iThresh,pacSet);
	iSymsCoded += codeZeroTreeRefineBandBiLevel(WVT_HH,iThresh,pacSet);
	
	if(getParent()!=NULL)
		iSymsCoded += getParent()->codeZeroTreeRefineBiLevel(iThresh,pacSet);
	
	return iSymsCoded;
}

Int CWvtQuantTree::codeZeroTreeRefineBandBiLevel(FilterBank band,Int iThresh,
												  WvtBiLevelACSet *pacSet)
{
	Int i,iSymbol;
	const PixelI *ppxliInput = m_piiResidual[band]->pixels();
	PixelI *ppxliQuant = (PixelI *)m_piiQuant[band]->pixels();
	CWvtArithCoderModel *pacModel = pacSet->pacModel;
	CWvtArithCoder *pacCoder = pacSet->ppacCoder[pacSet->iCurrCoder];
	Int iSymsCoded = 0;

	for(i=getArea();i;i--,ppxliInput++,ppxliQuant++)
	{
		if(*ppxliQuant>0)
		{
			if(*ppxliInput>=*ppxliQuant)
			{
				*ppxliQuant += iThresh;
				iSymbol=ACSYMBL_INC;
			}
			else
			{
				*ppxliQuant -= iThresh;
				iSymbol=ACSYMBL_DEC;
			}
			pacCoder->encodeSymbol(*pacModel,iSymbol);
			iSymsCoded++;
		}
		else if(*ppxliQuant<0)
		{
			if(*ppxliInput<=*ppxliQuant)
			{
				*ppxliQuant -= iThresh;
				iSymbol=ACSYMBL_INC;
			}
			else
			{
				*ppxliQuant += iThresh;
				iSymbol=ACSYMBL_DEC;
			}
			pacCoder->encodeSymbol(*pacModel,iSymbol);
			iSymsCoded++;
		}
	}

	return iSymsCoded;
}

void CWvtQuantTree::codeBiLevelQ(CWvtCoderChannel *pCoderChan,Bool bEnableSpatial,
								 Bool bEnableSnr,CWvtCodedStream *pwvtCodedStream,
								 CIntImage *piiSave)
{
	// zero tree is coded from low to high decomp bands
	// for "no spatial scalability", these bands are transmitted together
	// for each snr level
	// for "spatial scalability", each snr set for a spatial band is saved in
	// a different bitstream and these are later put together in sequence
	// for "snr scalability", snr layers are also broken up

	CWvtQuantTree *pwvtqtParentSave = m_pParent;
	m_pParent=NULL;

	Int iSpatialLevels = 1;
	CWvtQuantTree *pwvtqt = this;
	while((pwvtqt = pwvtqt->getChild())!=NULL)
		iSpatialLevels++;
	
	Int iNumCoders;
	if(bEnableSpatial==TRUE)
		iNumCoders=iSpatialLevels;
	else
		iNumCoders=1;

	pwvtCodedStream->setSpatialLevels(iNumCoders);

	Int iMax = loadGetMax();

	Int iMinThresh = pCoderChan->getSNRACQuant(0,0);
	Int iThresh = iMinThresh;
	
	/*if(iMax<iThresh)
		{assert(FALSE);} // to be fixed
	*/

	while(iMax >= (iThresh<<1))
		iThresh<<=1;

	WvtBiLevelACSet acSet;
	CWvtArithCoderModel acModel(AC_ADAPTIVE,NSYMSBL);
	acSet.pacModel = &acModel;
	acSet.iNumCoders = iNumCoders;
	acSet.ppacCoder = new CWvtArithCoder * [iNumCoders];
	
	Int i;
	Bool bInitPassFlag=TRUE;
	Int iSNRLevel = 0;
	Int iSymsCoded;

	pwvtCodedStream->setQuant(iThresh);

#if(DEBUG_OPS)
	printf("  Bi-level Quantiser (");
	if(iNumCoders>1)
		printf("separate spatial streams, ");
	else
		printf("combined spatial streams, ");
	if(bEnableSnr)
		printf("separate snr streams)\n");
	else
		printf("combined snr streams)\n");
#endif

	while(TRUE)
	{
#if(DEBUG_OPS)		
		printf("  Snr level = %d\n",iSNRLevel);
#endif
		iSNRLevel++;

		// find new coefficients above threshold
		threshBiLevel(WVT_LH,iThresh,bInitPassFlag);
		threshBiLevel(WVT_HL,iThresh,bInitPassFlag);
		threshBiLevel(WVT_HH,iThresh,bInitPassFlag);

		// make zero tree
		makeZeroTreeBiLevel(WVT_LH);
		makeZeroTreeBiLevel(WVT_HL);
		makeZeroTreeBiLevel(WVT_HH);

		// set up ac coder(s)
		if(bEnableSnr==TRUE || bInitPassFlag==TRUE)
			for(i=0;i<acSet.iNumCoders;i++)
			{
				acSet.ppacCoder[i] = new CWvtArithCoder;
				acSet.ppacCoder[i]->encodeStart();
				pwvtCodedStream->attachACStream(i,acSet.ppacCoder[i]);
			}

		// code zero tree
		iSymsCoded = getRoot()->codeZeroTreeBiLevel(iThresh,&acSet);
#if(DEBUG_OPS)
		printf("    Main pass: threshold = %d, total symbols = %d\n",
			iThresh,iSymsCoded);
#endif

		// get ready for "continue" or "eof"
		acModel.loadModel(rgiFreqTableEnd);

		iThresh>>=1;
		if(iThresh<iMinThresh)  // minthresh is >=1
			break;

		// code "continue"
		for(i=0;i<acSet.iNumCoders;i++)
			acSet.ppacCoder[i]->encodeSymbol(acModel,ACSYMBL_CNT);
	
		// refine previous values
		iSymsCoded = getRoot()->codeZeroTreeRefineBiLevel(iThresh>>1,&acSet);
#if(DEBUG_OPS)
		printf("    Refinement pass: threshold = %d, total symbols = %d\n",
			iThresh,iSymsCoded);
#endif

		// if multiple snr mode, tidy up ac coder streams
		if(bEnableSnr==TRUE)
			for(i=0;i<acSet.iNumCoders;i++)
			{
				acSet.ppacCoder[i]->encodeEnd();
#if(DEBUG_OPS)
				printf("  Total bits (spatial level %d, snr level %d) = %d\n",
					i,iSNRLevel-1,acSet.ppacCoder[i]->getBitCount());
#endif
			}

		bInitPassFlag=FALSE;
	}

	// tidy up ac coder streams
	for(i=0;i<acSet.iNumCoders;i++)
	{
		acSet.ppacCoder[i]->encodeSymbol(acModel,ACSYMBL_EOF);
		acSet.ppacCoder[i]->encodeEnd();
#if(DEBUG_OPS)
		printf("  Total bits (spatial level %d, snr level %d) = %d\n",
			i,iSNRLevel-1,acSet.ppacCoder[i]->getBitCount());
#endif
	}

	// save image if needed
	if(piiSave!=NULL)
	{
		copyQuantToWvt(WVT_LH);
		copyQuantToWvt(WVT_HL);
		copyQuantToWvt(WVT_HH);
		m_pwvtDecomp->inverseTransform(piiSave);
		copyResidualToWvt(WVT_LH);
		copyResidualToWvt(WVT_HL);
		copyResidualToWvt(WVT_HH);
	}

	delete acSet.ppacCoder;

	m_pParent=pwvtqtParentSave;
}

void CWvtQuantTree::copyQuantToWvt(FilterBank band)
{
	Int i;
	PixelF *ppxlfDst = (PixelF *)m_pwvtDecomp->data(band)->pixels();
	const PixelI *ppxliSrc = m_piiQuant[band]->pixels();
	for(i=getArea();i;i--)
		*ppxlfDst++ = (PixelF)*ppxliSrc++;

	if(getChild()!=NULL)
		getChild()->copyQuantToWvt(band);
}

void CWvtQuantTree::zeroQuant(FilterBank band)
{
	Int i;
	PixelI *ppxliDst = (PixelI *)m_piiQuant[band]->pixels();
	for(i=getArea();i;i--)
		*ppxliDst++ = 0;

	if(getChild()!=NULL)
		getChild()->zeroQuant(band);
}

void CWvtQuantTree::initMap(FilterBank band)
{
	Int i;
	U8 *pucMap = m_pucMap[band];

	for(i=getArea();i;i--)
		*pucMap++ = ZT_XXX;

	if(getChild()!=NULL)
		getChild()->initMap(band);
}

void CWvtQuantTree::copyResidualToWvt(FilterBank band)
{
	Int i;
	PixelF *ppxlfDst = (PixelF *)m_pwvtDecomp->data(band)->pixels();
	const PixelI *ppxliSrc = m_piiResidual[band]->pixels();
	for(i=getArea();i;i--)
		*ppxlfDst++ = (PixelF)*ppxliSrc++;

	if(getChild()!=NULL)
		getChild()->copyResidualToWvt(band);
}

void CWvtQuantTree::decodeZeroTreeVert(WvtArithCoderSet *pacSet,FilterBank band,
									   Int iX,Int iY,Bool bIgnoreMap,Int iQuant)
{
	Int iOffset = iX + iY*getWidth();
	U8 *pucMap = m_pucMap[band] + iOffset;
	PixelI *ppxliQuant = (PixelI *)m_piiQuant[band]->pixels() + iOffset;
	CWvtQuantTree *pwvtqtParent = getParent();
	CWvtArithCoder *pacCoder = pacSet->pacCoder;
	Int iVal,iXX,iYY;

	if(pwvtqtParent==NULL || (bIgnoreMap==FALSE && *pucMap==ZT_VAL))
	{
		// leaf or previously known to be VAL
		iVal = pacCoder->decodeSymbol(*(pacSet->pacModelValz));
		iVal = InvACReorder(iVal);
		*ppxliQuant += InvACQuant(iVal,iQuant);
	}
	else
	{
		switch(pacCoder->decodeSymbol(*(pacSet->pacModelType)))
		{
		case ACSYM_VAL:
			*pucMap=ZT_VAL;
			iVal = pacCoder->decodeSymbol(*(pacSet->pacModelValz));
			iVal = InvACReorder(iVal);
			*ppxliQuant += InvACQuant(iVal,iQuant);

			// also do the children
			iXX = iX<<1;
			iYY = iY<<1;
			pwvtqtParent->decodeZeroTreeVert(pacSet,band,iXX,iYY,bIgnoreMap,
				iQuant);
			pwvtqtParent->decodeZeroTreeVert(pacSet,band,iXX+1,iYY,bIgnoreMap,
				iQuant);
			pwvtqtParent->decodeZeroTreeVert(pacSet,band,iXX,iYY+1,bIgnoreMap,
				iQuant);
			pwvtqtParent->decodeZeroTreeVert(pacSet,band,iXX+1,iYY+1,bIgnoreMap,
				iQuant);
			break;

		case ACSYM_ZTR:
			*pucMap = ZT_ZTR;
			break;

		case ACSYM_VZTR:
			*pucMap = ZT_VZTR;
			iVal = pacCoder->decodeSymbol(*(pacSet->pacModelValnz))+1;
			iVal = InvACReorder(iVal);
			*ppxliQuant += InvACQuant(iVal,iQuant);
			break;

		default:
			assert(FALSE); // error condition
			break;
		}
	}
}

void CWvtQuantTree::decodeSingleQ(CInBitStream &bs,Int iLength,Int iQuant,
								  CIntImage *piiSave)
{
	CWvtArithCoder acCoder;
	WvtArithCoderSet acSet;

	CWvtQuantTree *pwvtqtParentSave = m_pParent;
	m_pParent=NULL;

	acCoder.decodeStart(bs,iLength);	
	
	Int iSymsValz = acCoder.decodeVLCValue(NUMBITS_AC_NUMSYMS);
	Int iSymsValnz = acCoder.decodeVLCValue(NUMBITS_AC_NUMSYMS);
#if(DEBUG_DECODE)
	printf("Decoding\n");
#endif
#if(DEBUG_BITS)
	printf("   Numsyms valz\t\t= %d\n",iSymsValz);
	printf("   Numsyms valnz\t= %d\n",iSymsValnz);
#endif
	CWvtArithCoderModel acModelType(AC_ADAPTIVE,NSYMS); // 0=VAL, 1=VZTR, 2=ZTR
	CWvtArithCoderModel acModelValz(AC_ADAPTIVE,iSymsValz); // 0=0, 1=1 etc
	CWvtArithCoderModel acModelValnz(AC_ADAPTIVE,iSymsValnz); // 0=1, 1=2 etc
	
	acSet.pacCoder = &acCoder;
	acSet.pacModelType = &acModelType;
	acSet.pacModelValz = &acModelValz;
	acSet.pacModelValnz = &acModelValnz;

	zeroQuant(WVT_HL);
	zeroQuant(WVT_LH);
	zeroQuant(WVT_HH);

	CWvtQuantTree *pwvtqtRoot = getRoot();
	Int iWidth = pwvtqtRoot->getWidth();
	Int iHeight = pwvtqtRoot->getHeight();
	Int iX,iY;

	// decodes tree and adds to Quant tree
	for(iY=0;iY<iHeight;iY++)
		for(iX=0;iX<iWidth;iX++)
		{
			pwvtqtRoot->decodeZeroTreeVert(&acSet,WVT_HL,iX,iY,TRUE,iQuant);
			pwvtqtRoot->decodeZeroTreeVert(&acSet,WVT_LH,iX,iY,TRUE,iQuant);
			pwvtqtRoot->decodeZeroTreeVert(&acSet,WVT_HH,iX,iY,TRUE,iQuant);
		}

	acCoder.decodeEnd();

	if(piiSave!=NULL)
	{
		copyQuantToWvt(WVT_HL);
		copyQuantToWvt(WVT_LH);
		copyQuantToWvt(WVT_HH);

		// inverse transform the wavelet tree
		//m_pwvtDecomp->printDecompTree();
		m_pwvtDecomp->inverseTransform(piiSave);
	}

	m_pParent=pwvtqtParentSave;
}

void CWvtQuantTree::decodeZeroTreeHoriz(WvtArithCoderSet *pacSet,Int iQuant)
{
	Int i,iVal;
	U8 *pucMapHL = m_pucMap[WVT_HL];
	U8 *pucMapLH = m_pucMap[WVT_LH];
	U8 *pucMapHH = m_pucMap[WVT_HH];
	PixelI *ppxliQuantHL = (PixelI *)m_piiQuant[WVT_HL]->pixels();
	PixelI *ppxliQuantLH = (PixelI *)m_piiQuant[WVT_LH]->pixels();
	PixelI *ppxliQuantHH = (PixelI *)m_piiQuant[WVT_HH]->pixels();
	CWvtArithCoder *pacCoder = pacSet->pacCoder;

	if(getParent()==NULL)
	{
		for(i=getArea();i;i--,ppxliQuantHL++,ppxliQuantLH++,ppxliQuantHH++,
			pucMapHL++,pucMapLH++,pucMapHH++)
		{
			if(*pucMapHL!=ZT_IGN)
			{
				iVal = pacCoder->decodeSymbol(*(pacSet->pacModelValz));
				iVal = InvACReorder(iVal);
				*ppxliQuantHL += InvACQuant(iVal,iQuant);
			}

			if(*pucMapLH!=ZT_IGN)
			{
				iVal = pacCoder->decodeSymbol(*(pacSet->pacModelValz));
				iVal = InvACReorder(iVal);
				*ppxliQuantLH += InvACQuant(iVal,iQuant);
			}

			if(*pucMapHH!=ZT_IGN)
			{
				iVal = pacCoder->decodeSymbol(*(pacSet->pacModelValz));
				iVal = InvACReorder(iVal);
				*ppxliQuantHH += InvACQuant(iVal,iQuant);
			}
		}	
	}
	else
	{
		// not leaf node
		Int iX,iY;
		Int iWidth = getWidth();
		Int iHeight = getHeight();
		Int iPrevWidth = iWidth<<1;
		U8 *pucPrevMapHL = getParent()->m_pucMap[WVT_HL];
		U8 *pucPrevMapLH = getParent()->m_pucMap[WVT_LH];
		U8 *pucPrevMapHH = getParent()->m_pucMap[WVT_HH];

		for(iY=0;iY<iHeight;iY++,
			pucPrevMapHL+=iPrevWidth,pucPrevMapLH+=iPrevWidth,pucPrevMapHH+=iPrevWidth)
			for(iX=0;iX<iWidth;iX++,
				ppxliQuantHL++,ppxliQuantLH++,ppxliQuantHH++,
				pucMapHL++,pucMapLH++,pucMapHH++,
				pucPrevMapHL+=2,pucPrevMapLH+=2,pucPrevMapHH+=2)
			{
				decodeZeroTreeNode(pacSet,iQuant,pucMapHL,ppxliQuantHL,
					pucPrevMapHL,iPrevWidth);
				decodeZeroTreeNode(pacSet,iQuant,pucMapLH,ppxliQuantLH,
					pucPrevMapLH,iPrevWidth);
				decodeZeroTreeNode(pacSet,iQuant,pucMapHH,ppxliQuantHH,
					pucPrevMapHH,iPrevWidth);
			}

		// recurse towards leaf
		getParent()->decodeZeroTreeHoriz(pacSet,iQuant);
	}
}

void CWvtQuantTree::decodeZeroTreeNode(WvtArithCoderSet *pacSet,Int iQuant,
									   U8 *pucMap,PixelI *ppxliQuant,U8 *pucPrevMap,
									   Int iPrevWidth)
{
	Bool bIgnoreFlag = TRUE;
	CWvtArithCoder *pacCoder = pacSet->pacCoder;
	Int iVal;

	if(*pucMap!=ZT_IGN)
	{
		if(*pucMap==ZT_VAL)
		{
			// previously a VAL, so just decode value
			iVal = pacCoder->decodeSymbol(*(pacSet->pacModelValz));
			iVal = InvACReorder(iVal);
			*ppxliQuant += InvACQuant(iVal,iQuant);
			bIgnoreFlag=FALSE;
		}
		else
		{
			switch(pacCoder->decodeSymbol(*(pacSet->pacModelType)))
			{
			case ACSYM_VAL:
				*pucMap=ZT_VAL;
				iVal = pacCoder->decodeSymbol(*(pacSet->pacModelValz));
				iVal = InvACReorder(iVal);
				*ppxliQuant += InvACQuant(iVal,iQuant);
				bIgnoreFlag=FALSE;
				break;

			case ACSYM_ZTR:
				*pucMap = ZT_ZTR;
				break;

			case ACSYM_VZTR:
				*pucMap = ZT_VZTR;
				iVal = pacCoder->decodeSymbol(*(pacSet->pacModelValnz))+1;
				iVal = InvACReorder(iVal);
				*ppxliQuant += InvACQuant(iVal,iQuant);
				break;

			default:
				assert(FALSE); // error condition
				break;
			}
		}
	}

	// set up next level
	if(bIgnoreFlag==TRUE)
	{
		pucPrevMap[0]=pucPrevMap[1]=ZT_IGN;
		pucPrevMap[iPrevWidth]=pucPrevMap[iPrevWidth+1]=ZT_IGN;
	}
	else
	{
		if(pucPrevMap[0]!=ZT_VAL)
			pucPrevMap[0]=ZT_XXX;
		if(pucPrevMap[1]!=ZT_VAL)
			pucPrevMap[1]=ZT_XXX;
		if(pucPrevMap[iPrevWidth]!=ZT_VAL)
			pucPrevMap[iPrevWidth]=ZT_XXX;
		if(pucPrevMap[iPrevWidth+1]!=ZT_VAL)
			pucPrevMap[iPrevWidth+1]=ZT_XXX;
	}
}

void CWvtQuantTree::decodeMultipleQ(CInBitStream &bs,Int iLength,Int iQuant,
									CIntImage *piiSave,Bool bInitFlag)
{
	if(bInitFlag==TRUE)
	{
		zeroQuant(WVT_HL);
		zeroQuant(WVT_LH);
		zeroQuant(WVT_HH);
		initMap(WVT_HL);
		initMap(WVT_LH);
		initMap(WVT_HH);
	}

	if(iLength>0)
	{
		// decode the tree

		CWvtQuantTree *pwvtqtParentSave = m_pParent;
		m_pParent=NULL;

		CWvtArithCoder acCoder;
		WvtArithCoderSet acSet;

		acCoder.decodeStart(bs,iLength);	
	
		Int iSymsValz = acCoder.decodeVLCValue(NUMBITS_AC_NUMSYMS);
		Int iSymsValnz = acCoder.decodeVLCValue(NUMBITS_AC_NUMSYMS);
#if(DEBUG_BITS)
		printf("      Numsyms valz\t= %d\n",iSymsValz);
		printf("      Numsyms valnz\t= %d\n",iSymsValnz);
#endif

		CWvtArithCoderModel acModelType(AC_ADAPTIVE,NSYMS); // 0=VAL, 1=VZTR, 2=ZTR
		CWvtArithCoderModel acModelValz(AC_ADAPTIVE,iSymsValz); // 0=0, 1=1 etc
		CWvtArithCoderModel acModelValnz(AC_ADAPTIVE,iSymsValnz); // 0=1, 1=2 etc
	
		acSet.pacCoder = &acCoder;
		acSet.pacModelType = &acModelType;
		acSet.pacModelValz = &acModelValz;
		acSet.pacModelValnz = &acModelValnz;

		getRoot()->decodeZeroTreeHoriz(&acSet,iQuant);

		acCoder.decodeEnd();

		m_pParent=pwvtqtParentSave;
	}

	if(piiSave!=NULL)
	{
		copyQuantToWvt(WVT_HL);
		copyQuantToWvt(WVT_LH);
		copyQuantToWvt(WVT_HH);

		//m_pwvtDecomp->printDecompTree();
		// inverse transform the wavelet tree
		m_pwvtDecomp->inverseTransform(piiSave);
	}
}

void CWvtQuantTree::decodeBiLevelQ(CWvtCodedStream *pcsStream,CIntImage *piiSave,
								   Bool bEnableSnr,Int iReqSNR,Int iLevelsInternal)
{
	Int iNumCoders = pcsStream->getSpatialLevels();
	assert(iNumCoders==1 && iLevelsInternal>=1 ||
		iNumCoders>=1 && iLevelsInternal==1);
	
	Int iValidLevels = iLevelsInternal;
	if(iNumCoders>iValidLevels)
		iValidLevels=iNumCoders;

	CWvtQuantTree *pwvtqtLast = getRoot();
	for(;iValidLevels>1;iValidLevels--)
		pwvtqtLast = pwvtqtLast->getParent();

	CWvtQuantTree *pwvtqtSave = pwvtqtLast->getParent();
	pwvtqtLast->m_pParent = NULL;

	WvtBiLevelACSet acSet;
	CWvtArithCoderModel acModel(AC_ADAPTIVE,NSYMSBL);
	acSet.pacModel = &acModel;
	acSet.iNumCoders = iNumCoders;
	acSet.ppacCoder = new CWvtArithCoder * [iNumCoders];
	Int *piThresh = new Int [iNumCoders];
	Int i,iSym;
	for(i=0;i<iNumCoders;i++)
		piThresh[i] = pcsStream->getACQuant(i,0);

	pwvtqtLast->zeroQuant(WVT_HL);
	pwvtqtLast->zeroQuant(WVT_LH);
	pwvtqtLast->zeroQuant(WVT_HH);
	pwvtqtLast->initMap(WVT_HL);
	pwvtqtLast->initMap(WVT_LH);
	pwvtqtLast->initMap(WVT_HH);

	Int iSNRLevel = 0;
	Int iSNRCounter = 0;
	Bool bInitPassFlag = TRUE;
#if(DEBUG_DECODE)
	printf("Decoding: ");
#endif
	while(TRUE)
	{
#if(DEBUG_DECODE)
		printf("[%d]",iSNRCounter);
#endif

		// initialise decoders
		if(bEnableSnr==TRUE || bInitPassFlag==TRUE)
			for(i=0;i<iNumCoders;i++)
			{
				acSet.ppacCoder[i] = pcsStream->getACStream(i,iSNRLevel);
				acSet.ppacCoder[i]->decodeStart();
			}

		// decode zero tree
		getRoot()->decodeZeroTreeBiLevel(piThresh,&acSet);

		// get ready for continue or eof
		acModel.loadModel(rgiFreqTableEnd);

		for(i=0;i<iNumCoders;i++)
		{
			if(piThresh[i]>1)
				piThresh[i]>>=1;
			// check for eof/cont
			iSym = acSet.ppacCoder[i]->decodeSymbol(acModel);
			assert(iSym==ACSYMBL_CNT || iSym==ACSYMBL_EOF);
		}
		if(iSym==ACSYMBL_EOF)
			break;

		// refinement pass
		getRoot()->decodeZeroTreeRefineBiLevel(piThresh,&acSet);

		iSNRCounter++;
		if(bEnableSnr==TRUE)
		{
			if(iReqSNR==iSNRLevel)
				break;
			iSNRLevel++;
			for(i=0;i<iNumCoders;i++)
				acSet.ppacCoder[i]->decodeEnd();  // not really needed but consistent
		}
		bInitPassFlag=FALSE;	
	}
#if(DEBUG_DECODE)
	printf("\n");
#endif

	for(i=0;i<iNumCoders;i++)
		acSet.ppacCoder[i]->decodeEnd();  // not really needed but consistent

	delete piThresh;
	delete acSet.ppacCoder;

	if(piiSave!=NULL)
	{
		copyQuantToWvt(WVT_HL);
		copyQuantToWvt(WVT_LH);
		copyQuantToWvt(WVT_HH);

		// inverse transform the wavelet tree
		//m_pwvtDecomp->printDecompTree();
		m_pwvtDecomp->inverseTransform(piiSave);
	}

	pwvtqtLast->m_pParent = pwvtqtSave;
}

void CWvtQuantTree::decodeZeroTreeBiLevel(Int *piThresh,WvtBiLevelACSet *pacSet)
{
	if(getChild()==NULL)		// root
		pacSet->iCurrCoder=0;	// initialise first time
	else if(pacSet->iNumCoders>1)
		pacSet->iCurrCoder++;	// increment if multi spatial

	if(getParent()==NULL)
		pacSet->pacModel->loadModel(rgiFreqTableLeaf);	// leaf
	else if(getChild()==NULL || pacSet->iNumCoders>1)
		pacSet->pacModel->loadModel(rgiFreqTableCoef);	// root or multi spatial

	decodeZeroTreeBandBiLevel(WVT_LH,piThresh,pacSet);
	decodeZeroTreeBandBiLevel(WVT_HL,piThresh,pacSet);
	decodeZeroTreeBandBiLevel(WVT_HH,piThresh,pacSet);

	if(getParent()!=NULL)
		getParent()->decodeZeroTreeBiLevel(piThresh,pacSet);

}

void CWvtQuantTree::decodeZeroTreeRefineBiLevel(Int *piThresh,WvtBiLevelACSet *pacSet)
{
	if(getChild()==NULL)		// root
		pacSet->iCurrCoder=0;	// initialise first time
	else if(pacSet->iNumCoders>1)
		pacSet->iCurrCoder++;	// increment if multi spatial
	
	if(getChild()==NULL || pacSet->iNumCoders>1) // root or multi spatial
		pacSet->pacModel->loadModel(rgiFreqTableRefine);

	decodeZeroTreeRefineBandBiLevel(WVT_LH,piThresh,pacSet);
	decodeZeroTreeRefineBandBiLevel(WVT_HL,piThresh,pacSet);
	decodeZeroTreeRefineBandBiLevel(WVT_HH,piThresh,pacSet);
	
	if(getParent()!=NULL)
		getParent()->decodeZeroTreeRefineBiLevel(piThresh,pacSet);
}

void CWvtQuantTree::decodeZeroTreeBandBiLevel(FilterBank band,Int *piThresh,
											  WvtBiLevelACSet *pacSet)
{
	CWvtArithCoderModel *pacModel = pacSet->pacModel;
	CWvtArithCoder *pacCoder = pacSet->ppacCoder[pacSet->iCurrCoder];
	Int iThresh = piThresh[pacSet->iCurrCoder];
	PixelI *ppxliQuant = (PixelI *)m_piiQuant[band]->pixels();

	Int iX,iY,iSymbol;
	Int iWidth = getWidth();
	Int iHeight = getHeight();
	Int iPrevWidth = iWidth<<1;
	U8 *pucPrevMap = NULL;
	Bool bValidMap = FALSE;
	if(getParent()!=NULL)
	{
		pucPrevMap = getParent()->m_pucMap[band];
		bValidMap=TRUE;
	}
	U8 *pucMap = m_pucMap[band];
	Bool bIgnoreFlag;

	for(iY=0;iY<iHeight;iY++,pucPrevMap+=iPrevWidth)
		for(iX=0;iX<iWidth;iX++,ppxliQuant++,pucMap++,pucPrevMap+=2)
		{
			bIgnoreFlag=TRUE;
			if(*pucMap!=ZT_IGN)
			{
				bIgnoreFlag=FALSE;
				if(*pucMap!=ZT_VAL)
				{
					iSymbol = pacCoder->decodeSymbol(*pacModel);
					if(iSymbol==ACSYMBL_POS)
					{
						*pucMap=ZT_VAL;
						*ppxliQuant += iThresh+(iThresh>>1);
					}
					else if(iSymbol==ACSYMBL_NEG)
					{
						*pucMap=ZT_VAL;
						*ppxliQuant -= iThresh+(iThresh>>1);
					}
					else if(iSymbol==ACSYMBL_ZTR)
					{
						*pucMap=ZT_ZTR;
						bIgnoreFlag=TRUE;
					}
					else if(iSymbol==ACSYMBL_IZ)
					{
						*pucMap=ZT_Z;
					}
					else
					{
						assert(FALSE);
					}
				}
			}
			if(bValidMap==TRUE)
			{
				if(bIgnoreFlag==TRUE)
				{
					if(pucPrevMap[0]!=ZT_VAL)
						pucPrevMap[0]=ZT_IGN;
					if(pucPrevMap[1]!=ZT_VAL)
						pucPrevMap[1]=ZT_IGN;
					if(pucPrevMap[iPrevWidth]!=ZT_VAL)
						pucPrevMap[iPrevWidth]=ZT_IGN;
					if(pucPrevMap[iPrevWidth+1]!=ZT_VAL)
						pucPrevMap[iPrevWidth+1]=ZT_IGN;
				}
				else
				{
					if(pucPrevMap[0]!=ZT_VAL)
						pucPrevMap[0]=ZT_XXX;
					if(pucPrevMap[1]!=ZT_VAL)
						pucPrevMap[1]=ZT_XXX;
					if(pucPrevMap[iPrevWidth]!=ZT_VAL)
						pucPrevMap[iPrevWidth]=ZT_XXX;
					if(pucPrevMap[iPrevWidth+1]!=ZT_VAL)
						pucPrevMap[iPrevWidth+1]=ZT_XXX;
				}
			}
		}
}

void CWvtQuantTree::decodeZeroTreeRefineBandBiLevel(FilterBank band,Int *piThresh,
													WvtBiLevelACSet *pacSet)
{
	CWvtArithCoderModel *pacModel = pacSet->pacModel;
	CWvtArithCoder *pacCoder = pacSet->ppacCoder[pacSet->iCurrCoder];
	Int iThresh = piThresh[pacSet->iCurrCoder]>>1;
	PixelI *ppxliQuant = (PixelI *)m_piiQuant[band]->pixels();
	Int i,iSymbol;
	if(iThresh<1)
		iThresh=1;

	for(i=getArea();i;i--,ppxliQuant++)
	{
		if(*ppxliQuant>0)
		{
			iSymbol=pacCoder->decodeSymbol(*pacModel);
			if(iSymbol==ACSYMBL_INC)
				*ppxliQuant += iThresh;
			else if(iSymbol==ACSYMBL_DEC)
				*ppxliQuant -= iThresh;
			else
				assert(FALSE);
		}
		else if(*ppxliQuant<0)
		{
			iSymbol=pacCoder->decodeSymbol(*pacModel);
			if(iSymbol==ACSYMBL_INC)
				*ppxliQuant -= iThresh;
			else if(iSymbol==ACSYMBL_DEC)
				*ppxliQuant += iThresh;
			else
				assert(FALSE);
		}
	}
}
