/*************************************************************************

This software module was originally developed by 

	Ming-Chieh Lee (mingcl@microsoft.com), Microsoft Corporation
	Wei-ge Chen (wchen@microsoft.com), Microsoft Corporation
	Bruce Lin (blin@microsoft.com), Microsoft Corporation
	Chuang Gu (chuanggu@microsoft.com), Microsoft Corporation
	(date: March, 1996)
and edited by 

	Wei Wu (weiwu@stallion.risc.rockwell.com), Rockwell Science Center
	Homer Chen (homer@jupiter.risc.rockwell.com), Rockwell Science Center

in the course of development of the MPEG-4 Video (ISO/IEC 14496-2). 
This software module is an implementation of a part of one or more MPEG-4 Video tools 
as specified by the MPEG-4 Video. 
ISO/IEC gives users of the MPEG-4 Video free license to this software module or modifications 
thereof for use in hardware or software products claiming conformance to the MPEG-4 Video. 
Those intending to use this software module in hardware or software products are advised that its use may infringe existing patents. 
The original developer of this software module and his/her company, 
the subsequent editors and their companies, 
and ISO/IEC have no liability for use of this software module or modifications thereof in an implementation. 
Copyright is not released for non MPEG-4 Video conforming products. 
Microsoft retains full right to use the code for his/her own purpose, 
assign or donate the code to a third party and to inhibit third parties from using the code for non <MPEG standard> conforming products. 
This copyright notice must be included in all copies or derivative works. 

Copyright (c) 1996, 1997.

Module Name:

	vopSeEnc.hpp

Abstract:

	Encoder for one VO.

Revision History:
	Sept. 30, 1997: Error resilient tools added by Toshiba

*************************************************************************/

#include <stdio.h>
#include <fstream.h>
#include <math.h>
#include <stdlib.h>

#include "typeapi.h"
#include "codehead.h"
#include "global.h"
#include "entropy/bitstrm.h"
#include "entropy/entropy.h"
#include "entropy/huffman.h"
#include "mode.h"
#include "dct.h"
#include "vd_grid.h" // VDS:
#include "vd_vector.h" // VDS:
#include "vopses.h"
#include "vopseenc.h"

#ifdef __MFC_
#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

#define new DEBUG_NEW				   
#endif // __MFC_

CVideoObjectEncoder::~CVideoObjectEncoder ()
{
	delete m_pvopcOrig;
	delete m_pvopcRefOrig1;
	delete m_pvopcRefOrig0;

	delete m_puciRefQZoom0;
	delete m_puciRefQZoom1;
	delete m_pfdct;
	delete [] m_rgdSNR;

	// bitstream stuff
	delete [] m_pchBitsBuffer; 
	delete m_pbitstrmOut;
	delete m_pentrencSet;
	delete [] m_pchShapeBitsBuffer;
	delete m_pbitstrmShape;

	// shape
	delete [] m_rgiSubBlkIndx16x16;
	delete [] m_rgiSSubBlkIndx16x16;
	delete [] m_rgiSubBlkIndx18x18;
	delete [] m_rgiSubBlkIndx20x20;
	delete [] m_rgiPxlIndx12x12;
	delete [] m_rgiPxlIndx8x8;

	// B-VOP MB buffer
	delete m_puciDirectPredMB;
	delete m_puciInterpPredMB;
	delete m_piiDirectErrorMB; 
	delete m_piiInterpErrorMB; 
}

CVideoObjectEncoder::CVideoObjectEncoder (
	UInt uiVOId, 
	VOLMode& volmd, 
	VOPMode& vopmd, 
	UInt nFirstFrame,
	UInt nLastFrame,
	Int iSessionWidth,
	Int iSessionHeight,
	Bool bRateControl,
	UInt uiBudget,
	ostream* pstrmTrace, // trace outstream
	UInt uiWarpingAccuracy, // for sprite warping
	Int iNumOfPnts, // for sprite warping
	CRct rctSpt	// rct sprite
) :
	CVideoObject (),
	m_nFirstFrame (nFirstFrame), m_nLastFrame (nLastFrame), 
	m_iBufferSize (uiBudget), m_bRateControl (bRateControl),
	m_pvopcOrig (NULL),
	m_rgiSubBlkIndx16x16 (NULL),
	m_rgiSSubBlkIndx16x16 (NULL),
	m_rgiSubBlkIndx18x18 (NULL),
	m_rgiSubBlkIndx20x20 (NULL),
	m_rgiPxlIndx12x12 (NULL),
	m_rgiPxlIndx8x8 (NULL)
	//, m_uiTotalPrev (0)
{
	// sprite stuff
	if (iNumOfPnts >= 0) {
		m_uiSprite = 1;
		m_uiWarpingAccuracy = uiWarpingAccuracy;
		m_iNumOfPnts = iNumOfPnts;
		m_rgstSrcQ = new CSiteD [m_iNumOfPnts];
		m_rgstDstQ = new CSiteD [m_iNumOfPnts];
		m_rctSpt = rctSpt;
	}
	else 
		m_uiSprite = 0;

	m_pchBitsBuffer = new Char [iSessionWidth * iSessionHeight * 2];	//we think this is enof for 4:2:0; if crushes, increase the buffer
	m_pbitstrmOut = new COutBitStream (m_pchBitsBuffer, 0, pstrmTrace);
	m_pentrencSet = new CEntropyEncoderSet (*m_pbitstrmOut);

	// shape cache
	m_pchShapeBitsBuffer = new Char [MB_SIZE * MB_SIZE * 2]; // same as above
	m_pbitstrmShape = new COutBitStream (m_pchShapeBitsBuffer, 0, pstrmTrace);
	m_pbitstrmShapeMBOut = m_pbitstrmOut; // initially the output stream (only changes for inter)

	m_uiVOId = uiVOId;
	m_volmd = volmd;

	Int iClockRate = m_volmd.iClockRate;
	assert (iClockRate >= 1 && iClockRate < 65536);
	for (m_iNumBitsTimeIncr = 1; m_iNumBitsTimeIncr <= 15; m_iNumBitsTimeIncr++)	{	
		if (iClockRate == 1)			
			break;
		iClockRate = (iClockRate >> 1);
	}

	m_vopmd = vopmd;
	decideMVInfo ();
	m_vopmd.iRoundingControl = 0;
	Int iMod = iSessionWidth % MB_SIZE;
	Int iSessionWidthRound = (iMod > 0) ? iSessionWidth + MB_SIZE - iMod : iSessionWidth;
	iMod = iSessionHeight % MB_SIZE;
	Int iSessionHeightRound = (iMod > 0) ? iSessionHeight + MB_SIZE - iMod : iSessionHeight;
	m_rctRefFrameY = CRct (
		-EXPANDY_REF_FRAME, -EXPANDY_REF_FRAME, 
		EXPANDY_REF_FRAME + iSessionWidthRound, EXPANDY_REF_FRAME + iSessionHeightRound
	);
	m_rctRefFrameUV = m_rctRefFrameY.downSampleBy2 ();
	m_pvopcOrig = new CVOPU8YUVBA (m_volmd.fAUsage, m_rctRefFrameY);
	m_pvopcRefOrig0 = new CVOPU8YUVBA (m_volmd.fAUsage, m_rctRefFrameY);
	m_pvopcRefOrig1 = new CVOPU8YUVBA (m_volmd.fAUsage, m_rctRefFrameY);
	allocateVOLMembers (iSessionWidthRound, iSessionHeightRound);

	CRct rctFrameZoom = m_rctRefFrameY.upSampleBy2 ();
	m_puciRefQZoom0 = new CU8Image (rctFrameZoom);

	m_puciRefQZoom1 = new CU8Image (rctFrameZoom);

	m_pfdct = new CFwdBlockDCT;

	// B-VOP MB buffer
	m_puciDirectPredMB = new CU8Image (CRct (0, 0, MB_SIZE, MB_SIZE));
	m_puciInterpPredMB = new CU8Image (CRct (0, 0, MB_SIZE, MB_SIZE));
	m_ppxlcDirectPredMBY = (PixelC*) m_puciDirectPredMB->pixels (); 
	m_ppxlcInterpPredMBY = (PixelC*) m_puciInterpPredMB->pixels (); 

	m_piiDirectErrorMB = new CIntImage (CRct (0, 0, MB_SIZE, MB_SIZE)); 
	m_piiInterpErrorMB = new CIntImage (CRct (0, 0, MB_SIZE, MB_SIZE)); 
	m_ppxliDirectErrorMBY = (PixelI*) m_piiDirectErrorMB->pixels (); 
	m_ppxliInterpErrorMBY = (PixelI*) m_piiInterpErrorMB->pixels (); 

	// with shape
	if (m_volmd.fAUsage != RECTANGLE) {
		m_rgiSubBlkIndx16x16 = computeShapeSubBlkIndex (4, 16);
		m_rgiSSubBlkIndx16x16 = computeShapeSubBlkIndex (2, 16);
		m_rgiSubBlkIndx18x18 = computeShapeSubBlkIndex (4, 18);
		m_rgiSubBlkIndx20x20 = computeShapeSubBlkIndex (4, 20);
		m_rgiPxlIndx12x12 = new Int [8 * 8];		
		Int* piPxl = m_rgiPxlIndx12x12;
		UInt i, j;
		for (i = 2; i < 10; i++)
			for (j = 2; j < 10; j++)
				*piPxl++ = i * 12 + j;
		m_rgiPxlIndx8x8 = new Int [4 * 4];		
		piPxl = m_rgiPxlIndx8x8;
		for (i = 2; i < 6; i++)
			for (j = 2; j < 6; j++)
				*piPxl++ = i * 8 + j;
	}

	m_statsVOL.reset ();
	codeVOHead ();
	codeVOLHead (iSessionWidth, iSessionHeight);//, rctSprite);
	m_statsVOL.nBitsStuffing += m_pbitstrmOut->flush ();
	m_statRC.resetSkipMode ();
	m_rgdSNR = (m_volmd.fAUsage == EIGHT_BIT) ? new Double [4] : new Double [3];

	// some fixed variables	
	m_iFrameWidthZoomY = m_iFrameWidthY * 2;
	m_iFrameWidthZoomUV = m_iFrameWidthUV * 2;
	m_iFrameWidthZoomYx2Minus2MB = m_iFrameWidthY * 2 * 2 - 2 * MB_SIZE;
	m_iFrameWidthZoomYx2Minus2Blk = m_iFrameWidthY * 2 * 2 - 2 * BLOCK_SIZE;
	
	if (m_volmd.fAUsage == RECTANGLE) {
		m_rctCurrVOPY = CRct (0, 0, iSessionWidthRound, iSessionHeightRound);
		m_pvopcOrig->setBoundRct (m_rctCurrVOPY);
		m_rctCurrVOPUV = m_rctCurrVOPY.downSampleBy2 ();
		
		m_rctRefVOPY0 = m_rctCurrVOPY;
		m_rctRefVOPY0.expand (EXPANDY_REFVOP);
		m_rctRefVOPUV0 = m_rctRefVOPY0.downSampleBy2 ();
		
		m_rctRefVOPY1 = m_rctRefVOPY0;
		m_rctRefVOPUV1 = m_rctRefVOPUV0;
		
		m_rctRefVOPZoom0 = m_rctRefVOPY0.upSampleBy2 ();
		m_pvopcRefOrig0->setBoundRct (m_rctRefVOPY0);
		m_pvopcRefOrig1->setBoundRct (m_rctRefVOPY1);
		computeVOLConstMembers (); // these VOP members are the same for all frames
	}
}

Bool CVideoObjectEncoder::skipTest(Time t,VOPpredType vopPredType)
{
	if (m_bRateControl) {
		if (!m_statRC.firstFrame ()) {// test whether we should skip this frame
			if (m_statRC.skipThisFrame ())
				return TRUE;
		} 
		else
			m_statRC.resetFirstFrame ();
	}
	return FALSE;
}

Void CVideoObjectEncoder::encode (
	Bool bVOPVisible, 
	Time t,
	VOPpredType vopPredType
//	const CVOPU8YUVBA* pvopfRefBaseLayer
)
{
	m_t = t;
	m_vopmd.vopPredType = vopPredType;
	if (m_uiSprite == 1 && m_t>=0)
		m_vopmd.vopPredType = SPRITE;

	m_statsVOP.reset ();
	m_pbitstrmOut->reset ();
	cout << "\nVOP " << m_uiVOId << "\n";
	
	if(!bVOPVisible)
	{
#ifdef __TRACE_AND_STATS_
		m_pbitstrmOut -> trace (m_t, "VOP_Time");
#endif // __TRACE_AND_STATS_
		codeNonCodedVOPHead ();
		cout << "\tTime..." << m_t << "\n\tNot coded.\n\n";
		cout.flush ();
		m_statsVOP.nBitsStuffing += m_pbitstrmOut->flush ();
#ifdef __TRACE_AND_STATS_
		m_statsVOP.dSNRY = m_statsVOP.dSNRU = m_statsVOP.dSNRV = 0;
		m_statsVOP.print (TRUE);
		m_statsVOL += m_statsVOP;
#endif // __TRACE_AND_STATS_
		return;
	}

	// set time stamps
	if(vopPredType==IVOP || vopPredType==PVOP)
	{
		m_tPastRef = m_tFutureRef;
		m_tFutureRef = m_t;
		m_iBCount = 0;
	}

	// count B-VOPs
	if(vopPredType==BVOP)
		m_iBCount++;

	if (!m_bGOV_Reset)						// reset at the begining of GOV; no reference frames will be updated
		updateAllRefVOPs ();				// update all reconstructed VOP's
	m_bGOV_Reset = FALSE;					// GOV header decoding can reset it.
	if (m_vopmd.vopPredType == PVOP) {
		m_vopmd.iRoundingControl ^= 0x00000001;
		//memset((PixelC*) m_puciRefQZoom0->pixels(),128,m_puciRefQZoom0->where().area());
		m_rctRefVOPZoom0 = m_rctRefVOPY0.upSampleBy2 ();
		biInterpolateY (m_pvopcRefQ0, m_rctRefVOPY0, m_puciRefQZoom0, m_rctRefVOPZoom0, m_vopmd.iRoundingControl);
	}
	else if (m_iBCount == 1) {
		//memset((PixelC*) m_puciRefQZoom0->pixels(),128,m_puciRefQZoom0->where().area());
		//memset((PixelC*) m_puciRefQZoom1->pixels(),128,m_puciRefQZoom1->where().area());
		m_rctRefVOPZoom0 = m_rctRefVOPY0.upSampleBy2 ();
		biInterpolateY (m_pvopcRefQ0, m_rctRefVOPY0, m_puciRefQZoom0, m_rctRefVOPZoom0, 0);	//done again for rounding control
		m_rctRefVOPZoom1 = m_rctRefVOPY1.upSampleBy2 ();
		biInterpolateY (m_pvopcRefQ1, m_rctRefVOPY1, m_puciRefQZoom1, m_rctRefVOPZoom1, 0);
	}

	if (m_volmd.fAUsage != RECTANGLE)
		resetBYPlane ();
	if (m_volmd.bOriginalForME) {
		if (m_vopmd.vopPredType == PVOP)
			swapVOPU8Pointers (m_pvopcRefOrig0, m_pvopcRefOrig1);
	}

	// prepare for the current original data
	// extend the currVOP size to multiples of MBSize, will put zero on extended pixels
	if (m_volmd.fAUsage != RECTANGLE) {
		findTightBoundingBox ();
		findBestBoundingBox ();
		m_rctCurrVOPY = m_pvopcOrig->whereBoundY ();
		m_rctCurrVOPUV = m_pvopcOrig->whereBoundUV ();
		setRefStartingPointers (); // need to compute the starting pointers of the refVOP's (corresponding to the current Rct)
		computeVOPMembers ();
	}

#ifdef __TRACE_AND_STATS_
	m_pbitstrmOut -> trace (m_t, "VOP_Time");
#endif // __TRACE_AND_STATS_
	codeVOPHead ();
	cout << "\t" << "Time..." << m_t;
	switch(vopPredType)
	{
	case IVOP:
		cout << "\tIVOP";
		break;
	case PVOP:
		cout << "\tPVOP (reference: t=" << m_tPastRef <<")";
		break;
	case BVOP:
		cout << "\tBVOP (past ref: t=" << m_tPastRef
			<< ", future ref: t=" << m_tFutureRef <<")";
		break;
	default:
		break;
	}
	cout << "\n";
	cout.flush ();
	
	encodeVOP ();

	m_statsVOP.nBitsStuffing += m_pbitstrmOut->flush ();

#ifdef __TRACE_AND_STATS_
	if (m_vopmd.vopPredType == BVOP)
		computeSNRs (m_pvopcCurrQ);
	else
		computeSNRs (m_pvopcRefQ1);
	m_statsVOP.dSNRY = m_rgdSNR [0];
	m_statsVOP.dSNRU = m_rgdSNR [1];
	m_statsVOP.dSNRV = m_rgdSNR [2];
	if (m_volmd.fAUsage == EIGHT_BIT)
		m_statsVOP.dSNRA = m_rgdSNR [3];
	m_statsVOP.print (TRUE);
	m_statsVOL += m_statsVOP;
#endif // __TRACE_AND_STATS_
	
	// rate control
	if (m_bRateControl) {
		assert (m_volmd.volType == BASE_LAYER);	//don't no know to do RC for scalability yet
		if (m_vopmd.vopPredType == IVOP) { // !!!! check whether CRct has been expanded
			Double dMad = 
				(m_pvopcOrig->getPlane (Y_PLANE)->sumAbs (m_rctCurrVOPY)) / 
				(Double) (m_rctCurrVOPY.area ());
			m_statRC.reset (
				m_nFirstFrame, 
				m_nLastFrame, 
				m_volmd.iTemporalRate, 
				m_iBufferSize, 
				dMad, 
				m_statsVOL.nBitsTotal
			);
		}
		else if (m_vopmd.vopPredType == PVOP)
			m_statRC.updateRCModel (m_statsVOP.total (), m_statsVOP.head ());
	}

	// store the direct mode data (by swaping pointers)
	if (m_vopmd.vopPredType != BVOP) {
		if(m_volmd.fAUsage != RECTANGLE)
			saveShapeMode();
		CMBMode* pmbmdTmp = m_rgmbmd;
		m_rgmbmd = m_rgmbmdRef;
		m_rgmbmdRef = pmbmdTmp;
//		if (m_vopmd.vopPredType == PVOP) {
		CMotionVector* pmvTmp = m_rgmv;
		m_rgmv = m_rgmvRef;
		m_rgmvRef = pmvTmp;
//		}
	}
	
	if (m_volmd.fAUsage != RECTANGLE) {
		if (m_vopmd.vopPredType != BVOP) {
			m_iNumMBRef = m_iNumMB;
			m_iNumMBXRef = m_iNumMBX;
			m_iNumMBYRef = m_iNumMBY;
			m_iOffsetForPadY = m_rctRefFrameY.offset (m_rctCurrVOPY.left, m_rctCurrVOPY.top);
			m_iOffsetForPadUV = m_rctRefFrameUV.offset (m_rctCurrVOPUV.left, m_rctCurrVOPUV.top);
			m_rctPrevNoExpandY = m_rctCurrVOPY;
			m_rctPrevNoExpandUV = m_rctCurrVOPUV;
			
			m_rctRefVOPY1 = m_rctCurrVOPY;
			m_rctRefVOPY1.expand (EXPANDY_REFVOP);
			m_rctRefVOPUV1 = m_rctCurrVOPUV;
			m_rctRefVOPUV1.expand (EXPANDUV_REFVOP);
			m_pvopcRefQ1->setBoundRct (m_rctRefVOPY1);
		}

		//this is ac/dc pred stuff; clean the memory
		Int nBlk = (m_volmd.fAUsage == EIGHT_BIT) ? 10 : 6;
		delete [] m_rgblkmCurrMB;
		for (Int iMB = 0; iMB < m_iNumMBX; iMB++)	{
			for (Int iBlk = 0; iBlk < nBlk; iBlk++)	{
				delete [] (m_rgpmbmAbove [iMB]->rgblkm) [iBlk];
				delete [] (m_rgpmbmCurr  [iMB]->rgblkm) [iBlk];
			}
			delete [] m_rgpmbmAbove [iMB]->rgblkm;
			delete m_rgpmbmAbove [iMB];
			delete [] m_rgpmbmCurr [iMB]->rgblkm;
			delete m_rgpmbmCurr [iMB];
		}
		delete [] m_rgpmbmAbove;
		delete [] m_rgpmbmCurr;
	}

	if (m_volmd.bOriginalForME) {
		if (m_vopmd.vopPredType == IVOP || m_vopmd.vopPredType == PVOP)
			updateAllOrigRefVOPs (); // update all original reference (MotionEsti) VOP's
	}
}

Void CVideoObjectEncoder::encode ( // VDS: overload function
	Bool bVOPVisible, 
	Time t,
	VOPpredType vopPredType,
	CGrid* pgrid,
	vector ViewPos,
	vector AimPos,
	Int FOV,
	Int rwin_width,
	Int rwin_height,
	CU8Image *allmaskY,
	CU8Image *allmaskUV
//	const CVOPU8YUVBA* pvopfRefBaseLayer
)
{
	m_t = t;
	m_vopmd.vopPredType = vopPredType;
	assert(vopPredType==IVOP); // VDS: add
	if (m_uiSprite == 1 && m_t>=0)
		m_vopmd.vopPredType = SPRITE;

	m_statsVOP.reset ();
	m_pbitstrmOut->reset ();
	cout << "\nVOP " << m_uiVOId << "\n";
	
	if(!bVOPVisible)
	{
#ifdef __TRACE_AND_STATS_
		m_pbitstrmOut -> trace (m_t, "VOP_Time");
#endif // __TRACE_AND_STATS_
		codeNonCodedVOPHead ();
		cout << "\tTime..." << m_t << "\n\tNot coded.\n\n";
		cout.flush ();
		m_statsVOP.nBitsStuffing += m_pbitstrmOut->flush ();
#ifdef __TRACE_AND_STATS_
		m_statsVOP.dSNRY = m_statsVOP.dSNRU = m_statsVOP.dSNRV = 0;
		m_statsVOP.print (TRUE);
		m_statsVOL += m_statsVOP;
#endif // __TRACE_AND_STATS_
		return;
	}

	// set time stamps
	if(vopPredType==IVOP || vopPredType==PVOP)
	{
		m_tPastRef = m_tFutureRef;
		m_tFutureRef = m_t;
		m_iBCount = 0;
	}

	// count B-VOPs
	if(vopPredType==BVOP)
		m_iBCount++;

	if (!m_bGOV_Reset)						// reset at the begining of GOV; no reference frames will be updated
		updateAllRefVOPs ();				// update all reconstructed VOP's
	m_bGOV_Reset = FALSE;					// GOV header decoding can reset it.
	if (m_vopmd.vopPredType == PVOP) {
		m_vopmd.iRoundingControl ^= 0x00000001;
		//memset((PixelC*) m_puciRefQZoom0->pixels(),128,m_puciRefQZoom0->where().area());
		m_rctRefVOPZoom0 = m_rctRefVOPY0.upSampleBy2 ();
		biInterpolateY (m_pvopcRefQ0, m_rctRefVOPY0, m_puciRefQZoom0, m_rctRefVOPZoom0, m_vopmd.iRoundingControl);
	}
	else if (m_iBCount == 1) {
		//memset((PixelC*) m_puciRefQZoom0->pixels(),128,m_puciRefQZoom0->where().area());
		//memset((PixelC*) m_puciRefQZoom1->pixels(),128,m_puciRefQZoom1->where().area());
		m_rctRefVOPZoom0 = m_rctRefVOPY0.upSampleBy2 ();
		biInterpolateY (m_pvopcRefQ0, m_rctRefVOPY0, m_puciRefQZoom0, m_rctRefVOPZoom0, 0);	//done again for rounding control
		m_rctRefVOPZoom1 = m_rctRefVOPY1.upSampleBy2 ();
		biInterpolateY (m_pvopcRefQ1, m_rctRefVOPY1, m_puciRefQZoom1, m_rctRefVOPZoom1, 0);
	}

	if (m_volmd.fAUsage != RECTANGLE)
		resetBYPlane ();
	if (m_volmd.bOriginalForME) {
		if (m_vopmd.vopPredType == PVOP)
			swapVOPU8Pointers (m_pvopcRefOrig0, m_pvopcRefOrig1);
	}

	// prepare for the current original data
	// extend the currVOP size to multiples of MBSize, will put zero on extended pixels
	if (m_volmd.fAUsage != RECTANGLE) {
		findTightBoundingBox ();
		findBestBoundingBox ();
		m_rctCurrVOPY = m_pvopcOrig->whereBoundY ();
		m_rctCurrVOPUV = m_pvopcOrig->whereBoundUV ();
		setRefStartingPointers (); // need to compute the starting pointers of the refVOP's (corresponding to the current Rct)
		computeVOPMembers ();
	}

#ifdef __TRACE_AND_STATS_
	m_pbitstrmOut -> trace (m_t, "VOP_Time");
#endif // __TRACE_AND_STATS_
	codeVOPHead ();
	cout << "\t" << "Time..." << m_t;
	switch(vopPredType)
	{
	case IVOP:
		cout << "\tIVOP";
		break;
	case PVOP:
		cout << "\tPVOP (reference: t=" << m_tPastRef <<")";
		break;
	case BVOP:
		cout << "\tBVOP (past ref: t=" << m_tPastRef
			<< ", future ref: t=" << m_tFutureRef <<")";
		break;
	default:
		break;
	}
	cout << "\n";
	cout.flush ();
	
/* VDS: replace
	encodeVOP ();
*/
	encodeNSForIVOP(pgrid, ViewPos, AimPos, FOV, rwin_width, rwin_height, allmaskY, allmaskUV);

	m_statsVOP.nBitsStuffing += m_pbitstrmOut->flush ();

#ifdef __TRACE_AND_STATS_
	if (m_vopmd.vopPredType == BVOP)
		computeSNRs (m_pvopcCurrQ);
	else
		computeSNRs (m_pvopcRefQ1);
	m_statsVOP.dSNRY = m_rgdSNR [0];
	m_statsVOP.dSNRU = m_rgdSNR [1];
	m_statsVOP.dSNRV = m_rgdSNR [2];
	if (m_volmd.fAUsage == EIGHT_BIT)
		m_statsVOP.dSNRA = m_rgdSNR [3];
	m_statsVOP.print (TRUE);
	m_statsVOL += m_statsVOP;
#endif // __TRACE_AND_STATS_
	
	// rate control
	if (m_bRateControl) {
		assert (m_volmd.volType == BASE_LAYER);	//don't no know to do RC for scalability yet
		if (m_vopmd.vopPredType == IVOP) { // !!!! check whether CRct has been expanded
			Double dMad = 
				(m_pvopcOrig->getPlane (Y_PLANE)->sumAbs (m_rctCurrVOPY)) / 
				(Double) (m_rctCurrVOPY.area ());
			m_statRC.reset (
				m_nFirstFrame, 
				m_nLastFrame, 
				m_volmd.iTemporalRate, 
				m_iBufferSize, 
				dMad, 
				m_statsVOL.nBitsTotal
			);
		}
		else if (m_vopmd.vopPredType == PVOP)
			m_statRC.updateRCModel (m_statsVOP.total (), m_statsVOP.head ());
	}

	// store the direct mode data (by swaping pointers)
	if (m_vopmd.vopPredType != BVOP) {
		if(m_volmd.fAUsage != RECTANGLE)
			saveShapeMode();
		CMBMode* pmbmdTmp = m_rgmbmd;
		m_rgmbmd = m_rgmbmdRef;
		m_rgmbmdRef = pmbmdTmp;
//		if (m_vopmd.vopPredType == PVOP) {
		CMotionVector* pmvTmp = m_rgmv;
		m_rgmv = m_rgmvRef;
		m_rgmvRef = pmvTmp;
//		}
	}
	
	if (m_volmd.fAUsage != RECTANGLE) {
		if (m_vopmd.vopPredType != BVOP) {
			m_iNumMBRef = m_iNumMB;
			m_iNumMBXRef = m_iNumMBX;
			m_iNumMBYRef = m_iNumMBY;
			m_iOffsetForPadY = m_rctRefFrameY.offset (m_rctCurrVOPY.left, m_rctCurrVOPY.top);
			m_iOffsetForPadUV = m_rctRefFrameUV.offset (m_rctCurrVOPUV.left, m_rctCurrVOPUV.top);
			m_rctPrevNoExpandY = m_rctCurrVOPY;
			m_rctPrevNoExpandUV = m_rctCurrVOPUV;
			
			m_rctRefVOPY1 = m_rctCurrVOPY;
			m_rctRefVOPY1.expand (EXPANDY_REFVOP);
			m_rctRefVOPUV1 = m_rctCurrVOPUV;
			m_rctRefVOPUV1.expand (EXPANDUV_REFVOP);
			m_pvopcRefQ1->setBoundRct (m_rctRefVOPY1);
		}

		//this is ac/dc pred stuff; clean the memory
		Int nBlk = (m_volmd.fAUsage == EIGHT_BIT) ? 10 : 6;
		delete [] m_rgblkmCurrMB;
		for (Int iMB = 0; iMB < m_iNumMBX; iMB++)	{
			for (Int iBlk = 0; iBlk < nBlk; iBlk++)	{
				delete [] (m_rgpmbmAbove [iMB]->rgblkm) [iBlk];
				delete [] (m_rgpmbmCurr  [iMB]->rgblkm) [iBlk];
			}
			delete [] m_rgpmbmAbove [iMB]->rgblkm;
			delete m_rgpmbmAbove [iMB];
			delete [] m_rgpmbmCurr [iMB]->rgblkm;
			delete m_rgpmbmCurr [iMB];
		}
		delete [] m_rgpmbmAbove;
		delete [] m_rgpmbmCurr;
	}

/* VDS: no update of original data!
	if (m_volmd.bOriginalForME) {
		if (m_vopmd.vopPredType == IVOP || m_vopmd.vopPredType == PVOP)
			updateAllOrigRefVOPs (); // update all original reference (MotionEsti) VOP's
	}
*/
}

Void CVideoObjectEncoder::codeVOHead ()
{
	static UInt uiNumBits = NUMBITS_START_CODE_PREFIX + NUMBITS_VO_START_CODE + NUMBITS_VO_ID;
	m_pbitstrmOut -> putBits (START_CODE_PREFIX, NUMBITS_START_CODE_PREFIX, "VO_Start_Code");
	m_pbitstrmOut -> putBits (VO_START_CODE, NUMBITS_VO_START_CODE, "VO_Start_Code");		//plus 3 bits
	m_pbitstrmOut -> putBits (m_uiVOId, NUMBITS_VO_ID, "VO_Id");
	m_statsVOL.nBitsHead += uiNumBits;
	cout << "VO Overhead" << "\t\t" << uiNumBits << "\n\n";
	cout.flush ();
}

Void CVideoObjectEncoder::codeVOLHead (Int iSessionWidth, Int iSessionHeight)//, const CRct& rctSprite)
{
	m_pbitstrmOut -> putBits (START_CODE_PREFIX, NUMBITS_START_CODE_PREFIX, "VOL_Start_Code");
	m_pbitstrmOut -> putBits (VOL_START_CODE, NUMBITS_VOL_START_CODE, "VOL_Start_Code");		//plus 4 bits
	m_pbitstrmOut -> putBits ((Int) 0, NUMBITS_VOL_ID, "VOL_Id");

	if(m_volmd.bShapeOnly==TRUE)
	{
		m_pbitstrmOut -> putBits ((Int)2, NUMBITS_VOL_SHAPE, "VOL_Shape_Type");
		m_pbitstrmOut->putBits (m_volmd.iClockRate, NUMBITS_TIME_RESOLUTION, "VOL_Time_Increment_Resolution"); 
		m_statsVOL.nBitsHead += NUMBITS_TIME_RESOLUTION;
		m_statsVOL.nBitsHead += NUMBITS_START_CODE_PREFIX + NUMBITS_VOL_START_CODE + NUMBITS_VOL_ID + NUMBITS_VOL_SHAPE;
	}
	else
	{
		m_pbitstrmOut -> putBits ((Int) m_volmd.fAUsage, NUMBITS_VOL_SHAPE, "VOL_Shape_Type"); 
		m_pbitstrmOut->putBits (m_volmd.iClockRate, NUMBITS_TIME_RESOLUTION, "VOL_Time_Increment_Resolution"); 
		m_statsVOL.nBitsHead += NUMBITS_TIME_RESOLUTION;
		m_statsVOL.nBitsHead += NUMBITS_START_CODE_PREFIX + NUMBITS_VOL_START_CODE + NUMBITS_VOL_ID + NUMBITS_VOL_SHAPE;
		if (m_volmd.fAUsage == RECTANGLE) {
			m_pbitstrmOut -> putBits (MARKER_BIT, MARKER_BIT, "Marker_Bit");
			m_statsVOL.nBitsStuffing += MARKER_BIT;
			m_pbitstrmOut -> putBits (iSessionWidth, NUMBITS_VOP_WIDTH, "VOL_Width"); 
			m_pbitstrmOut -> putBits (MARKER_BIT, MARKER_BIT, "Marker_Bit");
			m_statsVOL.nBitsStuffing += MARKER_BIT;
			m_pbitstrmOut -> putBits (iSessionHeight, NUMBITS_VOP_HEIGHT, "VOL_Height"); 
			m_statsVOL.nBitsHead += NUMBITS_VOP_WIDTH + NUMBITS_VOP_HEIGHT;
		}
	}

	if(m_volmd.bShapeOnly==TRUE)
	{
		cout << "VOL Overhead" << "\t\t" << m_statsVOL.nBitsHead << "\n\n";
		cout.flush ();
		return;
	}

#ifdef __VERIFICATION_MODEL_
	m_pbitstrmOut -> putBits (MARKER_BIT, MARKER_BIT, "Marker_Bit");
	m_statsVOL.nBitsStuffing += MARKER_BIT;
	m_pbitstrmOut -> putBits (1, (UInt) 1, "VOL_NOT_8_BIT_VIDEO");	
	m_statsVOL.nBitsHead++;
#endif //__VERIFICATION_MODEL_

	m_pbitstrmOut -> putBits (m_volmd.bAdvPredDisable, (UInt) 1, "VOL_OBMC_Disable");
	m_statsVOL.nBitsHead++;

#ifdef __VERIFICATION_MODEL_
	if (m_volmd.fAUsage != RECTANGLE)	{
		m_pbitstrmOut -> putBits (0, (UInt) 4, "VOL_Shape_Effect");		//no effect
		m_statsVOL.nBitsHead += 4;
	}
#endif //__VERIFICATION_MODEL_

	// code sprite info
	m_pbitstrmOut -> putBits (m_uiSprite, NUMBITS_SPRITE_USAGE, "VOL_Sprite_Usage");		
	m_statsVOL.nBitsHead += NUMBITS_SPRITE_USAGE;
	if (m_uiSprite == 1) {
		m_pbitstrmOut -> putBits (m_rctSpt.width, NUMBITS_SPRITE_HDIM, "SPRT_hdim");
		m_statsVOL.nBitsHead += NUMBITS_SPRITE_HDIM; 
		m_pbitstrmOut -> putBits (MARKER_BIT, MARKER_BIT, "Marker_Bit");
		m_statsVOL.nBitsStuffing += MARKER_BIT;
		m_pbitstrmOut -> putBits (m_rctSpt.height (), NUMBITS_SPRITE_VDIM, "SPRT_vdim");
		m_statsVOL.nBitsHead += NUMBITS_SPRITE_VDIM;
		m_pbitstrmOut -> putBits (MARKER_BIT, MARKER_BIT, "Marker_Bit");
		m_statsVOL.nBitsStuffing += MARKER_BIT;
		if (m_rctSpt.left >= 0) {
			m_pbitstrmOut -> putBits (m_rctSpt.left, NUMBITS_SPRITE_LEFT_EDGE, "SPRT_Left_Edge");
		} else {
			m_pbitstrmOut -> putBits (m_rctSpt.left & 0x00001FFFF, NUMBITS_SPRITE_LEFT_EDGE, "SPRT_Left_Edge");
		}
		m_statsVOL.nBitsHead += NUMBITS_SPRITE_LEFT_EDGE;
		m_pbitstrmOut -> putBits (MARKER_BIT, MARKER_BIT, "Marker_Bit");
		m_statsVOL.nBitsStuffing += MARKER_BIT;
		if (m_rctSpt.top >= 0) {
			m_pbitstrmOut -> putBits (m_rctSpt.top, NUMBITS_SPRITE_LEFT_EDGE, "SPRT_Top_Edge");
		} else {
			m_pbitstrmOut -> putBits (m_rctSpt.top & 0x00001FFFF, NUMBITS_SPRITE_LEFT_EDGE, "SPRT_Top_Edge");
		}
		m_statsVOL.nBitsHead += NUMBITS_SPRITE_TOP_EDGE; 
		m_pbitstrmOut -> putBits (MARKER_BIT, MARKER_BIT, "Marker_Bit");
		m_statsVOL.nBitsStuffing += MARKER_BIT;
		if (m_volmd.fAUsage == RECTANGLE) {// initial sprite piece
			m_pbitstrmOut -> putBits (m_rctSpt.width, NUMBITS_SPRITE_HDIM, "SPRT_hdim");
			m_statsVOL.nBitsHead += NUMBITS_SPRITE_HDIM; 
			m_pbitstrmOut -> putBits (MARKER_BIT, MARKER_BIT, "Marker_Bit");
			m_statsVOL.nBitsStuffing += MARKER_BIT;
			m_pbitstrmOut -> putBits (m_rctSpt.height (), NUMBITS_SPRITE_VDIM, "SPRT_vdim");
			m_statsVOL.nBitsHead += NUMBITS_SPRITE_VDIM;
			m_pbitstrmOut -> putBits (MARKER_BIT, MARKER_BIT, "Marker_Bit");
			m_statsVOL.nBitsStuffing += MARKER_BIT;
			m_pbitstrmOut -> putBits (m_rctSpt.left - m_rctSpt.left, NUMBITS_SPRITE_LEFT_EDGE, "SPRT_Left_Edge");
			m_statsVOL.nBitsHead += NUMBITS_SPRITE_LEFT_EDGE;
			m_pbitstrmOut -> putBits (MARKER_BIT, MARKER_BIT, "Marker_Bit");
			m_statsVOL.nBitsStuffing += MARKER_BIT;
			m_pbitstrmOut -> putBits (m_rctSpt.top - m_rctSpt.top, NUMBITS_SPRITE_LEFT_EDGE, "SPRT_Top_Edge");
			m_statsVOL.nBitsHead += NUMBITS_SPRITE_LEFT_EDGE;
		}
		m_pbitstrmOut -> putBits (m_iNumOfPnts, NUMBITS_NUM_SPRITE_POINTS, "SPRT_Num_Pnt");
		m_statsVOL.nBitsHead += NUMBITS_NUM_SPRITE_POINTS;
		m_pbitstrmOut -> putBits (m_uiWarpingAccuracy, NUMBITS_WARPING_ACCURACY, "SPRT_Warping_Accuracy");
		m_statsVOL.nBitsHead += NUMBITS_WARPING_ACCURACY;
		Bool bLightChange = 0;
		m_pbitstrmOut -> putBits (bLightChange, 1);
		m_statsVOL.nBitsHead++;
	}

#ifdef __VERIFICATION_MODEL_
	if (m_volmd.fAUsage != RECTANGLE)
		m_pbitstrmOut->putBits (TRUE, 1, "VOL_SADCT_Disable");	//no supported
#endif //__VERIFICATION_MODEL_

	m_pbitstrmOut -> putBits ((Int) m_volmd.fQuantizer, 1, "VOL_Quant_Type"); 
	m_statsVOL.nBitsHead++;
	if (m_volmd.fQuantizer == Q_MPEG) {
		m_pbitstrmOut -> putBits (m_volmd.bLoadIntraMatrix, 1, "VOL_Load_Q_Matrix (intra)");
		m_statsVOL.nBitsHead++;
		if (m_volmd.bLoadIntraMatrix) {
			UInt i = 0;
			do {
				m_pbitstrmOut -> putBits (m_volmd.rgiIntraQuantizerMatrix [grgiStandardZigzag[i]], NUMBITS_QMATRIX, "VOL_Quant_Matrix(intra)");
				m_statsVOL.nBitsHead += NUMBITS_QMATRIX;
			} while (m_volmd.rgiIntraQuantizerMatrix [grgiStandardZigzag[i]] != 0 && ++i < BLOCK_SQUARE_SIZE);
			for (UInt j = i; j < BLOCK_SQUARE_SIZE; j++) {
				m_volmd.rgiIntraQuantizerMatrix [j] = m_volmd.rgiIntraQuantizerMatrix [grgiStandardZigzag[i - 1]];
			}
		}
		m_pbitstrmOut -> putBits (m_volmd.bLoadInterMatrix, 1, "VOL_Load_Q_Matrix (inter)");
		m_statsVOL.nBitsHead++;
		if (m_volmd.bLoadInterMatrix) {
			UInt i = 0;
			do {
				m_pbitstrmOut -> putBits (m_volmd.rgiInterQuantizerMatrix [grgiStandardZigzag[i]], NUMBITS_QMATRIX, "VOL_Quant_Matrix(intra)");
				m_statsVOL.nBitsHead += NUMBITS_QMATRIX;
			} while (m_volmd.rgiInterQuantizerMatrix [grgiStandardZigzag[i]] != 0 && ++i < BLOCK_SQUARE_SIZE);
			for (UInt j = i; j < BLOCK_SQUARE_SIZE; j++) {
				m_volmd.rgiInterQuantizerMatrix [j] = m_volmd.rgiInterQuantizerMatrix [grgiStandardZigzag[i - 1]];
			}
		}
		if (m_volmd.fAUsage == EIGHT_BIT)	{
			m_pbitstrmOut -> putBits (m_volmd.bLoadIntraMatrixAlpha, 1, "VOL_Load_Alpha_Q_Matrix (intra)");
			m_statsVOL.nBitsHead++;
			if (m_volmd.bLoadIntraMatrixAlpha) {
				for (UInt i = 0; i < BLOCK_SQUARE_SIZE; i++) {
					m_pbitstrmOut -> putBits (m_volmd.rgiIntraQuantizerMatrixAlpha [grgiStandardZigzag[i]], NUMBITS_QMATRIX, "VOL_Alpha_Quant_Matrix(intra)");
					m_statsVOL.nBitsHead += NUMBITS_QMATRIX;
				}
			}
			m_pbitstrmOut -> putBits (m_volmd.bLoadInterMatrixAlpha, 1, "VOL_Load_Alpha_Q_Matrix (inter)");
			m_statsVOL.nBitsHead++;
			if (m_volmd.bLoadInterMatrixAlpha) {
				for (UInt i = 0; i < BLOCK_SQUARE_SIZE; i++) {
					m_pbitstrmOut -> putBits (m_volmd.rgiInterQuantizerMatrixAlpha [grgiStandardZigzag[i]], NUMBITS_QMATRIX, "VOL_Alpha_Quant_Matrix(inter)");
					m_statsVOL.nBitsHead += NUMBITS_QMATRIX;
				}
			}
		}
	}
	m_pbitstrmOut -> putBits (m_volmd.bErrorResilientDisable, 1, "VOL_error_resilient_disable");
	if ( !m_volmd.bErrorResilientDisable )  {
		m_pbitstrmOut -> putBits (m_volmd.bDataPartitioning, 1, "VOL_data_partitioning");
		m_pbitstrmOut -> putBits (m_volmd.bReversibleVlc, 1, "VOL_reversible_vlc");
	}

	m_pbitstrmOut -> putBits (m_volmd.bIntraPredictionDisable, 1, "VOL_intra_acdc_pred_disable");
	m_statsVOL.nBitsHead += 2;
#ifdef __VERIFICATION_MODEL_
	m_pbitstrmOut -> putBits (m_volmd.bSeprMotionTexture, NUMBITS_SEP_MOTION_TEXTURE, "VOL_Sep_Mot_Texture");
	m_statsVOL.nBitsHead += NUMBITS_SEP_MOTION_TEXTURE;
#endif //__VERIFICATION_MODEL_
	m_pbitstrmOut -> putBits (m_volmd.volType == ENHN_LAYER, 1, "VOL_Scalability");
	m_statsVOL.nBitsHead++;
	if (m_volmd.volType == ENHN_LAYER)	{
		m_pbitstrmOut -> putBits (0, 1, "VOL_Ref_Layer_Id");
		m_pbitstrmOut -> putBits (0, 1, "VOL_Ref_Layer_Sampling_Dir");
		m_pbitstrmOut -> putBits (m_volmd.iEnhnUpFactor, 5, "VOL_Horizontal_Sampling_Factor");			//The vm is not very clear about this.
		m_pbitstrmOut -> putBits (1, 5, "VOL_Horizontal_Sampling_Factor_Ref");		//Always 1
		m_pbitstrmOut -> putBits (m_volmd.iEnhnUpFactor, 5, "VOL_Horizontal_Vertical_Sampling_Factor");			//The vm is not very clear about this.
		m_pbitstrmOut -> putBits (1, 5, "VOL_Vertical_Sampling_Factor_Ref");		//Always 1
		m_pbitstrmOut -> putBits (0, 1, "VOL_Ehnancement_Type");		//Always 1
		m_statsVOL.nBitsHead += 2 + 20 + 1;
	}
	cout << "VOL Overhead" << "\t\t" << m_statsVOL.nBitsHead << "\n\n";
	cout.flush ();
}

Void CVideoObjectEncoder::codeGOVHead ()
{
	// Start code
	m_pbitstrmOut -> putBits (START_CODE_PREFIX, NUMBITS_START_CODE_PREFIX, "GOV_Start_Code");
	m_pbitstrmOut -> putBits (GOV_START_CODE, NUMBITS_GOV_START_CODE, "GOV_Start_Code");
	m_statsVOL.nBitsHead += NUMBITS_START_CODE_PREFIX + NUMBITS_GOV_START_CODE;

	// Time code
	m_pbitstrmOut -> putBits (0, NUMBITS_GOV_TIMECODE_HOUR, "GOV_Time_Code_Hour");
	m_statsVOL.nBitsHead += NUMBITS_GOV_TIMECODE_HOUR;
	m_pbitstrmOut -> putBits (0, NUMBITS_GOV_TIMECODE_MIN, "GOV_Time_Code_Min");
	m_statsVOL.nBitsHead += NUMBITS_GOV_TIMECODE_MIN;
	m_pbitstrmOut -> putBits (MARKER_BIT, MARKER_BIT, "Marker_Bit");
	m_statsVOL.nBitsStuffing += MARKER_BIT;
	m_pbitstrmOut -> putBits (0, NUMBITS_GOV_TIMECODE_SEC, "GOV_Time_Code_Sec");
	m_statsVOL.nBitsHead += NUMBITS_GOV_TIMECODE_SEC;
}

Void CVideoObjectEncoder::codeVOPHeadInitial()
{
	// Start code	
	m_pbitstrmOut -> putBits (START_CODE_PREFIX, NUMBITS_START_CODE_PREFIX, "VOP_Start_Code");
	m_pbitstrmOut -> putBits (VOP_START_CODE, NUMBITS_VOP_START_CODE, "VOP_Start_Code");
	m_statsVOP.nBitsHead += NUMBITS_START_CODE_PREFIX + NUMBITS_VOP_START_CODE;

	// prediction type
	m_pbitstrmOut -> putBits (m_vopmd.vopPredType, NUMBITS_VOP_PRED_TYPE, "VOP_Pred_Type");
	m_statsVOP.nBitsHead += NUMBITS_VOP_PRED_TYPE;

	// Time reference
	Time tCurrSec = m_t / SOURCE_FRAME_RATE;		//current time in second
	m_nBitsModuloBase = tCurrSec - ((m_vopmd.vopPredType != BVOP) ? m_tModuloBaseDecd : m_tModuloBaseDisp);
	assert(m_nBitsModuloBase<=31);
	m_pbitstrmOut -> putBits ((Int) 0xFFFFFFFE, (UInt) m_nBitsModuloBase + 1, "VOP_Modulo_Time_Base");
	m_statsVOP.nBitsHead += m_nBitsModuloBase + 1;
	m_iVopTimeIncr = m_t * m_volmd.iClockRate / SOURCE_FRAME_RATE - tCurrSec * m_volmd.iClockRate;
	m_pbitstrmOut -> putBits (m_iVopTimeIncr, m_iNumBitsTimeIncr, "VOP_Time_Incr");
	m_statsVOP.nBitsHead += m_iNumBitsTimeIncr;
	if (m_vopmd.vopPredType != BVOP)	{	//update modulo time base
		m_tModuloBaseDisp = m_tModuloBaseDecd;							//of the most recently displayed I/Pvop
		m_tModuloBaseDecd = tCurrSec;							//of the most recently decoded I/Pvop
	}
}

Void CVideoObjectEncoder::codeNonCodedVOPHead ()
{
	codeVOPHeadInitial();

	m_pbitstrmOut -> putBits (0, 1, "VOP_Coded");
	m_statsVOP.nBitsHead++;
}

Void CVideoObjectEncoder::codeVOPHead ()
{
	codeVOPHeadInitial();

	m_pbitstrmOut -> putBits (1, 1, "VOP_Coded");
	m_statsVOP.nBitsHead++;

	if(m_volmd.bShapeOnly==FALSE)
	{
#ifdef __VERIFICATION_MODEL_
		m_pbitstrmOut -> putBits (m_vopmd.iIntraDcSwitchThr, 3, "VOP_DC_VLC_Switch");
		m_statsVOP.nBitsHead += 3;
#endif //__VERIFICATION_MODEL_
		if (m_vopmd.vopPredType == PVOP)	{
			m_pbitstrmOut -> putBits (m_vopmd.iRoundingControl, 1, "VOP_Rounding_Type");
			m_statsVOP.nBitsHead++;
		}
	}
	if (m_volmd.fAUsage != RECTANGLE) {	
		m_pbitstrmOut->putBits (m_rctCurrVOPY.width, NUMBITS_VOP_WIDTH, "VOP_Width");
		m_pbitstrmOut->putBits ((Int) 1, 1, "Marker"); // marker bit
		m_pbitstrmOut->putBits (m_rctCurrVOPY.height (), NUMBITS_VOP_HEIGHT, "VOP_Height");
		m_pbitstrmOut->putBits ((Int) 1, 1, "Marker"); // marker bit
		Int iSign = (m_rctCurrVOPY.left < 0) ? 1 : 0;
		m_pbitstrmOut->putBits (iSign, 1, "VOP_Hori_Ref_Sgn");
		m_pbitstrmOut->putBits (abs (m_rctCurrVOPY.left), NUMBITS_VOP_HORIZONTAL_SPA_REF - 1, "VOP_Hori_Ref");
		m_pbitstrmOut->putBits ((Int) 1, 1, "Marker"); // marker bit
		iSign = (m_rctCurrVOPY.top < 0) ? 1 : 0;
		m_pbitstrmOut->putBits (iSign, 1, "VOP_Vert_Ref_Sgn");
		m_pbitstrmOut->putBits (abs (m_rctCurrVOPY.top), NUMBITS_VOP_VERTICAL_SPA_REF - 1, "VOP_Vert_Ref");
		m_statsVOP.nBitsHead += NUMBITS_VOP_WIDTH + 1 + NUMBITS_VOP_HEIGHT + 1 + 1 +
				NUMBITS_VOP_HORIZONTAL_SPA_REF - 1 + 1 + 1 + NUMBITS_VOP_VERTICAL_SPA_REF - 1;
#ifdef __VERIFICATION_MODEL_
		if(m_volmd.bShapeOnly==FALSE)
		{
			m_pbitstrmOut->putBits (FALSE, 1, "VOP_CR");//VM8 doesn't support vop level size conversion
			m_statsVOP.nBitsHead++;
		}
#endif //__VERIFICATION_MODEL_
		m_pbitstrmOut->putBits (m_volmd.bNoCrChange, 1, "VOP_CR_Change_Disable");
		m_statsVOP.nBitsHead++;
	}
	
	if(m_volmd.bShapeOnly==TRUE)
		return;

#ifndef __VERIFICATION_MODEL_
	m_pbitstrmOut -> putBits (m_vopmd.iIntraDcSwitchThr, 3, "VOP_DC_VLC_Switch");
	m_statsVOP.nBitsHead += 3;
#endif //__VERIFICATION_MODEL_

	m_pbitstrmOut -> putBits (FALSE, 1, "VOP_Interlaced");		//don't support interlace 
	m_statsVOP.nBitsHead++;

	if (m_vopmd.vopPredType == BVOP) {
		m_pbitstrmOut -> putBits (m_vopmd.intDBQuant, NUMBITS_VOP_DBQUANT, "VOP_DBQUANT");
		m_statsVOP.nBitsHead += NUMBITS_VOP_DBQUANT;
		if (m_vopmd.intDBQuant == 0)
			m_vopmd.intStepB = min ((m_vopmd.intStep * 5) >> 2, 31);
		else if (m_vopmd.intDBQuant == 1)
			m_vopmd.intStepB = min ((m_vopmd.intStep * 6) >> 2, 31);
		else if (m_vopmd.intDBQuant == 2)
			m_vopmd.intStepB = min ((m_vopmd.intStep * 7) >> 2, 31);
		else if (m_vopmd.intDBQuant == 3)
			m_vopmd.intStepB = min ((m_vopmd.intStep * 8) >> 2, 31);
		else
			assert (FALSE);
		if (m_volmd.fAUsage == EIGHT_BIT)	{
			if (m_vopmd.intDBQuant == 0)
				m_vopmd.intStepBAlpha = min ((m_vopmd.intStepAlpha * 5) >> 2, 63);
			else if (m_vopmd.intDBQuant == 1)
				m_vopmd.intStepBAlpha = min ((m_vopmd.intStepAlpha * 6) >> 2, 63);
			else if (m_vopmd.intDBQuant == 2)
				m_vopmd.intStepBAlpha = min ((m_vopmd.intStepAlpha * 7) >> 2, 63);
			else if (m_vopmd.intDBQuant == 3)
				m_vopmd.intStepBAlpha = min ((m_vopmd.intStepAlpha * 8) >> 2, 63);
			else
				assert (FALSE);
		}
	}
	else
	{
		Int stepToBeCoded = (m_vopmd.vopPredType == PVOP) ? m_vopmd.intStep : m_vopmd.intStepI;
		m_pbitstrmOut -> putBits (stepToBeCoded, NUMBITS_VOP_QUANTIZER, "VOP_Quantier");
		m_statsVOP.nBitsHead += NUMBITS_VOP_QUANTIZER;
		if (m_volmd.fAUsage == EIGHT_BIT)
		{
			stepToBeCoded = (m_vopmd.vopPredType == PVOP) ? m_vopmd.intStepAlpha : m_vopmd.intStepIAlpha;
			m_pbitstrmOut -> putBits (stepToBeCoded, NUMBITS_VOP_ALPHA_QUANTIZER, "VOP_Alpha_Quantier");
			m_statsVOP.nBitsHead += NUMBITS_VOP_QUANTIZER;
		}
	}
	if (m_vopmd.vopPredType != IVOP) {
		m_pbitstrmOut -> putBits (m_vopmd.mvInfoForward.uiFCode, NUMBITS_VOP_FCODE, "VOP_Fcode_Forward");
		m_statsVOP.nBitsHead += NUMBITS_VOP_FCODE;
	}
	if (m_vopmd.vopPredType == BVOP) {
		m_pbitstrmOut -> putBits (m_vopmd.mvInfoBackward.uiFCode, NUMBITS_VOP_FCODE, "VOP_Fcode_Backward");
		m_statsVOP.nBitsHead += NUMBITS_VOP_FCODE;
	}
	if (m_volmd.bErrorResilientDisable == FALSE) {
		VideoPacketResetVOP();
	}
}

// for sprite
/*
Void CVideoObjectEncoder::codeVOPHead (const CSiteD* rgstDest, CRct rctWarp)
{
	// Start code
	static UInt uiNumBitsStart = NUMBITS_START_CODE + NUMBITS_VOP_START;
	m_pbitstrmOut -> putBits (START_CODE, NUMBITS_START_CODE, "VOP_Start_Code");
	m_pbitstrmOut -> putBits ((Int) 0, (UInt) NUMBITS_VOP_START, "VOP_Start_Code");		//plus 8 bits
	m_statsVOP.nBitsHead += uiNumBitsStart;

	// Time reference
#ifndef FRAME_RATE				//assuming input source is always 30f/s
#define FRAME_RATE 30
#endif
	Time tCurrSec = m_t / FRAME_RATE;		//current time in second
	m_nBitsModuloBase = tCurrSec - m_tModuloBase;
	m_pbitstrmOut -> putBits ((Int) 0xFFFFFFFE, (UInt) m_nBitsModuloBase + 1, "VOP_Modulo_Time_Base");
	m_statsVOP.nBitsHead += m_nBitsModuloBase;
	if (m_nBitsModuloBase > 0 && m_vopmd.vopPredType != BVOP)
		m_tModuloBase = tCurrSec;
	Int iVopTimeIncr = m_t * 1000 / FRAME_RATE - tCurrSec * 1000;
	m_pbitstrmOut -> putBits (iVopTimeIncr, NUMBITS_VOP_TIMEINCR, "VOP_Time_Incr");
	m_statsVOP.nBitsHead += NUMBITS_VOP_TIMEINCR;

	m_pbitstrmOut -> putBits (m_vopmd.vopPredType, NUMBITS_VOP_PRED_TYPE, "VOP_Pred_Type");
	m_statsVOP.nBitsHead += NUMBITS_VOP_PRED_TYPE;
	m_pbitstrmOut -> putBits (m_vopmd.iIntraDcSwitchThr, 3, "VOP_DC_VLC_Switch");
	m_statsVOP.nBitsHead += 3;
	if (m_vopmd.vopPredType == SPRITE && m_iNumOfPnts > 0)
		m_statsVOP.nBitsMV += codeWarpPoints (rgstDest, rctWarp);
	if (m_volmd.fAUsage != RECTANGLE) {	
		m_pbitstrmOut -> putBits (m_rctOrgY.width (), NUMBITS_VOP_WIDTH, "VOP_Width");
		m_pbitstrmOut -> putBits ((Int) 1, 1, "Marker"); // marker bit
		m_pbitstrmOut -> putBits (m_rctOrgY.height (), NUMBITS_VOP_HEIGHT, "VOP_Height");
		Int iSign = (m_rctOrgY.left < 0) ? 0 : 1;
		m_pbitstrmOut -> putBits (iSign, 1, "VOP_Hori_Ref_Sgn");
		m_pbitstrmOut -> putBits (abs(m_rctOrgY.left), NUMBITS_VOP_HORIZONTAL_SPA_REF - 1, "VOP_Hori_Ref");
		m_pbitstrmOut -> putBits ((Int) 1, 1, "Marker"); // marker bit
		iSign = (m_rctOrgY.top < 0) ? 0 : 1;
		m_pbitstrmOut -> putBits (iSign, 1, "VOP_Vert_Ref_Sgn");
		m_pbitstrmOut -> putBits (abs(m_rctOrgY.top), NUMBITS_VOP_VERTICAL_SPA_REF - 1, "VOP_Vert_Ref");
		if (m_vopmd.vopPredType == SPRITE)
			return;
		m_pbitstrmOut -> putBits (0, 1, "VOP_CR");		//don't support vop level size conversion
		m_pbitstrmOut -> putBits (m_volmd.bNoCrChange, 1, "VOP_CR_Change_Disable");
		m_pbitstrmOut -> putBits (TRUE, 1, "VOP_SADCT_Disable");
		m_statsVOP.nBitsHead += NUMBITS_VOP_WIDTH + 1 + NUMBITS_VOP_HEIGHT + 
								NUMBITS_VOP_HORIZONTAL_SPA_REF + 1 + NUMBITS_VOP_VERTICAL_SPA_REF + 1 + 1 + 1;
	}
	if (m_vopmd.vopPredType == SPRITE)
		return;	
	if (m_vopmd.vopPredType == BVOP) {
		m_pbitstrmOut -> putBits (m_vopmd.intDBQuant, NUMBITS_VOP_DBQUANT, "VOP_DBQUANT");
		m_statsVOP.nBitsHead += NUMBITS_VOP_DBQUANT;
		if (m_vopmd.intDBQuant == 0)
			m_vopmd.intStepB = (m_vopmd.intStep * 5) >> 2;
		else if (m_vopmd.intDBQuant == 1)
			m_vopmd.intStepB = (m_vopmd.intStep * 6) >> 2;
		else if (m_vopmd.intDBQuant == 2)
			m_vopmd.intStepB = (m_vopmd.intStep * 7) >> 2;
		else if (m_vopmd.intDBQuant == 3)
			m_vopmd.intStepB = (m_vopmd.intStep * 8) >> 2;
		else
			assert (FALSE);
		if (m_volmd.fAUsage == EIGHT_BIT)	{
			if (m_vopmd.intDBQuant == 0)
				m_vopmd.intStepBAlpha = (m_vopmd.intStepAlpha * 5) >> 2;
			else if (m_vopmd.intDBQuant == 1)
				m_vopmd.intStepBAlpha = (m_vopmd.intStepAlpha * 6) >> 2;
			else if (m_vopmd.intDBQuant == 2)
				m_vopmd.intStepBAlpha = (m_vopmd.intStepAlpha * 7) >> 2;
			else if (m_vopmd.intDBQuant == 3)
				m_vopmd.intStepBAlpha = (m_vopmd.intStepAlpha * 8) >> 2;
			else
				assert (FALSE);
		}
	}
	else {
		Int stepToBeCoded = (m_vopmd.vopPredType == PVOP) ? m_vopmd.intStep : m_vopmd.intStepI;
		m_pbitstrmOut -> putBits (stepToBeCoded, NUMBITS_VOP_QUANTIZER, "VOP_Quantier");
		m_statsVOP.nBitsHead += NUMBITS_VOP_QUANTIZER;
		if (m_volmd.fAUsage == EIGHT_BIT)	{
			stepToBeCoded = (m_vopmd.vopPredType == PVOP) ? m_vopmd.intStepAlpha : m_vopmd.intStepIAlpha;
			m_pbitstrmOut -> putBits (stepToBeCoded, NUMBITS_VOP_ALPHA_QUANTIZER, "VOP_Alpha_Quantier");
			m_statsVOP.nBitsHead += NUMBITS_VOP_QUANTIZER;
		}
	}
}
*/
/*
UInt CVideoObjectEncoder::codeWarpPoints (const CSiteD* rgstDest, CRct rctWarp)
{
	UInt nBits = 0;
	assert (m_iNumOfPnts > 0);
	CSiteD rgstSrcQ [4];
	rgstSrcQ [0] = CSiteD (rctWarp.left, rctWarp.top);
	rgstSrcQ [1] = CSiteD (rctWarp.right, rctWarp.top);
	rgstSrcQ [2] = CSiteD (rctWarp.left, rctWarp.bottom);
	rgstSrcQ [3] = CSiteD (rctWarp.right, rctWarp.bottom);

	for (Int i = 0; i < m_iNumOfPnts; i++) {
		// quantization	rgstSrc
		CoordD x = rgstSrcQ [i].x, y = rgstSrcQ [i].y;
		Int rnd = (x > 0) ? (Int) (2.0 * x + .5) : (Int) (2.0 * x - .5);
		m_rgstSrcQ [i].x = (CoordD) rnd / (CoordD) 2.;
		rnd = (y > 0) ? (Int) (2.0 * y + .5) : (Int) (2.0 * y - .5);
		m_rgstSrcQ [i].y = (CoordD) rnd / (CoordD) 2.;
		// quantization	rgstDest
		x = rgstDest [i].x;
		y = rgstDest [i].y;
		rnd = (x > 0) ? (Int) (2.0 * x + .5) : (Int) (2.0 * x - .5);
		m_rgstDstQ [i].x = (CoordD) rnd / (CoordD) 2.;
		rnd = (y > 0) ? (Int) (2.0 * y + .5) : (Int) (2.0 * y - .5);
		m_rgstDstQ [i].y = (CoordD) rnd / (CoordD) 2.;
	}

	// Dest corner points MV encoding
	// construct u, v, du, dv
	Int rgiU [4], rgiV [4], rgiDU [4], rgiDV [4];
	Int j;
	for (j = 0; j < m_iNumOfPnts; j++) { 	//here both m_rgstDstQ and m_rgstSrcQ are in fulpel accuracy
		rgiU [j] = (Int) (2 * m_rgstDstQ [j].x -  2 * (m_rgstSrcQ [j].x - m_rgstSrcQ [0].x));
		rgiV [j] = (Int) (2 * m_rgstDstQ [j].y -  2 * (m_rgstSrcQ [j].y - m_rgstSrcQ [0].y));
	}
	rgiDU [0] = rgiU [0];									rgiDV [0] = rgiV [0];
	rgiDU [1] = rgiU [1] - rgiU [0];						rgiDV [1] = rgiV [1] - rgiV [0];
	rgiDU [2] = rgiU [2] - rgiU [0];						rgiDV [2] = rgiV [2] - rgiV [0];
	rgiDU [3] = rgiU [3] - rgiU [2] - rgiU [1] + rgiU [0];	rgiDV [3] = rgiV [3] - rgiV [2] - rgiV [1] + rgiV [0];
	// du, dv encoding
	Int rgiWrpPnt0Del [2];
	COutBitStream* pobstrmWrpPt0 = m_pEntrencSet->m_pentrencWrpPnt -> bitstream ();
	pobstrmWrpPt0->trace (m_rgstDstQ [0], "SPRT_Warp_Point_Q");
	for (j = 0; j < m_iNumOfPnts; j++) { 
		rgiWrpPnt0Del [0] = rgiDU [j];		//make them half pel units
		rgiWrpPnt0Del [1] = rgiDV [j];		//make them half pel units
		for (UInt iXorY = 0; iXorY < 2; iXorY++)	{
			assert (rgiWrpPnt0Del [iXorY] >= -2047 && rgiWrpPnt0Del [iXorY] <= 2047);
			UInt uiAbsWrpPnt0Del = (UInt) abs (rgiWrpPnt0Del [iXorY]);
			Long lSz = 0;
			for (Int i = 11; i > 0; i--)	{
				if (uiAbsWrpPnt0Del & (1 << (i - 1)))	{
					lSz = i;
					break;
				}
			}
			m_pEntrencSet->m_pentrencWrpPnt->encodeSymbol(lSz, "SPRT_Warp_Point");		// huffman encode 

			if (rgiWrpPnt0Del [iXorY] > 0)
				pobstrmWrpPt0->putBits (rgiWrpPnt0Del [iXorY], lSz, "SPRT_Warp_Point_Sgn");	//fix length code
			else if (rgiWrpPnt0Del[iXorY] < 0)
				pobstrmWrpPt0->putBits (~abs(rgiWrpPnt0Del [iXorY]), lSz, "SPRT_Warp_Point_Sgn");		//fix length code
			nBits += lSz + lSz;
		}
	}
	// Src corner points encoding
	for (j = 0; j < m_iNumOfPnts; j++) { 
		rgiWrpPnt0Del [0] = (Int) ((m_rgstSrcQ [j].x));	//make them single pel units
		rgiWrpPnt0Del [1] = (Int) ((m_rgstSrcQ [j].y));
		for (UInt iXorY = 0; iXorY < 2; iXorY++)	{
			assert (rgiWrpPnt0Del [iXorY] >= -2047 && rgiWrpPnt0Del [iXorY] <= 2047);
			UInt uiAbsWrpPnt0Del = (UInt) abs (rgiWrpPnt0Del [iXorY]);
			Long lSz = 0;
			for (Int i = 11; i > 0; i--)	{
				if (uiAbsWrpPnt0Del & (1 << (i - 1)))	{
					lSz = i;
					break;
				}
			}
			m_pEntrencSet->m_pentrencWrpPnt->encodeSymbol(lSz, "VOP boundingBox");		// huffman encode 

			if (rgiWrpPnt0Del [iXorY] > 0)
				pobstrmWrpPt0->putBits (rgiWrpPnt0Del [iXorY], lSz, "VOP boundingBox");	//fix length code
			else if (rgiWrpPnt0Del[iXorY] < 0)
				pobstrmWrpPt0->putBits (~abs(rgiWrpPnt0Del [iXorY]), lSz, "VOP boundingBox");		//fix length code
			nBits += lSz + lSz;
		}
	}
	return nBits;
}
*/

/*Void CVideoObjectEncoder::decidePredType ()
{	
	static Int iLastPossibleVOP = ((m_nLastFrame - m_nFirstFrame) / m_volmd.iTemporalRate) 
								  * m_volmd.iTemporalRate;	//relative frm no of last vop on the sampling grid
	if (m_uiSprite == 1) {
		if (m_t == -1)
			m_vopmd.vopPredType = IVOP;
		else
			m_vopmd.vopPredType = SPRITE;
	}
//	else if (m_tEncodedVOPCounter % m_volmd.iPeriodOfIVOP == 0) {
	else if ((m_t / m_volmd.iTemporalRate) % m_volmd.iPeriodOfIVOP == 0) {
		m_vopmd.vopPredType = IVOP;
		m_tDistanceBetwIPVOP = (m_t - m_tPrevIorPVOPCounter) / m_volmd.iTemporalRate;
		m_tPrevIorPVOPCounter = m_t;
		m_tEncodedVOPCounter++;
		m_tBCounter = 0;
	}
	else if (m_tEncodedVOPCounter == 1 ||
			(m_tEncodedVOPCounter - 1) % m_volmd.iPeriodOfPVOP == 0 ||
			 m_t == iLastPossibleVOP) {
		m_vopmd.vopPredType = (m_volmd.iPeriodOfIVOP == m_volmd.iPeriodOfPVOP) ? IVOP : PVOP;	//1st case: all I 
		m_tDistanceBetwIPVOP = (m_t - m_tPrevIorPVOPCounter) / m_volmd.iTemporalRate;
		m_tPrevIorPVOPCounter = m_t;
		m_tEncodedVOPCounter++;
		m_tBCounter = 0;
	}
	else {
		m_vopmd.vopPredType = BVOP;
		m_tEncodedVOPCounter++;
		m_tBCounter++;
	}
}*/

Void CVideoObjectEncoder::decideMVInfo ()
{
	assert (m_vopmd.iSearchRangeForward <= 128);		//seems a reasonable constraint
	assert (m_vopmd.iSearchRangeBackward <= 128);		//seems a reasonable constraint
/*	if (m_bSprite == 1) {
		m_volmd.mvInfo.uiRange = 128;
		m_volmd.mvInfo.uiFCode = 3;
	}
	else {
*/
		if (m_vopmd.iSearchRangeForward <= 16) {
			m_vopmd.mvInfoForward.uiRange = 32;
			m_vopmd.mvInfoForward.uiFCode = 1;
			if (m_vopmd.iSearchRangeForward == 16)
				m_vopmd.iSearchRangeForward--; // to avoid out of range for half pel search
		}
		else if (m_vopmd.iSearchRangeForward <= 32) {
			m_vopmd.mvInfoForward.uiRange = 64;
			m_vopmd.mvInfoForward.uiFCode = 2;
			if (m_vopmd.iSearchRangeForward == 32)
				m_vopmd.iSearchRangeForward--; // to avoid out of range for half pel search
		}
		else if (m_vopmd.iSearchRangeForward <= 128) {
			m_vopmd.mvInfoForward.uiRange = 128;
			m_vopmd.mvInfoForward.uiFCode = 3;
			if (m_vopmd.iSearchRangeForward == 64)
				m_vopmd.iSearchRangeForward--; // to avoid out of range for half pel search
		}
		else {
			m_vopmd.mvInfoForward.uiRange = 256;
			m_vopmd.mvInfoForward.uiFCode = 4;
			if (m_vopmd.iSearchRangeForward == 256)
				m_vopmd.iSearchRangeForward--; // to avoid out of range for half pel search
		}
//	}
		m_vopmd.mvInfoForward.uiScaleFactor = 1 << (m_vopmd.mvInfoForward.uiFCode - 1);

		if (m_vopmd.iSearchRangeBackward <= 16) {
			m_vopmd.mvInfoBackward.uiRange = 32;
			m_vopmd.mvInfoBackward.uiFCode = 1;
			if (m_vopmd.iSearchRangeBackward == 16)
				m_vopmd.iSearchRangeBackward--; // to avoid out of range for half pel search
		}
		else if (m_vopmd.iSearchRangeBackward <= 32) {
			m_vopmd.mvInfoBackward.uiRange = 64;
			m_vopmd.mvInfoBackward.uiFCode = 2;
			if (m_vopmd.iSearchRangeBackward == 32)
				m_vopmd.iSearchRangeBackward--; // to avoid out of range for half pel search
		}
		else if (m_vopmd.iSearchRangeBackward <= 128) {
			m_vopmd.mvInfoBackward.uiRange = 128;
			m_vopmd.mvInfoBackward.uiFCode = 3;
			if (m_vopmd.iSearchRangeBackward == 64)
				m_vopmd.iSearchRangeBackward--; // to avoid out of range for half pel search
		}
		else {
			m_vopmd.mvInfoBackward.uiRange = 256;
			m_vopmd.mvInfoBackward.uiFCode = 4;
			if (m_vopmd.iSearchRangeBackward == 256)
				m_vopmd.iSearchRangeBackward--; // to avoid out of range for half pel search
		}
//	}
		m_vopmd.mvInfoBackward.uiScaleFactor = 1 << (m_vopmd.mvInfoBackward.uiFCode - 1);

}

Void CVideoObjectEncoder::updateAllOrigRefVOPs ()
{
	assert (m_vopmd.vopPredType != BVOP);
	swapVOPU8Pointers (m_pvopcOrig, m_pvopcRefOrig1);
	m_pvopcRefOrig1->setBoundRct (m_rctRefVOPY1);
	m_pvopcOrig->setBoundRct (m_rctCurrVOPY);
	repeatPadYOrA ((PixelC*) m_pvopcRefOrig1->pixelsY () + m_iOffsetForPadY, m_pvopcRefOrig1);
}

Void CVideoObjectEncoder::copyCurrToRefOrig1Y ()
{
	m_pvopcRefOrig0->setAndExpandBoundRctOnly (m_pvopcOrig->whereBoundY (), EXPANDY_REFVOP); // will adjust the starting pixel pointer too
	const PixelC* ppxlcOrigY = m_pvopcOrig->pixelsBoundY ();
	PixelC* ppxlcRefOrig0Y = (PixelC*) m_pvopcRefOrig0->pixelsBoundY ();
	for (CoordI y = 0; y < m_pvopcOrig->whereBoundY ().height (); y++) {
		memcpy (ppxlcRefOrig0Y, ppxlcOrigY, m_pvopcOrig->whereBoundY ().width);
		ppxlcOrigY += m_pvopcOrig->whereBoundY ().width;
		ppxlcRefOrig0Y += m_pvopcRefOrig0->whereBoundY ().width;
	}
}
/*
Void CVideoObjectEncoder::scaleMV (
	Int& iVLC, 
	UInt& uiResidual, 
	Int iDiffMVcomponent)
{
	assert (iDiffMVcomponent >= (-3 * (Int) m_volmd.mvInfo.uiRange));
	assert (iDiffMVcomponent < 3 * (Int) m_volmd.mvInfo.uiRange);
	if (iDiffMVcomponent < (-1 * (Int) m_volmd.mvInfo.uiRange))
		iDiffMVcomponent += 2 * m_volmd.mvInfo.uiRange;
	else if (iDiffMVcomponent > (Int) (m_volmd.mvInfo.uiRange - 1))
		iDiffMVcomponent -= 2 * m_volmd.mvInfo.uiRange;

	if (iDiffMVcomponent == 0)	{									//nothing to do
		iVLC = 0;
		uiResidual = 0;
	}
	else	{
		if (m_volmd.mvInfo.uiScaleFactor == 1)	{									//simple no-scale case
			iVLC = iDiffMVcomponent;
			uiResidual = 0;
		}
		else {														//stupid scaling
			UInt uiAbsDiffMVcomponent = abs (iDiffMVcomponent);
			uiResidual = (uiAbsDiffMVcomponent - 1) % m_volmd.mvInfo.uiScaleFactor;																	
			UInt uiVLCmagnitude = (uiAbsDiffMVcomponent 
								- uiResidual + (m_volmd.mvInfo.uiScaleFactor - 1)) / m_volmd.mvInfo.uiScaleFactor;	// absorb ++ into here (m_volmd.mvInfo.uiScaleFactor - 1)
			iVLC = uiVLCmagnitude * sign (iDiffMVcomponent);
		}
	}
}

UInt CVideoObjectEncoder::sendDiffMV (const CVector2D& vctDiffMV)
{
	Int iVLC;
	UInt uiResidual;
	UInt nBits = 0;
	Long lsymbol;

	// send the bits
	Int iDiffMvX = (Int) (vctDiffMV.x * 2.0);
	if (iDiffMvX <  -1 * (Int) m_volmd.mvInfo.uiRange	||
		iDiffMvX > (Int) (m_volmd.mvInfo.uiRange - 1))	{
		m_pbitstrmOut->putBits ((Int) SPRITE_MV_ESC, (UInt) NUMBITS_SPRITE_MV_ESC, "SPRT_Esc");
		m_pbitstrmOut->putBits (signOf (iDiffMvX), (UInt) 1, "SPRT_Esc_Mv_Sgn");
		m_pbitstrmOut->putBits (abs (iDiffMvX), (UInt) (NUMBITS_SPRITE_MV_ESC - 1), "SPRT_Esc_Mv");
		nBits += NUMBITS_SPRITE_MV_ESC + NUMBITS_SPRITE_MV_ESC;
	}
	else	{
		scaleMV (iVLC, uiResidual, iDiffMvX);
		lsymbol = iVLC + 32;
		nBits += m_pEntrencSet->m_pentrencMV->encodeSymbol (lsymbol, "SPRT_MV_Magnitude");
		if (iVLC != 0) {
			m_pbitstrmOut->putBits ((Char) uiResidual, m_volmd.mvInfo.uiFCode - 1, "SPRT_MV_Residual");
			nBits += m_volmd.mvInfo.uiFCode - 1;
		}
		else
			assert (uiResidual == 0);
	}
	Int iDiffMvY = (Int) (vctDiffMV.y * 2.0);
	if (iDiffMvY <  -1 * (Int) m_volmd.mvInfo.uiRange	||
		iDiffMvY > (Int) (m_volmd.mvInfo.uiRange - 1))	{
		m_pbitstrmOut->putBits ((Int) SPRITE_MV_ESC, (UInt) NUMBITS_SPRITE_MV_ESC, "SPRT_Esc");
		m_pbitstrmOut->putBits (signOf (iDiffMvY), (UInt) 1, "SPRT_Esc_Mv_Sgn");
		m_pbitstrmOut->putBits (abs (iDiffMvY), (UInt) (NUMBITS_SPRITE_MV_ESC - 1), "SPRT_Esc_Mv");
		nBits += NUMBITS_SPRITE_MV_ESC + NUMBITS_SPRITE_MV_ESC;	 
	}
	else	{
		scaleMV (iVLC, uiResidual, iDiffMvY);
		lsymbol = iVLC + 32;
		nBits += m_pEntrencSet->m_pentrencMV->encodeSymbol (lsymbol, "SPRT_MV_Magnitude");
		if (iVLC != 0) {
			m_pbitstrmOut->putBits ((Char) uiResidual, m_volmd.mvInfo.uiFCode - 1, "SPRT_MV_Residual");
			nBits += m_volmd.mvInfo.uiFCode - 1;
		}
		else
			assert (uiResidual == 0);
	}
	return nBits;
}
*/

Void CVideoObjectEncoder::findTightBoundingBox ()
{
	const CRct& rctOrig = m_pvopcOrig->whereY ();
	CoordI left = rctOrig.right - 1;
	CoordI top = rctOrig.bottom - 1;
	CoordI right = rctOrig.left;
	CoordI bottom = rctOrig.top;
	const PixelC* ppxlcBY = m_pvopcOrig->pixelsBY ();
	for (CoordI y = rctOrig.top; y < rctOrig.bottom; y++) {
		for (CoordI x = rctOrig.left; x < rctOrig.right; x++) {
			if (*ppxlcBY != transpValue) {
				left = min (left, x);
				top = min (top, y);
				right = max (right, x);
				bottom = max (bottom, y);
			}								
			ppxlcBY++;
		}
	}
	right++;
	bottom++;
	if (left % 2 != 0)
		left--;
	if (top % 2 != 0)
		top--;
/*
	Int iMod = (right - left) % MB_SIZE;
	if (iMod > 0)
		right += MB_SIZE - iMod;
	iMod = (bottom - top) % MB_SIZE;
	if (iMod > 0)
		bottom += MB_SIZE - iMod;
*/
	m_pvopcOrig->setBoundRct (CRct (left, top, right, bottom));
}

Void CVideoObjectEncoder::findBestBoundingBox ()
{
	CoordI iLeftCurr   = m_pvopcOrig->whereBoundY ().left;
	CoordI iTopCurr    = m_pvopcOrig->whereBoundY ().top;
	CoordI iRightCurr  = m_pvopcOrig->whereBoundY ().right;
	CoordI iBottomCurr = m_pvopcOrig->whereBoundY ().bottom;
	CoordI iLeftMost   = max (iLeftCurr - MB_SIZE + 2, 0);
	CoordI iTopMost    = max (iTopCurr  - MB_SIZE + 2, 0);			//can't be negative

	Int nNonTransparentMBMin = m_pvopcOrig->whereBoundY ().area ();		//abitrary large number
	Int iLeftOfBBox, iTopOfBBox;										//control postions
	Int iLeftBest = iLeftMost;
	Int iTopBest  = iTopMost;
	for (iLeftOfBBox = iLeftMost; iLeftOfBBox <= iLeftCurr; iLeftOfBBox += 2)	{
		for (iTopOfBBox = iTopMost; iTopOfBBox <= iTopCurr; iTopOfBBox += 2)	{
			// counting for each control position
			const PixelC* ppxlcBYMBRow = m_pvopcOrig->getPlane(BY_PLANE)->pixels (iLeftOfBBox, iTopOfBBox);
			Int nNonTransparentMB = 0;
			Bool bKeepCounting = TRUE;
			Int iLeftOfMB, iTopOfMB;
			for (iTopOfMB = iTopOfBBox; iTopOfMB < iBottomCurr; iTopOfMB += MB_SIZE) {
				const PixelC* ppxlcBYMB = ppxlcBYMBRow;
				for (iLeftOfMB = iLeftOfBBox; iLeftOfMB < iRightCurr; iLeftOfMB += MB_SIZE) {
					const PixelC* ppxlcBY = ppxlcBYMB;
					UInt nOpaquePixel = 0;
					for (UInt iY = 0; iY < MB_SIZE; iY++)	{
						for (UInt iX = 0; iX < MB_SIZE; iX++)	{
							nOpaquePixel += *ppxlcBY++;
						}
						ppxlcBY += m_iFrameWidthY - MB_SIZE;
					}
					if (nOpaquePixel != 0)
						nNonTransparentMB++;
					if (nNonTransparentMB > nNonTransparentMBMin)	{
						bKeepCounting = FALSE;
						break;
					}
					ppxlcBYMB += MB_SIZE;
				}
				if (!bKeepCounting)
					break;
				ppxlcBYMBRow += m_iFrameWidthYxMBSize;
			}
			if (nNonTransparentMB <= nNonTransparentMBMin)	{
				nNonTransparentMBMin = nNonTransparentMB;
				iLeftBest = iLeftOfBBox;
				iTopBest = iTopOfBBox;
			}
		}
	}
	m_pvopcOrig->setBoundRct (CRct (
		iLeftBest, 
		iTopBest, 
		iLeftBest + ((iRightCurr - iLeftBest - 1) / MB_SIZE + 1) * MB_SIZE, 
		iTopBest  + ((iBottomCurr - iTopBest - 1) / MB_SIZE + 1) * MB_SIZE
	));
}

Void CVideoObjectEncoder::computeSNRs (const CVOPU8YUVBA* pvopcCurrQ)
{
	if (m_volmd.fAUsage == RECTANGLE) {
		SNRYorA (m_pvopcOrig->pixelsBoundY (), pvopcCurrQ->pixelsY () + m_iStartInRefToCurrRctY, m_rgdSNR [0]);
		SNRUV (pvopcCurrQ);
	}
	else {
		SNRYorAWithShape (m_pvopcOrig->pixelsBoundY (), pvopcCurrQ->pixelsY () + m_iStartInRefToCurrRctY, m_rgdSNR [0]);
		SNRUVWithShape (pvopcCurrQ);
		if (m_volmd.fAUsage == EIGHT_BIT)
			SNRYorAWithShape (m_pvopcOrig->pixelsBoundA (), pvopcCurrQ->pixelsA () + m_iStartInRefToCurrRctY, m_rgdSNR [3]);
	}
}

Void CVideoObjectEncoder::SNRYorA (
	const PixelC* ppxlcOrig, const PixelC* ppxlcRef1, // these two should point to the left-top of the rctOrig's bounding box 
	Double& dSNR
)
{
	CoordI x, y;
	Int iSqrDiff = 0;
	for (y = 0; y < m_pvopcOrig->whereBoundY ().height (); y++) {
		for (x = 0; x < m_pvopcOrig->whereBoundY ().width; x++) {
			Int iDiff = ppxlcOrig [x] - ppxlcRef1 [x];
			iSqrDiff += iDiff * iDiff;
		}
		ppxlcOrig += m_pvopcOrig->whereY ().width;
		ppxlcRef1 += m_pvopcRefQ1->whereY ().width;
	}
	if (iSqrDiff == 0) 
		dSNR = 1000000.0;
	else
		dSNR = (log10 ((255.0 * 255.0 * m_pvopcOrig->whereBoundY ().area ()) / (Double) iSqrDiff) * 10.0);
}

Void CVideoObjectEncoder::SNRUV (const CVOPU8YUVBA* pvopcCurrQ)
{
	CoordI x, y;
	Int iSqrDiffU = 0, iSqrDiffV = 0, iDiff;
	const PixelC* ppxlcOrigU = m_pvopcOrig->pixelsBoundU (); 
	const PixelC* ppxlcOrigV = m_pvopcOrig->pixelsBoundV (); 
	const PixelC* ppxlcRef1U = pvopcCurrQ->pixelsU () + m_iStartInRefToCurrRctUV;
	const PixelC* ppxlcRef1V = pvopcCurrQ->pixelsV () + m_iStartInRefToCurrRctUV;
	for (y = 0; y < m_pvopcOrig->whereBoundUV ().height (); y++) {
		for (x = 0; x < m_pvopcOrig->whereBoundUV ().width; x++) {
			iDiff = ppxlcOrigU [x] - ppxlcRef1U [x];
			iSqrDiffU += iDiff * iDiff;
			iDiff = ppxlcOrigV [x] - ppxlcRef1V [x];
			iSqrDiffV += iDiff * iDiff;
		}
		ppxlcOrigU += m_pvopcOrig->whereUV ().width;
		ppxlcRef1U += m_pvopcRefQ1->whereUV ().width;
		
		ppxlcOrigV += m_pvopcOrig->whereUV ().width;
		ppxlcRef1V += m_pvopcRefQ1->whereUV ().width;
	}
	if (iSqrDiffU == 0) 
		m_rgdSNR [1] = 1000000.0;
	else
		m_rgdSNR [1] = (log10 ((Double) (255.0 * 255.0 * m_pvopcOrig->whereBoundUV ().area ()) / (Double) iSqrDiffU) * 10.0);

	if (iSqrDiffV == 0) 
		m_rgdSNR [2] = 1000000.0;
	else
		m_rgdSNR [2] = (log10 ((Double) (255.0 * 255.0 * m_pvopcOrig->whereBoundUV ().area ()) / (Double) iSqrDiffV) * 10.0);
}

Void CVideoObjectEncoder::SNRYorAWithShape (
	const PixelC* ppxlcOrig, const PixelC* ppxlcRef1, // these two should point to the left-top of the rctOrig's bounding box 
	Double& dSNR
)
{
	CoordI x, y;
	Int iSqrDiff = 0, iNumNonTransp = 0;
	const PixelC* ppxlcOrigBY = m_pvopcOrig->pixelsBoundBY ();
	for (y = 0; y < m_pvopcOrig->whereBoundY ().height (); y++) {
		for (x = 0; x < m_pvopcOrig->whereBoundY ().width; x++) {
			if (ppxlcOrigBY [x] != transpValue) {
				Int iDiff = ppxlcOrig [x] - ppxlcRef1 [x];
				iSqrDiff += iDiff * iDiff;
				iNumNonTransp++;
			}
		}
		ppxlcOrig += m_pvopcOrig->whereY ().width;
		ppxlcOrigBY += m_pvopcOrig->whereY ().width;
		ppxlcRef1 += m_pvopcRefQ1->whereY ().width;
	}
	if (iSqrDiff == 0) 
		dSNR = 1000000.0;
	else
		dSNR = (log10 ((Double) (255.0 * 255.0 * iNumNonTransp) / (Double) iSqrDiff) * 10.0);
}

Void CVideoObjectEncoder::SNRUVWithShape (const CVOPU8YUVBA* pvopcCurrQ)
{
	CoordI x, y;
	Int iSqrDiffU = 0, iSqrDiffV = 0, iNumNonTransp = 0, iDiff;
	const PixelC* ppxlcOrigU = m_pvopcOrig->pixelsBoundU (); 
	const PixelC* ppxlcOrigV = m_pvopcOrig->pixelsBoundV (); 
	const PixelC* ppxlcRef1U = pvopcCurrQ->pixelsU () + m_iStartInRefToCurrRctUV;
	const PixelC* ppxlcRef1V = pvopcCurrQ->pixelsV () + m_iStartInRefToCurrRctUV;
	const PixelC* ppxlcOrigBUV = m_pvopcOrig->pixelsBoundBUV ();
	for (y = 0; y < m_pvopcOrig->whereBoundUV ().height (); y++) {
		for (x = 0; x < m_pvopcOrig->whereBoundUV ().width; x++) {
			if (ppxlcOrigBUV [x] != transpValue) {
				iDiff = ppxlcOrigU [x] - ppxlcRef1U [x];
				iSqrDiffU += iDiff * iDiff;
				iDiff = ppxlcOrigV [x] - ppxlcRef1V [x];
				iSqrDiffV += iDiff * iDiff;
				iNumNonTransp++;
			}
		}
		ppxlcOrigBUV += m_pvopcOrig->whereUV ().width;
		ppxlcOrigU += m_pvopcOrig->whereUV ().width;
		ppxlcRef1U += m_pvopcRefQ1->whereUV ().width;
		
		ppxlcOrigV += m_pvopcOrig->whereUV ().width;
		ppxlcRef1V += m_pvopcRefQ1->whereUV ().width;
	}
	if (iSqrDiffU == 0) 
		m_rgdSNR [1] = 1000000.0;
	else
		m_rgdSNR [1] = (log10 ((Double) (255.0 * 255.0 * iNumNonTransp) / (Double) iSqrDiffU) * 10.0);

	if (iSqrDiffV == 0) 
		m_rgdSNR [2] = 1000000.0;
	else
		m_rgdSNR [2] = (log10 ((Double) (255.0 * 255.0 * iNumNonTransp) / (Double) iSqrDiffV) * 10.0);
}

Void CVideoObjectEncoder::biInterpolateY (
	const CVOPU8YUVBA* pvopcRefQ, const CRct& rctRefVOP,
	CU8Image* puciRefQZoom, const CRct& rctRefVOPZoom, Int iRoundingControl
)
{
	const PixelC* ppxliOrg = pvopcRefQ->pixelsBoundY ();
	const PixelC* ppxliOrgBot;
	PixelC* ppxliZoom = (PixelC*) puciRefQZoom->pixels (rctRefVOPZoom.left, rctRefVOPZoom.top);
	PixelC* ppxliZoomBot;
	ppxliOrgBot = ppxliOrg + m_iFrameWidthY;
	ppxliZoomBot = ppxliZoom + m_iFrameWidthZoomY;
	Int iHeightPrevMinus1 = rctRefVOP.height () - 1;
	Int iWidthPrevMinus1 = rctRefVOP.width - 1;
	CoordI x, y;
	for (y = 0; y < iHeightPrevMinus1; y++) {
		const PixelC* ppxliOrgRow = ppxliOrg;
		const PixelC* ppxliOrgBotRow = ppxliOrgBot;
		PixelC* ppxliZoomRow = ppxliZoom;
		PixelC* ppxliZoomBotRow = ppxliZoomBot;
		for (x = 0; x < iWidthPrevMinus1; x++, ppxliZoomRow += 2, ppxliZoomBotRow += 2) {
			*ppxliZoomRow = *ppxliOrgRow;
//			*(ppxliZoomRow + 1) = (*ppxliOrgRow + *(ppxliOrgRow + 1) + 1) >> 1;
//			*ppxliZoomBotRow = (*ppxliOrgRow + *ppxliOrgBotRow + 1) >> 1;
//			*(ppxliZoomBotRow + 1) = 
//				(*ppxliOrgRow + *ppxliOrgBotRow + 
//				*(ppxliOrgRow++ + 1) + *(ppxliOrgBotRow++ + 1) + 2
//			) >> 2;
			*(ppxliZoomRow + 1) = (*ppxliOrgRow + *(ppxliOrgRow + 1) + 1 - iRoundingControl) >> 1;
			*ppxliZoomBotRow = (*ppxliOrgRow + *ppxliOrgBotRow + 1 - iRoundingControl) >> 1;
			*(ppxliZoomBotRow + 1) = 
				(*ppxliOrgRow + *ppxliOrgBotRow + 
				*(ppxliOrgRow++ + 1) + *(ppxliOrgBotRow++ + 1) + 2 - iRoundingControl
			) >> 2;

		}
		*ppxliZoomRow = *(ppxliZoomRow + 1) = *ppxliOrgRow;
		// it is better to use / 2 instead of >> 1 only at this place; Don't know why. Please don't rm it.
//		*ppxliZoomBotRow = *(ppxliZoomBotRow + 1) = (*ppxliOrgRow++ + *ppxliOrgBotRow++ + 1) / 2; // the last pixels of every row do not need average
		*ppxliZoomBotRow = *(ppxliZoomBotRow + 1) = (*ppxliOrgRow++ + *ppxliOrgBotRow++ + 1 - iRoundingControl) / 2; // the last pixels of every row do not need average
		
		ppxliOrg += m_iFrameWidthY;
		ppxliOrgBot += m_iFrameWidthY;
		ppxliZoom += m_iFrameWidthZoomY * 2;
		ppxliZoomBot += m_iFrameWidthZoomY * 2;
	}
	PixelC* ppxliZoomLastLnToSec = ppxliZoom;
	for (x = 0; x < iWidthPrevMinus1; x++, ppxliZoomLastLnToSec += 2) {
		*ppxliZoomLastLnToSec = *ppxliOrg;
		*(ppxliZoomLastLnToSec + 1) = (*ppxliOrg + *(ppxliOrg++ + 1) + 1) >> 1;
	}
	*ppxliZoomLastLnToSec = *(ppxliZoomLastLnToSec + 1) = *ppxliOrg;
	memcpy (ppxliZoomBot, ppxliZoom, rctRefVOPZoom.width);
}
