
/******************************************************************************

This software module was originally developed by
	Stefan HORBELT (horbelt@desun1.epfl.ch) LTS-EPFL
	Frederic JORDAN (jordan@ltssg3.epfl.ch) LTS-EPFL
        Wei Wu (weiwu@stallion.risc.rockwell.com),  Rockwell Science Center
in the course of development of the MPEG-4 Video (ISO/IEC 14496-2).
This software module is an implementation of a part of one or more MPEG-4 Video
tools as specified by the MPEG-4 Video. ISO/IEC gives users of the MPEG-4 Video
free license to this software module or modifications thereof for use in
hardware or software products claiming conformance to the MPEG-4 Video. Those
intending to use this software module in hardware or software products are advised
that its use may infringe existing patents. The original developer of this software
module and his/her company, the subsequent editors and their companies, and ISO/IEC
have no liability for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Video conforming
products. LTS-EPFL and Rockwell Science Center retains full right to use the 
code for his/her own
purpose, assign or donate the code to a third party and to inhibit third parties from
using the code for non <MPEG standard> conforming products. This copyright notice
must be included in all copies or derivative works. Copyright (c) 1997.

******************************************************************************/

#include "vd_vector.h"
#include "vd_projector.h"
#include <fstream.h>
#include <iostream.h>
#include <stdlib.h>
#include <iomanip.h>
#include <math.h>

#define M_PI            3.14159265358979323846

/************************************************************************/
/* ACTION:																*/
/*	Creates a projector object.											*/
/* INPUT:																*/
/*	p: Viewer position													*/
/*	d: Aim point position												*/
/* OUTPUT:																*/
/*	vpos: Viewer position												*/
/*	vdir: Aim point position											*/
/*	cos: Projection area of a given	face								*/
/*	dist: Distance to a given face										*/	
/*  beta: Rotation angle of a given face								*/
/************************************************************************/	
projector::projector (vector p, vector d)
{
	vpos.x = p.x;
	vpos.y = p.y;
	vpos.z = p.z;
	
	vdir.x = d.x;
	vdir.y = d.y;
	vdir.z = d.z;
	alpha = 0;
	dist = 1;
	beta = 0;
}


/************************************************************************/
/* ACTION:																*/
/*	Returns the projections values of the triangle face, taking account */
/*  relative position of the viewer.									*/
/*  This is computed on the basis of an orthogonal projection.			*/
/* INPUT:																*/
/*	v0, v1, v2: Coords of the 3 vertices defining the triangle			*/
/*	[center]:   optional center of viewing                              */
/* OUTPUT:																*/
/*	cos: Cosinus of the angle between triangle plane and the observer	*/
/*	beta: Rotation angle of the mesh face								*/
/************************************************************************/	
void	projector::triangle (vector v0, vector v1, vector v2)
{ 
    // standard is orthogonal projection: center=vdir;
	triangle (v0, v1, v2, vdir); 
}

void	projector::triangle (vector v0, vector v1, vector v2,  vector center)
{
	vector v01;
	vector v02;
	vector vv;
	vector n,  n2;
	
	vector u;
	double K;
	double cos_beta;
	
	/* Computes the 2 vectors defining the edges of the triangle	*/
	v01 = vec (v0, v1);
	v02 = vec (v0, v2);
			
	/* Computes the vector normal to the triangle					*/
	n = vect_prod (v01, v02);
	
	/* Computes the vector between viewer position and aim point	*/
	vv = vec (vpos, center);
	
	/* Computes the cosinus of the angle between trangle plane and	*/
	/* viewer vector												*/
	/* This value gives an evalutation of projected area of triangle*/
	/* If = 1, then plane is normal to viewer direction and we		*/
	/* have to transmit full texture information. If = 0, it is		*/
	/* parallel and we need to transmit nothing in the viewer		*/
	/* direction.													*/
	alpha = scal_prod (n, vv)/(norm(n)*norm(vv));
	if(alpha<0) {
		alpha=0;
		beta=0;
		return;
	}
	
	n2=vector(n.x/norm(n), n.y/norm(n), n.z/norm(n));
	n=n2;
/* ADD by weiwu: Computes the projection of viewer vector in (n, v01) plan */
        n2 = vect_prod (n, v01);
        K = norm(n2);
        n2 = vector(n2.x/K, n2.y/K, n2.z/K);
        K = scal_prod(n2, vv);
        u.x = vv.x - K*n2.x;
        u.y = vv.y - K*n2.y;
        u.z = vv.z - K*n2.z;
        alpha = scal_prod(u, n)/norm(u);
/* ADD by weiwu: Computes the projection of viewer vector in (n, v02) plan */
        n2 = vect_prod (n, v02);
        K = norm(n2);
        n2 = vector(n2.x/K, n2.y/K, n2.z/K);
        K = scal_prod(n2, vv);
        u.x = vv.x - K*n2.x;
        u.y = vv.y - K*n2.y;
        u.z = vv.z - K*n2.z;
        beta = scal_prod(u, n)/norm(u);

	/* Computes the projection of viewer vector in triangle plane	*/
/* ADD
	K = scal_prod (n, vv)/norm(vv)/norm(n);
	u.x = vv.x - K*n.x*norm(vv);	
	u.y = vv.y - K*n.y*norm(vv);	
	u.z = vv.z - K*n.z*norm(vv);
	
	cos_beta = scal_prod (v01, u)/(norm(v01)*norm(u));
	beta = 180*acos(cos_beta)/M_PI;
	beta = scal_prod (v01, u)/(norm(v01)*norm(u));
*/
}



/************************************************************************/
/* ACTION:																*/
/*	Returns the max projection value of the 2 triangles defining a quad.*/
/*  Also computes the distance between viewer and quad center			*/
/* INPUT:																*/
/*	v0, v1, v2, v3: Coords of the 3 vertices defining the triangle		*/
/*  pers==1: Optional flag for Perspective Mode                         */
/* OUTPUT:																*/
/*	cos: Cosinus of the angle between triangle plane and the observer	*/
/*	dist: Distance between viewer and center of quad					*/
/************************************************************************/
void	projector::quad (vector v0, vector v1, vector v2, vector v3)
{
	quad (v0, v1, v2, v3, 0);
}
	
void	projector::quad (vector v0, vector v1, vector v2, vector v3,  int pers)
{	
	double		p1, p2;	
	double		p3, p4;
	vector	c, v;

	/* Computes the quad center	position								*/
	c.x = (v0.x + v1.x + v2.x + v3.x)/4;
	c.y = (v0.y + v1.y + v2.y + v3.y)/4;
	c.z = (v0.z + v1.z + v2.z + v3.z)/4;
	
	/* Computes the distance between observer and quad center			*/
	v = vec (c, vpos);
/* Masked by Wei Wu
	dist = norm (v);
*/

    if (pers!=1) // fallback to orthogonal
		c=vdir;
		
	/* Computes the projection cosinus for the first triangle of quad	*/
	triangle (v0, v2, v1, c);
	p1 = alpha;
	p3 = beta;

	triangle (v3, v1, v2, c);
	p2 = alpha;
	p4 = beta;

	/* We take the highest value in order to optimize quality			*/			
	if (p1>p2)
		alpha = p1;
	else
		alpha = p2;
	
	if (alpha<0) alpha=0;	

	if (p3>p4)
		beta = p3;
	else
		beta = p4;
	
	if (beta<0) beta=0;	
}

/************************************************************************/
/* ACTION:																*/
/*	Returns the angle between observer-vertex and observer-aim vectors	*/
/*  Used to know if a vertex is in the field of view or not				*/
/* INPUT:																*/
/*	c: Vertex position													*/
/* OUTPUT:																*/
/*	clip_angle():	angle in degrees									*/
/* LIMITATION: This routine works only correct for squared windows,     */
/*             as it cuts with a squared field of view                  */
/* TODO: Drop this limitation											*/
/************************************************************************/	
double	projector::clip_angle (vector c)
{
	vector	v, r, p, u, n;
	double	h_angle, v_angle, angle, k;
	
	/* Computes observer-vertex vector									*/
	v = vec (vpos, c);
	
	/* Computes observer-aim vector										*/
	r = vec (vpos, vdir);

	/* This is the vector in horizontal viewing plane					*/
	p.x = 1; p.y = 0; p.z = 0;
	
	/* Computes the vector normal to horizontal viewing plane			*/
	n = vect_prod (p, r);

	/* Computes the projection of v on this plane						*/
	k = scal_prod (v, n)/scal_prod (n, n);
	u.x = v.x - k*n.x;
	u.y = v.y - k*n.y;
	u.z = v.z - k*n.z;

	/* Computes the angle in horizontal plane of observer-quad vector	*/
	h_angle = acos (scal_prod (u, r)/(norm(u)*norm(r)));
	
	/* This is the vector in vertical viewing plane						*/
	p.x = 0; p.y = 0; p.z = 1;
	
	/* Computes the vector normal to vertical viewing plane				*/
	n = vect_prod (r, p);
	
	/* Computes the projection of v on this plane						*/
	k = scal_prod (v, n)/scal_prod (n, n);
	u.x = v.x - k*n.x;
	u.y = v.y - k*n.y;
	u.z = v.z - k*n.z;		
	
	/* Computes the angle in vertical plane of observer-quad vector	*/
	v_angle = acos (scal_prod (u, r)/(norm(u)*norm(r)));

	if (h_angle>v_angle)
		angle = h_angle*180/M_PI;
	else
		angle = v_angle*180/M_PI;
				
	return (angle);
}

double projector::plane_angle(vector p,  vector c)
{
        double angle, k;
        vector v, r, n, u;

        /* Computes observer-vertex vector
                                */
        v = vec (vpos, c);

        /* Computes observer-aim vector
                                */
        r = vec (vpos, vdir);

        /* Computes the vector normal to vertical viewing plane
        */
        n = vect_prod (r, p);
        
        /* Computes the projection of v on this plane
                */
        k = scal_prod (v, n)/scal_prod (n, n);
        u.x = v.x - k*n.x;
        u.y = v.y - k*n.y;
        u.z = v.z - k*n.z;
        
        /* Computes the angle in p-plane of observer-quad vector        */
        angle = acos (scal_prod (u, r)/(norm(u)*norm(r)));
        
        return angle;
}

int     projector::in_fov (vector c1,  vector c2,  vector c3,  vector c4, double fov, double w_over_h,  vector vup)
{
        vector p, r ;
        double v_angle=fov /2.0 /180.0*M_PI;
        double h_angle=atan( w_over_h * tan(v_angle));
 
        /* Computes observer-aim vector
                                */
        r = vec (vpos, vdir);
 
        /* This is the vector in vertical viewing plane
                */
        //p.x = 0; p.y = 0; p.z = 1;
 
        if(
                plane_angle(vup, c1) > v_angle &&
                plane_angle(vup, c2) > v_angle &&
                plane_angle(vup, c3) > v_angle &&
                plane_angle(vup, c4) > v_angle
          )
          return 0; // not in field of view
 
         
        /* This is the vector in horizontal viewing plane
                */
        //p.x = 1; p.y = 0; p.z = 0;
        p=vect_prod (vup, r);   // vup x r
 
        if(
                plane_angle(p, c1) > h_angle &&
                plane_angle(p, c2) > h_angle &&
                plane_angle(p, c3) > h_angle &&
                plane_angle(p, c4) > h_angle
          )
          return 0; // not in field of view
 
        return 1; // in field of view
}

/************************************************************************/
/* ACTION:																*/
/*	Returns vector joining vertex v0 to v1								*/
/* INPUT:																*/
/*	v0, v1: Coords of the 2 vertices defining the vector				*/
/* OUTPUT:																*/
/*	vec: Coords of vector v0v1											*/
/************************************************************************/	
vector	projector::vec (vector v0, vector v1)
{
	vector v;
	
	v.x = v1.x - v0.x;
	v.y = v1.y - v0.y;
	v.z = v1.z - v0.z;
	
	return (v);
}

/************************************************************************/
/* ACTION:																*/
/*	Returns vectorial product											*/
/* INPUT:																*/
/*	v0, v1: Coords of 2 vectors											*/
/* OUTPUT:																*/
/*	vec_prod(): Coords of vector v=v0^v1									*/
/************************************************************************/	
vector	projector::vect_prod (vector v0, vector v1)
{
	vector v;
	
	v.x =   v0.y*v1.z - v0.z*v1.y;
	v.y = - v0.x*v1.z + v0.z*v1.x;
	v.z =   v0.x*v1.y - v0.y*v1.x;
	
	return (v);	
}

/************************************************************************/
/* ACTION:																*/
/*	Returns Scalar product												*/
/* INPUT:																*/
/*	v0, v1: Coords of 2 vectors											*/
/* OUTPUT:																*/
/*	scal_prod(): Scalar product of vector v0 and v1						*/
/************************************************************************/	
c_type	projector::scal_prod (vector v0, vector v1)
{
	c_type r=0;
	
	r = v0.x*v1.x + v0.y*v1.y + v0.z*v1.z;
		
	return (r);
}

/************************************************************************/
/* ACTION:																*/
/*	Returns norm of vector												*/
/* INPUT:																*/
/*	v: Vector															*/
/* OUTPUT:																*/
/*	norm(): Norm of vector												*/
/************************************************************************/	
c_type	projector::norm (vector v)
{
	c_type r;

	r = scal_prod (v, v);

	return (sqrt(r));
}


double	projector::alpha_val (void)
{
	return (alpha);
}

double	projector::dist_val (void)
{
	return (dist);
}

double	projector::beta_val (void)
{
	return (beta);
}

double projector::cal_dist (vector v0, vector v1, vector v2, vector v3)
{
	vector c, v;

	c.x = (v0.x + v1.x + v2.x + v3.x)/4;
        c.y = (v0.y + v1.y + v2.y + v3.y)/4;
        c.z = (v0.z + v1.z + v2.z + v3.z)/4;

	v = vec (c, vpos);
	dist = norm (v);
	
	return dist;
}
